Type: Package
Package: dbplyr
Title: A 'dplyr' Back End for Databases
Version: 2.6.0
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre")),
    person("Maximilian", "Girlich", role = "aut"),
    person("Edgar", "Ruiz", role = "aut"),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: A 'dplyr' back end for databases that allows you to work with
    remote database tables as if they are in-memory data frames.  Basic
    features work with any database that has a 'DBI' back end; more
    advanced features require 'SQL' translation to be provided by the
    package author.
License: MIT + file LICENSE
URL: https://dbplyr.tidyverse.org/, https://github.com/tidyverse/dbplyr
BugReports: https://github.com/tidyverse/dbplyr/issues
Depends: R (>= 4.1)
Imports: blob (>= 1.2.0), cli (>= 3.6.1), DBI (>= 1.1.3), dplyr (>=
        1.1.2), glue (>= 1.6.2), lifecycle (>= 1.0.3), magrittr,
        methods, pillar (>= 1.9.0), purrr (>= 1.0.1), R6 (>= 2.2.2),
        rlang (>= 1.1.1), tibble (>= 3.2.1), tidyr (>= 1.3.0),
        tidyselect (>= 1.2.1), utils, vctrs (>= 0.6.3), withr (>=
        2.5.0)
Suggests: adbcdrivermanager, adbcsqlite, adbi, bit64, covr, knitr,
        Lahman, nycflights13, odbc (>= 1.4.2), RJDBC, RMariaDB (>=
        1.2.2), rmarkdown, RPostgres (>= 1.4.5), RPostgreSQL, RSQLite
        (>= 2.3.8), testthat (>= 3.1.10)
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
Encoding: UTF-8
Language: en-gb
Collate: 'verb-copy-inline.R' 'verb-copy-to.R' 'db-sql.R' 'db.R'
        'utils-check.R' 'import-standalone-types-check.R'
        'import-standalone-obj-type.R' 'utils.R' 'sql.R' 'escape.R'
        'translate-sql-cut.R' 'translate-sql-string.R'
        'translate-sql-aggregate.R' 'translate-sql-scalar.R'
        'translate-sql-helpers.R' 'translate-sql-window.R'
        'translate-sql-conditional.R' 'backend-.R' 'backend-access.R'
        'backend-adbc.R' 'backend-db2.R' 'backend-hana.R'
        'backend-hive.R' 'backend-impala.R' 'backend-jdbc.R'
        'backend-mssql.R' 'backend-mysql.R' 'backend-odbc.R'
        'backend-oracle.R' 'backend-postgres.R'
        'backend-postgres-old.R' 'backend-redshift.R'
        'backend-snowflake.R' 'backend-spark-sql.R' 'backend-sqlite.R'
        'backend-teradata.R' 'backward-compatibility.R'
        'bind-queries.R' 'data-cache.R' 'data-lahman.R'
        'data-nycflights13.R' 'db-io.R' 'dbplyr.R' 'ident.R'
        'import-standalone-s3-register.R' 'join-by-compat.R'
        'join-cols-compat.R' 'lazy-ops.R' 'memdb.R' 'optimise-utils.R'
        'progress.R' 'query-base.R' 'query-join.R' 'query-rf-join.R'
        'query-select.R' 'query-semi-join.R' 'query-set-op.R'
        'query-union.R' 'query.R' 'remote.R' 'rows.R' 'schema.R'
        'sql-build.R' 'sql-clause.R' 'sql-dialect.R' 'sql-glue.R'
        'sql-quote.R' 'sql-superseded.R' 'src-sql.R' 'src_dbi.R'
        'table-name.R' 'tbl-lazy.R' 'tbl-sql.R' 'tidyeval-across.R'
        'tidyeval.R' 'translate-sql.R' 'utils-format.R'
        'verb-arrange.R' 'verb-collapse.R' 'verb-collect.R'
        'verb-compute.R' 'verb-count.R' 'verb-distinct.R'
        'verb-do-query.R' 'verb-do.R' 'verb-expand.R' 'verb-explain.R'
        'verb-fill.R' 'verb-filter.R' 'verb-group_by.R' 'verb-head.R'
        'verb-joins.R' 'verb-mutate.R' 'verb-pivot-longer.R'
        'verb-pivot-wider.R' 'verb-pull.R' 'verb-select.R'
        'verb-set-ops.R' 'verb-slice.R' 'verb-summarise.R'
        'verb-uncount.R' 'verb-window.R' 'with-dialect.R' 'zzz.R'
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-17 12:33:03 UTC; hadleywickham
Author: Hadley Wickham [aut, cre],
  Maximilian Girlich [aut],
  Edgar Ruiz [aut],
  Posit Software, PBC [cph, fnd]
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2026-06-17 21:50:07 UTC
Built: R 4.5.2; ; 2026-06-17 23:13:16 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$",tbl_lazy)
S3method("[",dbplyr_table_path)
S3method("[",sql)
S3method("[[",dbplyr_table_path)
S3method("[[",sql)
S3method(add_count,tbl_lazy)
S3method(anti_join,tbl_lazy)
S3method(arrange,tbl_lazy)
S3method(as.data.frame,tbl_lazy)
S3method(as.data.frame,tbl_sql)
S3method(auto_copy,tbl_sql)
S3method(c,dbplyr_table_path)
S3method(collapse,tbl_sql)
S3method(collect,tbl_sql)
S3method(compute,tbl_sql)
S3method(copy_to,src_sql)
S3method(count,tbl_lazy)
S3method(cross_join,tbl_lazy)
S3method(db_col_types,DBIConnection)
S3method(db_col_types,MariaDBConnection)
S3method(db_col_types,MySQL)
S3method(db_col_types,MySQLConnection)
S3method(db_col_types,PostgreSQL)
S3method(db_col_types,PostgreSQLConnection)
S3method(db_col_types,PqConnection)
S3method(db_col_types,SQLiteConnection)
S3method(db_col_types,TestConnection)
S3method(db_col_types,default)
S3method(db_collect,DBIConnection)
S3method(db_compute,"Spark SQL")
S3method(db_compute,DBIConnection)
S3method(db_connection_describe,DBIConnection)
S3method(db_connection_describe,MariaDBConnection)
S3method(db_connection_describe,MySQL)
S3method(db_connection_describe,MySQLConnection)
S3method(db_connection_describe,OdbcConnection)
S3method(db_connection_describe,PostgreSQL)
S3method(db_connection_describe,PostgreSQLConnection)
S3method(db_connection_describe,PqConnection)
S3method(db_connection_describe,SQLiteConnection)
S3method(db_copy_to,"Spark SQL")
S3method(db_copy_to,DBIConnection)
S3method(db_sql_render,"Microsoft SQL Server")
S3method(db_sql_render,DBIConnection)
S3method(db_sql_render,sql_dialect)
S3method(db_sql_render,sql_dialect_mssql)
S3method(db_table_drop_if_exists,DBIConnection)
S3method(db_table_drop_if_exists,OraConnection)
S3method(db_table_drop_if_exists,Oracle)
S3method(dbplyr_as_join_by,character)
S3method(dbplyr_as_join_by,default)
S3method(dbplyr_as_join_by,dplyr_join_by)
S3method(dbplyr_as_join_by,list)
S3method(dbplyr_edition,"Microsoft SQL Server")
S3method(dbplyr_edition,"Spark SQL")
S3method(dbplyr_edition,ACCESS)
S3method(dbplyr_edition,AdbiConnection)
S3method(dbplyr_edition,DB2Connection)
S3method(dbplyr_edition,HDB)
S3method(dbplyr_edition,Hive)
S3method(dbplyr_edition,Impala)
S3method(dbplyr_edition,JDBCConnection)
S3method(dbplyr_edition,MariaDBConnection)
S3method(dbplyr_edition,MySQL)
S3method(dbplyr_edition,MySQLConnection)
S3method(dbplyr_edition,OdbcConnection)
S3method(dbplyr_edition,OraConnection)
S3method(dbplyr_edition,Oracle)
S3method(dbplyr_edition,PostgreSQL)
S3method(dbplyr_edition,PostgreSQLConnection)
S3method(dbplyr_edition,PqConnection)
S3method(dbplyr_edition,Redshift)
S3method(dbplyr_edition,RedshiftConnection)
S3method(dbplyr_edition,SQLiteConnection)
S3method(dbplyr_edition,Snowflake)
S3method(dbplyr_edition,Teradata)
S3method(dbplyr_edition,TestConnection)
S3method(dbplyr_edition,default)
S3method(dbplyr_fill0,"Microsoft SQL Server")
S3method(dbplyr_fill0,ACCESS)
S3method(dbplyr_fill0,DBIConnection)
S3method(dbplyr_fill0,HDB)
S3method(dbplyr_fill0,MariaDBConnection)
S3method(dbplyr_fill0,MySQL)
S3method(dbplyr_fill0,MySQLConnection)
S3method(dbplyr_fill0,PostgreSQL)
S3method(dbplyr_fill0,PostgreSQLConnection)
S3method(dbplyr_fill0,PqConnection)
S3method(dbplyr_fill0,SQLiteConnection)
S3method(dbplyr_fill0,sql_dialect)
S3method(dbplyr_fill0,sql_dialect_access)
S3method(dbplyr_fill0,sql_dialect_hana)
S3method(dbplyr_fill0,sql_dialect_mariadb)
S3method(dbplyr_fill0,sql_dialect_mssql)
S3method(dbplyr_fill0,sql_dialect_mysql)
S3method(dbplyr_fill0,sql_dialect_postgres)
S3method(dbplyr_fill0,sql_dialect_sqlite)
S3method(dbplyr_write_table,DBIConnection)
S3method(dbplyr_write_table,OraConnection)
S3method(dbplyr_write_table,Oracle)
S3method(dim,tbl_lazy)
S3method(dimnames,tbl_lazy)
S3method(distinct,tbl_lazy)
S3method(do,tbl_sql)
S3method(dplyr::group_by_drop_default,tbl_lazy)
S3method(dplyr::intersect,tbl_lazy)
S3method(dplyr::setdiff,tbl_lazy)
S3method(dplyr::union,tbl_lazy)
S3method(dplyr::union_all,tbl_lazy)
S3method(escape,"NULL")
S3method(escape,Date)
S3method(escape,POSIXt)
S3method(escape,blob)
S3method(escape,character)
S3method(escape,data.frame)
S3method(escape,default)
S3method(escape,double)
S3method(escape,factor)
S3method(escape,ident)
S3method(escape,ident_q)
S3method(escape,integer)
S3method(escape,integer64)
S3method(escape,list)
S3method(escape,logical)
S3method(escape,reactivevalues)
S3method(escape,sql)
S3method(explain,tbl_sql)
S3method(filter_out,tbl_lazy)
S3method(flatten_query,base_query)
S3method(flatten_query,multi_join_query)
S3method(flatten_query,rf_join_query)
S3method(flatten_query,select_query)
S3method(flatten_query,semi_join_query)
S3method(flatten_query,set_op_query)
S3method(flatten_query,union_query)
S3method(flatten_query,values_query)
S3method(format,dbplyr_catalog)
S3method(format,dbplyr_schema)
S3method(format,ident)
S3method(format,sql)
S3method(format,src_sql)
S3method(full_join,tbl_lazy)
S3method(group_by,tbl_lazy)
S3method(group_size,tbl_sql)
S3method(group_vars,tbl_lazy)
S3method(groups,tbl_lazy)
S3method(head,tbl_lazy)
S3method(inner_join,tbl_lazy)
S3method(left_join,tbl_lazy)
S3method(mutate,tbl_lazy)
S3method(n_groups,tbl_sql)
S3method(names,sql_variant)
S3method(names,tbl_lazy)
S3method(op_frame,lazy_query)
S3method(op_frame,tbl_lazy)
S3method(op_grps,lazy_query)
S3method(op_grps,tbl_lazy)
S3method(op_sort,lazy_query)
S3method(op_sort,tbl_lazy)
S3method(op_vars,lazy_base_query)
S3method(op_vars,lazy_base_values_query)
S3method(op_vars,lazy_multi_join_query)
S3method(op_vars,lazy_rf_join_query)
S3method(op_vars,lazy_select_query)
S3method(op_vars,lazy_semi_join_query)
S3method(op_vars,lazy_set_op_query)
S3method(op_vars,lazy_union_query)
S3method(op_vars,tbl_lazy)
S3method(print,dbplyr_catalog)
S3method(print,dbplyr_schema)
S3method(print,dbplyr_sql_options)
S3method(print,dbplyr_table_path)
S3method(print,ident)
S3method(print,lazy_query)
S3method(print,query)
S3method(print,sql)
S3method(print,sql_clause)
S3method(print,sql_dialect)
S3method(print,sql_variant)
S3method(print,tbl_lazy)
S3method(print,tbl_sql)
S3method(pull,tbl_sql)
S3method(reframe,tbl_lazy)
S3method(relocate,tbl_lazy)
S3method(remote_table_path,lazy_base_local_query)
S3method(remote_table_path,lazy_base_remote_query)
S3method(remote_table_path,lazy_query)
S3method(remote_table_path,tbl_lazy)
S3method(rename,tbl_lazy)
S3method(rename_with,tbl_lazy)
S3method(right_join,tbl_lazy)
S3method(rows_append,tbl_lazy)
S3method(rows_delete,tbl_lazy)
S3method(rows_insert,tbl_lazy)
S3method(rows_patch,tbl_lazy)
S3method(rows_update,tbl_lazy)
S3method(rows_upsert,tbl_lazy)
S3method(same_src,src_sql)
S3method(same_src,tbl_lazy)
S3method(same_src,tbl_sql)
S3method(select,tbl_lazy)
S3method(semi_join,tbl_lazy)
S3method(show_query,tbl_lazy)
S3method(slice,tbl_lazy)
S3method(slice_head,tbl_lazy)
S3method(slice_max,tbl_lazy)
S3method(slice_min,tbl_lazy)
S3method(slice_sample,tbl_lazy)
S3method(slice_tail,tbl_lazy)
S3method(sql_build,lazy_base_local_query)
S3method(sql_build,lazy_base_remote_query)
S3method(sql_build,lazy_base_values_query)
S3method(sql_build,lazy_multi_join_query)
S3method(sql_build,lazy_rf_join_query)
S3method(sql_build,lazy_select_query)
S3method(sql_build,lazy_semi_join_query)
S3method(sql_build,lazy_set_op_query)
S3method(sql_build,lazy_union_query)
S3method(sql_build,tbl_lazy)
S3method(sql_cast_dispatch,Date)
S3method(sql_cast_dispatch,POSIXct)
S3method(sql_cast_dispatch,blob)
S3method(sql_cast_dispatch,character)
S3method(sql_cast_dispatch,factor)
S3method(sql_cast_dispatch,integer)
S3method(sql_cast_dispatch,integer64)
S3method(sql_cast_dispatch,logical)
S3method(sql_cast_dispatch,numeric)
S3method(sql_cast_dispatch,sql)
S3method(sql_dialect,"Microsoft SQL Server")
S3method(sql_dialect,"Spark SQL")
S3method(sql_dialect,ACCESS)
S3method(sql_dialect,AdbiConnection)
S3method(sql_dialect,DB2Connection)
S3method(sql_dialect,HDB)
S3method(sql_dialect,Hive)
S3method(sql_dialect,Impala)
S3method(sql_dialect,JDBCConnection)
S3method(sql_dialect,MariaDBConnection)
S3method(sql_dialect,MySQL)
S3method(sql_dialect,MySQLConnection)
S3method(sql_dialect,OdbcConnection)
S3method(sql_dialect,OraConnection)
S3method(sql_dialect,Oracle)
S3method(sql_dialect,PostgreSQL)
S3method(sql_dialect,PostgreSQLConnection)
S3method(sql_dialect,PqConnection)
S3method(sql_dialect,Redshift)
S3method(sql_dialect,RedshiftConnection)
S3method(sql_dialect,SQLiteConnection)
S3method(sql_dialect,Snowflake)
S3method(sql_dialect,Teradata)
S3method(sql_dialect,default)
S3method(sql_dialect,sql_dialect)
S3method(sql_escape_date,DBIConnection)
S3method(sql_escape_date,sql_dialect)
S3method(sql_escape_date,sql_dialect_access)
S3method(sql_escape_date,sql_dialect_postgres)
S3method(sql_escape_datetime,DBIConnection)
S3method(sql_escape_datetime,sql_dialect)
S3method(sql_escape_datetime,sql_dialect_access)
S3method(sql_escape_datetime,sql_dialect_mariadb)
S3method(sql_escape_datetime,sql_dialect_mysql)
S3method(sql_escape_datetime,sql_dialect_postgres)
S3method(sql_escape_ident,DBIConnection)
S3method(sql_escape_ident,TestConnection)
S3method(sql_escape_ident,sql_dialect)
S3method(sql_escape_logical,DBIConnection)
S3method(sql_escape_logical,sql_dialect)
S3method(sql_escape_logical,sql_dialect_access)
S3method(sql_escape_logical,sql_dialect_mssql)
S3method(sql_escape_logical,sql_dialect_sqlite)
S3method(sql_escape_raw,DBIConnection)
S3method(sql_escape_raw,sql_dialect)
S3method(sql_escape_raw,sql_dialect_mssql)
S3method(sql_escape_string,default)
S3method(sql_expr_matches,default)
S3method(sql_join_suffix,DBIConnection)
S3method(sql_join_suffix,sql_dialect)
S3method(sql_query_append,DBIConnection)
S3method(sql_query_append,sql_dialect)
S3method(sql_query_append,sql_dialect_mssql)
S3method(sql_query_delete,DBIConnection)
S3method(sql_query_delete,sql_dialect)
S3method(sql_query_delete,sql_dialect_mssql)
S3method(sql_query_explain,DBIConnection)
S3method(sql_query_explain,sql_dialect)
S3method(sql_query_explain,sql_dialect_oracle)
S3method(sql_query_explain,sql_dialect_postgres)
S3method(sql_query_explain,sql_dialect_redshift)
S3method(sql_query_explain,sql_dialect_sqlite)
S3method(sql_query_fields,DBIConnection)
S3method(sql_query_fields,sql_dialect)
S3method(sql_query_insert,DBIConnection)
S3method(sql_query_insert,sql_dialect)
S3method(sql_query_insert,sql_dialect_mssql)
S3method(sql_query_insert,sql_dialect_postgres)
S3method(sql_query_join,DBIConnection)
S3method(sql_query_join,sql_dialect)
S3method(sql_query_join,sql_dialect_mariadb)
S3method(sql_query_join,sql_dialect_mysql)
S3method(sql_query_multi_join,DBIConnection)
S3method(sql_query_multi_join,sql_dialect)
S3method(sql_query_multi_join,sql_dialect_access)
S3method(sql_query_rows,DBIConnection)
S3method(sql_query_rows,sql_dialect)
S3method(sql_query_save,DBIConnection)
S3method(sql_query_save,sql_dialect)
S3method(sql_query_save,sql_dialect_mssql)
S3method(sql_query_save,sql_dialect_oracle)
S3method(sql_query_select,DBIConnection)
S3method(sql_query_select,sql_dialect)
S3method(sql_query_select,sql_dialect_access)
S3method(sql_query_select,sql_dialect_db2)
S3method(sql_query_select,sql_dialect_mssql)
S3method(sql_query_select,sql_dialect_oracle)
S3method(sql_query_select,sql_dialect_teradata)
S3method(sql_query_semi_join,DBIConnection)
S3method(sql_query_semi_join,sql_dialect)
S3method(sql_query_set_op,DBIConnection)
S3method(sql_query_set_op,sql_dialect)
S3method(sql_query_set_op,sql_dialect_hive)
S3method(sql_query_set_op,sql_dialect_sqlite)
S3method(sql_query_union,DBIConnection)
S3method(sql_query_union,sql_dialect)
S3method(sql_query_update_from,DBIConnection)
S3method(sql_query_update_from,sql_dialect)
S3method(sql_query_update_from,sql_dialect_mariadb)
S3method(sql_query_update_from,sql_dialect_mssql)
S3method(sql_query_update_from,sql_dialect_mysql)
S3method(sql_query_upsert,DBIConnection)
S3method(sql_query_upsert,sql_dialect)
S3method(sql_query_upsert,sql_dialect_mariadb)
S3method(sql_query_upsert,sql_dialect_mssql)
S3method(sql_query_upsert,sql_dialect_mysql)
S3method(sql_query_upsert,sql_dialect_oracle)
S3method(sql_query_upsert,sql_dialect_postgres)
S3method(sql_query_upsert,sql_dialect_sqlite)
S3method(sql_query_wrap,DBIConnection)
S3method(sql_query_wrap,sql_dialect)
S3method(sql_render,base_query)
S3method(sql_render,lazy_query)
S3method(sql_render,multi_join_query)
S3method(sql_render,rf_join_query)
S3method(sql_render,select_query)
S3method(sql_render,semi_join_query)
S3method(sql_render,set_op_query)
S3method(sql_render,tbl_lazy)
S3method(sql_render,union_query)
S3method(sql_render,values_query)
S3method(sql_returning_cols,DBIConnection)
S3method(sql_returning_cols,sql_dialect)
S3method(sql_returning_cols,sql_dialect_mssql)
S3method(sql_set_op_method,DBIConnection)
S3method(sql_set_op_method,sql_dialect)
S3method(sql_set_op_method,sql_dialect_oracle)
S3method(sql_table_analyze,DBIConnection)
S3method(sql_table_analyze,sql_dialect)
S3method(sql_table_analyze,sql_dialect_access)
S3method(sql_table_analyze,sql_dialect_db2)
S3method(sql_table_analyze,sql_dialect_hana)
S3method(sql_table_analyze,sql_dialect_hive)
S3method(sql_table_analyze,sql_dialect_impala)
S3method(sql_table_analyze,sql_dialect_mariadb)
S3method(sql_table_analyze,sql_dialect_mssql)
S3method(sql_table_analyze,sql_dialect_mysql)
S3method(sql_table_analyze,sql_dialect_oracle)
S3method(sql_table_analyze,sql_dialect_snowflake)
S3method(sql_table_analyze,sql_dialect_spark_sql)
S3method(sql_table_analyze,sql_dialect_teradata)
S3method(sql_table_index,DBIConnection)
S3method(sql_table_index,sql_dialect)
S3method(sql_table_temporary,DBIConnection)
S3method(sql_table_temporary,sql_dialect)
S3method(sql_table_temporary,sql_dialect_hana)
S3method(sql_table_temporary,sql_dialect_mssql)
S3method(sql_table_temporary,sql_dialect_oracle)
S3method(sql_translation,DBIConnection)
S3method(sql_translation,sql_dialect)
S3method(sql_translation,sql_dialect_access)
S3method(sql_translation,sql_dialect_db2)
S3method(sql_translation,sql_dialect_hana)
S3method(sql_translation,sql_dialect_hive)
S3method(sql_translation,sql_dialect_impala)
S3method(sql_translation,sql_dialect_mariadb)
S3method(sql_translation,sql_dialect_mssql)
S3method(sql_translation,sql_dialect_mysql)
S3method(sql_translation,sql_dialect_odbc)
S3method(sql_translation,sql_dialect_oracle)
S3method(sql_translation,sql_dialect_postgres)
S3method(sql_translation,sql_dialect_redshift)
S3method(sql_translation,sql_dialect_snowflake)
S3method(sql_translation,sql_dialect_spark_sql)
S3method(sql_translation,sql_dialect_sqlite)
S3method(sql_translation,sql_dialect_teradata)
S3method(sql_values_subquery,DBIConnection)
S3method(sql_values_subquery,sql_dialect)
S3method(sql_values_subquery,sql_dialect_db2)
S3method(sql_values_subquery,sql_dialect_hana)
S3method(sql_values_subquery,sql_dialect_mariadb)
S3method(sql_values_subquery,sql_dialect_mssql)
S3method(sql_values_subquery,sql_dialect_mysql)
S3method(sql_values_subquery,sql_dialect_oracle)
S3method(sql_values_subquery,sql_dialect_postgres)
S3method(sql_values_subquery,sql_dialect_redshift)
S3method(src_tbls,src_sql)
S3method(summarise,tbl_lazy)
S3method(table_path_components,"Microsoft SQL Server")
S3method(table_path_components,DBIConnection)
S3method(table_path_components,sql_dialect)
S3method(table_path_components,sql_dialect_mssql)
S3method(tail,tbl_lazy)
S3method(tally,tbl_lazy)
S3method(tbl,src_dbi)
S3method(tbl_format_header,tbl_sql)
S3method(tbl_vars,tbl_lazy)
S3method(tidyr::complete,tbl_lazy)
S3method(tidyr::expand,tbl_lazy)
S3method(tidyr::fill,tbl_lazy)
S3method(tidyr::pivot_longer,tbl_lazy)
S3method(tidyr::pivot_wider,tbl_lazy)
S3method(tidyr::replace_na,tbl_lazy)
S3method(tidyselect_data_has_predicates,tbl_lazy)
S3method(tidyselect_data_proxy,tbl_lazy)
S3method(transmute,tbl_lazy)
S3method(ungroup,tbl_lazy)
S3method(union_all,tbl_lazy)
S3method(values_prepare,DBIConnection)
S3method(values_prepare,sql_dialect)
S3method(values_prepare,sql_dialect_sqlite)
export(.sql)
export(as.sql)
export(as_table_path)
export(base_agg)
export(base_no_win)
export(base_odbc_agg)
export(base_odbc_scalar)
export(base_odbc_win)
export(base_scalar)
export(base_win)
export(bind_queries)
export(build_sql)
export(check_table_path)
export(copy_inline)
export(copy_lahman)
export(copy_nycflights13)
export(db_col_types)
export(db_collect)
export(db_compute)
export(db_connection_describe)
export(db_copy_to)
export(db_sql_render)
export(db_table_drop_if_exists)
export(dbplyr_edition)
export(dbplyr_pivot_wider_spec)
export(dbplyr_uncount)
export(dialect_access)
export(dialect_ansi)
export(dialect_db2)
export(dialect_hana)
export(dialect_hive)
export(dialect_impala)
export(dialect_mariadb)
export(dialect_mssql)
export(dialect_mysql)
export(dialect_odbc)
export(dialect_oracle)
export(dialect_postgres)
export(dialect_redshift)
export(dialect_snowflake)
export(dialect_spark_sql)
export(dialect_sqlite)
export(dialect_teradata)
export(escape)
export(escape_ansi)
export(get_returned_rows)
export(has_lahman)
export(has_nycflights13)
export(has_returned_rows)
export(ident)
export(ident_q)
export(in_catalog)
export(in_schema)
export(is.ident)
export(is.sql)
export(is_table_path)
export(lahman_mysql)
export(lahman_postgres)
export(lahman_sqlite)
export(lahman_srcs)
export(last_sql)
export(lazy_base_query)
export(lazy_frame)
export(lazy_multi_join_query)
export(lazy_query)
export(lazy_rf_join_query)
export(lazy_select_query)
export(lazy_semi_join_query)
export(lazy_set_op_query)
export(lazy_union_query)
export(local_memdb_frame)
export(memdb)
export(memdb_frame)
export(named_commas)
export(new_sql_dialect)
export(nycflights13_postgres)
export(nycflights13_sqlite)
export(op_frame)
export(op_grps)
export(op_sort)
export(op_vars)
export(partial_eval)
export(remote_con)
export(remote_name)
export(remote_query)
export(remote_query_plan)
export(remote_src)
export(remote_table)
export(rf_join_query)
export(select_query)
export(semi_join_query)
export(set_op_query)
export(simulate_access)
export(simulate_dbi)
export(simulate_hana)
export(simulate_hive)
export(simulate_impala)
export(simulate_mariadb)
export(simulate_mssql)
export(simulate_mysql)
export(simulate_odbc)
export(simulate_oracle)
export(simulate_postgres)
export(simulate_redshift)
export(simulate_snowflake)
export(simulate_spark_sql)
export(simulate_sqlite)
export(simulate_teradata)
export(sql)
export(sql_aggregate)
export(sql_aggregate_2)
export(sql_aggregate_n)
export(sql_build)
export(sql_call2)
export(sql_cast)
export(sql_check_na_rm)
export(sql_cot)
export(sql_dialect)
export(sql_escape_date)
export(sql_escape_datetime)
export(sql_escape_ident)
export(sql_escape_logical)
export(sql_escape_raw)
export(sql_escape_string)
export(sql_expr)
export(sql_expr_matches)
export(sql_glue)
export(sql_glue2)
export(sql_indent_subquery)
export(sql_infix)
export(sql_join_suffix)
export(sql_log)
export(sql_not_supported)
export(sql_optimise)
export(sql_options)
export(sql_paste)
export(sql_paste_infix)
export(sql_prefix)
export(sql_query_append)
export(sql_query_delete)
export(sql_query_explain)
export(sql_query_fields)
export(sql_query_insert)
export(sql_query_join)
export(sql_query_multi_join)
export(sql_query_rows)
export(sql_query_save)
export(sql_query_select)
export(sql_query_semi_join)
export(sql_query_set_op)
export(sql_query_union)
export(sql_query_update_from)
export(sql_query_upsert)
export(sql_query_wrap)
export(sql_quote)
export(sql_random)
export(sql_render)
export(sql_returning_cols)
export(sql_runif)
export(sql_set_op_method)
export(sql_str_sub)
export(sql_substr)
export(sql_table_analyze)
export(sql_table_index)
export(sql_table_temporary)
export(sql_translation)
export(sql_translator)
export(sql_try_cast)
export(sql_variant)
export(sql_vector)
export(src_dbi)
export(src_memdb)
export(src_sql)
export(table_path_components)
export(table_path_name)
export(tbl_lazy)
export(tbl_memdb)
export(tbl_sql)
export(translate_sql)
export(translate_sql_)
export(union_query)
export(win_absent)
export(win_aggregate)
export(win_aggregate_2)
export(win_cumulative)
export(win_current_frame)
export(win_current_group)
export(win_current_order)
export(win_over)
export(win_rank)
export(win_recycled)
export(window_frame)
export(window_order)
export(with_dialect)
import(rlang)
importFrom(R6,R6Class)
importFrom(cli,cli_abort)
importFrom(dplyr,add_count)
importFrom(dplyr,anti_join)
importFrom(dplyr,arrange)
importFrom(dplyr,auto_copy)
importFrom(dplyr,case_when)
importFrom(dplyr,coalesce)
importFrom(dplyr,collapse)
importFrom(dplyr,collect)
importFrom(dplyr,compute)
importFrom(dplyr,copy_to)
importFrom(dplyr,count)
importFrom(dplyr,cross_join)
importFrom(dplyr,distinct)
importFrom(dplyr,do)
importFrom(dplyr,explain)
importFrom(dplyr,filter)
importFrom(dplyr,filter_out)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,group_size)
importFrom(dplyr,group_vars)
importFrom(dplyr,groups)
importFrom(dplyr,inner_join)
importFrom(dplyr,intersect)
importFrom(dplyr,join_by)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,n_groups)
importFrom(dplyr,pull)
importFrom(dplyr,reframe)
importFrom(dplyr,relocate)
importFrom(dplyr,rename)
importFrom(dplyr,rename_with)
importFrom(dplyr,right_join)
importFrom(dplyr,rows_append)
importFrom(dplyr,rows_delete)
importFrom(dplyr,rows_insert)
importFrom(dplyr,rows_patch)
importFrom(dplyr,rows_update)
importFrom(dplyr,rows_upsert)
importFrom(dplyr,same_src)
importFrom(dplyr,select)
importFrom(dplyr,semi_join)
importFrom(dplyr,setdiff)
importFrom(dplyr,show_query)
importFrom(dplyr,slice)
importFrom(dplyr,slice_head)
importFrom(dplyr,slice_max)
importFrom(dplyr,slice_min)
importFrom(dplyr,slice_sample)
importFrom(dplyr,slice_tail)
importFrom(dplyr,src_tbls)
importFrom(dplyr,summarise)
importFrom(dplyr,tally)
importFrom(dplyr,tbl)
importFrom(dplyr,tbl_vars)
importFrom(dplyr,transmute)
importFrom(dplyr,ungroup)
importFrom(dplyr,union)
importFrom(dplyr,union_all)
importFrom(glue,glue)
importFrom(lifecycle,deprecated)
importFrom(methods,initialize)
importFrom(pillar,tbl_format_header)
importFrom(stats,setNames)
importFrom(stats,update)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tidyselect,everything)
importFrom(tidyselect,tidyselect_data_has_predicates)
importFrom(tidyselect,tidyselect_data_proxy)
importFrom(utils,head)
importFrom(utils,tail)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# dbplyr <a href='https://dbplyr.tidyverse.org'><img src='man/figures/logo.png' align="right" height="139" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/dbplyr)](https://cran.r-project.org/package=dbplyr)
[![R-CMD-check](https://github.com/tidyverse/dbplyr/workflows/R-CMD-check/badge.svg)](https://github.com/tidyverse/dbplyr/actions)
[![Codecov test
coverage](https://codecov.io/gh/tidyverse/dbplyr/branch/main/graph/badge.svg)](https://app.codecov.io/gh/tidyverse/dbplyr?branch=main)
<!-- badges: end -->

## Overview

dbplyr is the database backend for [dplyr](https://dplyr.tidyverse.org).
It allows you to use remote database tables as if they are in-memory
data frames by automatically converting dplyr code into SQL.

To learn more about why you might use dbplyr instead of writing SQL, see
`vignette("sql")`. To learn more about the details of the SQL
translation, see `vignette("translation-verb")` and
`vignette("translation-function")`.

## Installation

``` r
# The easiest way to get dbplyr is to install the whole tidyverse:
install.packages("tidyverse")

# Alternatively, install just dbplyr:
install.packages("dbplyr")

# Or the the development version from GitHub:
# install.packages("devtools")
devtools::install_github("tidyverse/dbplyr")
```

## Usage

dbplyr is designed to work with database tables as if they were local
data frames. To demonstrate this I’ll first create an in-memory SQLite
database and copy over a dataset:

``` r
library(dplyr, warn.conflicts = FALSE)

con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
copy_to(con, mtcars)
```

Note that you don’t actually need to load dbplyr with `library(dbplyr)`;
dplyr automatically loads it for you when it sees you working with a
database. Database connections are coordinated by the DBI package. Learn
more at <https://dbi.r-dbi.org/>

Now you can retrieve a table using `tbl()` (see `?tbl_dbi` for more
details). Printing it just retrieves the first few rows:

``` r
mtcars2 <- tbl(con, "mtcars")
mtcars2
#> # Source:   table<mtcars> [?? x 11]
#> # Database: sqlite 3.39.4 [:memory:]
#>      mpg   cyl  disp    hp  drat    wt  qsec    vs    am  gear  carb
#>    <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl> <dbl>
#>  1  21       6  160    110  3.9   2.62  16.5     0     1     4     4
#>  2  21       6  160    110  3.9   2.88  17.0     0     1     4     4
#>  3  22.8     4  108     93  3.85  2.32  18.6     1     1     4     1
#>  4  21.4     6  258    110  3.08  3.22  19.4     1     0     3     1
#>  5  18.7     8  360    175  3.15  3.44  17.0     0     0     3     2
#>  6  18.1     6  225    105  2.76  3.46  20.2     1     0     3     1
#>  7  14.3     8  360    245  3.21  3.57  15.8     0     0     3     4
#>  8  24.4     4  147.    62  3.69  3.19  20       1     0     4     2
#>  9  22.8     4  141.    95  3.92  3.15  22.9     1     0     4     2
#> 10  19.2     6  168.   123  3.92  3.44  18.3     1     0     4     4
#> # … with more rows
```

All dplyr calls are evaluated lazily, generating SQL that is only sent
to the database when you request the data.

``` r
# lazily generates query
summary <- mtcars2 %>% 
  group_by(cyl) %>% 
  summarise(mpg = mean(mpg, na.rm = TRUE)) %>% 
  arrange(desc(mpg))

# see query
summary %>% show_query()
#> <SQL>
#> SELECT `cyl`, AVG(`mpg`) AS `mpg`
#> FROM `mtcars`
#> GROUP BY `cyl`
#> ORDER BY `mpg` DESC

# execute query and retrieve results
summary %>% collect()
#> # A tibble: 3 × 2
#>     cyl   mpg
#>   <dbl> <dbl>
#> 1     4  26.7
#> 2     6  19.7
#> 3     8  15.1
```

## Code of Conduct

Please note that the dbplyr project is released with a [Contributor Code
of Conduct](https://dbplyr.tidyverse.org/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

#' Database versions of the nycflights13 data
#'
#' These functions cache the data from the `nycflights13` database in
#' a local database, for use in examples and vignettes. Indexes are created
#' to making joining tables on natural keys efficient.
#'
#' @keywords internal
#' @name nycflights13
NULL

# nocov start
#' @export
#' @rdname nycflights13
#' @param path location of SQLite database file
nycflights13_sqlite <- function(path = NULL) {
  cache_computation("nycflights_sqlite", {
    path <- db_location(path, "nycflights13.sqlite")
    message("Caching nycflights db at ", path)
    con <- DBI::dbConnect(RSQLite::SQLite(), dbname = path, create = TRUE)
    copy_nycflights13(con)
  })
}

#' @export
#' @rdname nycflights13
#' @param dbname,... Arguments passed on to [src_postgres()]
nycflights13_postgres <- function(dbname = "nycflights13", ...) {
  cache_computation("nycflights_postgres", {
    message("Caching nycflights db in postgresql db ", dbname)
    con <- DBI::dbConnect(RPostgres::Postgres(), dbname = dbname, ...)
    copy_nycflights13(con)
  })
}

#' @rdname nycflights13
#' @export
has_nycflights13 <- function(type = c("sqlite", "postgresql"), ...) {
  if (!requireNamespace("nycflights13", quietly = TRUE)) return(FALSE)

  type <- match.arg(type)

  succeeds(switch(
    type,
    sqlite = nycflights13_sqlite(...),
    postgres = nycflights13_postgres(...)
  ), quiet = TRUE)
}


#' @export
#' @rdname nycflights13
copy_nycflights13 <- function(con, ...) {
  all <- utils::data(package = "nycflights13")$results[, 3]
  unique_index <- list(
    airlines = list("carrier"),
    planes =   list("tailnum")
  )
  index <- list(
    airports = list("faa"),
    flights =  list(c("year", "month", "day"), "carrier", "tailnum", "origin", "dest"),
    weather =  list(c("year", "month", "day"), "origin")
  )

  tables <- setdiff(all, DBI::dbListTables(con))

  # Create missing tables
  for (table in tables) {
    df <- getExportedValue("nycflights13", table)
    message("Creating table: ", table)

    copy_to(
      con, df, table,
      unique_indexes = unique_index[[table]],
      indexes = index[[table]],
      temporary = FALSE
    )
  }
  invisible(con)
}
# nocov end

#' @importFrom pillar tbl_format_header style_subtle align
#' @export
tbl_format_header.tbl_sql <- function(x, setup, ...) {
  # The setup object may know the total number of rows
  desc <- tbl_desc(x, rows_total = setup$rows_total)
  named_header <- tbl_sum_tbl_sql(x, desc = desc)

  # Adapted from pillar
  header <- paste0(
    align(paste0(names(named_header), ":")),
    " ",
    named_header
  )

  style_subtle(paste0("# ", header))
}

join_cols <- function(x_names,
                      y_names,
                      by,
                      ...,
                      suffix = c(".x", ".y"),
                      keep = NULL,
                      error_call = caller_env()) {
  check_dots_empty0(...)

  if (is_false(keep) && any(by$condition != "==")) {
    abort(
      "Can't set `keep = FALSE` when using an inequality, rolling, or overlap join.",
      call = error_call
    )
  }

  check_duplicate_vars(x_names, "x", error_call = error_call)
  check_duplicate_vars(y_names, "y", error_call = error_call)

  check_join_vars(by$x, x_names, by$condition, "x", error_call = error_call)
  check_join_vars(by$y, y_names, by$condition, "y", error_call = error_call)

  suffix <- standardise_join_suffix(suffix, error_call = error_call)

  x_by <- set_names(match(by$x, x_names), by$x)
  y_by <- set_names(match(by$y, y_names), by$y)

  x_loc <- seq_along(x_names)
  names(x_loc) <- x_names
  if (is_null(keep)) {
    # In x_out, equi key variables need to keep the same name, and non-equi
    # key variables and aux variables need suffixes for duplicates that appear
    # in y_out. This is equivalent to `keep = TRUE` for the non-equi keys and
    # `keep = FALSE` for the equi keys.
    equi <- by$condition == "=="
    y_aux <- setdiff(y_names, c(by$x[equi], by$y[equi]))
    x_ignore <- by$x[equi]
    x_check <- !x_names %in% x_ignore
    names(x_loc)[x_check] <- add_suffixes(x_names[x_check], c(x_ignore, y_aux), suffix$x)
  } else if (is_false(keep)) {
    # In x_out, key variables need to keep the same name, and aux
    # variables need suffixes for duplicates that appear in y_out
    y_aux <- setdiff(y_names, c(by$x, by$y))
    x_ignore <- by$x
    x_check <- !x_names %in% x_ignore
    names(x_loc)[x_check] <- add_suffixes(x_names[x_check], c(x_ignore, y_aux), suffix$x)
  } else {
    # In x_out, key variables and aux variables need suffixes
    # for duplicates that appear in y_out
    names(x_loc) <- add_suffixes(x_names, y_names, suffix$x)
  }

  y_loc <- seq_along(y_names)
  names(y_loc) <- add_suffixes(y_names, x_names, suffix$y)
  if (is_null(keep)) {
    equi <- by$condition == "=="
    y_ignore <- by$y[equi]
    y_loc <- y_loc[!y_names %in% y_ignore]
  } else if (is_false(keep)) {
    y_ignore <- by$y
    y_loc <- y_loc[!y_names %in% y_ignore]
  }

  # key = named locations to use for matching
  # out = named locations to use in output
  list(
    x = list(key = x_by, out = x_loc),
    y = list(key = y_by, out = y_loc)
  )
}

check_join_vars <- function(vars,
                            names,
                            condition,
                            input,
                            ...,
                            error_call = caller_env()) {
  check_dots_empty0(...)

  if (!is.character(vars)) {
    message <- "Join columns in `{input}` must be character vectors."
    cli_abort(message, call = error_call)
  }

  na <- is.na(vars)
  if (any(na)) {
    bullets <- c(
      "Join columns in `{input}` can't be `NA`.",
      x = "Problem at position {err_vars(na)}."
    )
    abort(bullets, call = error_call)
  }

  # Columns are allowed to appear in more than one non-equi condition
  # (but not in a mix of non-equi and equi conditions).
  # When non-equi conditions are present, `keep` can't be `FALSE` so we don't
  # have to worry about merging into the same key column multiple times (#6499).
  non_equi <- condition != "=="
  vars <- c(vars[!non_equi], unique(vars[non_equi]))

  dup <- duplicated(vars)
  if (any(dup)) {
    vars <- unique(vars[dup])

    bullets <- c(
      "Join columns in `{input}` must be unique.",
      x = "Problem with {err_vars(vars)}."
    )

    cli_abort(bullets, call = error_call)
  }

  missing <- setdiff(vars, names)
  if (length(missing) > 0) {
    bullets <- c(
      "Join columns in `{input}` must be present in the data.",
      x = "Problem with {err_vars(missing)}."
    )
    cli_abort(bullets, call = error_call)
  }
}

check_duplicate_vars <- function(vars,
                                 input,
                                 ...,
                                 error_call = caller_env()) {
  check_dots_empty0(...)

  dup <- duplicated(vars)

  if (any(dup)) {
    bullets <- c(
      "Input columns in `{input}` must be unique.",
      x = "Problem with {err_vars(vars[dup])}."
    )
    cli_abort(bullets, call = error_call)
  }
}

standardise_join_suffix <- function(x,
                                    ...,
                                    error_call = caller_env()) {
  check_dots_empty0(...)

  if (!is.character(x) || length(x) != 2) {
    bullets <- c(
      "`suffix` must be a character vector of length 2, not {.obj_type_friendly {x}} of length {length(x)}."
    )
    cli_abort(bullets, call = error_call)
  }

  if (any(is.na(x))) {
    msg <- "`suffix` can't be `NA`."
    cli_abort(msg, call = error_call)
  }

  list(x = x[[1]], y = x[[2]])
}

# `join_cols()` checks that `x` and `y` are individually unique,
# which plays into assumptions made here
add_suffixes <- function(x, y, suffix) {
  if (identical(suffix, "")) {
    return(x)
  }

  x <- c(y, x)

  # Never marks the "first" duplicate (i.e. never anything in `y`)
  dup <- duplicated(x)

  while (any(dup)) {
    x[dup] <- paste0(x[dup], suffix)
    dup <- duplicated(x)
  }

  loc <- seq2(length(y) + 1L, length(x))
  x <- x[loc]

  x
}

#' Perform arbitrary computation on remote backend
#'
#' @inheritParams dplyr::do
#' @param .chunk_size The size of each chunk to pull into R. If this number is
#'   too big, the process will be slow because R has to allocate and free a lot
#'   of memory. If it's too small, it will be slow, because of the overhead of
#'   talking to the database.
#' @export
#' @importFrom dplyr do
do.tbl_sql <- function(.data, ..., .chunk_size = 1e4L) {
  groups_sym <- groups(.data)

  if (length(groups_sym) == 0) {
    .data <- collect(.data)
    return(do(.data, ...))
  }

  args <- quos(...)
  named <- named_args(args)

  # Create data frame of labels
  labels <- .data %>%
    select(!!! groups_sym) %>%
    summarise() %>%
    collect()

  con <- .data$src$con

  n <- nrow(labels)
  m <- length(args)

  out <- replicate(m, vector("list", n), simplify = FALSE)
  names(out) <- names(args)
  p <- progress_estimated(n * m, min_time = 2)

  # Create ungrouped data frame suitable for chunked retrieval
  query <- Query$new(con, db_sql_render(con, ungroup(.data)), op_vars(.data))

  # When retrieving in pages, there's no guarantee we'll get a complete group.
  # So we always assume the last group in the chunk is incomplete, and leave
  # it for the next. If the group size is large than chunk size, it may
  # take a couple of iterations to get the entire group, but that should
  # be an unusual situation.
  last_group <- NULL
  i <- 0

  # Assumes `chunk` to be ordered with group columns first
  gvars <- seq_along(groups_sym)

  # Initialise a data mask for tidy evaluation
  env <- env(empty_env())
  mask <- new_data_mask(env)

  query$fetch_paged(.chunk_size, function(chunk) {
    if (!is_null(last_group)) {
      chunk <- rbind(last_group, chunk)
    }

    # Create an id for each group
    grouped <- chunk %>% dplyr::group_by(!!! syms(names(chunk)[gvars]))

    if (utils::packageVersion("dplyr") < "0.7.9") {
      index <- attr(grouped, "indices") # nocov
      # convert from 0-index
      index <- lapply(index, `+`, 1L) # nocov
    } else {
      index <- dplyr::group_rows(grouped)
    }

    n <- length(index)

    last_group <<- chunk[index[[length(index)]], , drop = FALSE]

    for (j in seq_len(n - 1)) {
      cur_chunk <- chunk[index[[j]], , drop = FALSE]

      # Update pronouns within the data mask
      env$. <- cur_chunk
      env$.data <- cur_chunk

      for (k in seq_len(m)) {
        out[[k]][[i + j]] <<- eval_tidy(args[[k]], mask)
        p$tick()$print()
      }
    }
    i <<- i + (n - 1)
  })

  # Process last group
  if (!is_null(last_group)) {
    env$. <- last_group
    last_group <- env$.data
    for (k in seq_len(m)) {
      out[[k]][[i + 1]] <- eval_tidy(args[[k]], mask)
      p$tick()$print()
    }
  }

  if (!named) {
    label_output_dataframe(labels, out, group_vars(.data))
  } else {
    label_output_list(labels, out, group_vars(.data))
  }
}

# Helper functions -------------------------------------------------------------

label_output_dataframe <- function(labels, out, groups) {
  data_frame <- vapply(out[[1]], is.data.frame, logical(1))
  if (any(!data_frame)) {
    cli_abort(c(
      "Results must be data frames",
      "Problems at positions {which(!data_frame)}"
    ))
  }

  rows <- vapply(out[[1]], nrow, numeric(1))
  out <- dplyr::bind_rows(out[[1]])

  if (!is.null(labels)) {
    # Remove any common columns from labels
    labels <- labels[setdiff(names(labels), names(out))]

    # Repeat each row to match data
    labels <- labels[rep(seq_len(nrow(labels)), rows), , drop = FALSE]
    rownames(labels) <- NULL

    dplyr::grouped_df(dplyr::bind_cols(labels, out), groups)
  } else {
    dplyr::rowwise(out) # nocov
  }
}

label_output_list <- function(labels, out, groups) {
  if (!is.null(labels)) {
    labels[names(out)] <- out
    dplyr::rowwise(labels)
  } else {
     # nocov start
    class(out) <- "data.frame"
    attr(out, "row.names") <- .set_row_names(length(out[[1]]))
    dplyr::rowwise(out)
     # nocov end
  }
}

named_args <- function(args) {
  # Arguments must either be all named or all unnamed.
  named <- sum(names2(args) != "")
  if (!(named == 0 || named == length(args))) {
    cli_abort(
      "Arguments to {.fun do} must either be all named or all unnamed"
    )
  }
  if (named == 0 && length(args) > 1) {
    cli_abort("Can only supply single unnamed argument to {.fun do}")
  }

  # Check for old syntax
  if (named == 1 && names(args) == ".f") {
    cli_abort(
      "{.fun do} syntax changed in dplyr 0.2. Please see documentation for details"
    )
  }

  named != 0
}

#' Database I/O generics
#'
#' @description
#' These generics are responsible for getting data into and out of the
#' database. They should be used a last resort - only use them when you can't
#' make a backend work by providing methods for DBI generics, or for dbplyr's
#' SQL generation generics. They tend to be most needed when a backend has
#' special handling of temporary tables.
#'
#' * `db_copy_to()` implements [copy_to.src_sql()] by calling
#'   `db_write_table()` (which calls [DBI::dbWriteTable()]) to transfer the
#'   data, then optionally adds indexes (via [sql_table_index()]) and
#'   analyses (via [sql_table_analyze()]).
#'
#' * `db_compute()` implements [compute.tbl_sql()] by calling
#'   [sql_query_save()] to create the table, then optionally adds indexes
#'   (via [sql_table_index()]) and analyses (via [sql_table_analyze()]).
#'
#' * `db_collect()` implements [collect.tbl_sql()] using [DBI::dbSendQuery()]
#'   and [DBI::dbFetch()].
#'
#' * `db_table_temporary()` is used for databases that have special naming
#'   schemes for temporary tables (e.g. SQL server and SAP HANA require
#'   temporary tables to start with `#`)
#'
#' @keywords internal
#' @family generic
#' @name db-io
#' @aliases NULL
NULL

#' @export
#' @rdname db-io
db_copy_to <-  function(con,
                        table,
                        values,
                        ...,
                        overwrite = FALSE,
                        types = NULL,
                        temporary = TRUE,
                        unique_indexes = NULL,
                        indexes = NULL,
                        analyze = TRUE,
                        in_transaction = TRUE) {
  check_table_id(table)
  check_bool(overwrite)
  check_character(types, allow_null = TRUE)
  check_named(types)
  check_bool(temporary)
  check_bool(analyze)
  check_dots_used()
  check_bool(in_transaction)

  UseMethod("db_copy_to")
}
#' @export
db_copy_to.DBIConnection <- function(con,
                                     table,
                                     values,
                                     ...,
                                     overwrite = FALSE,
                                     types = NULL,
                                     temporary = TRUE,
                                     unique_indexes = NULL,
                                     indexes = NULL,
                                     analyze = TRUE,
                                     in_transaction = TRUE) {
  table <- as_table_path(table, con)
  new <- db_table_temporary(con, table, temporary)
  table <- new$table
  temporary <- new$temporary
  call <- current_env()

  with_transaction(
    con,
    in_transaction,
    "Can't copy data to table {.field {format(table, con = con)}}.",
    {
      table <- dplyr::db_write_table(con, table,
        types = types,
        values = values,
        temporary = temporary,
        overwrite = overwrite,
        ...
      )
      create_indexes(con, table, unique_indexes, unique = TRUE)
      create_indexes(con, table, indexes)
      if (analyze) dbplyr_analyze(con, table)
    }
  )

  table
}

#' @export
#' @rdname db-io
db_compute <- function(con,
                       table,
                       sql,
                       ...,
                       overwrite = FALSE,
                       temporary = TRUE,
                       unique_indexes = list(),
                       indexes = list(),
                       analyze = TRUE,
                       in_transaction = TRUE) {
  check_table_id(table)
  check_scalar_sql(sql)
  check_bool(overwrite)
  check_bool(temporary)
  check_bool(in_transaction)
  check_dots_used()

  UseMethod("db_compute")
}
#' @export
db_compute.DBIConnection <- function(con,
                                     table,
                                     sql,
                                     ...,
                                     overwrite = FALSE,
                                     temporary = TRUE,
                                     unique_indexes = list(),
                                     indexes = list(),
                                     analyze = TRUE,
                                     in_transaction = FALSE) {
  table <- as_table_path(table, con)
  new <- db_table_temporary(con, table, temporary)
  table <- new$table
  temporary <- new$temporary

  with_transaction(
    con,
    in_transaction,
    "Can't copy query to table {.field {format(table, con = con)}}.",
    {
      table <- dbplyr_save_query(
        con,
        sql,
        table,
        ...,
        temporary = temporary,
        overwrite = overwrite
      )
      create_indexes(con, table, unique_indexes, unique = TRUE)
      create_indexes(con, table, indexes)
      if (analyze) dbplyr_analyze(con, table)
    }
  )

  table
}

#' @export
#' @rdname db-io
db_collect <- function(con, sql, n = -1, warn_incomplete = TRUE, ...) {
  check_dots_used()
  UseMethod("db_collect")
}
#' @export
db_collect.DBIConnection <- function(con, sql, n = -1, warn_incomplete = TRUE, ...) {
  res <- dbSendQuery(con, sql)
  on.exit(dbClearResult(res), add = TRUE)

  out <- dbFetch(res, n = n)
  if (warn_incomplete) {
    res_warn_incomplete(res, "n = Inf")
  }

  out
}


#' @export
#' @importFrom dplyr db_write_table
db_write_table.DBIConnection <- function(con,
                                         table,
                                         types,
                                         values,
                                         temporary = TRUE,
                                         ...,
                                         overwrite = FALSE) {
  check_table_path(table)
  check_character(types, allow_null = TRUE)
  check_named(types)
  check_bool(temporary)
  check_bool(overwrite)

  withCallingHandlers(
    dbWriteTable(
      con,
      name = SQL(table),
      value = values,
      field.types = types,
      temporary = temporary,
      overwrite = overwrite,
      ...,
      row.names = FALSE
    ),
    error = function(cnd) {
      msg <- "Can't write table table {.field {format(table, con = con)}}."
      cli_abort(msg, parent = cnd)
    }
  )

  table
}

# Utility functions ------------------------------------------------------------

create_indexes <- function(con, table, indexes = NULL, unique = FALSE, ...) {
  if (is.null(indexes)) {
    return()
  }

  indexes <- as.list(indexes)
  for (index in indexes) {
    dbplyr_create_index(con, table, index, unique = unique, ...)
  }
}

with_transaction <- function(con,
                             in_transaction,
                             msg,
                             code,
                             call = caller_env(),
                             env = caller_env()) {
  if (in_transaction) {
    dbBegin(con)
    on.exit(dbRollback(con))
  }

  withCallingHandlers(
    code,
    error = function(cnd) {
      cli_abort(msg, parent = cnd, call = call, .envir = env)
    }
  )

  if (in_transaction) {
    on.exit()
    dbCommit(con)
  }
}

#' @export
#' @rdname db-io
db_table_temporary <- function(con, table, temporary, ...) {
  check_dots_used()
  UseMethod("db_table_temporary")
}
#' @export
db_table_temporary.DBIConnection <- function(con, table, temporary, ...) {
  list(
    table = table,
    temporary = temporary
  )
}

#' @importFrom stats setNames update
#' @importFrom utils head tail
#' @importFrom glue glue
#' @importFrom cli cli_abort
#' @importFrom dplyr n join_by
#' @import rlang
#' @import DBI
#' @importFrom tibble tibble as_tibble
#' @importFrom magrittr %>%
#' @importFrom lifecycle deprecated
#' @keywords internal
"_PACKAGE"

# nocov start
.onLoad <- function(...) {
  # lazy S3method() directive for this case only works in 4.3 and later
  s3_register("dplyr::filter", "tbl_lazy")

  methods::setOldClass(c("ident_q", "ident", "character"), ident_q())
  methods::setOldClass(c("ident", "character"), ident())
  methods::setOldClass(c("sql", "character"), sql())

  base_scalar$`%>%` <- magrittr::`%>%`
}

# Silence R CMD check note:
# ** checking whether the namespace can be loaded with stated dependencies ... NOTE
# Warning in .undefineMethod("initialize", Class, classWhere) :
#   no generic function 'initialize' found
#
# I'm not sure why this is necessary, but I suspect it's due to the use of
# setOldClass onLoad
#' @importFrom methods initialize
NULL

# nocov end

#' @export
#' @rdname sql_build
lazy_set_op_query <- function(x,
                              y,
                              type,
                              all,
                              call = caller_env()) {
  check_lazy_query(x, call = call)
  check_lazy_query(y, call = call)
  check_string(type, call = call)
  check_bool(all, call = call)

  lazy_query(
    query_type = "set_op",
    x = x,
    y = y,
    type = type,
    all = all
  )
}

#' @export
print.lazy_set_op_query <- function(x, ...) {
  cat_line("<SQL ", toupper(x$type), ">")

  cat_line("X:")
  cat_line(indent_print(sql_build(x$x, simulate_dbi())))

  cat_line("Y:")
  cat_line(indent_print(sql_build(x$y, simulate_dbi())))
}

#' @export
op_vars.lazy_set_op_query <- function(op) {
  union(op_vars(op$x), op_vars(op$y))
}

#' @export
sql_build.lazy_set_op_query <- function(op, con, ..., sql_options = NULL) {
  # add_op_set_op() ensures that both have same variables
  set_op_query(
    sql_optimise(sql_build(op$x, con, sql_options = sql_options), con),
    sql_optimise(sql_build(op$y, con, sql_options = sql_options), con),
    type = op$type,
    all = op$all
  )
}

#' @export
#' @rdname sql_build
lazy_union_query <- function(x,
                             unions,
                             call = caller_env()) {
  check_lazy_query(x, call = call)

  lazy_query(
    query_type = "union",
    x = x,
    unions = unions
  )
}

#' @export
print.lazy_union_query <- function(x, ...) {
  cat_line("<SQL ", toupper(x$type), ">")

  cat_line("X:")
  cat_line(indent_print(sql_build(x$x, simulate_dbi())))

  cat_line("Y:")
  cat_line(indent_print(sql_build(x$y, simulate_dbi())))
}

#' @export
op_vars.lazy_union_query <- function(op) {
  purrr::reduce(op$unions$table, ~ union(.x, op_vars(.y$lazy_query)), .init = op_vars(op$x))
}

#' @export
sql_build.lazy_union_query <- function(op, con, ..., sql_options = NULL) {
  # add_union() ensures that both have same variables
  unions <- list(
    table = purrr::map(
      op$unions$table,
      ~ sql_optimise(sql_build(.x, con, sql_options = sql_options), con)
    ),
    all = op$unions$all
  )

  union_query(
    sql_optimise(sql_build(op$x, con, sql_options = sql_options), con),
    unions
  )
}

#' Generate SQL from R expressions
#'
#' Low-level building block for generating SQL from R expressions.
#' Strings are escaped; names become bare SQL identifiers. User infix
#' functions have `%` stripped.
#'
#' Using `sql_expr()` in package will require use of [globalVariables()]
#' to avoid `R CMD check` NOTES. This is a small amount of additional pain,
#' which I think is worthwhile because it leads to more readable translation
#' code.
#'
#' @param x A quasiquoted expression
#' @param con Connection to use for escaping. Will be set automatically when
#'   called from a function translation.
#' @param .fn Function name (as string, call, or symbol)
#' @param ... Arguments to function
#' @keywords internal
#' @export
#' @examples
#' con <- simulate_dbi() # not necessary when writing translations
#'
#' sql_expr(f(x + 1), con = con)
#' sql_expr(f("x", "y"), con = con)
#' sql_expr(f(x, y), con = con)
#'
#' x <- ident("x")
#' sql_expr(f(!!x, y), con = con)
#'
#' sql_expr(cast("x" %as% DECIMAL), con = con)
#' sql_expr(round(x) %::% numeric, con = con)
#'
#' sql_call2("+", quote(x), 1, con = con)
#' sql_call2("+", "x", 1, con = con)
sql_expr <- function(x, con = sql_current_con()) {
  x <- enexpr(x)
  x <- replace_expr(x, con = con)
  sql(x)
}

#' @export
#' @rdname sql_expr
sql_call2 <- function(.fn, ..., con = sql_current_con()) {
  fn <- call2(.fn, ...)
  fn <- replace_expr(fn, con = con)
  sql(fn)
}


replace_expr <- function(x, con) {
  if (is.atomic(x) || blob::is_blob(x)) {
    as.character(escape(unname(x), con = con))
  } else if (is.name(x)) {
    as.character(x)
  # } else if (is.call(x) && identical(x[[1]], quote(I))) {
  #   escape(ident(as.character(x[[2]])))
  } else if (is.call(x)) {
    fun <- toupper(as.character(x[[1]]))
    args <- lapply(x[-1], replace_expr, con = con)

    if (is_infix_base(fun)) {
      if (length(args) == 1) {
        paste0(fun, args[[1]])
      } else {
        paste0(args[[1]], " ", fun, " ", args[[2]])
      }
    } else if (is_infix_user(fun)) {
      fun <- substr(fun, 2, nchar(fun) - 1)
      paste0(args[[1]], " ", fun, " ", args[[2]])
    } else if (fun == "(") {
      paste0("(", paste0(args, collapse = ", "), ")")
    } else {
      paste0(fun, "(", paste0(args, collapse = ", "), ")")
    }

  } else {
    x # nocov
  }

}

test_that("atomic vectors are escaped", {
  con <- simulate_dbi()

  expect_equal(sql_expr(2, con = con), sql("2.0"))
  expect_equal(sql_expr("x", con = con), sql("'x'"))
})

test_that("user infix functions have % stripped", {
  con <- simulate_dbi()

  expect_equal(sql_expr(x %like% y, con = con), sql("x LIKE y"))
})

test_that("string function names are not quoted", {
  con <- simulate_dbi()

  f <- "foo"
  expect_equal(sql_expr((!!f)(), con = con), sql("FOO()"))
})

test_that("correct number of parens", {
  con <- simulate_dbi()

  expect_equal(sql_expr((1L), con = con), sql("(1)"))
})

#' SQL set operations
#'
#' These are methods for the dplyr generics `dplyr::intersect()`,
#' `dplyr::union()`, and `dplyr::setdiff()`. They are translated to
#' `INTERSECT`, `UNION`, and `EXCEPT` respectively.
#'
#' @inheritParams left_join.tbl_lazy
#' @param ... Not currently used; provided for future extensions.
#' @param all If `TRUE`, includes all matches in output, not just unique rows.
#' @exportS3Method dplyr::intersect
#' @importFrom dplyr intersect
intersect.tbl_lazy <- function(x, y, copy = FALSE, ..., all = FALSE) {
  lazy_query <- add_set_op(x, y, "INTERSECT", copy = copy, ..., all = all)

  x$lazy_query <- lazy_query
  x
}
#' @importFrom dplyr union
#' @exportS3Method dplyr::union
#' @rdname intersect.tbl_lazy
union.tbl_lazy <- function(x, y, copy = FALSE, ..., all = FALSE) {
  lazy_query <- add_union(x, y, all = all, copy = copy, ...)

  x$lazy_query <- lazy_query
  x
}
#' @export
#' @importFrom dplyr union_all
#' @exportS3Method dplyr::union_all
#' @rdname intersect.tbl_lazy
union_all.tbl_lazy <- function(x, y, copy = FALSE, ...) {
  lazy_query <- add_union(x, y, all = TRUE, copy = copy, ...)

  x$lazy_query <- lazy_query
  x
}
#' @importFrom dplyr setdiff
#' @exportS3Method dplyr::setdiff
#' @rdname intersect.tbl_lazy
setdiff.tbl_lazy <- function(x, y, copy = FALSE, ..., all = FALSE) {
  lazy_query <- add_set_op(x, y, "EXCEPT", copy = copy, ..., all = all)

  x$lazy_query <- lazy_query
  x
}

add_union <- function(x, y, all, copy = FALSE, ..., call = caller_env()) {
  y <- auto_copy(x, y, copy)
  check_set_op_sqlite(x, y, call = call)

  # Ensure each has same variables
  vars <- union(op_vars(x), op_vars(y))

  x_lq <- x$lazy_query
  if (inherits(x_lq, "lazy_union_query")) {
    tmp <- list(lazy_query = x_lq$x)
    class(tmp) <- "tbl_lazy"
    x_lq$x <- fill_vars(tmp, vars)$lazy_query
    x_lq$unions$table <- purrr::map(x_lq$unions$table, function(table) fill_vars(table, vars))
    y <- fill_vars(y, vars)

    x_lq$unions$table <- c(x_lq$unions$table, list(y))
    x_lq$unions$all <- c(x_lq$unions$all, all)

    return(x_lq)
  }

  x <- fill_vars(x, vars)

  unions <- list(
    table = list(fill_vars(y, vars)),
    all = all
  )

  lazy_union_query(
    x$lazy_query,
    unions,
    call = call
  )
}

add_set_op <- function(x, y, type, copy = FALSE, ..., all = FALSE, call = caller_env()) {
  y <- auto_copy(x, y, copy)
  check_set_op_sqlite(x, y, call = call)

  # Ensure each has same variables
  vars <- union(op_vars(x), op_vars(y))
  x <- fill_vars(x, vars)
  y <- fill_vars(y, vars)

  lazy_set_op_query(
    x$lazy_query, y$lazy_query,
    type = type,
    all = all,
    call = call
  )
}

check_set_op_sqlite <- function(x, y, call) {
  if (inherits(x$src$con, "SQLiteConnection")) {
    # LIMIT only part the compound-select-statement not the select-core.
    #
    # https://www.sqlite.org/syntax/compound-select-stmt.html
    # https://www.sqlite.org/syntax/select-core.html

    if (!is.null(x$lazy_query$limit) || !is.null(y$lazy_query$limit)) {
      cli_abort("SQLite does not support set operations on LIMITs", call = call)
    }
  }
}

fill_vars <- function(x, vars) {
  x_vars <- op_vars(x)
  if (identical(x_vars, vars)) {
    return(x)
  }

  new_vars <- lapply(set_names(vars), function(var) {
    if (var %in% x_vars) {
      sym(var)
    } else {
      NA
    }
  })

  transmute(x, !!!new_vars)
}

#' Create, modify, and delete columns
#'
#' These are methods for the dplyr [mutate()] and [transmute()] generics.
#' They are translated to computed expressions in the `SELECT` clause of
#' the SQL query.
#'
#' @inheritParams arrange.tbl_lazy
#' @inheritParams dplyr::mutate
#' @inherit arrange.tbl_lazy return
#' @param .order <[`data-masking`][rlang::args_data_masking]> A selection of
#'   columns to control ordering for window functions within this
#'   [mutate()] call. Use `c()` to order by multiple columns, e.g.
#'   `.order = c(x, y)`. Each column can be wrapped in [desc()] to specify
#'   descending order. Equivalent to calling [window_order()] before and
#'   clearing it after the [mutate()].
#' @param .frame A length-2 numeric vector specifying the bounds for
#'   window function frames. The first element is the lower bound (use `-Inf`
#'   for "unbounded preceding") and the second is the upper bound (use `Inf`
#'   for "unbounded following", `0` for "current row"). Equivalent to calling
#'   [window_frame()] before and clearing it after the [mutate()].
#' @export
#' @importFrom dplyr mutate
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(x = 1:5, y = 5:1)
#' db |>
#'   mutate(a = (x + y) / 2, b = sqrt(x^2L + y^2L)) |>
#'   show_query()
#'
#' # dbplyr automatically creates subqueries as needed
#' db |>
#'   mutate(x1 = x + 1, x2 = x1 * 2) |>
#'   show_query()
#'
#' # `.order` and `.frame` control window functions
#' db <- memdb_frame(g = c(1, 1, 2, 2, 2), x = c(5, 3, 1, 4, 2))
#' db |>
#'   mutate(rolling_sum = sum(x), .by = g, .order = x, .frame = c(-2, 2)) |>
#'   show_query()
mutate.tbl_lazy <- function(
  .data,
  ...,
  .by = NULL,
  .order = NULL,
  .frame = NULL,
  .keep = c("all", "used", "unused", "none"),
  .before = NULL,
  .after = NULL
) {
  keep <- arg_match(.keep)

  by <- compute_by({{ .by }}, .data, by_arg = ".by", data_arg = ".data")
  .data <- set_by_groups(.data, by)

  names_original <- colnames(.data)

  order <- compute_order(.data, {{ .order }})
  old_order <- .data$lazy_query$order_vars
  if (!is.null(order)) {
    .data$lazy_query$order_vars <- order
  }

  frame <- compute_frame(.frame)
  old_frame <- .data$lazy_query$frame
  if (!is.null(frame)) {
    .data$lazy_query$frame <- frame
  }

  layer_info <- get_mutate_layers(.data, ...)
  used <- layer_info$used_vars
  layers <- layer_info$layers

  # The layers may contain `var = quo(NULL)` at this point.
  # They are removed in `add_mutate()`.
  out <- .data
  for (layer in layers) {
    out$lazy_query <- add_mutate(out$lazy_query, layer)
  }

  out <- clear_by_groups(out, by)

  if (!is.null(order)) {
    out$lazy_query$order_vars <- old_order
  }

  if (!is.null(frame)) {
    out$lazy_query$frame <- old_frame
  }

  out <- mutate_relocate(
    out = out,
    before = {{ .before }},
    after = {{ .after }},
    names_original = names_original
  )

  names_new <- layer_info$modified_vars
  names_groups <- by$names

  out <- mutate_keep(
    out = out,
    keep = keep,
    used = used,
    names_new = names_new,
    names_groups = names_groups
  )

  out
}

#' @export
#' @importFrom dplyr transmute
transmute.tbl_lazy <- function(.data, ...) {
  layer_info <- get_mutate_layers(.data, ...)

  for (layer in layer_info$layers) {
    .data$lazy_query <- add_mutate(.data$lazy_query, layer)
  }

  # Retain expression columns in order of their appearance
  cols_expr <- layer_info$modified_vars

  # Retain untouched group variables up front
  cols_group <- group_vars(.data)
  cols_group <- setdiff(cols_group, cols_expr)

  cols_retain <- c(cols_group, cols_expr)

  select(.data, all_of(cols_retain))
}

# helpers -----------------------------------------------------------------

compute_order <- function(.data, order, error_call = caller_env()) {
  order <- enquo(order)
  if (quo_is_null(order)) {
    return(NULL)
  }

  # Expand c() calls to support multiple order columns
  order_expr <- quo_get_expr(order)
  if (is_call(order_expr, "c")) {
    order_exprs <- call_args(order_expr)
  } else {
    order_exprs <- list(order_expr)
  }

  order <- partial_eval_dots(.data, !!!order_exprs, .named = FALSE)
  names(order) <- NULL
  check_window_order_dots(order, arg = "order", call = error_call)
  order
}

compute_frame <- function(frame, error_call = caller_env()) {
  if (is.null(frame)) {
    return(NULL)
  }

  check_frame_range(frame, call = error_call)
  list(range = frame)
}

add_mutate <- function(lazy_query, exprs) {
  # drop NULLs
  exprs <- purrr::discard(exprs, is_null)

  if (is_projection(exprs)) {
    # Special case selecting/renaming/reordering done in mutate
    sel_vars <- purrr::map_chr(exprs, as_string)
    add_select(lazy_query, sel_vars)
  } else if (can_inline_mutate(lazy_query)) {
    lazy_query$select <- new_lazy_select(
      exprs,
      group_vars = op_grps(lazy_query),
      order_vars = op_sort(lazy_query),
      frame = op_frame(lazy_query)
    )
    lazy_query
  } else {
    lazy_select_query(
      x = lazy_query,
      select_operation = "mutate",
      select = exprs
    )
  }
}

# Special optimisation when applied to pure projection() - this is
# conservative and we could expand to any op_select() if combined with
# the logic in get_mutate_layers()
can_inline_mutate <- function(lazy_query) {
  if (!is_lazy_select_query(lazy_query)) {
    return(FALSE)
  }

  select <- lazy_query$select
  if (!is_pure_projection(select$expr, select$name)) {
    return(FALSE)
  }

  if (!lazy_query$select_operation %in% c("select", "mutate")) {
    return(FALSE)
  }

  if (is_true(lazy_query$distinct)) {
    return(FALSE)
  }

  TRUE
}

# Split mutate expressions in independent layers, e.g.
#
# `get_mutate_layers(lf, b = a + 1, c = a - 1, d = b + 1)`
#
# creates two layers:
# 1) a = a, b = a + 1, c = a - 1
#    because `b` and `c` are independent of each other they can be on the
#    same layer
# 2) a = a, b = b, c = c, d = b + 1
#    because `d` depends on `b` it must be on a new layer
get_mutate_layers <- function(.data, ..., error_call = caller_env()) {
  dots <- as.list(enquos(..., .named = TRUE))
  dot_names <- names2(exprs(...))
  was_named <- have_name(exprs(...))

  layer_modified_vars <- character()
  all_modified_vars <- character()
  used_vars <- character()
  all_vars <- op_vars(.data)
  final_removed_cols <- character()

  # Each dot may contain an `across()` expression which can refer to freshly
  # created variables. So, it is necessary to keep track of the current data
  # to partially evaluate the dot.
  cur_data <- .data
  cur_layer <- syms(set_names(all_vars))
  layers <- list()

  for (i in seq_along(dots)) {
    dot <- dots[[i]]
    dot_name <- dot_names[[i]]
    quosures <- partial_eval_quo(
      dot,
      cur_data,
      dot_name,
      error_call = error_call,
      was_named = was_named[[i]]
    )

    if (!is.list(quosures)) {
      quosures <- set_names(list(quosures), names(dots)[[i]])
    }
    quosures <- unclass(quosures)
    cols_result <- get_mutate_dot_cols(quosures, all_vars)

    if (any(cols_result$used_vars %in% layer_modified_vars)) {
      layers <- append(layers, list(cur_layer))

      cur_layer <- syms(set_names(names(cur_layer)))
      layer_modified_vars <- character()
    }

    used_vars <- c(used_vars, cols_result$used_vars)
    layer_modified_vars <- c(layer_modified_vars, cols_result$modified_vars)
    all_modified_vars <- c(all_modified_vars, cols_result$modified_vars)

    cur_layer <- purrr::list_assign(cur_layer, !!!cols_result$cols)
    all_vars <- c(all_vars, setdiff(cols_result$modified_vars, all_vars))

    cols <- set_names(syms(names(cur_layer)))
    cols <- purrr::list_assign(cur_layer, !!!cols_result$cols)
    cur_data$lazy_query <- add_mutate(cur_data$lazy_query, cols)

    removed_cols <- cols_result$removed_cols
    # Track columns that are removed and not later re-added
    final_removed_cols <- union(
      setdiff(final_removed_cols, cols_result$modified_vars),
      removed_cols
    )
    cur_data$lazy_query <- add_select(
      cur_data$lazy_query,
      set_names(setdiff(all_vars, removed_cols))
    )
  }

  # Remove NULL entries and columns that were ultimately set to NULL from final layer
  cur_layer <- purrr::discard(cur_layer, is_null)
  if (length(final_removed_cols) > 0) {
    cur_layer <- cur_layer[!names(cur_layer) %in% final_removed_cols]
    all_vars <- setdiff(all_vars, final_removed_cols)
  }

  list(
    layers = append(layers, list(cur_layer)),
    modified_vars = all_modified_vars,
    used_vars = set_names(all_vars %in% used_vars, all_vars)
  )
}

get_mutate_dot_cols <- function(quosures, all_vars) {
  cols <- list()
  modified_vars <- character()
  used_vars <- character()
  var_is_null <- logical()

  for (k in seq_along(quosures)) {
    cur_expr <- quosures[[k]]
    cur_var <- names(quosures)[[k]]

    if (is_null(cur_expr)) {
      var_is_null[[cur_var]] <- TRUE
      cols[[cur_var]] <- cur_expr
      modified_vars <- setdiff(modified_vars, cur_var)
      next
    }

    var_is_null[[cur_var]] <- FALSE
    cols[[cur_var]] <- cur_expr

    used_vars <- c(used_vars, all_names(cur_expr))
    modified_vars <- c(modified_vars, cur_var)
  }

  list(
    cols = cols,
    used_vars = used_vars,
    modified_vars = modified_vars,
    removed_cols = names2(var_is_null)[var_is_null]
  )
}

mutate_relocate <- function(out, before, after, names_original) {
  before <- enquo(before)
  after <- enquo(after)

  if (quo_is_null(before) && quo_is_null(after)) {
    return(out)
  }

  # Only change the order of completely new columns that
  # didn't exist in the original data
  names <- colnames(out)
  names <- setdiff(names, names_original)

  relocate(
    out,
    all_of(names),
    .before = !!before,
    .after = !!after
  )
}

mutate_keep <- function(out, keep, used, names_new, names_groups) {
  names <- colnames(out)

  if (keep == "all") {
    names_out <- names
  } else {
    names_keep <- switch(
      keep,
      used = names(used)[used],
      unused = names(used)[!used],
      none = character(),
      cli::cli_abort("Unknown value of {.arg keep}.", .internal = TRUE)
    )
    names_out <- intersect(names, c(names_new, names_groups, names_keep))
  }

  select(out, all_of(names_out))
}

#' @export
#' @rdname sql_build
lazy_multi_join_query <- function(x,
                                  joins,
                                  table_names,
                                  vars,
                                  group_vars = op_grps(x),
                                  order_vars = op_sort(x),
                                  frame = op_frame(x),
                                  call = caller_env()) {
  check_lazy_query(x, call = call)

  check_has_names(joins, c("table", "type", "by_x_table_id", "by"))
  check_character(joins$type, call = call)

  check_has_names(table_names, c("name", "from"), call = call)
  check_character(table_names$name, call = call)
  check_character(table_names$from, call = call)

  check_has_names(vars, c("name", "table", "var"), call = call)
  check_character(vars$name, call = call)
  check_integer(vars$table, call = call)
  check_character(vars$var, call = call)

  lazy_query(
    query_type = "multi_join",
    x = x,
    joins = joins,
    table_names = table_names,
    vars = vars,
    group_vars = group_vars,
    order_vars = order_vars,
    frame = frame
  )
}

#' @export
#' @rdname sql_build
lazy_rf_join_query <- function(x,
                               y,
                               type,
                               by,
                               table_names,
                               vars,
                               group_vars = op_grps(x),
                               order_vars = op_sort(x),
                               frame = op_frame(x),
                               call = caller_env()) {
  check_lazy_query(x, call = call)
  check_lazy_query(y, call = call)

  check_character(type, call = call)

  check_has_names(table_names, c("name", "from"), call = call)
  check_character(table_names$name, call = call)
  check_character(table_names$from, call = call)

  check_has_names(vars, c("name", "x", "y"), call = call)
  check_character(vars$name, call = call)
  check_character(vars$x, call = call)
  check_character(vars$y, call = call)

  lazy_query(
    query_type = "rf_join",
    x = x,
    y = y,
    type = type,
    by = by,
    table_names = table_names,
    vars = vars,
    group_vars = group_vars,
    order_vars = order_vars,
    frame = frame
  )
}

join_check_by <- function(by, call) {
  if (!vctrs::vec_is_list(by) && !inherits(by, "dplyr_join_by")) {
    cli_abort("{.arg by} must be a list", .internal = TRUE)
  }
  check_character(by$x, call = call)
  check_character(by$y, call = call)
  if (vctrs::vec_size(by$x) != vctrs::vec_size(by$y)) {
    cli_abort("{.arg by$x} and {.arg by$y} must have the same size", .internal = TRUE)
  }
  check_string(by$x_as, arg = "by$x_as", call = call)
  check_string(by$y_as, arg = "by$y_as", call = call)
}

#' @export
#' @rdname sql_build
lazy_semi_join_query <- function(x,
                                 y,
                                 vars,
                                 anti,
                                 by,
                                 where,
                                 group_vars = op_grps(x),
                                 order_vars = op_sort(x),
                                 frame = op_frame(x),
                                 call = caller_env()) {
  check_lazy_query(x, call = call)
  check_lazy_query(y, call = call)
  check_bool(anti, call = call)
  join_check_by(by, call)

  lazy_query(
    query_type = "semi_join",
    x = x,
    y = y,
    anti = anti,
    by = by,
    vars = vars,
    where = where,
    group_vars = group_vars,
    order_vars = order_vars,
    frame = frame
  )
}

#' @export
print.lazy_semi_join_query <- function(x, ...) {
  cat_line("<SQL ", if (x$anti) "ANTI" else "SEMI", " JOIN>")

  cat_line("By:")
  cat_line(indent(paste0(x$by$x, "-", x$by$y)))

  if (length(x$where)) {
    cat_line("Where:")
    where <- purrr::map_chr(x$where, as_label)
    cat_line(indent(paste0(where)))
  }

  cat_line("X:")
  cat_line(indent_print(sql_build(x$x, simulate_dbi())))

  cat_line("Y:")
  cat_line(indent_print(sql_build(x$y, simulate_dbi())))
}

#' @export
op_vars.lazy_multi_join_query <- function(op) {
  op$vars$name
}
#' @export
op_vars.lazy_rf_join_query <- function(op) {
  op$vars$name
}
#' @export
op_vars.lazy_semi_join_query <- function(op) {
  op$vars$name
}

#' @export
sql_build.lazy_multi_join_query <- function(op, con, ..., sql_options = NULL) {

  table_names_out <- generate_join_table_names(op$table_names, con)

  tables <- set_names(c(list(op$x), op$joins$table), table_names_out)
  table_vars <- purrr::map(tables, op_vars)
  select_sql <- sql_multi_join_vars(
    con,
    op$vars,
    table_vars,
    use_star = sql_options$use_star,
    qualify_all_columns = sql_options$qualify_all_columns
  )

  op$joins$table <- purrr::map(
    op$joins$table,
    ~ sql_optimise(sql_build(.x, con, sql_options = sql_options), con)
  )
  op$joins$by <- purrr::map2(
    op$joins$by, seq_along(op$joins$by),
    function(by, i) {
      by$x_as <- table_names_out[op$joins$by_x_table_id[[i]]]
      by$y_as <- table_names_out[i + 1L]
      by
    }
  )

  multi_join_query(
    x = sql_optimise(sql_build(op$x, con, sql_options = sql_options), con),
    joins = op$joins,
    table_names = table_names_out,
    select = select_sql
  )
}

generate_join_table_names <- function(table_names, con) {
  names <- table_path_name(table_names$name, con)
  table_name_length_max <- max(nchar(names))

  if (length(table_names$name) != 2) {
    table_names_repaired <- vctrs::vec_as_names(names, repair = "unique", quiet = TRUE)
    may_repair_name <- table_names$from != "as"
    names[may_repair_name] <- table_names_repaired[may_repair_name]
  } else{
    names <- join_two_table_alias(names, table_names$from)
  }

  # avoid database aliases exceeding the database-specific maximum length
  abbr_names <- abbreviate(
    names,
    # arbitrarily floor at identifier limit for Postgres backend to avoid unnecessarily truncating reasonable lengths
    # Source: https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
    # "By default, NAMEDATALEN is 64 so the maximum identifier length is 63 bytes."
    minlength = max(table_name_length_max, 63),
    # Explicitly set `strict = FALSE` (the default) to ensure table names are unique;
    # NB: non-zero (but low) chance that name is longer than DB limit
    strict = FALSE,
    named = FALSE,
    # Mitigation for non-zero chance of strings over limit:
    # don't over anchor to the start of the string,
    # since we opt to add qualifiers (...1, _{R/L}HS, etc.) to end of table name
    method = "both.sides"
  )

  as_table_paths(abbr_names, con)
}

#' @export
sql_build.lazy_rf_join_query <- function(op, con, ..., sql_options = NULL) {
  table_names_out <- generate_join_table_names(op$table_names, con)

  vars_classic <- as.list(op$vars)
  vars_classic$all_x <- op_vars(op$x)
  vars_classic$all_y <- op_vars(op$y)

  by <- op$by
  by$x_as <- table_names_out[[1]]
  by$y_as <- table_names_out[[2]]

  select <- sql_rf_join_vars(
    con,
    type = op$type,
    vars = vars_classic,
    x_as = by$x_as,
    y_as = by$y_as,
    use_star = sql_options$use_star,
    qualify_all_columns = sql_options$qualify_all_columns
  )

  join_query(
    sql_optimise(sql_build(op$x, con, sql_options = sql_options), con),
    sql_optimise(sql_build(op$y, con, sql_options = sql_options), con),
    select = select,
    type = op$type,
    by = by,
    suffix = NULL, # it seems like the suffix is not used for rendering
    na_matches = by$na_matches
  )
}

#' @export
sql_build.lazy_semi_join_query <- function(op, con, ..., sql_options = NULL) {
  vars_prev <- op_vars(op$x)
  if (sql_options$use_star &&
      identical(op$vars$var, op$vars$name) &&
      identical(op$vars$var, vars_prev)) {
    vars <- sql_star(con, op$by$x_as)
  } else {
    vars <- ident(set_names(op$vars$var, op$vars$name))
  }

  y_vars <- op_vars(op$y)
  replacements <- purrr::map(y_vars, ~ call2("$", call2("sql", op$by$y_as), sym(.x)))
  where <- purrr::map(
    op$where,
    ~ replace_sym(.x, y_vars, replacements)
  )

  where_sql <- translate_sql_(where, con = con, context = list(clause = "WHERE"))

  semi_join_query(
    sql_optimise(sql_build(op$x, con, sql_options = sql_options), con),
    sql_optimise(sql_build(op$y, con, sql_options = sql_options), con),
    vars = vars,
    anti = op$anti,
    by = op$by,
    where = where_sql,
    na_matches = op$by$na_matches
  )
}

test_that("lazy_semi_join_query() checks arguments", {
  by0 <- list(
    x = "x",
    y = "x",
    x_as = ident("LHS"),
    y_as = ident("RHS"),
    na_matches = "never"
  )
  lmod <- purrr::list_modify

  my_lazy_semi_join_query <- function(
    x = lazy_frame(x = 1, y = 2)$lazy_query,
    y = lazy_frame(x = 1, z = 2)$lazy_query,
    anti = FALSE,
    by = by0
  ) {
    lazy_semi_join_query(
      x = x,
      y = y,
      vars = tibble(name = "x", var = "x"),
      anti = anti,
      by = by
    )
  }

  expect_snapshot(error = TRUE, {
    (my_lazy_semi_join_query(x = lazy_frame(x = 1)))
    (my_lazy_semi_join_query(y = lazy_frame(x = 1)))
  })

  expect_snapshot(error = TRUE, {
    (my_lazy_semi_join_query(by = lmod(by0, x = 1)))
  })

  expect_snapshot(error = TRUE, {
    (my_lazy_semi_join_query(anti = NA))
  })
})

test_that("generated sql doesn't change unexpectedly", {
  lf <- lazy_frame(x = 1, y = 2)
  expect_snapshot(semi_join(lf, lf))
  expect_snapshot(anti_join(lf, lf))
})

test_that("sql_on query doesn't change unexpectedly", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 3)

  expect_snapshot(semi_join(lf1, lf2, sql_on = "LHS.y < RHS.z"))
  expect_snapshot(anti_join(lf1, lf2, sql_on = "LHS.y < RHS.z"))
})

test_that("lazy_semi_join_query() checks arguments", {
  by0 <- list(x = "x", y = "x", x_as = ident("LHS"), y_as = ident("RHS"))
  lmod <- purrr::list_modify

  my_lazy_semi_join_query <- function(x = lazy_frame(x = 1, y = 2)$lazy_query,
                                      y = lazy_frame(x = 1, z = 2)$lazy_query,
                                      anti = FALSE,
                                      by = by0,
                                      na_matches = "never") {
    lazy_semi_join_query(
      x = x,
      y = y,
      anti = anti,
      by = by,
      na_matches = na_matches
    )
  }

  expect_snapshot(error = TRUE, {
    (my_lazy_semi_join_query(x = lazy_frame(x = 1)))
    (my_lazy_semi_join_query(y = lazy_frame(x = 1)))
  })

  expect_snapshot(error = TRUE, {
    (my_lazy_semi_join_query(by = lmod(by0, x = 1)))
  })

  expect_snapshot(error = TRUE, {
    (my_lazy_semi_join_query(anti = NA))
    (my_lazy_semi_join_query(na_matches = "sometimes"))
  })
})

#' Expand SQL tables to include all possible combinations of values
#'
#' @description
#' This is a method for the [tidyr::expand] generics. It doesn't sort the
#' result explicitly, so the order might be different to what `expand()`
#' returns for data frames.
#'
#' @param data A lazy data frame backed by a database query.
#' @param ... Specification of columns to expand. See [tidyr::expand] for
#' more details.
#' @inheritParams tibble::as_tibble
#' @inherit arrange.tbl_lazy return
#' @examplesIf rlang::is_installed("tidyr", version = "1.0.0")
#' fruits <- memdb_frame(
#'   type   = c("apple", "orange", "apple", "orange", "orange", "orange"),
#'   year   = c(2010, 2010, 2012, 2010, 2010, 2012),
#'   size = c("XS", "S",  "M", "S", "S", "M"),
#'   weights = rnorm(6)
#' )
#'
#' # All possible combinations ---------------------------------------
#' fruits %>% tidyr::expand(type)
#' fruits %>% tidyr::expand(type, size)
#'
#' # Only combinations that already appear in the data ---------------
#' fruits %>% tidyr::expand(nesting(type, size))
expand.tbl_lazy <- function(data, ..., .name_repair = "check_unique") {
  dots <- purrr::discard(quos(...), quo_is_null)

  if (is_empty(dots)) {
    cli_abort("Must supply variables in `...`")
  }

  distinct_tbl_vars <- purrr::map(dots, extract_expand_dot_vars, call = current_env())

  # now that `nesting()` has been unpacked resolve name conflicts
  out_names <- names(exprs_auto_name(purrr::flatten(distinct_tbl_vars)))
  out_names_repaired <- vctrs::vec_as_names(out_names, repair = .name_repair)

  ns <- lengths(distinct_tbl_vars)
  indices <- vctrs::vec_rep_each(seq_along(distinct_tbl_vars), ns)
  out_names_list <- vctrs::vec_split(out_names_repaired, indices)$val

  distinct_tables <- purrr::map2(
    distinct_tbl_vars, out_names_list,
    ~ {
      args <- set_names(.x, .y)
      distinct(data, !!!args)
    }
  )

  by <- group_vars(data)
  if (is_empty(by)) {
    # We want a cross join if `by` is empty which is done with
    # a `full_join()` and `by = character()`
    purrr::reduce(distinct_tables, full_join, by = character())
  } else {
    # In this case a `full_join()` and a `left_join()` produce the same result
    # but a `left_join()` produces much nicer SQL for SQLite
    purrr::reduce(distinct_tables, left_join, by = group_vars(data))
  }
}

extract_expand_dot_vars <- function(dot, call) {
  # ugly hack to deal with `nesting()`
  if (!quo_is_call(dot, name = "nesting", ns = c("", "tidyr"))) {
    return(list(quo_get_expr(dot)))
  }

  x_expr <- quo_get_expr(dot)
  args <- call_args(x_expr)

  repair <- args[[".name_repair"]] %||% "check_unique"
  args[[".name_repair"]] <- NULL

  args_named <- exprs_auto_name(args)
  nms <- vctrs::vec_as_names(names(args_named), repair = repair, call = call)
  exprs(!!!set_names(args_named, nms))
}

#' Complete a SQL table with missing combinations of data
#'
#' Turns implicit missing values into explicit missing values. This is a method
#' for the [tidyr::complete()] generic.
#'
#' @inheritParams expand.tbl_lazy
#' @param fill A named list that for each variable supplies a single value to
#' use instead of NA for missing combinations.
#'
#' @inherit arrange.tbl_lazy return
#'
#' @examplesIf rlang::is_installed("tidyr", version = "1.0.0")
#' df <- memdb_frame(
#'   group = c(1:2, 1),
#'   item_id = c(1:2, 2),
#'   item_name = c("a", "b", "b"),
#'   value1 = 1:3,
#'   value2 = 4:6
#' )
#'
#' df %>% tidyr::complete(group, nesting(item_id, item_name))
#'
#' # You can also choose to fill in missing values
#' df %>% tidyr::complete(group, nesting(item_id, item_name), fill = list(value1 = 0))
complete.tbl_lazy <- function(data, ..., fill = list()) {
  full <- tidyr::expand(data, ...)

  full <- full_join(full, data, by = colnames(full))
  tidyr::replace_na(full, replace = fill)
}

#' Replace NAs with specified values
#'
#' This is a method for the [tidyr::replace_na()] generic.
#'
#' @param data A pair of lazy data frame backed by database queries.
#' @param replace A named list of values, with one value for each column that
#' has NA values to be replaced.
#' @param ... Unused; included for compatibility with generic.
#'
#' @inherit arrange.tbl_lazy return
#'
#' @examplesIf rlang::is_installed("tidyr", version = "1.0.0")
#' df <- memdb_frame(x = c(1, 2, NA), y = c("a", NA, "b"))
#' df %>% tidyr::replace_na(list(x = 0, y = "unknown"))
replace_na.tbl_lazy <- function(data, replace = list(), ...) {
  stopifnot(is_list(replace))
  stopifnot(is_empty(replace) || is_named(replace))
  replace <- replace[names(replace) %in% colnames(data)]

  if (is_empty(replace)) {
    return(data)
  }

  coalesce_expr <- purrr::imap(
    replace,
    function(value, col) {
      quo(coalesce(!!sym(col), !!value))
    }
  )

  mutate(data, !!!coalesce_expr)
}

globalVariables(c("coalesce", "expand", "replace_na"))

#' SQL helpers for window functions
#'
#' @description
#' These functions help you create custom window SQL translations when
#' implementing a new backend. They are typically used within [sql_translator()]
#' to define how R window functions should be translated to SQL.
#'
#' * `win_over()` makes it easy to generate the window function specification.
#' * `win_absent()`, `win_rank()`, `win_aggregate()`, and `win_cumulative()`
#'   provide helpers for constructing common types of window functions.
#' * `win_current_group()` and `win_current_order()` allow you to access
#'   the grouping and order context set up by [group_by()] and [arrange()].
#'
#' @param expr The window expression.
#' @param partition Variables to partition over.
#' @param order Variables to order by.
#' @param frame A numeric vector of length two defining the frame.
#' @param f The name of an SQL function as a string.
#' @param empty_order A logical value indicating whether to order by NULL if
#'   `order` is not specified.
#' @param con A [sql_dialect] object or database connection. Connections are
#'   supported for backward compatibility.
#' @family SQL translation helpers
#' @name sql_translation_window
#' @export
#' @examples
#' con <- dialect_ansi()
#'
#' win_over(sql("avg(x)"), con = con)
#' win_over(sql("avg(x)"), "y", con = con)
#' win_over(sql("avg(x)"), order = "y", con = con)
#' win_over(sql("avg(x)"), order = c("x", "y"), con = con)
#' win_over(sql("avg(x)"), frame = c(-Inf, 0), order = "y", con = con)
win_over <- function(
  expr,
  partition = NULL,
  order = NULL,
  frame = NULL,
  con = sql_current_con()
) {
  if (length(partition) > 0) {
    partition <- as_ident_or_sql(partition)
    partition <- sql_glue2(con, "PARTITION BY {partition}")
  }

  if (length(order) > 0) {
    order <- as_ident_or_sql(order)
    order <- sql_glue2(con, "ORDER BY {order}")
  }
  if (length(frame) > 0) {
    if (length(order) == 0) {
      cli::cli_warn(c(
        "Windowed expression `{expr}` does not have explicit order.",
        i = "Please use {.fun arrange}, {.fun window_order}, or {.arg .order} to make deterministic."
      ))
    }

    if (is.numeric(frame)) {
      frame <- rows(frame[1], frame[2])
    }
    frame <- sql_glue2(con, "ROWS {.sql frame}")
  }

  over <- sql_collapse(c(partition, order, frame), parens = TRUE)

  if (sql_context$register_windows) {
    win_register(over)
  } else {
    over <- win_get(over, con)
  }

  sql_glue2(con, "{expr} OVER {over}")
}

win_register_activate <- function() {
  sql_context$register_windows <- TRUE
}

win_register_deactivate <- function() {
  sql_context$register_windows <- FALSE
}

win_register <- function(over) {
  sql_context$windows <- sql(c(sql_context$windows, over))
}

win_register_names <- function() {
  windows <- sql_context$windows %||% character()

  window_count <- vctrs::vec_count(windows, sort = "location")
  window_count <- vctrs::vec_slice(window_count, window_count$count > 1)
  window_count$name <- paste0(
    "win",
    seq_along(window_count$key),
    recycle0 = TRUE
  )
  window_count$key <- window_count$key

  sql_context$window_names <- window_count
  window_count
}

win_get <- function(over, con) {
  windows <- sql_context$window_names

  if (vctrs::vec_in(over, windows$key)) {
    id <- vctrs::vec_match(over, windows$key)
    sql_escape_ident(con, windows$name[[id]])
  } else {
    over
  }
}

win_reset <- function() {
  sql_context$window_names <- NULL
  sql_context$windows <- sql()
}

rows <- function(from = -Inf, to = 0) {
  if (from >= to) {
    cli_abort("{.arg from} ({from}) must be less than {.arg to} ({to})")
  }

  dir <- function(x) if (x < 0) "PRECEDING" else "FOLLOWING"
  val <- function(x) if (is.finite(x)) as.integer(abs(x)) else "UNBOUNDED"
  bound <- function(x) {
    if (x == 0) {
      sql("CURRENT ROW")
    } else {
      sql(paste(val(x), dir(x)))
    }
  }

  if (to == 0) {
    sql(bound(from))
  } else {
    sql_glue("BETWEEN {bound(from)} AND {bound(to)}")
  }
}

#' @rdname sql_translation_window
#' @export
win_rank <- function(f, empty_order = FALSE) {
  check_bool(empty_order)

  function(order = NULL) {
    group <- win_current_group()
    order <- unwrap_order_expr(enexpr(order), f = f)
    con <- sql_current_con()

    if (!is_null(order)) {
      order_over <- translate_sql_(order, con = con)

      # Ensuring NULLs get a NULL rank in SQL requires two pieces:
      # * We need to ensure that a NULL gets a rank of NULL. We can't do this
      #   inside of RANK() so we use a CASE WHEN outside of it
      # * We need to ensure those NULLs don't affect the ranks of other
      #   values. We do this by PARTITIONING the NULLs into their own bucket
      #   (NULLS LAST is not yet widely supported)

      order_symbols <- purrr::map_if(
        order,
        \(x) is_call(x, "desc", n = 1L),
        \(x) call_args(x)[[1L]]
      )

      is_na_exprs <- purrr::map(order_symbols, \(sym) expr(is.na(!!sym)))
      is_missing <- purrr::reduce(is_na_exprs, function(lhs, rhs) {
        call2("|", lhs, rhs)
      })
      # e.g. X IS NULL OR Y IS NULL
      not_missing <- call2("!", is_missing)

      cond <- translate_sql(
        (case_when(!!is_missing ~ 1L, TRUE ~ 0L)),
        con = con
      )
      group <- sql(group, cond)
    } else {
      order_over <- win_current_order()
      if (empty_order & is_empty(order_over)) {
        # For certain backends (e.g., Snowflake), need a subquery that returns
        # a constant to work if no ordering is specified
        # https://stackoverflow.com/questions/44105691/row-number-without-order-by
        order_over <- sql("(SELECT NULL)")
      }
    }

    rank_sql <- win_over(
      sql_glue("{.sql f}()"),
      partition = group,
      order = order_over,
      frame = win_current_frame()
    )

    if (is_null(order)) {
      rank_sql
    } else {
      translate_sql(case_when(!!not_missing ~ !!rank_sql), con = con)
    }
  }
}

#' @rdname sql_translation_window
#' @export
win_aggregate <- function(f) {
  function(x, na.rm = FALSE) {
    sql_check_na_rm(na.rm)
    frame <- win_current_frame()

    win_over(
      sql_glue("{.sql f}({x})"),
      partition = win_current_group(),
      order = if (!is.null(frame)) win_current_order(),
      frame = frame
    )
  }
}

#' @rdname sql_translation_window
#' @export
win_aggregate_2 <- function(f) {
  function(x, y) {
    frame <- win_current_frame()

    win_over(
      sql_glue("{.sql f}({x}, {y})"),
      partition = win_current_group(),
      order = if (!is.null(frame)) win_current_order(),
      frame = frame
    )
  }
}

#' @rdname sql_translation_window
#' @usage NULL
#' @export
win_recycled <- win_aggregate


#' @rdname sql_translation_window
#' @export
win_cumulative <- function(f) {
  function(x, order = NULL) {
    win_over(
      sql_glue("{.sql f}({x})"),
      partition = win_current_group(),
      order = order %||% win_current_order(),
      frame = c(-Inf, 0)
    )
  }
}

sql_nth <- function(
  x,
  n,
  order_by = NULL,
  na_rm = FALSE,
  ignore_nulls = c("inside", "outside", "bool"),
  error_call = caller_env()
) {
  check_bool(na_rm, call = error_call)
  ignore_nulls <- arg_match(ignore_nulls, error_call = error_call)
  con <- sql_current_con()

  frame <- win_current_frame()
  args <- translate_sql(!!x, con = con)
  if (n == 1) {
    f <- "FIRST_VALUE"
  } else if (is.infinite(n) && n > 0) {
    f <- "LAST_VALUE"
    frame <- frame %||% c(-Inf, Inf)
  } else {
    f <- "NTH_VALUE"
    if (is.numeric(n)) {
      n <- as.integer(n)
    }
    args <- sql_glue2(con, "{args}, {n}")
  }

  if (na_rm) {
    if (ignore_nulls == "inside") {
      sql_expr <- "{.sql f}({args} IGNORE NULLS)"
    } else if (ignore_nulls == "outside") {
      sql_expr <- "{.sql f}({args}) IGNORE NULLS"
    } else {
      sql_expr <- "{.sql f}({args}, TRUE)"
    }
  } else {
    sql_expr <- "{.sql f}({args})"
  }

  win_over(
    sql_glue2(con, sql_expr),
    win_current_group(),
    order_by %||% win_current_order(),
    frame
  )
}

#' @rdname sql_translation_window
#' @export
win_absent <- function(f) {
  force(f)

  function(...) {
    cli_abort("Window function {.fun {f}} is not supported by this database.")
  }
}


# API to set default partitioning etc -------------------------------------

# Use a global variable to communicate state of partitioning between
# tbl and sql translator. This isn't the most amazing design, but it keeps
# things loosely coupled and is easy to understand.
sql_context <- new.env(parent = emptyenv())
sql_context$group_by <- NULL
sql_context$order_by <- NULL
sql_context$con <- NULL
# Used to carry additional information needed for special cases
sql_context$context <- list()

sql_context$register_windows <- FALSE
sql_context$windows <- NULL
sql_context$window_names <- NULL


set_current_con <- function(con) {
  old <- sql_context$con
  sql_context$con <- con
  invisible(old)
}

set_win_current_group <- function(vars) {
  check_character(vars, allow_null = TRUE)

  old <- sql_context$group_by
  sql_context$group_by <- vars
  invisible(old)
}

set_win_current_order <- function(vars) {
  check_character(vars, allow_null = TRUE)

  old <- sql_context$order_by
  sql_context$order_by <- vars
  invisible(old)
}

set_win_current_frame <- function(frame) {
  check_frame_range(frame)

  old <- sql_context$frame
  sql_context$frame <- frame
  invisible(old)
}
#' @export
#' @rdname sql_translation_window
win_current_group <- function() sql_context$group_by

#' @export
#' @rdname sql_translation_window
win_current_order <- function() sql_context$order_by

#' @export
#' @rdname sql_translation_window
win_current_frame <- function() sql_context$frame

# Not exported, because you shouldn't need it
sql_current_con <- function() {
  sql_context$con
}

local_con <- function(con, frame = caller_env()) {
  check_con(con)

  old <- sql_context$con
  withr::defer(sql_context$con <- old, envir = frame)

  sql_context$con <- con
  invisible(old)
}

# Functions to manage information for special cases
set_current_context <- function(context) {
  old <- sql_context$context
  sql_context$context <- context
  invisible(old)
}

sql_current_context <- function() sql_context$context

local_context <- function(x, env = parent.frame()) {
  old <- set_current_context(x)
  withr::defer(set_current_context(old), envir = env)
  invisible()
}

# Where translation -------------------------------------------------------

uses_window_fun <- function(x, con, lq) {
  check_list(x)

  calls <- unlist(lapply(x, all_calls))
  any(calls %in% window_funs(con))
}

window_funs <- function(con = simulate_dbi()) {
  env_names(sql_translation(con)$window)
}

is_aggregating <- function(x, non_group_cols, agg_f) {
  if (is_symbol(x)) {
    xc <- as_name(x)
    return(!(xc %in% non_group_cols))
  }

  if (is_call(x)) {
    fname <- as.character(x[[1]])
    if (fname %in% agg_f) {
      return(TRUE)
    }

    return(all(purrr::map_lgl(x[-1], is_aggregating, non_group_cols, agg_f)))
  }

  return(TRUE)
}

#' @noRd
#' @examples
#' translate_window_where(quote(1))
#' translate_window_where(quote(x))
#' translate_window_where(quote(x == 1))
#' translate_window_where(quote(x == 1 && y == 2))
#' translate_window_where(quote(n() > 10))
#' translate_window_where(quote(rank() > cumsum(AB)))
translate_window_where <- function(expr, window_funs = NULL) {
  window_funs <- window_funs %||% window_funs()

  switch(
    typeof(expr),
    logical = ,
    integer = ,
    double = ,
    complex = ,
    character = ,
    string = ,
    symbol = window_where(expr, list()),
    language = {
      if (is_formula(expr)) {
        translate_window_where(f_rhs(expr), window_funs)
      } else if (is_call(expr, name = window_funs)) {
        name <- unique_column_name()
        window_where(sym(name), set_names(list(expr), name))
      } else {
        args <- lapply(
          expr[-1],
          translate_window_where,
          window_funs = window_funs
        )
        expr <- call2(node_car(expr), splice(lapply(args, "[[", "expr")))

        window_where(
          expr = expr,
          comp = unlist(lapply(args, "[[", "comp"), recursive = FALSE)
        )
      }
    },
    cli_abort("Unknown type: {typeof(expr)}") # nocov
  )
}

#' @noRd
#' @examples
#' translate_window_where_all(list(quote(x == 1), quote(n() > 2)))
#' translate_window_where_all(list(quote(cumsum(x) == 10), quote(n() > 2)))
translate_window_where_all <- function(x, window_funs = NULL) {
  window_funs <- window_funs %||% window_funs()
  out <- lapply(x, translate_window_where, window_funs = window_funs)

  list(
    expr = unlist(lapply(out, "[[", "expr"), recursive = FALSE),
    comp = unlist(lapply(out, "[[", "comp"), recursive = FALSE)
  )
}

window_where <- function(expr, comp) {
  stopifnot(is.call(expr) || is.name(expr) || is.atomic(expr) || is.null(expr))
  check_list(comp)

  list(
    expr = expr,
    comp = comp
  )
}

#' Miscellaneous database generics
#'
#' @description
#' These are used when creating a new dbplyr backend and should generally not
#' be called directly.
#'
#' * `db_connection_describe()` provides a short string describing the
#'   database connection, helping users tell which database a table comes
#'   from. It should be a single line, and ideally less than 60 characters wide.
#'
#' * `dbplyr_edition()` declares which version of the dbplyr API you want.
#'
#' * `db_col_types()` introspects an existing table and returns a named
#'   character vector mapping column names to their database-native SQL types.
#'   The result is suitable for use as the `field.types` argument of
#'   [DBI::dbWriteTable()], allowing dbplyr to preserve column types when
#'   copying data with `rows_*()`.
#'
#'   Returns `NULL` if the backend does not implement introspection.
#'
#' @param con A database connection.
#' @param table A table identifier, or `NULL`. Use a string to refer to tables
#'   in the current schema/catalog or `I()` to refer to tables in other
#'   schemas/catalogs.
#' @param call The execution environment of a currently running function, used
#'   to report errors.
#' @family generic
#' @keywords internal
#' @name db-misc
#' @aliases NULL
NULL

#' @export
#' @rdname db-misc
db_connection_describe <- function(con, ...) {
  check_dots_used()
  UseMethod("db_connection_describe")
}
# nocov start
#' @export
db_connection_describe.DBIConnection <- function(con, ...) {
  class(con)[[1]]
}
# nocov end

#' @rdname db-misc
#' @export
sql_join_suffix <- function(con, suffix, ...) {
  UseMethod("sql_join_suffix", sql_dialect(con))
}
#' @export
sql_join_suffix.DBIConnection <- function(con, suffix, ...) {
  suffix %||% c(".x", ".y")
}

#' @export
sql_join_suffix.sql_dialect <- sql_join_suffix.DBIConnection

#' @rdname db-misc
#' @export
db_sql_render <- function(con, sql, ..., cte = FALSE, sql_options = NULL) {
  check_bool(cte)
  if (cte) {
    lifecycle::deprecate_warn(
      when = "2.4.0",
      what = "db_sql_render(cte)",
      with = I("db_sql_render(sql_options = sql_options(cte = TRUE))")
    )
    sql_options <- sql_options %||% sql_options(cte = TRUE)
  }

  sql_options <- sql_options %||% sql_options()

  out <- db_sql_render_dispatch(con, sql, ..., sql_options = sql_options)
  the$last_sql <- out
  return(out)

  # Unusued declaration for roxygen2
  UseMethod("db_sql_render")
}

db_sql_render_dispatch <- function(con, sql, ..., sql_options) {
  UseMethod("db_sql_render")
}
#' @export
db_sql_render.DBIConnection <- function(
  con,
  sql,
  ...,
  cte = FALSE,
  sql_options = NULL
) {
  sql_render(sql, con = con, ..., sql_options = sql_options)
}

#' @export
db_sql_render.sql_dialect <- db_sql_render.DBIConnection

#' @rdname db-misc
#' @export
db_col_types <- function(con, table, call) {
  if (is_null(table)) {
    return(NULL)
  }

  UseMethod("db_col_types")
}

#' @export
db_col_types.TestConnection <- function(con, table, call) {
  NULL
}

#' @export
db_col_types.DBIConnection <- function(con, table, call) {
  NULL
}
# add a default method so that packages that haven't implemented `db_col_types()`
# keep working, e.g. {Pool}
#' @export
db_col_types.default <- function(con, table, call) {
  NULL
}

#' Options for generating SQL
#'
#' Customise SQL generation with these three options.
#'
#' @param cte If `FALSE`, the default, subqueries are used. If `TRUE` common
#'   table expressions are used.
#' @param use_star If `TRUE`, the default, `*` is used to select all columns of
#'   a table. If `FALSE` all columns are explicitly selected.
#' @param qualify_all_columns If `FALSE`, the default, columns are only
#'   qualified with the table they come from if the same column name appears in
#'   multiple tables.
#'
#' @return A <dbplyr_sql_options> object.
#' @export
#'
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#' lf1 <- lazy_frame(key = 1, a = 1, b = 2)
#' lf2 <- lazy_frame(key = 1, a = 1, c = 3)
#'
#' result <- left_join(lf1, lf2, by = "key") |>
#'   filter(c >= 3)
#'
#' show_query(result)
#' sql_options <- sql_options(cte = TRUE, qualify_all_columns = TRUE)
#' show_query(result, sql_options = sql_options)
sql_options <- function(
  cte = FALSE,
  use_star = TRUE,
  qualify_all_columns = FALSE
) {
  check_bool(cte)
  check_bool(use_star)
  check_bool(qualify_all_columns)

  data <- list(
    cte = cte,
    use_star = use_star,
    qualify_all_columns = qualify_all_columns
  )
  class(data) <- "dbplyr_sql_options"
  data
}

as_sql_options <- function(
  x,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  if (is.null(x)) {
    x <- sql_options()
    return(x)
  }

  if (!inherits(x, "dbplyr_sql_options")) {
    stop_input_type(
      x = x,
      what = I("an object created by `sql_options()`"),
      arg = error_arg,
      call = error_call
    )
  }

  x
}

#' @export
print.dbplyr_sql_options <- function(x, ...) {
  if (x$cte) {
    cte <- "Use CTE"
  } else {
    cte <- "Use subqueries"
  }

  if (x$use_star) {
    star <- "Use {.val SELECT *} where possible"
  } else {
    star <- "Explicitly select all columns"
  }

  if (x$qualify_all_columns) {
    join_table_name <- "Qualify all columns"
  } else {
    join_table_name <- "Qualify only ambiguous columns"
  }

  msg <- c(cte, star, join_table_name)
  msg <- set_names(msg, "*")
  cli::cli_inform(msg)
}

#' @rdname db-misc
#' @export
dbplyr_edition <- function(con) {
  UseMethod("dbplyr_edition")
}
#' @export
dbplyr_edition.default <- function(con) {
  1L
}

check_2ed <- function(con, call = caller_env()) {
  if (is_sql_dialect(con)) {
    return(TRUE)
  }

  edition <- dbplyr_edition(con)
  if (edition >= 2) {
    return(invisible())
  }

  class <- class(con)[[1]]
  cli_abort(
    c(
      "<{class}> uses dbplyr's 1st edition interface, which is no longer supported.",
      i = "Please contact the maintainer of the package for a solution."
    ),
    call = call
  )
}

test_that("correctly inlines across all verbs", {
  lf <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, y = 2)

  # single table verbs
  expect_selects(lf |> arrange(x) |> union(lf2), 2)
  expect_selects(lf |> distinct() |> union(lf2), 2)
  expect_selects(lf |> filter(x == 1) |> union(lf2), 2)
  expect_selects(lf |> head(1) |> union(lf2), 2)
  expect_selects(lf |> mutate(z = x + 1) |> select(-z) |> union(lf2), 2)
  expect_selects(lf |> select(x, y) |> union(lf2), 2)
  expect_selects(lf |> summarise(x = mean(x), y = mean(y)) |> union(lf2), 2)

  # two table verbs
  lf3 <- lazy_frame(x = 1)
  expect_selects(lf |> left_join(lf3, by = "x") |> union(lf2), 2)
  expect_selects(lf |> right_join(lf3, by = "x") |> union(lf2), 2)
  expect_selects(lf |> semi_join(lf3, by = "x") |> union(lf2), 3)
  expect_selects(lf |> union(lf2) |> union(lf2), 3)
})

test_that("column order is matched", {
  df1 <- local_memdb_frame(x = 1, y = 2)
  df2 <- local_memdb_frame(y = 1, x = 2)

  out <- collect(union(df1, df2))
  expect_equal(out, tibble(x = c(1, 2), y = c(2, 1)))
})

test_that("missing columns filled with NULL", {
  df1 <- local_memdb_frame(x = 1)
  df2 <- local_memdb_frame(y = 2)

  out <- collect(union_all(df1, df2))
  expect_equal(out, tibble(x = c(1, NA), y = c(NA, 2)))
})

# SQL generation ----------------------------------------------------------

test_that("basic set ops work as expected", {
  db1 <- local_memdb_frame(x = c(1, 3))
  db2 <- local_memdb_frame(x = c(1, 2))

  expect_equal(sort(pull(union(db1, db2))), c(1, 2, 3))
  expect_equal(sort(pull(union_all(db1, db2))), c(1, 1, 2, 3))
  expect_equal(sort(pull(intersect(db1, db2))), 1)
  expect_equal(sort(pull(setdiff(db1, db2))), 3)
  expect_equal(sort(pull(setdiff(db2, db1))), 2)
})

test_that("basic set ops generate expected SQL", {
  db1 <- lazy_frame(x = 1, .name = "db1")
  db2 <- lazy_frame(x = 1, .name = "db2")

  expect_snapshot({
    union(db1, db2)
    union_all(db1, db2)
    intersect(db1, db2)
    setdiff(db1, db2)
  })
})

test_that("can combine multiple unions in one query", {
  lf1 <- lazy_frame(x = 1, y = 1, .name = "lf1")
  lf2 <- lazy_frame(y = 1, .name = "lf2")
  lf3 <- lazy_frame(z = 1, .name = "lf3")

  with_cte <- sql_options(cte = TRUE)
  lf_union <- lf1 |> union_all(lf2) |> union(lf3)

  expect_snapshot(lf_union |> show_query())
  # and with cte
  expect_snapshot(lf_union |> show_query(sql_options = with_cte))

  out <- lf_union |> mutate(a = x + y) |> sql_build()
  expect_equal(out$select, sql("*", "\"x\" + \"y\" AS \"a\""))
})

test_that("set ops correctly quote reused queries in CTEs", {
  lf <- lazy_frame(x = 1, .name = "lf") |> mutate(y = x + 1)
  expect_snapshot(
    union_all(lf, lf) |> show_query(sql_options = sql_options(cte = TRUE))
  )
})

test_that("SQLite warns if set op attempted when tbl has LIMIT", {
  mf <- local_memdb_frame(x = 1:2)
  m1 <- head(mf, 1)

  expect_error(union(mf, m1), "does not support")
  expect_error(union(m1, mf), "does not support")
})

test_that("but it can work with another backend", {
  con <- test_postgres()
  db <- local_db_table(con, tibble(x = 1:2), "db")

  out <- collect(head(union_all(db, db), 2))
  expect_equal(nrow(out), 2)
})

test_that("intersect works with copy = 'temp-table'", {
  df1 <- local_memdb_frame(x = 1:3)
  df2 <- tibble(x = 2:4)

  out <- intersect(df1, df2, copy = "temp-table") |> collect()
  expect_equal(out, tibble(x = 2:3))
})

test_that("intersect works with copy = 'inline'", {
  df1 <- local_memdb_frame(x = 1:3)
  df2 <- tibble(x = 2:4)

  out <- intersect(df1, df2, copy = "inline") |> collect()
  expect_equal(out, tibble(x = 2:3))
})

test_that("adbc_vendor_to_dialect maps known vendors", {
  expect_s3_class(adbc_vendor_to_dialect("SQLite"), "sql_dialect_sqlite")
  expect_s3_class(adbc_vendor_to_dialect("PostgreSQL"), "sql_dialect_postgres")
  expect_s3_class(adbc_vendor_to_dialect("MySQL"), "sql_dialect_mysql")
  expect_s3_class(adbc_vendor_to_dialect("TiDB"), "sql_dialect_mysql")
  expect_s3_class(adbc_vendor_to_dialect("MariaDB"), "sql_dialect_mariadb")
})

test_that("adbc_vendor_to_dialect is case-insensitive", {
  expect_s3_class(adbc_vendor_to_dialect("sqlite"), "sql_dialect_sqlite")
  expect_s3_class(adbc_vendor_to_dialect("POSTGRESQL"), "sql_dialect_postgres")
})

test_that("adbc_vendor_to_dialect falls back to ODBC for unknown vendors", {
  expect_s3_class(adbc_vendor_to_dialect("Unknown"), "sql_dialect_odbc")
})

test_that("sql_dialect.AdbiConnection queries the live driver", {
  skip_if_not_installed("adbi")
  skip_if_not_installed("adbcdrivermanager")

  con <- DBI::dbConnect(adbi::adbi("adbcsqlite"), uri = ":memory:")
  on.exit(DBI::dbDisconnect(con))

  expect_s3_class(sql_dialect(con), "sql_dialect_sqlite")
  expect_equal(dbplyr_edition(con), 2L)

  DBI::dbWriteTable(con, "df", data.frame(x = 1:3, y = c("a", "b", "c")))
  out <- tbl(con, "df") |> dplyr::filter(x > 1) |> dplyr::collect()
  expect_equal(out, tibble::tibble(x = 2:3, y = c("b", "c")))
})

#' DB2 backend
#'
#' @description
#' This backend supports IBM DB2 databases, typically accessed via ODBC.
#' Use `dialect_db2()` with `lazy_frame()` to see simulated SQL without
#' connecting to a live database.
#'
#' Key differences for this backend are:
#'
#' * Uses `FETCH FIRST n ROWS ONLY` instead of `LIMIT n`
#' * Uses double quotes for identifier quoting
#' * `paste()` uses `||`
#' * DB2-specific data type names for casts (e.g. `VARCHAR(255)`, `DOUBLE`)
#' * Date component extraction via `YEAR()`, `MONTH()`, ..., `DAYOFYEAR()`,
#'   `DAYOFWEEK()`, `QUARTER()`, `WEEK()`
#' * `str_flatten()` uses `LISTAGG`
#' * Statistical summaries `sd()`, `var()`, `cor()`, `cov()`
#' * `runif()` translates to `RAND()`
#' * Regular expression functions (DB2 11.1+)
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-db2
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_db2())
#' lf |> head()
#' lf |> transmute(x = paste0(d, " times"))
#' lf |> summarise(x = sd(b, na.rm = TRUE))
NULL

#' @export
#' @rdname backend-db2
dialect_db2 <- function() {
  new_sql_dialect(
    "db2",
    quote_identifier = function(x) sql_quote(x, '"')
  )
}

#' @export
sql_dialect.DB2Connection <- function(con) {
  dialect_db2()
}

#' @export
dbplyr_edition.DB2Connection <- function(con) {
  2L
}

#' @export
sql_query_select.sql_dialect_db2 <- function(
  con,
  select,
  from,
  where = NULL,
  group_by = NULL,
  having = NULL,
  window = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  ...,
  subquery = FALSE,
  lvl = 0
) {
  sql_select_clauses(
    select = sql_clause_select(select, distinct),
    from = sql_clause_from(from),
    where = sql_clause_where(where),
    group_by = sql_clause_group_by(group_by),
    having = sql_clause_having(having),
    window = sql_clause_window(window),
    order_by = sql_clause_order_by(order_by, subquery, limit),
    # DB2 uses FETCH FIRST instead of LIMIT
    limit = if (!is.null(limit)) {
      limit <- as.integer(limit)
      sql_glue2(con, "FETCH FIRST {limit} ROWS ONLY")
    },
    lvl = lvl
  )
}

#' @export
sql_translation.sql_dialect_db2 <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_odbc_scalar,
      # https://www.ibm.com/docs/en/db2-for-zos/12.0.0?topic=elements-casting-between-data-types
      as.character = sql_cast("VARCHAR(255)"),
      as.numeric = sql_cast("DOUBLE"),
      as.double = sql_cast("DOUBLE"),
      as.integer = sql_cast("INTEGER"),
      as.integer64 = sql_cast("BIGINT"),
      as.Date = sql_cast("DATE"),
      as.POSIXct = sql_cast("TIMESTAMP"),

      # https://www.ibm.com/docs/en/db2/9.7?topic=functions-rand
      runif = function(n = n(), min = 0, max = 1) {
        sql_runif("RAND()", n = {{ n }}, min = min, max = max)
      },

      # string -----------------------------------------------------------------
      # https://www.ibm.com/docs/en/db2-for-zos/12.0.0?topic=functions-concat
      paste = sql_paste_infix(" ", "||", sql_cast("VARCHAR(255)")),
      paste0 = sql_paste_infix("", "||", sql_cast("VARCHAR(255)")),
      str_c = sql_paste_infix("", "||", sql_cast("VARCHAR(255)")),

      # DB2 supports POSITION(substring IN string)
      # https://www.db2tutorial.com/db2-string-functions/db2-locate/
      str_locate = function(string, pattern) {
        sql_glue("LOCATE({pattern}, {string})")
      },
      str_detect = function(string, pattern, negate = FALSE) {
        sql_str_pattern_switch(
          string = string,
          pattern = {{ pattern }},
          negate = negate,
          f_fixed = sql_str_detect_fixed_position("detect")
        )
      },
      str_starts = function(string, pattern, negate = FALSE) {
        sql_str_pattern_switch(
          string = string,
          pattern = {{ pattern }},
          negate = negate,
          f_fixed = sql_str_detect_fixed_position("start")
        )
      },
      str_ends = function(string, pattern, negate = FALSE) {
        sql_str_pattern_switch(
          string = string,
          pattern = {{ pattern }},
          negate = negate,
          f_fixed = sql_str_detect_fixed_position("end")
        )
      },

      # https://www.ibm.com/docs/en/db2/11.5.x?topic=functions-regexp-replace
      # 4th argument is starting position (default: 1 => first char of string)
      # 5th argument is occurrence (default: 0 => match all occurrences)
      str_replace = function(string, pattern, replacement) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement}, 1, 1)")
      },
      str_replace_all = function(string, pattern, replacement) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement})")
      },
      str_remove = function(string, pattern) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, '', 1, 1)")
      },
      str_remove_all = function(string, pattern) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, '')")
      },
      str_squish = function(string) {
        sql_glue("LTRIM(RTRIM(REGEXP_REPLACE({string}, '\\s+', ' ')))")
      },

      # lubridate --------------------------------------------------------------
      # https://www.ibm.com/docs/en/db2/11.5.x?topic=functions-dayofweek
      today = \() sql("CURRENT DATE"),
      now = \() sql("CURRENT TIMESTAMP"),

      yday = sql_prefix("DAYOFYEAR", 1),
      wday = function(x, label = FALSE, abbr = TRUE, week_start = NULL) {
        check_bool(label)
        check_bool(abbr)
        check_number_whole(week_start, allow_null = TRUE)

        if (!label) {
          week_start <- week_start %||% getOption("lubridate.week.start", 7)
          # DB2 DAYOFWEEK: 1=Sunday, 7=Saturday
          offset <- as.integer(7 - week_start)
          sql_glue("MOD(DAYOFWEEK({x}) + {offset} - 1, 7) + 1")
        } else if (abbr) {
          sql_glue("SUBSTR(DAYNAME({x}), 1, 3)")
        } else {
          sql_glue("DAYNAME({x})")
        }
      },
      quarter = function(x, with_year = FALSE, fiscal_start = 1) {
        check_bool(with_year)
        check_unsupported_arg(fiscal_start, 1, backend = "DB2")

        if (with_year) {
          sql_glue("(CHAR(YEAR({x})) || '.' || CHAR(QUARTER({x})))")
        } else {
          sql_glue("QUARTER({x})")
        }
      },
      week = sql_prefix("WEEK_ISO", 1),
      isoweek = sql_prefix("WEEK_ISO", 1),
      isoyear = sql_prefix("YEAR", 1),

      # clock ------------------------------------------------------------------
      # https://db2portal.blogspot.com/2022/01/db2-date-and-time-arithmetic.html
      add_days = function(x, n, ...) {
        check_dots_empty()
        sql_glue("({x} + {n} DAYS)")
      },
      add_years = function(x, n, ...) {
        check_dots_empty()
        sql_glue("({x} + {n} YEARS)")
      },

      difftime = function(time1, time2, tz, units = "days") {
        check_unsupported_arg(tz)
        check_unsupported_arg(units, allowed = "days")

        sql_glue("(DAYS(CAST({time1} AS DATE)) - DAYS(CAST({time2} AS DATE)))")
      }
    ),
    sql_translator(
      .parent = base_odbc_agg,
      # https://www.ibm.com/docs/SSEPGG_11.5.0/com.ibm.db2.luw.sql.ref.doc/doc/r0002319.html
      var = sql_aggregate("VAR_SAMP", "var"),
      cor = sql_aggregate_2("CORRELATION"),
      cov = sql_aggregate_2("COVARIANCE_SAMP"),
      # https://www.db2tutorial.com/db2-aggregate-functions/db2-listagg/
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        sql_glue("LISTAGG({x}, {collapse})")
      }
    ),
    sql_translator(
      .parent = base_odbc_win,
      var = win_aggregate("VAR_SAMP"),
      cor = win_aggregate_2("CORRELATION"),
      cov = win_aggregate_2("COVARIANCE_SAMP"),
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        win_over(
          sql_glue("LISTAGG({x}, {collapse})"),
          partition = win_current_group(),
          order = win_current_order()
        )
      }
    )
  )
}

#' @export
sql_table_analyze.sql_dialect_db2 <- function(con, table, ...) {
  # https://www.ibm.com/support/knowledgecenter/en/SSEPGG_11.1.0/com.ibm.db2.luw.admin.cmd.doc/doc/r0001980.html
  sql_glue2(
    con,
    "RUNSTATS ON TABLE {.tbl table} WITH DISTRIBUTION AND DETAILED INDEXES ALL"
  )
}

#' @export
sql_values_subquery.sql_dialect_db2 <- sql_values_subquery_column_alias

#' Escape/quote a value
#'
#' @description
#' `escape()` turns R values into SQL literals. It implements double dispatch
#' via two sets of generics: first `escape()` dispatches on the class of `x`,
#' then that method calls `sql_escape_ident()`, `sql_escape_logical()`, etc,
#' which dispatch on `con`.
#'
#' These generics translate individual values into SQL. The core
#' generics are [DBI::dbQuoteIdentifier()] and [DBI::dbQuoteString()]
#' for quoting identifiers and strings, but dbplyr needs additional
#' tools for inserting logical, date, date-time, and raw values into
#' queries.
#'
#' @param x An object to escape. Existing sql vectors will be left as is,
#'   character vectors are escaped with single quotes, numeric vectors have
#'   trailing `.0` added if they're whole numbers, identifiers are
#'   escaped with double quotes.
#' @param parens,collapse Controls behaviour when multiple values are supplied.
#'   `parens` should be a logical flag, or if `NA`, will wrap in
#'   parens if length > 1.
#'
#'   Default behaviour: lists are always wrapped in parens and separated by
#'   commas, identifiers are separated by commas and never wrapped,
#'   atomic vectors are separated by spaces and wrapped in parens if needed.
#' @param con A [sql_dialect] object or database connection. Connections are
#'   supported for backward compatibility.
#' @family generic
#' @returns A [sql] vector.
#' @aliases db-quote
#' @export
#' @examples
#' con <- dialect_ansi()
#'
#' # Doubles vs. integers
#' escape(1:5, con = con)
#' escape(c(1, 5.4), con = con)
#'
#' # String vs known sql vs. sql identifier
#' escape("X", con = con)
#' escape(sql("X"), con = con)
#' escape(ident("X"), con = con)
#'
#' # Escaping is idempotent
#' escape("X", con = con)
#' escape(escape("X", con = con), con = con)
#'
#' # Database specific generics
#' sql_escape_logical(con, c(TRUE, FALSE, NA))
#' sql_escape_date(con, Sys.Date())
#' sql_escape_date(con, Sys.time())
#' sql_escape_raw(con, charToRaw("hi"))
escape <- function(x, parens = NA, collapse = " ", con = NULL) {
  check_con(con)

  UseMethod("escape")
}

# ident -------------------------------------------------------------------

#' @export
escape.ident <- function(x, parens = FALSE, collapse = ", ", con = NULL) {
  y <- set_names(sql_escape_ident(con, x), names(x))
  sql_vector(y, parens, collapse, con = con)
}

#' @export
#' @rdname escape
sql_escape_ident <- function(con, x) {
  UseMethod("sql_escape_ident", sql_dialect(con))
}
#' @export
sql_escape_ident.DBIConnection <- function(con, x) {
  sql(DBI::dbQuoteIdentifier(con, x))
}

#' @export
sql_escape_ident.sql_dialect <- function(con, x) {
  # Needed because UseMethod hack only affects dispatch, not value
  sql_dialect(con)$quote_identifier(x)
}

# logical -----------------------------------------------------------------

#' @export
escape.logical <- function(x, parens = NA, collapse = ", ", con = NULL) {
  sql_vector(sql_escape_logical(con, x), parens, collapse, con = con)
}

#' @rdname escape
#' @export
sql_escape_logical <- function(con, x) {
  UseMethod("sql_escape_logical", sql_dialect(con))
}
#' @export
sql_escape_logical.DBIConnection <- function(con, x) {
  y <- as.character(x)
  y[is.na(x)] <- "NULL"
  sql(y)
}

#' @export
sql_escape_logical.sql_dialect <- sql_escape_logical.DBIConnection

# factor ------------------------------------------------------------------

#' @export
escape.factor <- function(x, parens = NA, collapse = ", ", con = NULL) {
  x <- as.character(x)
  escape.character(x, parens = parens, collapse = collapse, con = con)
}

# Date --------------------------------------------------------------------

#' @export
escape.Date <- function(x, parens = NA, collapse = ", ", con = NULL) {
  sql_vector(sql_escape_date(con, x), parens, collapse, con = con)
}

#' @export
#' @rdname escape
sql_escape_date <- function(con, x) {
  UseMethod("sql_escape_date", sql_dialect(con))
}
#' @export
sql_escape_date.DBIConnection <- function(con, x) {
  sql_escape_string(con, as.character(x))
}

#' @export
sql_escape_date.sql_dialect <- sql_escape_date.DBIConnection

# POSIXt ------------------------------------------------------------------

#' @export
escape.POSIXt <- function(x, parens = NA, collapse = ", ", con = NULL) {
  sql_vector(sql_escape_datetime(con, x), parens, collapse, con = con)
}

#' @export
#' @rdname escape
sql_escape_datetime <- function(con, x) {
  UseMethod("sql_escape_datetime", sql_dialect(con))
}
#' @export
sql_escape_datetime.DBIConnection <- function(con, x) {
  x <- strftime(x, "%Y-%m-%dT%H:%M:%OSZ", tz = "UTC")
  sql_escape_string(con, x)
}

#' @export
sql_escape_datetime.sql_dialect <- sql_escape_datetime.DBIConnection

# character ---------------------------------------------------------------

#' @export
escape.character <- function(x, parens = NA, collapse = ", ", con = NULL) {
  sql_vector(sql_escape_string(con, x), parens, collapse, con = con)
}

#' @export
#' @rdname escape
sql_escape_string <- function(con, x) {
  UseMethod("sql_escape_string")
}
#' @export
sql_escape_string.default <- function(con, x) {
  if (inherits(x, "SQL")) {
    sql(unclass(x))
  } else {
    sql(sql_quote(x, "'"))
  }
}

# double ------------------------------------------------------------------

#' @export
escape.double <- function(x, parens = NA, collapse = ", ", con = NULL) {
  out <- ifelse(is_whole_number(x), sprintf("%.1f", x), as.character(x))

  # Special values
  out[is.na(x)] <- "NULL"
  inf <- is.infinite(x)
  out[inf & x > 0] <- "'Infinity'"
  out[inf & x < 0] <- "'-Infinity'"

  sql_vector(out, parens, collapse, con = con)
}

is_whole_number <- function(x) {
  trunc(x) == x
}

# integer -----------------------------------------------------------------

#' @export
escape.integer <- function(x, parens = NA, collapse = ", ", con = NULL) {
  x[is.na(x)] <- "NULL"
  sql_vector(x, parens, collapse, con = con)
}

#' @export
escape.integer64 <- function(x, parens = NA, collapse = ", ", con = NULL) {
  x <- as.character(x)
  x[is.na(x)] <- "NULL"
  sql_vector(x, parens, collapse, con = con)
}

# blob --------------------------------------------------------------------

#' @export
escape.blob <- function(x, parens = NA, collapse = ", ", con = NULL) {
  pieces <- vapply(x, sql_escape_raw, character(1), con = con)
  sql_vector(pieces, isTRUE(parens) || length(pieces) > 1, collapse, con = con)
}

#' @export
#' @rdname escape
sql_escape_raw <- function(con, x) {
  UseMethod("sql_escape_raw", sql_dialect(con))
}
#' @export
sql_escape_raw.DBIConnection <- function(con, x) {
  # Unlike the other escape functions, this is not vectorised because
  # raw "vectors" are scalars in this content

  if (is.null(x)) {
    sql("NULL")
  } else {
    # SQL-99 standard for BLOB literals
    # https://crate.io/docs/sql-99/en/latest/chapters/05.html#blob-literal-s
    sql(paste0(c("X'", format(x), "'"), collapse = ""))
  }
}

#' @export
sql_escape_raw.sql_dialect <- sql_escape_raw.DBIConnection

# NULL --------------------------------------------------------------------

#' @export
escape.NULL <- function(x, parens = NA, collapse = " ", con = NULL) {
  sql("NULL")
}

# sql ---------------------------------------------------------------------

#' @export
escape.sql <- function(x, parens = NULL, collapse = NULL, con = NULL) {
  sql_vector(x, isTRUE(parens), collapse, con = con)
}

# list --------------------------------------------------------------------

#' @export
escape.list <- function(x, parens = TRUE, collapse = ", ", con = NULL) {
  pieces <- vapply(x, escape, character(1), con = con)
  sql_vector(pieces, parens, collapse, con = con)
}

# errors ------------------------------------------------------------------

#' @export
escape.data.frame <- function(x, parens = TRUE, collapse = ", ", con = NULL) {
  error_embed("a data.frame", "df$x")
}

#' @export
escape.reactivevalues <- function(
  x,
  parens = TRUE,
  collapse = ", ",
  con = NULL
) {
  error_embed("shiny inputs", "input$x")
}

#' @export
escape.default <- function(x, parens = TRUE, collapse = ", ", con = NULL) {
  error_embed(obj_type_friendly(x), "x")
}

# Also used in default_ops() for reactives
error_embed <- function(type, expr) {
  cli_abort(
    c(
      "Cannot translate {type} to SQL.",
      `i` = "Do you want to force evaluation in R with (e.g.) `!!{expr}` or `local({expr})`?"
    ),
    call = NULL
  )
}

# helpers -----------------------------------------------------------------

#' @export
#' @rdname escape
sql_vector <- function(x, parens = NA, collapse = " ", con = NULL) {
  check_bool(parens, allow_na = TRUE)
  check_string(collapse, allow_null = TRUE)
  check_con(con)

  if (is.na(parens)) {
    parens <- length(x) > 1L
  }

  x <- names_to_as(con, x)
  sql_collapse(x, collapse, parens)
}

sql_collapse <- function(x, collapse = " ", parens = FALSE) {
  x <- paste(x, collapse = collapse)
  if (parens) {
    x <- paste0("(", x, ")", recycle0 = TRUE)
  }
  sql(x)
}

# Strips names from named sql() vector by converting them to AS identifiers
names_to_as <- function(con, x, names = names2(x)) {
  if (length(x) == 0) {
    return(sql())
  }

  names_esc <- sql_escape_ident(con, names)
  as_sql <- style_kw(" AS ")
  as <- ifelse(names == "" | names_esc == x, "", paste0(as_sql, names_esc))

  sql(paste0(x, as))
}

# Identifiers ------------------------------------------------------------------

ei <- function(...) unclass(escape(ident(c(...)), con = dialect_ansi()))

test_that("identifiers get identifier quoting", {
  expect_equal(ei("x"), '"x"')
})

test_that("identifiers are comma separated", {
  expect_equal(ei("x", "y"), '"x", "y"')
})

test_that("identifier names become AS", {
  expect_equal(ei(x = "y"), '"y" AS "x"')
})

# sql_collapse ------------------------------------------------------------------

test_that("sql_collapse collapses with separator", {
  expect_equal(sql_collapse(sql("a", "b")), sql("a b"))
  expect_equal(sql_collapse(sql("a", "b"), collapse = ", "), sql("a, b"))
})

test_that("sql_collapse handles 0-length inputs", {
  expect_equal(sql_collapse(character()), sql(""))
  expect_equal(sql_collapse(character(), collapse = NULL), sql())

  expect_equal(sql_collapse(character(), parens = TRUE), sql("()"))
  expect_equal(sql_collapse(character(), collapse = NULL, parens = TRUE), sql())
})

test_that("sql_collapse controls parens", {
  expect_equal(sql_collapse(sql("a")), sql("a"))
  expect_equal(sql_collapse(sql("a", "b")), sql("a b"))

  expect_equal(sql_collapse(sql("a"), parens = TRUE), sql("(a)"))
  expect_equal(sql_collapse(sql("a", "b"), parens = TRUE), sql("(a b)"))
})

# Numeric ------------------------------------------------------------------

test_that("missing values become null", {
  con <- dialect_ansi()

  expect_equal(escape(NA, con = con), sql("NULL"))
  expect_equal(escape(NA_real_, con = con), sql("NULL"))
  expect_equal(escape(NA_integer_, con = con), sql("NULL"))
  expect_equal(escape(NA_character_, con = con), sql("NULL"))
})

test_that("-Inf and Inf are expanded and quoted", {
  con <- dialect_ansi()
  expect_equal(escape(Inf, con = con), sql("'Infinity'"))
  expect_equal(escape(-Inf, con = con), sql("'-Infinity'"))
})

test_that("can escape integer64 values", {
  con <- dialect_ansi()
  skip_if_not_installed("bit64")

  expect_equal(
    escape(bit64::as.integer64(NA), con = con),
    sql("NULL")
  )
  expect_equal(
    escape(bit64::as.integer64("123456789123456789"), con = con),
    sql("123456789123456789")
  )
})

test_that("recognises integerish numerics", {
  expect_equal(is_whole_number(c(1.1, 1.0, 1.000001)), c(FALSE, TRUE, FALSE))
  con <- dialect_ansi()
  expect_equal(
    escape(c(1.1, 1, 1.000001), con = con),
    sql("(1.1, 1.0, 1.000001)")
  )
})

# Logical -----------------------------------------------------------------

test_that("logical is SQL-99 compatible (by default)", {
  con <- dialect_ansi()
  expect_equal(escape(TRUE, con = con), sql("TRUE"))
  expect_equal(escape(FALSE, con = con), sql("FALSE"))
  expect_equal(escape(NA, con = con), sql("NULL"))
})

# String -----------------------------------------------------------------

test_that("can escape strings", {
  con <- dialect_ansi()

  expect_equal(escape("x", con = con), sql("'x'"))
  expect_equal(escape(DBI::SQL("x"), con = con), sql("x"))
})

# Date-time ---------------------------------------------------------------

test_that("date and date-times are converted to ISO 8601", {
  con <- dialect_ansi()
  x1 <- ISOdatetime(2000, 1, 2, 3, 4, 5, tz = "America/New_York")
  x2 <- as.Date(x1)
  expect_equal(escape(x1, con = con), sql("'2000-01-02T08:04:05Z'"))
  expect_equal(escape(x2, con = con), sql("'2000-01-02'"))
})

# Raw -----------------------------------------------------------------

test_that("raw is SQL-99 compatible (by default)", {
  con <- dialect_ansi()
  expect_equal(escape(blob::as_blob(raw(0)), con = con), sql("X''"))
  expect_equal(
    escape(blob::as_blob(as.raw(c(0x01, 0x02, 0x03))), con = con),
    sql("X'010203'")
  )
  expect_equal(
    escape(blob::as_blob(as.raw(c(0x00, 0xff))), con = con),
    sql("X'00ff'")
  )
})

test_that("can translate raw to blob spec", {
  con <- dialect_ansi()

  expect_equal(sql_escape_raw(con, NULL), sql("NULL"))
  expect_equal(sql_escape_raw(con, charToRaw("abc")), sql("X'616263'"))
})

# Factor ------------------------------------------------------------------

test_that("factors are translated", {
  con <- dialect_ansi()
  expect_equal(escape(factor(c("a", "b")), con = con), sql("('a', 'b')"))
})

# Helpful errors --------------------------------------------------------

test_that("con must not be NULL", {
  expect_snapshot(error = TRUE, escape("a"))
  expect_snapshot(error = TRUE, sql_vector("a"))
})

test_that("other objects get informative error", {
  lf <- lazy_frame(x = 1)

  input <- structure(list(), class = "reactivevalues")
  x <- structure(function() "y", class = "reactive")
  df <- data.frame(x = 1)

  expect_snapshot(
    {
      lf |> filter(x == input)
      lf |> filter(x == x())
      lf |> filter(x == df)
      lf |> filter(x == mean)
    },
    error = TRUE
  )
})

# names_to_as() -----------------------------------------------------------

test_that("names_to_as() correctly aliases", {
  con <- dialect_ansi()

  # no alias when name is missing
  expect_equal(names_to_as(con, c("name")), sql("name"))
  # no alias when (quoted) name and value are identical
  expect_equal(names_to_as(con, c(name = '"name"')), sql('"name"'))
  # alias when name and value are different
  expect_equal(names_to_as(con, c(new = "old")), sql('old AS "new"'))
})

#' Pivot data from wide to long
#'
#' @description
#' `pivot_longer()` "lengthens" data, increasing the number of rows and
#' decreasing the number of columns. The inverse transformation is
#' `tidyr::pivot_wider()]
#'
#' Learn more in `vignette("pivot", "tidyr")`.
#'
#' While most functionality is identical there are some differences to
#' `pivot_longer()` on local data frames:
#' * the output is sorted differently/not explicitly,
#' * the coercion of mixed column types is left to the database,
#' * `values_ptypes` NOT supported.
#'
#' Note that `build_longer_spec()` and `pivot_longer_spec()` do not work with
#' remote tables.
#'
#' @details
#' The SQL translation basically works as follows:
#'
#' 1. split the specification by its key columns i.e. by variables crammed
#' into the column names.
#' 2. for each part in the splitted specification `transmute()` `data` into the
#' following columns
#'   * id columns i.e. columns that are not pivotted
#'   * key columns
#'   * value columns i.e. columns that are pivotted
#' 3. combine all the parts with `union_all()`
#'
#' @param data A data frame to pivot.
#' @param cols Columns to pivot into longer format.
#' @param names_to A string specifying the name of the column to create
#'   from the data stored in the column names of `data`.
#' @param names_prefix A regular expression used to remove matching text
#'   from the start of each variable name.
#' @param names_sep,names_pattern If `names_to` contains multiple values,
#'   these arguments control how the column name is broken up.
#' @param names_repair What happens if the output has invalid column names?
#' @param values_to A string specifying the name of the column to create
#'   from the data stored in cell values. If `names_to` is a character
#'   containing the special `.value` sentinel, this value will be ignored,
#'   and the name of the value column will be derived from part of the
#'   existing column names.
#' @param values_drop_na If `TRUE`, will drop rows that contain only `NA`s
#'   in the `value_to` column.
#' @param names_transform,values_transform A list of column name-function pairs.
#' @param names_ptypes A list of column name-prototype pairs.
#' @param values_ptypes Not supported.
#' @param ... Additional arguments passed on to methods.
#' @examplesIf rlang::is_installed("tidyr", version = "1.0.0")
#' # See vignette("pivot") for examples and explanation
#'
#' # Simplest case where column names are character data
#' memdb_frame(
#'   id = c("a", "b"),
#'   x = 1:2,
#'   y = 3:4
#' ) %>%
#'   tidyr::pivot_longer(-id)
pivot_longer.tbl_lazy <- function(data,
                                  cols,
                                  names_to = "name",
                                  names_prefix = NULL,
                                  names_sep = NULL,
                                  names_pattern = NULL,
                                  names_ptypes = NULL,
                                  names_transform = NULL,
                                  names_repair = "check_unique",
                                  values_to = "value",
                                  values_drop_na = FALSE,
                                  values_ptypes,
                                  values_transform = NULL,
                                  ...) {
  if (!is_missing(values_ptypes)) {
    cli_abort("The {.arg values_ptypes} argument is not supported for remote back-ends")
  }

  rlang::check_dots_empty()

  cols <- enquo(cols)
  spec <- tidyr::build_longer_spec(tidyselect_data_proxy(data), !!cols,
    names_to = names_to,
    values_to = values_to,
    names_prefix = names_prefix,
    names_sep = names_sep,
    names_pattern = names_pattern,
    names_ptypes = names_ptypes,
    names_transform = names_transform
  )

  dbplyr_pivot_longer_spec(data, spec,
    names_repair = names_repair,
    values_drop_na = values_drop_na,
    values_transform = values_transform
  )
}

dbplyr_pivot_longer_spec <- function(data,
                                     spec,
                                     names_repair = "check_unique",
                                     values_drop_na = FALSE,
                                     values_transform = NULL) {
  spec <- tidyr::check_pivot_spec(spec)
  # .seq col needed if different input columns are mapped to the same output
  # column
  spec <- deduplicate_spec(spec, data)

  id_cols <- syms(setdiff(colnames(data), spec$.name))
  repair_info <- apply_name_repair_pivot_longer(id_cols, spec, names_repair)
  id_cols <- repair_info$id_cols
  spec <- repair_info$spec

  spec_split <- vctrs::vec_split(spec, spec[, -(1:2)])

  call <- current_env()
  value_names <- unique(spec$.value)
  values_transform <- check_list_of_functions(values_transform, value_names, "values_transform", call)

  nms_map <- tibble(
    name = colnames(spec_split$key),
    name_mapped = ifelse(
      name %in% unlist(spec_split$key),
      paste0("..", name),
      name
    )
  )
  spec_split$key <- set_names(spec_split$key, nms_map$name_mapped)

  data_long_list <- purrr::map(
    vctrs::vec_seq_along(spec_split),
    function(idx) {
      row <- spec_split$val[[idx]][, 1:2]
      keys <- spec_split$key[idx, ]
      keys$.seq <- NULL

      measure_cols_exprs <- get_measure_column_exprs(
        row[[".name"]],
        row[[".value"]],
        values_transform,
        data = data,
        call = call
      )

      transmute(
        data,
        !!!id_cols,
        !!!keys,
        !!!measure_cols_exprs
      )
    }
  )

  data_long <- purrr::reduce(data_long_list, union_all)


  if (values_drop_na) {
    value_cols <- unique(spec$.value)

    data_long <- dplyr::filter_at(
      data_long,
      value_cols,
      dplyr::all_vars(!is.na(.))
    )
  }

  data_long %>%
    rename(!!!tibble::deframe(nms_map))
}

get_measure_column_exprs <- function(name, value, values_transform, data, call) {
  measure_cols <- set_names(syms(name), value)
  purrr::imap(
    measure_cols,
    ~ {
      f_trans <- values_transform[[.y]]

      if (is_null(f_trans)) {
        .x
      } else {
        resolve_fun(f_trans, .x, data, call)
      }
    }
  )
}

apply_name_repair_pivot_longer <- function(id_cols, spec, names_repair) {
  # Calculates how the column names in `pivot_longer()` need to be repaired
  # and applies this to the `id_cols` and the `spec` argument:
  # * The names of `id_cols` are the repaired id column names
  # * The `spec` columns after the third column are renamed to the repaired name
  # * The entries in the `value` column of `spec` are changed to the repaired name

  nms_map_df <- vctrs::vec_rbind(
    tibble(from = "id_cols", name = as.character(id_cols)),
    tibble(from = "measure_cols", name = colnames(spec)[-(1:2)]),
    tibble(from = "value_cols", name = unique(spec[[".value"]]))
  ) %>%
    mutate(name_rep = vctrs::vec_as_names(name, repair = names_repair))
  nms_map <- split(nms_map_df, nms_map_df$from)

  id_cols <- purrr::set_names(id_cols, nms_map$id_cols$name_rep)

  colnames(spec)[-(1:2)] <- nms_map$measure_cols$name_rep

  value_nms_map <- purrr::set_names(
    nms_map$value_cols$name_rep,
    nms_map$value_cols$name
  )
  spec$.value <- dplyr::recode(spec$.value, !!!value_nms_map)

  list(id_cols = id_cols, spec = spec)
}

# The following is copy-pasted from `tidyr`

# nocov start
# Ensure that there's a one-to-one match from spec to data by adding
# a special .seq variable which is automatically removed after pivotting.
deduplicate_spec <- function(spec, df) {
  # COPIED FROM tidyr

  # Ensure each .name has a unique output identifier
  key <- spec[setdiff(names(spec), ".name")]
  if (vctrs::vec_duplicate_any(key)) {
    pos <- vctrs::vec_group_loc(key)$loc
    seq <- vector("integer", length = nrow(spec))
    for (i in seq_along(pos)) {
      seq[pos[[i]]] <- seq_along(pos[[i]])
    }
    spec$.seq <- seq
  }

  # Match spec to data, handling duplicated column names
  col_id <- vctrs::vec_match(names(df), spec$.name)
  has_match <- !is.na(col_id)

  if (!vctrs::vec_duplicate_any(col_id[has_match])) {
    return(spec)
  }

  spec <- vctrs::vec_slice(spec, col_id[has_match])
  # Need to use numeric indices because names only match first
  spec$.name <- seq_along(df)[has_match]

  pieces <- vctrs::vec_split(seq_len(nrow(spec)), col_id[has_match])
  copy <- integer(nrow(spec))
  for (i in seq_along(pieces$val)) {
    idx <- pieces$val[[i]]
    copy[idx] <- seq_along(idx)
  }

  spec$.seq <- copy
  spec
}

check_list_of_functions <- function(x, names, arg, call = caller_env()) {
  # mostly COPIED FROM tidyr
  if (is.null(x)) {
    x <- set_names(list(), character())
  }

  if (!vctrs::vec_is_list(x)) {
    x <- rep_named(names, list(x))
  }

  if (length(x) > 0L && !is_named(x)) {
    cli_abort("All elements of {.arg {arg}} must be named.", call = call)
  }

  if (vctrs::vec_duplicate_any(names(x))) {
    cli_abort("The names of {.arg {arg}} must be unique.", call = call)
  }

  # Silently drop user supplied names not found in the data
  x <- x[intersect(names(x), names)]

  x
}
# nocov end

globalVariables(".")

#' Backend: ODBC
#'
#' @description
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology. Key differences for this backend
#' are minor translations for common data types.
#'
#' Use `simulate_odbc()` with `lazy_frame()` to see simulated SQL without
#' converting to live access database.
#'
#' @name backend-odbc
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, d = 2, c = "z", con = simulate_odbc())
#' lf %>% transmute(x = as.numeric(b))
#' lf %>% transmute(x = as.integer(b))
#' lf %>% transmute(x = as.character(b))
NULL

#' @export
#' @rdname backend-odbc
simulate_odbc <- function() simulate_dbi("OdbcConnection")

#' @export
dbplyr_edition.OdbcConnection <- function(con) {
  2L
}

#' @export
sql_translation.OdbcConnection <- function(con) {
  sql_variant(
    base_odbc_scalar,
    base_odbc_agg,
    base_odbc_win
  )
}

#' @export
#' @rdname sql_variant
#' @format NULL
base_odbc_scalar <- sql_translator(
  .parent = base_scalar,
  as.numeric = sql_cast("DOUBLE"),
  as.double = sql_cast("DOUBLE"),
  as.integer = sql_cast("INT"),
  as.character = sql_cast("STRING")
)

#' @export
#' @rdname sql_variant
#' @format NULL
base_odbc_agg <- sql_translator(
  .parent = base_agg,
  sd = sql_aggregate("STDDEV_SAMP", "sd")
)

#' @export
#' @rdname sql_variant
#' @format NULL
base_odbc_win <- sql_translator(
  .parent = base_win,
  sd = win_aggregate("STDDEV_SAMP"),
)

# nocov start
#' @export
db_connection_describe.OdbcConnection <- function(con) {
  info <- DBI::dbGetInfo(con)

  host <- if (info$servername == "") "localhost" else info$servername
  port <- if (info$port == "") "" else paste0(":", info$port)

  paste0(
    info$dbms.name, " ", info$db.version,
    "[", info$username, "@", host, port,
    "/", info$dbname, "]"
  )
}
# nocov end

utils::globalVariables("EXP")

# Backward compatibility ---------------------------------------------------------
# These functions are provided for backward compatibility with packages that
# used internal dbplyr functions. They should not be used in new code.

# Used by duckdb
glue_sql2 <- function(
  .con,
  ...,
  .sep = "",
  .envir = parent.frame(),
  .open = "{",
  .close = "}",
  .na = DBI::SQL("NULL"),
  .null = "",
  .comment = "#",
  .literal = FALSE,
  .trim = TRUE
) {
  sql(glue(
    ...,
    .sep = .sep,
    .envir = .envir,
    .open = .open,
    .close = .close,
    .na = .na,
    .null = .null,
    .comment = .comment,
    .literal = .literal,
    .transformer = sql_quote_transformer(.con),
    .trim = .trim
  ))
}

sql_quote_transformer <- function(connection) {
  function(text, envir) {
    collapse_regex <- "[*][[:space:]]*$"
    should_collapse <- grepl(collapse_regex, text)
    if (should_collapse) {
      text <- sub(collapse_regex, "", text)
    }

    type_regex <- "^\\.(tbl|col|name|from|kw|val) (.*)"
    m <- regexec(type_regex, text)
    is_quoted <- any(m[[1]] != -1)
    if (is_quoted) {
      matches <- regmatches(text, regexec(type_regex, text))[[1]]

      type <- matches[[2]]
      value <- matches[[3]]
    } else {
      value <- text
      type <- "raw"
    }
    value <- eval(parse(text = value, keep.source = FALSE), envir)
    glue_check_collapse(type, should_collapse)

    if (type == "tbl") {
      value <- as_table_path(value, connection)
    } else if (type == "from") {
      value <- as_table_source(value, connection)
    } else if (type == "col") {
      if (is_bare_character(value)) {
        value <- ident(value)
      }
    } else if (type == "name") {
      # allowed should be `ident`, `ident_q` (maybe), string
      if (is_bare_character(value)) {
        value <- ident(value)
      }
    } else if (type == "kw") {
      value <- sql(style_kw(value))
    } else if (type == "val") {
      # keep as is
    } else if (type == "raw") {
      if (!is.sql(value) && !is_string(value)) {
        stop_input_type(value, what = c("a string", "scalar SQL"))
      }
    }

    if (type == "val") {
      if (should_collapse) {
        value <- escape(
          value,
          collapse = ", ",
          parens = FALSE,
          con = connection
        )
      } else {
        value <- escape(value, con = connection)
      }
    } else if (type %in% c("tbl", "from", "col", "name")) {
      value <- escape(value, collapse = NULL, parens = FALSE, con = connection)
      if (should_collapse) {
        value <- paste0(unclass(value), collapse = ", ")
      }
    }

    # TODO use `vctrs::vec_check_size(value, size = 1L)`
    if (length(value) != 1) {
      cli_abort("{.arg value} must have size 1, not {length(value)}.")
    }

    unclass(value)
  }
}

glue_check_collapse <- function(type, collapse) {
  if (type %in% c("col", "val")) {
    return()
  }

  if (collapse) {
    cli_abort(
      "Collapsing is only allowed for {.val col} and {.val val}, not for {.val {type}}."
    )
  }
}


# Used by RPresto

#' Optimise a SQL query
#'
#' No longer used.
#'
#' @export
#' @keywords internal
sql_optimise <- function(x, con = NULL, ..., subquery = FALSE) {
  UseMethod("sql_optimise")
}

# Used by ckanr, implyr, sergeant
#' Create a "sql src" object
#'
#' Deprecated: please use directly use a `DBIConnection` object instead.
#'
#' @param subclass Name of subclass.
#' @param con The connection object.
#' @param ... Other arguments passed on to individual methods.
#' @export
#' @keywords internal
src_sql <- function(subclass, con, ...) {
  lifecycle::deprecate_stop(when = "1.4.0", what = "src_sql()")
}

# Used by dm, etl, starwarsdb
#' @importFrom dplyr src_tbls
#' @export
src_tbls.src_sql <- function(x, ...) {
  lifecycle::deprecate_soft("2.6.0", "src_tbls()")
  DBI::dbListTables(x$con)
}

#' @importFrom dplyr same_src
#' @export
same_src.src_sql <- function(x, y) {
  if (!inherits(y, "src_sql")) {
    return(FALSE)
  }
  identical(x$con, y$con)
}

# dplyr:::print.src calls format, so we need to retain this method
#' @export
format.src_sql <- function(x, ...) {
  paste0(
    "src:  ",
    db_connection_describe(x$con),
    "\n",
    wrap("tbls: ", paste0(sort(src_tbls(x)), collapse = ", "))
  )
}

# Used by healthdb
#' Escape ANSI characters
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' Use [escape()] instead.
#'
#' @param x Object to escape.
#' @param parens,collapse Controls parens and collapsing. Passed on to
#'   [escape()].
#' @export
#' @keywords internal
escape_ansi <- function(x, parens = NA, collapse = "") {
  lifecycle::deprecate_soft("2.6.0", "escape_ansi()", "escape()")
  escape(x, parens = parens, collapse = collapse, con = simulate_dbi())
}

# nocov start

#' Create a "sql src" object
#'
#' Deprecated: please use directly use a `DBIConnection` object instead.
#'
#' @keywords internal
#' @export
#' @param subclass name of subclass. "src_sql" is an abstract base class, so you
#'   must supply this value. `src_` is automatically prepended to the
#'   class name
#' @param con the connection object
#' @param ... fields used by object
src_sql <- function(subclass, con, ...) {
  subclass <- paste0("src_", subclass)
  structure(list(con = con, ...), class = c(subclass, "src_sql", "src"))
}


#' @importFrom dplyr same_src
#' @export
same_src.src_sql <- function(x, y) {
  if (!inherits(y, "src_sql")) return(FALSE)
  identical(x$con, y$con)
}

#' @importFrom dplyr src_tbls
#' @export
src_tbls.src_sql <- function(x, ...) {
  dbListTables(x$con)
}

#' @export
format.src_sql <- function(x, ...) {
  paste0(
    "src:  ", dbplyr_connection_describe(x$con), "\n",
    wrap("tbls: ", paste0(sort(src_tbls(x)), collapse = ", "))
  )
}
# nocov end

#' Edit individual rows in the underlying database table
#'
#' @description
#' These are methods for the dplyr [rows_insert()], [`rows_append()`],
#' [`rows_update()`], [`rows_patch()`], [`rows_upsert()`], and [`rows_delete()`]
#' generics.
#'
#' When `in_place = TRUE` these verbs do not generate `SELECT` queries, but
#' instead directly modify the underlying data using `INSERT`, `UPDATE`, or
#' `DELETE` operators. This will require that you have write access to
#' the database: the connection needs permission to insert, modify or delete
#' rows, but not to alter the structure of the table.
#'
#' The default, `in_place = FALSE`, generates equivalent lazy tables (using
#' `SELECT` queries) that allow previewing the result without actually
#' modifying the underlying table on the database.
#'
#' @export
#' @inheritParams left_join.tbl_lazy
#' @param x A lazy table.
#'   For `in_place = TRUE`, this must be a table instantiated with [tbl()] or
#'   [compute()], not to a lazy query. The [remote_name()] function is used to
#'   determine the name of the table to be updated.
#' @param y A lazy table, data frame, or data frame extensions (e.g. a tibble).
#' @param by An unnamed character vector giving the key columns. The key columns
#'   must exist in both `x` and `y`. Keys typically uniquely identify each row,
#'   but this is only enforced for the key values of `y` when `rows_update()`,
#'   `rows_patch()`, or `rows_upsert()` are used.
#'
#'   By default, we use the first column in `y`, since the first column is
#'   a reasonable place to put an identifier variable.
#' @param conflict For `rows_insert()`, how should keys in `y` that conflict
#'   with keys in `x` be handled? A conflict arises if there is a key in `y`
#'   that already exists in `x`.
#'
#'   One of:
#'   - `"error"`, the default, is not supported for database tables. To get the
#'     same behaviour add a unique index on the `by` columns and use
#'     `rows_append()`.
#'   - `"ignore"` will ignore rows in `y` with keys that conflict with keys in
#'     `x`.
#' @param unmatched For `rows_update()`, `rows_patch()`, and `rows_delete()`,
#'   how should keys in `y` that are unmatched by the keys in `x` be handled?
#'
#'   One of:
#'   - `"error"`, the default, is not supported for database tables. Add a
#'     foreign key constraint on the `by` columns of `y` to let the database
#'     check this behaviour for you.
#'   - `"ignore"` will ignore rows in `y` with keys that are unmatched by the
#'     keys in `x`.
#' @param in_place  Should `x` be modified in place? If `FALSE` will
#'   generate a `SELECT` query that returns the modified table; if `TRUE`
#'   will modify the underlying table using a DML operation (`INSERT`, `UPDATE`,
#'   `DELETE` or similar).
#' @param returning Columns to return. See [get_returned_rows()] for details.
#' @param method A string specifying the method to use. This is only relevant for
#'   `in_place = TRUE`.
#'
#' @importFrom dplyr rows_insert
#' @returns A new `tbl_lazy` of the modified data.
#'   With `in_place = FALSE`, the result is a lazy query that prints visibly,
#'   because the purpose of this operation is to preview the results.
#'   With `in_place = TRUE`, `x` is returned invisibly,
#'   because the purpose of this operation is the side effect of modifying rows
#'   in the table behind `x`.
#' @rdname rows-db
#' @examples
#' library(dplyr)
#'
#' con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#' DBI::dbExecute(con, "CREATE TABLE Ponies (
#'    id INTEGER PRIMARY KEY AUTOINCREMENT,
#'    name TEXT,
#'    cutie_mark TEXT
#' )")
#'
#' ponies <- tbl(con, "Ponies")
#'
#' applejack <- copy_inline(con, data.frame(
#'   name = "Apple Jack",
#'   cutie_mark = "three apples"
#' ))
#'
#' # The default behavior is to generate a SELECT query
#' rows_insert(ponies, applejack, conflict = "ignore")
#' # And the original table is left unchanged:
#' ponies
#'
#' # You can also choose to modify the table with in_place = TRUE:
#' rows_insert(ponies, applejack, conflict = "ignore", in_place = TRUE)
#' # In this case `rows_insert()` returns nothing and the underlying
#' # data is modified
#' ponies
rows_insert.tbl_lazy <- function(
  x,
  y,
  by = NULL,
  ...,
  conflict = c("error", "ignore"),
  copy = "none",
  in_place = FALSE,
  returning = NULL,
  method = NULL
) {
  check_dots_empty()
  rows_check_in_place(x, in_place)
  table <- target_table(x, in_place)

  conflict <- rows_check_conflict(conflict)

  rows_check_containment(x, y)
  y <- rows_auto_copy(x, y, copy = copy)

  by <- rows_check_by(by, y)

  rows_check_key(x, by, "x")
  rows_check_key(y, by, "y", unique = TRUE)

  new_columns <- setdiff(colnames(y), by)

  returning_cols <- rows_check_returning(x, returning, enexpr(returning))

  if (!is_null(table)) {
    sql <- sql_query_insert(
      con = remote_con(x),
      table = table,
      from = sql_render(y, remote_con(x), lvl = 1),
      insert_cols = colnames(y),
      by = by,
      ...,
      conflict = conflict,
      returning_cols = returning_cols,
      method = method
    )

    rows_get_or_execute(x, sql, returning_cols)
  } else {
    out <- union_all(x, anti_join(y, x, by = by))

    if (!is_empty(returning_cols)) {
      # Need to `union_all()` with `x` so that all columns of `x` exist in the result
      returned_rows <- anti_join(y, x, by = by) |>
        union_all(x |> filter(0 == 1)) |>
        select(!!!returning_cols) |>
        collect()
      out <- set_returned_rows(out, returned_rows)
    }

    out
  }
}

#' @inheritParams dplyr::rows_append
#'
#' @export
#' @importFrom dplyr rows_append
#' @rdname rows-db
rows_append.tbl_lazy <- function(
  x,
  y,
  ...,
  copy = "none",
  in_place = FALSE,
  returning = NULL
) {
  check_dots_empty()
  rows_check_in_place(x, in_place)
  table <- target_table(x, in_place)

  rows_check_containment(x, y)
  y <- rows_auto_copy(x, y, copy = copy)

  returning_cols <- rows_check_returning(x, returning, enexpr(returning))

  if (!is_null(table)) {
    sql <- sql_query_append(
      con = remote_con(x),
      table = table,
      from = sql_render(y, remote_con(x), lvl = 1),
      insert_cols = colnames(y),
      ...,
      returning_cols = returning_cols
    )

    rows_get_or_execute(x, sql, returning_cols)
  } else {
    out <- union_all(x, y)

    if (!is_empty(returning_cols)) {
      # Need to `union_all()` with `x` so that all columns of `x` exist in the result
      returned_rows <- union_all(y, x |> filter(0 == 1)) |>
        select(!!!returning_cols) |>
        collect()
      out <- set_returned_rows(out, returned_rows)
    }

    out
  }
}

#' @inheritParams dplyr::rows_update
#'
#' @export
#' @importFrom dplyr rows_update
#' @rdname rows-db
rows_update.tbl_lazy <- function(
  x,
  y,
  by = NULL,
  ...,
  unmatched = c("error", "ignore"),
  copy = "none",
  in_place = FALSE,
  returning = NULL
) {
  check_dots_empty()
  rows_check_in_place(x, in_place)
  table <- target_table(x, in_place)

  rows_check_containment(x, y)
  y <- rows_auto_copy(x, y, copy = copy)

  by <- rows_check_by(by, y)

  rows_check_key(x, by, "x")
  rows_check_key(y, by, "y", unique = TRUE)

  unmatched <- rows_check_ummatched(unmatched)

  new_columns <- setdiff(colnames(y), by)

  returning_cols <- rows_check_returning(x, returning, enexpr(returning))

  if (!is_null(table)) {
    # TODO handle `returning_cols` here
    if (is_empty(new_columns)) {
      return(invisible(x))
    }

    con <- remote_con(x)
    update_cols <- setdiff(colnames(y), by)
    update_values <- set_names(
      sql_table_prefix(con, "...y", update_cols),
      update_cols
    )

    sql <- sql_query_update_from(
      con = con,
      table = table,
      from = sql_render(y, remote_con(y), lvl = 1),
      by = by,
      update_values = update_values,
      ...,
      returning_cols = returning_cols
    )

    rows_get_or_execute(x, sql, returning_cols)
  } else {
    existing_columns <- setdiff(colnames(x), new_columns)
    updated <- x |>
      select(!!!existing_columns) |>
      inner_join(y, by = by)

    if (is_empty(new_columns)) {
      out <- x
    } else {
      unchanged <- anti_join(x, y, by = by)
      out <- union_all(unchanged, updated)
    }

    if (!is_empty(returning_cols)) {
      returned_rows <- updated |>
        select(!!!returning_cols) |>
        collect()
      out <- set_returned_rows(out, returned_rows)
    }

    out
  }
}

#' @inheritParams dplyr::rows_patch
#'
#' @export
#' @importFrom dplyr rows_patch
#' @rdname rows-db
rows_patch.tbl_lazy <- function(
  x,
  y,
  by = NULL,
  ...,
  unmatched = c("error", "ignore"),
  copy = "none",
  in_place = FALSE,
  returning = NULL
) {
  check_dots_empty()
  rows_check_in_place(x, in_place)
  table <- target_table(x, in_place)

  rows_check_containment(x, y)
  y <- rows_auto_copy(x, y, copy = copy)

  by <- rows_check_by(by, y)

  rows_check_key(x, by, "x")
  rows_check_key(y, by, "y", unique = TRUE)

  unmatched <- rows_check_ummatched(unmatched)

  new_columns <- setdiff(colnames(y), by)

  returning_cols <- rows_check_returning(x, returning, enexpr(returning))

  if (!is_null(table)) {
    # TODO handle `returning_cols` here
    if (is_empty(new_columns)) {
      return(invisible(x))
    }

    con <- remote_con(x)

    update_cols <- setdiff(colnames(y), by)
    update_values <- sql_coalesce(
      sql_table_prefix(con, table, update_cols),
      sql_table_prefix(con, "...y", update_cols)
    )
    update_values <- set_names(update_values, update_cols)

    sql <- sql_query_update_from(
      con = con,
      table = table,
      from = sql_render(y, remote_con(y), lvl = 1),
      by = by,
      update_values = update_values,
      ...,
      returning_cols = returning_cols
    )

    rows_get_or_execute(x, sql, returning_cols)
  } else {
    to_patch <- inner_join(
      x,
      y,
      by = by,
      suffix = c("", "...y")
    )

    patch_columns_y <- paste0(new_columns, "...y")
    patch_quos <-
      lapply(
        seq_along(new_columns),
        function(.x) {
          quo(coalesce(!!sym(new_columns[.x]), !!sym(patch_columns_y[.x])))
        }
      ) |>
      rlang::set_names(new_columns)
    if (is_empty(new_columns)) {
      patched <- to_patch
      out <- x
    } else {
      patched <- to_patch |>
        mutate(!!!patch_quos) |>
        select(-all_of(patch_columns_y))
      unchanged <- anti_join(x, y, by = by)
      out <- union_all(unchanged, patched)
    }

    if (!is_empty(returning_cols)) {
      returned_rows <- patched |>
        select(!!!returning_cols) |>
        collect()
      out <- set_returned_rows(out, returned_rows)
    }

    out
  }
}

#' @export
#' @inheritParams dplyr::rows_upsert
#'
#' @importFrom dplyr rows_upsert
#' @rdname rows-db
rows_upsert.tbl_lazy <- function(
  x,
  y,
  by = NULL,
  ...,
  copy = "none",
  in_place = FALSE,
  returning = NULL,
  method = NULL
) {
  check_dots_empty()
  rows_check_in_place(x, in_place)
  table <- target_table(x, in_place)

  rows_check_containment(x, y)
  y <- rows_auto_copy(x, y, copy = copy)

  by <- rows_check_by(by, y)

  rows_check_key(x, by, "x")
  rows_check_key(y, by, "y", unique = TRUE)

  returning_cols <- rows_check_returning(x, returning, enexpr(returning))

  new_columns <- setdiff(colnames(y), by)

  if (!is_null(table)) {
    # TODO use `rows_insert()` here
    if (is_empty(new_columns)) {
      return(invisible(x))
    }

    sql <- sql_query_upsert(
      con = remote_con(x),
      table = table,
      from = sql_render(y, remote_con(x), lvl = 1),
      by = by,
      update_cols = setdiff(colnames(y), by),
      ...,
      returning_cols = returning_cols,
      method = method
    )

    rows_get_or_execute(x, sql, returning_cols)
  } else {
    inserted <- anti_join(y, x, by = by)

    if (is_empty(new_columns)) {
      unchanged <- x
      upserted <- inserted
    } else {
      unchanged <- anti_join(x, y, by = by)
      existing_columns <- setdiff(colnames(x), new_columns)
      updated <- x |>
        select(!!!existing_columns) |>
        inner_join(y, by = by)
      upserted <- union_all(updated, inserted)
    }

    out <- union_all(unchanged, upserted)

    if (!is_empty(returning_cols)) {
      returned_rows <- upserted |>
        select(!!!returning_cols) |>
        collect()
      out <- set_returned_rows(out, returned_rows)
    }

    out
  }
}

#' @export
#' @inheritParams dplyr::rows_delete
#'
#' @importFrom dplyr rows_delete
#' @rdname rows-db
rows_delete.tbl_lazy <- function(
  x,
  y,
  by = NULL,
  ...,
  unmatched = c("error", "ignore"),
  copy = "none",
  in_place = FALSE,
  returning = NULL
) {
  check_dots_empty()
  rows_check_in_place(x, in_place)
  table <- target_table(x, in_place)

  rows_check_containment(x, y)
  y <- rows_auto_copy(x, y, copy = copy)

  by <- rows_check_by(by, y)

  rows_check_key(x, by, "x")
  rows_check_key(y, by, "y")

  unmatched <- rows_check_ummatched(unmatched)

  returning_cols <- rows_check_returning(x, returning, enexpr(returning))

  extra <- setdiff(colnames(y), by)
  if (!is_empty(extra)) {
    message <- glue("Ignoring extra `y` columns: ", commas(tick(extra)))
    inform(
      message,
      class = c("dplyr_message_delete_extra_cols", "dplyr_message")
    )
  }

  if (!is_null(table)) {
    sql <- sql_query_delete(
      con = remote_con(x),
      table = table,
      from = sql_render(y, remote_con(x), lvl = 2),
      by = by,
      ...,
      returning_cols = returning_cols
    )

    rows_get_or_execute(x, sql, returning_cols)
  } else {
    out <- anti_join(x, y, by = by)

    if (!is_empty(returning_cols)) {
      returned_rows <- semi_join(x, y, by = by) |>
        select(!!!returning_cols) |>
        collect()
      out <- set_returned_rows(out, returned_rows)
    }

    out
  }
}

set_returned_rows <- function(x, returned_rows) {
  attr(x, "returned_rows") <- as_tibble(returned_rows)
  x
}

#' Extract and check the `RETURNING` rows
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `get_returned_rows()` extracts the `RETURNING` rows produced by
#' [rows_insert()], [rows_append()], [rows_update()], [rows_upsert()],
#' or [rows_delete()] if these are called with the `returning` argument.
#' An error is raised if this information is not available.
#'
#' @param x A lazy tbl.
#'
#' @return For `get_returned_rows()`, a tibble.
#'
#' @export
#' @examples
#' library(dplyr)
#'
#' con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#' DBI::dbExecute(con, "CREATE TABLE Info (
#'    id INTEGER PRIMARY KEY AUTOINCREMENT,
#'    number INTEGER
#' )")
#' info <- tbl(con, "Info")
#'
#' rows1 <- copy_inline(con, data.frame(number = c(1, 5)))
#' rows_insert(info, rows1, conflict = "ignore", in_place = TRUE)
#' info
#'
#' # If the table has an auto incrementing primary key, you can use
#' # the returning argument + `get_returned_rows()` its value
#' rows2 <- copy_inline(con, data.frame(number = c(13, 27)))
#' info <- rows_insert(
#'   info,
#'   rows2,
#'   conflict = "ignore",
#'   in_place = TRUE,
#'   returning = id
#' )
#' info
#' get_returned_rows(info)
get_returned_rows <- function(x) {
  out <- attr(x, "returned_rows", TRUE)
  if (is.null(out)) {
    cli_abort("No returned rows available.")
  }
  out
}

#' has_returned_rows()
#'
#' `has_returned_rows()` checks if `x` has stored RETURNING rows produced by
#' [rows_insert()], [rows_append()], [rows_update()], [rows_upsert()],
#' or [rows_delete()].
#'
#' @param x A lazy tbl.
#'
#' @return For `has_returned_rows()`, a scalar logical.
#'
#' @rdname get_returned_rows
#' @export
has_returned_rows <- function(x) {
  !identical(attr(x, "returned_rows"), NULL)
}

sql_coalesce <- function(x, y) {
  sql(paste0("COALESCE(", x, ", ", y, ")"))
}

# check helpers -----------------------------------------------------------

rows_check_by <- function(by, y, ..., error_call = caller_env()) {
  check_dots_empty()

  if (is.null(by)) {
    if (ncol(y) == 0L) {
      cli_abort("{.arg y} must have at least one column.", call = error_call)
    }

    by <- colnames(y)[[1]]

    inform(
      message = glue("Matching, by = \"{by}\""),
      class = c("dplyr_message_matching_by", "dplyr_message")
    )
  }

  if (!is.character(by)) {
    cli_abort("{.arg by} must be a character vector.", call = error_call)
  }
  if (is_empty(by)) {
    cli_abort("{.arg by} must specify at least 1 column.", call = error_call)
  }
  if (!all(names2(by) == "")) {
    cli_abort("{.arg by} must be unnamed.", call = error_call)
  }

  by
}

rows_check_containment <- function(x, y, ..., error_call = caller_env()) {
  check_dots_empty()

  bad <- setdiff(colnames(y), colnames(x))

  if (!is_empty(bad)) {
    bad <- err_vars(bad)

    message <- c(
      "All columns in {.arg y} must exist in {.arg x}.",
      i = "The following columns only exist in {.arg y}: {.field {bad}}."
    )

    cli_abort(message, call = error_call)
  }

  invisible()
}

rows_check_key <- function(
  x,
  by,
  arg,
  ...,
  unique = FALSE,
  error_call = caller_env()
) {
  check_dots_empty()

  missing <- setdiff(by, colnames(x))

  if (!is_empty(missing)) {
    missing <- err_vars(missing)

    message <- c(
      "All columns specified through {.arg by} must exist in {.arg x} and {.arg y}.",
      i = "The following columns are missing from {.arg {arg}}: {.field {missing}}."
    )

    cli_abort(message, call = error_call)
  }
}

rows_check_in_place <- function(df, in_place, call = caller_env()) {
  check_bool(in_place, call = call)

  if (!in_place) {
    return()
  }

  if (inherits(df, "tbl_TestConnection")) {
    cli_abort(
      "{.code in_place = TRUE} does not work for simulated connections.",
      call = call
    )
  }
}

rows_check_conflict <- function(conflict, error_call = caller_env()) {
  conflict <- arg_match(
    arg = conflict,
    values = c("error", "ignore"),
    error_arg = "conflict",
    error_call = error_call
  )

  check_unsupported_arg(conflict, "ignore", call = error_call)
  conflict
}

rows_check_ummatched <- function(unmatched, error_call = caller_env()) {
  unmatched <- arg_match(
    arg = unmatched,
    values = c("error", "ignore"),
    error_arg = "ummatched",
    error_call = error_call
  )

  check_unsupported_arg(unmatched, "ignore", call = error_call)

  unmatched
}

rows_check_returning <- function(
  df,
  returning,
  returning_expr,
  error_call = caller_env()
) {
  locs <- tidyselect::eval_select(returning_expr, df, error_call = error_call)
  returning_cols <- set_names(colnames(df)[locs], names(locs))

  if (is_empty(returning_cols)) {
    return(returning_cols)
  }

  if (inherits(df, "tbl_TestConnection")) {
    cli_abort(
      "{.arg returning} does not work for simulated connections.",
      call = error_call
    )
  }

  returning_cols
}

err_vars <- function(x) {
  if (is.logical(x)) {
    x <- which(x)
  }
  if (is.character(x)) {
    x <- encodeString(x, quote = "`")
  }

  glue::glue_collapse(
    x,
    sep = ", ",
    last = if (length(x) <= 2) " and " else ", and "
  )
}

commas <- function(...) paste0(..., collapse = ", ")

tick <- function(x) {
  ifelse(is.na(x), "NA", encodeString(x, quote = "`"))
}

# other helpers -----------------------------------------------------------

target_table <- function(x, in_place) {
  # Never touch target table with `in_place = FALSE`
  if (!is_true(in_place)) {
    return(NULL)
  }

  table <- remote_table_path(x)
  if (is_null(table)) {
    cli_abort(
      "Can't determine name for target table. Set {.code in_place = FALSE} to return a lazy table."
    )
  }

  table
}

rows_prep <- function(con, table, from, by, lvl = 0) {
  join_by <- new_join_by(by, x_as = "...y", y_as = table)
  where <- sql_join_tbls(con, by = join_by, na_matches = "never")

  list(
    from = sql_query_wrap(con, from, join_by$x_as, lvl = lvl),
    where = where
  )
}

rows_insert_prep <- function(con, table, from, cols, by, lvl = 0) {
  out <- rows_prep(con, table, from, by, lvl = lvl)

  join_by <- new_join_by(by, x_as = table, y_as = "...y")
  where <- sql_join_tbls(con, by = join_by, na_matches = "never")
  table_sql <- sql_escape_table_source(con, table)
  out$conflict_clauses <- sql_clause_where_exists(table_sql, where, not = TRUE)

  cols_sql <- sql_escape_ident(con, cols)
  out$insert_clause <- sql_clause_insert(cols_sql, table_sql)

  out
}

rows_auto_copy <- function(x, y, copy, call = caller_env()) {
  if (same_src(x, y)) {
    return(y)
  }

  table <- remote_table_path(x)
  x_types <- db_col_types(remote_con(x), table, call)

  if (!is_null(x_types)) {
    rows_check_containment(x, y, error_call = call)
    x_types <- x_types[colnames(y)]
  }

  dbplyr_auto_copy(x, y, copy = copy, types = x_types, call = call)
}

rows_get_or_execute <- function(x, sql, returning_cols, call = caller_env()) {
  error <- "Can't modify database table {.val {remote_name(x)}}."
  con <- remote_con(x)

  if (is_empty(returning_cols)) {
    db_execute(con, sql, error, call = call)
  } else {
    returned_rows <- db_get_query(con, sql, error, call = call)
    x <- set_returned_rows(x, returned_rows)
  }

  invisible(x)
}

# Copy in deprecated progress_estimated() from dplyr
# nocov start
progress_estimated <- function(n, min_time = 0) {
  Progress$new(n, min_time = min_time)
}

#' @importFrom R6 R6Class
Progress <- R6::R6Class("Progress",
  public = list(
    n = NULL,
    i = 0,
    init_time = NULL,
    stopped = FALSE,
    stop_time = NULL,
    min_time = NULL,
    last_update = NULL,

    initialize = function(n, min_time = 0, ...) {
      self$n <- n
      self$min_time <- min_time
      self$begin()
    },

    begin = function() {
      "Initialise timer. Call this before beginning timing."
      self$i <- 0
      self$last_update <- self$init_time <- now()
      self$stopped <- FALSE
      self
    },

    pause = function(x) {
      "Sleep for x seconds. Useful for testing."
      Sys.sleep(x)
      self
    },

    width = function() {
      getOption("width") - nchar("|100% ~ 99.9 h remaining") - 2
    },

    tick = function() {
      "Process one element"
      if (self$stopped) return(self)

      if (self$i == self$n) cli_abort("No more ticks")
      self$i <- self$i + 1
      self
    },

    stop = function() {
      if (self$stopped) return(self)

      self$stopped <- TRUE
      self$stop_time <- now()
      self
    },

    print = function(...) {
      if (!isTRUE(getOption("dplyr.show_progress")) || # user sepecifies no progress
        !interactive() || # not an interactive session
        !is.null(getOption("knitr.in.progress"))) { # dplyr used within knitr document
        return(invisible(self))
      }

      now_ <- now()
      if (now_ - self$init_time < self$min_time || now_ - self$last_update < 0.05) {
        return(invisible(self))
      }
      self$last_update <- now_

      if (self$stopped) {
        overall <- show_time(self$stop_time - self$init_time)
        if (self$i == self$n) {
          cat_line("Completed after ", overall)
          cat("\n")
        } else {
          cat_line("Killed after ", overall)
          cat("\n")
        }
        return(invisible(self))
      }

      avg <- (now() - self$init_time) / self$i
      time_left <- (self$n - self$i) * avg
      nbars <- trunc(self$i / self$n * self$width())

      cat_line(
        "|", str_rep("=", nbars), str_rep(" ", self$width() - nbars), "|",
        format(round(self$i / self$n * 100), width = 3), "% ",
        "~", show_time(time_left), " remaining"
      )

      invisible(self)
    }
  )
)

str_rep <- function(x, i) {
  paste(rep.int(x, i), collapse = "")
}

show_time <- function(x) {
  if (x < 60) {
    paste(round(x), "s")
  } else if (x < 60 * 60) {
    paste(round(x / 60), "m")
  } else {
    paste(round(x / (60 * 60)), "h")
  }
}

now <- function() proc.time()[[3]]
# nocov end

#' PostgreSQL backend
#'
#' @description
#' This backend supports PostgreSQL databases, typically accessed via a
#' `PqConnection` created by [DBI::dbConnect()]. Use `dialect_postgres()` with
#' `lazy_frame()` to see simulated SQL without connecting to a live database.
#'
#' Key differences for this backend are:
#'
#' * Many stringr functions
#' * lubridate date-time extraction functions
#' * More standard statistical summaries
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-postgres
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_postgres())
#' lf |> summarise(x = sd(b, na.rm = TRUE))
#' lf |> summarise(y = cor(b, c), z = cov(b, c))
NULL

#' @export
#' @rdname backend-postgres
dialect_postgres <- function() {
  new_sql_dialect(
    "postgres",
    quote_identifier = function(x) sql_quote(x, '"'),
    has_window_clause = TRUE
  )
}

#' @export
#' @rdname backend-postgres
simulate_postgres <- function() simulate_dbi("PqConnection")

#' @export
sql_dialect.PqConnection <- function(con) {
  dialect_postgres()
}

#' @export
sql_dialect.PostgreSQL <- function(con) {
  dialect_postgres()
}

#' @export
dbplyr_edition.PostgreSQL <- function(con) {
  2L
}
#' @export
dbplyr_edition.PqConnection <- dbplyr_edition.PostgreSQL

#' @export
db_connection_describe.PqConnection <- function(con, ...) {
  info <- DBI::dbGetInfo(con)
  host <- if (info$host == "") "localhost" else info$host

  paste0(
    "postgres ",
    info$serverVersion,
    " [",
    info$username,
    "@",
    host,
    ":",
    info$port,
    "/",
    info$dbname,
    "]"
  )
}
#' @export
db_connection_describe.PostgreSQL <- db_connection_describe.PqConnection

postgres_grepl <- function(
  pattern,
  x,
  ignore.case = FALSE,
  perl = FALSE,
  fixed = FALSE,
  useBytes = FALSE
) {
  # https://www.postgresql.org/docs/current/static/functions-matching.html#FUNCTIONS-POSIX-TABLE
  check_unsupported_arg(perl, FALSE, backend = "PostgreSQL")
  check_unsupported_arg(fixed, FALSE, backend = "PostgreSQL")
  check_unsupported_arg(useBytes, FALSE, backend = "PostgreSQL")
  check_bool(ignore.case)

  if (ignore.case) {
    sql_glue("({x}) ~* ({pattern})")
  } else {
    sql_glue("({x}) ~ ({pattern})")
  }
}
postgres_round <- function(x, digits = 0L) {
  digits <- as.integer(digits)
  sql_glue("ROUND(({x})::numeric, {digits})")
}

# https://neon.com/postgresql/postgresql-date-functions/postgresql-make_interval
# All MAKE_INTERVAL() parameters are integer typed, except `secs`, so we must
# coerce the argument to an integer. R doubles are escaped with a trailing `.0`
# (e.g. `months(1)` gives `1.0`), which MAKE_INTERVAL() rejects.
postgres_period <- function(x, unit) {
  if (unit != "secs") {
    if (is.numeric(x)) {
      x <- as.integer(x)
    } else if (is.ident(x)) {
      x <- sql_glue("CAST({x} AS integer)")
    }
  }
  sql_glue("MAKE_INTERVAL({.sql unit} => {x})")
}

#' @export
sql_translation.sql_dialect_postgres <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_scalar,
      # https://www.postgresql.org/docs/current/functions-comparison.html
      is_distinct_from = \(x, y) sql_glue("({x}) IS DISTINCT FROM ({y})"),
      is_not_distinct_from = function(x, y) {
        sql_glue("({x}) IS NOT DISTINCT FROM ({y})")
      },
      bitwXor = sql_infix("#"),
      log10 = \(x) sql_glue("LOG({x})"),
      log = sql_log(),
      cot = sql_cot(),
      round = postgres_round,
      grepl = postgres_grepl,

      paste = sql_paste(" "),
      paste0 = sql_paste(""),

      # stringr functions
      # https://www.postgresql.org/docs/9.1/functions-string.html
      # https://www.postgresql.org/docs/9.1/functions-matching.html#FUNCTIONS-POSIX-REGEXP
      str_c = sql_paste(""),

      str_locate = function(string, pattern) {
        sql_glue("STRPOS({string}, {pattern})")
      },
      # https://www.postgresql.org/docs/9.1/functions-string.html
      str_detect = function(string, pattern, negate = FALSE) {
        sql_str_pattern_switch(
          string = string,
          pattern = {{ pattern }},
          negate = negate,
          f_fixed = sql_str_detect_fixed_position("detect"),
          f_regex = function(string, pattern, negate = FALSE) {
            if (isTRUE(negate)) {
              sql_glue("!({string} ~ {pattern})")
            } else {
              sql_glue("{string} ~ {pattern}")
            }
          }
        )
      },
      str_ends = function(string, pattern, negate = FALSE) {
        sql_str_pattern_switch(
          string = string,
          pattern = {{ pattern }},
          negate = negate,
          f_fixed = sql_str_detect_fixed_position("end")
        )
      },
      # https://www.postgresql.org/docs/current/functions-matching.html
      str_like = function(string, pattern, ignore_case = deprecated()) {
        ignore_case <- deprecate_ignore_case(ignore_case)
        if (ignore_case) {
          sql_glue("{string} ILIKE {pattern}")
        } else {
          sql_glue("{string} LIKE {pattern}")
        }
      },
      str_ilike = function(string, pattern) {
        sql_glue("{string} ILIKE {pattern}")
      },
      str_replace = function(string, pattern, replacement) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement})")
      },
      str_replace_all = function(string, pattern, replacement) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement}, 'g')")
      },
      str_squish = function(string) {
        sql_glue("LTRIM(RTRIM(REGEXP_REPLACE({string}, '\\s+', ' ', 'g')))")
      },
      str_remove = function(string, pattern) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, '')")
      },
      str_remove_all = function(string, pattern) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, '', 'g')")
      },
      str_starts = function(string, pattern, negate = FALSE) {
        sql_str_pattern_switch(
          string = string,
          pattern = {{ pattern }},
          negate = negate,
          f_fixed = sql_str_detect_fixed_position("start")
        )
      },

      # lubridate functions
      # https://www.postgresql.org/docs/9.1/functions-datetime.html
      day = function(x) {
        sql_glue("EXTRACT(DAY FROM {x})")
      },
      mday = function(x) {
        sql_glue("EXTRACT(DAY FROM {x})")
      },
      wday = function(x, label = FALSE, abbr = TRUE, week_start = NULL) {
        check_bool(label)
        check_bool(abbr)
        check_number_whole(week_start, allow_null = TRUE)
        if (!label) {
          week_start <- week_start %||% getOption("lubridate.week.start", 7)
          offset <- as.integer(7 - week_start)
          sql_glue("EXTRACT('dow' FROM DATE({x}) + {offset}) + 1")
        } else if (label && !abbr) {
          sql_glue("TO_CHAR({x}, 'Day')")
        } else if (label && abbr) {
          sql_glue("SUBSTR(TO_CHAR({x}, 'Day'), 1, 3)")
        } else {
          cli_abort("Unrecognized arguments to {.arg wday}")
        }
      },
      yday = \(x) sql_glue("EXTRACT(DOY FROM {x})"),
      week = function(x) {
        sql_glue("FLOOR((EXTRACT(DOY FROM {x}) - 1) / 7) + 1")
      },
      isoweek = function(x) {
        sql_glue("EXTRACT(WEEK FROM {x})")
      },
      month = function(x, label = FALSE, abbr = TRUE) {
        check_bool(label)
        check_bool(abbr)
        if (!label) {
          sql_glue("EXTRACT(MONTH FROM {x})")
        } else {
          if (abbr) {
            sql_glue("TO_CHAR({x}, 'Mon')")
          } else {
            sql_glue("TO_CHAR({x}, 'Month')")
          }
        }
      },
      quarter = function(x, with_year = FALSE, fiscal_start = 1) {
        check_bool(with_year)
        check_unsupported_arg(fiscal_start, 1, backend = "PostgreSQL")

        if (with_year) {
          sql_glue(
            "(EXTRACT(YEAR FROM {x}) || '.' || EXTRACT(QUARTER FROM {x}))"
          )
        } else {
          sql_glue("EXTRACT(QUARTER FROM {x})")
        }
      },
      isoyear = function(x) {
        sql_glue("EXTRACT(YEAR FROM {x})")
      },

      seconds = function(x) postgres_period(x, "secs"),
      minutes = function(x) postgres_period(x, "mins"),
      hours = function(x) postgres_period(x, "hours"),
      days = function(x) postgres_period(x, "days"),
      weeks = function(x) postgres_period(x, "weeks"),
      months = function(x) postgres_period(x, "months"),
      years = function(x) postgres_period(x, "years"),

      # https://www.postgresql.org/docs/current/functions-datetime.html#FUNCTIONS-DATETIME-TRUNC
      floor_date = function(x, unit = "seconds") {
        unit <- arg_match(
          unit,
          c(
            "second",
            "minute",
            "hour",
            "day",
            "week",
            "month",
            "quarter",
            "year"
          )
        )
        sql_glue("DATE_TRUNC({unit}, {x})")
      },

      # clock ---------------------------------------------------------------
      add_days = function(x, n, ...) {
        check_dots_empty()
        sql_glue("({x} + {n}*INTERVAL'1 day')")
      },
      add_years = function(x, n, ...) {
        check_dots_empty()
        sql_glue("({x} + {n}*INTERVAL'1 year')")
      },
      date_build = function(year, month = 1L, day = 1L, ..., invalid = NULL) {
        check_unsupported_arg(invalid, allow_null = TRUE)
        sql_glue("MAKE_DATE({year}, {month}, {day})")
      },
      date_count_between = function(start, end, precision, ..., n = 1L) {
        check_dots_empty()
        check_unsupported_arg(precision, allowed = "day")
        check_unsupported_arg(n, allowed = 1L)

        sql_glue("{end} - {start}")
      },
      get_year = function(x) {
        sql_glue("DATE_PART('year', {x})")
      },
      get_month = function(x) {
        sql_glue("DATE_PART('month', {x})")
      },
      get_day = function(x) {
        sql_glue("DATE_PART('day', {x})")
      },

      difftime = function(time1, time2, tz, units = "days") {
        check_unsupported_arg(tz)
        check_unsupported_arg(units, allowed = "days")

        sql_glue("(CAST({time1} AS DATE) - CAST({time2} AS DATE))")
      },
    ),
    sql_translator(
      .parent = base_agg,
      cor = sql_aggregate_2("CORR"),
      cov = sql_aggregate_2("COVAR_SAMP"),
      sd = sql_aggregate("STDDEV_SAMP", "sd"),
      var = sql_aggregate("VAR_SAMP", "var"),
      all = sql_aggregate("BOOL_AND", "all"),
      any = sql_aggregate("BOOL_OR", "any"),
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        sql_glue("STRING_AGG({x}, {collapse})")
      }
    ),
    sql_translator(
      .parent = base_win,
      cor = win_aggregate_2("CORR"),
      cov = win_aggregate_2("COVAR_SAMP"),
      sd = win_aggregate("STDDEV_SAMP"),
      var = win_aggregate("VAR_SAMP"),
      all = win_aggregate("BOOL_AND"),
      any = win_aggregate("BOOL_OR"),
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        win_over(
          sql_glue("STRING_AGG({x}, {collapse})"),
          partition = win_current_group(),
          order = win_current_order()
        )
      },
      median = sql_win_not_supported("median", "PostgreSQL"),
      quantile = sql_win_not_supported("quantile", "PostgreSQL")
    )
  )
}

# http://www.postgresql.org/docs/9.3/static/sql-explain.html
#' @export
sql_query_explain.sql_dialect_postgres <- function(
  con,
  sql,
  format = "text",
  ...
) {
  format <- match.arg(format, c("text", "json", "yaml", "xml"))

  if (!is.null(format)) {
    sql_glue2(con, "EXPLAIN (FORMAT {format}) {.sql sql}")
  } else {
    sql_glue2(con, "EXPLAIN {.sql sql}")
  }
}

#' @export
sql_query_insert.sql_dialect_postgres <- function(
  con,
  table,
  from,
  insert_cols,
  by,
  conflict = c("error", "ignore"),
  ...,
  returning_cols = NULL,
  method = NULL
) {
  table <- as_table_path(table, con)

  check_string(method, allow_null = TRUE)
  method <- method %||% "on_conflict"
  arg_match(method, c("on_conflict", "where_not_exists"), error_arg = "method")
  if (method == "where_not_exists") {
    return(NextMethod("sql_query_insert"))
  }

  # https://stackoverflow.com/questions/17267417/how-to-upsert-merge-insert-on-duplicate-update-in-postgresql
  # https://www.sqlite.org/lang_UPSERT.html
  conflict <- rows_check_conflict(conflict)

  parts <- rows_insert_prep(con, table, from, insert_cols, by, lvl = 0)
  by_sql <- sql_escape_ident(con, by)

  clauses <- list(
    parts$insert_clause,
    sql_clause_select(sql("*")),
    sql_clause_from(parts$from),
    sql_clause("ON CONFLICT", by_sql, parens = TRUE),
    {
      if (conflict == "ignore") sql("DO NOTHING")
    },
    sql_returning_cols(con, returning_cols, table)
  )
  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_upsert.sql_dialect_postgres <- function(
  con,
  table,
  from,
  by,
  update_cols,
  ...,
  returning_cols = NULL,
  method = NULL
) {
  check_string(method, allow_null = TRUE)
  method <- method %||% "on_conflict"
  arg_match(method, c("cte_update", "on_conflict"), error_arg = "method")

  if (method == "cte_update") {
    return(NextMethod("sql_query_upsert"))
  }

  sql_query_upsert_on_conflict(
    con,
    table,
    from,
    by,
    update_cols,
    returning_cols = returning_cols
  )
}

# Shared upsert implementation for Postgres and SQLite (ON CONFLICT syntax)
# https://stackoverflow.com/questions/17267417/how-to-upsert-merge-insert-on-duplicate-update-in-postgresql
# https://www.sqlite.org/lang_UPSERT.html
sql_query_upsert_on_conflict <- function(
  con,
  table,
  from,
  by,
  update_cols,
  returning_cols = NULL
) {
  table <- as_table_path(table, con)
  parts <- rows_prep(con, table, from, by, lvl = 0)

  insert_cols <- c(by, update_cols)
  insert_cols_sql <- sql_escape_ident(con, insert_cols)

  update_values <- set_names(
    sql_table_prefix(con, "excluded", update_cols),
    update_cols
  )
  update_cols <- sql_escape_ident(con, update_cols)

  by_sql <- sql_escape_ident(con, by)
  table_sql <- sql_escape_table_source(con, table)
  clauses <- list(
    sql_clause_insert(insert_cols_sql, into = table_sql),
    sql_clause_select(insert_cols_sql),
    sql_clause_from(parts$from),
    # `WHERE true` is required for SQLite
    sql("WHERE true"),
    sql_clause("ON CONFLICT", by_sql, parens = TRUE),
    sql("DO UPDATE"),
    sql_clause_set(update_cols, update_values),
    sql_returning_cols(con, returning_cols, table)
  )
  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_values_subquery.sql_dialect_postgres <- sql_values_subquery_column_alias

#' @export
sql_escape_date.sql_dialect_postgres <- function(con, x) {
  sql(paste0(sql_quote(as.character(x), "'"), "::date"))
}

#' @export
sql_escape_datetime.sql_dialect_postgres <- function(con, x) {
  x <- strftime(x, "%Y-%m-%dT%H:%M:%OSZ", tz = "UTC")
  sql(paste0(sql_quote(x, "'"), "::timestamp"))
}

#' @export
db_col_types.PqConnection <- function(con, table, call) {
  table <- as_table_path(table, con, error_call = call)

  sql <- sql_glue2(con, "SELECT * FROM {.tbl table} LIMIT 0")
  res <- DBI::dbSendQuery(con, sql)
  on.exit(DBI::dbClearResult(res))

  DBI::dbFetch(res, n = 0)
  col_info_df <- DBI::dbColumnInfo(res)
  set_names(col_info_df[[".typname"]], col_info_df[["name"]])
}

#' @export
db_col_types.PostgreSQL <- db_col_types.PqConnection

test_that("with_dialect() validates inputs", {
  con <- simulate_dbi()

  expect_snapshot(error = TRUE, {
    with_dialect(NULL, dialect_postgres())
    with_dialect(con, "postgres")
  })
})

test_that("with_dialect() returns correct dialect", {
  con <- simulate_dbi()
  wrapped <- with_dialect(con, dialect_postgres())

  expect_s3_class(sql_dialect(wrapped), "sql_dialect_postgres")
})

test_that("SQL generation uses specified dialect", {
  con <- with_dialect(memdb(), dialect_postgres())
  mtcars <- local_db_table(con, mtcars, "mtcars")
  expect_snapshot(mtcars |> summarise(mpg = sd(mpg), .by = cyl) |> show_query())
})

test_that("verbs work through wrapper", {
  skip_if_not_installed("RSQLite")

  # For simple operations there's no difference between sqlite and postgres
  wrapped <- with_dialect(memdb(), dialect_postgres())

  # copy_to()
  db_mtcars <- copy_to(wrapped, mtcars, name = "mtcars")
  withr::defer(DBI::dbRemoveTable(wrapped, "mtcars"))
  expect_true(DBI::dbExistsTable(wrapped, "mtcars"))

  # compute()
  result <- db_mtcars |>
    filter(cyl == 4) |>
    select(cyl, vs, am) |>
    compute(name = "mtcars_4cyl")
  expect_true(DBI::dbExistsTable(wrapped, "mtcars_4cyl"))
  withr::defer(DBI::dbRemoveTable(wrapped, "mtcars_4cyl"))

  # collect()
  collected <- db_mtcars |>
    filter(mpg > 30) |>
    collect()
  expect_equal(nrow(collected), 4)
})

test_that("simulate_postgres() still works", {
  expect_translation(simulate_postgres(), x + 1, '"x" + 1.0')
})

test_that("custom scalar translated correctly", {
  con <- dialect_postgres()

  expect_translation(con, bitwXor(x, 128L), "\"x\" # 128")
  expect_translation(con, log10(x), "LOG(\"x\")")
  expect_translation(con, log(x), "LN(\"x\")")
  expect_translation(con, log(x, 2), "LOG(\"x\") / LOG(2.0)")
  expect_translation(con, cot(x), "1 / TAN(\"x\")")
  expect_translation(con, round(x, digits = 1.1), 'ROUND(("x")::numeric, 1)')
  expect_translation(con, grepl("exp", x), "(\"x\") ~ ('exp')")
  expect_translation(con, grepl("exp", x, TRUE), "(\"x\") ~* ('exp')")
  expect_translation(con, substr("test", 2, 3), "SUBSTR('test', 2, 2)")
})

test_that("custom stringr functions translated correctly", {
  con <- dialect_postgres()

  expect_translation(con, str_detect(x, y), "\"x\" ~ \"y\"")
  expect_translation(con, str_detect(x, y, negate = TRUE), "!(\"x\" ~ \"y\")")
  expect_translation(con, str_like(x, y), "\"x\" LIKE \"y\"")
  expect_translation(con, str_ilike(x, y), "\"x\" ILIKE \"y\"")

  expect_translation(
    con,
    str_replace(x, y, z),
    "REGEXP_REPLACE(\"x\", \"y\", \"z\")"
  )
  expect_translation(
    con,
    str_replace_all(x, y, z),
    "REGEXP_REPLACE(\"x\", \"y\", \"z\", 'g')"
  )
  expect_translation(
    con,
    str_squish(x),
    "LTRIM(RTRIM(REGEXP_REPLACE(\"x\", '\\s+', ' ', 'g')))"
  )
  expect_translation(con, str_remove(x, y), "REGEXP_REPLACE(\"x\", \"y\", '')")
  expect_translation(
    con,
    str_remove_all(x, y),
    "REGEXP_REPLACE(\"x\", \"y\", '', 'g')"
  )

  expect_translation(
    con,
    str_detect(x, fixed("%0")),
    "POSITION('%0' in \"x\") > 0"
  )
  expect_translation(
    con,
    str_starts(x, fixed("%0")),
    "POSITION('%0' in \"x\") = 1"
  )
  expect_translation(
    con,
    str_ends(x, fixed("%0")),
    "POSITION('%0' in \"x\") = ((LENGTH(\"x\") - LENGTH('%0')) + 1)"
  )
})

test_that("two variable aggregates are translated correctly", {
  con <- dialect_postgres()

  expect_translation(con, cor(x, y), "CORR(\"x\", \"y\")", window = FALSE)
  expect_translation(
    con,
    cor(x, y),
    "CORR(\"x\", \"y\") OVER ()",
    window = TRUE
  )
})

test_that("pasting translated correctly", {
  con <- dialect_postgres()

  expect_translation(
    con,
    paste(x, y),
    "CONCAT_WS(' ', \"x\", \"y\")",
    window = FALSE
  )
  expect_translation(
    con,
    paste0(x, y),
    "CONCAT_WS('', \"x\", \"y\")",
    window = FALSE
  )

  expect_snapshot(
    error = TRUE,
    translate_sql(paste0(x, collapse = ""), con = con, window = FALSE)
  )
})

test_that("postgres mimics two argument log", {
  con <- dialect_postgres()

  expect_translation(con, log(x), "LN(\"x\")")
  expect_translation(con, log(x, 10), "LOG(\"x\") / LOG(10.0)")
  expect_translation(con, log(x, 10L), "LOG(\"x\") / LOG(10)")
})

test_that("custom lubridate functions translated correctly", {
  con <- dialect_postgres()

  expect_translation(con, day(x), "EXTRACT(DAY FROM \"x\")")
  expect_translation(con, mday(x), "EXTRACT(DAY FROM \"x\")")
  expect_translation(con, yday(x), "EXTRACT(DOY FROM \"x\")")
  expect_translation(
    con,
    week(x),
    "FLOOR((EXTRACT(DOY FROM \"x\") - 1) / 7) + 1"
  )
  expect_translation(con, isoweek(x), "EXTRACT(WEEK FROM \"x\")")
  expect_translation(con, quarter(x), "EXTRACT(QUARTER FROM \"x\")")
  expect_translation(
    con,
    quarter(x, with_year = TRUE),
    "(EXTRACT(YEAR FROM \"x\") || '.' || EXTRACT(QUARTER FROM \"x\"))"
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(quarter(x, fiscal_start = 2), con = con)
  )
  expect_translation(con, isoyear(x), "EXTRACT(YEAR FROM \"x\")")

  expect_translation(con, seconds(x), "MAKE_INTERVAL(secs => \"x\")")
  expect_translation(
    con,
    days(x),
    "MAKE_INTERVAL(days => CAST(\"x\" AS integer))"
  )

  # Whole-number doubles are rendered as integer literals (#1848)
  expect_translation(con, seconds(1.5), "MAKE_INTERVAL(secs => 1.5)")
  expect_translation(con, days(1), "MAKE_INTERVAL(days => 1)")
  expect_translation(con, months(1), "MAKE_INTERVAL(months => 1)")

  expect_translation(con, floor_date(x, "month"), "DATE_TRUNC('month', \"x\")")
  expect_translation(con, floor_date(x, "week"), "DATE_TRUNC('week', \"x\")")
})

test_that("custom clock functions translated correctly", {
  con <- dialect_postgres()
  expect_translation(con, add_years(x, 1), "(\"x\" + 1.0*INTERVAL'1 year')")
  expect_translation(con, add_days(x, 1), "(\"x\" + 1.0*INTERVAL'1 day')")
  expect_error(
    translate_sql(add_days(x, 1, "dots", "must", "be empty"), con = con),
    class = "rlib_error_dots_nonempty"
  )
  expect_translation(con, date_build(2020, 1, 1), "MAKE_DATE(2020.0, 1.0, 1.0)")
  expect_translation(
    con,
    date_build(year_column, 1L, 1L),
    "MAKE_DATE(\"year_column\", 1, 1)"
  )
  expect_translation(
    con,
    get_year(date_column),
    "DATE_PART('year', \"date_column\")"
  )
  expect_translation(
    con,
    get_month(date_column),
    "DATE_PART('month', \"date_column\")"
  )
  expect_translation(
    con,
    get_day(date_column),
    "DATE_PART('day', \"date_column\")"
  )
  expect_translation(
    con,
    date_count_between(date_column_1, date_column_2, "day"),
    "\"date_column_2\" - \"date_column_1\""
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      date_count_between(date_column_1, date_column_2, "year"),
      con = con
    )
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      date_count_between(
        date_column_1,
        date_column_2,
        "day",
        n = 5
      ),
      con = con
    )
  )
})

test_that("difftime is translated correctly", {
  con <- dialect_postgres()
  expect_translation(
    con,
    difftime(start_date, end_date, units = "days"),
    "(CAST(\"start_date\" AS DATE) - CAST(\"end_date\" AS DATE))"
  )

  expect_translation(
    con,
    difftime(start_date, end_date),
    "(CAST(\"start_date\" AS DATE) - CAST(\"end_date\" AS DATE))"
  )

  expect_snapshot(
    error = TRUE,
    translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      difftime(
        start_date,
        end_date,
        tz = "UTC",
        units = "days"
      ),
      con = con
    )
  )
})

test_that("custom window functions translated correctly", {
  con <- dialect_postgres()

  expect_snapshot(
    error = TRUE,
    translate_sql(quantile(x, 0.3, na.rm = TRUE), con = con, window = TRUE)
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(median(x, na.rm = TRUE), con = con, window = TRUE)
  )
})

test_that("custom SQL translation", {
  lf <- lazy_frame(x = 1, con = dialect_postgres())
  expect_snapshot(left_join(lf, lf, by = "x", na_matches = "na"))

  con <- dialect_postgres()
  expect_snapshot(
    copy_inline(con, tibble(x = integer(), y = character())) |> remote_query()
  )
  expect_snapshot(
    copy_inline(con, tibble(x = 1:2, y = letters[1:2])) |> remote_query()
  )
})

test_that("`sql_query_insert()` works", {
  con <- dialect_postgres()
  df_y <- lazy_frame(
    a = 2:3,
    b = c(12L, 13L),
    c = -(2:3),
    d = c("y", "z"),
    con = con,
    .name = "df_y"
  ) |>
    mutate(c = c + 1)

  expect_snapshot(
    error = TRUE,
    (sql_query_insert(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      insert_cols = colnames(df_y),
      by = c("a", "b"),
      conflict = "error",
      returning_cols = c("a", b2 = "b")
    ))
  )

  expect_snapshot(
    sql_query_insert(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      insert_cols = colnames(df_y),
      by = c("a", "b"),
      conflict = "ignore",
      returning_cols = c("a", b2 = "b")
    )
  )
})

test_that("`sql_query_upsert()` with method = 'on_conflict' is correct", {
  con <- dialect_postgres()
  df_y <- lazy_frame(
    a = 2:3,
    b = c(12L, 13L),
    c = -(2:3),
    d = c("y", "z"),
    con = con,
    .name = "df_y"
  ) |>
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_upsert(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      by = c("c", "d"),
      update_cols = c("a", "b"),
      returning_cols = c("a", b2 = "b"),
      method = "on_conflict"
    )
  )
})


# live database -----------------------------------------------------------

test_that("can explain", {
  db <- copy_to(
    test_postgres(),
    data.frame(x = 1:3),
    name = "test",
    overwrite = TRUE
  )
  expect_snapshot(db |> mutate(y = x + 1) |> explain())
})

test_that("can overwrite temp tables", {
  src <- test_postgres()
  copy_to(src, mtcars, "mtcars", overwrite = TRUE)
  withr::defer(DBI::dbRemoveTable(src, "mtcars"))
  expect_no_error(copy_to(src, mtcars, "mtcars", overwrite = TRUE))
})

test_that("copy_inline works", {
  src <- test_postgres()
  df <- tibble(
    lgl = TRUE,
    int = 1L,
    dbl = 1.5,
    chr = "a",
    date = as.Date("2020-01-01", tz = "UTC"),
    dtt = as.POSIXct("2020-01-01 01:23:45", tz = "UTC")
  )

  expect_equal(copy_inline(src, df) |> collect(), df)
})

test_that("can insert with returning", {
  con <- test_postgres()

  df_x <- tibble(a = 1L, b = 11L, c = 1L, d = "a")
  x <- local_db_table(con, df_x, "df_x")

  df_y <- tibble(a = 1:3, b = 11:13, c = -(1:3), d = c("x", "y", "z"))
  y <- local_db_table(con, df_y, "df_y") |>
    mutate(c = c + 1)

  # This errors because there is no unique constraint on (`a`, `b`)
  expect_snapshot(
    error = TRUE,
    {
      rows_insert(
        x,
        y,
        by = c("a", "b"),
        in_place = TRUE,
        conflict = "ignore",
        returning = everything(),
        method = "on_conflict"
      )
    },
    transform = snap_transform_dbi
  )

  expect_no_error(
    rows_insert(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      conflict = "ignore",
      returning = everything(),
      method = "where_not_exists"
    )
  )

  x <- local_db_table(con, df_x, "df_x2", overwrite = TRUE)
  dbplyr_create_index(con, "df_x2", columns = c("a", "b"), unique = TRUE)

  expect_equal(
    rows_insert(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      conflict = "ignore",
      returning = everything(),
      method = "on_conflict"
    ) |>
      get_returned_rows(),
    tibble(
      a = 2:3,
      b = 12:13,
      c = c(-1L, -2L),
      d = c("y", "z")
    )
  )

  expect_equal(
    collect(x),
    tibble(a = 1:3, b = 11:13, c = c(1L, -1L, -2L), d = c("a", "y", "z"))
  )
})

test_that("can use `rows_*()` inside a transaction #1183", {
  con <- test_postgres()

  local_db_table(con, tibble(a = 1:2e3, b = 2, x = "a"), "df_x")

  expect_no_error(
    DBI::dbWithTransaction(
      con,
      {
        dbplyr:::db_col_types(con, "df_x", rlang::current_env())
        DBI::dbGetQuery(con, "SELECT * FROM df_x LIMIT 1")
      }
    )
  )
})

test_that("casts `y` column for local df", {
  con <- test_postgres()

  DBI::dbExecute(con, "CREATE SCHEMA dbplyr_test_schema")
  withr::defer(DBI::dbExecute(con, "DROP SCHEMA dbplyr_test_schema CASCADE"))
  df <- tibble(id = 1L, val = 10L, arr = "{1,2}")
  types <- c(id = "bigint", val = "bigint", arr = "integer[]")
  local_db_table(con, value = df, types = types, temporary = FALSE, "df_x")
  table2 <- DBI::Id(schema = "dbplyr_test_schema", table = "df_x2")
  local_db_table(con, value = df, types = types, temporary = FALSE, table2)

  y <- tibble(
    id = "2",
    val = 20,
    arr = "{1, 2, 3}"
  )

  out <- tibble(
    id = bit64::as.integer64(1:2),
    val = bit64::as.integer64(c(10L, 20L)),
    arr = structure(c("{1,2}", "{1,2,3}"), class = "pq__int4")
  )
  expect_equal(
    rows_append(
      tbl(con, "df_x"),
      y,
      copy = TRUE,
      in_place = FALSE
    ) |>
      collect(),
    out
  )

  rows_append(
    tbl(con, "df_x"),
    y,
    copy = TRUE,
    in_place = TRUE
  )

  # also works with schema
  rows_append(
    tbl(con, table2),
    y,
    copy = TRUE,
    in_place = TRUE
  )

  expect_equal(tbl(con, "df_x") |> collect(), out)

  types_expected <- c(id = "int8", val = "int8", arr = "_int4")
  expect_equal(db_col_types(con, table2), types_expected)
  expect_equal(db_col_types(con, in_schema("public", "df_x")), types_expected)
})

test_that("can upsert with returning", {
  con <- test_postgres()

  df_x <- tibble(a = 1:2, b = 11:12, c = 1:2, d = c("a", "b"))
  x <- local_db_table(con, df_x, "df_x")

  df_y <- tibble(a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"))
  y <- local_db_table(con, df_y, "df_y") |>
    mutate(c = c + 1)

  # Errors because there is no unique index
  expect_snapshot(
    error = TRUE,
    {
      rows_upsert(
        x,
        y,
        by = c("a", "b"),
        in_place = TRUE,
        returning = everything(),
        method = "on_conflict"
      )
    },
    transform = snap_transform_dbi
  )

  # DBI method does not need a unique index
  expect_no_error(
    rows_upsert(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      returning = everything(),
      method = "cte_update"
    )
  )

  x <- local_db_table(con, df_x, "df_x2")
  dbplyr_create_index(con, "df_x2", columns = c("a", "b"), unique = TRUE)

  expect_equal(
    rows_upsert(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      returning = everything(),
      method = "on_conflict"
    ) |>
      get_returned_rows() |>
      arrange(a),
    tibble(
      a = 2:3,
      b = 12:13,
      c = c(-1L, -2L),
      d = c("y", "z")
    )
  )

  expect_equal(
    collect(x),
    tibble(a = 1:3, b = 11:13, c = c(1L, -1L, -2L), d = c("a", "y", "z"))
  )
})

test_that("correctly escapes dates", {
  con <- test_postgres()

  dd <- as.Date("2022-03-04")
  expect_equal(escape(dd, con = con), sql("'2022-03-04'::date"))
})

test_that("uses FETCH FIRST instead of LIMIT", {
  mf <- lazy_frame(x = 1, y = 2, con = dialect_db2())
  expect_snapshot(mf |> head())
  expect_snapshot(mf |> head(10))
})

test_that("uses double quotes for identifiers", {
  con <- dialect_db2()
  expect_translation(con, first_name, '"first_name"')
})

test_that("queries with WHERE and ORDER BY work", {
  mf <- lazy_frame(x = 1, y = 2, con = dialect_db2())
  expect_snapshot(
    mf |>
      filter(x > 0) |>
      arrange(y) |>
      head(5)
  )
})

test_that("custom casts translated correctly", {
  con <- dialect_db2()

  expect_translation(con, as.character(x), 'CAST("x" AS VARCHAR(255))')
  expect_translation(con, as.numeric(x), 'CAST("x" AS DOUBLE)')
  expect_translation(con, as.double(x), 'CAST("x" AS DOUBLE)')
  expect_translation(con, as.integer(x), 'CAST("x" AS INTEGER)')
  expect_translation(con, as.integer64(x), 'CAST("x" AS BIGINT)')
})

test_that("pasting translated correctly", {
  con <- dialect_db2()

  expect_translation(con, paste(x, y), '"x" || \' \' || "y"')
  expect_translation(con, paste0(x, y), '"x" || "y"')
  expect_translation(con, str_c(x, y), '"x" || "y"')
})

test_that("runif uses RAND()", {
  con <- dialect_db2()
  expect_translation(con, runif(n()), "RAND()")
})

test_that("string functions translated correctly", {
  con <- dialect_db2()

  expect_translation(
    con,
    str_replace(x, y, z),
    'REGEXP_REPLACE("x", "y", "z", 1, 1)'
  )
  expect_translation(
    con,
    str_replace_all(x, y, z),
    'REGEXP_REPLACE("x", "y", "z")'
  )
  expect_translation(
    con,
    str_remove(x, y),
    'REGEXP_REPLACE("x", "y", \'\', 1, 1)'
  )
  expect_translation(
    con,
    str_remove_all(x, y),
    'REGEXP_REPLACE("x", "y", \'\')'
  )
  expect_translation(
    con,
    str_squish(x),
    "LTRIM(RTRIM(REGEXP_REPLACE(\"x\", '\\s+', ' ')))"
  )
  expect_translation(con, str_locate(x, "ab"), "LOCATE('ab', \"x\")")
  expect_translation(
    con,
    str_detect(x, fixed("ab")),
    "POSITION('ab' in \"x\") > 0"
  )
  expect_translation(
    con,
    str_starts(x, fixed("ab")),
    "POSITION('ab' in \"x\") = 1"
  )
})

test_that("custom lubridate functions translated correctly", {
  con <- dialect_db2()

  expect_translation(con, today(), "CURRENT DATE")
  expect_translation(con, now(), "CURRENT TIMESTAMP")
  expect_translation(con, yday(x), 'DAYOFYEAR("x")')
  expect_translation(con, quarter(x), 'QUARTER("x")')
  expect_translation(
    con,
    quarter(x, with_year = TRUE),
    "(CHAR(YEAR(\"x\")) || '.' || CHAR(QUARTER(\"x\")))"
  )
  expect_translation(con, week(x), 'WEEK_ISO("x")')
  expect_translation(con, isoweek(x), 'WEEK_ISO("x")')
  expect_translation(con, isoyear(x), 'YEAR("x")')

  expect_snapshot(translate_sql(wday(x), con = con))
  expect_translation(con, wday(x, label = TRUE), 'SUBSTR(DAYNAME("x"), 1, 3)')
  expect_translation(
    con,
    wday(x, label = TRUE, abbr = FALSE),
    'DAYNAME("x")'
  )

  expect_snapshot(
    error = TRUE,
    translate_sql(quarter(x, fiscal_start = 2), con = con)
  )
})

test_that("custom clock functions translated correctly", {
  con <- dialect_db2()

  expect_translation(con, add_days(x, 1L), '("x" + 1 DAYS)')
  expect_translation(con, add_years(x, 1L), '("x" + 1 YEARS)')
  expect_error(
    translate_sql(add_days(x, 1, "must", "be empty"), con = con),
    class = "rlib_error_dots_nonempty"
  )
})

test_that("difftime is translated correctly", {
  con <- dialect_db2()
  expect_translation(
    con,
    difftime(end_date, start_date, units = "days"),
    "(DAYS(CAST(\"end_date\" AS DATE)) - DAYS(CAST(\"start_date\" AS DATE)))"
  )

  expect_snapshot(
    error = TRUE,
    translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
  )
})

test_that("statistical aggregates translated correctly", {
  con <- dialect_db2()

  expect_translation(
    con,
    sd(x, na.rm = TRUE),
    'STDDEV_SAMP("x")',
    window = FALSE
  )
  expect_translation(
    con,
    var(x, na.rm = TRUE),
    'VAR_SAMP("x")',
    window = FALSE
  )
  expect_translation(
    con,
    cor(x, y),
    'CORRELATION("x", "y")',
    window = FALSE
  )
  expect_translation(
    con,
    cov(x, y),
    'COVARIANCE_SAMP("x", "y")',
    window = FALSE
  )
})

test_that("str_flatten uses LISTAGG", {
  con <- dialect_db2()

  expect_translation(
    con,
    str_flatten(x, ", ", na.rm = TRUE),
    "LISTAGG(\"x\", ', ')",
    window = FALSE
  )
  expect_translation(
    con,
    str_flatten(x, ", ", na.rm = TRUE),
    "LISTAGG(\"x\", ', ') OVER ()",
    window = TRUE
  )
})

test_that("sql_table_analyze uses RUNSTATS", {
  con <- dialect_db2()
  expect_snapshot(
    sql_table_analyze(con, ident("mytable"))
  )
})

test_that("copy_inline uses VALUES with column aliases", {
  con <- dialect_db2()
  df <- tibble::tibble(x = 1:2, y = c("a", "b"))
  expect_snapshot(copy_inline(con, df) |> remote_query())
})

#' SQL escaping/quoting generics
#'
#' These generics translate individual values into SQL. The core
#' generics are [DBI::dbQuoteIdentifier()] and[DBI::dbQuoteString]
#' for quoting identifiers and strings, but dbplyr needs additional
#' tools for inserting logical, date, date-time, and raw values into
#' queries.
#'
#' @keywords internal
#' @family generic
#' @name db-quote
#' @aliases NULL
#' @examples
#' con <- simulate_dbi()
#' sql_escape_logical(con, c(TRUE, FALSE, NA))
#' sql_escape_date(con, Sys.Date())
#' sql_escape_date(con, Sys.time())
#' sql_escape_raw(con, charToRaw("hi"))
NULL

#' @rdname db-quote
#' @export
sql_escape_logical <- function(con, x) {
  UseMethod("sql_escape_logical")
}
#' @export
sql_escape_logical.DBIConnection <- function(con, x) {
  y <- as.character(x)
  y[is.na(x)] <- "NULL"
  y
}

#' @export
#' @rdname db-quote
sql_escape_date <- function(con, x) {
  UseMethod("sql_escape_date")
}
#' @export
sql_escape_date.DBIConnection <- function(con, x) {
  sql_escape_string(con, as.character(x))
}

#' @export
#' @rdname db-quote
sql_escape_datetime <- function(con, x) {
  UseMethod("sql_escape_datetime")
}
#' @export
sql_escape_datetime.DBIConnection <- function(con, x) {
  x <- strftime(x, "%Y-%m-%dT%H:%M:%OSZ", tz = "UTC")
  sql_escape_string(con, x)
}

#' @export
#' @rdname db-quote
sql_escape_raw <- function(con, x) {
  UseMethod("sql_escape_raw")
}
#' @export
sql_escape_raw.DBIConnection <- function(con, x) {
  # Unlike the other escape functions, this is not vectorised because
  # raw "vectors" are scalars in this content

  if (is.null(x)) {
    "NULL"
  } else {
    # SQL-99 standard for BLOB literals
    # https://crate.io/docs/sql-99/en/latest/chapters/05.html#blob-literal-s
    paste0(c("X'", format(x), "'"), collapse = "")
  }
}

#' Build a SQL string.
#'
#' This is a convenience function that should prevent sql injection attacks
#' (which in the context of dplyr are most likely to be accidental not
#' deliberate) by automatically escaping all expressions in the input, while
#' treating bare strings as sql. This is unlikely to prevent any serious
#' attack, but should make it unlikely that you produce invalid sql.
#'
#' This function should be used only when generating `SELECT` clauses,
#' other high level queries, or for other syntax that has no R equivalent.
#' For individual function translations, prefer [sql_expr()].
#'
#' @param ... input to convert to SQL. Use [sql()] to preserve
#'   user input as is (dangerous), and [ident()] to label user
#'   input as sql identifiers (safe)
#' @param .env the environment in which to evaluate the arguments. Should not
#'   be needed in typical use.
#' @param con database connection; used to select correct quoting characters.
#' @keywords internal
#' @export
#' @examples
#' con <- simulate_dbi()
#' build_sql("SELECT * FROM TABLE", con = con)
#' x <- "TABLE"
#' build_sql("SELECT * FROM ", x, con = con)
#' build_sql("SELECT * FROM ", ident(x), con = con)
#' build_sql("SELECT * FROM ", sql(x), con = con)
#'
#' # http://xkcd.com/327/
#' name <- "Robert'); DROP TABLE Students;--"
#' build_sql("INSERT INTO Students (Name) VALUES (", name, ")", con = con)
build_sql <- function(..., .env = parent.frame(), con = sql_current_con()) {
  check_con(con)

  escape_expr <- function(x, con) {
    # If it's a string, leave it as is
    if (is.character(x)) return(x)

    val <- eval_bare(x, .env)
    # Skip nulls, so you can use if statements like in paste
    if (is.null(val)) return("")

    escape(val, con = con)
  }

  pieces <- purrr::map_chr(enexprs(...), escape_expr, con = con)
  sql(paste0(pieces, collapse = ""))
}

#' A dbplyr specific version of glue
#'
#' Similar to the inline markup of cli this function makes SQL generation easier
#' and safer by providing a couple of types. For example
#'
#' ```
#'   glue_sql2(
#'      con,
#'      "CREATE ", if (unique) "UNIQUE ", "INDEX {.name name}",
#'      " ON {.tbl table} ({.col columns*})"
#'    )
#' ```
#'
#' The following types are supported:
#'
#' * .tbl A table identifier, e.g. `DBI::Id()`. Converted via `as_table_ident()`.
#' * .from A subquery or a table identifier. Converted via `as_from()`.
#' * .name A name, e.g. for an index or a subquery. Can be a string or a scalar (quoted) ident.
#' * .col A column or multiple columns if expression ends with `*`.
#' * .kw An SQL keyword - e.g. `SELECT` or `WHERE` - that should be highlighted.
#' * .val Any value - e.g. an integer vector, a Date, SQL - which is escaped as
#'   usual via `escape()`.
#'
#' If no type is specified the value must be a string or scalar SQL and it isn't
#' escaped or collapsed.
#'
#' @noRd
#'
#' @examples
#' glue_sql2(con, "COLLECT STATISTICS {.tbl table}")
#' glue_sql2(con, "{f}({.val x}, {.val y})")
glue_sql2 <- function(.con,
                      ...,
                      .sep = "",
                      .envir = parent.frame(),
                      .open = "{",
                      .close = "}",
                      .na = DBI::SQL("NULL"),
                      .null = "",
                      .comment = "#",
                      .literal = FALSE,
                      .trim = TRUE) {
  sql(glue(
    ...,
    .sep = .sep,
    .envir = .envir,
    .open = .open,
    .close = .close,
    .na = .na,
    .null = .null,
    .comment = .comment,
    .literal = .literal,
    .transformer = sql_quote_transformer(.con),
    .trim = .trim
  ))
}

sql_quote_transformer <- function(connection) {
  function(text, envir) {
    collapse_regex <- "[*][[:space:]]*$"
    should_collapse <- grepl(collapse_regex, text)
    if (should_collapse) {
      text <- sub(collapse_regex, "", text)
    }

    type_regex <- "^\\.(tbl|col|name|from|kw|val) (.*)"
    m <- regexec(type_regex, text)
    is_quoted <- any(m[[1]] != -1)
    if (is_quoted) {
      matches <- regmatches(text, regexec(type_regex, text))[[1]]

      type <- matches[[2]]
      value <- matches[[3]]
    } else {
      value <- text
      type <- "raw"
    }
    value <- eval(parse(text = value, keep.source = FALSE), envir)
    glue_check_collapse(type, should_collapse)

    if (type == "tbl") {
      value <- as_table_path(value, connection)
    } else if (type == "from") {
      value <- as_table_source(value, connection)
    } else if (type == "col") {
      if (is_bare_character(value)) {
        value <- ident(value)
      }
    } else if (type == "name") {
      # allowed should be `ident`, `ident_q` (maybe), string
      if (is_bare_character(value)) {
        value <- ident(value)
      }
    } else if (type == "kw") {
      value <- sql(style_kw(value))
    } else if (type == "val") {
      # keep as is
    } else if (type == "raw") {
      if (!is.sql(value) && !is_string(value)) {
        stop_input_type(
          value,
          what = c("a string", "scalar SQL")
        )
      }
    }

    if (type == "val") {
      if (should_collapse) {
        value <- escape(value, collapse = ", ", parens = FALSE, con = connection)
      } else {
        value <- escape(value, con = connection)
      }
    } else if (type %in% c("tbl", "from", "col", "name")) {
      value <- escape(value, collapse = NULL, parens = FALSE, con = connection)
      if (should_collapse) {
        value <- paste0(unclass(value), collapse = ", ")
      }
    }

    # TODO use `vctrs::vec_check_size(value, size = 1L)`
    if (length(value) != 1) {
      cli_abort("{.arg value} must have size 1, not {length(value)}.")
    }

    unclass(value)
  }
}

glue_check_collapse <- function(type, collapse) {
  if (type %in% c("col", "val")) {
    return()
  }

  if (collapse) {
    cli_abort("Collapsing is only allowed for {.val col} and {.val val}, not for {.val {type}}.")
  }
}

test_that("build_sql() requires connection", {
  x <- ident("TABLE")
  expect_snapshot(error = TRUE, build_sql("SELECT * FROM ", x))
})

test_that("glue_sql() interpolates .tbl correctly", {
  con <- simulate_dbi()
  expect_equal(glue_sql2("{.tbl 'table'}", .con = con), sql("`table`"))
  tbl <- "table"
  expect_equal(glue_sql2("{.tbl tbl}", .con = con), sql("`table`"))
  tbl <- ident("table")
  expect_equal(glue_sql2("{.tbl tbl}", .con = con), sql("`table`"))
  tbl <- in_schema("schema", "table")
  expect_equal(glue_sql2("{.tbl tbl}", .con = con), sql("`schema`.`table`"))
  tbl <- in_catalog("catalog", "schema", "table")
  expect_equal(glue_sql2("{.tbl tbl}", .con = con), sql("`catalog`.`schema`.`table`"))
})

test_that("glue_sql() interpolates .from correctly", {
  con <- simulate_dbi()
  expect_equal(glue_sql2("{.from 'table'}", .con = con), sql("`table`"))
  tbl <- "table"
  expect_equal(glue_sql2("{.from tbl}", .con = con), sql("`table`"))
  tbl <- ident("table")
  expect_equal(glue_sql2("{.from tbl}", .con = con), sql("`table`"))
  tbl <- in_schema("schema", "table")
  expect_equal(glue_sql2("{.from tbl}", .con = con), sql("`schema`.`table`"))
  tbl <- in_catalog("catalog", "schema", "table")
  expect_equal(glue_sql2("{.from tbl}", .con = con), sql("`catalog`.`schema`.`table`"))

  expect_equal(glue_sql2("{.from sql('(SELECT * FROM df_x)')}", .con = con), sql("(SELECT * FROM df_x)"))
})

test_that("glue_sql() interpolates .name correctly", {
  con <- simulate_dbi()
  expect_equal(glue_sql2("{.name 'table'}", .con = con), sql("`table`"))
  expect_equal(glue_sql2("{.name ident('table')}", .con = con), sql("`table`"))
  expect_equal(glue_sql2("{.name ident_q('ta ble')}", .con = con), sql("ta ble"))
})

test_that("glue_sql() interpolates .col correctly", {
  con <- simulate_dbi()
  expect_equal(glue_sql2("{.col 'x'}", .con = con), sql("`x`"))
  expect_equal(glue_sql2("{.col ident('x')}", .con = con), sql("`x`"))
})

test_that("glue_sql() interpolates .kw correctly", {
  withr::local_options(dbplyr_use_colour = TRUE)
  local_reproducible_output(crayon = TRUE)
  withr::local_options(dbplyr_highlight = cli::combine_ansi_styles("blue"))
  con <- simulate_dbi()
  expect_equal(glue_sql2("{.kw 'FROM'}", .con = con), sql("\033[34mFROM\033[39m"))
})

test_that("glue_sql() checks size", {
  con <- simulate_dbi()
  x <- c("a", "b")
  expect_snapshot(error = TRUE, {
    glue_sql2("{.col x}", .con = con)
    glue_sql2("{.col character()}", .con = con)
  })
})

test_that("glue_sql() can collapse", {
  con <- simulate_dbi()
  x <- c("a", "b")
  expect_equal(glue_sql2("{.col x*}", .con = con), sql("`a`, `b`"))
  expect_equal(glue_sql2("{.val x*}", .con = con), sql("'a', 'b'"))

  expect_snapshot(error = TRUE, {
    glue_sql2("{.tbl x*}", .con = con)
    glue_sql2("{.name x*}", .con = con)
    glue_sql2("{.from x*}", .con = con)
  })
})

#' Override the SQL dialect for a connection
#'
#' @description
#' `with_dialect()` overrides the default dialect assigned to a connection.
#' This is useful when dbplyr guesses the dialect incorrectly, which is most
#' likely to occur with ODBC/JDBC/ADBC backends.
#'
#' @param con A database connection (class `DBIConnection`).
#' @param dialect A dialect object created by a `dialect_*()` function
#'   (e.g., [dialect_postgres()], [dialect_sqlite()]).
#' @return A connection object that uses the specified dialect for SQL
#'   generation.
#' @export
#' @examples
#' # Wrap an in-memory SQLite connection to use Postgres dialect
#' con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#' pg_con <- with_dialect(con, dialect_postgres())
#'
#' # SQL generation uses Postgres syntax
#' lf <- lazy_frame(x = 1, con = pg_con)
#' lf |> dplyr::mutate(y = sd(x))
#'
#' DBI::dbDisconnect(con)
with_dialect <- function(con, dialect) {
  check_con(con)
  if (!is_sql_dialect(dialect)) {
    cli_abort("{.arg dialect} must be a dialect object.")
  }

  # This is a little hacky, but it seems to be ok to add additional attributes
  # to S4 objects, and namespacing the attribute ensures that there's little
  # chance of collision.
  attr(con, "dbplyr_dialect") <- dialect
  con
}

#' @export
#' @rdname sql_build
lazy_semi_join_query <- function(
  x,
  y,
  vars,
  anti,
  by,
  where,
  group_vars = op_grps(x),
  order_vars = op_sort(x),
  frame = op_frame(x),
  call = caller_env()
) {
  check_lazy_query(x, call = call)
  check_lazy_query(y, call = call)
  check_bool(anti, call = call)
  check_semi_join_by(by, call)

  lazy_query(
    query_type = "semi_join",
    x = x,
    y = y,
    anti = anti,
    by = by,
    vars = vars,
    where = where,
    group_vars = group_vars,
    order_vars = order_vars,
    frame = frame
  )
}

check_semi_join_by <- function(by, call) {
  if (!vctrs::vec_is_list(by) && !inherits(by, "dplyr_join_by")) {
    cli_abort("{.arg by} must be a list", .internal = TRUE)
  }
  check_character(by$x, call = call)
  check_character(by$y, call = call)
  if (vctrs::vec_size(by$x) != vctrs::vec_size(by$y)) {
    cli_abort(
      "{.arg by$x} and {.arg by$y} must have the same size",
      .internal = TRUE
    )
  }
  check_string(by$x_as, arg = "by$x_as", call = call)
  check_string(by$y_as, arg = "by$y_as", call = call)
}


#' @export
op_vars.lazy_semi_join_query <- function(op) {
  op$vars$name
}

#' @export
sql_build.lazy_semi_join_query <- function(op, con, ..., sql_options = NULL) {
  vars_prev <- op_vars(op$x)
  if (
    sql_options$use_star &&
      identical(op$vars$var, op$vars$name) &&
      identical(op$vars$var, vars_prev)
  ) {
    vars <- sql_star(con, op$by$x_as)
  } else {
    vars <- set_names(sql_escape_ident(con, op$vars$var), op$vars$name)
  }

  # We've introduced aliases to disambiguate the internal and external tables
  # so need to update the existing WHERE clause
  y_vars <- op_vars(op$y)
  y_as <- op$by$y_as
  replacements <- lapply(y_vars, \(var) sql_glue2(con, "{.tbl y_as}.{.id var}"))
  where <- replace_sym(op$where, y_vars, replacements)
  where_sql <- translate_sql_(
    where,
    con = con,
    context = list(clause = "WHERE")
  )

  semi_join_query(
    sql_build(op$x, con, sql_options = sql_options),
    sql_build(op$y, con, sql_options = sql_options),
    vars = vars,
    anti = op$anti,
    by = op$by,
    where = where_sql,
    na_matches = op$by$na_matches
  )
}

# Built query -------------------------------------------------------------

#' @export
#' @rdname sql_build
semi_join_query <- function(
  x,
  y,
  vars,
  anti = FALSE,
  by = NULL,
  where = NULL,
  na_matches = FALSE
) {
  query(
    "semi_join",
    x = x,
    y = y,
    vars = vars,
    anti = anti,
    by = by,
    where = where,
    na_matches = na_matches
  )
}

#' @export
sql_render.semi_join_query <- function(
  query,
  con = NULL,
  ...,
  sql_options = NULL,
  subquery = FALSE,
  lvl = 0
) {
  from_x <- sql_render(query$x, con, ..., subquery = TRUE, lvl = lvl + 1)
  from_y <- sql_render(query$y, con, ..., subquery = TRUE, lvl = lvl + 1)

  dbplyr_query_semi_join(
    con,
    from_x,
    from_y,
    vars = query$vars,
    anti = query$anti,
    by = query$by,
    where = query$where,
    lvl = lvl
  )
}

#' @export
flatten_query.semi_join_query <- function(qry, query_list, con) {
  flatten_query_2_tables(qry, query_list, con)
}

# SQL generation ----------------------------------------------------------

#' @rdname db-sql
#' @export
sql_query_semi_join <- function(
  con,
  x,
  y,
  anti,
  by,
  where,
  vars,
  ...,
  lvl = 0
) {
  check_dots_used()
  UseMethod("sql_query_semi_join", sql_dialect(con))
}
#' @export
sql_query_semi_join.DBIConnection <- function(
  con,
  x,
  y,
  anti,
  by,
  where,
  vars,
  ...,
  lvl = 0
) {
  x <- dbplyr_sql_subquery(con, x, name = by$x_as)
  y <- dbplyr_sql_subquery(con, y, name = by$y_as)

  on <- sql_join_tbls(con, by, na_matches = by$na_matches)

  exists <- if (anti) "NOT EXISTS" else "EXISTS"

  lines <- list(
    sql_clause_select(names_to_as(con, vars)),
    sql_clause_from(x),
    sql_glue2(con, "WHERE {.sql exists} ("),
    # lvl = 1 because they are basically in a subquery
    sql_clause("SELECT 1 FROM", y, lvl = 1),
    sql_clause_where(sql(c(on, where)), lvl = 1),
    sql(")")
  )
  sql_format_clauses(lines, lvl)
}

#' @export
sql_query_semi_join.sql_dialect <- sql_query_semi_join.DBIConnection

dbplyr_query_semi_join <- function(
  con,
  x,
  y,
  anti = FALSE,
  by = NULL,
  ...,
  lvl = 0
) {
  check_2ed(con)
  sql_query_semi_join(con, x, y, anti = anti, by = by, ..., lvl = lvl)
}

#' MS Access backend
#'
#' @description
#' This backend supports Microsoft Access databases, typically accessed via
#' odbc. Use `dialect_access()` with `lazy_frame()` to see simulated SQL without
#' connecting to a live database.
#'
#' Key differences for this backend are:
#'
#' * `SELECT` uses `TOP`, not `LIMIT`
#' * Non-standard types and mathematical functions
#' * String concatenation uses `&`
#' * No `ANALYZE` equivalent
#' * `TRUE` and `FALSE` converted to 1 and 0
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-access
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#' lf <- lazy_frame(x = 1, y = 2, z = "a", con = dialect_access())
#'
#' lf |> head()
#' lf |> mutate(y = as.numeric(y), z = sqrt(x^2 + 10))
#' lf |> mutate(a = paste0(z, " times"))
NULL

#' @export
#' @rdname backend-access
dialect_access <- function() {
  new_sql_dialect(
    "access",
    quote_identifier = function(x) sql_quote(x, '"'),
    has_window_clause = TRUE
  )
}

#' @export
#' @rdname backend-access
simulate_access <- function() simulate_dbi("ACCESS")

#' @export
sql_dialect.ACCESS <- function(con) {
  dialect_access()
}

#' @export
dbplyr_edition.ACCESS <- function(con) {
  2L
}

# sql_ generics --------------------------------------------

#' @export
sql_query_select.sql_dialect_access <- function(
  con,
  select,
  from,
  where = NULL,
  group_by = NULL,
  having = NULL,
  window = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  ...,
  subquery = FALSE,
  lvl = 0
) {
  sql_select_clauses(
    select = sql_clause_select(select, distinct, top = limit),
    from = sql_clause_from(from),
    where = sql_clause_where(where),
    group_by = sql_clause_group_by(group_by),
    having = sql_clause_having(having),
    window = sql_clause_window(window),
    order_by = sql_clause_order_by(order_by, subquery, limit),
    lvl = lvl
  )
}

#' @export
sql_translation.sql_dialect_access <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_scalar,
      # Much of this translation comes from: https://www.techonthenet.com/access/functions/

      # Conversion
      as.numeric = sql_prefix("CDBL"),
      as.double = sql_prefix("CDBL"),
      # as.integer() always rounds down. CInt does not, but Int does
      as.integer = sql_prefix("INT"),
      as.logical = sql_prefix("CBOOL"),
      as.character = sql_prefix("CSTR"),
      as.Date = sql_prefix("CDATE"),

      # Math
      exp = sql_prefix("EXP"),
      log = sql_prefix("LOG"),
      log10 = function(x) {
        sql_glue("LOG({x}) / LOG(10)")
      },
      sqrt = sql_prefix("SQR"),
      sign = sql_prefix("SGN"),
      floor = sql_prefix("INT"),
      # Nearly add 1, then drop off the decimal. This results in the equivalent to ceiling()
      ceiling = function(x) {
        sql_glue("INT({x} + 0.9999999999)")
      },
      ceil = function(x) {
        sql_glue("INT({x} + 0.9999999999)")
      },
      # There is no POWER function in Access. It uses ^ instead
      `^` = function(x, y) {
        sql_glue("{x} ^ {y}")
      },

      # Strings
      nchar = sql_prefix("LEN"),
      tolower = sql_prefix("LCASE"),
      toupper = sql_prefix("UCASE"),
      # Pull `left` chars from the left, then `right` chars from the right to replicate substr
      substr = function(x, start, stop) {
        right <- stop - start + 1
        left <- stop
        sql_glue("RIGHT(LEFT({x}, {left}), {right})")
      },
      trimws = sql_prefix("TRIM"),
      # No support for CONCAT in Access
      paste = sql_paste_infix(" ", "&", function(x) sql_glue("CSTR({x})")),
      paste0 = sql_paste_infix("", "&", function(x) sql_glue("CSTR({x})")),

      # Logic
      # Access always returns -1 for True and 0 for False
      is.null = sql_prefix("ISNULL"),
      is.na = sql_prefix("ISNULL"),
      # IIF() is like ifelse()
      ifelse = function(test, yes, no) {
        sql_glue("IIF({test}, {yes}, {no})")
      },
      # Access uses <> for inequality
      `!=` = sql_infix("<>"),

      # Coalesce doesn't exist in Access.
      # NZ() only works while in Access, not with the Access driver
      # IIF(ISNULL()) is the best way to construct this
      coalesce = function(x, y) {
        sql_glue("IIF(ISNULL({x}), {y}, {x})")
      },

      # pmin/pmax for 2 columns
      pmin = function(x, y) {
        sql_glue("IIF({x} <= {y}, {x}, {y})")
      },

      pmax = function(x, y) {
        sql_glue("IIF({x} <= {y}, {y}, {x})")
      },

      # Dates
      Sys.Date = sql_prefix("DATE")
    ),

    sql_translator(
      .parent = base_agg,
      sd = sql_aggregate("STDEV"),
      var = sql_aggregate("VAR"),
      cor = sql_not_supported("cor"),
      cov = sql_not_supported("cov"),

      # Count(Distinct *) does not work in Access
      # This would work but we don't know the table name when translating:
      # SELECT Count(*) FROM (SELECT DISTINCT * FROM table_name) AS T
      n_distinct = sql_not_supported("n_distinct"),
    ),

    # Window functions not supported in Access
    sql_translator(.parent = base_no_win)
  )
}

# db_ generics -----------------------------------

#' @export
sql_table_analyze.sql_dialect_access <- function(con, table, ...) {
  # Do nothing. Access doesn't support an analyze / update statistics function
  NULL # nocov
}

# Util -------------------------------------------

#' @export
sql_escape_logical.sql_dialect_access <- function(con, x) {
  # Access uses a convention of -1 as True and 0 as False
  y <- ifelse(x, -1, 0)
  y[is.na(x)] <- "NULL"
  sql(y)
}

#' @export
sql_escape_date.sql_dialect_access <- function(con, x) {
  # Access delimits dates using octothorpes, and uses YYYY-MM-DD
  y <- format(x, "#%Y-%m-%d#")
  y[is.na(x)] <- "NULL"
  sql(y)
}

#' @export
sql_escape_datetime.sql_dialect_access <- function(con, x) {
  # Access delimits datetimes using octothorpes, and uses YYYY-MM-DD HH:MM:SS
  # Timezones are not supported in Access
  y <- format(x, "#%Y-%m-%d %H:%M:%S#")
  y[is.na(x)] <- "NULL"
  sql(y)
}

#' @export
sql_query_multi_join.sql_dialect_access <- function(
  con,
  x,
  joins,
  table_names,
  by_list,
  select,
  ...,
  lvl = 0
) {
  if (vctrs::vec_duplicate_any(table_names)) {
    cli_abort("{.arg table_names} must be unique.")
  }

  from <- dbplyr_sql_subquery(con, x, name = table_names[[1]], lvl = lvl)
  names <- table_names[-1]
  tables <- joins$table
  types <- toupper(paste0(joins$type, " JOIN"))

  n_joins <- length(types)

  # MS Access requires: ((t1 JOIN t2 ON ...) JOIN t3 ON ...)
  # N joins need N opening parens, each ON clause followed by closing paren
  open_parens <- strrep("(", n_joins)
  from <- sql(paste0(open_parens, from))

  for (i in seq_len(n_joins)) {
    table <- dbplyr_sql_subquery(con, tables[[i]], name = names[[i]], lvl = lvl)
    by <- joins$by[[i]]
    on <- sql_join_tbls(con, by = by, na_matches = by$na_matches)

    from <- sql(paste0(
      paste0(from, "\n"),
      paste0(types[[i]], " ", table, "\n"),
      paste0("ON ", on, ")")
    ))
  }

  clauses <- list(
    sql_clause_select(select),
    sql_clause_from(from)
  )
  sql_format_clauses(clauses, lvl = lvl)
}

test_that("simulate_access() still works", {
  expect_translation(simulate_access(), x + 1, '"x" + 1.0')
})

test_that("custom scalar translated correctly", {
  con <- dialect_access()

  # Conversion
  expect_translation(con, as.numeric(x), 'CDBL("x")')
  expect_translation(con, as.double(x), 'CDBL("x")')
  expect_translation(con, as.integer(x), 'INT("x")')
  expect_translation(con, as.logical(x), 'CBOOL("x")')
  expect_translation(con, as.character(x), 'CSTR("x")')
  expect_translation(con, as.Date(x), 'CDATE("x")')
  # Math
  expect_translation(con, exp(x), 'EXP("x")')
  expect_translation(con, log(x), 'LOG("x")')
  expect_translation(con, log10(x), 'LOG("x") / LOG(10)')
  expect_translation(con, sqrt(x), 'SQR("x")')
  expect_translation(con, sign(x), 'SGN("x")')
  expect_translation(con, floor(x), 'INT("x")')
  expect_translation(con, ceiling(x), 'INT("x" + 0.9999999999)')
  expect_translation(con, ceil(x), 'INT("x" + 0.9999999999)')
  expect_translation(con, x^y, '"x" ^ "y"')
  # String
  expect_translation(con, nchar(x), 'LEN("x")')
  expect_translation(con, tolower(x), 'LCASE("x")')
  expect_translation(con, toupper(x), 'UCASE("x")')
  expect_translation(
    con,
    substr(x, 1, 2),
    'RIGHT(LEFT("x", 2.0), 2.0)'
  )
  expect_translation(con, paste(x), 'CSTR("x")')
  expect_translation(con, trimws(x), 'TRIM("x")')
  expect_translation(con, is.null(x), 'ISNULL("x")')
  expect_translation(con, is.na(x), 'ISNULL("x")')
  expect_translation(
    con,
    coalesce(x, y),
    'IIF(ISNULL("x"), "y", "x")'
  )
  expect_translation(con, pmin(x, y), 'IIF("x" <= "y", "x", "y")')
  expect_translation(con, pmax(x, y), 'IIF("x" <= "y", "y", "x")')
  expect_translation(con, Sys.Date(), "DATE()")
  # paste()
  expect_translation(
    con,
    paste(x, y, sep = "+"),
    '"x" & \'+\' & "y"'
  )
  expect_translation(con, paste0(x, y), '"x" & "y"')
  expect_snapshot(
    error = TRUE,
    translate_sql(paste(x, collapse = "-"), con = con)
  )
  # Logic
  expect_translation(
    con,
    ifelse(x, "true", "false"),
    'IIF("x", \'true\', \'false\')'
  )
})

test_that("custom aggregators translated correctly", {
  con <- dialect_access()

  expect_translation(
    con,
    sd(x, na.rm = TRUE),
    'STDEV("x")',
    window = FALSE
  )
  expect_translation(
    con,
    var(x, na.rm = TRUE),
    'VAR("x")',
    window = FALSE
  )

  expect_error(
    translate_sql(cor(x), window = FALSE, con = con),
    class = "dbplyr_error_unsupported_fn"
  )
  expect_error(
    translate_sql(cov(x), window = FALSE, con = con),
    class = "dbplyr_error_unsupported_fn"
  )
  expect_error(
    translate_sql(n_distinct(x), window = FALSE, con = con),
    class = "dbplyr_error_unsupported_fn"
  )
})

test_that("custom escaping works as expected", {
  con <- dialect_access()

  expect_equal(escape(TRUE, con = con), sql("-1"))
  expect_equal(escape(FALSE, con = con), sql("0"))
  expect_equal(escape(NA, con = con), sql("NULL"))

  expect_equal(escape(as.Date("2020-01-01"), con = con), sql("#2020-01-01#"))
  expect_equal(escape(as.Date(NA), con = con), sql("NULL"))

  expect_equal(
    escape(as.POSIXct("2020-01-01", tz = "UTC"), con = con),
    sql("#2020-01-01 00:00:00#")
  )
  expect_equal(escape(as.POSIXct(NA, tz = "UTC"), con = con), sql("NULL"))
})

test_that("queries translate correctly", {
  mf <- lazy_frame(x = 1, con = dialect_access())
  expect_snapshot(mf |> head())
})

test_that("multiple joins use parens #1576", {
  lf1 <- lazy_frame(x = 1, a = 1, .name = "lf1", con = dialect_access())
  lf2 <- lazy_frame(x = 1, b = 1, .name = "lf2", con = dialect_access())
  lf3 <- lazy_frame(x = 1, c = 1, .name = "lf3", con = dialect_access())
  lf4 <- lazy_frame(x = 1, d = 1, .name = "lf4", con = dialect_access())

  expect_snapshot(left_join(lf1, lf2, by = "x") |> inner_join(lf3, by = "x"))
  expect_snapshot(
    left_join(lf1, lf2, by = "x") |>
      inner_join(lf3, by = "x") |>
      left_join(lf4, by = "x")
  )
})

#' Build a SQL string.
#'
#' @description
#' `r lifecycle::badge("superseded")`
#' `build_sql()` is superseded in favor of [sql_glue2()].
#'
#' This is a convenience function that should prevent sql injection attacks
#' (which in the context of dplyr are most likely to be accidental not
#' deliberate) by automatically escaping all expressions in the input, while
#' treating bare strings as sql. This is unlikely to prevent any serious
#' attack, but should make it unlikely that you produce invalid sql.
#'
#' This function should be used only when generating `SELECT` clauses,
#' other high level queries, or for other syntax that has no R equivalent.
#' For individual function translations, prefer [sql_expr()].
#'
#' @param ... input to convert to SQL. Use [sql()] to preserve
#'   user input as is (dangerous), and [ident()] to label user
#'   input as sql identifiers (safe)
#' @param .env the environment in which to evaluate the arguments. Should not
#'   be needed in typical use.
#' @param con database connection; used to select correct quoting characters.
#' @keywords internal
#' @export
#' @examples
#' con <- dialect_ansi()
#'
#' # Old:
#' build_sql("SELECT * FROM ", ident("table"), con = con)
#' # New:
#' sql_glue2(con, "SELECT * FROM {.tbl 'table'}")
#'
#' # Old:
#' name <- "Robert"
#' build_sql("INSERT INTO students (name) VALUES (", name, ")", con = con)
#' # New:
#' sql_glue2(con, "INSERT INTO students (name) VALUES ({name})")
build_sql <- function(..., .env = parent.frame(), con = sql_current_con()) {
  check_con(con)

  escape_expr <- function(x, con) {
    # If it's a string, leave it as is
    if (is.character(x)) {
      return(x)
    }

    val <- eval_bare(x, .env)
    # Skip nulls, so you can use if statements like in paste
    if (is.null(val)) {
      return("")
    }

    escape(val, con = con)
  }

  pieces <- purrr::map_chr(enexprs(...), escape_expr, con = con)
  sql(paste0(pieces, collapse = ""))
}

#' Generate SQL from R expressions
#'
#' @description
#' `r lifecycle::badge("superseded")`
#' `sql_expr()` and `sql_call2()` are superseded in favour of [sql_glue()].
#'
#' Low-level building block for generating SQL from R expressions.
#' Strings are escaped; names become bare SQL identifiers. User infix
#' functions have `%` stripped.
#'
#' Using `sql_expr()` in package will require use of [globalVariables()]
#' to avoid `R CMD check` NOTES. This is a small amount of additional pain,
#' which I think is worthwhile because it leads to more readable translation
#' code.
#'
#' @param x A quasiquoted expression
#' @param con Connection to use for escaping. Will be set automatically when
#'   called from a function translation.
#' @param .fn Function name (as string, call, or symbol)
#' @param ... Arguments to function
#' @keywords internal
#' @export
#' @examples
#' con <- dialect_ansi() # not necessary when writing translations
#'
#' sql_expr(f(x + 1), con = con)
#' sql_expr(f("x", "y"), con = con)
#' sql_expr(f(x, y), con = con)
#'
#' x <- ident("x")
#' sql_expr(f(!!x, y), con = con)
#'
#' sql_expr(cast("x" %as% DECIMAL), con = con)
#' sql_expr(round(x) %::% numeric, con = con)
#'
#' sql_call2("+", quote(x), 1, con = con)
#' sql_call2("+", "x", 1, con = con)
sql_expr <- function(x, con = sql_current_con()) {
  x <- enexpr(x)
  x <- replace_expr(x, con = con)
  sql(x)
}

#' @export
#' @rdname sql_expr
sql_call2 <- function(.fn, ..., con = sql_current_con()) {
  fn <- call2(.fn, ...)
  fn <- replace_expr(fn, con = con)
  sql(fn)
}


replace_expr <- function(x, con) {
  if (is.null(x)) {
    "NULL"
  } else if (is.atomic(x) || blob::is_blob(x)) {
    as.character(escape(unname(x), con = con))
  } else if (is.name(x)) {
    as.character(x)
  } else if (is.call(x)) {
    fun <- toupper(as.character(x[[1]]))
    args <- lapply(x[-1], replace_expr, con = con)

    if (is_infix_base(fun)) {
      if (length(args) == 1) {
        paste0(fun, args[[1]])
      } else {
        paste0(args[[1]], " ", fun, " ", args[[2]])
      }
    } else if (is_infix_user(fun)) {
      fun <- substr(fun, 2, nchar(fun) - 1)
      paste0(args[[1]], " ", fun, " ", args[[2]])
    } else if (fun == "(") {
      paste0("(", paste0(args, collapse = ", "), ")")
    } else {
      paste0(fun, "(", paste0(args, collapse = ", "), ")")
    }
  } else {
    x # nocov
  }
}

#' Convert to sql (deprecated)
#'
#' `r lifecycle::badge("deprecated")`
#'
#' @keywords internal
#'
#' @param x Object to coerce
#' @param con Needed when `x` is directly supplied from the user so that
#'   schema specifications can be quoted using the correct identifiers.
#' @export
as.sql <- function(x, con) {
  lifecycle::deprecate_warn("2.6.0", "as.sql()", "as_table_path()")
  if (methods::is(x, "Id")) {
    sql(DBI::dbQuoteIdentifier(con, x))
  } else if (is_bare_character(x)) {
    ident(x)
  } else {
    x
  }
}

#' Declare a identifier as being pre-quoted.
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' No longer needed; please use [sql()] instead.
#'
#' @keywords internal
#' @export
ident_q <- function(...) {
  x <- c_character(...)
  structure(x, class = c("ident_q", "ident", "character"))
}

#' @export
escape.ident_q <- function(x, parens = FALSE, collapse = ", ", con = NULL) {
  sql_vector(x, parens, collapse, con = con)
}

# build_sql() ------------------------------------------------------------------

test_that("build_sql() requires connection", {
  withr::local_options(lifecycle_verbosity = "quiet")
  x <- ident("TABLE")
  expect_snapshot(error = TRUE, build_sql("SELECT * FROM ", x))
})

# sql_expr() -------------------------------------------------------------------

test_that("NULL becomes SQL NULL", {
  con <- dialect_ansi()
  expect_equal(sql_expr(NULL), sql("NULL"))
})

test_that("atomic vectors are escaped", {
  con <- dialect_ansi()

  expect_equal(sql_expr(2, con = con), sql("2.0"))
  expect_equal(sql_expr("x", con = con), sql("'x'"))
})

test_that("user infix functions have % stripped", {
  con <- dialect_ansi()

  expect_equal(sql_expr(x %like% y, con = con), sql("x LIKE y"))
})

test_that("string function names are not quoted", {
  con <- dialect_ansi()

  f <- "foo"
  expect_equal(sql_expr((!!f)(), con = con), sql("FOO()"))
})

test_that("correct number of parens", {
  con <- dialect_ansi()

  expect_equal(sql_expr((1L), con = con), sql("(1)"))
})

# as.sql() ---------------------------------------------------------------------

test_that("as.sql() is deprecated", {
  expect_snapshot(as.sql(ident("x")))
})

test_that("as.sql() only affects character vectors", {
  local_options(lifecycle_verbosity = "quiet")

  expect_equal(as.sql("x"), ident("x"))

  x1 <- ident("x")
  expect_equal(as.sql(x1), x1)

  x2 <- ident_q('"x"')
  expect_equal(as.sql(x2), x2)

  x3 <- sql("SELECT 1")
  expect_equal(as.sql(x3), x3)
})

test_that("as.sql() converts DBI::Id to quoted identifier", {
  local_options(lifecycle_verbosity = "quiet")
  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(con))

  expect_equal(
    as.sql(DBI::Id(table = "x"), con = con),
    sql("`x`")
  )
  expect_equal(
    as.sql(DBI::Id(schema = "s", table = "x"), con = con),
    sql("`s`.`x`")
  )
})

# ident_q() --------------------------------------------------------------------

test_that("quoted identifier correctly escaped", {
  con <- dialect_ansi()
  x2 <- ident_q('"x"')
  expect_equal(escape(x2, con = con), sql('"x"'))

  expect_equal(sql_vector(ident_q(), collapse = NULL, con = con), sql())
  expect_equal(
    sql_vector(ident_q(), parens = FALSE, collapse = "", con = con),
    sql("")
  )
  expect_equal(
    sql_vector(ident_q(), parens = TRUE, collapse = "", con = con),
    sql("()")
  )
})

#' Simulate database connections
#'
#' These functions generate S3 objects that have been designed to simulate
#' the action of a database connection, without actually having the database
#' available. Obviously, this simulation can only be incomplete, but most
#' importantly it allows us to simulate SQL generation for any database without
#' actually connecting to it.
#'
#' Simulated SQL always quotes identifies with `` `x` ``, and strings with
#' `'x'`.
#'
#' @keywords internal
#' @export
simulate_dbi <- function(class = character(), ...) {
  structure(
    list(),
    ...,
    class = c(class, "TestConnection", "DBIConnection")
  )
}

#' @export
dbplyr_edition.TestConnection <- function(con) 2L


sql_escape_ident <- function(con, x) {
  UseMethod("sql_escape_ident")
}
#' @export
sql_escape_ident.DBIConnection <- function(con, x) {
  dbQuoteIdentifier(con, x)
}
#' @export
sql_escape_ident.TestConnection <- function(con, x) {
  sql_quote(x, "`")
}

sql_escape_string <- function(con, x) {
  UseMethod("sql_escape_string")
}
#' @export
sql_escape_string.DBIConnection <- function(con, x) {
  dbQuoteString(con, x)
}
#' @export
sql_escape_string.TestConnection <- function(con, x) {
  sql_quote(x, "'")
}

deparse_all <- function(x) {
  x <- purrr::map_if(x, is_formula, f_rhs)
  purrr::map_chr(x, expr_text, width = 500L)
}

#' Provides comma-separated string out of the parameters
#' @export
#' @keywords internal
named_commas <- function(x) {
  if (is.list(x)) {
    x <- purrr::map_chr(x, as_label)
  } else {
    x <- as.character(x)
  }

  nms <- names2(x)
  out <- ifelse(nms == "", x, paste0(nms, " = ", x))
  paste0(out, collapse = ", ")
}

commas <- function(...) paste0(..., collapse = ", ")

unique_table_name <- function(prefix = "") {
  vals <- c(letters, LETTERS, 0:9)
  name <- paste0(sample(vals, 10, replace = TRUE), collapse = "")
  paste0(prefix, "dbplyr_", name)
}

unique_subquery_name <- function() {
  # Needs to use option so can reset at the start of each query
  i <- getOption("dbplyr_subquery_name", 0) + 1
  options(dbplyr_subquery_name = i)
  sprintf("q%02i", i)
}
unique_column_name <- function() {
  # Needs to use option so can reset at the start of each query
  i <- getOption("dbplyr_column_name", 0) + 1
  options(dbplyr_column_name = i)
  sprintf("col%02i", i)
}
unique_subquery_name_reset <- function() {
  options(dbplyr_subquery_name = 0)
}
unique_column_name_reset <- function() {
  options(dbplyr_column_name = 0)
}

succeeds <- function(x, quiet = FALSE) {
  tryCatch(
    {
      x
      TRUE
    },
    error = function(e) {
      if (!quiet)
        message("Error: ", e$message) # nocov
      FALSE
    }
  )
}

c_character <- function(...) {
  x <- c(...)
  if (length(x) == 0) {
    return(character())
  }

  if (!is.character(x)) {
    cli_abort("Character input expected")
  }

  x
}

cat_line <- function(...) cat(paste0(..., "\n"), sep = "")

# nocov start
res_warn_incomplete <- function(res, hint = "n = -1") {
  if (dbHasCompleted(res)) return()

  rows <- big_mark(dbGetRowCount(res))
  cli::cli_warn("Only first {rows} results retrieved. Use {hint} to retrieve all.")
}

add_temporary_prefix <- function(con, table, temporary = TRUE) {
  check_table_path(table)

  if (!temporary) {
    return(table)
  }

  pieces <- table_path_components(table, con)[[1]]
  table_name <- pieces[length(pieces)]

  if (substr(table_name, 1, 1) != "#") {
    new_name <- paste0("#", table_name)
    cli::cli_inform(
      paste0("Created a temporary table named ", new_name),
      class = c("dbplyr_message_temp_table", "dbplyr_message")
    )
    pieces[[length(pieces)]] <- new_name
    table <- make_table_path(pieces, con)
  }

  table
}
# nocov end

# Helper for testing
local_methods <- function(..., .frame = caller_env()) {
  local_bindings(..., .env = global_env(), .frame = .frame)
}

local_db_table <- function(con, value, name, ..., temporary = TRUE, envir = parent.frame()) {
  if (inherits(con, "Microsoft SQL Server") && temporary) {
    name <- paste0("#", name)
  }

  withr::defer(DBI::dbRemoveTable(con, name), envir = envir)
  copy_to(con, value, name, temporary = temporary, ...)
  tbl(con, name)
}

local_sqlite_connection <- function(envir = parent.frame()) {
  withr::local_db_connection(
    DBI::dbConnect(RSQLite::SQLite(), ":memory:"),
    .local_envir = envir
  )
}

test_that("can translate a table", {
  con <- local_sqlite_connection()
  df <- tibble(
    lgl = TRUE,
    int = 1L,
    dbl = 1.5,
    chr = "a",
    date = as.Date("2020-01-01", tz = "UTC"),
    dtt = as.POSIXct("2020-01-01 01:23:45", tz = "UTC")
  )

  expect_snapshot(copy_inline(con, df) |> remote_query())

  expect_equal(
    copy_inline(con, df) |> collect(),
    tibble(
      lgl = 1L,
      int = 1L,
      dbl = 1.5,
      chr = "a",
      date = "2020-01-01",
      dtt = "2020-01-01T01:23:45Z"
    )
  )

  expect_equal(
    copy_inline(
      con,
      tibble(date = as.Date(c("2020-01-01", "2020-01-02"), tz = "UTC"))
    ) |>
      collect(),
    tibble(date = c("2020-01-01", "2020-01-02"))
  )
})

test_that("can translate blob columns", {
  con <- local_sqlite_connection()

  df <- tibble(x = blob::blob(charToRaw("abc"), charToRaw("def")))
  db <- copy_inline(con, df)

  expect_snapshot(show_query(db))
  expect_equal(collect(db), df)
})

test_that("can translate 1-column tables", {
  con <- local_sqlite_connection()
  expect_snapshot(
    copy_inline(con, tibble(dbl = 1.5)) |>
      remote_query()
  )
})

test_that("zero row table works", {
  con <- local_sqlite_connection()
  expect_snapshot(
    copy_inline(con, tibble(dbl = numeric(), chr = character())) |>
      remote_query()
  )

  expect_snapshot(
    copy_inline(con, tibble(dbl = numeric())) |>
      remote_query()
  )
})

test_that("types argument works", {
  con <- local_sqlite_connection()

  df <- tibble(x = "1", y = 2L)
  expect_equal(copy_inline(con, df) |> collect(), df)

  expect_equal(
    copy_inline(con, df, types = c(x = "INTEGER", y = "TEXT")) |> collect(),
    tibble(x = 1L, y = "2")
  )
})

test_that("checks inputs", {
  con <- dialect_ansi()

  expect_snapshot({
    (expect_error(copy_inline(con, tibble())))
    (expect_error(copy_inline(con, lazy_frame(a = 1))))

    (expect_error(copy_inline(con, tibble(a = 1), types = c(b = "bigint"))))
    (expect_error(copy_inline(con, tibble(a = 1), types = c(b = 1))))
  })
})

test_that("can copy to from remote sources", {
  df <- tibble(x = 1:10)
  con1 <- local_sqlite_connection()
  df_1 <- copy_to(con1, df, "df1")

  # Create from tbl in same database
  df_2 <- copy_to(con1, df_1, "df2")
  expect_equal(collect(df_2), df)

  # Create from tbl in another data
  con2 <- local_sqlite_connection()
  df_3 <- copy_to(con2, df_1, "df3")
  expect_equal(collect(df_3), df)
})

test_that("copy_to() within same source uses compute() and ignores types", {
  con <- local_sqlite_connection()
  df_1 <- copy_to(con, tibble(x = 1L), "df1", types = c(x = "INT"))
  df_2 <- copy_to(con, df_1, "df2", types = c(x = "DOUBLE"))

  expect_equal(
    DBI::dbGetQuery(con, "SELECT typeof(x) FROM df2")[[1]],
    "integer"
  )
})

test_that("only overwrite existing table if explicitly requested", {
  con <- local_sqlite_connection()
  local_db_table(con, data.frame(x = 1:5), "df")

  expect_error(copy_to(con, data.frame(x = 1), name = "df"), "exists")
  expect_silent(copy_to(con, data.frame(x = 1), name = "df", overwrite = TRUE))
})

test_that("overwrite flag works when copying *within* db sources", {
  con <- local_sqlite_connection()
  df <- copy_to(con, tibble(x = 1), "df", temporary = FALSE)
  copy_to(con, df, "df2", temporary = FALSE)

  df2 <- tibble(x = 2)
  out <- copy_to(con, df2, name = "df2", temporary = FALSE, overwrite = TRUE)
  expect_equal(collect(out), df2)
})

test_that("can create a new table in non-default schema", {
  con <- local_sqlite_con_with_aux()

  df1 <- tibble(x = 1)
  df2 <- tibble(x = 2)

  db1 <- copy_to(con, df1, in_schema("aux", "df"), temporary = FALSE)
  expect_equal(collect(db1), df1)

  # And can overwrite
  db2 <- copy_to(
    con,
    df2,
    in_schema("aux", "df"),
    temporary = FALSE,
    overwrite = TRUE
  )
  expect_equal(collect(db2), df2)
})

test_that("df must be a local or remote table", {
  con <- local_sqlite_connection()

  expect_snapshot(error = TRUE, copy_to(con, list(x = 1), name = "df"))
})

# as_copy() ---------------------------------------------------------------

test_that("as_copy() converts TRUE/FALSE to enum", {
  expect_equal(as_copy(TRUE), "temp-table")
  expect_equal(as_copy(FALSE), "none")
})

test_that("as_copy() passes through valid strings", {
  expect_equal(as_copy("none"), "none")
  expect_equal(as_copy("temp-table"), "temp-table")
  expect_equal(as_copy("inline"), "inline")
})

test_that("as_copy() errors on invalid values", {
  expect_snapshot(as_copy("other"), error = TRUE)
})

# dbplyr_auto_copy() ----------------------------------------------------

test_that("dbplyr_auto_copy() returns y if same source", {
  df <- local_memdb_frame(x = 1:3)
  expect_identical(dbplyr_auto_copy(df, df, copy = FALSE), df)
})

test_that("dbplyr_auto_copy() errors when copy = FALSE and different sources", {
  df <- local_memdb_frame(x = 1:3)
  local_df <- tibble(x = 1:3)

  expect_snapshot(dbplyr_auto_copy(df, local_df, copy = FALSE), error = TRUE)
  expect_snapshot(dbplyr_auto_copy(df, local_df, copy = "none"), error = TRUE)
})

test_that("dbplyr_auto_copy() with copy = TRUE copies to temp table", {
  df <- local_memdb_frame(x = 1:3)
  local_df <- tibble(x = 1:3)

  out <- dbplyr_auto_copy(df, local_df, copy = TRUE)
  expect_s3_class(out, "tbl_sql")
  expect_equal(collect(out), local_df)
})

test_that("dbplyr_auto_copy() with copy = 'inline' uses copy_inline", {
  df <- local_memdb_frame(x = 1:3)
  local_df <- tibble(x = 1:3)

  out <- dbplyr_auto_copy(df, local_df, copy = "inline")
  expect_s3_class(out, "tbl_sql")
  expect_s3_class(out$lazy_query, "lazy_base_values_query")
  expect_equal(collect(out), local_df)
})

test_that("compute can create indexes", {
  # integration test to ensure that indexes argument passed all the way through
  db <- local_memdb_frame("db1", x = 5:1, y = 1:5, z = 10)

  db |> compute(indexes = c("x", "y"), name = "db2")
  indices <- DBI::dbGetQuery(memdb(), "PRAGMA index_list('db2');")
  expect_equal(indices$name, c("db2_y", "db2_x"))
  expect_equal(indices$unique, c(0, 0))

  db |> compute(unique_indexes = c("x", "y"), name = "db3")
  indices <- DBI::dbGetQuery(memdb(), "PRAGMA index_list('db3');")
  expect_equal(indices$name, c("db3_y", "db3_x"))
  expect_equal(indices$unique, c(1, 1))
})

test_that("unique index fails if values are duplicated", {
  mf <- local_memdb_frame(x = 5:1, y = "a")
  expect_error(compute(mf, unique_indexes = "y"))
})

test_that("index fails if columns are missing", {
  mf <- local_memdb_frame(x = 1)
  expect_snapshot({
    (expect_error(compute(mf, indexes = list(c("y", "x", "z"), "a"))))
    (expect_error(compute(mf, unique_indexes = list(c("y", "x", "z"), "a"))))
  })
})

test_that("compute creates correct column names", {
  out <- local_memdb_frame(x = 1) |>
    group_by(x) |>
    summarise(n = n()) |>
    compute() |>
    collect()

  expect_equal(out, tibble(x = 1, n = 1L))
})

test_that("compute keeps window and groups", {
  out <- local_memdb_frame(x = 1, y = 1) |>
    window_order(x) |>
    group_by(x, y) |>
    summarise(n = n(), .groups = "drop_last") |>
    compute()

  expect_equal(op_sort(out), list(expr(x)))
  expect_equal(op_grps(out), "x")
})

test_that("compute can handle named name", {
  con <- dialect_ansi()
  expect_equal(
    local_memdb_frame(x = 1:10) |>
      compute(name = c(x = unique_table_name())) |>
      collect(),
    tibble(x = 1:10)
  )
})

test_that("compute can handle schema", {
  df <- local_memdb_frame("db1", x = 1:10)
  withr::defer(DBI::dbRemoveTable(memdb(), "db1"))

  expect_equal(
    df |>
      compute(name = in_schema("main", "db1"), temporary = FALSE) |>
      collect(),
    tibble(x = 1:10)
  )

  # errors because name already exists
  expect_snapshot(
    df |> compute(name = in_schema("main", "db1"), temporary = FALSE),
    transform = snap_transform_dbi,
    error = TRUE
  )
})

test_that("compute(temporary = FALSE) without a name is deprecated", {
  df <- local_memdb_frame(x = 1:10)

  expect_snapshot(error = TRUE, df |> compute(temporary = FALSE))
})

test_that("sorting preserved across compute", {
  df1 <- local_memdb_frame(x = sample(10)) |> arrange(x)

  df2 <- compute(df1)
  expect_equal(get_expr(op_sort(df2)[[1]]), quote(x))
})

test_that("compute dispatches through custom tbl method", {
  local_methods(
    tbl.SQLiteConnection = function(src, from, ...) {
      out <- NextMethod()
      attr(out, "custom") <- TRUE
      out
    }
  )

  df <- local_memdb_frame(x = 1)
  out <- compute(df)
  expect_true(attr(out, "custom"))
})

test_that("Successful and not-successful commands are identified", {
  expect_true(succeeds("success"))
  expect_false(succeeds(x - 1, quiet = TRUE))
})

test_that("Returns error if no characters are passed", {
  expect_error(c_character(1, 2))
})

test_that("n_selects counts SELECT statements", {
  lf <- lazy_frame(x = 1)

  expect_equal(lf |> n_selects(), 1)
  expect_equal(lf |> filter(x > 1) |> n_selects(), 1)
  expect_equal(lf |> summarise(x) |> summarise(x) |> n_selects(), 2)
})

test_that("generated sql doesn't change unexpectedly", {
  lf <- lazy_frame(x = 1, y = 2)
  expect_snapshot(setdiff(lf, lf))
  expect_snapshot(intersect(lf, lf))

  expect_snapshot(setdiff(lf, lf, all = TRUE))
  expect_snapshot(intersect(lf, lf, all = TRUE))
})

test_that("sql_set_op_method can be customized", {
  local_methods(
    sql_set_op_method.TestConnection = function(con, op, ...) {
      switch(
        op,
        "INTERSECT" = "INTERSECT DISTINCT",
        "EXCEPT" = "EXCEPT DISTINCT",
        op
      )
    }
  )

  lf <- lazy_frame(x = 1)
  expect_match(remote_query(intersect(lf, lf)), "INTERSECT DISTINCT")
  expect_match(remote_query(setdiff(lf, lf)), "EXCEPT DISTINCT")
})

test_that("db_copy_to() wraps DBI errors", {
  con <- local_sqlite_connection()
  local_db_table(con, data.frame(x = 1), "tmp")

  # error when creating unique index
  expect_snapshot(
    (expect_error(
      db_copy_to(
        con = con,
        table = "tmp2",
        values = data.frame(x = c(1, 1)),
        unique_indexes = list("x")
      )
    )),
    transform = snap_transform_dbi
  )
})

test_that("db_copy_to() can overwrite a table", {
  con <- local_sqlite_connection()
  local_db_table(con, data.frame(x = 1), "tmp")

  # doesn't overwrite by default
  expect_snapshot(
    (expect_error(
      db_copy_to(
        con = con,
        table = "tmp",
        values = data.frame(x = c(1, 1))
      )
    )),
    transform = snap_transform_dbi
  )

  db_copy_to(
    con = con,
    table = "tmp",
    values = data.frame(x = c(1, 1)),
    overwrite = TRUE
  )
  expect_equal(DBI::dbReadTable(con, "tmp"), data.frame(x = c(1, 1)))
})

test_that("db_save_query() can overwrite a table", {
  con <- local_sqlite_connection()
  local_db_table(con, data.frame(x = 1), "tmp")

  # doesn't overwrite by default
  expect_snapshot(
    (expect_error(
      db_save_query(
        con = con,
        sql = "SELECT 2 FROM tmp",
        name = "tmp"
      )
    )),
    transform = snap_transform_dbi
  )

  db_save_query(
    con = con,
    sql = sql("SELECT 2 AS x"),
    name = "tmp",
    overwrite = TRUE
  )
  expect_equal(DBI::dbReadTable(con, "tmp"), data.frame(x = 2L))
})

test_that("set ops captures both tables", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 2)

  out <- union(lf1, lf2) |> sql_build()
  expect_s3_class(out, "union_query")
})

test_that("generated sql doesn't change unexpectedly", {
  lf <- lazy_frame(x = 1, y = 2)
  expect_snapshot(union(lf, lf))
  expect_snapshot(union_all(lf, lf))
})

test_that("sql_set_op_method can be customized", {
  local_methods(
    sql_set_op_method.TestConnection = function(con, op, ...) {
      switch(op, "UNION" = "UNION DISTINCT", op)
    }
  )

  lf <- lazy_frame(x = 1)
  expect_match(remote_query(union(lf, lf)), "UNION DISTINCT")
})

#' @export
#' @rdname sql_build
lazy_select_query <- function(
  x,
  select = NULL,
  where = NULL,
  group_by = NULL,
  having = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  group_vars = NULL,
  order_vars = NULL,
  frame = NULL,
  select_operation = c("select", "mutate", "summarise")
) {
  check_lazy_query(x, call = call)
  stopifnot(is.null(select) || is_lazy_sql_part(select))
  stopifnot(is_lazy_sql_part(where))
  # stopifnot(is.character(group_by))
  stopifnot(is_lazy_sql_part(order_by))
  check_number_whole(limit, allow_infinite = TRUE, allow_null = TRUE)
  check_bool(distinct)

  select <- select %||% syms(set_names(op_vars(x)))
  select_operation <- arg_match0(
    select_operation,
    c("select", "mutate", "summarise")
  )

  # inherit `group_vars`, `order_vars`, and `frame` from `from`
  group_vars <- group_vars %||% op_grps(x)
  order_vars <- order_vars %||% op_sort(x)
  frame <- frame %||% op_frame(x)

  if (select_operation == "mutate") {
    select <- new_lazy_select(
      select,
      group_vars = group_vars,
      order_vars = order_vars,
      frame = frame
    )
  } else {
    select <- new_lazy_select(select)
  }

  lazy_query(
    query_type = "select",
    x = x,
    select = select,
    where = where,
    group_by = group_by,
    order_by = order_by,
    distinct = distinct,
    limit = limit,
    select_operation = select_operation,
    group_vars = group_vars,
    order_vars = order_vars,
    frame = frame
  )
}

is_lazy_select_query <- function(x) {
  inherits(x, "lazy_select_query")
}

is_lazy_sql_part <- function(x) {
  if (is.null(x)) {
    return(TRUE)
  }
  if (!is.list(x)) {
    return(FALSE)
  }
  is_valid_expr <- function(item) {
    is_symbol(item) || is_call(item) || is_sql_literal(item) || is.sql(item)
  }
  purrr::every(x, is_valid_expr)
}

new_lazy_select <- function(
  vars,
  group_vars = character(),
  order_vars = NULL,
  frame = NULL
) {
  vctrs::vec_as_names(names2(vars), repair = "check_unique")

  var_names <- names(vars)
  vars <- unname(vars)

  tibble(
    name = var_names %||% character(),
    expr = vars %||% list(),
    group_vars = rep_along(vars, list(group_vars)),
    order_vars = rep_along(vars, list(order_vars)),
    frame = rep_along(vars, list(frame))
  )
}

#' @export
op_vars.lazy_select_query <- function(op) {
  op$select$name
}

#' @export
sql_build.lazy_select_query <- function(op, con, ..., sql_options = NULL) {
  alias <- remote_name(op$x, null_if_local = FALSE) %||% unique_subquery_name()
  from <- sql_build(op$x, con, sql_options = sql_options)
  select_sql_list <- get_select_sql(
    select = op$select,
    select_operation = op$select_operation,
    in_vars = op_vars(op$x),
    table_alias = alias,
    con = con,
    use_star = sql_options$use_star
  )
  where_sql <- translate_sql_(
    op$where,
    con = con,
    context = list(clause = "WHERE")
  )

  select_query(
    from = from,
    select = select_sql_list$select_sql,
    where = where_sql,
    group_by = translate_sql_(op$group_by, con = con),
    having = translate_sql_(op$having, con = con, window = FALSE),
    window = select_sql_list$window_sql,
    order_by = translate_sql_(op$order_by, con = con),
    distinct = op$distinct,
    limit = op$limit,
    from_alias = alias
  )
}

get_select_sql <- function(
  select,
  select_operation,
  in_vars,
  table_alias,
  con,
  use_star
) {
  if (select_operation == "summarise") {
    select_expr <- set_names(select$expr, select$name)
    select_sql <- translate_sql_(
      select_expr,
      con,
      window = FALSE,
      context = list(clause = "SELECT")
    )
    return(list(
      select_sql = names_to_as(con, select_sql),
      window_sql = sql()
    ))
  }

  if (use_star) {
    if (is_identity(select$expr, select$name, in_vars)) {
      out <- list(
        select_sql = sql_select_star(con, table_alias),
        window_sql = sql()
      )
      return(out)
    } else {
      select <- select_use_star(select, in_vars, table_alias, con)
    }
  }

  # translate once just to register windows
  win_register_activate()
  # Remove known windows before building the next query
  on.exit(win_reset(), add = TRUE)
  on.exit(win_register_deactivate(), add = TRUE)
  select_sql <- translate_select_sql(con, select)
  win_register_deactivate()

  named_windows <- win_register_names()
  if (nrow(named_windows) > 0 && sql_has_window_clause(con)) {
    # need to translate again and use registered windows names
    select_sql <- translate_select_sql(con, select)

    # build window sql
    names_esc <- sql_escape_ident(con, named_windows$name)
    window_sql <- sql(paste0(names_esc, " AS ", named_windows$key))
  } else {
    window_sql <- sql()
  }

  list(
    select_sql = select_sql,
    window_sql = window_sql
  )
}

select_use_star <- function(select, vars_prev, table_alias, con) {
  first_match <- vctrs::vec_match(vars_prev[[1]], select$name)
  if (is.na(first_match)) {
    return(select)
  }

  last <- first_match + length(vars_prev) - 1
  n <- vctrs::vec_size(select)

  if (n < last) {
    return(select)
  }

  test_cols <- vctrs::vec_slice(select, seq2(first_match, last))

  if (is_identity(test_cols$expr, test_cols$name, vars_prev)) {
    idx_start <- seq2(1, first_match - 1)
    idx_end <- seq2(last + 1, n)
    star <- sql_select_star(con, table_alias)
    vctrs::vec_rbind(
      vctrs::vec_slice(select, idx_start),
      tibble(name = "", expr = list(star)),
      vctrs::vec_slice(select, idx_end)
    )
  } else {
    select
  }
}

translate_select_sql <- function(con, select_df) {
  n <- vctrs::vec_size(select_df)
  out <- vctrs::vec_init(sql(), n)
  out <- set_names(out, select_df$name)
  for (i in seq2(1, n)) {
    out[[i]] <- translate_sql_(
      dots = select_df$expr[i],
      con = con,
      vars_group = translate_sql_(syms(select_df$group_vars[[i]]), con),
      vars_order = translate_sql_(
        select_df$order_vars[[i]],
        con,
        context = list(clause = "ORDER")
      ),
      vars_frame = select_df$frame[[i]][[1]],
      context = list(clause = "SELECT")
    )
  }

  names_to_as(con, out)
}

# Built query -------------------------------------------------------------

#' @export
#' @rdname sql_build
select_query <- function(
  from,
  select = sql("*"),
  where = sql(),
  group_by = sql(),
  having = sql(),
  window = sql(),
  order_by = sql(),
  limit = NULL,
  distinct = FALSE,
  from_alias = NULL
) {
  check_sql(select, allow_names = FALSE)
  check_sql(where)
  check_sql(group_by)
  check_sql(having)
  check_sql(window)
  check_sql(order_by)
  check_number_whole(limit, allow_infinite = TRUE, allow_null = TRUE)
  check_bool(distinct)
  check_string(from_alias, allow_null = TRUE)

  query(
    "select",
    from = from,
    select = select,
    where = where,
    group_by = group_by,
    having = having,
    window = window,
    order_by = order_by,
    distinct = distinct,
    limit = limit,
    from_alias = from_alias
  )
}

#' @export
sql_render.select_query <- function(
  query,
  con,
  ...,
  sql_options = NULL,
  subquery = FALSE,
  lvl = 0
) {
  from <- dbplyr_sql_subquery(
    con,
    sql_render(query$from, con, ..., subquery = TRUE, lvl = lvl + 1),
    name = query$from_alias,
    lvl = lvl
  )

  sql_query_select(
    con,
    names_to_as(con, query$select),
    from,
    where = query$where,
    group_by = query$group_by,
    having = query$having,
    window = query$window,
    order_by = query$order_by,
    limit = query$limit,
    distinct = query$distinct,
    ...,
    subquery = subquery,
    lvl = lvl
  )
}

warn_drop_order_by <- function() {
  # Rules according to SQLite: https://sqlite.org/forum/forumpost/878ca7a9be0862af?t=h
  # 1. There is no LIMIT clause in the subquery
  # 3. The subquery is not part of the FROM clause in an UPDATE-FROM statement
  # 4. The outer query does not use any aggregate functions other than the built-in count(), min(), and/or max() functions.
  # 5. Either the outer query has its own ORDER BY clause or else the subquery is one term of a join.
  #
  # (2. and 6. are left out as they are SQLite internal)
  warn(c(
    "ORDER BY is ignored in subqueries without LIMIT",
    i = "Do you need to move arrange() later in the pipeline or use window_order() instead?"
  ))
}

#' @export
flatten_query.select_query <- function(qry, query_list, con) {
  from <- qry$from
  query_list <- flatten_query(from, query_list, con)

  qry$from <- get_subquery_name(from, query_list)
  querylist_reuse_query(qry, query_list, con)
}

# SQL generation ----------------------------------------------------------

#' @rdname db-sql
#' @export
sql_query_select <- function(
  con,
  select,
  from,
  where = NULL,
  group_by = NULL,
  having = NULL,
  window = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  ...,
  subquery = FALSE,
  lvl = 0
) {
  check_dots_used()
  check_sql(select, allow_names = FALSE)

  UseMethod("sql_query_select", sql_dialect(con))
}

#' @export
sql_query_select.DBIConnection <- function(
  con,
  select,
  from,
  where = NULL,
  group_by = NULL,
  having = NULL,
  window = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  ...,
  subquery = FALSE,
  lvl = 0
) {
  sql_select_clauses(
    select = sql_clause_select(select, distinct),
    from = sql_clause_from(from),
    where = sql_clause_where(where),
    group_by = sql_clause_group_by(group_by),
    having = sql_clause_having(having),
    window = sql_clause_window(window),
    order_by = sql_clause_order_by(order_by, subquery, limit),
    limit = sql_clause_limit(con, limit),
    lvl = lvl
  )
}

#' @export
sql_query_select.sql_dialect <- sql_query_select.DBIConnection

test_that("can print lazy_select_query", {
  expect_snapshot(
    lazy_select_query(
      x = lazy_query_local(tibble(x = 1, y = 2), "df"),
      select = quos(
        x_mean = mean(x),
        y2 = y
      ),
      where = quos(y > 1, x == y - 2),
      group_by = quos("x")
    )
  )
})

#' Collapse a query into a subquery
#'
#' `collapse()` forces computation of a lazy query by wrapping it in a subquery.
#' This is not generally needed, but can be useful if you need to work around
#' database/dbplyr limitations.
#'
#' @export
#' @inheritParams collect.tbl_sql
#' @importFrom dplyr collapse
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(a = c(3, 4, 1, 2), b = c(5, 1, 2, NA))
#' db |> filter(a <= 2) |> show_query()
#' db |> filter(a <= 2) |> collapse() |> show_query()
collapse.tbl_sql <- function(x, ...) {
  sql <- db_sql_render(x$con, x)

  new_tbl_sql(x$con, sql, vars = colnames(x)) |>
    group_by(!!!syms(op_grps(x))) |>
    arrange(!!!op_sort(x))
}

#' @export
#' @rdname sql_build
lazy_select_query <- function(x,
                              select = NULL,
                              where = NULL,
                              group_by = NULL,
                              having = NULL,
                              order_by = NULL,
                              limit = NULL,
                              distinct = FALSE,
                              group_vars = NULL,
                              order_vars = NULL,
                              frame = NULL,
                              select_operation = c("select", "mutate", "summarise"),
                              message_summarise = NULL) {
  stopifnot(inherits(x, "lazy_query"))
  stopifnot(is.null(select) || is_lazy_sql_part(select))
  stopifnot(is_lazy_sql_part(where))
  # stopifnot(is.character(group_by))
  stopifnot(is_lazy_sql_part(order_by))
  stopifnot(is.null(limit) || (is.numeric(limit) && length(limit) == 1L))
  stopifnot(is.logical(distinct), length(distinct) == 1L)

  select <- select %||% syms(set_names(op_vars(x)))
  select_operation <- arg_match0(select_operation, c("select", "mutate", "summarise"))

  stopifnot(is.null(message_summarise) || is_string(message_summarise))

  # inherit `group_vars`, `order_vars`, and `frame` from `from`
  group_vars <- group_vars %||% op_grps(x)
  order_vars <- order_vars %||% op_sort(x)
  frame <- frame %||% op_frame(x)

  if (select_operation == "mutate") {
    select <- new_lazy_select(
      select,
      group_vars = group_vars,
      order_vars = order_vars,
      frame = frame
    )
  } else {
    select <- new_lazy_select(select)
  }

  lazy_query(
    query_type = "select",
    x = x,
    select = select,
    where = where,
    group_by = group_by,
    order_by = order_by,
    distinct = distinct,
    limit = limit,
    select_operation = select_operation,
    message_summarise = message_summarise,
    group_vars = group_vars,
    order_vars = order_vars,
    frame = frame
  )
}

is_lazy_sql_part <- function(x) {
  if (is.null(x)) return(TRUE)
  if (is_quosures(x)) return(TRUE)

  if (!is.list(x)) return(FALSE)
  purrr::every(x, ~ is_quosure(.x) || is_symbol(.x) || is_call(.x))
}

new_lazy_select <- function(vars, group_vars = character(), order_vars = NULL, frame = NULL) {
  vctrs::vec_as_names(names2(vars), repair = "check_unique")

  var_names <- names(vars)
  vars <- unname(vars)

  tibble(
    name = var_names %||% character(),
    expr = vars %||% list(),
    group_vars = rep_along(vars, list(group_vars)),
    order_vars = rep_along(vars, list(order_vars)),
    frame = rep_along(vars, list(frame))
  )
}

# projection = only select (including rename) from parent query
# identity = selects exactly the same variable as the parent query
is_lazy_select_query_simple <- function(x,
                                        ignore_group_by = FALSE,
                                        select = c("projection", "identity")) {
  select <- arg_match(select, c("projection", "identity"))
  if (!inherits(x, "lazy_select_query")) {
    return(FALSE)
  }

  if (select == "projection" && !is_projection(x$select$expr)) {
    return(FALSE)
  }

  if (select == "identity" && !is_select_identity(x$select, op_vars(x$x))) {
    return(FALSE)
  }

  if (!is_empty(x$where)) {
    return(FALSE)
  }
  if (!ignore_group_by && !is_empty(x$group_by)) {
    return(FALSE)
  }
  if (!is_empty(x$order_by)) {
    return(FALSE)
  }
  if (is_true(x$distinct)) {
    return(FALSE)
  }
  if (!is_empty(x$limit)) {
    return(FALSE)
  }

  TRUE
}

is_select_identity <- function(select, vars_prev) {
  is_identity(select$expr, select$name, vars_prev)
}


#' @export
print.lazy_select_query <- function(x, ...) {
  cat(
    "<SQL SELECT",
    if (x$distinct) " DISTINCT", ">\n",
    sep = ""
  )
  cat_line("From:")
  cat_line(indent_print(sql_build(x$x, simulate_dbi())))

  select <- purrr::set_names(x$select$expr, x$select$name)
  if (length(select))   cat("Select:   ", named_commas2(select), "\n", sep = "")
  if (length(x$where))    cat("Where:    ", named_commas2(x$where), "\n", sep = "")
  if (length(x$group_by)) cat("Group by: ", named_commas2(x$group_by), "\n", sep = "")
  if (length(x$order_by)) cat("Order by: ", named_commas2(x$order_by), "\n", sep = "")
  if (length(x$limit))    cat("Limit:    ", x$limit, "\n", sep = "")

  if (length(x$group_vars)) cat("group_vars: ", named_commas2(x$group_vars), "\n", sep = "")
  if (length(x$order_vars)) cat("order_vars: ", named_commas2(x$order_vars), "\n", sep = "")
  if (length(x$frame))    cat("frame:    ", x$frame, "\n", sep = "")
}

named_commas2 <- function(x) {
  x <- purrr::map_chr(x, as_label)
  nms <- names2(x)
  out <- ifelse(nms == "", x, paste0(nms, " = ", x))
  paste0(out, collapse = ", ")
}

#' @export
op_vars.lazy_query <- function(op) {
  op$select$name
}

#' @export
op_desc.lazy_query <- function(op) {
  "SQL"
}

#' @export
sql_build.lazy_select_query <- function(op, con, ...) {
  if (!is.null(op$message_summarise)) {
    inform(op$message_summarise)
  }

  select_sql_list <- get_select_sql(op$select, op$select_operation, op_vars(op$x), con)
  where_sql <- translate_sql_(op$where, con = con, context = list(clause = "WHERE"))

  select_query(
    from = sql_build(op$x, con),
    select = select_sql_list$select_sql,
    where = where_sql,
    group_by = translate_sql_(op$group_by, con = con),
    having = translate_sql_(op$having, con = con, window = FALSE),
    window = select_sql_list$window_sql,
    order_by = translate_sql_(op$order_by, con = con),
    distinct = op$distinct,
    limit = op$limit
  )
}

get_select_sql <- function(select, select_operation, in_vars, con) {
  if (select_operation == "summarise") {
    select_expr <- set_names(select$expr, select$name)
    select_sql_list <- translate_sql_(select_expr, con, window = FALSE, context = list(clause = "SELECT"))
    select_sql <- sql_vector(select_sql_list, parens = FALSE, collapse = NULL, con = con)
    return(list(select_sql = select_sql, window_sql = character()))
  }

  if (is_select_identity(select, in_vars)) {
    return(list(select_sql = sql("*"), window_sql = character()))
  }

  select <- select_use_star(select, in_vars, con)

  # translate once just to register windows
  win_register_activate()
  # Remove known windows before building the next query
  on.exit(win_reset(), add = TRUE)
  on.exit(win_register_deactivate(), add = TRUE)
  select_sql <- translate_select_sql(con, select)
  win_register_deactivate()

  named_windows <- win_register_names()
  if (nrow(named_windows) > 0 && supports_window_clause(con)) {
    # need to translate again and use registered windows names
    select_sql <- translate_select_sql(con, select)

    # build window sql
    names_esc <- sql_escape_ident(con, named_windows$name)
    window_sql <- sql(paste0(names_esc, " AS ", named_windows$key))
  } else {
    window_sql <- character()
  }

  list(
    select_sql = select_sql,
    window_sql = window_sql
  )
}

select_use_star <- function(select, vars_prev, con) {
  if (!supports_star_without_alias(con)) {
    return(select)
  }

  first_match <- vctrs::vec_match(vars_prev[[1]], select$name)
  if (is.na(first_match)) {
    return(select)
  }

  last <- first_match + length(vars_prev) - 1
  n <- vctrs::vec_size(select)

  if (n < last) {
    return(select)
  }

  test_cols <- vctrs::vec_slice(select, seq2(first_match, last))

  if (is_select_identity(test_cols, vars_prev)) {
    idx_start <- seq2(1, first_match - 1)
    idx_end <- seq2(last + 1, n)
    vctrs::vec_rbind(
      vctrs::vec_slice(select, idx_start),
      tibble(name = "", expr = list(sql("*"))),
      vctrs::vec_slice(select, idx_end)
    )
  } else {
    select
  }
}

translate_select_sql <- function(con, select_df) {
  select_df <- transmute(
    select_df,
    dots = set_names(expr, name),
    vars_group = .data$group_vars,
    vars_order = .data$order_vars,
    vars_frame = .data$frame
  )

  out <- purrr::pmap(
    select_df,
    function(dots, vars_group, vars_order, vars_frame) {
      translate_sql_(
        list(dots), con,
        vars_group = translate_sql_(syms(vars_group), con),
        vars_order = translate_sql_(vars_order, con, context = list(clause = "ORDER")),
        vars_frame = vars_frame[[1]],
        context = list(clause = "SELECT")
      )
    }
  )

  sql(unlist(out))
}

#' Build SQL strings with glue syntax
#'
#' @description
#' `sql_glue()` and `sql_glue2()` are designed to help dbplyr extenders
#' generate custom SQL. They differ only in whether or not they require
#' a connection. `sql_glue()` retrieves the ambient connection, making it
#' suitable for use inside [sql_translation()] methods; `sql_glue2()` requires
#' a connection, making it suitable for use inside all other `sql_` methods.
#'
#' As the name suggests, these functions use \pkg{glue} syntax to make it
#' easy to mix fixed SQL with varying user inputs. The default glue syntax, `{x}`,
#' will escape `x` using the database connection. If there are multiple values
#' in `x`, they'll be collapsed into a single string with `,`. If you want them
#' to be wrapped in `()`, use a `*` suffix, e.g. `{x*}`.
#'
#' You can also use type markers to control how the value is treated:
#'
#' * `{.sql x}`: `x` is literal SQL that should be interpolated as
#'   is, without additional escaping. `x` must be a string.
#' * `{.tbl x}`: `x` is a table identifier like a string, `I()`, or one of
#'   the older forms like `DBI::Id()` or `in_schema()`.
#' * `{.id x}`: `x` is a generic identifier, e.g. for a column or index.
#'    `x` must be a character vector
#'
#' @param con A [sql_dialect] object or database connection. Connections are
#'   supported for backward compatibility.
#' @param sql A string to interpolate.
#' @param envir Environment to evaluate `sql` in.
#' @return An SQL string.
#' @export
#' @examples
#' con <- dialect_ansi()
#'
#' tbl <- "my_table"
#' sql_glue2(con, "SELECT * FROM {.tbl tbl}")
#'
#' # Values are properly escaped
#' name <- "Robert'); DROP TABLE students;--"
#' sql_glue2(con, "INSERT INTO students (name) VALUES ({name})")
#'
#' # Control wrapping with *
#' x <- c("name", "age", "grade")
#' sql_glue2(con, "SELECT {.id x} FROM students")
#' sql_glue2(con, "SELECT * WHERE variable IN {x*}")
#' @export
sql_glue <- function(sql, envir = parent.frame()) {
  dbplyr_glue(sql_current_con(), sql, envir = envir)
}

#' @export
#' @rdname sql_glue
sql_glue2 <- function(con, sql, envir = parent.frame()) {
  check_con(con)
  dbplyr_glue(con, sql, envir = envir)
}


dbplyr_glue <- function(con, sql, envir = caller_env(), call = caller_env()) {
  sql(glue(
    sql,
    .envir = envir,
    .na = "NULL",
    .null = "",
    .transformer = function(text, envir) {
      glue_transformer(con, text, envir, call = call)
    }
  ))
}


glue_transformer <- function(con, text, envir, call = caller_env()) {
  parsed <- parse_glue_spec(text)
  if (!parsed$type %in% c("sql", "tbl", "id", "")) {
    cli::cli_abort(
      "Unknown marker {.val {parsed$type}} in {{{text}}}.",
      call = call
    )
  }

  if (parsed$value == "...") {
    value <- eval(quote(list(...)), envir)
  } else {
    value <- wrap_glue_error(
      eval(parse(text = parsed$value, keep.source = FALSE), envir),
      text,
      call
    )
  }

  # Coerce types that need coercion
  if (parsed$type == "sql") {
    if (!is_string(value)) {
      cli::cli_abort("{{{text}}} must be passed a string.", call = call)
    }
    value <- sql(value)
  } else if (parsed$type == "tbl") {
    value <- wrap_glue_error(as_table_source(value, con), text, call)
    value <- sql_escape_table_source(con, value)
    value <- sql_collapse(value, collapse = ", ", parens = parsed$collapse)
  } else if (parsed$type == "id") {
    if (!is_bare_character(value)) {
      cli::cli_abort(
        "{{{text}}} must be passed a character vector.",
        call = call
      )
    }
    value <- sql_escape_ident(con, value)
    value <- sql_collapse(value, collapse = ", ", parens = parsed$collapse)
  } else {
    value <- escape(value, collapse = ", ", parens = parsed$collapse, con = con)
  }

  unclass(value)
}

wrap_glue_error <- function(code, text, call = caller_env) {
  withCallingHandlers(
    code,
    error = function(e) {
      cli::cli_abort(
        "Failed to interpolate {{{text}}.",
        call = call,
        parent = e
      )
    }
  )
}

parse_glue_spec <- function(text) {
  # optional format string, like .id, .table
  # followed by R expression
  # followed by optional *, and whitespace
  re <- "^(\\.([^ ]+) )?\\s*([^*]+)([*]?)\\s*$"
  rc <- regexec(re, text)
  rm <- regmatches(text, rc)[[1]]

  list(
    type = rm[[3]],
    value = rm[[4]],
    collapse = rm[[5]] == "*"
  )
}

test_that("glue_sql() handles type casting", {
  con <- dialect_ansi()
  x <- "x"

  expect_equal(sql_glue2(con, "{x}"), sql("'x'"))
  expect_equal(sql_glue2(con, "{.id x}", ), sql('"x"'))
  expect_equal(sql_glue2(con, "{.sql x}", ), sql("x"))

  tbl1 <- "table"
  tbl2 <- I("schema.table")
  tbl3 <- in_schema("schema", "table")
  expect_equal(sql_glue2(con, "{.tbl tbl1}"), sql('"table"'))
  expect_equal(sql_glue2(con, "{.tbl tbl2}"), sql("schema.table"))
  expect_equal(sql_glue2(con, "{.tbl tbl3}"), sql('"schema"."table"'))
})

test_that("useful error if bad identifier", {
  con <- dialect_ansi()
  expect_snapshot(error = TRUE, {
    sql_glue2(con, "{.id 1}")
    sql_glue2(con, "{.tbl 1}")
    sql_glue2(con, "{.sql 1}")
  })
})

test_that("glue_sql() can collapse with and without parens", {
  con <- dialect_ansi()
  x <- c("a", "b")

  expect_equal(sql_glue2(con, "{x}"), sql("'a', 'b'"))
  expect_equal(sql_glue2(con, "{x*}"), sql("('a', 'b')"))
})


test_that("glue_sql() can interpolate ...", {
  con <- dialect_ansi()
  f <- function(...) {
    sql_glue2(con, "f({...})")
  }
  expect_equal(f(), sql("f()"))
  expect_equal(f("a"), sql("f('a')"))
  expect_equal(f("a", "b"), sql("f('a', 'b')"))
})

test_that("gives informative errors", {
  con <- dialect_ansi()
  x <- 1
  expect_snapshot(error = TRUE, {
    sql_glue2(con, "{y*}")
    sql_glue2(con, "{1 + }")
    sql_glue2(con, "{.bar x}")
  })
})

#' @export
#' @rdname sql_build
lazy_query <- function(query_type,
                       x,
                       ...,
                       group_vars = op_grps(x),
                       order_vars = op_sort(x),
                       frame = op_frame(x)) {
  # stopifnot(is.null(group_vars) || (is.character(group_vars) && is.null(names(group_vars))))
  stopifnot(is_lazy_sql_part(order_vars), is.null(names(order_vars)))
  stopifnot(is.null(frame) || is_integerish(frame$range, n = 2))

  structure(
    list(
      x = x,
      ...,
      group_vars = group_vars,
      order_vars = order_vars,
      frame = frame
    ),
    class = c(paste0("lazy_", query_type, "_query"), "lazy_query")
  )
}

#' Lazy operations
#'
#' @description
#' This set of S3 classes describe the action of dplyr verbs. These are
#' currently used for SQL sources to separate the description of operations
#' in R from their computation in SQL. This API is very new so is likely
#' to evolve in the future.
#'
#' `op_vars()` and `op_grps()` compute the variables and groups from
#' a sequence of lazy operations. `op_sort()` and `op_frame()` tracks the
#' order and frame for use in window functions.
#'
#' @keywords internal
#' @name lazy_ops
NULL

# op_grps -----------------------------------------------------------------

#' @export
#' @rdname lazy_ops
op_grps <- function(op) UseMethod("op_grps")
#' @export
op_grps.tbl_lazy <- function(op) op_grps(op$lazy_query)
#' @export
op_grps.lazy_query <- function(op) op$group_vars %||% character()

# op_vars -----------------------------------------------------------------

#' @export
#' @rdname lazy_ops
op_vars <- function(op) UseMethod("op_vars")
#' @export
op_vars.tbl_lazy <- function(op) op_vars(op$lazy_query)
#' @export
op_vars.lazy_base_query <- function(op) op$vars

# op_sort -----------------------------------------------------------------

#' @export
#' @rdname lazy_ops
op_sort <- function(op) UseMethod("op_sort")
#' @export
op_sort.tbl_lazy <- function(op) op_sort(op$lazy_query)
#' @export
op_sort.lazy_query <- function(op) {
  # Renaming (like for groups) cannot be done because:
  # * `order_vars` is a list of expressions
  # * variables needed in sorting can be dropped
  op$order_vars
}

# op_frame ----------------------------------------------------------------

#' @export
#' @rdname lazy_ops
op_frame <- function(op) UseMethod("op_frame")
#' @export
op_frame.tbl_lazy <- function(op) op_frame(op$lazy_query)
#' @export
op_frame.lazy_query <- function(op) op$frame

test_that("select quotes correctly", {
  out <- local_memdb_frame(x = 1, y = 1) |>
    select(x) |>
    collect()
  expect_equal(out, tibble(x = 1))
})

test_that("select can rename", {
  out <- local_memdb_frame(x = 1, y = 2) |>
    select(y = x) |>
    collect()
  expect_equal(out, tibble(y = 1))
})

test_that("two selects equivalent to one", {
  mf <- local_memdb_frame(a = 1, b = 1, c = 1, d = 2)

  out <- mf |>
    select(a:c) |>
    select(b:c) |>
    collect()

  expect_named(out, c("b", "c"))
})

test_that("correctly inlines across all verbs", {
  lf <- lazy_frame(x = 1, y = 2)

  # single table verbs
  expect_selects(lf |> arrange(x) |> select(x), 1)
  expect_selects(lf |> distinct() |> select(x), 2)
  expect_selects(lf |> filter(x == 1) |> select(x), 1)
  expect_selects(lf |> head(1) |> select(x), 1)
  expect_selects(lf |> mutate(z = x + 1) |> select(x), 1)
  expect_selects(lf |> select(y = x) |> select(y), 1)
  expect_selects(lf |> summarise(y = mean(x)) |> select(y), 1)

  # two table verbs
  lf2 <- lazy_frame(x = 1)
  expect_selects(lf |> left_join(lf2, by = "x") |> select(x), 1)
  expect_selects(lf |> right_join(lf2, by = "x") |> select(x), 1)
  expect_selects(lf |> semi_join(lf2, by = "x") |> select(x), 2)
  expect_selects(lf |> union(lf2) |> select(x), 3)
})

test_that("select after distinct produces subquery", {
  lf <- lazy_frame(x = 1, y = 1:2)
  expect_snapshot(
    lf |> distinct() |> select(x)
  )

  out <- lf |> distinct() |> select(x)
  lq <- out$lazy_query
  expect_false(lq$distinct)
  expect_true(lq$x$distinct)
})

test_that("select after arrange produces subquery, if needed", {
  lf <- lazy_frame(x = 1)

  # shouldn't inline
  out <- lf |> mutate(z = 2) |> arrange(x, z) |> select(x)
  # should inline
  out2 <- lf |> mutate(z = 2) |> arrange(x, z) |> select(x, z)

  inner_query <- out$lazy_query$x
  expect_s3_class(inner_query, "lazy_select_query")
  expect_equal(inner_query$order_by, list(expr(x), expr(z)))
  expect_equal(
    op_vars(inner_query),
    c("x", "z")
  )
  expect_equal(op_vars(out$lazy_query), "x")

  # order vars in a subquery are dropped
  expect_equal(
    inner_query[setdiff(names(inner_query), "order_vars")],
    out2$lazy_query[setdiff(names(out2$lazy_query), "order_vars")]
  )
})


test_that("rename/relocate after distinct is inlined #1141", {
  lf <- lazy_frame(x = 1, y = 1:2)
  expect_snapshot({
    lf |> distinct() |> rename(z = y)
    lf |> distinct() |> relocate(y)
  })

  out <- lf |> distinct() |> rename(z = y)
  lq <- out$lazy_query
  expect_true(lq$distinct)

  out <- lf |> distinct() |> relocate(y)
  lq <- out$lazy_query
  expect_true(lq$distinct)
})

test_that("select operates on mutated vars", {
  mf <- local_memdb_frame(x = c(1, 2, 3), y = c(3, 2, 1))
  expect_equal(mf |> mutate(x, z = x + y) |> select(z) |> pull(), c(4, 4, 4))
})

test_that("select renames variables (#317)", {
  mf <- local_memdb_frame(x = 1, y = 2)
  expect_equal(mf |> select(A = x) |> collect(), tibble(A = 1))
})

test_that("rename renames variables", {
  mf <- local_memdb_frame(x = 1, y = 2)
  expect_equal(mf |> rename(A = x) |> collect(), tibble(A = 1, y = 2))
})

test_that("can rename multiple vars", {
  mf <- local_memdb_frame(a = 1, b = 2)
  exp <- tibble(c = 1, d = 2)

  expect_equal(mf |> rename(c = a, d = b) |> collect(), exp)
  expect_equal(
    mf |> group_by(a) |> rename(c = a, d = b) |> collect(),
    exp |> group_by(c)
  )
})

test_that("can rename with a function", {
  mf <- local_memdb_frame(a = 1, b = 2)

  expect_named(mf |> rename_with(toupper) |> collect(), c("A", "B"))
  expect_named(mf |> rename_with(toupper, 1) |> collect(), c("A", "b"))
})

test_that("select preserves grouping vars", {
  mf <- local_memdb_frame(a = 1, b = 2) |> group_by(b)
  expect_snapshot(out <- mf |> select(a) |> collect())

  expect_named(out, c("b", "a"))
})

test_that("select handles order vars", {
  lf <- lazy_frame(x = 1, y = 1, z = 1)
  # can drop order vars
  expect_equal(lf |> window_order(y) |> select(-y) |> op_sort(), list())
  expect_equal(
    lf |> window_order(desc(y)) |> select(-y) |> op_sort(),
    list()
  )
  # can rename order vars
  expect_equal(
    lf |> window_order(y) |> select(y2 = y) |> op_sort(),
    list(expr(y2))
  )
  expect_equal(
    lf |> window_order(desc(y)) |> select(y2 = y) |> op_sort(),
    list(expr(desc(y2)))
  )
  # keeps sort order
  expect_equal(
    lf |> window_order(x, y) |> select(y2 = y, x) |> op_sort(),
    list(expr(x), expr(y2))
  )
})

test_that("select doesn't relocate grouping vars to the front", {
  mf <- local_memdb_frame(a = 1, b = 2) |> group_by(b)
  expect_equal(mf |> select(a, b) |> op_vars(), c("a", "b"))
})

test_that("relocate works", {
  mf <- local_memdb_frame(a = 1, b = 2, c = 1) |> group_by(b)

  out1 <- mf |> relocate(c) |> collect()
  expect_named(out1, c("c", "a", "b"))
  out2 <- mf |> relocate(a, .after = c) |> collect()
  expect_named(out2, c("b", "c", "a"))
})

test_that("relocate can rename variables", {
  mf <- local_memdb_frame(a = 1, b = 2, c = 1) |> group_by(b)

  out1 <- mf |> relocate(d = b) |> collect()
  expect_named(out1, c("d", "a", "c"))
  expect_equal(group_vars(out1), "d")
})

test_that("only add step if necessary", {
  lf <- lazy_frame(x = 1:3, y = 1:3)
  expect_equal(lf |> select(everything()), lf)
  expect_equal(lf |> select(x, y), lf)

  expect_equal(lf |> rename(x = x), lf)
  expect_equal(lf |> rename(), lf)

  expect_equal(lf |> relocate(x, y), lf)
  expect_equal(lf |> relocate(), lf)
})

test_that("select() after left_join() is inlined", {
  lf1 <- lazy_frame(x = 1, a = 1, .name = "lf1")
  lf2 <- lazy_frame(x = 1, b = 2, .name = "lf2")

  expect_snapshot(
    (out <- left_join(lf1, lf2, by = "x") |>
      select(b, x))
  )
  expect_equal(op_vars(out), c("b", "x"))

  expect_snapshot(
    (out <- left_join(lf1, lf2, by = "x") |>
      relocate(b))
  )
  expect_equal(op_vars(out), c("b", "x", "a"))
  expect_equal(out$lazy_query$vars$var, c("b", "x", "a"))
  expect_equal(out$lazy_query$vars$table, c(2L, 1L, 1L))

  out <- left_join(lf1, lf2, by = "x") |>
    transmute(b, x = x + 1)
  expect_s3_class(out$lazy_query, "lazy_select_query")
})

test_that("select() after filter() on join preserves qualified where", {
  lf1 <- lazy_frame(x = 1, y = 2, .name = "t1")
  lf2 <- lazy_frame(x = 1, z = 3, .name = "t2")

  # drop a column referenced in $where: must fall through to lazy_select_query
  # so the join's WHERE still qualifies z as t2.z.
  out <- lf1 |>
    left_join(lf2, by = "x") |>
    filter(z > 0) |>
    select(-z)
  expect_s3_class(out$lazy_query, "lazy_select_query")
  expect_snapshot(out)

  # rename a column referenced in $where: same handling.
  out <- lf1 |>
    left_join(lf2, by = "x") |>
    filter(z > 0) |>
    rename(zz = z)
  expect_s3_class(out$lazy_query, "lazy_select_query")
  expect_snapshot(out)

  # dropping a column not in $where stays inline.
  out <- lf1 |>
    left_join(lf2, by = "x") |>
    filter(z > 0) |>
    select(-y)
  expect_s3_class(out$lazy_query, "lazy_multi_join_query")
})

test_that("select() after semi_join() is inlined", {
  lf1 <- lazy_frame(x = 1, a = 1, .name = "lf1")
  lf2 <- lazy_frame(x = 1, b = 2, .name = "lf2")

  expect_snapshot(
    (out <- semi_join(lf1, lf2, by = "x") |>
      select(x, a2 = a))
  )
  expect_equal(op_vars(out), c("x", "a2"))

  expect_snapshot(
    (out <- anti_join(lf1, lf2, by = "x") |>
      relocate(a))
  )
  expect_equal(op_vars(out), c("a", "x"))

  out <- semi_join(lf1, lf2, by = "x") |>
    transmute(a, x = x + 1)
  expect_s3_class(out$lazy_query, "lazy_select_query")
})

test_that("select() after join handles previous select", {
  lf <- lazy_frame(x = 1, y = 1, z = 1) |>
    group_by(x, y, z) |>
    select(x, y2 = y, z) |>
    semi_join(
      lazy_frame(x = 1),
      by = "x"
    ) |>
    select(x2 = x, y3 = y2, z)

  expect_equal(op_vars(lf), c("x2", "y3", "z"))
  expect_equal(
    lf$lazy_query$vars,
    tibble(
      name = c("x2", "y3", "z"),
      var = c("x", "y", "z")
    )
  )
  expect_equal(op_grps(lf), c("x2", "y3", "z"))
  expect_snapshot(print(lf))

  lf2 <- lazy_frame(x = 1, y = 1, z = 1) |>
    group_by(x, y, z) |>
    select(x, y2 = y, z) |>
    left_join(
      lazy_frame(x = 1, y = 1),
      by = "x"
    ) |>
    select(x2 = x, y3 = y2, z)

  expect_equal(op_vars(lf2), c("x2", "y3", "z"))
  vars2 <- lf2$lazy_query$vars
  expect_equal(vars2$var, c("x", "y", "z"))
  expect_equal(vars2$table, c(1L, 1L, 1L))

  expect_equal(op_grps(lf2), c("x2", "y3", "z"))
  expect_snapshot(print(lf2))
})

test_that("select() afer join keeps grouping", {
  lf1 <- lazy_frame(x = 1, y = 1) |> group_by(y)
  lf2 <- lazy_frame(x = 1, z = 1) |> group_by(z)

  # just to be sure check without select/renaming
  expect_equal(left_join(lf1, lf2, by = "x") |> op_grps(), "y")

  # rename grouping variable
  expect_equal(
    left_join(lf1, lf2, by = "x") |>
      select(y2 = y) |>
      op_grps(),
    "y2"
  )
})

test_that("select() produces nice error messages", {
  lf <- lazy_frame(x = 1)

  expect_snapshot(error = TRUE, {
    lf |> select(non_existent)
    lf |> select(non_existent + 1)
  })

  expect_snapshot(error = TRUE, {
    lf |> relocate(non_existent)
    lf |> relocate(non_existent + 1)
  })

  expect_snapshot(error = TRUE, {
    # no name
    lf |> rename(x)
    # non-existing column
    lf |> rename(y = non_existent)
    lf |> rename(y = non_existent + 1)
  })

  expect_snapshot(error = TRUE, {
    lf |> rename_with(toupper, .cols = non_existent)
    lf |> rename_with(toupper, .cols = non_existent + 1)
  })
})

test_that("where() isn't suppored", {
  lf <- lazy_frame(x = 1)
  expect_snapshot(error = TRUE, {
    lf |> select(where(is.integer))
  })
})

test_that("arranged computed columns are not inlined away", {
  lf <- lazy_frame(x = 1)

  # shouldn't inline
  out <- lf |> mutate(z = 2) |> arrange(x, z) |> select(x)
  # should inline
  out2 <- lf |> mutate(z = 2) |> arrange(x, z) |> select(x, z)

  inner_query <- out$lazy_query$x
  expect_snapshot({
    lf |> mutate(z = 1) |> arrange(x, z) |> select(x)
  })
  expect_s3_class(inner_query, "lazy_select_query")
  expect_equal(inner_query$order_by, list(expr(x), expr(z)))
  expect_equal(op_vars(inner_query), c("x", "z"))
  expect_equal(op_vars(out$lazy_query), "x")
  expect_equal(
    # order vars in a subquery are dropped
    inner_query[setdiff(names(inner_query), "order_vars")],
    out2$lazy_query[setdiff(names(out2$lazy_query), "order_vars")]
  )
})

# sql_render --------------------------------------------------------------

test_that("multiple selects are collapsed", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_snapshot(lf |> select(2:1) |> select(2:1))
  expect_snapshot(lf |> select(2:1) |> select(2:1) |> select(2:1))
  expect_snapshot(lf |> select(x1 = x) |> select(x2 = x1))
})

test_that("mutate collapses over nested select", {
  lf <- lazy_frame(g = 0, x = 1, y = 2)

  expect_snapshot(lf |> mutate(a = 1, b = 2) |> select(a))
  expect_snapshot(lf |> mutate(a = 1, b = 2) |> select(x))
})

test_that("output is styled", {
  local_reproducible_output(crayon = TRUE)
  withr::local_options(dbplyr_highlight = cli::combine_ansi_styles("blue"))

  lf <- lazy_frame(x = 1, y = 1, z = 1)
  out <- lf |>
    group_by(x) |>
    mutate(y = mean(y, na.rm = TRUE), z = z + 1) |>
    filter(z == 1) |>
    left_join(lf, by = "x")

  expect_snapshot(show_query(out, sql_options = sql_options(cte = TRUE)))
})

# sql_build -------------------------------------------------------------

test_that("select picks variables", {
  out <- lazy_frame(x1 = 1, x2 = 1, x3 = 2) |>
    select(x1:x2) |>
    sql_build()

  expect_equal(out$select, sql("\"x1\"", "\"x2\""))
})

test_that("select renames variables", {
  out <- lazy_frame(x1 = 1, x2 = 1, x3 = 2) |>
    select(y = x1, z = x2) |>
    sql_build()

  expect_equal(out$select, sql("\"x1\" AS \"y\"", "\"x2\" AS \"z\""))
})

test_that("select can refer to variables in local env", {
  vars <- c("x", "y")
  out <- lazy_frame(x = 1, y = 1, z = 1) |>
    select(dplyr::one_of(vars)) |>
    sql_build()

  expect_equal(out$select, sql("\"x\"", "\"y\""))
})

test_that("rename preserves existing vars", {
  out <- lazy_frame(x = 1, y = 1) |>
    rename(z = y) |>
    sql_build()

  expect_equal(out$select, sql("\"x\"", "\"y\" AS \"z\""))
})


# ops ---------------------------------------------------------------------

test_that("select reduces variables", {
  out <- mtcars |> tbl_lazy() |> select(mpg:disp)
  expect_equal(op_vars(out), c("mpg", "cyl", "disp"))
})

test_that("rename preserves existing", {
  out <- tibble(x = 1, y = 2) |> tbl_lazy() |> rename(z = y)
  expect_equal(op_vars(out), c("x", "z"))
})

test_that("rename renames grouping vars", {
  df <- lazy_frame(a = 1, b = 2)
  expect_equal(df |> group_by(a) |> rename(c = a) |> op_grps(), "c")
})

test_that("mutate preserves grouping vars (#396)", {
  df <- lazy_frame(a = 1, b = 2, c = 3) |> group_by(a, b)
  expect_equal(df |> mutate(a = 1) |> op_grps(), c("a", "b"))
  expect_equal(df |> mutate(b = 1) |> op_grps(), c("a", "b"))
})

test_that("select after arrange(desc()) works (#1206)", {
  out <- lazy_frame(x = 1, y = 1) |>
    arrange(desc(x)) |>
    select(x)
  expect_equal(op_vars(out), "x")
})


# lazy_select_query -------------------------------------------------------

test_that("select, relocate, and rename work", {
  lf <- lazy_frame(x = 1, y = 1)

  expect_equal(
    select(lf, x)$lazy_query$select,
    new_lazy_select(exprs(x = x))
  )

  expect_equal(
    relocate(lf, y)$lazy_query$select,
    new_lazy_select(exprs(y = y, x = x))
  )

  expect_equal(
    rename(lf, b = y, a = x)$lazy_query$select,
    new_lazy_select(exprs(a = x, b = y))
  )
})

test_that("renaming handles groups correctly", {
  lf <- lazy_frame(x = 1, y = 1) |>
    group_by(x) |>
    rename(ax = x)

  result <- lf$lazy_query
  expect_equal(
    result$select,
    new_lazy_select(exprs(ax = x, y = y))
  )

  expect_equal(result$group_vars, "ax")
  expect_equal(op_grps(result), "ax")

  result <- rename(lf, x = ax)$lazy_query
  expect_equal(
    result$select,
    new_lazy_select(exprs(x = x, y = y))
  )

  expect_equal(result$group_vars, "x")
  expect_equal(op_grps(result), "x")

  # https://github.com/tidyverse/dbplyr/issues/928
  result <- lazy_frame(cyl = 1, vs = 1) |>
    rename(vs = cyl, new_vs = vs) |>
    group_by(vs)
  expect_equal(op_grps(result), "vs")
})

test_that("slice, head and tail aren't available", {
  lf <- lazy_frame(x = 1)
  expect_snapshot(error = TRUE, lf %>% slice())
  expect_snapshot(error = TRUE, lf %>% slice_head())
  expect_snapshot(error = TRUE, lf %>% slice_tail())
})

test_that("slice_min handles arguments", {
  db <- memdb_frame(x = c(1, 1, 2), id = c(1, 2, 3))

  expect_equal(db %>% slice_min(id) %>% pull(), 1)
  expect_equal(db %>% slice_min(x) %>% pull(), c(1, 2))
  expect_equal(db %>% slice_min(x, with_ties = FALSE) %>% pull(), 1)
  expect_equal(db %>% slice_min(id, n = 2) %>% pull(), c(1, 2))
  expect_equal(db %>% slice_min(id, prop = 0.5) %>% pull(), 1)

  expect_snapshot(error = TRUE, db %>% slice_min())
  expect_snapshot(error = TRUE, db %>% slice_min(id, prop = 0.5, with_ties = FALSE))
})

test_that("slice_max orders in opposite order", {
  db <- memdb_frame(x = c(1, 1, 2), id = c(1, 2, 3))

  expect_equal(db %>% slice_max(id) %>% pull(), 3)
  expect_equal(db %>% slice_max(x) %>% pull(), 3)
  expect_snapshot(error = TRUE, db %>% slice_max())
})

test_that("slice_sample errors when expected", {
  db <- memdb_frame(x = c(1, 1, 2), id = c(1, 2, 3))

  # Can't see how to test this, but interactive experimentation
  # shows that it doesn't always return the same result
  expect_error(db %>% slice_sample() %>% pull(), NA)

  expect_snapshot(error = TRUE, db %>% slice_sample(replace = TRUE))
  expect_snapshot(error = TRUE, db %>% slice_sample(weight_by = x))
  expect_snapshot(error = TRUE, db %>% slice_sample(prop = 0.5))
})

test_that("window_order is preserved", {
  db <- memdb_frame(x = c(1, 1, 2), id = c(1, 2, 3))
  sort <- db %>% window_order(x) %>% slice_min(id) %>% op_sort()
  expect_equal(length(sort), 1)
  expect_equal(get_expr(sort[[1]]), quote(x))
})

test_that("check_slice_size checks for common issues", {
  lf <- lazy_frame(x = c(1, 1, 2), id = c(1, 2, 3))

  expect_snapshot(error = TRUE, lf %>% slice_sample(n = 1, prop = 1))
  expect_snapshot(error = TRUE, lf %>% slice_sample(n = "a"))
  expect_snapshot(error = TRUE, lf %>% slice_sample(prop = "a"))
  expect_snapshot(error = TRUE, lf %>% slice_sample(n = -1))
  expect_snapshot(error = TRUE, lf %>% slice_sample(prop = -1))
})

test_that("slice_helper `by` errors use correct error context and correct `by_arg`", {
  df <- lazy_frame(x = 1)
  gdf <- group_by(df, x)

  expect_snapshot(error = TRUE, {
    slice_min(gdf, order_by = x, by = x)
    slice_max(gdf, order_by = x, by = x)
    slice_sample(gdf, n = 1, by = x)
  })
})

test_that("slice_min/max() work with `by`", {
  df <- memdb_frame(g = c(2, 2, 1, 1), x = c(1, 2, 3, 1))

  out <- slice_min(df, x, by = g) %>% arrange(g) %>% collect()
  expect_identical(out, tibble(g = c(1, 2), x = 1))
  expect_equal(group_vars(out), character())
  expect_identical(slice_max(df, x, by = g) %>% arrange(g) %>% collect(), tibble(g = c(1, 2), x = c(3, 2)))
})

test_that("slice_sample() works with `by`", {
  df <- tibble(g = c(2, 2, 2, 1), x = c(1, 2, 3, 1))
  df <- memdb_frame(g = c(2, 2, 2, 1), x = c(1, 2, 3, 1))
  expect_identical(
    slice_sample(df, n = 2, by = g) %>% pull(g) %>% sort(),
    c(1, 2, 2)
  )
})

test_that("correctly inlines across all verbs", {
  lf <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 3)

  # single table verbs
  suppressWarnings(expect_selects(
    lf |> arrange(x) |> left_join(lf2, by = "x"),
    2
  ))
  expect_selects(lf |> distinct() |> left_join(lf2, by = "x"), 2)
  expect_selects(lf |> filter(x == 1) |> left_join(lf2, by = "x"), 2)
  expect_selects(lf |> head(1) |> left_join(lf2, by = "x"), 2)
  expect_selects(lf |> mutate(a = x + 1) |> left_join(lf2, by = "x"), 2)
  expect_selects(lf |> select(x) |> left_join(lf2, by = "x"), 1)
  expect_selects(lf |> summarise(x = mean(x)) |> left_join(lf2, by = "x"), 2)

  # two table verbs
  lf3 <- lazy_frame(x = 1)
  expect_selects(lf |> left_join(lf2, by = "x") |> left_join(lf3, by = "x"), 1)
  expect_selects(lf |> right_join(lf2, by = "x") |> left_join(lf3, by = "x"), 2)
  expect_selects(lf |> semi_join(lf2, by = "x") |> left_join(lf3, by = "x"), 3)
  expect_selects(lf |> union(lf3) |> left_join(lf2, by = "x"), 3)
})

test_that("complete join pipeline works with SQLite", {
  df1 <- local_memdb_frame(x = 1:5)
  df2 <- local_memdb_frame(x = c(1, 3, 5), y = c("a", "b", "c"))

  out <- collect(left_join(df1, df2, by = "x"))
  expect_equal(out, tibble(x = 1:5, y = c("a", NA, "b", NA, "c")))

  df1 <- local_memdb_frame(x = 1:3, y = c("x", "y", "z"))
  df2 <- local_memdb_frame(
    "df4",
    x = c(1, 3, 5),
    y = c("a", "b", "c"),
    z = 11:13
  )

  expect_equal(
    collect(full_join(df1, df2, by = "x")),
    tibble(
      x = c(1, 2, 3, 5),
      y.x = c("x", "y", "z", NA),
      y.y = c("a", NA, "b", "c"),
      z = c(11, NA, 12, 13)
    )
  )
})

test_that("complete join pipeline works with SQLite and table alias", {
  df1 <- local_memdb_frame(x = 1:5)
  df2 <- local_memdb_frame(x = c(1, 3, 5), y = c("a", "b", "c"))

  out <- left_join(df1, df2, by = "x", x_as = "df1", y_as = "df2")
  expect_equal(out |> collect(), tibble(x = 1:5, y = c("a", NA, "b", NA, "c")))

  lf1 <- lazy_frame(x = 1:5, .name = "lf1")
  lf2 <- lazy_frame(x = c(1, 3, 5), y = c("a", "b", "c"), .name = "lf2")
  expect_snapshot(left_join(lf1, lf2, by = "x", x_as = "df1", y_as = "df2"))
})

test_that("complete semi join works with SQLite", {
  lf1 <- local_memdb_frame(x = c(1, 2), y = c(2, 3))
  lf2 <- local_memdb_frame(x = 1)

  lf3 <- inner_join(lf1, lf2, by = "x")
  expect_equal(op_vars(lf3), c("x", "y"))

  out <- collect(lf3)
  expect_equal(out, tibble(x = 1, y = 2))
})

test_that("complete semi join works with SQLite and table alias", {
  df1 <- local_memdb_frame(x = c(1, 2), y = c(2, 3))
  df2 <- local_memdb_frame(x = 1)

  out <- inner_join(df1, df2, by = "x", x_as = "df1", y_as = "df2")
  expect_equal(out |> collect(), tibble(x = 1, y = 2))

  lf1 <- lazy_frame(x = c(1, 2), y = c(2, 3))
  lf2 <- lazy_frame(x = 1)
  expect_snapshot(inner_join(lf1, lf2, by = "x", x_as = "df1", y_as = "df2"))
})

test_that("join works with in_schema", {
  con <- local_sqlite_connection()

  DBI::dbExecute(con, "ATTACH ':memory:' AS foo")
  local_db_table(
    con,
    tibble(x = 1:3, y = "a"),
    DBI::Id(schema = "foo", table = "df"),
    temporary = FALSE
  )
  local_db_table(
    con,
    tibble(x = 2:3, z = "b"),
    DBI::Id(schema = "foo", table = "df2"),
    temporary = FALSE
  )

  # same schema, different name
  df1 <- tbl(con, in_schema("foo", "df"))
  df2 <- tbl(con, in_schema("foo", "df2"))
  expect_equal(
    left_join(df1, df2, by = "x") |> collect(),
    tibble(x = 1:3, y = "a", z = c(NA, "b", "b"))
  )
  expect_snapshot(
    left_join(df1, df2, by = "x") |> remote_query()
  )

  # different schema, same name
  DBI::dbExecute(con, "ATTACH ':memory:' AS foo2")
  local_db_table(
    con,
    tibble(x = 2:3, z = "c"),
    DBI::Id(schema = "foo2", table = "df"),
    temporary = FALSE
  )
  df3 <- tbl(con, in_schema("foo2", "df"))
  expect_equal(
    left_join(df1, df3, by = "x") |> collect(),
    tibble(x = 1:3, y = "a", z = c(NA, "c", "c"))
  )
  expect_snapshot(
    left_join(df1, df3, by = "x") |> remote_query()
  )

  # use auto name for escaped table names
  df4 <- tbl(con, I("foo.df"))
  df5 <- tbl(con, I("foo2.df"))
  expect_snapshot(
    left_join(df4, df5, by = "x") |> remote_query()
  )
  out <- left_join(df4, df5, by = "x")
  expect_equal(
    out$lazy_query$table_names,
    tibble(
      name = table_path(c("foo.df", "foo2.df")),
      from = c("name", "name")
    )
  )
})

test_that("alias truncates long table names at database limit", {
  # Postgres has max identifier length of 63; ensure it's not exceeded when generating table alias
  # Source: https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
  con <- test_postgres()

  nm1 <- table_path(paste0("a", paste0(0:61 %% 10, collapse = "")))
  mf1 <- local_db_table(con, tibble(x = 1:3, y = "a"), unclass(nm1))

  nm2 <- table_path(paste0("b", paste0(0:61 %% 10, collapse = "")))
  mf2 <- local_db_table(con, tibble(x = 2:3, y = "b"), unclass(nm2))

  # 2 tables
  # aliased names are as expected
  self_join2_names <- generate_join_table_names(
    tibble::tibble(
      name = c(nm1, nm1),
      from = "name"
    ),
    con
  )
  expect_equal(max(nchar(self_join2_names)), 65)
  expect_equal(
    length(self_join2_names),
    length(unique(self_join2_names))
  )

  # joins correctly work
  self_join2 <- left_join(mf1, mf1, by = c("x", "y"))

  expect_equal(
    self_join2 |> collect(),
    tibble(x = 1:3, y = "a")
  )

  expect_snapshot(
    self_join2 |> remote_query()
  )

  # 3 tables
  # aliased names are as expected
  self_join3_names <- generate_join_table_names(
    tibble::tibble(
      name = c(nm1, nm1, nm2),
      from = "name"
    ),
    con
  )

  expect_equal(max(nchar(self_join3_names)), 65)
  expect_equal(
    length(self_join3_names),
    length(unique(self_join3_names))
  )

  # joins correctly work
  self_join3 <- left_join(mf1, mf1, by = c("x", "y")) |>
    inner_join(mf2, by = "x")

  expect_equal(
    self_join3 |> collect(),
    tibble(x = 2:3, y.x = "a", y.y = "b")
  )
  expect_snapshot(
    self_join3 |> remote_query()
  )
})

test_that("cross join via by = character() is deprecated", {
  df1 <- local_memdb_frame(x = 1:5)
  df2 <- local_memdb_frame(y = 1:5)

  expect_snapshot(error = TRUE, {
    inner_join(df1, df2, by = character())
    full_join(df1, df2, by = character())
  })
})

df1 <- local_memdb_frame(x = 1:5, y = 1:5)
df2 <- local_memdb_frame(a = 5:1, b = 1:5)
df3 <- local_memdb_frame(x = 1:5, z = 1:5)
df4 <- local_memdb_frame(a = 5:1, z = 5:1)

test_that("named by join by different x and y vars", {
  j1 <- collect(inner_join(df1, df2, c("x" = "a")))
  expect_equal(names(j1), c("x", "y", "b"))
  expect_equal(nrow(j1), 5)

  j2 <- collect(inner_join(df1, df2, c("x" = "a", "y" = "b")))
  expect_equal(names(j2), c("x", "y"))
  expect_equal(nrow(j2), 1)
})

test_that("named by join by same z vars", {
  j1 <- collect(inner_join(df3, df4, c("z" = "z")))
  expect_equal(nrow(j1), 5)
  expect_equal(names(j1), c("x", "z", "a"))
})

test_that("join with both same and different vars", {
  j1 <- collect(left_join(df1, df3, by = c("y" = "z", "x")))
  expect_equal(nrow(j1), 5)
  expect_equal(names(j1), c("x", "y"))
})

test_that("joining over arbitrary predicates", {
  j1 <- collect(left_join(
    df1,
    df2,
    x_as = "LHS",
    y_as = "RHS",
    sql_on = "LHS.x = RHS.b"
  ))
  j2 <- collect(left_join(df1, df2, by = c("x" = "b"))) |> mutate(b = x)
  expect_equal(j1, j2)

  j1 <- collect(left_join(
    df1,
    df3,
    x_as = "LHS",
    y_as = "RHS",
    sql_on = "LHS.x = RHS.z"
  ))
  j2 <- collect(left_join(df1, df3, by = c("x" = "z"))) |>
    mutate(z = x) |>
    rename(x.x = x)
  expect_equal(j1, j2)

  j1 <- collect(left_join(
    df1,
    df3,
    x_as = "LHS",
    y_as = "RHS",
    sql_on = "LHS.x = RHS.x"
  ))
  j2 <- collect(left_join(df1, df3, by = "x")) |>
    mutate(x.y = x) |>
    select(x.x = x, y, x.y, z)
  expect_equal(j1, j2)
})

test_that("inner join doesn't result in duplicated columns ", {
  expect_equal(colnames(inner_join(df1, df1, by = c("x", "y"))), c("x", "y"))
})

test_that("self-joins allowed with named by", {
  fam <- local_memdb_frame(id = 1:5, parent = c(NA, 1, 2, 2, 4))

  j1 <- fam |> left_join(fam, by = c("parent" = "id"))
  j2 <- fam |> inner_join(fam, by = c("parent" = "id"))

  expect_equal(op_vars(j1), c("id", "parent", "parent.y"))
  expect_equal(op_vars(j2), c("id", "parent", "parent.y"))
  expect_equal(nrow(collect(j1)), 5)
  expect_equal(nrow(collect(j2)), 4)

  j3 <- collect(semi_join(fam, fam, by = c("parent" = "id")))
  j4 <- collect(anti_join(fam, fam, by = c("parent" = "id")))

  expect_equal(j3, filter(collect(fam), !is.na(parent)))
  expect_equal(j4, filter(collect(fam), is.na(parent)))
})

test_that("suffix modifies duplicated variable names", {
  fam <- local_memdb_frame(id = 1:5, parent = c(NA, 1, 2, 2, 4))
  j1 <- collect(inner_join(
    fam,
    fam,
    by = c("parent" = "id"),
    suffix = c("1", "2")
  ))
  j2 <- collect(left_join(
    fam,
    fam,
    by = c("parent" = "id"),
    suffix = c("1", "2")
  ))

  expect_named(j1, c("id", "parent", "parent2"))
  expect_named(j2, c("id", "parent", "parent2"))
})

test_that("join variables always disambiguated (#2823)", {
  # Even if the new variable conflicts with an existing variable
  df1 <- local_memdb_frame(a = 1, b.x = 1, b = 1)
  df2 <- local_memdb_frame(a = 1, b = 1)

  both <- collect(left_join(df1, df2, by = "a"))
  expect_named(both, c("a", "b.x", "b.x.x", "b.y"))
})

test_that("join functions error on column not found for SQL sources #1928", {
  # Rely on dplyr to test precise code
  expect_error(
    left_join(
      local_memdb_frame(x = 1:5),
      local_memdb_frame(y = 1:5),
      by = "x"
    ),
    "present"
  )
  expect_error(
    left_join(
      local_memdb_frame(x = 1:5),
      local_memdb_frame(y = 1:5),
      by = "y"
    ),
    "present"
  )
  expect_error(
    left_join(
      local_memdb_frame(x = 1:5),
      local_memdb_frame(y = 1:5)
    ),
    "[Nn]o common variables"
  )
})

test_that("join check `x_as` and `y_as`", {
  x <- lazy_frame(a = 1, x = 1, .name = "x")
  y <- lazy_frame(a = 1, y = 1, .name = "y")

  expect_snapshot(error = TRUE, left_join(x, x, by = "x", y_as = c("A", "B")))
  expect_snapshot(
    error = TRUE,
    left_join(x, x, by = "x", x_as = "LHS", y_as = "LHS")
  )
})

test_that("join uses correct table alias", {
  x <- lazy_frame(a = 1, x = 1, .name = "x")
  y <- lazy_frame(a = 1, y = 1, .name = "y")

  # self joins
  table_paths <- sql_build(left_join(x, x, by = "a"))$table_names
  expect_equal(table_paths, table_path(c('"x_LHS"', '"x_RHS"')))

  table_paths <- sql_build(left_join(x, x, by = "a", x_as = "my_x"))$table_names
  expect_equal(table_paths, table_path(c('"my_x"', '"x"')))

  table_paths <- sql_build(left_join(x, x, by = "a", y_as = "my_y"))$table_names
  expect_equal(table_paths, table_path(c('"x"', '"my_y"')))

  table_paths <- sql_build(left_join(
    x,
    x,
    by = "a",
    x_as = "my_x",
    y_as = "my_y"
  ))$table_names
  expect_equal(table_paths, table_path(c('"my_x"', '"my_y"')))

  # x-y joins
  table_paths <- sql_build(left_join(x, y, by = "a"))$table_names
  expect_equal(table_paths, table_path(c('"x"', '"y"')))

  table_paths <- sql_build(left_join(x, y, by = "a", x_as = "my_x"))$table_names
  expect_equal(table_paths, table_path(c('"my_x"', '"y"')))

  table_paths <- sql_build(left_join(x, y, by = "a", y_as = "my_y"))$table_names
  expect_equal(table_paths, table_path(c('"x"', '"my_y"')))

  table_paths <- sql_build(left_join(
    x,
    y,
    by = "a",
    x_as = "my_x",
    y_as = "my_y"
  ))$table_names
  expect_equal(table_paths, table_path(c('"my_x"', '"my_y"')))

  # x_as same name as `y`
  table_paths <- sql_build(left_join(x, y, by = "a", x_as = "y"))$table_names
  expect_equal(table_paths, table_path(c('"y"', '"y...2"')))

  table_paths <- sql_build(left_join(
    x |> filter(x == 1),
    x,
    by = "x",
    y_as = "LHS"
  ))$table_names
  expect_equal(table_paths, table_path(c('"LHS...1"', '"LHS"')))

  # sql_on -> use alias or LHS/RHS
  table_paths <- sql_build(left_join(
    x,
    y,
    sql_on = sql("LHS.a = RHS.a")
  ))$table_names
  expect_equal(table_paths, table_path(c('"LHS"', '"RHS"')))

  table_paths <- sql_build(left_join(
    x,
    y,
    x_as = "my_x",
    sql_on = sql("my_x.a = RHS.a")
  ))$table_names
  expect_equal(table_paths, table_path(c('"my_x"', '"RHS"')))

  # triple join
  z <- lazy_frame(a = 1, z = 1, .name = "z")
  out <- left_join(x, y, by = "a") |>
    left_join(z, by = "a") |>
    sql_build()
  expect_equal(out$table_names, table_path(c('"x"', '"y"', '"z"')))

  # triple join where names need to be repaired
  out <- left_join(x, x, by = "a") |>
    left_join(z, by = "a") |>
    sql_build()
  expect_equal(out$table_names, table_path(c('"x...1"', '"x...2"', '"z"')))
})

test_that("select() before join is inlined", {
  lf <- lazy_frame(x1 = 10, a = 1, y = 3, .name = "lf1")
  lf2 <- lazy_frame(x2 = 10, b = 2, z = 4, .name = "lf2")

  test_vars <- function(lq, var, table) {
    expect_equal(lq$vars$name, c("a2", "x", "b"))
    expect_equal(lq$vars$var, var)
    expect_equal(lq$vars$table, table)

    expect_equal(lq$joins$by[[1]]$x, "x1")
    expect_equal(lq$joins$by[[1]]$y, "x2")
  }

  out_left <- left_join(
    lf |> select(a2 = a, x = x1),
    lf2 |> select(x = x2, b),
    by = "x"
  )
  test_vars(out_left$lazy_query, c("a", "x1", "b"), c(1, 1, 2))
  expect_equal(op_vars(out_left), c("a2", "x", "b"))
  expect_snapshot(out_left)

  out_inner <- inner_join(
    lf |> select(a2 = a, x = x1),
    lf2 |> select(x = x2, b),
    by = "x"
  )
  test_vars(out_inner$lazy_query, c("a", "x1", "b"), c(1, 1, 2))

  out_right <- right_join(
    lf |> select(a2 = a, x = x1),
    lf2 |> select(x = x2, b),
    by = "x"
  )$lazy_query

  expect_equal(out_right$vars$name, c("a2", "x", "b"))
  expect_equal(out_right$vars$x, c("a", NA, NA))
  expect_equal(out_right$vars$y, c(NA, "x2", "b"))
  expect_equal(out_right$by$x, "x1")
  expect_equal(out_right$by$y, "x2")

  out_full <- full_join(
    lf |> select(a2 = a, x = x1),
    lf2 |> select(x = x2, b),
    by = "x"
  )
  vars <- out_full$lazy_query$vars
  expect_equal(vars$name, c("a2", "x", "b"))
  expect_equal(vars$x, c("a", "x1", NA))
  expect_equal(vars$y, c(NA, "x2", "b"))

  out_cross <- cross_join(
    lf |> select(a2 = a, x = x1),
    lf2 |> select(x = x2, b)
  )
  vars <- out_cross$lazy_query$vars
  expect_equal(vars$name, c("a2", "x.x", "x.y", "b"))
  expect_equal(vars$var, c("a", "x1", "x2", "b"))
  expect_equal(vars$table, c(1, 1, 2, 2))

  # attributes like `group`, `sort`, `frame` is kept
  lf <- lazy_frame(x = 10, a = 1, b = 1, .name = "lf1")
  lf2 <- lazy_frame(x = 10, .name = "lf2")
  out_left <- left_join(
    lf |>
      group_by(a) |>
      arrange(a) |>
      window_frame(0, 1) |>
      select(x, a),
    lf2,
    by = "x"
  )
  expect_equal(op_grps(out_left), "a")
  expect_equal(op_sort(out_left), list(expr(a)))
  expect_equal(op_frame(out_left), list(range = c(0, 1)))
})

test_that("select() before join works for tables with same column name", {
  lf <- lazy_frame(id = 1, x = 1)
  lf2 <- lazy_frame(id = 12, x = 2)

  out <- left_join(
    lf |> rename(id1 = id),
    lf2 |> rename(id2 = id),
    by = "x"
  )

  lq <- out$lazy_query
  expect_equal(op_vars(lq), c("id1", "x", "id2"))
  expect_equal(lq$vars$var, c("id", "x", "id"))
  expect_equal(lq$vars$table, c(1, 1, 2))
})

test_that("named by works in combination with inlined select", {
  lf <- lazy_frame(id_x = 1, x = 1)
  lf2 <- lazy_frame(id_y = 12, x = 2)

  out <- left_join(
    lf |> select(id_x, x.x = x),
    lf2 |> select(id_y, x.y = x),
    by = c(id_x = "id_y", x.x = "x.y")
  )

  lq <- out$lazy_query
  expect_equal(op_vars(lq), c("id_x", "x.x"))
  expect_equal(lq$vars$var, c("id_x", "x"))
  expect_equal(lq$vars$table, c(1, 1))
  expect_equal(lq$joins$by[[1]]$x, c("id_x", "x"))
  expect_equal(lq$joins$by[[1]]$y, c("id_y", "x"))
})

test_that("rename works with duplicate column names in join_by (#1572)", {
  lf1 <- lazy_frame(x = 1, .name = "df")
  lf2 <- lazy_frame(y = 1, .name = "df")

  # between(x, z, z) produces by$y = c("z", "z")
  # both must be renamed back to "y"
  out <- left_join(lf1, lf2 |> rename(z = y), join_by(between(x, z, z)))

  lq <- out$lazy_query
  expect_equal(lq$joins$by[[1]]$x, c("x", "x"))
  expect_equal(lq$joins$by[[1]]$y, c("y", "y"))

  expect_snapshot(out)
})

test_that("suffix works in combination with inlined select", {
  lf <- lazy_frame(id = 1, x = 1)
  lf2 <- lazy_frame(id = 12, x = 2)

  out <- left_join(
    lf |> rename(x2 = x),
    lf2 |> rename(x2 = x),
    by = "id"
  )

  lq <- out$lazy_query
  expect_equal(op_vars(lq), c("id", "x2.x", "x2.y"))
  expect_equal(lq$vars$var, c("id", "x", "x"))
  expect_equal(lq$vars$table, c(1L, 1L, 2L))
})

test_that("select() before join is not inlined when using `sql_on`", {
  lf <- lazy_frame(x1 = 10, a = 1, y = 3)
  lf2 <- lazy_frame(x2 = 10, b = 2, z = 4)

  out <- left_join(
    lf |> select(a2 = a, x = x1),
    lf2 |> select(x = x2, b),
    sql_on = sql("LHS.x = RHS.x")
  )

  lq <- out$lazy_query
  expect_s3_class(lq$x, "lazy_select_query")
  expect_s3_class(lq$joins$table[[1]], "lazy_select_query")
  expect_equal(lq$vars$var, c("a2", "x", "x", "b"))
  expect_equal(lq$vars$table, c(1L, 1L, 2L, 2L))
})

test_that("select() before semi_join is inlined", {
  lf <- lazy_frame(x1 = 10, a = 1, y = 3, .name = "lf1")
  lf2 <- lazy_frame(x2 = 10, b = 2, z = 4, .name = "lf2")

  out_semi <- semi_join(
    lf |> select(a2 = a, x = x1),
    lf2 |> select(x = x2, b),
    by = "x"
  )
  lq <- out_semi$lazy_query
  expect_equal(op_vars(out_semi), c("a2", "x"))
  expect_equal(lq$vars, tibble(name = c("a2", "x"), var = c("a", "x1")))
  expect_equal(lq$by$x, "x1")
  expect_snapshot(out_semi)

  out_anti <- anti_join(
    lf |> select(a2 = a, x = x1),
    lf2 |> select(x = x2, b),
    by = "x"
  )
  lq <- out_anti$lazy_query
  expect_equal(lq$vars, tibble(name = c("a2", "x"), var = c("a", "x1")))
  expect_equal(lq$by$x, "x1")

  # attributes like `group`, `sort`, `frame` is kept
  lf <- lazy_frame(x = 10, a = 1, b = 1, .name = "lf1")
  lf2 <- lazy_frame(x = 10, .name = "lf2")
  out_semi <- semi_join(
    lf |>
      group_by(a) |>
      arrange(a) |>
      window_frame(0, 1) |>
      select(x, a),
    lf2,
    by = "x"
  )
  expect_equal(op_grps(out_semi), "a")
  expect_equal(op_sort(out_semi), list(expr(a)))
  expect_equal(op_frame(out_semi), list(range = c(0, 1)))
})

test_that("can combine full_join with other joins #1178", {
  lf1 <- lazy_frame(x = 1)
  lf2 <- lazy_frame(x = 1, y = 1)
  lf3 <- lazy_frame(x = 1, z = 1)

  # left join after full join
  expect_snapshot(
    full_join(lf1, lf2, by = "x") |>
      left_join(lf3, by = "x")
  )
  # full join after left join
  expect_snapshot(
    left_join(lf1, lf2, by = "x") |>
      full_join(lf3, by = "x")
  )
  # full join after full join
  expect_snapshot(
    full_join(lf1, lf2, by = "x") |>
      full_join(lf3, by = "x")
  )
})

test_that("select() before join is not inlined when using `sql_on`", {
  lf <- lazy_frame(x1 = 10, a = 1, y = 3, .name = "lf1")
  lf2 <- lazy_frame(x2 = 10, b = 2, z = 4, .name = "lf2")

  out <- semi_join(
    lf |> select(a2 = a, x = x1),
    lf2 |> select(x = x2, b),
    sql_on = sql("LHS.x = RHS.x")
  )

  lq <- out$lazy_query
  expect_s3_class(lq$x, "lazy_select_query")
  expect_equal(lq$vars, tibble(name = c("a2", "x"), var = c("a2", "x")))
})

test_that("filter() before semi join is inlined", {
  lf <- lazy_frame(x = 1, y = 2, z = 3)
  lf2 <- lazy_frame(x2 = 1, a = 2, b = 3, c = 4)

  out <- semi_join(
    lf,
    lf2 |>
      select(x = x2, a2 = a, b) |>
      filter(a2 == 1L, b == 2L),
    by = "x"
  )
  lq <- out$lazy_query
  expect_equal(
    lq$where,
    list(expr(a == 1L), expr(b == 2L))
  )
  expect_snapshot(out)
})

test_that("filter() before semi join is not when y has other operations", {
  lf <- lazy_frame(x = 1, y = 2, z = 3)
  lf2 <- lazy_frame(x = 1, a = 2, b = 3)

  out <- semi_join(
    lf,
    lf2 |>
      filter(a == 1L, b == 2L) |>
      mutate(a = a + 1),
    by = "x"
  )
  lq <- out$lazy_query
  expect_null(lq$where)

  out <- semi_join(
    lf,
    lf2 |>
      filter(a == 1L, b == 2L) |>
      distinct(),
    by = "x"
  )
  lq <- out$lazy_query
  expect_null(lq$where)

  out <- semi_join(
    lf,
    lf2 |>
      filter(a == 1L, b == 2L) |>
      head(10),
    by = "x"
  )
  lq <- out$lazy_query
  expect_null(lq$where)
})

test_that("filtered aggregates with subsequent select are not inlined away in semi_join (#1474)", {
  lf <- lazy_frame(x = 1, y = 2, z = 3)
  lf2 <- lazy_frame(x = 1, a = 2, b = 3)

  out <- semi_join(
    lf,
    lf2 |>
      dplyr::summarize(n = n(), .by = "x") |>
      filter(n == 1) |>
      select(x),
    by = "x"
  )
  lq <- out$lazy_query

  expect_equal(lq$y$having, list(expr(n() == 1)))
  expect_snapshot(out)
})

test_that("filtered window joins work in a semi_join", {
  df1 <- local_memdb_frame("df1", id = 1:5)
  df2 <- local_memdb_frame("df2", id = 1:5)

  df2_a <- df2 |> filter(row_number() <= 3)
  out <- anti_join(df1, df2_a, by = "id")
  expect_snapshot(show_query(out))

  expect_equal(collect(out), tibble(id = 4:5))
})

test_that("multiple joins create a single query", {
  lf <- lazy_frame(x = 1, a = 1, .name = "df1")
  lf2 <- lazy_frame(x = 1, b = 2, .name = "df2")
  lf3 <- lazy_frame(x = 1, b = 2, .name = "df3")

  out <- left_join(lf, lf2, by = "x") |>
    inner_join(lf3, by = "x")
  lq <- out$lazy_query
  expect_s3_class(lq, "lazy_multi_join_query")
  expect_equal(
    lq$table_names,
    tibble(
      name = table_path(c('"df1"', '"df2"', '"df3"')),
      from = "name"
    )
  )
  expect_equal(lq$vars$name, c("x", "a", "b.x", "b.y"))
  expect_equal(lq$vars$table, c(1L, 1L, 2L, 3L))
  expect_equal(lq$vars$var, c("x", "a", "b", "b"))

  expect_snapshot(out)
})

test_that("can join 4 tables with same column #1101", {
  # when there are
  # * at least 3 joins
  # * a variable appears in two table
  # * it is not joined by
  # * and it is renamed
  #
  # this produced a wrong query
  lf1 <- lazy_frame(x = 1, a = 2, .name = "lf1")
  lf2 <- lazy_frame(x = 2, b = 3, .name = "lf2")
  lf3 <- lazy_frame(x = 3, c = 4, .name = "lf3")
  lf4 <- lazy_frame(x = 4, a = 5, .name = "lf4") |>
    rename(a4 = a)

  out <- lf1 |>
    inner_join(lf2, by = "x") |>
    inner_join(lf3, by = "x") |>
    inner_join(lf4, by = "x")

  join_vars <- out$lazy_query$vars
  expect_equal(join_vars$name, c("x", "a", "b", "c", "a4"))
  expect_equal(join_vars$table, c(1L, 1L, 2L, 3L, 4L))
  expect_equal(join_vars$var, c("x", "a", "b", "c", "a"))
  # `lf4`.`a` AS `a4`
  expect_snapshot(remote_query(out))
})

test_that("multiple joins produce separate queries if using right/full join", {
  lf <- lazy_frame(x = 1, a = 1, .name = "df1")
  lf2 <- lazy_frame(x = 1, b = 2, .name = "df2")
  lf3 <- lazy_frame(x = 1, b = 2, .name = "df3")

  out <- left_join(lf, lf2, by = "x") |>
    right_join(lf3, by = "x")

  out_build <- out |> sql_build()
  expect_s3_class(out_build, "rf_join_query")
  expect_s3_class(out_build$x, "multi_join_query")

  expect_snapshot(remote_query(out))

  out_build <- left_join(lf, lf2, by = "x") |>
    full_join(lf3, by = "x") |>
    sql_build()

  expect_s3_class(out_build, "rf_join_query")
  expect_s3_class(out_build$x, "multi_join_query")
})

test_that("multiple joins can use by column from any table", {
  lf <- lazy_frame(x = 1, a = 1, .name = "df1")
  lf2 <- lazy_frame(x = 1, y = 2, .name = "df2")
  lf3 <- lazy_frame(x = 1, y = 2, b = 2, .name = "df3")

  out <- left_join(lf, lf2, by = "x") |>
    left_join(lf3, by = c("x", "y"))
  joins_metadata <- out$lazy_query$joins
  expect_equal(joins_metadata$by[[1]]$x, "x")
  expect_equal(joins_metadata$by[[2]]$x, c("x", "y"))
  expect_equal(joins_metadata$by[[1]]$y, "x")
  expect_equal(joins_metadata$by[[2]]$y, c("x", "y"))
  expect_equal(joins_metadata$by_x_table_id, list(1, c(1, 2)))

  lf <- lazy_frame(x = 1, a = 1, .name = "df1")
  lf2 <- lazy_frame(x = 1, y2 = 2, .name = "df2")
  lf3 <- lazy_frame(x = 1, y = 2, b = 2, .name = "df3")

  out <- left_join(
    lf,
    lf2 |> rename(y = y2),
    by = "x"
  ) |>
    left_join(lf3, by = c("x", "y"))

  joins_metadata <- out$lazy_query$joins
  expect_equal(joins_metadata$by[[1]]$x, "x")
  expect_equal(joins_metadata$by[[2]]$x, c("x", "y2"))
  expect_equal(joins_metadata$by[[1]]$y, "x")
  expect_equal(joins_metadata$by[[2]]$y, c("x", "y"))
  expect_equal(joins_metadata$by_x_table_id, list(1, c(1, 2)))
})

test_that("multi joins work with x_as", {
  lf <- lazy_frame(x = 1, a = 1, .name = "df1")
  lf2 <- lazy_frame(x = 1, b = 2, .name = "df2")
  lf3 <- lazy_frame(x = 1, b = 2, .name = "df3")

  out <- left_join(lf, lf2, by = "x", x_as = "lf1", y_as = "lf2") |>
    inner_join(lf3, by = "x", y_as = "lf3")
  lq <- out$lazy_query
  expect_s3_class(lq, "lazy_multi_join_query")
  expect_equal(
    lq$table_names,
    tibble(name = table_path(c('"lf1"', '"lf2"', '"lf3"')), from = "as")
  )

  # `x_as` provided twice with the same name -> one query
  out2 <- left_join(lf, lf2, by = "x", x_as = "lf1", y_as = "lf2") |>
    inner_join(lf3, by = "x", x_as = "lf1", y_as = "lf3")
  expect_equal(out, out2)

  # `x_as` provided twice with different names -> two queries
  out <- left_join(lf, lf2, by = "x", x_as = "lf1") |>
    inner_join(lf3, by = "x", x_as = "lf2")
  lq <- out$lazy_query
  expect_s3_class(lq, "lazy_multi_join_query")
  expect_s3_class(lq$x, "lazy_multi_join_query")

  # `x_as` name already used
  out <- left_join(lf, lf2, by = "x", y_as = "lf2") |>
    inner_join(lf3, by = "x", x_as = "lf2")
  lq <- out$lazy_query
  expect_s3_class(lq, "lazy_multi_join_query")
  expect_s3_class(lq$x, "lazy_multi_join_query")

  # `y_as` name already used
  out <- left_join(lf, lf2, by = "x", y_as = "lf2") |>
    inner_join(lf3, by = "x", y_as = "lf2")
  lq <- out$lazy_query
  expect_s3_class(lq, "lazy_multi_join_query")
  expect_s3_class(lq$x, "lazy_multi_join_query")

  out <- left_join(lf, lf2, by = "x", x_as = "lf2") |>
    inner_join(lf3, by = "x", y_as = "lf2")
  lq <- out$lazy_query
  expect_s3_class(lq, "lazy_multi_join_query")
  expect_s3_class(lq$x, "lazy_multi_join_query")
})

test_that("when keep = TRUE, left_join() preserves both sets of keys", {
  # when keys have different names
  df1 <- local_memdb_frame(a = c(2, 3), b = c(1, 2))
  df2 <- local_memdb_frame(x = c(3, 4), y = c(3, 4))
  out <- left_join(df1, df2, by = c("a" = "x"), keep = TRUE) |> collect()
  expect_equal(out$a, c(2, 3))
  expect_equal(out$x, c(NA, 3))

  # when keys have same name
  df3 <- local_memdb_frame(a = c(2, 3), b = c(1, 2))
  df4 <- local_memdb_frame(a = c(3, 4), y = c(3, 4))
  out <- left_join(df3, df4, by = c("a"), keep = TRUE) |> collect()
  expect_equal(out$a.x, c(2, 3))
  expect_equal(out$a.y, c(NA, 3))
})

test_that("when keep = TRUE, right_join() preserves both sets of keys", {
  # when keys have different names
  df1 <- local_memdb_frame(a = c(2, 3), b = c(1, 2))
  df2 <- local_memdb_frame(x = c(3, 4), y = c(3, 4))
  out <- right_join(df1, df2, by = c("a" = "x"), keep = TRUE) |> collect()
  expect_equal(out$a, c(3, NA))
  expect_equal(out$x, c(3, 4))

  # when keys have same name
  df3 <- local_memdb_frame(a = c(2, 3), b = c(1, 2))
  df4 <- local_memdb_frame(a = c(3, 4), y = c(3, 4))
  out <- right_join(df3, df4, by = c("a"), keep = TRUE) |> collect()
  expect_equal(out$a.x, c(3, NA))
  expect_equal(out$a.y, c(3, 4))
})

test_that("when keep = TRUE, full_join() preserves both sets of keys", {
  # when keys have different names
  df1 <- local_memdb_frame(a = c(2, 3), b = c(1, 2))
  df2 <- local_memdb_frame(x = c(3, 4), y = c(3, 4))
  out <- full_join(df1, df2, by = c("a" = "x"), keep = TRUE) |>
    collect() |>
    arrange(a)
  expect_equal(out$a, c(2, 3, NA))
  expect_equal(out$x, c(NA, 3, 4))

  # when keys have same name
  df3 <- local_memdb_frame(a = c(2, 3), b = c(1, 2))
  df4 <- local_memdb_frame(a = c(3, 4), y = c(3, 4))
  out <- full_join(df3, df4, by = c("a"), keep = TRUE) |>
    collect() |>
    arrange(a.x)
  expect_equal(out$a.x, c(2, 3, NA))
  expect_equal(out$a.y, c(NA, 3, 4))
})

test_that("when keep = TRUE, inner_join() preserves both sets of keys (#5581)", {
  # when keys have different names
  df1 <- local_memdb_frame(a = c(2, 3), b = c(1, 2))
  df2 <- local_memdb_frame(x = c(3, 4), y = c(3, 4))
  out <- inner_join(df1, df2, by = c("a" = "x"), keep = TRUE) |> collect()
  expect_equal(out$a, c(3))
  expect_equal(out$x, c(3))

  # when keys have same name
  df3 <- local_memdb_frame(a = c(2, 3), b = c(1, 2))
  df4 <- local_memdb_frame(a = c(3, 4), y = c(3, 4))
  out <- inner_join(df3, df4, by = c("a"), keep = TRUE) |> collect()
  expect_equal(out$a.x, c(3))
  expect_equal(out$a.y, c(3))
})

test_that("can't use `keep = FALSE` with non-equi conditions (#6499)", {
  join_by <- dplyr::join_by
  df1 <- local_memdb_frame(xl = c(1, 3), xu = c(4, 7))
  df2 <- local_memdb_frame(yl = c(2, 5, 8), yu = c(6, 8, 9))

  expect_snapshot(error = TRUE, {
    left_join(df1, df2, join_by(overlaps(xl, xu, yl, yu)), keep = FALSE)
  })

  # Would never make sense here.
  # Based on how the binary conditions are generated we'd merge:
  # - `yu` into `xl`
  # - `yl` into `xu`
  # Which results in `xl` and `xu` columns that don't maintain `xl <= xu`.
  expect_snapshot(error = TRUE, {
    full_join(df1, df2, join_by(overlaps(xl, xu, yl, yu)), keep = FALSE)
  })
})

test_that("by default, `by` columns omitted from `y` with equi-conditions, but not non-equi conditions", {
  # equi keys always keep the LHS name, regardless of whether of not a duplicate exists in the RHS
  # non-equi keys will get a suffix if a duplicate exists
  lf <- lazy_frame(x = 1, y = 1, z = 1)
  lf2 <- lazy_frame(x = 2, y = 1, z = 2)
  out <- right_join(
    lf,
    lf2,
    by = join_by(x == y, y > z),
    keep = NULL
  )
  vars <- out$lazy_query$vars
  expect_equal(vars$name, c("x", "y", "z.x", "x.y", "z.y"))
  expect_equal(vars$x, c(NA, "y", "z", NA, NA))
  expect_equal(vars$y, c("y", NA, NA, "x", "z"))

  # unless specifically requested with `keep = TRUE`
  lf <- lazy_frame(x = 1, y = 1, z = 1)
  lf2 <- lazy_frame(x = 2, y = 1, z = 2)
  out <- right_join(
    lf,
    lf2,
    by = join_by(x == y, y > z),
    keep = TRUE
  )
  vars <- out$lazy_query$vars
  expect_equal(vars$name, c("x.x", "y.x", "z.x", "x.y", "y.y", "z.y"))
  expect_equal(vars$x, c("x", "y", "z", NA, NA, NA))
  expect_equal(vars$y, c(NA, NA, NA, "x", "y", "z"))
})

test_that("can translate join conditions", {
  lf1 <- lazy_frame(a = 1, b = 1, c = 1)
  expect_snapshot({
    left_join(
      lf1,
      lf1,
      by = join_by(
        a == a,
        b >= b,
        c < c
      ),
      keep = TRUE
    )
  })
})

test_that("joins using `between(bounds =)` work as expected", {
  df1 <- local_memdb_frame(x = 1:5)
  df2 <- local_memdb_frame(lower = 2, upper = 4)

  out <- left_join(
    df1,
    df2,
    by = join_by(between(x, lower, upper, bounds = "[]"))
  ) |>
    collect()
  # out <- full_join(df1, df2, by = join_by(between(x, lower, upper, bounds = "[]")))
  expect_identical(out$lower, c(NA, 2, 2, 2, NA))
  expect_identical(out$upper, c(NA, 4, 4, 4, NA))

  out <- left_join(
    df1,
    df2,
    by = join_by(between(x, lower, upper, bounds = "[)"))
  ) |>
    collect()
  expect_identical(out$lower, c(NA, 2, 2, NA, NA))
  expect_identical(out$upper, c(NA, 4, 4, NA, NA))

  out <- left_join(
    df1,
    df2,
    by = join_by(between(x, lower, upper, bounds = "(]"))
  ) |>
    collect()
  expect_identical(out$lower, c(NA, NA, 2, 2, NA))
  expect_identical(out$upper, c(NA, NA, 4, 4, NA))

  out <- left_join(
    df1,
    df2,
    by = join_by(between(x, lower, upper, bounds = "()"))
  ) |>
    collect()
  expect_identical(out$lower, c(NA, NA, 2, NA, NA))
  expect_identical(out$upper, c(NA, NA, 4, NA, NA))
})

test_that("joins using `overlaps(bounds =)` work as expected", {
  df1 <- tibble(x_lower = c(1, 1, 3, 4), x_upper = c(2, 3, 4, 5))
  df2 <- tibble(y_lower = 2, y_upper = 4)

  expect_closed <- vctrs::vec_cbind(df1, vctrs::vec_c(df2, df2, df2, df2))
  mf1 <- local_memdb_frame(
    "df1",
    x_lower = c(1, 1, 3, 4),
    x_upper = c(2, 3, 4, 5)
  )
  mf2 <- local_memdb_frame(y_lower = 2, y_upper = 4)

  out <- left_join(
    mf1,
    mf2,
    by = join_by(overlaps(x_lower, x_upper, y_lower, y_upper, bounds = "[]"))
  ) |>
    collect()
  expect_identical(out, expect_closed)

  # `[)`, `(]`, and `()` all generate the same binary conditions but are useful
  # for consistency with `between(bounds =)`
  expect_open <- vctrs::vec_cbind(df1, vctrs::vec_c(NA, df2, df2, NA))

  out <- left_join(
    mf1,
    mf2,
    by = join_by(overlaps(x_lower, x_upper, y_lower, y_upper, bounds = "[)"))
  ) |>
    collect()
  expect_identical(out, expect_open)
  out <- left_join(
    mf1,
    mf2,
    by = join_by(overlaps(x_lower, x_upper, y_lower, y_upper, bounds = "(]"))
  ) |>
    collect()
  expect_identical(out, expect_open)
  out <- left_join(
    mf1,
    mf2,
    by = join_by(overlaps(x_lower, x_upper, y_lower, y_upper, bounds = "()"))
  ) |>
    collect()
  expect_identical(out, expect_open)
})

test_that("rolling joins aren't supported", {
  lf <- lazy_frame(x = 1, y = 1)

  expect_snapshot({
    (expect_error(left_join(lf, lf, join_by(closest(x >= y)))))
    (expect_error(semi_join(lf, lf, join_by(closest(x >= y)))))
  })
})

test_that("`na_matches` is validated", {
  df <- lazy_frame(x = 1)

  # Mutating joins
  expect_snapshot(error = TRUE, {
    left_join(df, df, by = "x", na_matches = 1)
  })
  expect_snapshot(error = TRUE, {
    left_join(df, df, by = "x", na_matches = "foo")
  })

  # Filtering joins
  expect_snapshot(error = TRUE, {
    semi_join(df, df, by = "x", na_matches = 1)
  })
  expect_snapshot(error = TRUE, {
    semi_join(df, df, by = "x", na_matches = "foo")
  })
})

test_that("using multiple gives an informative error", {
  lf <- lazy_frame(x = 1)

  expect_no_error(left_join(lf, lf, by = "x", multiple = NULL))
  expect_no_error(left_join(lf, lf, by = "x", multiple = "all"))

  expect_snapshot(error = TRUE, {
    left_join(lf, lf, by = "x", multiple = "first")
  })
})

test_that("using unmatched gives an informative error", {
  lf <- lazy_frame(x = 1)
  expect_no_error(left_join(lf, lf, by = "x", unmatched = "drop"))

  expect_snapshot(error = TRUE, {
    left_join(lf, lf, by = "x", unmatched = "error")
  })
})

test_that("using relationship gives an informative error", {
  lf <- lazy_frame(x = 1)
  expect_no_error(left_join(lf, lf, by = "x", relationship = NULL))
  expect_no_error(left_join(lf, lf, by = "x", relationship = "many-to-many"))

  expect_snapshot(error = TRUE, {
    left_join(lf, lf, by = "x", relationship = "one-to-one")
  })
})

# sql_build ---------------------------------------------------------------

test_that("join verbs generate expected ops", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 2)

  ji <- inner_join(lf1, lf2, by = "x")
  expect_s3_class(ji$lazy_query, "lazy_multi_join_query")
  expect_equal(ji$lazy_query$joins$type, "inner")

  jl <- left_join(lf1, lf2, by = "x")
  expect_s3_class(jl$lazy_query, "lazy_multi_join_query")
  expect_equal(jl$lazy_query$joins$type, "left")

  jr <- right_join(lf1, lf2, by = "x")
  expect_s3_class(jr$lazy_query, "lazy_rf_join_query")
  expect_equal(jr$lazy_query$type, "right")

  jf <- full_join(lf1, lf2, by = "x")
  expect_s3_class(jf$lazy_query, "lazy_rf_join_query")
  expect_equal(jf$lazy_query$type, "full")

  js <- semi_join(lf1, lf2, by = "x")
  expect_s3_class(js$lazy_query, "lazy_semi_join_query")
  expect_equal(js$lazy_query$anti, FALSE)

  ja <- anti_join(lf1, lf2, by = "x")
  expect_s3_class(ja$lazy_query, "lazy_semi_join_query")
  expect_equal(ja$lazy_query$anti, TRUE)
})

test_that("can optionally match NA values", {
  con <- dialect_postgres()
  lf1 <- lazy_frame(x = 1, .name = "lf1", con = con)
  lf2 <- lazy_frame(x = 1, .name = "lf2", con = con)
  expect_snapshot(left_join(lf1, lf2, by = "x", na_matches = "na"))
  expect_snapshot(semi_join(lf1, lf2, by = "x", na_matches = "na"))

  # even with non-equi conditions #1505
  lf3 <- lazy_frame(lower = 1, upper = 2, .name = "lf3")
  by <- join_by(between(x, lower, upper))
  expect_snapshot(left_join(lf1, lf3, by = by, na_matches = "na"))
})

test_that("join captures both tables", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 2)

  out <- inner_join(lf1, lf2, by = "x") |> sql_build()

  expect_s3_class(out, "multi_join_query")
  expect_equal(out$joins$type, "inner")
})

test_that("semi join captures both tables", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 2)

  out <- semi_join(lf1, lf2, by = "x") |> sql_build()

  expect_equal(out$anti, FALSE)
})

test_that("extra args generates error", {
  skip_if(getRversion() < "4.0.0")
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 2)

  expect_error(
    inner_join(lf1, lf2, by = "x", never_used = "na"),
    "used"
  )
})

test_that("suffix arg is checked", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 2)

  expect_snapshot({
    (expect_error(inner_join(lf1, lf2, by = "x", suffix = "a")))
    (expect_error(inner_join(lf1, lf2, by = "x", suffix = 1L)))
  })
})

test_that("copy = TRUE works", {
  lf <- local_memdb_frame(x = 1, y = 2)
  df <- tibble(x = 1, z = 3)

  expect_equal(
    inner_join(lf, df, by = "x", copy = TRUE) |> collect(),
    tibble(x = 1, y = 2, z = 3)
  )
})

test_that("copy = 'inline' works", {
  df1 <- local_memdb_frame(x = 1:3, y = c("a", "b", "c"))
  df2 <- tibble(x = 1L)

  out <- semi_join(df1, df2, by = "x", copy = "inline")
  expect_equal(collect(out), collect(df1)[1, ])
})

test_that("left_join/inner_join uses *", {
  lf1 <- lazy_frame(a = 1, b = 2, c = 1)
  lf2 <- lazy_frame(a = 1, b = 2, z = 1)

  con <- dialect_ansi()
  out <- lf1 |>
    left_join(lf2, by = c("a", "b")) |>
    sql_build()

  expect_equal(out$select, sql('"df_LHS".*', '"z"'))

  out <- lf1 |>
    left_join(lf2, by = c("a", "b")) |>
    sql_build(sql_options = sql_options(use_star = FALSE))

  expect_equal(
    out$select,
    sql(
      '"df_LHS"."a" AS "a"',
      '"df_LHS"."b" AS "b"',
      '"c"',
      '"z"'
    )
  )

  # also works after relocate
  out <- lf1 |>
    left_join(lf2, by = c("a", "b")) |>
    relocate(z) |>
    sql_build()

  expect_equal(out$select, sql('"z"', '"df_LHS".*'))

  # does not use * if variable are missing
  out <- lf1 |>
    left_join(lf2, by = c("a", "b")) |>
    select(a, c) |>
    sql_build()

  expect_equal(out$select, sql('"df_LHS"."a" AS "a"', '"c"'))

  # does not use * if variable names changed
  lf1 <- lazy_frame(a = 1, b = 2)
  lf2 <- lazy_frame(a = 1, b = 2)
  out <- lf1 |>
    left_join(lf2, by = c("a")) |>
    sql_build()

  expect_equal(
    out$select,
    sql('"df_LHS"."a" AS "a"', '"df_LHS"."b" AS "b.x"', '"df_RHS"."b" AS "b.y"')
  )
})

test_that("right_join uses *", {
  lf1 <- lazy_frame(a = 1, b = 2, c = 1)
  lf2 <- lazy_frame(a = 1, b = 2, z = 1)

  con <- dialect_ansi()
  out <- lf1 |>
    right_join(lf2, by = c("a", "b")) |>
    sql_build()

  # cannot use * without relocate or select
  expect_equal(
    out$select,
    sql('"df_RHS"."a" AS "a"', '"df_RHS"."b" AS "b"', '"c"', '"z"')
  )

  # also works after relocate
  out <- lf1 |>
    right_join(lf2, by = c("a", "b")) |>
    select(a, b, z, c) |>
    sql_build()

  expect_equal(
    out$select,
    sql('"df_RHS".*', '"c"')
  )

  # does not use * if `use_star = FALSE`
  out <- lf1 |>
    right_join(lf2, by = c("a", "b")) |>
    select(a, b, z, c) |>
    sql_build(sql_options = sql_options(use_star = FALSE))

  expect_equal(
    out$select,
    sql(
      '"df_RHS"."a" AS "a"',
      '"df_RHS"."b" AS "b"',
      '"z"',
      '"c"'
    )
  )

  # does not use * if variable are missing
  out <- lf1 |>
    right_join(lf2, by = c("a", "b")) |>
    select(a, z) |>
    sql_build()

  expect_equal(
    out$select,
    sql('"df_RHS"."a" AS "a"', '"z"')
  )

  # does not use * if variable names changed
  lf1 <- lazy_frame(a = 1, b = 2)
  lf2 <- lazy_frame(a = 1, b = 2)
  out <- lf1 |>
    right_join(lf2, by = c("a")) |>
    sql_build()

  expect_equal(
    out$select,
    sql('"df_RHS"."a" AS "a"', '"df_LHS"."b" AS "b.x"', '"df_RHS"."b" AS "b.y"')
  )
})

test_that("cross_join() errors if `...` is not empty (#1792)", {
  lf1 <- lazy_frame(a = 1)
  lf2 <- lazy_frame(b = 1)
  lf3 <- lazy_frame(c = 1)

  expect_snapshot(cross_join(lf1, lf2, lf3), error = TRUE)
})

test_that("cross_join uses *", {
  lf1 <- lazy_frame(a = 1, b = 1)
  lf2 <- lazy_frame(x = 1, y = 1)

  con <- dialect_ansi()
  out <- lf1 |>
    cross_join(lf2) |>
    sql_build()

  expect_equal(
    out$select,
    sql('"df_LHS".*', '"df_RHS".*')
  )

  # does not use * if `use_star = FALSE`
  out <- lf1 |>
    cross_join(lf2) |>
    sql_build(sql_options = sql_options(use_star = FALSE))

  expect_equal(
    out$select,
    sql('"a"', '"b"', '"x"', '"y"')
  )

  # also works after relocate
  out <- lf1 |>
    cross_join(lf2) |>
    select(x, y, a, b) |>
    sql_build()

  expect_equal(
    out$select,
    sql('"df_RHS".*', '"df_LHS".*')
  )

  out <- lf1 |>
    cross_join(lf2) |>
    select(x, a, b, y) |>
    sql_build()

  expect_equal(out$select, sql('"x"', '"df_LHS".*', '"y"'))

  out <- lf1 |>
    cross_join(lf2) |>
    select(a, x, y, b) |>
    sql_build()

  expect_equal(out$select, sql('"a"', '"df_RHS".*', '"b"'))
})

test_that("full_join() does not use *", {
  con <- dialect_ansi()
  lf1 <- lazy_frame(a = 1, b = 2)
  lf2 <- lazy_frame(a = 1, b = 2)

  out <- lf1 |>
    full_join(lf2, by = c("a", "b")) |>
    sql_build()

  expect_equal(
    out$select,
    sql(
      'COALESCE("df_LHS"."a", "df_RHS"."a") AS "a"',
      'COALESCE("df_LHS"."b", "df_RHS"."b") AS "b"'
    )
  )
})

test_that("joins reuse queries in cte mode", {
  lf1 <- lazy_frame(x = 1, .name = "lf1")
  lf <- lf1 |>
    inner_join(lf1, by = "x")

  expect_snapshot(
    left_join(lf, lf, by = "x") |>
      remote_query(sql_options = sql_options(cte = TRUE))
  )
})

test_that("joins correctly quote reused queries in CTEs", {
  lf <- lazy_frame(x = 1, .name = "lf") |> mutate(z = x * 2)
  expect_snapshot(
    left_join(lf, lf, by = join_by(z)) |>
      show_query(sql_options = sql_options(cte = TRUE))
  )
})

test_that("can force to qualify all columns", {
  lf1 <- lazy_frame(x = 1, a = 2, y = 1, .name = "lf1")
  lf2 <- lazy_frame(x = 1, a = 2, z = 1, .name = "lf2")

  unforced <- left_join(lf1, lf2, by = "x") |>
    sql_build(sql_options = sql_options(qualify_all_columns = FALSE))
  expect_equal(
    unforced$select,
    sql(
      '"lf1"."x" AS "x"',
      '"lf1"."a" AS "a.x"',
      '"y"',
      '"lf2"."a" AS "a.y"',
      '"z"'
    )
  )
  forced <- left_join(lf1, lf2, by = "x") |>
    sql_build(sql_options = sql_options(qualify_all_columns = TRUE))
  expect_equal(
    forced$select,
    sql(
      '"lf1"."x" AS "x"',
      '"lf1"."a" AS "a.x"',
      '"lf1"."y" AS "y"',
      '"lf2"."a" AS "a.y"',
      '"lf2"."z" AS "z"'
    )
  )
})

# ops ---------------------------------------------------------------------

test_that("joins get vars from both left and right", {
  out <- left_join(
    lazy_frame(x = 1, y = 1),
    lazy_frame(x = 2, z = 2),
    by = "x"
  )

  expect_equal(op_vars(out), c("x", "y", "z"))
})


test_that("semi joins get vars from left", {
  out <- semi_join(
    lazy_frame(x = 1, y = 1),
    lazy_frame(x = 2, z = 2),
    by = "x"
  )

  expect_equal(op_vars(out), c("x", "y"))
})

# Helpers -----------------------------------------------------------------

test_that("add_suffixes works if no suffix requested", {
  expect_equal(add_suffixes(c("x", "x"), "y", ""), c("x", "x"))
  expect_equal(add_suffixes(c("x", "y"), "y", ""), c("x", "y"))
})

test_that("expand completes all values", {
  expect_equal(
    memdb_frame(x = 1:2, y = 1:2) %>% tidyr::expand(x, y) %>% collect(),
    tibble(x = c(1, 1, 2, 2), y = c(1, 2, 1, 2))
  )

  expect_snapshot(lazy_frame(x = 1, y = 1) %>% tidyr::expand(x, y))
})

test_that("nesting doesn't expand values", {
  df <- tibble(x = 1:2, y = 1:2)
  expect_equal(
    tidyr::expand(memdb_frame(!!!df), nesting(x, y)) %>%
      collect(),
    df
  )

  df_lazy <- lazy_frame(!!!df)
  expect_snapshot(df_lazy %>% tidyr::expand(nesting(x, y)))
})

test_that("expand accepts expressions", {
  df <- lazy_frame(x = 1)
  expect_snapshot(tidyr::expand(df, round(x / 2)))
  expect_snapshot(tidyr::expand(df, nesting(x_half = round(x / 2), x1 = x + 1)))
})

test_that("works with tidyr::nesting", {
  df_lazy <- lazy_frame(x = 1:2, y = 1:2)
  expect_snapshot(df_lazy %>% tidyr::expand(tidyr::nesting(x, y)))
})

test_that("expand respects groups", {
  df <- tibble(
    a = c(1L, 1L, 2L),
    b = c(1L, 2L, 1L),
    c = c("b", "a", "a")
  )
  expect_equal(
    memdb_frame(!!!df) %>% group_by(a) %>% tidyr::expand(b, c) %>% collect(),
    tibble(
      a = c(1, 1, 1, 1, 2),
      b = c(1, 1, 2, 2, 1),
      c = c("b", "a", "b", "a", "a")
    ) %>%
      group_by(a)
  )

  df_lazy <- lazy_frame(!!!df)
  expect_snapshot(df_lazy %>% group_by(a) %>% tidyr::expand(b, c))
})

test_that("NULL inputs", {
  expect_snapshot(tidyr::expand(lazy_frame(x = 1), x, y = NULL))
})

test_that("expand() errors when expected", {
  expect_snapshot(error = TRUE, tidyr::expand(memdb_frame(x = 1)))
  expect_snapshot(error = TRUE, tidyr::expand(memdb_frame(x = 1), x = NULL))
})

test_that("nesting() respects .name_repair", {
  expect_snapshot(error = TRUE, 
    tidyr::expand(
      memdb_frame(x = 1, y = 1),
      nesting(x, x = x + 1)
    )
  )

  vars <- suppressMessages(
    tidyr::expand(
      memdb_frame(x = 1, y = 1),
      nesting(x, x = x + 1, .name_repair = "unique")
    ) %>%
      op_vars()
  )

  expect_equal(vars, c("x...1", "x...2"))
})

test_that("expand respect .name_repair", {
  vars <- suppressMessages(
    memdb_frame(x = integer(), z = integer()) %>%
      tidyr::expand(x, z = x, nesting(x), nesting(z), .name_repair = "unique") %>%
      op_vars()
  )

  expect_equal(vars, c("x...1", "z...2", "x...3", "z...4"))
})

# replace_na --------------------------------------------------------------

test_that("replace_na replaces missing values", {
  expect_equal(
    memdb_frame(x = c(1, NA), y = c(NA, "b")) %>%
      tidyr::replace_na(list(x = 0, y = "unknown")) %>%
      collect(),
    tibble(
      x = c(1, 0),
      y = c("unknown", "b")
    )
  )

  expect_snapshot(
    lazy_frame(x = 1, y = "a") %>% tidyr::replace_na(list(x = 0, y = "unknown"))
  )
})

test_that("replace_na ignores missing columns", {
  expect_snapshot(lazy_frame(x = 1) %>% tidyr::replace_na(list(not_there = 0)))
})

# complete ----------------------------------------------------------------

test_that("complete completes missing combinations", {
  df <- tibble(
    x = 1:2,
    y = 1:2,
    z = c("a", "b")
  )

  expect_equal(
    memdb_frame(!!!df) %>% tidyr::complete(x, y, fill = list(z = "c")) %>% collect(),
    tibble(
      x = c(1, 1, 2, 2),
      y = c(1, 2, 1, 2),
      z = c("a", "c", "c", "b")
    )
  )

  df_lazy <- lazy_frame(!!!df)
  expect_snapshot(df_lazy %>% tidyr::complete(x, y, fill = list(z = "c")))
})

test_that("reframe is not supported", {
  expect_snapshot(error = TRUE, {
    lazy_frame(x = 1) |> reframe()
  })
})

test_that("summarise peels off a single layer of grouping", {
  mf1 <- local_memdb_frame(x = 1, y = 1, z = 2) |> group_by(x, y)
  mf2 <- mf1 |> summarise(n = n())
  expect_equal(group_vars(mf2), "x")

  mf3 <- mf2 |> summarise(n = n())
  expect_equal(group_vars(mf3), character())
})

test_that("correctly inlines across all verbs", {
  lf <- lazy_frame(x = 1, y = 2)

  # single table verbs
  expect_selects(lf |> arrange(x) |> summarise(y = mean(x)), 1)
  expect_selects(lf |> distinct() |> summarise(y = mean(x)), 2)
  expect_selects(lf |> filter(x == 1) |> summarise(y = mean(x)), 1)
  expect_selects(lf |> head(1) |> summarise(y = mean(x)), 2)
  expect_selects(lf |> mutate(z = x + 1) |> summarise(y = mean(x)), 2)
  expect_selects(lf |> select(y = x) |> summarise(z = mean(y)), 2)
  expect_selects(lf |> summarise(y = mean(x)) |> summarise(z = mean(y)), 2)

  # two table verbs
  lf2 <- lazy_frame(x = 1)
  expect_selects(lf |> left_join(lf2, by = "x") |> summarise(y = mean(x)), 2)
  expect_selects(lf |> right_join(lf2, by = "x") |> summarise(y = mean(x)), 2)
  expect_selects(lf |> semi_join(lf2, by = "x") |> summarise(y = mean(x)), 3)

  #
  expect_selects(lf |> summarise(.by = x) |> summarise(n = n()), 2)
})

test_that("can performed grouped summarise with no inputs", {
  # used in n_groups()
  db <- local_memdb_frame("empty-summarise", x = c(1, 1, 1, 2, 2, 3))
  out <- db |> summarise(.by = x)
  expect_snapshot(out |> show_query())
  expect_equal(out |> collect(), tibble(x = c(1, 2, 3)))
})

test_that("generates minimal sql when possible", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_snapshot({
    lf |> arrange(x) |> summarise(y = mean(x))
    lf |> filter(x < 1) |> summarise(y = mean(x))
  })
})

test_that("summarise performs partial evaluation", {
  mf1 <- local_memdb_frame(x = 1)

  val <- 1
  mf2 <- mf1 |> summarise(y = x == val) |> collect()

  expect_equal(mf2$y, 1)
})

test_that("can't refer to freshly created variables", {
  lf1 <- lazy_frame(a = 1, b = 1)
  expect_snapshot({
    (expect_error(summarise(lf1, a_sum = sum(a), issue_col = sum(a_sum))))

    # works for variables created in `across()`
    # mentions `a_sum`
    (expect_error(summarise(
      lf1,
      across(c(a, b), list(sum = sum)),
      issue_col = sum(a_sum)
    )))
    # mentions `b_sum`
    (expect_error(summarise(
      lf1,
      across(c(a, b), list(sum = sum)),
      issue_col = sum(b_sum)
    )))

    # mentions `across()`
    (expect_error(summarise(
      lf1,
      a_sum = sum(a),
      issue_col = across(a_sum, sum)
    )))
  })
})

test_that("summarise(.groups=)", {
  df <- lazy_frame(x = 1, y = 2) |> group_by(x, y)

  expect_equal(df |> summarise() |> group_vars(), "x")
  expect_equal(df |> summarise(.groups = "drop_last") |> group_vars(), "x")
  expect_equal(df |> summarise(.groups = "drop") |> group_vars(), character())
  expect_equal(df |> summarise(.groups = "keep") |> group_vars(), c("x", "y"))

  expect_snapshot(error = TRUE, df |> summarise(.groups = "rowwise"))
})

test_that("summarise produces informative message about grouping", {
  local_mocked_bindings(summarise_verbose = function(...) TRUE)
  lf <- lazy_frame(x = 1, y = 2) |> group_by(x, y)
  expect_snapshot(. <- lf |> summarise())
})

test_that("summarise_verbose() detects caller environment correctly", {
  call <- quote(
    lazy_frame(x = 1, y = 2) |>
      dplyr::group_by(x, y) |>
      dplyr::summarise()
  )

  # should produce a message when called directly by user
  expect_message(eval_bare(call, env(global_env())))
  # should be silent when called in another package
  expect_no_message(eval_bare(call, asNamespace("testthat")))
})

test_that("summarise can modify grouping variables", {
  lf <- lazy_frame(g = 1, x = 1)

  expect_snapshot((result1 <- lf |> group_by(g) |> summarise(g = g + 1)))
  expect_equal(op_vars(result1), "g")
  expect_snapshot(
    (result2 <- lf |> group_by(g) |> summarise(x = x + 1, g = g + 1))
  )
  expect_equal(op_vars(result2), c("g", "x"))
})

test_that("across() does not select grouping variables", {
  df <- lazy_frame(g = 1, x = 1)

  # SELECT `g`, 0.0 AS `x`
  expect_snapshot(df |> group_by(g) |> summarise(across(.fns = ~0)))
})

test_that("summarise() after select() works #985", {
  df <- local_memdb_frame(g = 1, x = 1:3)
  expect_equal(
    df |>
      select(x) |>
      summarise(x = mean(x, na.rm = TRUE)) |>
      collect(),
    tibble(x = 2)
  )
})

# .by ----------------------------------------------------------------------

test_that("can group transiently using `.by`", {
  df <- local_memdb_frame(g = c(1, 1, 2, 1, 2), x = c(5, 2, 1, 2, 3))

  out <- summarise(df, x = mean(x, na.rm = TRUE), .by = g) |>
    arrange(g) |>
    collect()

  expect_identical(out$g, c(1, 2))
  expect_identical(out$x, c(3, 2))
  expect_equal(group_vars(out), character())
})

test_that("across doesn't select columns from `.by` #1493", {
  lf <- lazy_frame(g = 1, x = 1)

  out <- lf |>
    summarise(
      across(everything(), ~ sum(.x, na.rm = TRUE)),
      .by = g
    )

  expect_snapshot(out)
  expect_equal(sql_build(out)$select[1], sql("\"g\""))
})

test_that("can't use `.by` with `.groups`", {
  df <- lazy_frame(x = 1)

  expect_snapshot(error = TRUE, {
    summarise(df, .by = x, .groups = "drop")
  })
})

test_that("catches `.by` with grouped-df", {
  df <- lazy_frame(x = 1)
  gdf <- group_by(df, x)

  expect_snapshot(error = TRUE, {
    summarise(gdf, .by = x)
  })
})

# sql-render --------------------------------------------------------------

test_that("quoting for rendering summarized grouped table", {
  out <- local_memdb_frame("verb-summarise", x = 1) |>
    group_by(x) |>
    summarise(n = n())
  expect_snapshot(out |> sql_render())
  expect_equal(out |> collect(), tibble(x = 1, n = 1L))
})

# sql-build ---------------------------------------------------------------

test_that("summarise generates group_by and select", {
  out <- lazy_frame(g = 1) |>
    group_by(g) |>
    summarise(n = n()) |>
    sql_build()

  expect_equal(out$group_by, sql('"g"'))
  expect_equal(out$select, sql('"g"', 'COUNT(*) AS "n"'))
})


# ops ---------------------------------------------------------------------

test_that("summarise replaces existing", {
  out <- tibble(x = 1, y = 2) |> tbl_lazy() |> summarise(z = 1)
  expect_equal(op_vars(out), "z")
})

test_that("summarised vars are always named", {
  mf <- local_memdb_frame(a = 1)

  out1 <- mf |> summarise(1) |> op_vars()
  expect_equal(out1, "1")
})

test_that("grouped summary keeps groups", {
  out <- tibble(g = 1, x = 1) |>
    tbl_lazy() |>
    group_by(g) |>
    summarise(y = 1)
  expect_equal(op_vars(out), c("g", "y"))
})

test_that("summarise drops one grouping level", {
  df <- tibble(g1 = 1, g2 = 2, x = 3) |> tbl_lazy() |> group_by(g1, g2)
  out1 <- df |> summarise(y = 1)
  out2 <- out1 |> summarise(y = 2)

  expect_equal(op_grps(out1), "g1")
  expect_equal(op_grps(out2), character())
})


# lazy_select_query -------------------------------------------------------

test_that("can handle rename", {
  lf <- lazy_frame(x = 1:3, y = 3:1)

  out <- lf |>
    group_by(x) |>
    rename(ax = x, by = y) |>
    summarise(mean_by = mean(by, na.rm = TRUE))

  expect_equal(
    out$lazy_query,
    lazy_select_query(
      x = out$lazy_query$x,
      select = list(ax = sym("ax"), mean_by = expr(mean(by, na.rm = TRUE))),
      group_by = syms("ax"),
      select_operation = "summarise",
      group_vars = character()
    )
  )
})

#' Show generated SQL and query plan
#'
#' `show_sql()` displays the SQL query that will be dispatched to the database;
#' `explain()` displays both the SQL query and the query plan.
#'
#' @param ... For `explain()`, further arguments to [remote_query_plan()].
#'   For `show_query()`, ignored.
#' @param x An object to explain
#' @importFrom dplyr show_query
#' @export
#' @param use_colour Should the output be coloured?
#' @inheritParams remote_query
#' @name show_query
show_query.tbl_lazy <- function(
  x,
  ...,
  use_colour = TRUE,
  sql_options = NULL,
  cte = deprecated()
) {
  rlang::check_exclusive(sql_options, cte, .require = FALSE)
  if (lifecycle::is_present(cte)) {
    lifecycle::deprecate_warn(
      when = "2.4.0",
      what = "show_query(cte)",
      with = I("show_query(sql_options = sql_options(cte = TRUE))")
    )
    sql_options <- dbplyr::sql_options(cte = cte)
  }
  withr::local_options(list(dbplyr_use_colour = use_colour))
  sql <- remote_query(x, sql_options = sql_options)
  cat_line("<SQL>")
  cat_line(sql)
  invisible(x)
}

#' @importFrom dplyr explain
#' @export
#' @rdname show_query
explain.tbl_sql <- function(x, ...) {
  force(x)
  show_query(x)
  cat_line()
  cat_line("<PLAN>")
  cat_line(remote_query_plan(x, ...))

  invisible(x)
}

#' Create an SQL tbl (abstract)
#'
#' This function creates a lazy tbl object from a table in a database.
#' It's primarily intended for backend authors who need to create custom
#' subclasses; most users should use [tbl()] instead.
#'
#' @keywords internal
#' @export
#' @inheritParams tbl.src_dbi
#' @param subclass name of subclass
#' @param ... needed for agreement with generic. Not otherwise used.
#' @param check_from `r lifecycle::badge("deprecated")`
tbl_sql <- function(
  subclass,
  src,
  from,
  ...,
  vars = NULL,
  check_from = deprecated()
) {
  check_dots_empty()

  if (lifecycle::is_present(check_from)) {
    lifecycle::deprecate_stop("2.5.0", "tbl_sql(check_from)")
  }

  db_table(src$con, from, vars = vars, subclass = subclass)
}

find_variables <- function(con, from, call = caller_env()) {
  source <- as_table_source(from, con = con, error_call = call)

  withCallingHandlers(
    dbplyr_query_fields(con, source),
    error = function(err) {
      is_suspicious <- is_bare_string(from) && grepl(".", from, fixed = TRUE)
      if (!is_suspicious) {
        return()
      }

      cli::cli_abort(
        c(
          "Failed to find table {source}.",
          i = "Did you mean {.code from = I({.str {from}})}?"
        ),
        parent = err,
        call = call
      )
    }
  )
}

new_tbl_sql <- function(con, source, vars, subclass = NULL) {
  check_con(con)
  check_table_source(source)
  check_character(vars)

  lazy_query <- lazy_query_remote(source, vars)
  new_tbl_lazy(con, lazy_query, subclass = subclass)
}

# The goal of this function is to paper over all of the historical differences
# and provide a consistent forward looking interface, recognising that the
# primary different between the three cases (tbl(), lazy_frame(), and
# copy_inline()) is the base query
new_tbl_lazy <- function(con, query, subclass = NULL) {
  check_con(con)
  if (!inherits(query, "lazy_base_query")) {
    stop_input_type(query, "a <lazy_base_query>")
  }

  is_sql <- !inherits(query, "lazy_base_local_query")
  subclass <- c(
    subclass %||% class(con)[[1]],
    if (is_sql) c("dbi", "sql"),
    "lazy"
  )

  dplyr::make_tbl(
    subclass,
    con = con,
    src = src_dbi(con), # for backward compatibility
    lazy_query = query
  )
}

#' @importFrom dplyr same_src
#' @export
same_src.tbl_sql <- function(x, y) {
  inherits(y, "tbl_sql") && identical(x$con, y$con)
}

# Grouping methods -------------------------------------------------------------

#' @importFrom dplyr group_size
#' @export
group_size.tbl_sql <- function(x) {
  df <- x |>
    summarise(n = n()) |>
    collect()
  df$n
}

#' @importFrom dplyr n_groups
#' @export
n_groups.tbl_sql <- function(x) {
  if (length(groups(x)) == 0) {
    return(1L)
  }

  df <- x |>
    summarise() |>
    ungroup() |>
    summarise(n = n()) |>
    collect()
  df$n
}

# Standard data frame methods --------------------------------------------------

#' @export
print.tbl_sql <- function(x, ..., n = NULL, width = NULL, n_extra = NULL) {
  cat_line(format(x, ..., n = n, width = width, n_extra = n_extra))
  invisible(x)
}

#' @export
as.data.frame.tbl_sql <- function(
  x,
  row.names = NULL,
  optional = NULL,
  ...,
  n = Inf
) {
  as.data.frame(collect(x, n = n))
}

#' @importFrom pillar tbl_format_header
#' @export
tbl_format_header.tbl_sql <- function(x, setup, ...) {
  grps <- op_grps(x$lazy_query)
  sort <- op_sort(x$lazy_query)
  named_header <- c(
    # Can be overwritten by tbl_format_header.tbl_lazy:
    "A query" = paste0("?? x ", length(op_vars(x))),
    "Database" = db_connection_describe(x$con),
    "Groups" = if (length(grps) > 0) commas(grps),
    "Ordered by" = if (length(sort) > 0) commas(deparse_all(sort))
  )
  header <- paste0(
    pillar::align(paste0(names(named_header), ":")),
    " ",
    named_header
  )
  pillar::style_subtle(paste0("# ", header))
}

#' Create a local lazy tibble
#'
#' These functions are useful for testing SQL generation without having to
#' have an active database connection. See [simulate_dbi()] for a list
#' available database simulations.
#'
#' @keywords internal
#' @export
#' @examples
#' library(dplyr)
#' df <- data.frame(x = 1, y = 2)
#'
#' df_sqlite <- tbl_lazy(df, con = simulate_sqlite())
#' df_sqlite %>% summarise(x = sd(x, na.rm = TRUE)) %>% show_query()
tbl_lazy <- function(df, con = NULL, ..., name = "df") {
  check_dots_empty0(...)
  con <- con %||% sql_current_con() %||% simulate_dbi()
  subclass <- class(con)[[1]]

  name <- as_table_path(name, con)

  dplyr::make_tbl(
    purrr::compact(c(subclass, "lazy")),
    lazy_query = lazy_base_query(df, names(df), class = "local", name = name),
    src = src_dbi(con)
  )
}
methods::setOldClass(c("tbl_lazy", "tbl"))

#' @export
#' @rdname tbl_lazy
lazy_frame <- function(..., con = NULL, .name = "df") {
  con <- con %||% sql_current_con() %||% simulate_dbi()
  tbl_lazy(tibble(...), con = con, name = .name)
}

#' @export
dimnames.tbl_lazy <- function(x) {
  list(NULL, op_vars(x$lazy_query))
}

#' @export
dim.tbl_lazy <- function(x) {
  c(NA, length(op_vars(x$lazy_query)))
}

#' @export
print.tbl_lazy <- function(x, ...) {
  show_query(x)
}

#' @export
as.data.frame.tbl_lazy <- function(x, row.names, optional, ...) {
  cli_abort("Can not coerce {.cls tbl_lazy} to {.cls data.frame}")
}

#' @importFrom dplyr same_src
#' @export
same_src.tbl_lazy <- function(x, y) {
  inherits(y, "tbl_lazy")
}

#' @importFrom dplyr tbl_vars
#' @export
tbl_vars.tbl_lazy <- function(x) {
  op_vars(x$lazy_query)
}

#' @importFrom dplyr groups
#' @export
groups.tbl_lazy <- function(x) {
  lapply(group_vars(x), as.name)
}

# nocov start
#' @exportS3Method dplyr::group_by_drop_default
group_by_drop_default.tbl_lazy <- function(x) {
  TRUE
}
# nocov end

#' @importFrom dplyr group_vars
#' @export
group_vars.tbl_lazy <- function(x) {
  op_grps(x$lazy_query)
}

is_tbl_lazy <- function(x) {
  inherits(x, "tbl_lazy")
}

#' @importFrom tidyselect tidyselect_data_proxy tidyselect_data_has_predicates
#' @export
tidyselect_data_proxy.tbl_lazy <- function(x) {
  vars <- op_vars(x)
  out <- as_tibble(rep_named(vars, list(logical())), .name_repair = "minimal")
  group_by(out, !!!syms(group_vars(x)))
}

#' @export
tidyselect_data_has_predicates.tbl_lazy <- function(x) {
  FALSE
}

the <- new_environment()
the$warned_on_tbl_lazy_names <- FALSE

#' @export
names.tbl_lazy <- function(x) {
  should_inform <- rlang::env_is_user_facing(rlang::caller_env())
  if (should_inform) {
    cli::cli_inform(c(
      `!` = "The {.fn names} method of {.cls tbl_lazy} is for internal use only.",
      i = "Did you mean {.fn colnames}?"
    ))
  }
  NextMethod("names")
}

#' @export
`$.tbl_lazy` <- function(x, name) {
  unexpected_name <- !name %in% c("lazy_query", "src")

  if (unexpected_name) {
    cli::cli_abort(c(
      "The `$` method of {.cls tbl_lazy} is for internal use only.",
      i = "Use {.fn dplyr::pull} to get the values in a column."
    ))
  }
  NextMethod("$")
}

test_that("can pivot all cols to long", {
  pv <- memdb_frame(x = 1:2, y = 3:4) %>%
    tidyr::pivot_longer(x:y)

  expect_equal(
    pv %>% collect(),
    tibble(
      name = c("x", "x", "y", "y"),
      value = 1:4
    )
  )

  expect_snapshot(
    lazy_frame(x = 1:2, y = 3:4) %>%
      tidyr::pivot_longer(x:y)
  )
})

test_that("can add multiple columns from spec", {
  # add columns `a` and `b`
  sp <- tibble(
    .name = c("x", "y"),
    .value = "v",
    a = 11:12,
    b = 13:14
  )
  pv <- lazy_frame(x = 1:2, y = 3:4) %>%
    dbplyr_pivot_longer_spec(spec = sp)

  expect_equal(colnames(pv), c("a", "b", "v"))
  expect_snapshot(pv)
})

test_that("preserves original keys", {
  # preserves `x`
  pv <- lazy_frame(x = 1:2, y = 2, z = 1:2) %>%
    tidyr::pivot_longer(y:z)

  expect_equal(colnames(pv), c("x", "name", "value"))
  expect_snapshot(pv)
})

test_that("can drop missing values", {
  expect_snapshot(
    lazy_frame(x = c(1, NA), y = c(NA, 2)) %>%
      tidyr::pivot_longer(x:y, values_drop_na = TRUE)
  )
})

test_that("can handle missing combinations", {
  df <- tibble::tribble(
    ~id, ~x_1, ~x_2, ~y_2,
    "A",    1,    2,  "a",
    "B",    3,    4,  "b",
  )

  df_db <- memdb_frame(!!!df)

  pv_db <- tidyr::pivot_longer(
    df_db,
    -id,
    names_to = c(".value", "n"),
    names_sep = "_"
  )
  pv <- pv_db %>% collect()

  expect_named(pv, c("id", "n", "x", "y"))
  expect_equal(pv$x, c(1, 3, 2, 4))
  expect_equal(pv$y, c(NA, NA, "a", "b"))

  sql <- tidyr::pivot_longer(
    lazy_frame(!!!df),
    -id,
    names_to = c(".value", "n"),
    names_sep = "_"
  )
  expect_snapshot(sql)
})

test_that("can override default output column type", {
  expect_snapshot(
    lazy_frame(x = 1) %>%
      tidyr::pivot_longer(x, values_transform = list(value = as.character))
  )
})

test_that("values_transform can be a formula", {
  expect_snapshot(
    lazy_frame(x = 1) %>%
      tidyr::pivot_longer(x, values_transform = list(value = ~ as.character(.x)))
  )
})

test_that("`values_transform` works with single functions (#1284)", {
  df <- memdb_frame(x_1 = 1L, y_1 = 2L)

  res <- tidyr::pivot_longer(
    data = df,
    cols = everything(),
    names_to = c(".value", "set"),
    names_sep = "_",
    values_transform = as.character
  ) %>%
    collect()

  expect_identical(res$x, "1")
  expect_identical(res$y, "2")
})

test_that("`names_ptypes` and `names_transform`", {
  df <- memdb_frame(`1x2` = 1)

  res <- tidyr::pivot_longer(
    data = df,
    cols = `1x2`,
    names_to = c("one", "two"),
    names_sep = "x",
    names_transform = as.numeric
  ) %>%
    collect()

  expect_identical(res$one, 1)
  expect_identical(res$two, 2)

  res <- tidyr::pivot_longer(
    data = df,
    cols = `1x2`,
    names_to = c("one", "two"),
    names_sep = "x",
    names_transform = as.numeric,
    names_ptypes = integer()
  ) %>%
    collect()

  expect_identical(res$one, 1L)
  expect_identical(res$two, 2L)
})

test_that("`values_transform` is validated", {
  df <- memdb_frame(x = 1)

  expect_snapshot({
    (expect_error(tidyr::pivot_longer(df, x, values_transform = 1)))
    (expect_error(tidyr::pivot_longer(df, x, values_transform = list(~.x))))
  })
})

test_that("can pivot to multiple measure cols", {
  df <- lazy_frame(x = "x", y = 1)
  sp <- tibble::tribble(
    ~.name, ~.value, ~row,
    "x", "X", 1,
    "y", "Y", 1,
  )

  pv <- dbplyr_pivot_longer_spec(df, sp)
  expect_equal(colnames(pv), c("row", "X", "Y"))
  expect_snapshot(pv)
})

test_that("original col order is preserved", {
  df <- tibble::tribble(
    ~id, ~z_1, ~y_1, ~x_1, ~z_2,  ~y_2, ~x_2,
    "A",    1,    2,    3,     4,    5,    6,
    "B",    7,    8,    9,    10,   11,   12,
  ) %>%
    tbl_lazy()

  expect_equal(
    df %>%
      tidyr::pivot_longer(-id, names_to = c(".value", "n"), names_sep = "_") %>%
      colnames(),
    c("id", "n", "z", "y", "x")
  )
})

test_that(".value can be at any position in `names_to`", {
  samp <- tibble(
    i = 1:4,
    y_t1 = rnorm(4),
    y_t2 = rnorm(4),
    z_t1 = rep(3, 4),
    z_t2 = rep(-2, 4),
  )

  value_first <- tidyr::pivot_longer(
    lazy_frame(!!!samp), -i,
    names_to = c(".value", "time"), names_sep = "_"
  )

  expect_snapshot(value_first)

  samp2 <- dplyr::rename(samp, t1_y = y_t1,
                               t2_y = y_t2,
                               t1_z = z_t1,
                               t2_z = z_t2)

  value_second <- tidyr::pivot_longer(
    lazy_frame(!!!samp2), -i,
    names_to = c("time", ".value"), names_sep = "_"
  )

  expect_snapshot(value_second)

  cols <- c("i", "time", "y", "z")
  expect_equal(colnames(value_first), cols)
  expect_equal(colnames(value_second), cols)
})

test_that("grouping is preserved", {
  df <- memdb_frame(g = 1, x1 = 1, x2 = 2)
  out <- df %>%
    group_by(g) %>%
    tidyr::pivot_longer(x1:x2, names_to = "x", values_to = "v")
  expect_equal(group_vars(out), "g")
})

test_that("can pivot column with name equal to `names_to`", {
  df <- memdb_frame(id = 1:2, name2 = c("x", "y"))
  expect_equal(
    df %>%
      tidyr::pivot_longer(name2, names_to = "name2") %>%
      collect(),
    tibble(
      id = 1:2,
      name2 = "name2",
      value = c("x", "y")
    )
  )
})

test_that("can repair names", {
  df <- memdb_frame(id = 1, x = "a", y = "r", name = "nm", value = "val")

  expect_snapshot(out <- df %>% tidyr::pivot_longer(c(x, y), names_repair = "unique"))
  expect_equal(colnames(out), c("id", "name...2", "value...3", "name...4", "value...5"))

  expect_equal(
    collect(out),
    tibble(
      id = 1,
      name...2 = "nm",
      value...3 = "val",
      name...4 = c("x", "y"),
      value...5 = c("a", "r")
    )
  )
})

test_that("values_ptype is not supported", {
  expect_snapshot(
    error = TRUE,
    lazy_frame(x = 1:2, y = 3:4) %>%
      tidyr::pivot_longer(x:y, values_ptypes = character())
  )
})

test_that("mutate computed before summarise", {
  mf <- memdb_frame(x = c(1, 2, 3), y = c(9, 8, 7))

  out <- mutate(mf, z = x + y) %>%
    summarise(sum_z = sum(z, na.rm = TRUE)) %>%
    collect()

  expect_equal(out$sum_z, 30)
})

test_that("two mutates equivalent to one", {
  mf <- memdb_frame(x = c(1, 5, 9), y = c(3, 12, 11))

  df1 <- mf %>% mutate(x2 = x * 2, y4 = y * 4) %>% collect()
  df2 <- mf %>% collect() %>% mutate(x2 = x * 2, y4 = y * 4)
  compare_tbl(df1, df2)
})

test_that("can refer to fresly created values", {
  out1 <- memdb_frame(x1 = 1) %>%
    mutate(x2 = x1 + 1, x3 = x2 + 1, x4 = x3 + 1) %>%
    collect()
  expect_equal(out1, tibble(x1 = 1, x2 = 2, x3 = 3, x4 = 4))

  out2 <- copy_to_test("sqlite", tibble(x = 1), name = "multi_mutate") %>%
    mutate(x = x + 1, x = x + 2, x = x + 4)
  expect_equal(collect(out2), tibble(x = 8))
  expect_snapshot(show_query(out2))
})

test_that("transmute includes all needed variables", {
  lf <- lazy_frame(x = 1, y = 2)
  out <- transmute(lf, x = x / 2, x2 = x + y)
  expect_equal(op_vars(out$lazy_query$x), c("x", "y"))
  expect_snapshot(out)
})

test_that("queries are not nested unnecessarily", {
  # Should only be one query deep
  sql <- memdb_frame(x = 1) %>%
    mutate(y = x + 1, a = y + 1, b = y + 1) %>%
    sql_build()

  expect_s3_class(sql$from, "select_query")
  expect_s3_class(sql$from$from, "ident")
})

test_that("maintains order of existing columns (#3216, #3223)", {
  lazy <- lazy_frame(x = 1, y = 2) %>%
    mutate(z = 3, y = 4, y = 5)

  expect_equal(op_vars(lazy), c("x", "y", "z"))
})

test_that("supports overwriting variables (#3222)", {
  df <- memdb_frame(x = 1, y = 2) %>%
    mutate(y = 4, y = 5) %>%
    collect()
  expect_equal(df, tibble(x = 1, y = 5))

  df <- memdb_frame(x = 1, y = 2) %>%
    mutate(y = 4, y = y + 1) %>%
    collect()
  expect_equal(df, tibble(x = 1, y = 5))

  df <- memdb_frame(x = 1, y = 2) %>%
    mutate(y = 4, y = x + 4) %>%
    collect()
  expect_equal(df, tibble(x = 1, y = 5))
})

test_that("across() does not select grouping variables", {
  df <- lazy_frame(g = 1, x = 1)

  # SELECT `g`, 0.0 AS `x`
  expect_snapshot(df %>% group_by(g) %>% mutate(across(.fns = ~ 0)))
  expect_snapshot(df %>% group_by(g) %>% transmute(across(.fns = ~ 0)))
})

test_that("transmute() keeps grouping variables", {
  lf <- lazy_frame(g = 1, x = 1) %>% group_by(g)
  expect_equal(lf %>% transmute(x = 1) %>% op_vars(), c("g", "x"))
  expect_equal(lf %>% transmute(x = 1, g) %>% op_vars(), c("x", "g"))
})

test_that("across() can access previously created variables", {
  out <- memdb_frame(x = 1) %>% mutate(y = 2, across(y, sqrt))
  lf <- lazy_frame(x = 1) %>% mutate(y = 2, across(y, sqrt))

  expect_equal(
    collect(out),
    tibble(x = 1, y = sqrt(2))
  )

  expect_equal(
    op_vars(out),
    c("x", "y")
  )

  expect_snapshot(remote_query(lf))
})

test_that("new columns take precedence over global variables", {
  y <- "global var"
  db <- memdb_frame(data.frame(x = 1)) %>% mutate(y = 2, z = y + 1)
  lf <- lazy_frame(data.frame(x = 1)) %>% mutate(y = 2, z = y + 1)

  expect_equal(
    collect(db),
    tibble(x = 1, y = 2, z = 3)
  )

  expect_snapshot(remote_query(lf))
})

test_that("constants do not need a new query", {
  expect_equal(
    lazy_frame(x = 1, y = 2) %>% mutate(z = 2, z = 3) %>% remote_query(),
    sql("SELECT *, 3.0 AS `z`\nFROM `df`")
  )
})

test_that("mutate() produces nice error messages", {
  options(lifecycle_verbosity = "quiet")
  expect_snapshot(error = TRUE, {
    lazy_frame(x = 1) %>% mutate(z = non_existent + 1)

    # `...` cannot be evaluated
    lazy_frame(x = 1) %>% mutate(across(x, mean, na.rm = z))

    # `.fns` cannot be evaluated
    lazy_frame(x = 1) %>% mutate(across(x, .fns = "a"))
  })
})

test_that("empty mutate returns input", {
  df <- lazy_frame(x = 1)
  gf <- group_by(df, x)

  expect_equal(mutate(df), df)
  expect_equal(mutate(df, .by = x), df)
  expect_equal(mutate(gf), gf)

  expect_equal(mutate(df, !!!list()), df)
  expect_equal(mutate(df, !!!list(), .by = x), df)
  expect_equal(mutate(gf, !!!list()), gf)
})

# .by -------------------------------------------------------------------------

test_that("can group transiently using `.by`", {
  df <- memdb_frame(g = c(1, 1, 2, 1, 2), x = c(5, 2, 1, 2, 3))

  out <- mutate(df, x = mean(x), .by = g) %>%
    arrange(g) %>%
    collect()

  expect_identical(out$g, c(1, 1, 1, 2, 2))
  expect_identical(out$x, c(3, 3, 3, 2, 2))
  expect_equal(group_vars(out), character())
})

test_that("can `NULL` out the `.by` column", {
  df <- lazy_frame(x = 1:3, y = 1:3)
  out <- mutate(df, x = NULL, .by = x)

  expect_identical(op_vars(out), "y")
  expect_identical(remote_query(out), sql("SELECT `y`\nFROM `df`"))
})

# SQL generation -----------------------------------------------------------

test_that("mutate generates new variables and replaces existing", {
  df1 <- memdb_frame(x = 1)
  out <- df1 %>% mutate(x = 2, y = 3) %>% collect()
  expect_equal(out, tibble(x = 2, y = 3))
})

test_that("transmute only returns new variables", {
  df1 <- memdb_frame(x = 1)
  out <- df1 %>% transmute(y = 3) %>% collect()
  expect_equal(out, tibble(y = 3))
})

test_that("mutate calls windowed versions of sql functions", {
  df1 <- memdb_frame(x = 1:4, g = rep(c(1, 2), each = 2))
  out <- df1 %>% group_by(g) %>% mutate(r = row_number(x)) %>% collect()
  expect_equal(out$r, c(1, 2, 1, 2))
})

test_that("recycled aggregates generate window function", {
  df1 <- memdb_frame(x = as.numeric(1:4), g = rep(c(1, 2), each = 2))
  out <- df1 %>%
    group_by(g) %>%
    mutate(r = x - mean(x, na.rm = TRUE)) %>%
    collect()

  expect_equal(out$r, c(-0.5, 0.5, -0.5, 0.5))
})

test_that("cumulative aggregates generate window function", {
  df1 <- memdb_frame(x = 1:4, g = rep(c(1, 2), each = 2))
  out <- df1 %>%
    group_by(g) %>%
    arrange(x) %>%
    mutate(r = cumsum(x)) %>%
    collect()

  expect_equal(out$r, c(1, 3, 3, 7))
})

test_that("mutate overwrites previous variables", {
  df <- memdb_frame(x = 1:5) %>%
    mutate(x = x + 1) %>%
    mutate(x = x + 1) %>%
    collect()

  expect_equal(names(df), "x")
  expect_equal(df$x, 1:5 + 2)
})

test_that("sequence of operations work", {
  out <- memdb_frame(x = c(1, 2, 3, 4)) %>%
    select(y = x) %>%
    mutate(z = 2 * y) %>%
    filter(z == 2) %>%
    collect()

  expect_equal(out, tibble(y = 1, z = 2))
})

# .before, .after, .keep ------------------------------------------------------

test_that(".keep = 'unused' keeps variables explicitly mentioned", {
  df <- lazy_frame(x = 1, y = 2)
  out <- mutate(df, x1 = x + 1, y = y, .keep = "unused")
  expect_equal(op_vars(out), c("y", "x1"))
})

test_that(".keep = 'used' not affected by across()", {
  df <- lazy_frame(x = 1, y = 2, z = 3, a = "a", b = "b", c = "c")

  # This must evaluate every column in order to figure out if should
  # be included in the set or not, but that shouldn't be counted for
  # the purposes of "used" variables
  out <- mutate(df, across(c("x", "y", "z"), identity), .keep = "unused")
  expect_equal(op_vars(out), op_vars(df))
})

test_that(".keep = 'used' keeps variables used in expressions", {
  df <- lazy_frame(a = 1, b = 2, c = 3, x = 1, y = 2)
  out <- mutate(df, xy = x + y, .keep = "used")
  expect_equal(op_vars(out), c("x", "y", "xy"))
})

test_that(".keep = 'none' only keeps grouping variables", {
  df <- lazy_frame(x = 1, y = 2)
  gf <- group_by(df, x)

  expect_equal(op_vars(mutate(df, z = 1, .keep = "none")), "z")
  expect_equal(op_vars(mutate(gf, z = 1, .keep = "none")), c("x", "z"))
})

test_that(".keep = 'none' retains original ordering (#5967)", {
  df <- lazy_frame(x = 1, y = 2)
  expect_equal(
    df %>% mutate(y = 1, x = 2, .keep = "none") %>% op_vars(),
    c("x", "y")
  )

  # even when grouped
  gf <- group_by(df, x)
  expect_equal(
    gf %>% mutate(y = 1, x = 2, .keep = "none") %>% op_vars(),
    c("x", "y")
  )
})

test_that("can use .before and .after to control column position", {
  df <- lazy_frame(x = 1, y = 2)
  expect_equal(mutate(df, z = 1) %>% op_vars(), c("x", "y", "z"))
  expect_equal(mutate(df, z = 1, .before = 1) %>% op_vars(), c("z", "x", "y"))
  expect_equal(mutate(df, z = 1, .after = 1) %>% op_vars(), c("x", "z", "y"))

  # but doesn't affect order of existing columns
  df <- lazy_frame(x = 1, y = 2)
  expect_equal(mutate(df, x = 1, .after = y) %>% op_vars(), c("x", "y"))
})

test_that(".keep and .before/.after interact correctly", {
  df <- lazy_frame(x = 1, y = 1, z = 1, a = 1, b = 2, c = 3) %>%
    group_by(a, b)

  expect_equal(mutate(df, d = 1, x = 2, .keep = "none") %>% op_vars(), c("x", "a", "b", "d"))
  expect_equal(mutate(df, d = 1, x = 2, .keep = "none", .before = "a") %>% op_vars(), c("x", "d", "a", "b"))
  expect_equal(mutate(df, d = 1, x = 2, .keep = "none", .after = "a") %>% op_vars(), c("x", "a", "d", "b"))
})

test_that("dropping column with `NULL` then readding it retains original location", {
  df <- lazy_frame(x = 1, y = 2, z = 3, a = 4)
  df <- group_by(df, z)

  expect_equal(
    mutate(df, y = NULL, y = 3, .keep = "all") %>% op_vars(),
    c("x", "y", "z", "a")
  )
  expect_equal(
    mutate(df, b = a, y = NULL, y = 3, .keep = "used") %>% op_vars(),
    c("y", "z", "a", "b")
  )
  expect_equal(
    mutate(df, b = a, y = NULL, y = 3, .keep = "unused") %>% op_vars(),
    c("x", "y", "z", "b")
  )

  # It isn't treated as a "new" column
  expect_equal(mutate(df, y = NULL, y = 3, .keep = "all", .before = x) %>% op_vars(), c("x", "y", "z", "a"))
})

test_that(".keep= always retains grouping variables (#5582)", {
  df <- lazy_frame(x = 1, y = 2, z = 3) %>% group_by(z)
  expect_equal(
    df %>% mutate(a = x + 1, .keep = "none") %>% op_vars(),
    c("z", "a")
  )
  expect_equal(
    df %>% mutate(a = x + 1, .keep = "all") %>% op_vars(),
    c("x", "y", "z", "a")
  )
  expect_equal(
    df %>% mutate(a = x + 1, .keep = "used") %>% op_vars(),
    c("x", "z", "a")
  )
  expect_equal(
    df %>% mutate(a = x + 1, .keep = "unused") %>% op_vars(),
    c("y", "z", "a")
  )
})


# sql_render --------------------------------------------------------------

test_that("quoting for rendering mutated grouped table", {
  out <- memdb_frame(x = 1, y = 2) %>% mutate(y = x)
  expect_match(out %>% sql_render, "^SELECT `x`, `x` AS `y`\nFROM `[^`]*`$")
  expect_equal(out %>% collect, tibble(x = 1, y = 1))
})

test_that("mutate generates subqueries as needed", {
  lf <- lazy_frame(x = 1, con = simulate_sqlite())

  expect_snapshot(lf %>% mutate(x = x + 1, x = x + 1))
  expect_snapshot(lf %>% mutate(x1 = x + 1, x2 = x1 + 1))
})

test_that("mutate collapses over nested select", {
  lf <- lazy_frame(g = 0, x = 1, y = 2)

  expect_snapshot(lf %>% select(x:y) %>% mutate(x = x * 2, y = y * 2))
  expect_snapshot(lf %>% select(y:x) %>% mutate(x = x * 2, y = y * 2))
})

test_that("mutate() uses star", {
  lf <- lazy_frame(x = 1, y = 1)

  expect_equal(
    lf %>% mutate(z = 1L) %>% remote_query(),
    sql("SELECT *, 1 AS `z`\nFROM `df`")
  )

  expect_equal(
    lf %>% mutate(a = 1L, .before = 1) %>% remote_query(),
    sql("SELECT 1 AS `a`, *\nFROM `df`")
  )

  expect_equal(
    lf %>% transmute(a = 1L, x, y, z = 2L) %>% remote_query(),
    sql("SELECT 1 AS `a`, *, 2 AS `z`\nFROM `df`")
  )
})

# sql_build ---------------------------------------------------------------

test_that("mutate generates simple expressions", {
  out <- lazy_frame(x = 1) %>%
    mutate(y = x + 1L) %>%
    sql_build()

  expect_equal(out$select, sql('*', y = '`x` + 1'))
})

test_that("mutate can drop variables with NULL", {
  out <- lazy_frame(x = 1, y = 1) %>%
    mutate(y = NULL)

  expect_named(sql_build(out)$select, "x")
  expect_equal(op_vars(out), "x")
})

test_that("var = NULL works when var is in original data", {
  lf <- lazy_frame(x = 1) %>% mutate(x = 2, z = x*2, x = NULL)
  expect_equal(sql_build(lf)$select, sql(z = "`x` * 2.0"))
  expect_equal(op_vars(lf), "z")
  expect_snapshot(remote_query(lf))
})

test_that("var = NULL when var is in final output", {
  lf <- lazy_frame(x = 1) %>% mutate(y = NULL, y = 3)
  expect_equal(sql_build(lf)$select, sql("*", y = "3.0"))
  expect_equal(op_vars(lf), c("x", "y"))
  expect_snapshot(remote_query(lf))
})

test_that("temp var with nested arguments", {
  lf <- lazy_frame(x = 1) %>% mutate(y = 2, z = y*2, y = NULL)
  expect_equal(sql_build(lf)$select, sql(x = "`x`", z = "`y` * 2.0"))
  expect_equal(op_vars(lf), c("x", "z"))
  expect_snapshot(remote_query(lf))
})

test_that("mutate_all generates correct sql", {
  out <- lazy_frame(x = 1, y = 1) %>%
    dplyr::mutate_all(~ . + 1L) %>%
    sql_build()

  expect_equal(out$select, sql(x = '`x` + 1', y = '`y` + 1'))

  out <- lazy_frame(x = 1) %>%
    dplyr::mutate_all(list(one = ~ . + 1L, two = ~ . + 2L)) %>%
    sql_build()
  expect_equal(out$select, sql('*', one = '`x` + 1', two = '`x` + 2'))
})

test_that("mutate_all scopes nested quosures correctly", {
  num <- 10L
  out <- lazy_frame(x = 1, y = 1) %>%
    dplyr::mutate_all(~ . + num) %>%
    sql_build()

  expect_equal(out$select, sql(x = '`x` + 10', y = '`y` + 10'))
})


# ops ---------------------------------------------------------------------

test_that("mutate adds new", {
  out <- tibble(x = 1) %>% tbl_lazy() %>% mutate(y = x + 1, z = y + 1)
  expect_equal(op_vars(out), c("x", "y", "z"))
})

test_that("mutated vars are always named", {
  mf <- dbplyr::memdb_frame(a = 1)

  out2 <- mf %>% mutate(1) %>% op_vars()
  expect_equal(out2, c("a", "1"))
})

test_that("group_by with .add = TRUE adds groups", {
  mf <- memdb_frame(x = 1:3, y = 1:3)
  gf1 <- mf %>% group_by(x, y)
  gf2 <- mf %>% group_by(x) %>% group_by(y, .add = TRUE)

  expect_equal(group_vars(gf1), c("x", "y"))
  expect_equal(group_vars(gf2), c("x", "y"))
})

test_that("warns about add argument ", {
  mf <- memdb_frame(x = 1:3, y = 1:3)
  expect_warning(
    gf <- mf %>% group_by(x) %>% group_by(y, add = TRUE),
    "deprecated"
  )
  expect_equal(group_vars(gf), c("x", "y"))
})

test_that("errors for .drop = FALSE", {
  expect_snapshot(
    error = TRUE,
    lazy_frame(x = 1:3, y = 1:3) %>% group_by(y, .drop = FALSE)
  )
})

test_that("informative errors for missing variables", {
  expect_snapshot({
    (expect_error(lazy_frame(x = 1:3) %>% group_by(y)))
  })
})

test_that("group_by() produces nice error messages", {
  lf <- lazy_frame(x = 1)

  expect_snapshot(error = TRUE, {
    lf %>% group_by(z = non_existent + 1)
    lf %>% group_by(across(non_existent))
  })
})

test_that("collect, collapse and compute preserve grouping", {
  g <- memdb_frame(x = 1:3, y = 1:3) %>% group_by(x, y)

  expect_equal(group_vars(compute(g)), c("x", "y"))
  expect_equal(group_vars(collapse(g)), c("x", "y"))
  expect_equal(group_vars(collect(g)), c("x", "y"))
})

test_that("joins preserve grouping", {
  g <- memdb_frame(x = 1:3, y = 1:3) %>% group_by(x)

  expect_equal(group_vars(inner_join(g, g, by = c("x", "y"))), "x")
  expect_equal(group_vars(left_join(g, g, by = c("x", "y"))), "x")
  expect_equal(group_vars(semi_join(g, g, by = c("x", "y"))), "x")
  expect_equal(group_vars(anti_join(g, g, by = c("x", "y"))), "x")
})

test_that("group_by can perform mutate", {
  mf <- memdb_frame(x = 3:1, y = 1:3)
  out <- mf %>%
    group_by(z = x + y) %>%
    summarise(n = n()) %>%
    collect()

  expect_equal(out, tibble(z = 4L, n = 3L))
})

test_that("group_by handles empty dots", {
  lf <- lazy_frame(x = 1) %>% group_by(x)

  expect_equal(lf %>% group_by() %>% group_vars(), character())
  expect_equal(lf %>% group_by(!!!list()) %>% group_vars(), character())
  expect_equal(lf %>% group_by(.add = TRUE) %>% group_vars(), c("x"))
})

test_that("only adds step if necessary", {
  lf <- lazy_frame(x = 1, y = 1)
  expect_equal(lf %>% group_by(), lf)

  expect_equal(lf %>% ungroup(), lf)
  expect_equal(lf %>% ungroup(x), lf)

  lf_grouped <- lf %>% group_by(x)
  lf_grouped2 <- lf_grouped %>% group_by(x)
  expect_equal(lf_grouped, lf_grouped2)
  expect_equal(lf_grouped %>% ungroup(y), lf_grouped)

  out <- lf_grouped %>% mutate(y = y - mean(y)) %>% group_by()
  expect_equal(group_vars(out), character())
})

test_that("group_by mutate is not optimised away", {
  lf <- lazy_frame(x = 1) %>% group_by(x)

  expect_equal(
    lf %>% group_by(x = x + 1) %>% remote_query(),
    sql("SELECT `x` + 1.0 AS `x`\nFROM `df`")
  )
})


# sql_build ---------------------------------------------------------------

test_that("ungroup drops PARTITION BY", {
  out <- lazy_frame(x = 1) %>%
    group_by(x) %>%
    ungroup() %>%
    mutate(x = sum(x)) %>%
    sql_build()
  expect_equal(out$select, sql(x = 'SUM(`x`) OVER ()'))
})

# ops ---------------------------------------------------------------------

test_that("group_by overrides existing groups", {
  df <- tibble(g1 = 1, g2 = 2, x = 3) %>% tbl_lazy()

  out1 <- df %>% group_by(g1)
  expect_equal(op_grps(out1), "g1")

  out2 <- out1 %>% group_by(g2)
  expect_equal(op_grps(out2), "g2")
})

test_that("group_by increases grouping if add = TRUE", {
  df <- tibble(g1 = 1, g2 = 2, x = 3) %>% tbl_lazy()

  out <- df %>% group_by(g1) %>% group_by(g2, .add = TRUE)
  expect_equal(op_grps(out), c("g1", "g2"))
})


test_that("ungroup drops all groups", {
  out1 <- lazy_frame(g1 = 1, g2 = 2) %>%
    group_by(g1, g2) %>%
    ungroup()

  out2 <- lazy_frame(g1 = 1, g2 = 2) %>%
    group_by(g1, g2) %>%
    ungroup() %>%
    rename(g3 = g1)

  expect_equal(op_grps(out1), character())
  expect_equal(op_grps(out2), character())
})

test_that("ungroup() produces nice error messages", {
  expect_snapshot(error = TRUE, {
    lazy_frame(x = 1) %>% ungroup(non_existent)
  })
})

test_that("bind_queries combines multiple lazy tables", {
  lf1 <- lazy_frame(x = 1, y = "a", .name = "lf1")
  lf2 <- lazy_frame(x = 2, y = "b", .name = "lf2")
  lf3 <- lazy_frame(x = 3, y = "c", .name = "lf3")

  expect_snapshot(bind_queries(lf1, lf2, lf3))
})

test_that("bind_queries works with splice operator", {
  lf1 <- lazy_frame(x = 1, .name = "lf1")
  lf2 <- lazy_frame(x = 2, .name = "lf2")
  queries <- list(lf1, lf2)

  expect_snapshot(bind_queries(!!!queries))
})

test_that("bind_queries checks its inputs", {
  lf <- lazy_frame(x = 1)

  expect_snapshot(error = TRUE, {
    bind_queries()
    bind_queries(lf, 1)
  })
})

test_that("adds src class", {
  tb <- tbl_lazy(mtcars, con = simulate_sqlite())
  expect_s3_class(tb, "tbl_SQLiteConnection")
})

test_that("argument src is deprecated", {
  expect_snapshot(dummy <- tbl_lazy(mtcars, src = simulate_sqlite()))
})

test_that("cannot convert tbl_lazy to data.frame", {
  expect_snapshot(
    error = TRUE,
    as.data.frame(tbl_lazy(mtcars, src = simulate_sqlite()))
  )
})

test_that("dim() works for tbl_lazy", {
  expect_equal(dim(tbl_lazy(mtcars)), c(NA, 11))
})

test_that("has print method", {
  expect_snapshot(tbl_lazy(mtcars))
})

test_that("support colwise variants", {
  mf <- memdb_frame(x = 1:5, y = factor(letters[1:5]))
  exp <- mf %>% collect() %>% mutate(y = as.character(y))

  expect_message(
    mf1 <- dplyr::mutate_if(mf, is.factor, as.character),
    "on the first 100 rows"
  )
  compare_tbl(mf1, exp)

  mf2 <- dplyr::mutate_at(mf, "y", as.character)
  compare_tbl(mf2, exp)
})

test_that("base source of tbl_lazy is always 'df'", {
  out <- lazy_frame(x = 1, y = 5) %>% sql_build()
  expect_equal(out, ident("df"))
})

test_that("head limits rows", {
  db <- head(memdb_frame(x = 1:100), 10)

  expect_equal(sql_build(db)$limit, 10)
  expect_equal(nrow(collect(db)), 10)
})

test_that("two heads are equivalent to one", {
  out <- lazy_frame(x = 1:10) %>% head(3) %>% head(5)
  expect_equal(out$lazy_query$limit, 3)
})

test_that("non-integer automatically truncated", {
  out <- lazy_frame(x = 1:10) %>% head(3.5)
  expect_equal(out$lazy_query$limit, 3)
})

test_that("can get 0 rows", {
  db <- memdb_frame(x = 1)
  out <- collect(head(db, 0))
  expect_equal(out, tibble(x = double()))
})

test_that("n must be valid", {
  db <- memdb_frame(x = 1)

  expect_error(head(db, "x"), "non-negative integer")
  expect_error(head(db, 1:2), "non-negative integer")
  expect_error(head(db, -1), "non-negative integer")
  expect_error(head(db, Inf), NA)
})

test_that("tail not supported", {
  lf <- lazy_frame(x = 1)
  expect_error(tail(lf), "not supported")
})

test_that("generates expected SQL for common situations", {
  db <- lazy_frame(g = 1, x = 2)

  expect_snapshot(db |> count(g))
  expect_snapshot(db |> count(g, wt = x))
  expect_snapshot(db |> count(g, sort = TRUE))

  expect_snapshot(db |> add_count(g, sort = TRUE))
  expect_snapshot(db |> group_by(g) |> add_count())
})

test_that("preserves group of input", {
  db <- lazy_frame(g = 1, x = 2)
  expect_equal(db |> count(g) |> group_vars(), character())
  expect_equal(db |> group_by(g) |> count() |> group_vars(), "g")
  expect_equal(db |> group_by(g, x) |> count() |> group_vars(), c("g", "x"))

  expect_equal(
    db |> group_by(g, x) |> add_count(g) |> group_vars(),
    c("g", "x")
  )
  expect_equal(db |> add_count(g) |> group_vars(), character())
})

test_that("count() does not create groups", {
  # https://github.com/tidyverse/dbplyr/issues/940
  db <- local_memdb_frame(
    x = c(1, 1, 1, 2, 2),
    y = c("a", "a", "b", "c", "c")
  )

  df <- db |>
    count(x, y) |>
    collect()

  expect <- tibble(
    x = c(1, 1, 2),
    y = c("a", "b", "c"),
    n = c(2L, 1L, 2L)
  )

  expect_equal(group_vars(df), character())
  expect_identical(df, expect)
})

test_that("informs if n column already present, unless overridden", {
  df1 <- lazy_frame(n = c(1, 1, 2, 2, 2))
  expect_message(out <- count(df1, n), "already present")
  expect_equal(colnames(out), c("n", "nn"))

  # not a good idea, but supported
  expect_message(out <- count(df1, n, name = "n"), NA)
  expect_equal(colnames(out), "n")

  expect_message(out <- count(df1, n, name = "nn"), NA)
  expect_equal(colnames(out), c("n", "nn"))

  df2 <- lazy_frame(n = c(1, 1, 2, 2, 2), nn = 1:5)
  expect_message(out <- count(df2, n), "already present")
  expect_equal(colnames(out), c("n", "nn"))

  expect_message(out <- count(df2, n, nn), "already present")
  expect_equal(colnames(out), c("n", "nn", "nnn"))
})

test_that(".drop is not supported", {
  expect_snapshot(error = TRUE, {
    lazy_frame(g = 1) |>
      add_count(.drop = TRUE)
  })
})

test_that("generated sql doesn't change unexpectedly", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_snapshot(right_join(lf, lf))
  expect_snapshot(full_join(lf, lf))
})

test_that("disambiguate variables that only differ in case", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(X = 1, y = 2)
  expect_snapshot(full_join(lf1, lf2, by = "y"))
})

test_that("sql_on query doesn't change unexpectedly", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 3)

  expect_snapshot(right_join(lf1, lf2, sql_on = "LHS.y < RHS.z"))
  expect_snapshot(full_join(lf1, lf2, sql_on = "LHS.y < RHS.z"))
})

test_that("sql_rf_join_vars generates expected SQL", {
  con <- dialect_ansi()

  # full_join(lf(x, a), lf(x, b), by = "x")
  expect_equal(
    sql_rf_join_vars(
      con,
      type = "full",
      vars = list(
        name = c("x", "a.x", "a.y", "b"),
        x = c("x", "a", NA, NA),
        y = c("x", NA, "a", "b"),
        all_x = c("x", "a"),
        all_y = c("x", "a", "b")
      ),
      x_as = table_path("left"),
      y_as = table_path("right"),
      use_star = TRUE,
      qualify_all_columns = FALSE
    ),
    sql(
      'COALESCE("left"."x", "right"."x") AS "x"',
      '"left"."a" AS "a.x"',
      '"right"."a" AS "a.y"',
      '"b"'
    )
  )

  # disambiguate variables that only differ in case
  expect_equal(
    sql_rf_join_vars(
      con,
      type = "full",
      vars = list(
        name = c("a", "b.x", "b.y"),
        x = c("a", "b", NA),
        y = c("a", NA, "B"),
        all_x = c("a", "b"),
        all_y = c("a", "B")
      ),
      x_as = table_path("left"),
      y_as = table_path("right"),
      use_star = TRUE,
      qualify_all_columns = FALSE
    ),
    sql(
      'COALESCE("left"."a", "right"."a") AS "a"',
      '"left"."b" AS "b.x"',
      '"right"."B" AS "b.y"'
    )
  )
})

#' SQLite backend
#'
#' @description
#' This backend supports SQLite databases, typically accessed via
#' a `SQLiteConnection` created by [DBI::dbConnect()]. Use `dialect_sqlite()`
#' with `lazy_frame()` to see simulated SQL without connecting to a live
#' database.
#'
#' Key differences for this backend are:
#'
#' * Uses non-standard `LOG()` function
#' * Date-time extraction functions from lubridate
#' * Custom median translation
#' * Right and full joins are simulated using left joins
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-sqlite
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_sqlite())
#' lf |> transmute(x = paste(c, " times"))
#' lf |> transmute(x = log(b), y = log(b, base = 2))
NULL

#' @export
#' @rdname backend-sqlite
dialect_sqlite <- function() {
  new_sql_dialect(
    "sqlite",
    quote_identifier = function(x) sql_quote(x, "`"),
    has_window_clause = TRUE
  )
}

#' @export
#' @rdname backend-sqlite
simulate_sqlite <- function() simulate_dbi("SQLiteConnection")

#' @export
sql_dialect.SQLiteConnection <- function(con) {
  dialect_sqlite()
}

#' @export
dbplyr_edition.SQLiteConnection <- function(con) {
  2L
}

#' @export
db_connection_describe.SQLiteConnection <- function(con, ...) {
  paste0("sqlite ", sqlite_version(), " [", con@dbname, "]")
}

#' @export
db_col_types.SQLiteConnection <- function(con, table, call = caller_env()) {
  path <- as_table_path(table, con, error_call = call)
  parts <- table_path_components(path, con)[[1]]

  if (length(parts) == 1) {
    sql <- sql_glue2(
      con,
      "SELECT name, type FROM pragma_table_info({parts})"
    )
  } else {
    schema <- parts[[length(parts) - 1]]
    name <- parts[[length(parts)]]
    sql <- sql_glue2(
      con,
      "SELECT name, type FROM pragma_table_info({name}, {schema})"
    )
  }
  col_info_df <- DBI::dbGetQuery(con, sql)
  set_names(col_info_df[["type"]], col_info_df[["name"]])
}

#' @export
sql_query_explain.sql_dialect_sqlite <- function(con, sql, ...) {
  sql_glue2(con, "EXPLAIN QUERY PLAN {sql}")
}

#' @export
sql_query_set_op.sql_dialect_sqlite <- sql_query_set_op.sql_dialect_hive

#' @export
sql_query_upsert.sql_dialect_sqlite <- function(
  con,
  table,
  from,
  by,
  update_cols,
  ...,
  returning_cols = NULL,
  method = NULL
) {
  sql_query_upsert_on_conflict(
    con,
    table,
    from,
    by,
    update_cols,
    returning_cols = returning_cols
  )
}

sqlite_version <- function() {
  numeric_version(RSQLite::rsqliteVersion()[[2]])
}

# SQL methods -------------------------------------------------------------

#' @export
sql_translation.sql_dialect_sqlite <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_scalar,
      # https://sqlite.org/lang_expr.html#isisnot
      is_distinct_from = \(x, y) sql_glue("({x}) IS NOT ({y})"),
      is_not_distinct_from = \(x, y) sql_glue("({x}) IS ({y})"),
      as.numeric = sql_cast("REAL"),
      as.double = sql_cast("REAL"),
      log = function(x, base = exp(1)) {
        if (base != exp(1)) {
          sql_glue("LOG({x}) / LOG({base})")
        } else {
          sql_glue("LOG({x})")
        }
      },
      paste = sql_paste_infix(" ", "||"),
      paste0 = sql_paste_infix("", "||"),
      # https://www.sqlite.org/lang_corefunc.html#maxoreunc
      pmin = sql_aggregate_n("MIN", "pmin"),
      pmax = sql_aggregate_n("MAX", "pmax"),

      runif = function(n = n(), min = 0, max = 1) {
        # https://stackoverflow.com/a/23785593/7529482
        sql_runif(
          "(0.5 + RANDOM() / 18446744073709551616.0)",
          n = {{ n }},
          min = min,
          max = max
        )
      },

      # lubridate,
      today = function() {
        sql_glue("DATE('now')")
      },
      now = \() sql_glue("DATETIME('now')"),
      # https://modern-sql.com/feature/extract#proprietary-strftime
      year = \(x) sql_glue("CAST(STRFTIME('%Y', {x}) AS NUMERIC)"),
      month = \(x) sql_glue("CAST(STRFTIME('%m', {x}) AS NUMERIC)"),
      mday = \(x) sql_glue("CAST(STRFTIME('%d', {x}) AS NUMERIC)"),
      day = \(x) sql_glue("CAST(STRFTIME('%d', {x}) AS NUMERIC)"),
      hour = \(x) sql_glue("CAST(STRFTIME('%H', {x}) AS NUMERIC)"),
      minute = \(x) sql_glue("CAST(STRFTIME('%M', {x}) AS NUMERIC)"),
      second = \(x) sql_glue("CAST(STRFTIME('%f', {x}) AS REAL)"),
      yday = \(x) sql_glue("CAST(STRFTIME('%j', {x}) AS NUMERIC)"),
    ),
    sql_translator(
      .parent = base_agg,
      sd = sql_aggregate("STDEV", "sd"),
      median = sql_aggregate("MEDIAN"),
      quantile = sql_not_supported("quantile"),
    ),
    if (sqlite_version() >= "3.25") {
      sql_translator(
        .parent = base_win,
        sd = win_aggregate("STDEV"),
        median = win_absent("median"),
        quantile = sql_not_supported("quantile"),
      )
    } else {
      base_no_win # nocov
    }
  )
}

#' @export
sql_escape_logical.sql_dialect_sqlite <- function(con, x) {
  y <- as.character(as.integer(x))
  y[is.na(x)] <- "NULL"
  sql(y)
}

#' @export
values_prepare.sql_dialect_sqlite <- function(con, df) {
  needs_escape <- purrr::map_lgl(
    df,
    \(col) methods::is(col, "Date") || inherits(col, "POSIXct")
  )
  purrr::modify_if(
    df,
    needs_escape,
    \(col) escape(col, con = con, parens = FALSE, collapse = NULL)
  )
}

test_that("simulate_sqlite() still works", {
  expect_translation(simulate_sqlite(), x + 1, "`x` + 1.0")
})

test_that("logicals translated to integers", {
  expect_equal(escape(FALSE, con = dialect_sqlite()), sql("0"))
  expect_equal(escape(TRUE, con = dialect_sqlite()), sql("1"))
  expect_equal(escape(NA, con = dialect_sqlite()), sql("NULL"))
})

test_that("vectorised translations", {
  con <- dialect_sqlite()

  expect_translation(con, paste(x, y), '`x` || \' \' || `y`')
  expect_translation(con, paste0(x, y), '`x` || `y`')
})

test_that("pmin and max become MIN and MAX", {
  con <- dialect_sqlite()

  expect_translation(con, pmin(x, y, na.rm = TRUE), 'MIN(`x`, `y`)')
  expect_translation(con, pmax(x, y, na.rm = TRUE), 'MAX(`x`, `y`)')
})

test_that("sqlite mimics two argument log", {
  con <- dialect_sqlite()

  expect_translation(con, log(x), 'LOG(`x`)')
  expect_translation(con, log(x, 10), 'LOG(`x`) / LOG(10.0)')
})

test_that("date-time", {
  con <- dialect_sqlite()

  expect_translation(con, today(), "DATE('now')")
  expect_translation(con, now(), "DATETIME('now')")
})

test_that("custom aggregates translated", {
  con <- dialect_sqlite()

  expect_translation(
    con,
    median(x, na.rm = TRUE),
    'MEDIAN(`x`)',
    window = FALSE
  )
  expect_translation(con, sd(x, na.rm = TRUE), 'STDEV(`x`)', window = FALSE)
  expect_error(
    translate_sql(quantile(x, 0.5, na.rm = TRUE), con = con, window = FALSE),
    class = "dbplyr_error_unsupported_fn"
  )
  expect_error(
    translate_sql(quantile(x, 0.5, na.rm = TRUE), con = con, window = TRUE),
    class = "dbplyr_error_unsupported_fn"
  )
})

test_that("custom SQL translation", {
  con <- dialect_sqlite()

  lf <- lazy_frame(x = 1, con = dialect_sqlite())
  expect_snapshot(left_join(lf, lf, by = "x", na_matches = "na"))

  expect_snapshot(translate_sql(runif(n()), con = con))
})

test_that("case_when translates correctly to ELSE when TRUE ~ is used", {
  con <- dialect_sqlite()
  expect_snapshot(
    translate_sql(
      case_when(
        x == 1L ~ "yes",
        x == 0L ~ "no",
        TRUE ~ "undefined"
      ),
      con = con
    )
  )
})


# live database -----------------------------------------------------------

test_that("as.numeric()/as.double() get custom translation", {
  mf <- local_memdb_frame(x = 1L)

  out <- mf |> mutate(x1 = as.numeric(x), x2 = as.double(x)) |> collect()
  expect_type(out$x1, "double")
  expect_type(out$x2, "double")
})

test_that("date extraction agrees with R", {
  db <- local_memdb_frame(x = "2000-01-02 03:40:50.5")
  out <- db |>
    transmute(
      year = year(x),
      month = month(x),
      day = day(x),
      hour = hour(x),
      minute = minute(x),
      second = second(x),
      yday = yday(x),
    ) |>
    collect() |>
    as.list()

  expect_equal(
    out,
    list(
      year = 2000,
      month = 1,
      day = 2,
      hour = 3,
      minute = 40,
      second = 50.5,
      yday = 2
    )
  )
})

test_that("can explain a query", {
  db <- copy_to(
    memdb(),
    data.frame(x = 1:5),
    name = "test",
    indexes = list("x"),
    overwrite = TRUE
  )
  expect_snapshot(db |> filter(x > 2) |> explain())
})

test_that("db_col_types() returns column types (#1821)", {
  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  withr::defer(DBI::dbDisconnect(con))

  DBI::dbExecute(con, "CREATE TABLE df (id INTEGER, val REAL, name TEXT)")
  expect_equal(
    db_col_types(con, "df"),
    c(id = "INTEGER", val = "REAL", name = "TEXT")
  )

  # including with schema
  DBI::dbExecute(con, "ATTACH DATABASE ':memory:' AS aux")
  DBI::dbExecute(con, "CREATE TABLE aux.df (a INTEGER, b TEXT)")
})

#' SQL helpers for scalar functions
#'
#' @description
#' These functions help you create custom scalar SQL translations when
#' implementing a new backend. They are typically used within [sql_translator()]
#' to define how R functions should be translated to SQL.
#'
#' * `sql_infix()` creates SQL infix operators like `+`, `-`, `*`, `/`.
#' * `sql_prefix()` creates SQL prefix functions like `ABS()`, `SQRT()`.
#' * `sql_cast()` creates SQL cast expressions like `CAST(x AS type)`.
#' * `sql_try_cast()` creates SQL try_cast expressions (for safe casting).
#' * `sql_log()` creates a SQL logarithm function with optional base.
#' * `sql_cot()` creates a SQL cotangent function (as `1 / TAN(x)`).
#' * `sql_runif()` creates a SQL expression for generating uniform random
#'   numbers.
#'
#' @param f The name of the SQL function as a string.
#' @family SQL translation helpers
#' @name sql_translation_scalar
NULL

#' @rdname sql_translation_scalar
#' @param pad If `TRUE`, the default, pad the infix operator with spaces.
#' @export
sql_infix <- function(f, pad = TRUE) {
  # Unquoting involving infix operators easily create abstract syntax trees
  # without parentheses where they are needed for printing and translation.
  # For example `expr(!!expr(2 - 1) * x))`
  #
  # See https://adv-r.hadley.nz/quasiquotation.html#non-standard-ast
  # for more information.
  #
  # This is fixed with `escape_infix_expr()`
  # see https://github.com/tidyverse/dbplyr/issues/634
  check_string(f)
  check_bool(pad)

  function(x, y) {
    x <- escape_infix_expr(enexpr(x), x)
    y <- escape_infix_expr(enexpr(y), y)

    if (is.null(x)) {
      if (pad) {
        sql <- "{.sql f} {y}"
      } else {
        sql <- "{.sql f}{y}"
      }
    } else {
      if (pad) {
        sql <- "{x} {.sql f} {y}"
      } else {
        sql <- "{x}{.sql f}{y}"
      }
    }
    sql_glue(sql)
  }
}

escape_infix_expr <- function(xq, x, escape_unary_minus = FALSE) {
  infix_calls <- c("+", "-", "*", "/", "%%", "^")
  is_infix <- is_call(xq, infix_calls, n = 2)
  is_unary_minus <- escape_unary_minus &&
    is_call(xq, "-", n = 1) &&
    !is_atomic(x, n = 1)

  if (is_infix || is_unary_minus) {
    sql(paste0("(", x, ")"))
  } else {
    x
  }
}

#' @rdname sql_translation_scalar
#' @param n For `sql_prefix()`, an optional number of arguments to expect.
#'   Will signal error if not correct.
#' @export
sql_prefix <- function(f, n = NULL) {
  check_string(f)

  function(...) {
    args <- list(...)
    if (!is.null(n) && ...length() != n) {
      cli_abort(
        "Invalid number of args to SQL function {f}",
        i = "Expecting {n} and got {length(args)}"
      )
    }
    if (any(names2(args) != "")) {
      cli::cli_warn("Named arguments ignored for SQL {f}")
    }
    sql_glue("{.sql f}({...})")
  }
}

#' @rdname sql_translation_scalar
#' @param type SQL type name as a string.
#' @export
sql_cast <- function(type) {
  function(x) {
    sql_glue("CAST({x} AS {.sql type})")
  }
}

#' @rdname sql_translation_scalar
#' @export
sql_try_cast <- function(type) {
  function(x) {
    sql_glue("TRY_CAST({x} AS {.sql type})")
    # try_cast available in MSSQL 2012+
  }
}

#' @rdname sql_translation_scalar
#' @export
sql_log <- function() {
  function(x, base = exp(1)) {
    if (isTRUE(all.equal(base, exp(1)))) {
      sql_glue("LN({x})")
    } else {
      sql_glue("LOG({x}) / LOG({base})")
    }
  }
}


#' @rdname sql_translation_scalar
#' @export
sql_cot <- function() {
  function(x) {
    sql_glue("1 / TAN({x})")
  }
}

#' @rdname sql_translation_scalar
#' @param rand_expr A string giving an SQL expression that generates a
#'   random number between 0 and 1, e.g. `"RANDOM()"`.
#' @param min,max Range of random values.
#' @export
sql_runif <- function(rand_expr, n = n(), min = 0, max = 1) {
  rand_expr <- enexpr(rand_expr)
  if (!is_string(rand_expr)) {
    # for backward compatibility
    rand_expr <- sql_expr(!!rand_expr)
  }

  n_expr <- quo_get_expr(enquo(n))
  if (!is_call(n_expr, "n", n = 0)) {
    cli_abort("Only {.code n = n()} is supported.")
  }

  range <- max - min
  if (range != 1) {
    rand_expr <- sql_glue("{.sql rand_expr} * {range}")
  }

  if (min != 0) {
    rand_expr <- sql_glue("{.sql rand_expr} + {min}")
  }

  sql(rand_expr)
}

test_that("can translate infix expression without parentheses", {
  con <- dialect_ansi()
  expect_translation(con, !!expr(2 - 1) * x, '(2.0 - 1.0) * "x"')
  expect_translation(con, !!expr(2 / 1) * x, '(2.0 / 1.0) * "x"')
  expect_translation(con, !!expr(2 * 1) - x, '(2.0 * 1.0) - "x"')
})

test_that("unary minus works with expressions", {
  con <- dialect_ansi()
  expect_translation(con, -!!expr(x + 2), '-("x" + 2.0)')
  expect_translation(con, --x, '--"x"')
})

test_that("sql_infix generates expected output (#1345)", {
  local_con(dialect_ansi())
  x <- ident_q("x")
  y <- ident_q("y")

  expect_equal(sql_infix("-", pad = FALSE)(x, y), sql("x-y"))
  expect_equal(sql_infix("-", pad = FALSE)(NULL, y), sql("-y"))
  expect_equal(sql_infix("-")(x, y), sql("x - y"))
  expect_equal(sql_infix("-")(NULL, y), sql("- y"))
})

test_that("sql_prefix checks arguments", {
  con <- dialect_ansi()
  sin_db <- sql_prefix("SIN", 1)

  expect_snapshot(error = TRUE, sin_db(sin(1, 2)))
  expect_snapshot(error = TRUE, sin_db(sin(a = 1)))
})

test_that("runif is translated", {
  con <- dialect_ansi()
  expect_translation(con, runif(n()), "RANDOM()")
  expect_translation(con, runif(n(), max = 2), "RANDOM() * 2.0")
  expect_translation(con, runif(n(), min = 1, max = 2), "RANDOM() + 1.0")
  expect_translation(con, runif(n(), min = 1, max = 3), "RANDOM() * 2.0 + 1.0")

  expect_translation_snapshot(con, runif(2), error = TRUE)
})

test_that("sql_runif() still works with an expression", {
  local_con(dialect_ansi())

  expect_equal(sql_runif(RANDOM()), sql("RANDOM()"))
  expect_equal(sql_runif(RANDOM(), min = 1, max = 2), sql("RANDOM() + 1.0"))
})

#' Subset, rename, and reorder columns using their names
#'
#' @description
#' These are methods for the dplyr [select()], [rename()], and [relocate()]
#' generics. They generate the `SELECT` clause of the SQL query.
#'
#' These functions do not support predicate functions, i.e. you can
#' not use `where(is.numeric)` to select all numeric variables.
#'
#' @inheritParams arrange.tbl_lazy
#' @inheritParams dplyr::select
#' @export
#' @importFrom dplyr select
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(x = 1, y = 2, z = 3)
#' db |> select(-y) |> show_query()
#' db |> relocate(z) |> show_query()
#' db |> rename(first = x, last = z) |> show_query()
select.tbl_lazy <- function(.data, ...) {
  loc <- tidyselect::eval_select(expr(c(...)), .data)
  loc <- ensure_group_vars(loc, .data, notify = TRUE)
  new_vars <- set_names(colnames(.data)[loc], names(loc))

  .data$lazy_query <- add_select(.data$lazy_query, new_vars)
  .data
}

#' @rdname select.tbl_lazy
#' @importFrom dplyr rename
#' @export
rename.tbl_lazy <- function(.data, ...) {
  loc <- tidyselect::eval_rename(expr(c(...)), .data)

  new_vars <- set_names(colnames(.data), colnames(.data))
  names(new_vars)[loc] <- names(loc)

  .data$lazy_query <- add_select(.data$lazy_query, new_vars)
  .data
}

#' @rdname select.tbl_lazy
#' @importFrom dplyr rename_with
#' @importFrom tidyselect everything
#' @inheritParams dplyr::rename_with
#' @export
rename_with.tbl_lazy <- function(.data, .fn, .cols = everything(), ...) {
  .fn <- as_function(.fn)
  cols <- tidyselect::eval_select(enquo(.cols), .data)

  new_vars <- set_names(op_vars(.data))
  names(new_vars)[cols] <- .fn(new_vars[cols], ...)

  .data$lazy_query <- add_select(.data$lazy_query, new_vars)
  .data
}

#' @rdname select.tbl_lazy
#' @importFrom dplyr relocate
#' @inheritParams dplyr::relocate
#' @export
relocate.tbl_lazy <- function(.data, ..., .before = NULL, .after = NULL) {
  loc <- tidyselect::eval_relocate(
    expr(c(...)),
    data = .data,
    before = enquo(.before),
    after = enquo(.after),
    before_arg = ".before",
    after_arg = ".after"
  )

  new_vars <- set_names(colnames(.data)[loc], names(loc))

  .data$lazy_query <- add_select(.data$lazy_query, new_vars)
  .data
}

# op_select ---------------------------------------------------------------

add_select <- function(lazy_query, vars) {
  check_character(vars)
  vars_data <- op_vars(lazy_query)

  if (is_identity(syms(vars), names(vars), vars_data)) {
    return(lazy_query)
  }

  # Update grouping/ordering with new variable names
  lazy_query$group_vars <- rename_groups(op_grps(lazy_query), vars)
  lazy_query$order_vars <- rename_order(op_sort(lazy_query), vars)

  is_join <- inherits(lazy_query, "lazy_multi_join_query") ||
    inherits(lazy_query, "lazy_rf_join_query") ||
    inherits(lazy_query, "lazy_semi_join_query")
  if (is_join) {
    if (can_inline_select_into_join(lazy_query, vars)) {
      idx <- vctrs::vec_match(vars, vars_data)

      lazy_query$vars <- vctrs::vec_slice(lazy_query$vars, idx)
      lazy_query$vars$name <- names(vars)
      lazy_query
    } else {
      lazy_select_query(
        x = lazy_query,
        select_operation = "select",
        select = syms(vars)
      )
    }
  } else if (can_inline_select(lazy_query, vars)) {
    idx <- vctrs::vec_match(vars, vars_data)

    lazy_query$select <- vctrs::vec_slice(lazy_query$select, idx)
    lazy_query$select$name <- names(vars)
    lazy_query
  } else {
    lazy_select_query(
      x = lazy_query,
      select_operation = "select",
      select = syms(vars)
    )
  }
}

# select() rewrites the join's projection
# * WHERE is back-qualified at sql_build using vars$name
#   => columns referenced in WHERE must survive the select unchanged
can_inline_select_into_join <- function(lazy_query, vars) {
  if (!inherits(lazy_query, "lazy_multi_join_query")) {
    return(TRUE)
  }
  if (length(lazy_query$where) == 0) {
    return(TRUE)
  }

  where_names <- unique(unlist(lapply(lazy_query$where, all_names)))
  identity_kept <- unname(vars[names(vars) == vars])
  all(where_names %in% identity_kept)
}

# select() modifies the SELECT clause
# * ORDER BY can reference SELECT columns by name
#   => computed columns used in ORDER BY must be preserved
# * DISTINCT operates on the SELECT result
#   => can only inline bijective projections (same columns, different order)
can_inline_select <- function(lazy_query, vars) {
  if (!is_lazy_select_query(lazy_query)) {
    return(FALSE)
  }

  is_mutate <- !purrr::map_lgl(lazy_query$select$expr, is_symbol)
  computed_columns <- lazy_query$select$name[is_mutate]
  order_vars <- purrr::map_chr(lazy_query$order_by, as_label)
  ordered_present <- all(intersect(computed_columns, order_vars) %in% vars)
  if (!ordered_present) {
    return(FALSE)
  }

  if (is_true(lazy_query$distinct)) {
    identical(sort(unname(vars)), op_vars(lazy_query))
  } else {
    TRUE
  }
}

rename_groups <- function(group_vars, select_vars) {
  old2new <- set_names(names(select_vars), select_vars)
  renamed <- group_vars %in% names(old2new)
  group_vars[renamed] <- old2new[group_vars[renamed]]
  group_vars
}

rename_order <- function(order_vars, select_vars) {
  # Drop any ordering variables not used in the output
  order_names <- purrr::map_chr(order_vars, \(var) all_names(var)[[1]])
  order_vars <- order_vars[order_names %in% select_vars]

  # Rename the remaining
  order_vars[] <- replace_sym(order_vars, select_vars, syms(names(select_vars)))
  order_vars
}

# Helpers ----------------------------------------------------------------------

# Selection, rename, or relocation
is_projection <- function(exprs) {
  purrr::every(exprs, is_symbol)
}

# Selection or relocation
is_pure_projection <- function(exprs, names) {
  if (!is_projection(exprs)) {
    return(FALSE)
  }

  expr_vars <- purrr::map_chr(unname(exprs), as_string)
  identical(expr_vars, names)
}

# Selects all variables
is_identity <- function(exprs, names, names_prev) {
  if (!is_pure_projection(exprs, names)) {
    return(FALSE)
  }

  identical(names, names_prev)
}

ensure_group_vars <- function(loc, data, notify = TRUE) {
  group_loc <- match(group_vars(data), colnames(data))
  missing <- setdiff(group_loc, loc)

  if (length(missing) > 0) {
    vars <- colnames(data)[missing]
    if (notify) {
      cli::cli_inform("Adding missing grouping variables: {.var {vars}}")
    }
    loc <- c(set_names(missing, vars), loc)
  }

  loc
}

#' Pivot data from long to wide
#'
#' @description
#' `pivot_wider()` "widens" data, increasing the number of columns and
#' decreasing the number of rows. The inverse transformation is
#' `pivot_longer()`. Learn more in `vignette("pivot", "tidyr")`.
#'
#' `pivot_wider()` on database tables comes with some caveats, please make
#' sure to read below for details.
#'
#' # Caveats
#'
#' ## `pivot_wider()` is eager
#'
#' Note that `pivot_wider()` cannot be lazy because we need to look
#' at the data to figure out what the new column names will be.
#' If you have a long-running query you have two options:
#'
#' * Temporarily store the result of the query via `compute()`.
#' * Create a spec before and use `dbplyr_pivot_wider_spec()` - dbplyr's version
#'   of `tidyr::pivot_wider_spec()`.
#'
#' ## You must supply `values_fn`
#'
#' The big difference to `pivot_wider()` for local data frames is that
#' `values_fn` must not be `NULL`. By default it is `max()` which yields
#' the same results as for local data frames if three conditions are true:
#'
#' 1. The combination of `id_cols` and `value` uniquely identify an observation.
#' 1. The column has a comparable type (e.g. numeric, date-time, or
#'    (for most databases) string).
#' 1. `values_fill` is `NULL`.
#'
#' If either the second or third condition is not met, you must supply a
#' custom `values_fn`. Unfortunately there is no generally available alternative
#' and you'll need to look for something database specific, like `FIRST()`
#' or `ANY_VALUE()`.
#'
#' # How does it work?
#'
#' The translation to SQL code basically works as follows:
#'
#' 1. Get unique keys in `names_from` column.
#' 2. For each key value generate an expression of the form:
#'     ```sql
#'     values_fn(
#'       CASE WHEN (`names from column` == `key value`) THEN (`value column`) END
#'     ) AS `output column`
#'     ```
#' 3. Group data by id columns.
#' 4. Summarise the grouped data with the expressions from step 2.
#'
#' @param data A lazy data frame backed by a database query.
#' @param id_cols A set of columns that uniquely identifies each observation.
#' @param id_expand Unused; included for compatibility with the generic.
#' @param names_from,values_from A pair of
#'   arguments describing which column (or columns) to get the name of the
#'   output column (`names_from`), and which column (or columns) to get the
#'   cell values from (`values_from`).
#'
#'   If `values_from` contains multiple values, the value will be added to the
#'   front of the output column.
#' @param names_prefix String added to the start of every variable name.
#' @param names_sep If `names_from` or `values_from` contains multiple
#'   variables, this will be used to join their values together into a single
#'   string to use as a column name.
#' @param names_glue Instead of `names_sep` and `names_prefix`, you can supply
#'   a glue specification that uses the `names_from` columns (and special
#'   `.value`) to create custom column names.
#' @param names_sort Should the column names be sorted? If `FALSE`, the default,
#'   column names are ordered by first appearance.
#' @param names_vary When `names_from` identifies a column (or columns) with
#'   multiple unique values, and multiple `values_from` columns are provided,
#'   in what order should the resulting column names be combined?
#'
#'   - `"fastest"` varies `names_from` values fastest, resulting in a column
#'     naming scheme of the form: `value1_name1, value1_name2, value2_name1,
#'     value2_name2`. This is the default.
#'
#'   - `"slowest"` varies `names_from` values slowest, resulting in a column
#'     naming scheme of the form: `value1_name1, value2_name1, value1_name2,
#'     value2_name2`.
#' @param names_repair What happens if the output has invalid column names?
#' @param names_expand Should the values in the `names_from` columns be expanded
#'   by [expand()] before pivoting? This results in more columns, the output
#'   will contain column names corresponding to a complete expansion of all
#'   possible values in `names_from`. Additionally, the column names will be
#'   sorted, identical to what `names_sort` would produce.
#' @param values_fill Optionally, a (scalar) value that specifies what each
#'   `value` should be filled in with when missing. Be careful when using this
#'   in combination with the default `values_fn`.
#' @param values_fn A function, the default is `max()`, applied to the `value`
#' in each cell in the output. In contrast to local data frames it must not be
#' `NULL`.
#' @param unused_fn Optionally, a function applied to summarize the values from
#'   the unused columns (i.e. columns not identified by `id_cols`,
#'   `names_from`, or `values_from`).
#'
#'   The default drops all unused columns from the result.
#'
#'   This can be a named list if you want to apply different aggregations
#'   to different unused columns.
#'
#'   `id_cols` must be supplied for `unused_fn` to be useful, since otherwise
#'   all unspecified columns will be considered `id_cols`.
#'
#'   This is similar to grouping by the `id_cols` then summarizing the
#'   unused columns using `unused_fn`.
#' @param ... Unused; included for compatibility with generic.
#'
#' @examplesIf rlang::is_installed("tidyr", version = "1.0.0")
#' memdb_frame(
#'   id = 1,
#'   key = c("x", "y"),
#'   value = 1:2
#' ) |>
#'   tidyr::pivot_wider(
#'     id_cols = id,
#'     names_from = key,
#'     values_from = value
#'   )
#' @exportS3Method tidyr::pivot_wider
pivot_wider.tbl_lazy <- function(
  data,
  ...,
  id_cols = NULL,
  id_expand = FALSE,
  names_from = name,
  names_prefix = "",
  names_sep = "_",
  names_glue = NULL,
  names_sort = FALSE,
  names_vary = "fastest",
  names_expand = FALSE,
  names_repair = "check_unique",
  values_from = value,
  values_fill = NULL,
  values_fn = ~ max(.x, na.rm = TRUE),
  unused_fn = NULL
) {
  rlang::check_dots_empty()
  check_unsupported_arg(id_expand, FALSE)

  names_from <- enquo(names_from)
  values_from <- enquo(values_from)

  spec <- dbplyr_build_wider_spec(
    data,
    names_from = !!names_from,
    values_from = !!values_from,
    names_prefix = names_prefix,
    names_sep = names_sep,
    names_glue = names_glue,
    names_sort = names_sort,
    names_vary = names_vary,
    names_expand = names_expand,
    error_call = current_env()
  )

  id_cols <- build_wider_id_cols_expr(
    data = data,
    id_cols = {{ id_cols }},
    names_from = !!names_from,
    values_from = !!values_from
  )

  dbplyr_pivot_wider_spec(
    data = data,
    spec = spec,
    ...,
    names_repair = names_repair,
    id_cols = !!id_cols,
    id_expand = FALSE,
    values_fill = values_fill,
    values_fn = values_fn,
    unused_fn = unused_fn,
    error_call = current_env()
  )
}

dbplyr_build_wider_spec <- function(
  data,
  names_from = name,
  values_from = value,
  names_prefix = "",
  names_sep = "_",
  names_glue = NULL,
  names_sort = FALSE,
  names_vary = "fastest",
  names_expand = FALSE,
  error_call = current_env()
) {
  if (!inherits(data, "tbl_sql")) {
    cli_abort(c(
      "{.fun dbplyr_build_wider_spec} doesn't work with local lazy tibbles.",
      i = "Use {.fun memdb_frame} together with {.fun show_query} to see the SQL code."
    ))
  }

  # prepare a minimal local tibble we can pass to `tidyr::build_wider_spec`
  # 1. create a tibble with unique values in the `names_from` column
  # row_ids <- vec_unique(data[names_from])
  names_from <- tidyselect::eval_select(enquo(names_from), data) |> names()
  if (is_empty(names_from)) {
    cli_abort("{.arg names_from} must select at least one column.")
  }
  distinct_data <- collect(distinct(data, !!!syms(names_from)))

  # 2. add `values_from` column
  values_from <- tidyselect::eval_select(enquo(values_from), data) |> names()
  if (is_empty(values_from)) {
    cli_abort("{.arg values_from} must select at least one column.")
  }
  dummy_data <- vctrs::vec_cbind(
    distinct_data,
    !!!rlang::rep_named(values_from, list(TRUE)),
    .name_repair = "check_unique"
  )

  tidyr::build_wider_spec(
    dummy_data,
    names_from = !!enquo(names_from),
    values_from = !!enquo(values_from),
    names_prefix = names_prefix,
    names_sep = names_sep,
    names_glue = names_glue,
    names_sort = names_sort,
    names_vary = names_vary,
    names_expand = names_expand,
    error_call = error_call
  )
}

#' @inheritParams rlang::args_error_context
#' @inheritParams tidyr::pivot_wider_spec
#' @export
#' @rdname pivot_wider.tbl_lazy
dbplyr_pivot_wider_spec <- function(
  data,
  spec,
  ...,
  names_repair = "check_unique",
  id_cols = NULL,
  id_expand = FALSE,
  values_fill = NULL,
  values_fn = ~ max(.x, na.rm = TRUE),
  unused_fn = NULL,
  error_call = current_env()
) {
  check_unsupported_arg(id_expand, FALSE)
  spec <- tidyr::check_pivot_spec(spec)

  names_from_cols <- names(spec)[-(1:2)]
  values_from_cols <- vctrs::vec_unique(spec$.value)

  id_cols <- select_wider_id_cols(
    data = data,
    id_cols = {{ id_cols }},
    names_from_cols = names_from_cols,
    values_from_cols = values_from_cols,
    error_call = error_call
  )

  values_fn <- check_list_of_functions(
    values_fn,
    values_from_cols,
    call = error_call
  )

  unused_cols <- setdiff(
    colnames(data),
    c(id_cols, names_from_cols, values_from_cols)
  )
  unused_fn <- check_list_of_functions(
    unused_fn,
    unused_cols,
    call = error_call
  )
  unused_cols <- names(unused_fn)

  if (is.null(values_fill)) {
    values_fill <- list()
  } else if (is_scalar(values_fill)) {
    values_fill <- rep_named(values_from_cols, list(values_fill))
  } else if (!vctrs::vec_is_list(values_fill)) {
    cli::cli_abort(
      "{.arg values_fill} must be {.code NULL}, a scalar, or a named list, not {.obj_type_friendly {values_fill}}.",
      call = error_call
    )
  }
  values_fill <- values_fill[intersect(names(values_fill), values_from_cols)]

  missing_values <- setdiff(values_from_cols, names(values_fn))
  if (!is_empty(missing_values)) {
    cli_abort(
      "{.arg values_fn} must specify a function for each col in {.arg values_from}"
    )
  }

  n_unused_fn <- length(unused_fn)

  unused_col_expr <- vector("list", length = n_unused_fn)
  names(unused_col_expr) <- unused_cols

  for (i in seq_len(n_unused_fn)) {
    unused_col <- unused_cols[[i]]
    unused_fn_i <- unused_fn[[i]]

    unused_col_expr[[i]] <- resolve_fun(
      unused_fn_i,
      sym(unused_col),
      call = error_call
    )
  }

  spec_idx <- set_names(vctrs::vec_seq_along(spec), spec$.name)
  pivot_exprs <- with_indexed_errors(
    purrr::map(
      spec_idx,
      ~ build_pivot_wider_exprs(.x, spec, values_fill, values_fn, call = NULL)
    ),
    message = function(cnd) {
      col <- spec[[".value"]][cnd$location]
      cli::format_inline("Can't pivot column {.field {col}}:")
    },
    .error_call = error_call
  )

  non_id_cols <- c(names_from_cols, values_from_cols)
  key_vars <- setdiff(id_cols, non_id_cols)
  data_grouped <- group_by(data, !!!syms(key_vars), .add = TRUE)

  group_names <- group_vars(data_grouped)
  out_nms <- c(group_names, names(pivot_exprs))
  out_nms_repaired <- vctrs::vec_as_names(out_nms, repair = names_repair)

  if (!is_empty(group_names)) {
    out_nms_repaired <- out_nms_repaired[-(1:length(group_names))]
  }
  pivot_exprs <- set_names(pivot_exprs, out_nms_repaired)

  data_grouped |>
    summarise(
      !!!pivot_exprs,
      !!!unused_col_expr,
      .groups = "drop"
    ) |>
    group_by(!!!syms(group_vars(data)))
}

utils::globalVariables(c("name", "value"))

build_wider_id_cols_expr <- function(
  data,
  id_cols = NULL,
  names_from = name,
  values_from = value,
  call = caller_env()
) {
  # COPIED FROM tidyr
  names_from <- tidyselect::eval_select(
    enquo(names_from),
    data,
    allow_rename = FALSE,
    error_call = error_call
  )

  values_from <- tidyselect::eval_select(
    enquo(values_from),
    data,
    allow_rename = FALSE,
    error_call = error_call
  )

  out <- select_wider_id_cols(
    data = data,
    id_cols = {{ id_cols }},
    names_from_cols = names(names_from),
    values_from_cols = names(values_from),
    error_call = error_call
  )

  expr(c(!!!out))
}

select_wider_id_cols <- function(
  data,
  id_cols = NULL,
  names_from_cols = character(),
  values_from_cols = character(),
  error_call = caller_env()
) {
  # COPIED FROM tidyr
  id_cols <- enquo(id_cols)
  sim_data <- tidyselect_data_proxy(data)

  # Remove known non-id-cols so they are never selected
  sim_data <- sim_data[setdiff(
    names(sim_data),
    c(names_from_cols, values_from_cols)
  )]

  if (quo_is_null(id_cols)) {
    # Default selects everything in `sim_data` after non-id-cols have been removed
    return(names(sim_data))
  }

  withCallingHandlers(
    id_cols <- tidyselect::eval_select(
      enquo(id_cols),
      sim_data,
      allow_rename = FALSE,
      error_call = error_call
    ),
    vctrs_error_subscript_oob = function(cnd) {
      rethrow_id_cols_oob(cnd, names_from_cols, values_from_cols, error_call)
    }
  )

  names(id_cols)
}

rethrow_id_cols_oob <- function(cnd, names_from_cols, values_from_cols, call) {
  i <- cnd[["i"]]

  check_string(i, .internal = TRUE)

  if (i %in% names_from_cols) {
    stop_id_cols_oob(i, "names_from", call = call)
  } else if (i %in% values_from_cols) {
    stop_id_cols_oob(i, "values_from", call = call)
  } else {
    # Zap this special handler, throw the normal condition
    zap()
  }
}

stop_id_cols_oob <- function(i, arg, call) {
  cli::cli_abort(
    c(
      "`id_cols` can't select a column already selected by `{arg}`.",
      i = "Column `{i}` has already been selected."
    ),
    parent = NA,
    call = call
  )
}

build_pivot_wider_exprs <- function(
  row_id,
  spec,
  values_fill,
  values_fn,
  call
) {
  values_col <- spec[[".value"]][row_id]
  fill_value <- values_fill[[values_col]]

  keys <- vctrs::vec_slice(spec[, -(1:2)], row_id)
  keys_cond <- purrr::imap(
    keys,
    function(value, name) {
      if (is.na(value)) {
        expr(is.na(!!sym(name)))
      } else {
        expr(!!sym(name) == !!value)
      }
    }
  )
  keys_cond <- purrr::reduce(keys_cond, ~ expr(!!.x & !!.y))

  case_expr <- expr(ifelse(!!keys_cond, !!sym(values_col), !!fill_value))

  agg_fn <- values_fn[[values_col]]
  resolve_fun(
    agg_fn,
    case_expr,
    arg = paste0("values_fn$", values_col),
    call = call
  )
}

is_scalar <- function(x) {
  if (is.null(x)) {
    return(FALSE)
  }

  if (is.list(x)) {
    (length(x) == 1) && !have_name(x)
  } else {
    length(x) == 1
  }
}


resolve_fun <- function(x, var, arg = caller_arg(x), call = caller_env()) {
  if (is_formula(x) || is_function(x)) {
    .fn_expr <- across_fun(x, env = empty_env(), fn = "across")
    exec(.fn_expr, var, NULL)
  } else {
    cli_abort(
      "Can't convert {.arg {arg}}, {.code {as_label(x)}}, to a function.",
      call = call
    )
  }
}

spec <- tibble(
  .name = c("x", "y"),
  .value = "val",
  key = c("x", "y")
)

spec1 <- tibble(.name = "x", .value = "val", key = "x")

test_that("can pivot all cols to wide", {
  expect_equal(
    memdb_frame(key = c("x", "y", "z"), val = 1:3) %>%
      tidyr::pivot_wider(names_from = key, values_from = val) %>%
      collect(),
    tibble(x = 1, y = 2, z = 3)
  )

  spec <- tibble(
    .name = c("x", "y", "z"),
    .value = "val",
    key = c("x", "y", "z")
  )

  expect_snapshot(
    lazy_frame(key = c("x", "y", "z"), val = 1:3) %>%
      dbplyr_pivot_wider_spec(spec)
  )
})

test_that("non-pivoted cols are preserved", {
  df <- lazy_frame(a = 1, key = c("x", "y"), val = 1:2)

  expect_equal(
    dbplyr_pivot_wider_spec(df, spec) %>% op_vars(),
    c("a", "x", "y")
  )
})

test_that("implicit missings turn into explicit missings", {
  df <- memdb_frame(a = 1:2, key = c("x", "y"), val = 1:2)

  expect_equal(
    memdb_frame(a = 1:2, key = c("x", "y"), val = 1:2) %>%
      tidyr::pivot_wider(names_from = key, values_from = val) %>%
      collect(),
    tibble(a = 1:2, x = c(1, NA), y = c(NA, 2))
  )

  expect_snapshot(
    lazy_frame(a = 1:2, key = c("x", "y"), val = 1:2) %>%
      dbplyr_pivot_wider_spec(spec)
  )
})

test_that("error when overwriting existing column", {
  df <- memdb_frame(
    a = c(1, 1),
    key = c("a", "b"),
    val = c(1, 2)
  )
  expect_snapshot(error = TRUE,
    tidyr::pivot_wider(df, names_from = key, values_from = val)
  )
})

test_that("`names_repair` happens after spec column reorganization (#1107)", {
  df <- memdb_frame(
    test = c("a", "b"),
    name = c("test", "test2"),
    value = c(1, 2)
  )

  out <- tidyr::pivot_wider(df, names_repair = ~make.unique(.x)) %>%
    collect()

  expect_identical(out$test, c("a", "b"))
  expect_identical(out$test.1, c(1, NA))
  expect_identical(out$test2, c(NA, 2))
})

test_that("minimal `names_repair` doesn't overwrite a value column that collides with key column (#1107)", {
  skip("`grouped_df()` needs a `name_repair` argument")
  # `collect.tbl_sql()` does not work with duplicated names
  df <- memdb_frame(
    test = c("a", "b"),
    name = c("test", "test2"),
    value = c(1, 2)
  )

  out <- tidyr::pivot_wider(df, names_repair = "minimal") %>%
    collect()

  expect_identical(out[[1]], c("a", "b"))
  expect_identical(out[[2]], c(1, NA))
  expect_identical(out[[3]], c(NA, 2))
})

test_that("grouping is preserved", {
  df <- lazy_frame(a = 1, key = "x", val = 2)

  expect_equal(
    df %>%
      dplyr::group_by(a) %>%
      dbplyr_pivot_wider_spec(spec1) %>%
      group_vars(),
    "a"
  )
})

# https://github.com/tidyverse/tidyr/issues/804
test_that("column with `...j` name can be used as `names_from`", {
  df <- memdb_frame(...8 = c("x", "y", "z"), val = 1:3)
  pv <- tidyr::pivot_wider(df, names_from = ...8, values_from = val) %>% collect()
  expect_named(pv, c("x", "y", "z"))
})


# column names -------------------------------------------------------------

test_that("dbplyr_build_wider_spec can handle multiple columns", {
  df <- memdb_frame(
    x = c("X", "Y"),
    y = 1:2,
    a = 1:2,
    b = 1:2
  )

  expect_equal(
    dbplyr_build_wider_spec(df, x:y, a:b),
    tibble::tribble(
      ~.name, ~.value,  ~x, ~y,
      "a_X_1",     "a", "X", 1L,
      "a_Y_2",     "a", "Y", 2L,
      "b_X_1",     "b", "X", 1L,
      "b_Y_2",     "b", "Y", 2L
    )
  )
})

# keys ---------------------------------------------------------

test_that("can override default keys", {
  df <- tibble::tribble(
    ~row, ~name, ~var, ~value,
    1,    "Sam", "age", 10,
    2,    "Sam", "height", 1.5,
    3,    "Bob", "age", 20,
  )

  df_db <- memdb_frame(!!!df)

  expect_equal(
    df_db %>%
      tidyr::pivot_wider(id_cols = name, names_from = var, values_from = value) %>%
      collect(),
    tibble::tribble(
      ~name, ~age, ~height,
      "Bob",   20,      NA,
      "Sam",   10,     1.5
    )
  )
})

test_that("`id_cols = everything()` excludes `names_from` and `values_from`", {
  df <- memdb_frame(key = "x", name = "a", value = 1L)

  expect_identical(
    tidyr::pivot_wider(df, id_cols = everything()) %>% collect(),
    tibble(key = "x", a = 1L)
  )

  spec <- dbplyr_build_wider_spec(df)

  expect_identical(
    dbplyr_pivot_wider_spec(df, spec, id_cols = everything()) %>% collect(),
    tibble(key = "x", a = 1L)
  )
})

test_that("pivoting a zero row data frame drops `names_from` and `values_from` (#1249)", {
  df <- memdb_frame(key = character(), name = character(), value = integer())

  expect_identical(
    tidyr::pivot_wider(df, names_from = name, values_from = value) %>% collect(),
    tibble(key = character())
  )
})

test_that("known bug - building a wider spec with a zero row data frame loses `values_from` info (#1249)", {
  # We can't currently change this behavior in `pivot_wider_spec()`,
  # for fear of breaking backwards compatibility

  df <- memdb_frame(key = character(), name = character(), value = integer())

  # Building the spec loses the fact that `value` was specified as `values_from`,
  # which would normally be in the `spec$.value` column
  spec <- dbplyr_build_wider_spec(df, names_from = name, values_from = value)

  # So pivoting with this spec accidentally keeps `value` around
  expect_identical(
    dbplyr_pivot_wider_spec(df, spec) %>% collect(),
    tibble(key = character(), value = integer())
  )

  # If you specify `id_cols` to be the `key` column, it works right
  expect_identical(
    dbplyr_pivot_wider_spec(df, spec, id_cols = key) %>% collect(),
    tibble(key = character())
  )

  # But `id_cols = everything()` won't work as intended, because we can't know
  # to remove `value` from `names(data)` before computing the tidy-selection
  expect_identical(
    dbplyr_pivot_wider_spec(df, spec, id_cols = everything()) %>% collect(),
    tibble(key = character(), value = integer())
  )
})

# non-unqiue keys ---------------------------------------------------------

test_that("values_fn can be a single function", {
  df <- lazy_frame(a = c(1, 1, 2), key = c("x", "x", "x"), val = c(1, 10, 100))

  expect_snapshot(
    suppressWarnings(dbplyr_pivot_wider_spec(df, spec1, values_fn = sum))
  )
})

test_that("values_fn can be a formula", {
  df <- lazy_frame(a = c(1, 1, 2), key = c("x", "x", "x"), val = c(1, 10, 100))

  expect_snapshot(dbplyr_pivot_wider_spec(df, spec1, values_fn = ~ sum(.x, na.rm = TRUE)))
})

test_that("values_fn can be a named list", {
  df <- lazy_frame(
    key = c("x", "x"),
    a = c(1, 2),
    b = c(3, 4)
  )

  spec <- tibble(
    .name = c("a_x", "b_x"),
    .value = c("a", "b"),
    key = "x"
  )

  dbplyr_pivot_wider_spec(
    df, spec,
    values_fn = list(a = sum, b = ~ sum(.x, na.rm = TRUE))
  )

  # must specify `values_fn` for every column
  expect_snapshot_error(
    dbplyr_pivot_wider_spec(df, spec, values_fn = list(a = sum))
  )
  # no function must be `NULL`
  expect_snapshot_error(
    dbplyr_pivot_wider_spec(df, spec, values_fn = list(a = sum, b = NULL))
  )
})

test_that("values_fn cannot be NULL", {
  df <- lazy_frame(a = 1, key = "x", val = 1)

  expect_snapshot(error = TRUE, dbplyr_pivot_wider_spec(df, spec1, values_fn = NULL))
})


# unused -------------------------------------------------------------------

test_that("`unused_fn` can summarize unused columns (#990)", {
  df <- memdb_frame(
    id = c(1, 1, 2, 2),
    unused1 = c(1, 2, 4, 3),
    unused2 = c(11, 12, 14, 13),
    name = c("a", "b", "a", "b"),
    value = c(1, 2, 3, 4)
  )

  # By name
  suppressWarnings(
    res <- tidyr::pivot_wider(df, id_cols = id, unused_fn = list(unused1 = max)) %>%
      collect()
  )
  expect_named(res, c("id", "a", "b", "unused1"))
  expect_identical(res$unused1, c(2, 4))

  # Globally
  suppressWarnings(
    res <- tidyr::pivot_wider(df, id_cols = id, unused_fn = min) %>%
      collect()
  )
  expect_named(res, c("id", "a", "b", "unused1", "unused2"))
  expect_identical(res$unused1, c(1, 3))
  expect_identical(res$unused2, c(11, 13))
})

test_that("`unused_fn` works with anonymous functions", {
  df <- memdb_frame(
    id = c(1, 1, 2, 2),
    unused = c(1, NA, 4, 3),
    name = c("a", "b", "a", "b"),
    value = c(1, 2, 3, 4)
  )

  res <- tidyr::pivot_wider(df, id_cols = id, unused_fn = ~mean(.x, na.rm = TRUE)) %>%
    collect()
  expect_identical(res$unused, c(1, 3.5))
})

test_that("`unused_fn` is validated", {
  df <- memdb_frame(id = 1, unused = 1, name = "a", value = 1)

  expect_snapshot(
    (expect_error(tidyr::pivot_wider(df, id_cols = id, unused_fn = 1)))
  )
})

# can fill missing cells --------------------------------------------------

test_that("can fill in missing cells", {
  spec <- tibble(
    .name = c("x", "y"),
    .value = "value",
    name = c("x", "y")
  )
  df <- memdb_frame(g = c(1, 2), name = c("x", "y"), value = c(1, 2))
  df_lazy <- lazy_frame(g = c(1, 2), name = c("x", "y"), value = c(1, 2))

  expect_equal(tidyr::pivot_wider(df) %>% pull(x), c(1, NA))

  expect_equal(tidyr::pivot_wider(df, values_fill = 0) %>% pull(x), c(1, 0))
  expect_snapshot(dbplyr_pivot_wider_spec(df_lazy, spec, values_fill = 0))

  expect_equal(
    tidyr::pivot_wider(df, values_fill = list(value = 0)) %>%
      pull(x),
    c(1, 0)
  )
  expect_snapshot(
    dbplyr_pivot_wider_spec(
      df_lazy,
      spec,
      values_fill = list(value = 0)
    )
  )
})

test_that("values_fill only affects missing cells", {
  df <- memdb_frame(g = c(1, 2), name = c("x", "y"), value = c(1, NA))
  dbplyr_build_wider_spec(df)
  out <- tidyr::pivot_wider(df, values_fill = 0) %>%
    collect()
  expect_equal(out$y, c(0, NA))
})

test_that("values_fill is checked", {
  lf <- lazy_frame(g = c(1, 2), name = c("x", "y"), value = c(1, NA))
  spec <- tibble(
    .name = c("x", "y"),
    .value = "value",
    name = .name
  )
  expect_snapshot(
    error = TRUE,
    dbplyr_pivot_wider_spec(lf, spec, values_fill = 1:2)
  )
})

# multiple values ----------------------------------------------------------

test_that("can pivot from multiple measure cols", {
  df <- memdb_frame(row = 1, var = c("x", "y"), a = 1:2, b = 3:4)
  pv <- tidyr::pivot_wider(df, names_from = var, values_from = c(a, b)) %>%
    collect()

  expect_named(pv, c("row", "a_x", "a_y", "b_x", "b_y"))
  expect_equal(pv$a_x, 1)
  expect_equal(pv$b_y, 4)
})

test_that("column order in output matches spec", {
  df <- tibble::tribble(
    ~hw,   ~name,  ~mark,   ~pr,
    "hw1", "anna",    95,  "ok",
    "hw2", "anna",    70, "meh",
  )

  # deliberately create weird order
  sp <- tibble::tribble(
    ~hw, ~.value,  ~.name,
    "hw1", "mark", "hw1_mark",
    "hw1", "pr",   "hw1_pr",
    "hw2", "pr",   "hw2_pr",
    "hw2", "mark", "hw2_mark",
  )

  pv <- dbplyr_pivot_wider_spec(lazy_frame(!!!df), sp)
  expect_equal(pv %>% op_vars(), c("name", sp$.name))
})

test_that("cannot pivot lazy frames", {
  expect_snapshot(error = TRUE, tidyr::pivot_wider(lazy_frame(name = "x", value = 1)))
})

# multiple names ----------------------------------------------------------

test_that("can pivot multiple from multiple names", {
  x <- tibble(
    seq = c(1, 1, 2, 2),
    name = rep(c("id", "name"), 2),
    value = c("01", "curie", "02", "arrhenius")
  )

  expect_equal(
    memdb_frame(x) %>%
      tidyr::pivot_wider(
        names_from = c(name, seq),
        values_from = value
      ) %>%
      collect(),
    tibble(id_1 = "01", name_1 = "curie", id_2 = "02", name_2 = "arrhenius")
  )
})


# pass through arguments --------------------------------------------------

test_that("can vary `names_from` values slowest (#839)", {
  df <- memdb_frame(
    name = c("name1", "name2"),
    value1 = c(1, 2),
    value2 = c(4, 5)
  )

  spec <- dbplyr_build_wider_spec(df, names_from = name, values_from = c(value1, value2))

  expect_identical(
    spec$.name,
    c("value1_name1", "value1_name2", "value2_name1", "value2_name2")
  )

  spec <- dbplyr_build_wider_spec(df, names_from = name, values_from = c(value1, value2), names_vary = "slowest")

  expect_identical(
    spec$.name,
    c("value1_name1", "value2_name1", "value1_name2", "value2_name2")
  )
})

test_that("`names_expand` does a cartesian expansion of `names_from` columns (#770)", {
  df <- memdb_frame(name1 = c("a", "b"), name2 = c("c", "d"), value = c(1, 2))
  spec <- dbplyr_build_wider_spec(df, names_from = c(name1, name2), names_expand = TRUE)
  expect_identical(spec$.name, c("a_c", "a_d", "b_c", "b_d"))
})


# checks arguments --------------------------------------------------------

test_that("`names_from` must be supplied if `name` isn't in `data` (#1240)", {
  df <- memdb_frame(key = "x", val = 1)
  expect_snapshot((expect_error(tidyr::pivot_wider(df, values_from = val))))
})

test_that("`values_from` must be supplied if `value` isn't in `data` (#1240)", {
  df <- memdb_frame(key = "x", val = 1)
  expect_snapshot((expect_error(tidyr::pivot_wider(df, names_from = key))))
})

test_that("`names_from` must identify at least 1 column (#1240)", {
  df <- memdb_frame(key = "x", val = 1)
  expect_snapshot(
    (expect_error(tidyr::pivot_wider(df, names_from = starts_with("foo"), values_from = val)))
  )
})

test_that("`values_from` must identify at least 1 column (#1240)", {
  df <- memdb_frame(key = "x", val = 1)
  expect_snapshot(
    (expect_error(tidyr::pivot_wider(df, names_from = key, values_from = starts_with("foo"))))
  )
})

#' @include backend-postgres.R
NULL

# nocov start
# RPostgreSQL connections use the postgres dialect
#' @export
dbplyr_edition.PostgreSQLConnection <- function(con) {
  2L
}

#' @export
sql_dialect.PostgreSQLConnection <- function(con) {
  dialect_postgres()
}

#' @export
db_connection_describe.PostgreSQLConnection <- db_connection_describe.PqConnection

#' @export
db_col_types.PostgreSQLConnection <- db_col_types.PqConnection
# nocov end

#' Override window order and frame
#'
#' These allow you to override the `PARTITION BY` and `ORDER BY` clauses
#' of window functions generated by grouped mutates.
#'
#' @inheritParams arrange.tbl_lazy
#' @param ... Variables to order by
#' @export
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(g = rep(1:2, each = 5), y = runif(10), z = 1:10)
#' db %>%
#'   window_order(y) %>%
#'   mutate(z = cumsum(y)) %>%
#'   show_query()
#'
#' db %>%
#'   group_by(g) %>%
#'   window_frame(-3, 0) %>%
#'   window_order(z) %>%
#'   mutate(z = sum(y)) %>%
#'   show_query()
window_order <- function(.data, ...) {
  if (!is_tbl_lazy(.data)) {
    msg <- "{.arg .data} must be a {.cls tbl_lazy}, not {.obj_type_friendly {(.data)}}."
    if (is.data.frame(.data)) {
      msg <- c(msg, i = "Did you mean to use {.fn arrange} instead?")
    }
    cli_abort(msg)
  }
  dots <- partial_eval_dots(.data, ..., .named = FALSE)
  names(dots) <- NULL
  dots <- check_window_order_dots(dots)

  .data$lazy_query <- add_order(.data, dots)
  .data
}

# We want to preserve this ordering (for window functions) without
# imposing an additional arrange, so we have a special op_order
add_order <- function(.data, dots) {
  .data$lazy_query$order_vars <- unname(dots)
  .data$lazy_query
}

check_window_order_dots <- function(dots, call = caller_env()) {
  for (i in seq_along(dots)) {
    x <- dots[[i]]
    if (is_quosure(x)) {
      x <- quo_get_expr(x)
    }

    if (is_call(x, "desc", n = 1)) {
      x <- call_args(x)[[1]]
    }

    dot <- dots[[i]]
    if (!is_symbol(x)) {
      dot <- as_label(dot)
      msg <- c(
        `!` = "Each element of {.code ...} must be a single column name or a column wrapped in {.fn desc}.",
        x = "Element {i} is {.code {dot}}."
      )
      cli_abort(msg, call = call)
    }

    dots[[i]] <- quo_get_expr(dot)
  }

  dots
}


# Frame -------------------------------------------------------------------

#' @export
#' @rdname window_order
#' @param from,to Bounds of the frame.
window_frame <- function(.data, from = -Inf, to = Inf) {
  if (!is_tbl_lazy(.data)) {
    cli_abort(
      "{.arg .data} must be a {.cls tbl_lazy}, not a {.cls {class(.data)}}."
    )
  }

  check_number_whole(from, allow_infinite = TRUE)
  check_number_whole(to, allow_infinite = TRUE)

  .data$lazy_query$frame <- list(range = c(from, to))
  .data
}

check_frame <- function(frame, call = caller_env()) {
  if (is.null(frame)) {
    return()
  }

  check_frame_range(frame$range)
}

check_frame_range <- function(range, call = caller_env()) {
  if (is.null(range)) {
    return()
  }

  vctrs::vec_assert(range, size = 2L, arg = "frame", call = call)
  check_number_whole(range[1], allow_infinite = TRUE, arg = "frame", call = call)
  check_number_whole(range[2], allow_infinite = TRUE, arg = "frame", call = call)
}

test_that("window_order errors for data frame", {
  expect_snapshot({
    (expect_error(window_order(data.frame(x = 1))))
    (expect_error(window_order("a")))
  })
})

test_that("window_order only accepts variables", {
  lf <- lazy_frame(x = 1, y = 1)
  expect_equal(window_order(lf, x, y) |> op_sort(), list(expr(x), expr(y)))
  expect_equal(
    window_order(lf, x, desc(y)) |> op_sort(),
    list(expr(x), expr(desc(y)))
  )

  expect_snapshot(error = TRUE, {
    window_order(lf, x + y)
    window_order(lf, foo())
    window_order(lf, desc(x + y))
  })
})

test_that("window order works afer renaming variable", {
  expect_snapshot({
    lazy_frame(x = 1, y = 1) |>
      window_order(y) |>
      rename(y2 = y) |>
      mutate(x_cum = cumsum(x))

    lazy_frame(x = 1, y = 1) |>
      rename(y2 = y) |>
      window_order(y2) |>
      mutate(x_cum = cumsum(x))
  })
})

test_that("window_frame errors for data frame", {
  expect_snapshot({
    (expect_error(window_frame(data.frame(x = 1))))
  })
})

test_that("window_order works with qualified desc()", {
  lf <- lazy_frame(y = runif(10), z = 1:10)

  out <- lf |>
    window_order(dplyr::desc(y)) |>
    mutate(z = cumsum(y))
  expect_snapshot(show_query(out))
})

test_that("filter captures local variables", {
  mf <- local_memdb_frame(x = 1:5, y = 5:1, id = 1:5)

  z <- 3
  expect_equal(mf |> filter(x > z) |> pull(), c(4, 5))
})

test_that("two filters equivalent to one", {
  unique_column_name_reset()
  mf <- local_memdb_frame("df", x = 1:5, y = 5:1, id = 1:5)

  df1 <- mf |> filter(x > 3) |> filter(y < 3)
  expect_equal(df1 |> pull(), c(4, 5))
  expect_snapshot(df1 |> show_query())

  df2 <- mf |> filter(mean(x, na.rm = TRUE) > 2) |> filter(y < 3)
  expect_equal(df2 |> pull(), c(4, 5))
  expect_snapshot(df2 |> show_query())
})

test_that("correctly inlines across all verbs", {
  lf <- lazy_frame(x = 1, y = 2)

  # single table verbs
  expect_selects(lf |> arrange(x) |> filter(x == 1), 1)
  expect_selects(lf |> distinct() |> filter(x == 1), 1)
  expect_selects(lf |> filter(x == 1) |> filter(y == 1), 1)
  expect_selects(lf |> head(1) |> filter(x == 1), 1)
  expect_selects(lf |> mutate(z = x + 1) |> filter(x == 1), 1)
  expect_selects(lf |> select(y = x) |> filter(y == 1), 1)
  expect_selects(lf |> summarise(y = mean(x)) |> filter(y == 1), 1)

  # two table verbs
  lf2 <- lazy_frame(x = 1)
  expect_selects(lf |> left_join(lf2, by = "x") |> filter(x == 1), 1)
  expect_selects(lf |> right_join(lf2, by = "x") |> filter(x == 1), 2)
  expect_selects(lf |> semi_join(lf2, by = "x") |> filter(x == 1), 3)
  expect_selects(lf |> union(lf2) |> filter(x == 1), 3)

  # special cases
  expect_selects(lf |> filter(mean(x) == 1), 2)
  expect_selects(lf |> mutate(x2 = mean(x)) |> filter(x == 1), 2)
  expect_selects(lf |> mutate(z = sql("x")) |> filter(z == 1), 2)
  expect_selects(lf |> mutate(z = x + 1) |> filter(z == 1), 2)

  # flag add_filter() for update if add_mutate ever gains inlining
  expect_selects(lf |> mutate(z = x + 1) |> filter(n() == 1), 3)
})

test_that("each argument gets implicit parens", {
  mf <- local_memdb_frame(
    v1 = c("a", "b", "a", "b"),
    v2 = c("b", "a", "a", "b"),
    v3 = c("a", "b", "c", "d"),
    id = 1:4
  )

  # want (v1 == "a" | v2 == "a") & v3 == "a"
  # NOT (v1 == "a" | v2 == "a" & v3 == "a")
  expect_equal(mf |> filter(v1 == "a" | v2 == "a", v3 == "a") |> pull(), 1)
})

test_that("only add step if necessary", {
  lf <- lazy_frame(x = 1:3, y = 1:3)
  expect_equal(lf |> filter(), lf)
})

test_that("errors for named input", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_snapshot(error = TRUE, filter(lf, x = 1))
  expect_snapshot(error = TRUE, filter(lf, y > 1, x = 1))
})

test_that(".preserve is not supported", {
  lf <- lazy_frame(x = 1:3, y = 1:3)
  expect_snapshot(error = TRUE, lf |> filter(x == 1, .preserve = TRUE))
})

test_that("filter() inlined after select()", {
  lf <- lazy_frame(x = 1, y = 2)

  out <- lf |>
    select(y) |>
    filter(y > 1)

  expect_equal(
    remote_query(out),
    sql("SELECT \"y\"\nFROM \"df\"\nWHERE (\"y\" > 1.0)")
  )

  out <- lf |>
    select(z = x) |>
    filter(z == 1)
  lq <- out$lazy_query
  expect_equal(lq$select$expr, list(sym("x")))
  expect_equal(lq$where, list(expr(x == 1)))
})

test_that("filter() inlined after mutate()", {
  lf <- lazy_frame(x = 1, y = 2)

  out <- lf |>
    mutate(x = x + 1) |>
    filter(y == 1)
  lq <- out$lazy_query
  expect_equal(lq$select$expr, list(expr(x + 1), expr(y)))
  expect_equal(lq$where, list(expr(y == 1)))

  # can rename variable used in `filter()`
  out <- lf |>
    rename(z = x) |>
    filter(z == 1)
  lq <- out$lazy_query
  expect_equal(lq$select$expr, list(sym("x"), sym("y")))
  expect_equal(lq$select$name, c("z", "y"))
  expect_equal(lq$where, list(expr(x == 1)))

  # does not rename field names of `$` that collide with renamed columns (#1812)
  ltr <- list(z = "A", x = "B")
  out <- lf |>
    rename(z = x) |>
    filter(z == ltr$z)
  expect_equal(out$lazy_query$where, list(expr(x == (!!ltr)$z)))

  # does not inline if uses mutated variable
  out2 <- lf |>
    mutate(x = x + 1) |>
    filter(x == 1)
  lq2 <- out2$lazy_query
  expect_equal(lq2$x$select$expr, list(expr(x + 1), sym("y")))
  expect_equal(lq2$select$expr, syms(c("x", "y")))
  expect_equal(lq2$where, list(expr(x == 1)))

  # does not inline if unclear whether uses mutated variable
  out3 <- lf |>
    mutate(x = x + 1) |>
    filter(y == sql("1"))
  lq3 <- out3$lazy_query
  expect_equal(lq3$select$expr, syms(c("x", "y")))
  expect_s3_class(lq3$x, "lazy_select_query")
  # sql("1") is evaluated during partial_eval
  expect_equal(lq3$where[[1]], expr(y == !!sql("1")))
})

test_that("filter isn't inlined after mutate with window function #1135", {
  lf <- lazy_frame(x = 1L, y = 1:2)
  out <- lf |>
    dplyr::mutate(z = sum(y, na.rm = TRUE)) |>
    dplyr::filter(y <= 1)

  lq <- out$lazy_query
  expect_equal(lq$select$expr, syms(c("x", "y", "z")))
  expect_equal(lq$where, list(expr(y <= 1)))
  expect_equal(lq$x$select$expr[[3]], expr(sum(y, na.rm = TRUE)))

  out2 <- lf |>
    dplyr::mutate(z = sql("SUM(y) OVER ()")) |>
    dplyr::filter(y <= 1)

  lq2 <- out2$lazy_query
  expect_equal(lq2$select$expr, syms(c("x", "y", "z")))
  expect_equal(lq2$where, list(expr(y <= 1)))
  # sql() is evaluated during partial_eval
  expect_equal(lq2$x$select$expr[[3]], sql("SUM(y) OVER ()"))
})

# .by -------------------------------------------------------------------------

test_that("can group transiently using `.by`", {
  df <- local_memdb_frame(g = c(1, 1, 2, 1, 2), x = c(5, 10, 1, 2, 3))

  out <- filter(df, x > mean(x), .by = g) |>
    arrange(g, x) |>
    collect()

  expect_identical(out$g, c(1, 2))
  expect_identical(out$x, c(10, 3))
  expect_equal(group_vars(out), character())
})

test_that("catches `.by` with grouped-df", {
  df <- lazy_frame(x = 1)
  gdf <- group_by(df, x)

  expect_snapshot(error = TRUE, {
    filter(gdf, .by = x)
  })
})

# SQL generation --------------------------------------------------------

test_that("filter calls windowed versions of sql functions", {
  df1 <- local_memdb_frame(x = 1:10, g = rep(c(1, 2), each = 5))

  out <- df1 |> group_by(g) |> filter(dplyr::row_number(x) < 3) |> collect()
  expect_equal(out$x, c(1L, 2L, 6L, 7L))
})

test_that("filter() can use window function and external vector - #1048", {
  to_filter <- 1:2
  expect_snapshot(
    lazy_frame(x = 1L) |>
      filter(x == max(x, na.rm = T), x %in% to_filter)
  )
})

test_that("recycled aggregates generate window function", {
  df1 <- local_memdb_frame(x = 1:10, g = rep(c(1, 2), each = 5))

  out <- df1 |>
    group_by(g) |>
    filter(x > mean(x, na.rm = TRUE)) |>
    collect()
  expect_equal(out$x, c(4L, 5L, 9L, 10L))
})

test_that("cumulative aggregates generate window function", {
  df1 <- local_memdb_frame(x = c(1:3, 2:4), g = rep(c(1, 2), each = 3))
  out <- df1 |>
    group_by(g) |>
    window_order(x) |>
    filter(cumsum(x) > 3)

  expect_equal(pull(out, x), c(3L, 3L, 4L))
})

test_that("filter() after summarise() uses `HAVING`", {
  lf <- lazy_frame(g = 1, h = 1, x = 1) |>
    group_by(g, h) |>
    summarise(x_mean = mean(x, na.rm = TRUE), .groups = "drop_last")
  mf <- local_memdb_frame(g = c(1, 1, 1, 2, 2), h = 1, x = 1:5) |>
    group_by(g, h) |>
    summarise(x_mean = mean(x, na.rm = TRUE), .groups = "drop_last")

  # use `HAVING`
  expect_snapshot((out <- lf |> filter(g == 1)))
  expect_equal(out$lazy_query$having, list(expr(g == 1)))
  expect_equal(out$lazy_query$group_by, list(sym("g"), sym("h")))
  expect_equal(op_grps(out), "g")

  expect_equal(
    mf |>
      filter(g == 1) |>
      collect(),
    tibble(g = 1, h = 1, x_mean = 2) |> group_by(g)
  )

  # Can use freshly aggregated column
  expect_snapshot((out <- lf |> filter(x_mean > 1)))
  expect_equal(out$lazy_query$having, list(expr(mean(x, na.rm = TRUE) > 1)))

  expect_equal(
    mf |>
      filter(x_mean > 3) |>
      collect(),
    tibble(g = 2, h = 1, x_mean = 4.5) |> group_by(g)
  )

  # multiple `filter()` combine instead of overwrite
  expect_snapshot(
    (out <- lf |>
      filter(g == 1) |>
      filter(g == 2))
  )
  expect_equal(out$lazy_query$having, list(expr(g == 1), expr(g == 2)))

  expect_snapshot(
    (out <- lf |>
      filter(g == 1) |>
      filter(h == 2))
  )
  expect_equal(out$lazy_query$having, list(expr(g == 1), expr(h == 2)))

  # `window_order()` and `window_frame()` do not matter
  out <- lazy_frame(g = 1, h = 1, x = 1) |>
    window_order(h) |>
    window_frame(-3) |>
    group_by(g, h) |>
    summarise(x_mean = mean(x, na.rm = TRUE), .groups = "drop_last") |>
    filter(x_mean > 1)

  lq <- out$lazy_query
  expect_equal(lq$having, list(expr(mean(x, na.rm = TRUE) > 1)))
  # TODO should the `order_vars` and the `frame` really survive `summarise()`?
  expect_equal(lq$order_vars, list(expr(h)))
  expect_equal(lq$frame, list(range = c(-3, Inf)))
})

test_that("`HAVING` supports expressions #1128", {
  lf <- lazy_frame(x = 1)

  expect_snapshot({
    lf |>
      summarise(x_sum = sum(x, na.rm = TRUE)) |>
      filter(!is.na(x_sum))
  })

  out <- lf |>
    summarise(x_sum = sum(x, na.rm = TRUE)) |>
    filter(!is.na(x_sum))
  expect_equal(out$lazy_query$having, list(expr(!is.na(sum(x, na.rm = TRUE)))))

  # correctly handles environments
  y <- 1L
  f <- function(lf, y = 2L) {
    lf |> summarise(x_sum = sum(x, na.rm = TRUE) - y)
  }

  out <- f(lf) |>
    filter(!is.na(x_sum + y))

  expect_equal(
    out$lazy_query$having,
    list(expr(!is.na(sum(x, na.rm = TRUE) - 2L + 1L)))
  )
})

test_that("filter() after mutate() does not use `HAVING`", {
  lf <- lazy_frame(g = 1, h = 1, x = 1) |>
    group_by(g, h) |>
    mutate(x_mean = mean(x, na.rm = TRUE))

  expect_snapshot((out <- lf |> filter(x_mean > 1)))
  lq <- out$lazy_query
  expect_s3_class(lq$x, "lazy_select_query")
})

test_that("filter() using a window function after summarise() does not use `HAVING`", {
  lf <- lazy_frame(g = 1, h = 1, x = 1) |>
    group_by(g, h) |>
    summarise(x_mean = mean(x, na.rm = TRUE), .groups = "drop_last")

  expect_snapshot((out <- lf |> filter(cumsum(x_mean) == 1)))
  lq <- out$lazy_query
  expect_s3_class(lq$x, "lazy_select_query")
})

# sql_build ---------------------------------------------------------------

test_that("filter generates simple expressions", {
  out <- lazy_frame(x = 1) |>
    filter(x > 1L) |>
    sql_build()

  expect_equal(out$where, sql('"x" > 1'))
})

# lazy_select_query -------------------------------------------------------

test_that("generates correct lazy_select_query", {
  lf <- lazy_frame(x = 1:3, y = 3:1)

  expect_equal(
    filter(lf, x > 1)$lazy_query,
    lazy_select_query(
      x = lf$lazy_query,
      select = syms(set_names(colnames(lf))),
      where = list(expr(x > 1))
    )
  )

  out <- lf |>
    filter(mean(x, na.rm = TRUE) > 1)

  expect_equal(
    out$lazy_query,
    lazy_select_query(
      x = out$lazy_query$x,
      select = syms(set_names(colnames(lf))),
      where = list(expr(col01 > 1))
    )
  )

  expect_equal(
    out$lazy_query$x,
    lazy_select_query(
      x = lf$lazy_query,
      select_operation = "mutate",
      select = list(
        x = sym("x"),
        y = sym("y"),
        col01 = expr(mean(x, na.rm = TRUE))
      )
    )
  )
})

# filter_out ------------------------------------------------------------------

test_that("filter_out keeps rows where the condition is FALSE or NA", {
  db <- memdb_frame(x = c(2, NA, 5, NA, 10), y = 1:5)
  out <- db |> filter_out(x < 5) |> collect()
  expect_equal(out$x, c(NA, 5, NA, 10))
  expect_equal(out$y, c(2L, 3L, 4L, 5L))
})

test_that("filter_out matches local dplyr semantics with multiple conditions", {
  local_df <- tibble(x = c(1, 2, NA), y = c(1, NA, 3))
  db <- memdb_frame(x = local_df$x, y = local_df$y)

  expect_equal(
    db |> filter_out(x == 1, y == 3) |> collect(),
    filter_out(local_df, x == 1, y == 3)
  )
})

test_that("filter_out uses backend-aware IS DISTINCT FROM TRUE", {
  filter_out_backend <- function(con) {
    lf <- lazy_frame(x = 1, y = 2, con = con)
    lf |> filter_out(x == 1, y == 3)
  }

  expect_snapshot({
    filter_out_backend(simulate_postgres())
    filter_out_backend(simulate_sqlite())
    filter_out_backend(simulate_dbi())
  })
})

test_that("filter_out with no conditions returns no rows", {
  lf <- lazy_frame(x = 1:3, y = 1:3)
  expect_equal(lf |> filter_out(), lf |> filter(FALSE))
})

test_that("filter_out errors for named input", {
  lf <- lazy_frame(x = 1, y = 2)
  expect_snapshot(error = TRUE, filter_out(lf, x = 1))
})

test_that("filter_out errors for .preserve = TRUE", {
  lf <- lazy_frame(x = 1:3, y = 1:3)
  expect_snapshot(error = TRUE, lf |> filter_out(x == 1, .preserve = TRUE))
})

test_that("filter_out supports .by", {
  db <- local_memdb_frame(g = c(1, 1, 2, 1, 2), x = c(5, 10, 1, 2, 3))
  out <- filter_out(db, x > mean(x, na.rm = TRUE), .by = g) |>
    arrange(g, x) |>
    collect()

  expect_equal(out$g, c(1, 1, 2))
  expect_equal(out$x, c(2, 5, 1))
})

test_that("filter_out captures local variables", {
  mf <- local_memdb_frame(x = 1:5, y = 5:1, id = 1:5)
  z <- 3
  expect_equal(mf |> filter_out(x > z) |> pull(), 1:3)
})

#' Subset the first rows
#'
#' @description
#' This is a method for the [head()] generic. It is usually translated to the
#' `LIMIT` clause of the SQL query. Because `LIMIT` is not an official part of
#' the SQL specification, some database use other clauses like `TOP` or
#' `FETCH ROWS`.
#'
#' Note that databases don't really have a sense of row order, so what "first"
#' means is subject to interpretation. Most databases will respect ordering
#' performed with `arrange()`, but it's not guaranteed. `tail()` is not
#' supported at all because the situation is even murkier for the "last" rows.
#'
#' @param x A lazy data frame backed by a database query.
#' @param n Number of rows to return
#' @param ... Not used.
#' @inherit arrange.tbl_lazy return
#' @export
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(x = 1:100)
#' db %>% head() %>% show_query()
#'
#' # Pretend we have data in a SQL server database
#' db2 <- lazy_frame(x = 1:100, con = simulate_mssql())
#' db2 %>% head() %>% show_query()
head.tbl_lazy <- function(x, n = 6L, ...) {
  if (!is.numeric(n) || length(n) != 1L || n < 0) {
    cli_abort("{.arg n} must be a non-negative integer")
  }
  n <- trunc(n)

  x$lazy_query <- add_head(x, n)
  x
}

add_head <- function(x, n) {
  lazy_query <- x$lazy_query
  if (!inherits(lazy_query, "lazy_select_query")) {
    lazy_query <- lazy_select_query(
      x = lazy_query,
      limit = n
    )

    return(lazy_query)
  }

  lazy_query$limit <- min(lazy_query$limit, n)
  lazy_query
}

#' @export
tail.tbl_lazy <- function(x, n = 6L, ...) {
  cli_abort("{.fun tail} is not supported by sql sources")
}

# table source = table id or sql

# table id =
# * interface to outside world; many ways to specify
# * always refers to exactly one table
# * but all converted to table name ASAP

# table path =
# * qualified table identifier (e.g. foo.bar.baz, bar.baz, baz)
# * always quoted
# * internal (and backend) use only; not user facing
# * can be vector containing multiple names
# * object names are always assumed to be table paths

# table_path --------------------------------------------------------------

table_path <- function(x) {
  check_character(x)
  x <- unname(x)
  structure(x, class = c("dbplyr_table_path", "character"))
}

# So you can do SQL(table_path("foo"))
setOldClass(c("dbplyr_table_path", "character"))


#' Table paths
#'
#' @description
#' dbplyr standardises all the ways of referring to a table (i.e. a single
#' string, a string wrapped in `I()`, a [DBI::Id()] and the results of
#' [in_schema()] and [in_catalog()]) into a table "path" of the form
#' `table`, `schema.table`, or `catalog.schema.path`. A table path is
#' always suitable for inlining into a query, so user input is quoted unless
#' it is wrapped in `I()`.
#'
#' This is primarily for internal usage, but you may need to work with it if
#' you're implementing a backend, and you need to compute with the table path,
#' not just pass it on unchanged to some other dbplyr function.
#'
#' * `is_table_path()` returns `TRUE` if the object is a `table_path`.
#' * `as_table_path()` coerces known table identifiers to a `table_path`.
#' * `check_table_path()` throws an error if the object is not a `table_path`.
#' * `table_path_name()` returns the last component of the table path (i.e.
#'   the name of the table).
#' * `table_path_components()` returns a list containing the components of each
#'    table path.
#'
#' A `table_path` object can technically be a vector of table paths, but
#' you will never see this in table paths constructed from user inputs.
#'
#' @keywords internal
#' @export
is_table_path <- function(x) {
  inherits(x, "dbplyr_table_path")
}

#' @export
print.dbplyr_table_path <- function(x, ...) {
  cat("<table_path> ", paste0(style_kw(x), collapse = ", "), "\n", sep = "")
}

#' @export
`[.dbplyr_table_path` <- function(x, ...) {
  table_path(NextMethod())
}

#' @export
`[[.dbplyr_table_path` <- function(x, ...) {
  table_path(NextMethod())
}

#' @export
`c.dbplyr_table_path` <- function(x, ...) {
  table_path(NextMethod())
}

make_table_path <- function(x, con, collapse = TRUE) {
  needs_quote <- !vapply(x, component_is_escaped, logical(1))

  x <- vapply(x, unclass, character(1))
  x[needs_quote] <- sql_escape_ident(con, x[needs_quote])
  if (collapse) {
    x <- paste0(x, collapse = ".")
  }

  table_path(x)
}

component_is_escaped <- function(x) {
  inherits(x, "AsIs") || is.sql(x) || inherits(x, "ident_q")
}

as_table_paths <- function(x, con) {
  make_table_path(x, con, collapse = FALSE)
}

#' @export
#' @rdname is_table_path
table_path_name <- function(x, con) {
  path <- as_table_path(x, con)
  components <- table_path_components(path, con)

  vapply(components, FUN.VALUE = character(1), function(x) {
    if (length(x) == 0) "" else x[[length(x)]]
  })
}

#' @export
#' @rdname is_table_path
table_path_components <- function(x, con) {
  UseMethod("table_path_components", con)
}

#' @export
table_path_components.default <- function(x, con) {
  quote_char <- substr(sql_escape_ident(con, ""), 1, 1)

  lapply(x, function(x) {
    scan(
      text = x,
      what = character(),
      quote = quote_char,
      quiet = TRUE,
      na.strings = character(),
      sep = "."
    )
  })
}

#' @export
escape.dbplyr_table_path <- function(x, parens = FALSE, collapse = ", ", con = NULL) {
  # names are always already escaped
  alias <- names2(x)
  table_path <- as_table_path(table_path_name(x, con), con)
  has_alias <- alias == "" | alias == table_path

  if (db_supports_table_alias_with_as(con)) {
    as_sql <- style_kw(" AS ")
  } else {
    as_sql <- " "
  }

  out <- ifelse(has_alias, unname(x), paste0(x, as_sql, alias))
  sql_vector(out, parens, collapse, con = con)
}

# table id ----------------------------------------------------------------

check_table_id <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (!is_table_id(x)) {
    stop_input_type(x, "a table identifier")
  }
}

is_table_id <- function(x) {
  is.ident(x) ||
    methods::is(x, "Id") ||
    is_catalog(x) ||
    is_schema(x) ||
    ((is.character(x) || is_table_path(x)) && length(x) == 1)
}

#' @export
#' @rdname is_table_path
check_table_path <- function(x,
                             error_arg = caller_arg(x),
                             error_call = caller_env()) {
  if (!is_table_path(x)) {
    cli::cli_abort(
      "{.arg {error_arg}} must be a <table_path>, not {.obj_type_friendly x}.",
      call = error_call,
      .internal = TRUE
    )
  }
}

#' @export
#' @rdname is_table_path
as_table_path <- function(x,
                          con,
                          error_arg = caller_arg(x),
                          error_call = caller_env()) {
  check_required(con)

  if (is_table_path(x)) {
    x
  } else if (is.sql(x)) {
    cli::cli_warn(
      c(
        "{.arg {error_arg}} uses SQL where a table identifier is expected.",
        i = "If you want to use a literal (unquoted) identifier use {.fn I} instead."
      )
    )
    table_path(unclass(x))
  } else if (inherits(x, "ident_q")) {
    table_path(paste0(x, collapse = "."))
  } else if (is.ident(x)) {
    make_table_path(unclass(x), con)
  } else if (methods::is(x, "Id")) {
    make_table_path(x@name, con)
  } else if (inherits(x, "dbplyr_catalog")) {
    make_table_path(list(x$catalog, x$schema, x$table), con)
  } else if (inherits(x, "dbplyr_schema")) {
    make_table_path(list(x$schema, x$table), con)
  } else if (inherits(x, "AsIs")) {
    check_string(unclass(x), allow_empty = FALSE, arg = error_arg, call = error_call)
    table_path(unclass(x))
  } else if (is.character(x)) {
    check_string(x, allow_empty = FALSE, arg = error_arg, call = error_call)
    make_table_path(x, con, collapse = FALSE)
  } else {
    cli::cli_abort(
      "{.arg {error_arg}} uses unknown specification for table name",
      error_call = error_call
    )
  }
}

# table source ------------------------------------------------------------

# Returns either SQL (representing a custom query) or a table name
as_table_source <- function(x, con, ..., error_arg = caller_arg(x), error_call = caller_env()) {
  if (is.sql(x)) {
    x
  } else if (is_table_id(x)) {
    as_table_path(x, con = con, error_arg = error_arg, error_call = error_call)
  } else {
    check_table_source(x, arg = error_arg, call = error_call)
  }
}

check_table_source <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (!is.sql(x) && !is_table_id(x)) {
    stop_input_type(x, "a table source (SQL or a table identifier)")
  }
}

test_that("table_path possess key methods", {
  expect_snapshot({
    name <- table_path(c("x", "y", "z"))
    name
  })

  expect_equal(name[c(1, 3)], table_path(c("x", "z")))
  expect_equal(name[[2]], table_path("y"))
  expect_equal(c(name[[1]], name[[2]]), table_path(c("x", "y")))
})

test_that("can check for table name", {
  foo <- function(y) check_table_path(y)
  expect_snapshot(foo(1), error = TRUE)
})

# as_table_path -----------------------------------------------------------

test_that("can coerce all user facing inputs", {
  con <- simulate_dbi()

  x_esc <- table_path("`x`")
  x_raw <- table_path("x")

  id <- table_path("x")
  expect_true(is_table_id(id))
  expect_equal(as_table_path(id, con), x_raw)

  id <- "x"
  expect_true(is_table_id(id))
  expect_equal(as_table_path(id, con), x_esc)

  id <- I("x")
  expect_true(is_table_id(id))
  expect_equal(as_table_path(id, con), x_raw)

  id <- ident("x")
  expect_true(is_table_id(id))
  expect_equal(as_table_path(id, con), x_esc)

  id <- ident_q("x")
  expect_true(is_table_id(id))
  expect_equal(as_table_path(id, con), x_raw)

  id <- DBI::Id(schema = "foo", table = "bar")
  expect_true(is_table_id(id))
  expect_equal(as_table_path(id, con), table_path("`foo`.`bar`"))

  # strip names, simulating DBI 1.2.0
  names(id@name) <- NULL
  expect_equal(as_table_path(id, con), table_path("`foo`.`bar`"))

  id <- in_schema("foo", "bar")
  expect_true(is_table_id(id))
  expect_equal(as_table_path(id, con), table_path("`foo`.`bar`"))

  id <- in_catalog("foo", "bar", "baz")
  expect_true(is_table_id(id))
  expect_equal(as_table_path(id, con), table_path("`foo`.`bar`.`baz`"))

  id <- in_catalog("foo", sql("bar"), ident_q("baz"))
  expect_true(is_table_id(id))
  expect_equal(as_table_path(id, con), table_path("`foo`.bar.baz"))
})

test_that("vectorised table_path is not a table_id", {
  expect_false(is_table_id(table_path(c("x", "y"))))
})

test_that("strips names", {
  con <- simulate_dbi()
  expect_equal(as_table_path(c(x = "x"), con), table_path("`x`"))

  id <- in_schema(c(x = "a"), "b")
  expect_equal(as_table_path(id, con), table_path("`a`.`b`"))

  id <- in_catalog("a", "b", c(x = "c"))
  expect_equal(as_table_path(id, con), table_path("`a`.`b`.`c`"))
})

test_that("as_table_path validates its inputs", {
  con <- simulate_dbi()
  expect_snapshot(error = TRUE, {
    as_table_path("x")
    as_table_path(c("x", "y"), con)
    as_table_path(1, con)
    as_table_path(I(1), con)
  })
})

test_that("as_table_path warns when using sql", {
  con <- simulate_dbi()
  expect_snapshot(as_table_path(sql("x"), con))
})

# components --------------------------------------------------------------

test_that("can parse components from path", {
  con <- simulate_dbi()

  expect_equal(
    table_path_components(table_path(c("x.y", "`x.y`.z", "`x.y`.`y.z`")), con),
    list(c("x", "y"), c("x.y", "z"), c("x.y", "y.z"))
  )
})

test_that("can extract names from path", {
  con <- simulate_dbi()

  expect_equal(
    table_path_name(table_path(c("x.y", "`x.y`.z", "x.`y.z`")), con),
    c("y", "z", "y.z")
  )
})

# nocov start
wrap <- function(..., indent = 0) {
  x <- paste0(..., collapse = "")
  wrapped <- strwrap(
    x,
    indent = indent,
    exdent = indent + 2,
    width = getOption("width")
  )

  paste0(wrapped, collapse = "\n")
}
# nocov end

indent <- function(x) {
  x <- paste0(x, collapse = "\n")
  paste0("  ", gsub("\n", "\n  ", x))
}

indent_print <- function(x) {
  indent(utils::capture.output(print(x)))
}

style_kw <- function(x) {
  if (!dbplyr_use_colour()) {
    return(x)
  }

  highlight <- dbplyr_highlight()
  if (is_false(highlight)) {
    return(x)
  }

  highlight(x)
}

# function for the thousand separator,
# returns "," unless it's used for the decimal point, in which case returns "."
'big_mark' <- function(x, ...) {
  mark <- if (identical(getOption("OutDec"), ",")) "." else ","
  formatC(x, big.mark = mark, ...)
}

dbplyr_use_colour <- function() {
  getOption("dbplyr_use_colour", FALSE)
}

dbplyr_highlight <- function() {
  highlight <- getOption("dbplyr_highlight", cli::combine_ansi_styles("blue"))

  if (is_true(highlight)) {
    highlight <- cli::combine_ansi_styles("blue")
  }

  if (is_false(highlight)) {
    return(FALSE)
  }

  if (!inherits(highlight, "cli_ansi_style")) {
    msg <- "{.envvar dbplyr_highlight} must be `NULL`, `FALSE` or a {.cls cli_ansi_style}."
    cli::cli_abort(msg)
  }

  highlight
}

#' Refer to a table in another schema/catalog
#'
#' @description
#' `in_schema()` and `in_catalog()` can be used to refer to tables outside of
#' the current catalog/schema. However, we now recommend using `I()` as it's
#' typically less typing.
#'
#' @param catalog,schema,table Names of catalog, schema, and table.
#'   These will be automatically quoted; use [sql()] to pass a raw name
#'   that won't get quoted.
#' @export
#' @keywords internal
#' @examples
#' # Previously:
#' in_schema("my_schema", "my_table")
#' in_catalog("my_catalog", "my_schema", "my_table")
#' in_schema(sql("my_schema"), sql("my_table"))
#'
#' # Now
#' I("my_schema.my_table")
#' I("my_catalog.my_schema.my_table")
#' I("my_schema.my_table")
#'
#' # Example using schemas with SQLite
#' con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#'
#' # Add auxiliary schema
#' tmp <- tempfile()
#' DBI::dbExecute(con, paste0("ATTACH '", tmp, "' AS aux"))
#'
#' library(dplyr, warn.conflicts = FALSE)
#' copy_to(con, iris, "df", temporary = FALSE)
#' copy_to(con, mtcars, I("aux.df"), temporary = FALSE)
#'
#' con |> tbl("df")
#' con |> tbl(I("aux.df"))
in_schema <- function(schema, table) {
  structure(
    list(
      schema = as_ident_or_sql(schema),
      table = as_ident_or_sql(table)
    ),
    class = "dbplyr_schema"
  )
}

#' @rdname in_schema
#' @export
in_catalog <- function(catalog, schema, table) {
  structure(
    list(
      schema = as_ident_or_sql(schema),
      table = as_ident_or_sql(table),
      catalog = as_ident_or_sql(catalog)
    ),
    class = "dbplyr_catalog"
  )
}

#' @export
format.dbplyr_schema <- function(x, ...) {
  con <- simulate_dbi()
  sql_glue2(simulate_dbi(), "{.tbl x$schema}.{.tbl x$table}")
}
#' @export
print.dbplyr_schema <- function(x, ...) {
  cat_line("<SCHEMA> ", format(x))
}

#' @export
format.dbplyr_catalog <- function(x, ...) {
  sql_glue2(simulate_dbi(), "{.tbl x$catalog}.{.tbl x$schema}.{.tbl x$table}")
}
#' @export
print.dbplyr_catalog <- function(x, ...) {
  cat_line("<CATALOG> ", format(x))
}

is_schema <- function(x) inherits(x, "dbplyr_schema")

is_catalog <- function(x) inherits(x, "dbplyr_catalog")

test_that("can create table idents", {
  expect_equal(
    new_table_ident(
      table = c("A", "B", "C"),
      schema = "schema"
    ),
    vctrs::new_rcrd(list(
      table = c("A", "B", "C"),
      schema = vctrs::vec_rep("schema", 3),
      catalog = vctrs::vec_rep(NA_character_, 3),
      quoted = vctrs::vec_rep(FALSE, 3),
      alias = vctrs::vec_rep(NA_character_, 3)
    ), class = "dbplyr_table_ident")
  )
})

test_that("is properly vectorised", {
  expect_equal(
    new_table_ident(
      table = c("A", "B"),
      schema = c("schema1", "schema2"),
      catalog = c("cat1", "cat2")
    ),
    vctrs::new_rcrd(list(
      table = c("A", "B"),
      schema = c("schema1", "schema2"),
      catalog = c("cat1", "cat2"),
      quoted = vctrs::vec_rep(FALSE, 2),
      alias = vctrs::vec_rep(NA_character_, 2)
    ), class = "dbplyr_table_ident")
  )

  expect_snapshot(error = TRUE, {
    new_table_ident(table = c("A", "B", "c"), schema = c("schema1", "schema2"))
  })
})

test_that("can't supply table and sql", {
  expect_snapshot(error = TRUE, {
    new_table_ident(schema = "my schema", table = "my table", quoted = TRUE)
  })
})

test_that("must supply table and schema when catalog is used", {
  expect_snapshot(error = TRUE, {
    new_table_ident(table = "my table", catalog = "cat")
    new_table_ident(schema = "schema", catalog = "cat")
  })

  # also works when schema is a vector
  expect_snapshot(error = TRUE, {
    new_table_ident(table = "my table", schema = c("my schema", NA), catalog = "cat")
  })
})

test_that("can't coerce or cast to character", {
  table <- new_table_ident(table = "table")
  expect_snapshot(error = TRUE, {
    c(table, "character")
    as.character(table)
  })
})

test_that("can print", {
  expect_snapshot({
    new_table_ident(table = "table")
    new_table_ident(schema = "schema", table = "table")
    new_table_ident(catalog = "catalog", schema = "schema", table = "table")
    new_table_ident(table = "`my schema`.`my table`", quoted = TRUE)
  })

  # is correctly vectorised
  expect_snapshot(
    new_table_ident(
      table   = c("`my schema`.`my table`", "table1",  "table2",   "table3"),
      schema  = c(                      NA,       NA, "schema2",  "schema3"),
      catalog = c(                      NA,       NA,        NA, "catalog3"),
      quoted  = c(                    TRUE,    FALSE,     FALSE,      FALSE)
    )
  )
})

test_that("as_table_ident works", {
  expect_equal(
    as_table_ident("table"),
    new_table_ident(table = "table")
  )

  expect_equal(
    as_table_ident(ident("table")),
    new_table_ident(table = "table")
  )

  expect_equal(
    as_table_ident(in_schema("schema", "table")),
    new_table_ident(schema = "schema", table = "table")
  )

  expect_equal(
    as_table_ident(in_catalog("catalog", "schema", "table")),
    new_table_ident(catalog = "catalog", schema = "schema", table = "table")
  )

  expect_equal(
    as_table_ident(DBI::Id(catalog = "catalog", schema = "schema", table = "table")),
    new_table_ident(catalog = "catalog", schema = "schema", table = "table")
  )

  table <- new_table_ident(table = "table")
  expect_equal(
    as_table_ident(table),
    table
  )

  withr::local_options(rlib_message_verbosity = "verbose")
  expect_snapshot({
    expect_equal(
      as_table_ident(ident_q("my schema.my table")),
      new_table_ident(table = "my schema.my table", quoted = TRUE)
    )
    expect_equal(
      as_table_ident(sql("my schema.my table")),
      new_table_ident(table = "my schema.my table", quoted = TRUE)
    )
  })
})

test_that("escaped as needed", {
  con <- simulate_dbi()
  out <- c(
    as_table_ident(ident("table1")),
    as_table_ident(in_schema("schema2", "table2")),
    as_table_ident(in_catalog("catalog3", "schema3", "table3"))
  )

  expect_equal(
    escape(out, collapse = NULL, con = simulate_dbi()),
    sql(
      "`table1`",
      "`schema2`.`table2`",
      "`catalog3`.`schema3`.`table3`"
    )
  )
})

test_that("can construct and print", {
  expect_snapshot(in_schema("schema", "table"))
  expect_snapshot(in_catalog("catalog", "schema", "table"))
})

test_that("can copy and collect with schema or Id", {
  con <- local_sqlite_con_with_aux()
  df <- tibble(x = 1:10)

  db <- copy_to(con, df, in_schema("aux", "db1"), temporary = FALSE)
  expect_equal(collect(db), df)
  expect_equal(collect(filter(db, x < 2)), df[1, ])

  db <- copy_to(
    con,
    df,
    DBI::Id(schema = "aux", table = "db2"),
    temporary = FALSE
  )
  expect_equal(collect(db), df)
  expect_equal(collect(filter(db, x < 2)), df[1, ])
})

#' JDBC backend
#'
#' @description
#' This backend supports databases accessed via `JDBCConnection` created by
#' `RJDBC::JDBC()` and [DBI::dbConnect()]. dbplyr automatically detects the
#' underlying database type from the JDBC connection class and uses the
#' appropriate SQL dialect.
#'
#' The following databases are recognized via their JDBC connection class:
#'
#' * PostgreSQL (`org.postgresql.*`): [dialect_postgres()]
#' * MySQL (`com.mysql.*`): [dialect_mysql()]
#' * MariaDB (`org.mariadb.*`): [dialect_mariadb()]
#' * SQLite (`org.sqlite.*`): [dialect_sqlite()]
#' * Oracle (`oracle.*`): [dialect_oracle()]
#' * SQL Server (`com.microsoft.sqlserver.*`): [dialect_mssql()]
#' * IBM DB2 (`com.ibm.db2.*`): [dialect_db2()]
#' * SAP HANA (`com.sap.db.*`): [dialect_hana()]
#' * Teradata (`com.teradata.*`): [dialect_teradata()]
#' * Apache Hive (`org.apache.hive.*`): [dialect_hive()]
#' * Apache Spark (`org.apache.spark.*` or `com.simba.spark.*`): [dialect_spark_sql()]
#' * Snowflake (`net.snowflake.*`): [dialect_snowflake()]
#' * Impala (`com.cloudera.impala.*`): [dialect_impala()]
#' * Amazon Redshift (`com.amazon.redshift.*`): [dialect_redshift()]
#'
#' If your database is not recognized, dbplyr will fall back to a generic ODBC
#' dialect. Please file an issue if you'd like support for additional databases.
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @seealso [with_dialect()] to use a different dialect if dbplyr guesses
#'   incorrectly, or a more specific translation is available.
#' @name backend-jdbc
#' @examples
#' # JDBC connections require the RJDBC package and a JDBC driver JAR file.
#' # Once connected, dbplyr automatically detects the database type:
#' #
#' # library(RJDBC)
#' # library(dplyr, warn.conflicts = FALSE)
#' #
#' # drv <- JDBC("org.postgresql.Driver", "postgresql.jar")
#' # con <- dbConnect(drv, "jdbc:postgresql://localhost/mydb", "user", "password")
#' # tbl(con, "my_table") |> filter(x > 1)
NULL

#' @export
sql_dialect.JDBCConnection <- function(con) {
  class_name <- jdbc_connection_class(con)
  jdbc_class_to_dialect(class_name)
}

#' @export
dbplyr_edition.JDBCConnection <- function(con) {
  2L
}

# Returns the JDBC connection class name
jdbc_connection_class <- function(con) {
  con@jc$getClass()$getName()
}

# Maps JDBC connection class name to the appropriate dialect
jdbc_class_to_dialect <- function(class_name) {
  if (startsWith(class_name, "org.postgresql.")) {
    dialect_postgres()
  } else if (startsWith(class_name, "org.mariadb.")) {
    # MariaDB (check before MySQL since MariaDB uses different driver)
    dialect_mariadb()
  } else if (startsWith(class_name, "com.mysql.")) {
    dialect_mysql()
  } else if (startsWith(class_name, "org.sqlite.")) {
    dialect_sqlite()
  } else if (startsWith(class_name, "oracle.")) {
    dialect_oracle()
  } else if (startsWith(class_name, "com.microsoft.sqlserver.")) {
    dialect_mssql()
  } else if (startsWith(class_name, "com.ibm.db2.")) {
    dialect_db2()
  } else if (startsWith(class_name, "com.sap.db.")) {
    dialect_hana()
  } else if (startsWith(class_name, "com.teradata.")) {
    dialect_teradata()
  } else if (startsWith(class_name, "org.apache.hive.")) {
    dialect_hive()
  } else if (
    startsWith(class_name, "org.apache.spark.") ||
      startsWith(class_name, "com.simba.spark.")
  ) {
    dialect_spark_sql()
  } else if (startsWith(class_name, "net.snowflake.")) {
    dialect_snowflake()
  } else if (
    startsWith(class_name, "com.cloudera.impala.") ||
      startsWith(class_name, "com.simba.impala.")
  ) {
    dialect_impala()
  } else if (startsWith(class_name, "com.amazon.redshift.")) {
    dialect_redshift()
  } else {
    # Fall back to generic ODBC dialect
    dialect_odbc()
  }
}

register_JDBC_dbExecute <- function(...) {
  # Protect against JDBC defined this
  if (methods::existsMethod(DBI::dbExecute, c("JDBCConnection", "character"))) {
    return()
  }
  methods::setMethod(
    DBI::dbExecute,
    c("JDBCConnection", "character"),
    function(conn, statement, ..., immediate = FALSE) {
      RJDBC::dbSendUpdate(conn, statement, ...)
    },
    where = globalenv()
  )
}

test_that("jdbc_class_to_dialect maps known databases", {
  expect_s3_class(
    jdbc_class_to_dialect("org.postgresql.jdbc.PgConnection"),
    "sql_dialect_postgres"
  )
  expect_s3_class(
    jdbc_class_to_dialect("org.mariadb.jdbc.Connection"),
    "sql_dialect_mariadb"
  )
  expect_s3_class(
    jdbc_class_to_dialect("com.mysql.cj.jdbc.ConnectionImpl"),
    "sql_dialect_mysql"
  )
  expect_s3_class(
    jdbc_class_to_dialect("org.sqlite.jdbc4.JDBC4Connection"),
    "sql_dialect_sqlite"
  )
  expect_s3_class(
    jdbc_class_to_dialect("oracle.jdbc.driver.T4CConnection"),
    "sql_dialect_oracle"
  )
  expect_s3_class(
    jdbc_class_to_dialect("com.microsoft.sqlserver.jdbc.SQLServerConnection"),
    "sql_dialect_mssql"
  )
  expect_s3_class(
    jdbc_class_to_dialect("com.ibm.db2.jcc.DB2Connection"),
    "sql_dialect_db2"
  )
  expect_s3_class(
    jdbc_class_to_dialect("com.sap.db.jdbc.HanaConnection"),
    "sql_dialect_hana"
  )
  expect_s3_class(
    jdbc_class_to_dialect("com.teradata.jdbc.TeraConnection"),
    "sql_dialect_teradata"
  )
  expect_s3_class(
    jdbc_class_to_dialect("org.apache.hive.jdbc.HiveConnection"),
    "sql_dialect_hive"
  )
  expect_s3_class(
    jdbc_class_to_dialect("org.apache.spark.sql.SparkSession"),
    "sql_dialect_spark_sql"
  )
  expect_s3_class(
    jdbc_class_to_dialect("com.simba.spark.jdbc.SparkConnection"),
    "sql_dialect_spark_sql"
  )
  expect_s3_class(
    jdbc_class_to_dialect("net.snowflake.client.jdbc.SnowflakeConnection"),
    "sql_dialect_snowflake"
  )
  expect_s3_class(
    jdbc_class_to_dialect("com.cloudera.impala.jdbc.ImpalaConnection"),
    "sql_dialect_impala"
  )
  expect_s3_class(
    jdbc_class_to_dialect("com.simba.impala.jdbc.ImpalaConnection"),
    "sql_dialect_impala"
  )
  expect_s3_class(
    jdbc_class_to_dialect("com.amazon.redshift.jdbc.RedshiftConnection"),
    "sql_dialect_redshift"
  )
})

test_that("jdbc_class_to_dialect falls back to ODBC for unknown databases", {
  expect_s3_class(
    jdbc_class_to_dialect("com.unknown.jdbc.Connection"),
    "sql_dialect_odbc"
  )
  expect_s3_class(
    jdbc_class_to_dialect("org.h2.jdbc.JdbcConnection"),
    "sql_dialect_odbc"
  )
  expect_s3_class(
    jdbc_class_to_dialect("org.hsqldb.jdbc.JDBCConnection"),
    "sql_dialect_odbc"
  )
})

#' @export
#' @rdname sql_variant
sql_paste <- function(default_sep, f = "CONCAT_WS") {

  function(..., sep = default_sep, collapse = NULL){
    check_collapse(collapse)
    sql_call2(f, sep, ...)
  }
}

#' @export
#' @rdname sql_variant
sql_paste_infix <- function(default_sep, op, cast) {
  force(default_sep)
  op <- as.symbol(paste0("%", op, "%"))
  force(cast)

  function(..., sep = default_sep, collapse = NULL){
    check_collapse(collapse)

    args <- list(...)
    if (length(args) == 1) {
      return(cast(args[[1]]))
    }

    if (sep == "") {
      infix <- function(x, y) sql_call2(op, x, y)
    } else {
      infix <- function(x, y) sql_call2(op, sql_call2(op, x, sep), y)
    }

    purrr::reduce(args, infix)
  }
}

check_collapse <- function(collapse) {
  if (is.null(collapse))
    return()

  cli_abort(c(
    "{.arg collapse} not supported in DB translation of {.fun paste}.",
    i = "Please use {.fun str_flatten} instead."
  ))
}

test_that("basic prefix operation", {
  local_con(simulate_dbi())

  paste <- sql_paste("")
  x <- ident("x")
  y <- ident("y")

  expect_equal(paste(x), sql("CONCAT_WS('', `x`)"))
  expect_equal(paste(x, y), sql("CONCAT_WS('', `x`, `y`)"))
  expect_equal(paste(x, y, sep = " "), sql("CONCAT_WS(' ', `x`, `y`)"))
})

test_that("basic infix operation", {
  local_con(simulate_dbi())

  paste <- sql_paste_infix("", "&&", function(x) sql_expr(cast((!!x) %as% text)))
  x <- ident("x")
  y <- ident("y")

  expect_equal(paste(x), sql("CAST(`x` AS text)"))
  expect_equal(paste(x, y), sql("`x` && `y`"))
  expect_equal(paste(x, y, sep = " "), sql("`x` && ' ' && `y`"))
})

#' ADBC backend
#'
#' @description
#' This backend supports databases accessed via `AdbiConnection` created by
#' `adbi::adbi()` and [DBI::dbConnect()]. dbplyr automatically detects the
#' underlying database type by querying the ADBC driver's vendor name and
#' uses the appropriate SQL dialect.
#'
#' The following vendors are recognized:
#'
#' * PostgreSQL and PostgreSQL-compatible systems (CrateDB, CockroachDB,
#'   Citus, Neon, ParadeDB, TimescaleDB, Yellowbrick, YugabyteDB, CedarDB):
#'   [dialect_postgres()]
#' * MySQL and MySQL-compatible systems (MariaDB, TiDB, Vitess): the
#'   appropriate dialect ([dialect_mariadb()] for MariaDB, [dialect_mysql()]
#'   otherwise)
#' * SQLite: [dialect_sqlite()]
#' * Microsoft SQL Server: [dialect_mssql()]
#' * Snowflake: [dialect_snowflake()]
#' * Amazon Redshift: [dialect_redshift()]
#'
#' If your database is not recognized, dbplyr will fall back to a generic ODBC
#' dialect. In this case, or if dbplyr guesses wrong, you can use
#' [with_dialect()] to choose a specific dialect.
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-adbc
#' @examples
#' # ADBC connections require the adbi package and an ADBC driver.
#' # Once connected, dbplyr automatically detects the database type:
#' #
#' # library(adbi)
#' # library(dplyr, warn.conflicts = FALSE)
#' #
#' # con <- DBI::dbConnect(adbi::adbi("adbcsqlite"), uri = ":memory:")
#' # tbl(con, "my_table") |> filter(x > 1)
NULL

#' @export
sql_dialect.AdbiConnection <- function(con) {
  vendor <- adbc_vendor_name(con)
  adbc_vendor_to_dialect(vendor)
}

#' @export
dbplyr_edition.AdbiConnection <- function(con) {
  2L
}

adbc_vendor_name <- function(con) {
  info <- as.list(adbcdrivermanager::adbc_connection_get_info(
    con@connection,
    info_codes = 0L # ADBC_INFO_VENDOR_NAME
  ))
  info$info_value$string_value
}

adbc_vendor_to_dialect <- function(vendor) {
  postgres <- dialect_postgres()
  mysql <- dialect_mysql()
  mssql <- dialect_mssql()
  redshift <- dialect_redshift()

  dialects <- list(
    sqlite = dialect_sqlite(),
    postgresql = postgres,
    cedardb = postgres,
    citus = postgres,
    cockroachdb = postgres,
    cratedb = postgres,
    neon = postgres,
    paradedb = postgres,
    timescaledb = postgres,
    yellowbrick = postgres,
    yugabytedb = postgres,
    mariadb = dialect_mariadb(),
    mysql = mysql,
    tidb = mysql,
    vitess = mysql,
    "microsoft sql server" = mssql,
    mssql = mssql,
    "sql server" = mssql,
    snowflake = dialect_snowflake(),
    redshift = redshift,
    "amazon redshift" = redshift
  )

  dialects[[tolower(vendor)]] %||% dialect_odbc()
}

#' @export
#' @rdname sql_build
lazy_multi_join_query <- function(
  x,
  joins,
  table_names,
  vars,
  distinct = FALSE,
  where = NULL,
  group_vars = op_grps(x),
  order_vars = op_sort(x),
  frame = op_frame(x),
  call = caller_env()
) {
  check_lazy_query(x, call = call)

  check_has_names(joins, c("table", "type", "by_x_table_id", "by"))
  check_character(joins$type, call = call)

  check_has_names(table_names, c("name", "from"), call = call)
  check_character(table_names$name, call = call)
  check_character(table_names$from, call = call)

  check_has_names(vars, c("name", "table", "var"), call = call)
  check_character(vars$name, call = call)
  check_integer(vars$table, call = call)
  check_character(vars$var, call = call)

  lazy_query(
    query_type = "multi_join",
    x = x,
    joins = joins,
    table_names = table_names,
    vars = vars,
    distinct = distinct,
    where = where,
    group_vars = group_vars,
    order_vars = order_vars,
    frame = frame
  )
}

#' @export
op_vars.lazy_multi_join_query <- function(op) {
  op$vars$name
}

#' @export
sql_build.lazy_multi_join_query <- function(op, con, ..., sql_options = NULL) {
  table_names_out <- generate_join_table_names(op$table_names, con)

  tables <- set_names(c(list(op$x), op$joins$table), table_names_out)
  table_vars <- purrr::map(tables, op_vars)
  select_sql <- sql_multi_join_select(
    con,
    op$vars,
    table_vars,
    use_star = sql_options$use_star,
    qualify_all_columns = sql_options$qualify_all_columns
  )

  op$joins$table <- purrr::map(
    op$joins$table,
    \(table) sql_build(table, con, sql_options = sql_options)
  )
  op$joins$by <- purrr::map2(
    op$joins$by,
    seq_along(op$joins$by),
    function(by, i) {
      by$x_as <- table_names_out[op$joins$by_x_table_id[[i]]]
      by$y_as <- table_names_out[i + 1L]
      by
    }
  )

  # WHERE occurs before SELECT so need to backtransform variable names with
  # qualified table.column references
  replacements <- purrr::map2(op$vars$var, op$vars$table, \(var, table_idx) {
    sql_table_prefix(con, table_names_out[[table_idx]], var)
  })
  where <- replace_sym(op$where, op$vars$name, replacements)
  where_sql <- translate_sql_(
    where,
    con = con,
    context = list(clause = "WHERE")
  )

  multi_join_query(
    x = sql_build(op$x, con, sql_options = sql_options),
    joins = op$joins,
    table_names = table_names_out,
    select = select_sql,
    where = where_sql,
    distinct = op$distinct
  )
}

generate_join_table_names <- function(table_names, con) {
  names <- table_path_name(table_names$name, con)
  table_name_length_max <- max(nchar(names))

  if (length(table_names$name) != 2) {
    table_names_repaired <- vctrs::vec_as_names(
      names,
      repair = "unique",
      quiet = TRUE
    )
    may_repair_name <- table_names$from != "as"
    names[may_repair_name] <- table_names_repaired[may_repair_name]
  } else {
    names <- join_two_table_alias(names, table_names$from)
  }

  # avoid database aliases exceeding the database-specific maximum length
  abbr_names <- abbreviate(
    names,
    # arbitrarily floor at identifier limit for Postgres backend to avoid unnecessarily truncating reasonable lengths
    # Source: https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
    # "By default, NAMEDATALEN is 64 so the maximum identifier length is 63 bytes."
    minlength = max(table_name_length_max, 63),
    # Explicitly set `strict = FALSE` (the default) to ensure table names are unique;
    # NB: non-zero (but low) chance that name is longer than DB limit
    strict = FALSE,
    named = FALSE,
    # Mitigation for non-zero chance of strings over limit:
    # don't over anchor to the start of the string,
    # since we opt to add qualifiers (...1, _{R/L}HS, etc.) to end of table name
    method = "both.sides"
  )

  as_table_paths(abbr_names, con)
}

# Built query -------------------------------------------------------------

multi_join_query <- function(
  x,
  joins,
  table_names,
  select,
  where = NULL,
  distinct = FALSE
) {
  query(
    "multi_join",
    x = x,
    joins = joins,
    table_names = table_names,
    select = select,
    where = where,
    distinct = distinct
  )
}

#' @export
sql_render.multi_join_query <- function(
  query,
  con = NULL,
  ...,
  sql_options = NULL,
  subquery = FALSE,
  lvl = 0
) {
  x <- sql_render(query$x, con, ..., subquery = TRUE, lvl = lvl + 1)
  query$joins$table <- purrr::map(
    query$joins$table,
    \(table) sql_render(table, con, ..., subquery = TRUE, lvl = lvl + 1)
  )

  sql_query_multi_join(
    con = con,
    x = x,
    joins = query$joins,
    table_names = query$table_names,
    by_list = query$by_list,
    select = query$select,
    where = query$where,
    distinct = query$distinct,
    lvl = lvl
  )
}

#' @export
flatten_query.multi_join_query <- function(qry, query_list, con) {
  x <- qry$x
  query_list_new <- flatten_query(x, query_list, con)
  qry$x <- get_subquery_name(x, query_list_new)

  for (i in vctrs::vec_seq_along(qry$joins)) {
    y <- qry$joins$table[[i]]
    query_list_new <- flatten_query(y, query_list_new, con)
    qry$joins$table[[i]] <- get_subquery_name(y, query_list_new)
  }

  # TODO reuse query
  name <- as_table_path(unique_subquery_name(), con)
  wrapped_query <- set_names(list(qry), name)

  query_list$queries <- c(query_list_new$queries, wrapped_query)
  query_list$name <- name
  query_list
}


# SQL generation ----------------------------------------------------------

#' @rdname db-sql
#' @export
sql_query_multi_join <- function(
  con,
  x,
  joins,
  table_names,
  by_list,
  select,
  where = NULL,
  ...,
  distinct = FALSE,
  lvl = 0
) {
  check_dots_used()
  UseMethod("sql_query_multi_join", sql_dialect(con))
}

#' @export
#' @param vars tibble with six columns:
#'   * `table` `<tbl_lazy>`: the tables to join with.
#'   * `type` `<character>`: the join type (left, right, inner, full).
#'   * `by_x`, `by_y` `<list_of<character>>`: The columns to join by
#'   * `by_x_table_id` `<list_of<integer>>`: The table index where the join column
#'     comes from. This needs to be a list because a the join columns might come
#'     from different tables
#'   * `on` `<character>`
#'   * `na_matches` `<character>`: Either `"na"` or `"never"`.
#' @param select A named SQL vector.
#' @param table_names `<character>` The names of the tables.
#' @noRd
#' @examples
#' # Left join with *
#' df1 <- lazy_frame(x = 1, y = 1)
#' df2 <- lazy_frame(x = 1, z = 1)
#' df3 <- lazy_frame(x = 1, z2 = 1)
#'
#' tmp <- left_join(df1, df2, by = "x") |>
#'   left_join(df3, by = c("x", z = "z2"))
#' tibble(
#'   table = list(df1, df2),
#'   type = c("left", "left"),
#'   by_x = list("x", c("x", "z")),
#'   by_y = list("x", c("x", "z2")),
#'   by_x_table_id = list(1L, c(1L, 2L)),
#'   on = c(NA, NA),
#'   na_matches = c("never", "never")
#' )
sql_query_multi_join.DBIConnection <- function(
  con,
  x,
  joins,
  table_names,
  by_list,
  select,
  where = NULL,
  ...,
  distinct = FALSE,
  lvl = 0
) {
  if (vctrs::vec_duplicate_any(table_names)) {
    cli_abort("{.arg table_names} must be unique.")
  }

  from <- dbplyr_sql_subquery(con, x, name = table_names[[1]], lvl = lvl)
  names <- table_names[-1]
  tables <- joins$table
  types <- toupper(paste0(joins$type, " JOIN"))

  n_joins <- length(types)
  out <- vector("list", n_joins * 2)

  for (i in seq_len(n_joins)) {
    table <- dbplyr_sql_subquery(con, tables[[i]], name = names[[i]], lvl = lvl)
    out[[2 * i - 1]] <- sql_clause(types[[i]], table)

    by <- joins$by[[i]]
    on <- sql_join_tbls(con, by = by, na_matches = by$na_matches)
    out[[2 * i]] <- sql_clause("ON", on, sep = " AND", parens = TRUE, lvl = 1)
  }

  clauses <- list2(
    sql_clause_select(select, distinct),
    sql_clause_from(from),
    !!!out,
    sql_clause_where(where)
  )
  sql_format_clauses(clauses, lvl = lvl)
}

#' @export
sql_query_multi_join.sql_dialect <- sql_query_multi_join.DBIConnection

# Helpers ----------------------------------------------------------------------

#' @param vars tibble with three columns:
#'   * `name` `<character>`: variable name in output.
#'   * `table` `<list_of<integer>>`: tables index in database.
#'   * `var` `<list_of<character>>`: variable names in database.
#'     If more than one, need to coalesce in output.
#' @param table_names `<character>`: table names indexed by `table`.
#' @param all_vars_list `<list_of<character>>` All variables in each table.
#' @noRd
sql_multi_join_select <- function(
  con,
  vars,
  table_vars,
  use_star = TRUE,
  qualify_all_columns = FALSE
) {
  all_vars <- tolower(unlist(table_vars, use.names = FALSE))
  if (qualify_all_columns) {
    duplicated_vars <- unique(all_vars)
  } else {
    duplicated_vars <- all_vars[vctrs::vec_duplicate_detect(all_vars)]
    duplicated_vars <- unique(duplicated_vars)
  }
  table_paths <- table_path(names(table_vars))

  out <- rep_named(vars$name, list())

  for (i in seq_along(table_paths)) {
    all_vars_i <- table_vars[[i]]
    vars_idx_i <- which(vars$table == i)
    used_vars_i <- vars$var[vars_idx_i]
    out_vars_i <- vars$name[vars_idx_i]

    if (
      use_star &&
        join_can_use_star(all_vars_i, used_vars_i, out_vars_i, vars_idx_i)
    ) {
      id <- vars_idx_i[[1]]
      out[[id]] <- sql_star(con, table_paths[i])
      names(out)[id] <- ""
    } else {
      out[vars_idx_i] <- purrr::map2(used_vars_i, i, function(var, table_idx) {
        sql_multi_join_var(
          con,
          var,
          table_paths[[table_idx]],
          duplicated_vars
        )
      })
    }
  }

  names_to_as(con, unlist(out))
}

join_can_use_star <- function(all_vars, used_vars, out_vars, idx) {
  # using `tbl.*` for a single variable is silly
  if (length(all_vars) <= 1) {
    return(FALSE)
  }

  # all variables need to be used in same order
  if (!identical(used_vars, all_vars)) {
    return(FALSE)
  }

  # they must not be renamed
  if (!identical(used_vars, out_vars)) {
    return(FALSE)
  }

  # the variables must form a sequence
  all(diff(idx) == 1)
}

sql_multi_join_var <- function(con, var, table_name, duplicated_vars) {
  if (tolower(var) %in% duplicated_vars) {
    sql_table_prefix(con, table_name, var)
  } else {
    sql_escape_ident(con, var)
  }
}

sql_join_tbls <- function(con, by, na_matches) {
  na_matches <- arg_match(na_matches, c("na", "never"))

  if (na_matches == "na" || length(by$x) + length(by$y) > 0) {
    lhs <- sql_table_prefix(con, by$x_as %||% "LHS", by$x)
    rhs <- sql_table_prefix(con, by$y_as %||% "RHS", by$y)

    by$condition[by$condition == "=="] <- "="
    join_params <- list(lhs = lhs, rhs = rhs, op = by$condition)

    if (na_matches == "na") {
      compare <- purrr::pmap_chr(join_params, function(lhs, rhs, op) {
        if (op == "=") {
          translate_sql(is_not_distinct_from(!!lhs, !!rhs), con = con)
        } else {
          sql_glue2(
            con,
            "({lhs} {.sql op} {rhs} OR ({lhs} IS NULL AND {rhs} IS NULL))"
          )
        }
      })
    } else {
      compare <- purrr::pmap_chr(join_params, function(lhs, rhs, op) {
        sql_glue2(con, "{lhs} {.sql op} {rhs}")
      })
    }

    sql(compare)
  } else if (length(by$on) > 0) {
    by$on
  }
}

sql_table_prefix <- function(con, table, var) {
  if (!is_bare_character(var)) {
    cli_abort("{.arg var} must be a bare character.", .internal = TRUE)
  }

  var <- sql_escape_ident(con, var)
  table <- sql_escape_ident(con, table_path_name(table, con))
  sql(paste0(table, ".", var))
}

sql_star <- function(con, table) {
  table <- table_path_name(table, con)
  sql_glue2(con, "{.id table}.*")
}

# For single-table SELECTs: use unqualified * unless backend requires prefix
sql_select_star <- function(con, table) {
  if (sql_has_star_table_prefix(con)) {
    sql_star(con, table)
  } else {
    sql("*")
  }
}

test_that("sql_build.lazy_multi_join_query() includes distinct", {
  lf1 <- lazy_frame(x = 1, y = 1)
  lf2 <- lazy_frame(x = 1, z = 2)

  out <- lf1 |>
    left_join(lf2, by = "x") |>
    distinct()

  query <- out$lazy_query
  expect_s3_class(query, "lazy_multi_join_query")
  built <- sql_build(out, dialect_ansi())
  expect_true(built$distinct)
})

test_that("sql_build.lazy_multi_join_query() includes where", {
  lf1 <- lazy_frame(x = 1, y = 1)
  lf2 <- lazy_frame(x = 1, z = 2)

  out <- lf1 |>
    left_join(lf2, by = "x") |>
    filter(y > 1, z < 5)

  query <- out$lazy_query
  expect_s3_class(query, "lazy_multi_join_query")
  expect_length(query$where, 2)

  built <- sql_build(out, dialect_ansi())
  expect_length(built$where, 2)
})

test_that("multi_join where clause uses qualified column names", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 3)

  expect_snapshot(left_join(lf1, lf2, by = "x") |> filter(y > 1))
  expect_snapshot(left_join(lf1, lf2, by = "x") |> filter(z > 1))
  expect_snapshot(left_join(lf1, lf2, by = "x") |> filter(y > 1, z < 5))

  lf3 <- lazy_frame(x = 1, y = 3)
  expect_snapshot(left_join(lf1, lf3, by = "x") |> filter(y.x > 1))
})

test_that("where clause is updated when vars are renamed by later join (#1770)", {
  lf1 <- lazy_frame(x = 1, y = 2, .name = "t1")
  lf2 <- lazy_frame(x = 1, z = 3, .name = "t2")
  lf3 <- lazy_frame(x = 1, z = 4, .name = "t3")

  # When z gets renamed to z.x, the filter should use the new name
  out <- lf1 |>
    left_join(lf2, by = "x") |>
    filter(z == 1) |>
    left_join(lf3, by = "x")

  expect_equal(out$lazy_query$where, list(expr(z.x == 1)))
})

test_that("generated sql doesn't change unexpectedly", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_snapshot(inner_join(lf, lf))
  expect_snapshot(left_join(lf, lf))
})

test_that("only disambiguates shared variables", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 2)
  expect_snapshot(left_join(lf1, lf2, by = c("x" = "x")))
  expect_snapshot(left_join(lf1, lf2, by = c("y" = "z")))
})

test_that("disambiguate variables that only differ in case", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(X = 1, y = 2)
  expect_snapshot(left_join(lf1, lf2, by = "y"))
})

test_that("sql_on query doesn't change unexpectedly", {
  lf1 <- lazy_frame(x = 1, y = 2)
  lf2 <- lazy_frame(x = 1, z = 3)

  expect_snapshot(inner_join(lf1, lf2, sql_on = "LHS.y < RHS.z"))
  expect_snapshot(left_join(lf1, lf2, sql_on = "LHS.y < RHS.z"))
})

test_that("sql_multi_join_select generates expected SQL", {
  con <- dialect_ansi()

  vars <- tibble(
    name = c("x", "a", "b"),
    var = c("x", "a", "b"),
    table = c(1L, 1L, 2L)
  )
  # left_join(lf(x, a), lf(x, b), by = "x")
  expect_equal(
    sql_multi_join_select(
      con,
      vars = vars,
      table_vars = list(left = c("x", "a"), right = c("x", "b"))
    ),
    sql('"left".*', '"b"')
  )

  # Left join with * - uses star for first table when all vars selected in order
  vars <- tibble(
    name = c("y", "a", "b", "c"),
    table = c(1L, 1L, 1L, 2L),
    var = c("y", "a", "b", "c")
  )
  expect_equal(
    sql_multi_join_select(
      con,
      vars = vars,
      table_vars = list(one = c("y", "a", "b"), two = c("y", "c"))
    ),
    sql('"one".*', '"c"')
  )

  # Left join with duplicated names - needs qualified column references
  vars <- tibble(
    name = c("y", "a.y", "a.x"),
    table = c(1L, 2L, 1L),
    var = c("x", "a", "a")
  )
  expect_equal(
    sql_multi_join_select(
      con,
      vars = vars,
      table_vars = list(one = c("x", "a"), two = c("x", "a"))
    ),
    sql('"one"."x" AS "y"', '"two"."a" AS "a.y"', '"one"."a" AS "a.x"')
  )
})

#' @importFrom R6 R6Class
NULL

Query <- R6::R6Class("Query",
  private = list(
    .vars = NULL
  ),
  public = list(
    con = NULL,
    sql = NULL,

    initialize = function(con, sql, vars) {
      self$con <- con
      self$sql <- sql
      private$.vars <- vars
    },

    # nocov start
    print = function(...) {
      cat("<Query> ", self$sql, "\n", sep = "")
      print(self$con)
    },

    fetch = function(n = -1L) {
      res <- dbSendQuery(self$con, self$sql)
      on.exit(dbClearResult(res))

      out <- dbFetch(res, n)
      res_warn_incomplete(res)
      out
    },
    # nocov end

    fetch_paged = function(chunk_size = 1e4, callback) {
      qry <- dbSendQuery(self$con, self$sql)
      on.exit(dbClearResult(qry))

      while (!dbHasCompleted(qry)) {
        chunk <- dbFetch(qry, chunk_size)
        callback(chunk)
      }

      invisible(TRUE)
    },

    # nocov start
    vars = function() {
      private$.vars
    },

    ncol = function() {
      length(self$vars())
    }
    # nocov end
  )
)

test_that("ungrouped data collected first", {
  out <- memdb_frame(x = 1:2) %>% do(head(.))
  expect_equal(out, tibble(x = 1:2))
})

test_that("named argument become list columns", {
  mf <- memdb_frame(
    g = rep(1:3, 1:3),
    x = 1:6
  ) %>% group_by(g)

  out <- mf %>% do(nrow = nrow(.), ncol = ncol(.))
  expect_equal(out$nrow, list(1, 2, 3))
  expect_equal(out$ncol, list(2, 2, 2))
})

test_that("unnamed results bound together by row", {
  mf <- memdb_frame(
    g = c(1, 1, 2, 2),
    x = c(3, 9, 4, 9)
  ) %>% group_by(g)

  first <- mf %>% do(head(., 1)) %>% ungroup()
  compare_tbl(first, tibble(g = c(1, 2), x = c(3, 4)))
})

test_that("unnamed results must be data frames", {
  mf <- memdb_frame(
    g = c(1, 1, 2, 2),
    x = c(3, 9, 4, 9)
  ) %>% group_by(g)

  expect_snapshot(error = TRUE, mf %>% do(nrow(.)))
})

test_that("Results respect select", {
  mf <- memdb_frame(
    g = c(1, 1, 2, 2),
    x = c(3, 9, 4, 9),
    y = 1:4,
    z = 4:1
  ) %>% group_by(g)

  expect_message(out <- mf %>% select(x) %>% do(ncol = ncol(.)))
  expect_equal(out$g, c(1, 2))
  expect_equal(out$ncol, list(2L, 2L))
})

test_that("results independent of chunk_size", {
  mf <- memdb_frame(
    g = rep(1:3, 1:3),
    x = 1:6
  ) %>% group_by(g)

  nrows <- function(group, n) {
    unlist(do(group, nrow = nrow(.), .chunk_size = n)$nrow)
  }

  expect_equal(nrows(mf, 1), c(1, 2, 3))
  expect_equal(nrows(mf, 2), c(1, 2, 3))
  expect_equal(nrows(mf, 10), c(1, 2, 3))
})

test_that("named argument become list columns", {
  mf <- memdb_frame(
    g = rep(1:3, 1:3),
    x = 1:6
  ) %>% group_by(g)

  # mix named and unnamed
  expect_snapshot(error = TRUE, mf %>% do(nrow = nrow(.), ncol(.)))

  # multiple unnamed
  expect_snapshot(error = TRUE, mf %>% do(nrow(.), ncol(.)))

  # old syntax
  expect_snapshot(error = TRUE, mf %>% do(.f = nrow))
})

#' @importFrom dplyr show_query
#' @export
show_query.tbl_lazy <- function(x, ..., cte = FALSE) {
  withr::local_options(list(dbplyr_use_colour = TRUE))
  sql <- remote_query(x, cte = cte)
  cat_line("<SQL>")
  cat_line(sql)
  invisible(x)
}

#' @importFrom dplyr explain
#' @export
explain.tbl_sql <- function(x, ...) {
  force(x)
  show_query(x)
  cat_line()
  cat_line("<PLAN>")
  cat_line(remote_query_plan(x, ...))

  invisible(x)
}

sql_cut <- function(x, breaks, labels = NULL, include.lowest = FALSE, right = TRUE) {
  breaks <- sort.int(as.double(breaks), na.last = TRUE)
  if (anyDuplicated(breaks)) {
    cli_abort("{.arg breaks} are not unique.")
  }
  n <- length(breaks)
  if (n == 1) {
    cli_abort("{.arg breaks} must have at least two values.")
  }
  if (any(is.na(breaks))) {
    cli_abort("{.arg breaks} values must not be missing.")
  }

  include.lowest <- vctrs::vec_cast(include.lowest, logical())
  vctrs::vec_assert(include.lowest, size = 1)
  labels <- check_cut_labels(labels, breaks, include.lowest, right)

  cases <- purrr::map(
    seq2(1, n - 2L),
    ~ {
      if (right) {
        expr(!!x <= !!breaks[[.x + 1]] ~ !!labels[[.x]])
      } else {
        expr(!!x < !!breaks[[.x + 1]] ~ !!labels[[.x]])
      }
    }
  )

  if (is.finite(breaks[[1]])) {
    if (right && !include.lowest) {
      first_case <- expr(!!x <= !!breaks[[1]] ~ NA)
    } else {
      first_case <- expr(!!x < !!breaks[[1]] ~ NA)
    }

    cases <- c(first_case, cases)
  }

  if (is.finite(breaks[[n]])) {
    if (right || include.lowest) {
      last_cases <- exprs(
        !!x <= !!breaks[[n]] ~ !!labels[[n - 1]],
        !!x > !!breaks[[n]] ~ NA
      )
    } else {
      last_cases <- exprs(
        !!x < !!breaks[[n]] ~ !!labels[[n - 1]],
        !!x >= !!breaks[[n]] ~ NA
      )
    }
  } else {
    if (right) {
      last_cases <- exprs(!!x > !!breaks[[n - 1]] ~ !!labels[[n - 1]])
    } else {
      last_cases <- exprs(!!x >= !!breaks[[n - 1]] ~ !!labels[[n - 1]])
    }
  }

  translate_sql(case_when(!!!cases, !!!last_cases))
}

check_cut_labels <- function(labels, breaks, include.lowest, right, call = caller_env()) {
  labels <- labels %||% TRUE

  n <- length(breaks)
  if (vctrs::vec_is(labels, character())) {
    return(vctrs::vec_recycle(labels, n - 1, x_arg = "labels", call = call))
  }

  vctrs::vec_assert(labels, size = 1L)
  if (!vctrs::vec_is(labels, logical()) || is.na(labels)) {
    cli_abort("{.arg labels} must be a bool or a character.", call = call)
  }

  if (is_false(labels)) {
    return(as.character(seq2(1, n)))
  }

  if (right) {
    labels <- paste0("(", breaks[seq2(1, n-1)], ",", breaks[seq2(2, n)], "]")
  } else {
    labels <- paste0("[", breaks[seq2(1, n-1)], ",", breaks[seq2(2, n)], ")")
  }

  if (include.lowest) {
    if (right) {
      substr(labels[[1]], 1, 1) <- "["
    } else {
      nchar <- nchar(labels[[n-1]], "chars")
      substr(labels[[n-1]], nchar, nchar) <- "]"
    }
  }

  labels
}

globalVariables(c("case_when"))

#' Oracle backend
#'
#' @description
#' This backend supports Oracle databases, typically accessed via
#' `OraConnection` created by [DBI::dbConnect()]. Use `dialect_oracle()` with
#' `lazy_frame()` to see simulated SQL without connecting to a live database.
#'
#' Key differences for this backend are:
#'
#' * Use `FETCH FIRST` instead of `LIMIT`
#' * Custom types
#' * `paste()` uses `||`
#' * Custom subquery generation (no `AS`)
#' * `setdiff()` uses `MINUS` instead of `EXCEPT`
#'
#' Note that versions of Oracle prior to 23c have limited supported for
#' `TRUE` and `FALSE` and you may need to use `1` and `0` instead.
#' See <https://oracle-base.com/articles/23/boolean-data-type-23> for
#' more details.
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-oracle
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_oracle())
#' lf |> transmute(x = paste0(c, " times"))
#' lf |> setdiff(lf)
NULL

#' @export
#' @rdname backend-oracle
dialect_oracle <- function() {
  new_sql_dialect(
    "oracle",
    quote_identifier = function(x) sql_quote(x, '"'),
    has_table_alias_with_as = FALSE,
    has_star_table_prefix = TRUE
  )
}

#' @export
#' @rdname backend-oracle
simulate_oracle <- function() simulate_dbi("Oracle")

#' @export
sql_dialect.Oracle <- function(con) {
  dialect_oracle()
}
#' @export
sql_dialect.OraConnection <- function(con) {
  dialect_oracle()
}

#' @export
dbplyr_edition.Oracle <- function(con) {
  2L
}

#' @export
sql_query_select.sql_dialect_oracle <- function(
  con,
  select,
  from,
  where = NULL,
  group_by = NULL,
  having = NULL,
  window = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  ...,
  subquery = FALSE,
  lvl = 0
) {
  sql_select_clauses(
    select = sql_clause_select(select, distinct),
    from = sql_clause_from(from),
    where = sql_clause_where(where),
    group_by = sql_clause_group_by(group_by),
    having = sql_clause_having(having),
    window = sql_clause_window(window),
    order_by = sql_clause_order_by(order_by, subquery, limit),
    # Requires Oracle 12c, released in 2013
    limit = if (!is.null(limit)) {
      limit <- as.integer(limit)
      sql_glue2(con, "FETCH FIRST {limit} ROWS ONLY")
    },
    lvl = lvl
  )
}

#' @export
sql_query_upsert.sql_dialect_oracle <- function(
  con,
  table,
  from,
  by,
  update_cols,
  ...,
  returning_cols = NULL,
  method = NULL
) {
  table <- as_table_path(table, con)

  method <- method %||% "merge"
  arg_match(method, c("merge", "cte_update"), error_arg = "method")
  if (method == "cte_update") {
    return(NextMethod("sql_query_upsert"))
  }

  # https://oracle-base.com/articles/9i/merge-statement
  parts <- rows_prep(con, table, from, by, lvl = 0)
  update_cols_esc <- sql_escape_ident(con, update_cols)
  update_values <- sql_table_prefix(con, "...y", update_cols)
  update_clause <- sql(paste0(update_cols_esc, " = ", update_values))

  insert_cols <- c(by, update_cols)
  insert_cols_esc <- sql_escape_ident(con, insert_cols)
  insert_values <- sql_table_prefix(con, "...y", insert_cols)

  table_sql <- sql_escape_table_source(con, table)
  clauses <- list(
    sql_clause("MERGE INTO", table_sql),
    sql_clause("USING", parts$from),
    sql_clause_on(parts$where, lvl = 1, parens = TRUE),
    sql("WHEN MATCHED THEN"),
    sql_clause("UPDATE SET", update_clause, lvl = 1),
    sql("WHEN NOT MATCHED THEN"),
    sql_clause_insert(insert_cols_esc, lvl = 1),
    sql_clause("VALUES", insert_values, parens = TRUE, lvl = 1),
    sql_returning_cols(con, returning_cols, table),
    sql(";")
  )
  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_translation.sql_dialect_oracle <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_odbc_scalar,
      # https://docs.oracle.com/cd/B19306_01/server.102/b14200/functions040.htm
      is_distinct_from = \(x, y) sql_glue("(decode({x}, {y}, 0, 1) = 1)"),
      is_not_distinct_from = \(x, y) sql_glue("(decode({x}, {y}, 0, 1) = 0)"),
      # Data type conversions are mostly based on this article
      # https://docs.oracle.com/cd/B19306_01/server.102/b14200/sql_elements001.htm

      # https://stackoverflow.com/questions/1171196
      as.character = sql_cast("VARCHAR2(255)"),
      # https://oracle-base.com/articles/misc/oracle-dates-timestamps-and-intervals
      as.Date = \(x) sql_glue("DATE {x}"),
      # bit64::as.integer64 can translate to BIGINT for some
      # vendors, which is equivalent to NUMBER(19) in Oracle
      # https://docs.oracle.com/cd/B19306_01/gateways.102/b14270/apa.htm
      as.integer64 = sql_cast("NUMBER(19)"),
      as.numeric = sql_cast("NUMBER"),
      as.double = sql_cast("NUMBER"),

      runif = function(n = n(), min = 0, max = 1) {
        sql_runif("DBMS_RANDOM.VALUE()", n = {{ n }}, min = min, max = max)
      },

      # string -----------------------------------------------------------------
      # https://docs.oracle.com/cd/B19306_01/server.102/b14200/operators003.htm#i997789
      paste = sql_paste_infix(" ", "||"),
      paste0 = sql_paste_infix("", "||"),
      str_c = sql_paste_infix("", "||"),

      # https://docs.oracle.com/en/database/oracle/oracle-database/19/sqlrf/REGEXP_REPLACE.html
      # 4th argument is starting position (default: 1 => first char of string)
      # 5th argument is occurrence (default: 0 => match all occurrences)
      str_replace = function(string, pattern, replacement) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement}, 1, 1)")
      },
      str_replace_all = function(string, pattern, replacement) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement})")
      },

      # lubridate --------------------------------------------------------------
      today = \() sql_glue("TRUNC(CURRENT_TIMESTAMP)"),
      now = \() sql("CURRENT_TIMESTAMP"),

      # clock ------------------------------------------------------------------
      add_days = function(x, n, ...) {
        check_dots_empty()
        sql_glue("({x} + NUMTODSINTERVAL({n}, 'day'))")
      },
      add_years = function(x, n, ...) {
        check_dots_empty()
        sql_glue("({x} + NUMTODSINTERVAL({n} * 365.25, 'day'))")
      },

      difftime = function(time1, time2, tz, units = "days") {
        if (!missing(tz)) {
          cli::cli_abort(
            "The {.arg tz} argument is not supported for SQL backends."
          )
        }

        if (units[1] != "days") {
          cli::cli_abort(
            'The only supported value for {.arg units} on SQL backends is "days"'
          )
        }

        sql_glue("CEIL(CAST({time2} AS DATE) - CAST({time1} AS DATE))")
      }
    ),
    base_odbc_agg,
    base_odbc_win
  )
}

#' @export
sql_query_explain.sql_dialect_oracle <- function(con, sql, ...) {
  # https://docs.oracle.com/en/database/oracle/oracle-database/19/tgsql/generating-and-displaying-execution-plans.html
  sql(
    sql_glue2(con, "EXPLAIN PLAN FOR {sql}"),
    sql_glue2(con, "SELECT PLAN_TABLE_OUTPUT FROM TABLE(DBMS_XPLAN.DISPLAY())")
  )
}

#' @export
sql_table_analyze.sql_dialect_oracle <- function(con, table, ...) {
  # Can't analyze private temporary tables
  if (is_oracle_temporary_table(table, con)) {
    return(NULL)
  }

  # https://docs.oracle.com/cd/B19306_01/server.102/b14200/statements_4005.htm
  sql_glue2(con, "ANALYZE TABLE {.tbl table} COMPUTE STATISTICS")
}

is_oracle_temporary_table <- function(table, con) {
  grepl("^ORA\\$PTT_", table_path_name(table, con), ignore.case = TRUE)
}

#' @export
sql_table_temporary.sql_dialect_oracle <- function(con, table, temporary, ...) {
  if (temporary && !is_oracle_temporary_table(table, con)) {
    new_name <- paste0("ORA$PTT_", table_path_name(table, con))
    cli::cli_inform(
      paste0("Created a temporary table named ", new_name),
      class = c("dbplyr_message_temp_table", "dbplyr_message")
    )
    table = table_path(new_name)
  }

  list(table = table, temporary = temporary)
}

#' @export
sql_query_save.sql_dialect_oracle <- function(
  con,
  sql,
  name,
  temporary = TRUE,
  ...
) {
  # Since db_table_temporary handles the prefix, `temporary` here is always
  # FALSE for temp tables (the name already has ORA$PTT_ prefix)

  # ON COMMIT PRESERVE ROWS creates a session-specific temporary table
  if (is_oracle_temporary_table(name, con)) {
    sql_glue2(
      con,
      "
      CREATE PRIVATE TEMPORARY TABLE {.tbl name}
      ON COMMIT PRESERVE DEFINITION
      AS
      {sql}
      "
    )
  } else {
    sql_glue2(con, "CREATE TABLE {.tbl name} AS\n{sql}")
  }
}

#' @export
db_table_drop_if_exists.Oracle <- function(con, table, ...) {
  # Private temporary tables don't appear in DBI::dbExistsTable() because
  # they're not in the normal catalog views, so we try to drop directly
  if (is_oracle_temporary_table(table, con)) {
    try(DBI::dbRemoveTable(con, DBI::SQL(table)), silent = TRUE)
  } else if (DBI::dbExistsTable(con, DBI::SQL(table))) {
    DBI::dbRemoveTable(con, DBI::SQL(table))
  }
}

#' @export
dbplyr_write_table.Oracle <- function(
  con,
  table,
  types,
  values,
  temporary = TRUE,
  ...,
  overwrite = FALSE
) {
  if (overwrite) {
    db_table_drop_if_exists(con, table)
  }

  # We can't use DBI::dbWriteTable() here because it doesn't support
  # Oracle's PRIVATE TEMPORARY TABLE syntax
  fields <- types %||% values
  create_sql <- oracle_sql_table_create(con, table, fields)
  DBI::dbExecute(con, create_sql)

  DBI::dbAppendTable(con, DBI::SQL(table), values)

  table
}

oracle_sql_table_create <- function(con, table, fields) {
  # Convert data frame to field types (like DBI:::sqlCreateTable_DBIConnection)
  if (is.data.frame(fields)) {
    fields <- vapply(fields, function(x) DBI::dbDataType(con, x), character(1))
  }

  # Quote column names and build field definitions
  field_names <- sql_escape_ident(con, names(fields))
  field_types <- unname(fields)
  fields_sql <- sql(paste0(field_names, " ", field_types))

  # Use PRIVATE TEMPORARY TABLE for temp tables (detected by ORA$PTT_ prefix)
  if (is_oracle_temporary_table(table, con)) {
    sql_glue2(
      con,
      "CREATE PRIVATE TEMPORARY TABLE {.tbl table}
      ({fields_sql})
      ON COMMIT PRESERVE DEFINITION"
    )
  } else {
    sql_glue2(
      con,
      "CREATE TABLE {.tbl table} 
      ({fields_sql})"
    )
  }
}

#' @export
sql_values_subquery.sql_dialect_oracle <- function(
  con,
  df,
  types,
  lvl = 0,
  ...
) {
  sql_values_subquery_union(con, df, types = types, lvl = lvl, from = "DUAL")
}

#' @export
sql_set_op_method.sql_dialect_oracle <- function(con, op, ...) {
  # Oracle uses MINUS instead of EXCEPT:
  # https://docs.oracle.com/cd/B19306_01/server.102/b14200/queries004.htm
  switch(op, "EXCEPT" = "MINUS", op)
}

# roacle package ----------------------------------------------------------

#' @export
dbplyr_edition.OraConnection <- dbplyr_edition.Oracle

#' @export
dbplyr_write_table.OraConnection <- dbplyr_write_table.Oracle

#' @export
db_table_drop_if_exists.OraConnection <- db_table_drop_if_exists.Oracle

test_that("simulate_oracle() still works", {
  expect_translation(simulate_oracle(), x + 1, '"x" + 1.0')
})

test_that("custom scalar functions translated correctly", {
  con <- dialect_oracle()

  expect_translation(
    con,
    as.character(x),
    'CAST("x" AS VARCHAR2(255))'
  )
  expect_translation(
    con,
    as.integer64(x),
    'CAST("x" AS NUMBER(19))'
  )
  expect_translation(con, as.double(x), 'CAST("x" AS NUMBER)')
  expect_translation(con, as.Date(x), 'DATE "x"')
  expect_translation(
    con,
    as.Date("2023-01-01"),
    "DATE '2023-01-01'"
  )
})

test_that("paste and paste0 translate correctly", {
  con <- dialect_oracle()

  expect_translation(con, paste(x, y), '"x" || \' \' || "y"')
  expect_translation(con, paste0(x, y), '"x" || "y"')
  expect_translation(con, str_c(x, y), '"x" || "y"')
})


test_that("string functions translate correctly", {
  con <- dialect_oracle()

  expect_snapshot({
    translate_sql(str_replace(col, "pattern", "replacement"), con = con)
    translate_sql(str_replace_all(col, "pattern", "replacement"), con = con)
  })
})

test_that("queries translate correctly", {
  mf <- lazy_frame(x = 1, con = dialect_oracle())
  expect_snapshot(mf |> head())
})

test_that("`sql_query_upsert()` is correct", {
  con <- dialect_oracle()
  df_y <- lazy_frame(
    a = 2:3,
    b = c(12L, 13L),
    c = -(2:3),
    d = c("y", "z"),
    con = con,
    .name = "df_y"
  ) |>
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_upsert(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      by = c("a", "b"),
      update_cols = c("c", "d"),
      returning_cols = c("a", b2 = "b"),
      method = "merge"
    )
  )
})

test_that("db_table_temporary adds ORA$PTT_ prefix", {
  con <- dialect_oracle()

  # Adds prefix (with message) for temporary tables
  expect_snapshot(
    result <- sql_table_temporary(con, table_path("tbl"), temporary = TRUE)
  )
  expect_equal(as.character(result$table), "ORA$PTT_tbl")
  expect_true(result$temporary)

  # Doesn't double-prefix if already has ORA$PTT_
  result <- sql_table_temporary(
    con,
    table_path("ORA$PTT_tbl"),
    temporary = TRUE
  )
  expect_equal(as.character(result$table), "ORA$PTT_tbl")
  expect_true(result$temporary)

  # Returns table unchanged for non-temporary
  result <- sql_table_temporary(con, table_path("tbl"), temporary = FALSE)
  expect_equal(as.character(result$table), "tbl")
  expect_false(result$temporary)
})

test_that("generates custom sql", {
  con <- dialect_oracle()

  expect_snapshot(sql_table_analyze(con, in_schema("schema", "tbl")))
  # Can't analyze private temporary tables
  expect_null(sql_table_analyze(con, table_path("ORA$PTT_tbl")))

  expect_snapshot(sql_query_explain(con, sql("SELECT * FROM foo")))

  lf <- lazy_frame(x = 1, con = con)
  expect_snapshot(left_join(lf, lf, by = "x", na_matches = "na"))

  # With ORA$PTT_ prefix -> creates PRIVATE TEMPORARY TABLE
  expect_snapshot(sql_query_save(con, sql("SELECT * FROM foo"), "ORA$PTT_tbl"))
  # Without ORA$PTT_ prefix -> creates regular TABLE
  expect_snapshot(sql_query_save(
    con,
    sql("SELECT * FROM foo"),
    "tbl",
    temporary = FALSE
  ))

  expect_snapshot(slice_sample(lf, n = 1))
})

test_that("oracle_sql_table_create generates correct SQL", {
  con <- dialect_oracle()

  expect_snapshot({
    # Temporary table (has ORA$PTT_ prefix)
    oracle_sql_table_create(
      con,
      table_path("ORA$PTT_test"),
      c(x = "INTEGER", y = "TEXT")
    )

    # Regular table
    oracle_sql_table_create(
      con,
      table_path("test"),
      c(x = "INTEGER", y = "TEXT")
    )
  })
})

test_that("copy_inline uses UNION ALL", {
  con <- dialect_oracle()
  y <- tibble::tibble(id = 1L, arr = "{1,2,3}")

  types <- c(id = "bigint", arr = "integer[]")
  expect_snapshot({
    copy_inline(con, y |> slice(0)) |> remote_query()
    copy_inline(con, y) |> remote_query()

    # with `types`
    copy_inline(con, y |> slice(0), types = types) |> remote_query()
    copy_inline(con, y, types = types) |> remote_query()
  })
})

test_that("custom clock functions translated correctly", {
  con <- dialect_oracle()
  expect_translation(
    con,
    add_years(x, 1),
    '("x" + NUMTODSINTERVAL(1.0 * 365.25, \'day\'))'
  )
  expect_translation(
    con,
    add_days(x, 1),
    '("x" + NUMTODSINTERVAL(1.0, \'day\'))'
  )
  expect_error(
    translate_sql(add_days(x, 1, "dots", "must", "be empty"), con = con),
    class = "rlib_error_dots_nonempty"
  )
})

test_that("difftime is translated correctly", {
  con <- dialect_oracle()
  expect_translation(
    con,
    difftime(start_date, end_date, units = "days"),
    'CEIL(CAST("end_date" AS DATE) - CAST("start_date" AS DATE))'
  )
  expect_translation(
    con,
    difftime(start_date, end_date),
    'CEIL(CAST("end_date" AS DATE) - CAST("start_date" AS DATE))'
  )

  expect_snapshot(
    error = TRUE,
    translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      difftime(
        start_date,
        end_date,
        tz = "UTC",
        units = "days"
      ),
      con = con
    )
  )
})

#' @export
#' @rdname sql_build
lazy_set_op_query <- function(x, y, type, all, call = caller_env()) {
  check_lazy_query(x, call = call)
  check_lazy_query(y, call = call)
  check_string(type, call = call)
  check_bool(all, call = call)

  if (all) {
    type <- paste0(type, " ALL")
  }

  lazy_query(
    query_type = "set_op",
    x = x,
    y = y,
    type = type
  )
}

#' @export
op_vars.lazy_set_op_query <- function(op) {
  union(op_vars(op$x), op_vars(op$y))
}

#' @export
sql_build.lazy_set_op_query <- function(op, con, ..., sql_options = NULL) {
  # add_op_set_op() ensures that both have same variables
  set_op_query(
    sql_build(op$x, con, sql_options = sql_options),
    sql_build(op$y, con, sql_options = sql_options),
    type = op$type
  )
}

# Built query -------------------------------------------------------------

#' @export
#' @rdname sql_build
set_op_query <- function(x, y, type) {
  query("set_op", x = x, y = y, type = type)
}

#' @export
sql_render.set_op_query <- function(
  query,
  con = NULL,
  ...,
  sql_options = NULL,
  subquery = FALSE,
  lvl = 0
) {
  sub_lvl <- lvl + !inherits(con, "SQLiteConnection")
  from_x <- sql_render(query$x, con, ..., subquery = FALSE, lvl = sub_lvl)
  from_y <- sql_render(query$y, con, ..., subquery = FALSE, lvl = sub_lvl)

  sql_query_set_op(
    con,
    from_x,
    from_y,
    method = query$type,
    lvl = lvl
  )
}

#' @export
flatten_query.set_op_query <- function(qry, query_list, con) {
  flatten_query_2_tables(qry, query_list, con)
}

# SQL generation ----------------------------------------------------------

#' @rdname db-sql
#' @export
sql_set_op_method <- function(con, op, ...) {
  UseMethod("sql_set_op_method", sql_dialect(con))
}
#' @export
sql_set_op_method.DBIConnection <- function(con, op, ...) {
  op
}
#' @export
sql_set_op_method.sql_dialect <- sql_set_op_method.DBIConnection

#' @rdname db-sql
#' @export
sql_query_set_op <- function(con, x, y, method, ..., lvl = 0) {
  check_dots_used()
  UseMethod("sql_query_set_op", sql_dialect(con))
}
#' @export
sql_query_set_op.DBIConnection <- function(con, x, y, method, ..., lvl = 0) {
  method <- sql_set_op_method(con, method)
  method <- style_kw(method)
  lines <- list(
    sql_indent_subquery(x, con = con, lvl = lvl),
    sql(method),
    sql_indent_subquery(y, con = con, lvl = lvl)
  )
  sql_format_clauses(lines, lvl)
}

#' @export
sql_query_set_op.sql_dialect <- sql_query_set_op.DBIConnection

#' Literal SQL escaping
#'
#' Use `sql()` to declare that a string is literal SQL and should be used
#' as is, without quoting.
#'
#' @param ... Character vectors that will be combined into a single SQL vector.
#' @param x Object to check if it is an sql object.
#' @export
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#' # sql() just adds a class
#' sql("x + 1")
#' is.sql(sql("x + 1"))
#'
#' # You can use it when you need to insert some literal SQL in a query
#' db <- memdb_frame(x = 1:3)
#' db |> mutate(y = sql("CAST(x as VARCHAR)"))
sql <- function(...) {
  x <- c_character(...)
  structure(x, class = c("sql", "character"))
}

#' @export
`[.sql` <- function(x, i) {
  sql(NextMethod())
}
#' @export
`[[.sql` <- function(x, i) {
  sql(NextMethod())
}

#' @rdname sql
#' @export
is.sql <- function(x) inherits(x, "sql")

#' @export
print.sql <- function(x, ...) cat(format(x, ...), sep = "\n")
#' @export
format.sql <- function(x, ...) {
  if (length(x) == 0) {
    paste0("<SQL> [empty]")
  } else {
    paste0("<SQL> ", x, ifelse(names2(x) == "", "", paste0(" AS ", names2(x))))
  }
}

test_that("sql() creates sql object from character inputs", {
  expect_s3_class(sql("x"), "sql")
  expect_equal(sql("x", "y"), sql(c("x", "y")))
})

test_that("sql() preserves names", {
  x <- sql(a = "x", b = "y")
  expect_equal(names(x), c("a", "b"))
})

test_that("sql() is idempotent", {
  expect_equal(sql(sql("x")), sql("x"))
  expect_equal(sql(sql("x"), sql("y")), sql("x", "y"))
})

test_that("is.sql() checks for sql class", {
  expect_true(is.sql(sql("x")))
  expect_false(is.sql("x"))
  expect_false(is.sql(1))
})

test_that("subsetting preserves sql class", {
  x <- sql("a", "b", "c")
  expect_s3_class(x[1], "sql")
  expect_s3_class(x[1:2], "sql")
  expect_s3_class(x[[1]], "sql")
})

test_that("can format sql", {
  expect_snapshot(sql())
  expect_snapshot(sql(a = "x", "y"))
})

test_that("collect equivalent to as.data.frame/as_tibble", {
  mf <- local_memdb_frame(letters = letters)

  expect_equal(as.data.frame(mf), data.frame(letters, stringsAsFactors = FALSE))
  expect_equal(as_tibble(mf), tibble(letters))
  expect_equal(collect(mf), tibble(letters))
})

test_that("explicit collection returns all data", {
  n <- 1e5 + 10 # previous default was 1e5
  big <- local_memdb_frame(x = seq_len(n))

  nrow1 <- big |> as.data.frame() |> nrow()
  nrow2 <- big |> as_tibble() |> nrow()
  nrow3 <- big |> collect() |> nrow()

  expect_equal(nrow1, n)
  expect_equal(nrow2, n)
  expect_equal(nrow3, n)
})

test_that("collect() handles DBI error", {
  mf <- local_memdb_frame(x = 1)
  expect_snapshot(
    (expect_error(mf |> mutate(a = sql("invalid sql")) |> collect())),
    transform = snap_transform_dbi
  )
})

test_that("sql_quote() returns empty sql() for empty input", {
  expect_equal(sql_quote(character(), "'"), sql())
})

test_that("sql_quote() converts NA to NULL", {
  expect_equal(sql_quote(NA_character_, "'"), sql("NULL"))
  expect_equal(sql_quote(c("x", NA), "'"), sql("'x'", "NULL"))
})

test_that("sql_quote() preserves names", {
  expect_equal(sql_quote(c(a = "x", b = "y"), "'"), sql(a = "'x'", b = "'y'"))
})

test_that("sql_quote() doubles quote characters in input", {
  expect_equal(sql_quote("it's", "'"), sql("'it''s'"))
  expect_equal(sql_quote("a]b", c("[", "]")), sql("[a]]b]"))
})

test_that("sql_quote() checks its inputs", {
  expect_snapshot(error = TRUE, {
    sql_quote(1)
    sql_quote("x", 1)
    sql_quote("x", character())
  })
})

test_that("check_sql() checks inputs", {
  f <- function(x) check_sql(x)

  expect_snapshot(f("x"), error = TRUE)
  expect_no_error(f(sql("x")))
})

test_that("check_sql() optionally checks names", {
  f <- function(x) check_sql(x, allow_names = FALSE)

  expect_no_error(f(sql("a")))
  expect_snapshot(f(sql(a = "x")), error = TRUE)
})

test_that("check_sql() optionally allows NULL", {
  f <- function(x) check_sql(x, allow_null = TRUE)

  expect_no_error(f(NULL))
  expect_snapshot(f("x"), error = TRUE)
})

sql_lines <- function(...) {
  sql(paste0(c(...), collapse = "\n"))
}

test_that("can translate cut", {
  con <- dialect_ansi()
  expect_translation(
    con,
    cut(x, 1:2),
    c(
      "CASE",
      "WHEN (\"x\" <= 1.0) THEN NULL",
      "WHEN (\"x\" <= 2.0) THEN '(1,2]'",
      "WHEN (\"x\" > 2.0) THEN NULL",
      "END"
    )
  )

  expect_translation(
    con,
    cut(x, 1:3),
    c(
      "CASE",
      "WHEN (\"x\" <= 1.0) THEN NULL",
      "WHEN (\"x\" <= 2.0) THEN '(1,2]'",
      "WHEN (\"x\" <= 3.0) THEN '(2,3]'",
      "WHEN (\"x\" > 3.0) THEN NULL",
      "END"
    )
  )
})

test_that("works with include.lowest = TRUE", {
  con <- dialect_ansi()
  expect_translation(
    con,
    cut(x, 1:3, include.lowest = TRUE),
    c(
      "CASE",
      "WHEN (\"x\" < 1.0) THEN NULL",
      "WHEN (\"x\" <= 2.0) THEN '[1,2]'",
      "WHEN (\"x\" <= 3.0) THEN '(2,3]'",
      "WHEN (\"x\" > 3.0) THEN NULL",
      "END"
    )
  )
})

test_that("works with right = FALSE", {
  con <- dialect_ansi()
  expect_translation(
    con,
    cut(x, 1:2, right = FALSE),
    c(
      "CASE",
      "WHEN (\"x\" < 1.0) THEN NULL",
      "WHEN (\"x\" < 2.0) THEN '[1,2)'",
      "WHEN (\"x\" >= 2.0) THEN NULL",
      "END"
    )
  )
})

test_that("works with right = FALSE and include.lowest = TRUE", {
  con <- dialect_ansi()
  expect_translation(
    con,
    cut(x, 1:2, right = FALSE, include.lowest = TRUE),
    c(
      "CASE",
      "WHEN (\"x\" < 1.0) THEN NULL",
      "WHEN (\"x\" <= 2.0) THEN '[1,2]'",
      "WHEN (\"x\" > 2.0) THEN NULL",
      "END"
    )
  )
})

test_that("works with labels = FALSE", {
  con <- dialect_ansi()
  expect_translation(
    con,
    cut(x, 1:3, labels = FALSE),
    c(
      "CASE",
      "WHEN (\"x\" <= 1.0) THEN NULL",
      "WHEN (\"x\" <= 2.0) THEN '1'",
      "WHEN (\"x\" <= 3.0) THEN '2'",
      "WHEN (\"x\" > 3.0) THEN NULL",
      "END"
    )
  )
})

test_that("works with labels a character vector", {
  con <- dialect_ansi()
  expect_translation(
    con,
    cut(x, 1:3, labels = c("a", "b")),
    c(
      "CASE",
      "WHEN (\"x\" <= 1.0) THEN NULL",
      "WHEN (\"x\" <= 2.0) THEN 'a'",
      "WHEN (\"x\" <= 3.0) THEN 'b'",
      "WHEN (\"x\" > 3.0) THEN NULL",
      "END"
    )
  )

  expect_translation_snapshot(
    con,
    cut(x, 1:3, labels = c("a", "b", "c")),
    error = TRUE
  )
})

test_that("can handle infinity", {
  con <- dialect_ansi()
  expect_translation(
    con,
    cut(x, c(-Inf, 0, 1, Inf)),
    c(
      "CASE",
      "WHEN (\"x\" <= 0.0) THEN '(-Inf,0]'",
      "WHEN (\"x\" <= 1.0) THEN '(0,1]'",
      "WHEN (\"x\" > 1.0) THEN '(1,Inf]'",
      "END"
    )
  )

  expect_translation(
    con,
    cut(x, c(-Inf, 0, 1, Inf), right = FALSE),
    c(
      "CASE",
      "WHEN (\"x\" < 0.0) THEN '[-Inf,0)'",
      "WHEN (\"x\" < 1.0) THEN '[0,1)'",
      "WHEN (\"x\" >= 1.0) THEN '[1,Inf)'",
      "END"
    )
  )

  expect_translation(
    con,
    cut(x, c(-Inf, 0, 1, Inf), include.lowest = TRUE),
    c(
      "CASE",
      "WHEN (\"x\" <= 0.0) THEN '[-Inf,0]'",
      "WHEN (\"x\" <= 1.0) THEN '(0,1]'",
      "WHEN (\"x\" > 1.0) THEN '(1,Inf]'",
      "END"
    )
  )
})

test_that("cut checks arguments", {
  con <- dialect_ansi()
  expect_translation_snapshot(con, cut(x, 1), error = TRUE)
  expect_translation_snapshot(con, cut(x, c(1, 1)), error = TRUE)
  expect_translation_snapshot(con, cut(x, c(1, 2, NA)), error = TRUE)
})

#' SQL Server backend
#'
#' @description
#' This backend supports Microsoft SQL Server, typically accessed via odbc. Use
#' `dialect_mssql()` with `lazy_frame()` to see simulated SQL without connecting
#' to a live database.
#'
#' Key differences for this backend are:
#'
#' - `SELECT` uses `TOP` not `LIMIT`
#' - Automatically prefixes `#` to create temporary tables. Add the prefix
#'   yourself to avoid the message.
#' - String basics: `paste()`, `substr()`, `nchar()`
#' - Custom types for `as.*` functions
#' - Lubridate extraction functions, `year()`, `month()`, `day()` etc
#' - Semi-automated bit <-> boolean translation (see below)
#' - stringr functions `str_detect()`, `str_starts()`, `str_ends()` with
#'   `fixed()` patterns work on all versions; regular expression patterns
#'   require SQL Server 2025+ (version 17.0)
#' - stringr functions `str_replace()`, `str_replace_all()`, `str_remove()`,
#'   `str_remove_all()`, `str_extract()`, and `str_count()` require SQL Server
#'   2025+ (version 17.0)
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @section Bit vs boolean:
#' SQL server uses two incompatible types to represent `TRUE` and `FALSE`
#' values:
#'
#' * The `BOOLEAN` type is the result of logical comparisons (e.g. `x > y`)
#'   and can be used `WHERE` but not to create new columns in `SELECT`.
#'   <https://learn.microsoft.com/en-us/sql/t-sql/language-elements/comparison-operators-transact-sql>
#'
#' * The `BIT` type is a special type of numeric column used to store
#'   `TRUE` and `FALSE` values, but can't be used in `WHERE` clauses.
#'   <https://learn.microsoft.com/en-us/sql/t-sql/data-types/bit-transact-sql?view=sql-server-ver15>
#'
#' dbplyr does its best to automatically create the correct type when needed,
#' but can't do it 100% correctly because it does not have a full type
#' inference system. This means that you many need to manually do conversions
#' from time to time.
#'
#' * To convert from bit to boolean use `x == 1`
#' * To convert from boolean to bit use `as.logical(if(x, 0, 1))`
#'
#' @param version Version of MS SQL to simulate. Currently, 11.0 and above
#'   will use `TRY_CAST()` instead of `CAST()`, and 17.0 and above will
#'   support regular expression patterns in stringr functions.
#' @name backend-mssql
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_mssql())
#' lf |> head()
#' lf |> transmute(x = paste(b, c, d))
#'
#' # Can use boolean as is:
#' lf |> filter(c > d)
#' # Need to convert from boolean to bit:
#' lf |> transmute(x = c > d)
#' # Can use boolean as is:
#' lf |> transmute(x = ifelse(c > d, "c", "d"))
NULL

#' @export
#' @rdname backend-mssql
dialect_mssql <- function(version = "15.0") {
  dialect <- new_sql_dialect(
    "mssql",
    quote_identifier = function(x) sql_quote(x, c("[", "]"))
  )
  dialect$version <- numeric_version(version)
  dialect
}

#' @export
#' @rdname backend-mssql
simulate_mssql <- function(version = "15.0") {
  simulate_dbi("Microsoft SQL Server", version = numeric_version(version))
}

#' @export
`sql_dialect.Microsoft SQL Server` <- function(con) {
  version <- mssql_version(con)
  dialect_mssql(version = version)
}

#' @export
`dbplyr_edition.Microsoft SQL Server` <- function(con) {
  2L
}

#' @export
`table_path_components.Microsoft SQL Server` <- function(x, con) {
  mssql_table_path_components(x)
}
#' @export
table_path_components.sql_dialect_mssql <- function(x, con) {
  mssql_table_path_components(x)
}

mssql_table_path_components <- function(x) {
  # SQL Server uses asymmetric quotes [identifier], which scan() can't handle
  # Also support standard SQL double-quote quoting "identifier"
  lapply(x, function(path) {
    # Split on . but not inside [] or ""
    # Pattern: match [anything], "anything", or non-dot characters
    matches <- gregexpr('\\[[^]]*\\]|"[^"]*"|[^.]+', path)
    components <- regmatches(path, matches)[[1]]
    # Remove the brackets or quotes from each component
    gsub('^\\[|\\]$|^"|"$', "", components)
  })
}

#' @export
sql_query_select.sql_dialect_mssql <- function(
  con,
  select,
  from,
  where = NULL,
  group_by = NULL,
  having = NULL,
  window = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  ...,
  subquery = FALSE,
  lvl = 0
) {
  sql_select_clauses(
    select = sql_clause_select(select, distinct, top = limit),
    from = sql_clause_from(from),
    where = sql_clause_where(where),
    group_by = sql_clause_group_by(group_by),
    having = sql_clause_having(having),
    window = sql_clause_window(window),
    order_by = sql_clause_order_by(order_by, subquery, limit),
    lvl = lvl
  )
}

#' @export
sql_query_insert.sql_dialect_mssql <- function(
  con,
  table,
  from,
  insert_cols,
  by,
  ...,
  conflict = c("error", "ignore"),
  returning_cols = NULL,
  method = NULL
) {
  table <- as_table_path(table, con)

  # https://stackoverflow.com/questions/25969/insert-into-values-select-from
  conflict <- rows_check_conflict(conflict)

  check_character(returning_cols, allow_null = TRUE)

  check_string(method, allow_null = TRUE)
  method <- method %||% "where_not_exists"
  arg_match(method, "where_not_exists", error_arg = "method")

  parts <- rows_insert_prep(con, table, from, insert_cols, by, lvl = 0)

  clauses <- list2(
    parts$insert_clause,
    sql_returning_cols(con, returning_cols, "INSERTED"),
    sql_clause_select(sql("*")),
    sql_clause_from(parts$from),
    !!!parts$conflict_clauses
  )

  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_append.sql_dialect_mssql <- function(
  con,
  table,
  from,
  insert_cols,
  ...,
  returning_cols = NULL
) {
  table <- as_table_path(table, con)
  parts <- rows_prep(con, table, from, by = list(), lvl = 0)

  insert_cols_sql <- sql_escape_ident(con, insert_cols)
  table_sql <- sql_escape_table_source(con, table)
  clauses <- list2(
    sql_clause_insert(insert_cols_sql, into = table_sql),
    sql_returning_cols(con, returning_cols, "INSERTED"),
    sql_clause_select(sql("*")),
    sql_clause_from(parts$from)
  )

  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_update_from.sql_dialect_mssql <- function(
  con,
  table,
  from,
  by,
  update_values,
  ...,
  returning_cols = NULL
) {
  table <- as_table_path(table, con)

  # https://stackoverflow.com/a/2334741/946850
  parts <- rows_prep(con, table, from, by, lvl = 0)
  update_cols <- sql_escape_ident(con, names(update_values))

  table_sql <- sql_escape_table_source(con, table)
  clauses <- list(
    sql_clause_update(table_sql),
    sql_clause_set(update_cols, update_values),
    sql_returning_cols(con, returning_cols, "INSERTED"),
    sql_clause_from(table_sql),
    sql_clause("INNER JOIN", parts$from),
    sql_clause_on(parts$where, lvl = 1)
  )
  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_upsert.sql_dialect_mssql <- function(
  con,
  table,
  from,
  by,
  update_cols,
  ...,
  returning_cols = NULL,
  method = NULL
) {
  table <- as_table_path(table, con)

  check_string(method, allow_null = TRUE)
  method <- method %||% "merge"
  arg_match(method, "merge", error_arg = "method")

  parts <- rows_prep(con, table, from, by, lvl = 0)

  update_cols_esc <- sql_escape_ident(con, update_cols)
  update_values <- sql_table_prefix(con, "...y", update_cols)
  update_clause <- sql(paste0(update_cols_esc, " = ", update_values))

  insert_cols <- c(by, update_cols)
  insert_cols_esc <- sql_escape_ident(con, insert_cols)
  insert_cols_qual <- sql_table_prefix(con, "...y", insert_cols)

  table_sql <- sql_escape_table_source(con, table)
  clauses <- list(
    sql_clause("MERGE INTO", table_sql),
    sql_clause("USING", parts$from),
    sql_clause_on(parts$where, lvl = 1),
    sql("WHEN MATCHED THEN"),
    sql_clause("UPDATE SET", update_clause, lvl = 1),
    sql("WHEN NOT MATCHED THEN"),
    sql_clause_insert(insert_cols_esc, lvl = 1),
    sql_clause("VALUES", insert_cols_qual, parens = TRUE, lvl = 1),
    sql_returning_cols(con, returning_cols, "INSERTED"),
    sql(";")
  )
  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_delete.sql_dialect_mssql <- function(
  con,
  table,
  from,
  by,
  ...,
  returning_cols = NULL
) {
  table <- as_table_path(table, con)
  parts <- rows_prep(con, table, from, by, lvl = 0)

  table_sql <- sql_escape_table_source(con, table)
  clauses <- list2(
    sql_clause("DELETE FROM", table_sql),
    sql_returning_cols(con, returning_cols, table = "DELETED"),
    !!!sql_clause_where_exists(parts$from, parts$where, not = FALSE)
  )
  sql_format_clauses(clauses, lvl = 0)
}

mssql_scalar_base <- function() {
  sql_translator(
    .parent = base_odbc_scalar,

    `!` = function(x) {
      if (mssql_needs_bit()) {
        x <- with_mssql_bool(x)
        sql_glue("~{mssql_as_bit(x)}")
      } else {
        sql_glue("NOT({x})")
      }
    },

    `!=` = mssql_infix_comparison("!="),
    `==` = mssql_infix_comparison("="),
    `<` = mssql_infix_comparison("<"),
    `<=` = mssql_infix_comparison("<="),
    `>` = mssql_infix_comparison(">"),
    `>=` = mssql_infix_comparison(">="),

    `&` = mssql_infix_boolean("&", "AND"),
    `&&` = mssql_infix_boolean("&", "AND"),
    `|` = mssql_infix_boolean("|", "OR"),
    `||` = mssql_infix_boolean("|", "OR"),

    `[` = function(x, i) {
      i <- with_mssql_bool(i)
      sql_glue("CASE WHEN ({i}) THEN ({x}) END")
    },

    bitwShiftL = sql_not_supported("bitwShiftL"),
    bitwShiftR = sql_not_supported("bitwShiftR"),

    `if` = function(condition, true, false = NULL, missing = NULL) {
      mssql_sql_if(
        enquo(condition),
        enquo(true),
        enquo(false),
        enquo(missing)
      )
    },
    if_else = function(condition, true, false, missing = NULL) {
      check_required(true)
      check_required(false)

      mssql_sql_if(
        enquo(condition),
        enquo(true),
        enquo(false),
        enquo(missing)
      )
    },
    ifelse = function(test, yes, no) {
      check_required(yes)
      check_required(no)

      mssql_sql_if(enquo(test), enquo(yes), enquo(no))
    },
    case_when = mssql_case_when,
    between = function(x, left, right) {
      context <- sql_current_context()
      if (context$clause == "WHERE") {
        sql_glue("{x} BETWEEN {left} AND {right}")
      } else {
        sql_glue("CASE WHEN {x} BETWEEN {left} AND {right} THEN 1 ELSE 0 END")
      }
    },

    as.logical = sql_cast("BIT"),

    as.Date = sql_cast("DATE"),
    as.numeric = sql_cast("FLOAT"),
    as.double = sql_cast("FLOAT"),
    as.character = sql_cast("VARCHAR(MAX)"),
    log = sql_prefix("LOG"),
    atan2 = sql_prefix("ATN2"),
    ceil = sql_prefix("CEILING"),
    ceiling = sql_prefix("CEILING"),

    is.null = mssql_is_null,
    is.na = mssql_is_null,

    runif = function(n = n(), min = 0, max = 1) {
      # https://stackoverflow.com/a/9039661
      sql_runif("RAND(CHECKSUM(NEWID()))", n = {{ n }}, min = min, max = max)
    },

    # string functions ------------------------------------------------
    nchar = sql_prefix("LEN"),
    paste = sql_paste_infix(" ", "+"),
    paste0 = sql_paste_infix("", "+"),
    substr = sql_substr("SUBSTRING"),
    substring = sql_substr("SUBSTRING"),

    # stringr functions
    str_length = sql_prefix("LEN"),
    str_c = sql_paste_infix("", "+"),
    # use COLLATE to ensure consistent behaviour https://learn.microsoft.com/en-us/sql/relational-databases/collations/set-or-change-the-column-collation?view=sql-server-ver17&source=recommendations
    str_like = function(string, pattern, ignore_case = deprecated()) {
      ignore_case <- deprecate_ignore_case(ignore_case)
      if (ignore_case) {
        sql_glue("{string} COLLATE Latin1_General_100_CI_AS LIKE {pattern}")
      } else {
        sql_glue("{string} COLLATE Latin1_General_100_CS_AS LIKE {pattern}")
      }
    },
    str_ilike = function(string, pattern) {
      sql_glue("{string} COLLATE Latin1_General_100_CI_AS LIKE {pattern}")
    },
    # Fixed pattern support for all SQL Server versions (regex requires 2025+)
    str_detect = function(string, pattern, negate = FALSE) {
      sql_str_pattern_switch(
        string = string,
        pattern = {{ pattern }},
        negate = negate,
        f_fixed = mssql_str_detect("detect")
      )
    },
    str_starts = function(string, pattern, negate = FALSE) {
      sql_str_pattern_switch(
        string = string,
        pattern = {{ pattern }},
        negate = negate,
        f_fixed = mssql_str_detect("start")
      )
    },
    str_ends = function(string, pattern, negate = FALSE) {
      sql_str_pattern_switch(
        string = string,
        pattern = {{ pattern }},
        negate = negate,
        f_fixed = mssql_str_detect("end")
      )
    },
    # no built in function: https://stackoverflow.com/questions/230138
    str_to_title = sql_not_supported("str_to_title"),
    # https://docs.microsoft.com/en-us/sql/t-sql/functions/substring-transact-sql?view=sql-server-ver15
    str_sub = sql_str_sub("SUBSTRING", "LEN", optional_length = FALSE),

    # lubridate ---------------------------------------------------------------
    # https://en.wikibooks.org/wiki/SQL_Dialects_Reference/Functions_and_expressions/Date_and_time_functions
    as_date = sql_cast("DATE"),

    # Using DATETIME2 as it complies with ANSI and ISO.
    # MS recommends DATETIME2 for new work:
    # https://docs.microsoft.com/en-us/sql/t-sql/data-types/datetime-transact-sql?view=sql-server-2017
    as_datetime = sql_cast("DATETIME2"),

    today = \() sql_glue("CAST(SYSDATETIME() AS DATE)"),

    # https://docs.microsoft.com/en-us/sql/t-sql/functions/datepart-transact-sql?view=sql-server-2017
    year = \(x) sql_glue("DATEPART(YEAR, {x})"),
    day = \(x) sql_glue("DATEPART(DAY, {x})"),
    mday = \(x) sql_glue("DATEPART(DAY, {x})"),
    yday = \(x) sql_glue("DATEPART(DAYOFYEAR, {x})"),
    hour = \(x) sql_glue("DATEPART(HOUR, {x})"),
    minute = \(x) sql_glue("DATEPART(MINUTE, {x})"),
    second = \(x) sql_glue("DATEPART(SECOND, {x})"),

    month = function(x, label = FALSE, abbr = TRUE) {
      check_bool(label)
      if (!label) {
        sql_glue("DATEPART(MONTH, {x})")
      } else {
        check_unsupported_arg(abbr, FALSE, backend = "SQL Server")
        sql_glue("DATENAME(MONTH, {x})")
      }
    },

    quarter = function(x, with_year = FALSE, fiscal_start = 1) {
      check_bool(with_year)
      check_unsupported_arg(fiscal_start, 1, backend = "SQL Server")

      if (with_year) {
        sql_glue("(DATENAME(YEAR, {x}) + '.' + DATENAME(QUARTER, {x}))")
      } else {
        sql_glue("DATEPART(QUARTER, {x})")
      }
    },

    # clock ---------------------------------------------------------------
    add_days = function(x, n, ...) {
      check_dots_empty()
      sql_glue("DATEADD(DAY, {n}, {x})")
    },
    add_years = function(x, n, ...) {
      check_dots_empty()
      sql_glue("DATEADD(YEAR, {n}, {x})")
    },
    date_build = function(year, month = 1L, day = 1L, ..., invalid = NULL) {
      check_unsupported_arg(invalid, allow_null = TRUE)
      sql_glue("DATEFROMPARTS({year}, {month}, {day})")
    },
    get_year = function(x) {
      sql_glue("DATEPART(YEAR, {x})")
    },
    get_month = function(x) {
      sql_glue("DATEPART(MONTH, {x})")
    },
    get_day = function(x) {
      sql_glue("DATEPART(DAY, {x})")
    },
    date_count_between = function(start, end, precision, ..., n = 1L) {
      check_dots_empty()
      check_unsupported_arg(precision, allowed = "day")
      check_unsupported_arg(n, allowed = 1L)
      # https://learn.microsoft.com/en-us/sql/t-sql/functions/datediff-big-transact-sql?view=sql-server-ver17
      sql_glue("DATEDIFF_BIG(DAY, {start}, {end})")
    },

    difftime = function(time1, time2, tz, units = "days") {
      check_unsupported_arg(tz)
      check_unsupported_arg(units, allowed = "days")

      sql_glue("DATEDIFF_BIG(DAY, {time2}, {time1})")
    }
  )
}

# MSSQL 2012
mssql_scalar_11 <- function() {
  sql_translator(
    .parent = mssql_scalar_base(),
    as.logical = sql_try_cast("BIT"),
    as.Date = sql_try_cast("DATE"),
    as.POSIXct = sql_try_cast("DATETIME2"),
    as.numeric = sql_try_cast("FLOAT"),
    as.double = sql_try_cast("FLOAT"),

    # In SQL server, CAST (even with TRY) of INTEGER and BIGINT appears
    # fill entire columns with NULL if parsing single value fails:
    # https://gist.github.com/DavidPatShuiFong/7b47a9804a497b605e477f1bf6c38b37
    # So we parse to NUMERIC (which doesn't have this problem), then to the
    # target type
    as.integer = function(x) {
      sql_glue("TRY_CAST(TRY_CAST({x} AS NUMERIC) AS INT)")
    },
    as.integer64 = function(x) {
      sql_glue("TRY_CAST(TRY_CAST({x} AS NUMERIC(38, 0)) AS BIGINT)")
    },
    as.character = sql_try_cast("VARCHAR(MAX)"),
    as_date = sql_try_cast("DATE"),
    as_datetime = sql_try_cast("DATETIME2")
  )
}

# MSSQL 2025 - regex functions
# https://learn.microsoft.com/en-us/sql/t-sql/functions/regexp-like-transact-sql
mssql_scalar_17 <- function() {
  sql_translator(
    .parent = mssql_scalar_11(),
    str_detect = function(string, pattern, negate = FALSE) {
      sql_str_pattern_switch(
        string = string,
        pattern = {{ pattern }},
        negate = negate,
        f_fixed = mssql_str_detect("detect"),
        f_regex = function(string, pattern, negate = FALSE) {
          if (isTRUE(negate)) {
            mssql_as_bit(sql_glue("NOT REGEXP_LIKE({string}, {pattern})"))
          } else {
            mssql_as_bit(sql_glue("REGEXP_LIKE({string}, {pattern})"))
          }
        }
      )
    },
    str_starts = function(string, pattern, negate = FALSE) {
      sql_str_pattern_switch(
        string = string,
        pattern = {{ pattern }},
        negate = negate,
        f_fixed = mssql_str_detect("start"),
        f_regex = function(string, pattern, negate = FALSE) {
          pattern <- sql_glue("'^' + {pattern}")
          if (isTRUE(negate)) {
            mssql_as_bit(sql_glue("NOT REGEXP_LIKE({string}, {pattern})"))
          } else {
            mssql_as_bit(sql_glue("REGEXP_LIKE({string}, {pattern})"))
          }
        }
      )
    },
    str_ends = function(string, pattern, negate = FALSE) {
      sql_str_pattern_switch(
        string = string,
        pattern = {{ pattern }},
        negate = negate,
        f_fixed = mssql_str_detect("end"),
        f_regex = function(string, pattern, negate = FALSE) {
          pattern <- sql_glue("{pattern} + '$'")
          if (isTRUE(negate)) {
            mssql_as_bit(sql_glue("NOT REGEXP_LIKE({string}, {pattern})"))
          } else {
            mssql_as_bit(sql_glue("REGEXP_LIKE({string}, {pattern})"))
          }
        }
      )
    },
    str_replace = function(string, pattern, replacement) {
      sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement}, 1, 1)")
    },
    str_replace_all = function(string, pattern, replacement) {
      sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement})")
    },
    str_remove = function(string, pattern) {
      sql_glue("REGEXP_REPLACE({string}, {pattern}, '', 1, 1)")
    },
    str_remove_all = function(string, pattern) {
      sql_glue("REGEXP_REPLACE({string}, {pattern}, '')")
    },
    str_extract = function(string, pattern) {
      sql_glue("REGEXP_SUBSTR({string}, {pattern})")
    },
    str_count = function(string, pattern) {
      sql_glue("REGEXP_COUNT({string}, {pattern})")
    }
  )
}

#' @export
sql_translation.sql_dialect_mssql <- function(con) {
  # Needed because UseMethod hack only affects dispatch, not value
  version <- sql_dialect(con)$version
  mssql_scalar <- if (version >= "17.0") {
    mssql_scalar_17()
  } else if (version >= "11.0") {
    mssql_scalar_11()
  } else {
    mssql_scalar_base()
  }

  sql_variant(
    mssql_scalar,
    sql_translator(
      .parent = base_odbc_agg,
      # https://learn.microsoft.com/en-us/sql/t-sql/functions/count-big-transact-sql?view=sql-server-ver17
      n = \() sql("COUNT_BIG(*)"),
      sd = sql_aggregate("STDEV", "sd"),
      var = sql_aggregate("VAR", "var"),
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        sql_glue("STRING_AGG({x}, {collapse})")
      },

      median = sql_agg_not_supported("median", "SQL Server"),
      quantile = sql_agg_not_supported("quantile", "SQL Server"),
      all = mssql_bit_int_bit(sql_aggregate("MIN")),
      any = mssql_bit_int_bit(sql_aggregate("MAX"))
    ),
    sql_translator(
      .parent = base_odbc_win,
      sd = win_aggregate("STDEV"),
      var = win_aggregate("VAR"),
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        win_over(
          sql_glue("STRING_AGG({x}, {collapse})"),
          partition = win_current_group(),
          order = win_current_order()
        )
      },

      # percentile_cont needs `OVER()` in mssql
      # https://docs.microsoft.com/en-us/sql/t-sql/functions/percentile-cont-transact-sql?view=sql-server-ver15
      median = sql_median("PERCENTILE_CONT", "ordered", window = TRUE),
      quantile = sql_quantile("PERCENTILE_CONT", "ordered", window = TRUE),
      first = function(x, order_by = NULL, na_rm = FALSE) {
        sql_nth(
          x = x,
          n = 1L,
          order_by = order_by,
          na_rm = na_rm,
          ignore_nulls = "outside"
        )
      },
      last = function(x, order_by = NULL, na_rm = FALSE) {
        sql_nth(
          x = x,
          n = Inf,
          order_by = order_by,
          na_rm = na_rm,
          ignore_nulls = "outside"
        )
      },
      nth = function(x, n, order_by = NULL, na_rm = FALSE) {
        sql_nth(
          x = x,
          n = n,
          order_by = order_by,
          na_rm = na_rm,
          ignore_nulls = "outside"
        )
      },
      all = mssql_bit_int_bit(win_aggregate("MIN")),
      any = mssql_bit_int_bit(win_aggregate("MAX")),
      row_number = win_rank("ROW_NUMBER", empty_order = TRUE),

      n_distinct = function(x, na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        cli_abort(
          "No translation available in `mutate()`/`filter()` for SQL server."
        )
      }
    )
  )
}

mssql_version <- function(con) {
  if (inherits(con, "TestConnection")) {
    attr(con, "version")
  } else {
    numeric_version(DBI::dbGetInfo(con)$db.version) # nocov
  }
}

#' @export
sql_escape_raw.sql_dialect_mssql <- function(con, x) {
  if (is.null(x)) {
    "NULL"
  } else {
    # SQL Server binary constants should be prefixed with 0x
    # https://docs.microsoft.com/en-us/sql/t-sql/data-types/constants-transact-sql?view=sql-server-ver15#binary-constants
    paste0(c("0x", format(x)), collapse = "")
  }
}

#' @export
sql_table_analyze.sql_dialect_mssql <- function(con, table, ...) {
  # https://docs.microsoft.com/en-us/sql/t-sql/statements/update-statistics-transact-sql
  sql_glue2(con, "UPDATE STATISTICS {.tbl table}")
}

# SQL server does not use CREATE TEMPORARY TABLE and instead prefixes
# temporary table names with #
# <https://docs.microsoft.com/en-us/previous-versions/sql/sql-server-2008-r2/ms177399%28v%3dsql.105%29#temporary-tables>
#' @export
sql_table_temporary.sql_dialect_mssql <- function(con, table, temporary, ...) {
  list(
    table = add_temporary_prefix(con, table, temporary = temporary),
    temporary = FALSE
  )
}


#' @export
sql_query_save.sql_dialect_mssql <- function(
  con,
  sql,
  name,
  temporary = TRUE,
  ...
) {
  # https://stackoverflow.com/q/16683758/946850
  sql_glue2(con, "SELECT * INTO {.tbl name} FROM (\n  {sql}\n) AS temp")
}

#' @export
sql_values_subquery.sql_dialect_mssql <- sql_values_subquery_column_alias

#' @export
sql_returning_cols.sql_dialect_mssql <- function(con, cols, table, ...) {
  arg_match(table, values = c("DELETED", "INSERTED"))
  returning_cols <- sql_named_cols(con, cols, table = table)

  sql_clause("OUTPUT", returning_cols)
}

# Bit vs boolean ----------------------------------------------------------

mssql_needs_bit <- function() {
  context <- sql_current_context()
  identical(context$clause, "SELECT") || identical(context$clause, "ORDER")
}

with_mssql_bool <- function(code) {
  local_context(list(clause = ""))
  code
}

mssql_as_bit <- function(x) {
  if (mssql_needs_bit()) {
    sql_glue("CAST(CASE WHEN {x} THEN 1 ELSE 0 END AS BIT)")
  } else {
    x
  }
}

mssql_is_null <- function(x) {
  mssql_as_bit(sql_is_null({{ x }}))
}

mssql_infix_comparison <- function(f) {
  check_string(f)
  function(x, y) {
    mssql_as_bit(sql_glue("{x} {.sql f} {y}"))
  }
}

mssql_infix_boolean <- function(if_bit, if_bool) {
  force(if_bit)
  force(if_bool)

  function(x, y) {
    if (mssql_needs_bit()) {
      x <- with_mssql_bool(x)
      y <- with_mssql_bool(y)
      mssql_as_bit(sql_glue("{x} {.sql if_bool} {y}"))
    } else {
      sql_glue("{x} {.sql if_bool} {y}")
    }
  }
}

mssql_sql_if <- function(
  cond,
  if_true,
  if_false = quo(NULL),
  missing = quo(NULL)
) {
  with_mssql_bool(sql_if(cond, if_true, if_false, missing))
}

mssql_case_when <- function(...) {
  with_mssql_bool(sql_case_when(...))
}

mssql_bit_int_bit <- function(f) {
  # bit fields must be cast to numeric before aggregating (e.g. min/max).
  function(x, na.rm = FALSE) {
    f_wrapped <- purrr::compose(
      sql_cast("BIT"),
      purrr::partial(f, na.rm = na.rm),
      sql_cast("INT")
    )

    f_wrapped(x)
  }
}

# SQL Server uses CHARINDEX(substring, string) instead of INSTR(string, substring)
mssql_str_detect <- function(type = c("detect", "start", "end")) {
  type <- arg_match(type)

  function(string, pattern, negate = FALSE) {
    index_sql <- sql_glue("CHARINDEX({pattern}, {string})")

    if (negate) {
      sql <- switch(
        type,
        detect = sql_glue("{index_sql} = 0"),
        start = sql_glue("{index_sql} != 1"),
        end = sql_glue("{index_sql} != (LEN({string}) - LEN({pattern})) + 1")
      )
    } else {
      sql <- switch(
        type,
        detect = sql_glue("{index_sql} > 0"),
        start = sql_glue("{index_sql} = 1"),
        end = sql_glue("{index_sql} = (LEN({string}) - LEN({pattern})) + 1")
      )
    }
    mssql_as_bit(sql)
  }
}

#' @export
sql_escape_logical.sql_dialect_mssql <- function(con, x) {
  dplyr::if_else(x, "1", "0", "NULL")
}

#' @export
`db_sql_render.Microsoft SQL Server` <- function(
  con,
  sql,
  ...,
  cte = FALSE,
  sql_options = NULL
) {
  sql <- mssql_render_preprocess(sql)
  NextMethod()
}

#' @export
db_sql_render.sql_dialect_mssql <- function(
  con,
  sql,
  ...,
  cte = FALSE,
  sql_options = NULL
) {
  sql <- mssql_render_preprocess(sql)
  NextMethod()
}

mssql_render_preprocess <- function(sql) {
  # Post-process WHERE to cast logicals from BIT to BOOLEAN.
  sql$lazy_query <- purrr::modify_tree(
    sql$lazy_query,
    is_node = \(x) is.list(x) || inherits(x, "lazy_query"),
    post = mssql_update_where_clause
  )
  sql
}

mssql_update_where_clause <- function(qry) {
  if (!inherits(qry, "lazy_query") || !has_name(qry, "where")) {
    return(qry)
  }

  qry$where <- lapply(
    qry$where,
    \(x) set_expr(x, bit_to_boolean(get_expr(x)))
  )
  qry
}

bit_to_boolean <- function(x_expr) {
  if (is_atomic(x_expr) || is_symbol(x_expr)) {
    expr(cast(!!x_expr %AS% BIT) == 1L)
  } else if (is_call(x_expr, c("|", "&", "||", "&&", "!", "("))) {
    idx <- seq2(2, length(x_expr))
    x_expr[idx] <- lapply(x_expr[idx], bit_to_boolean)
    x_expr
  } else {
    x_expr
  }
}

utils::globalVariables(c("cast", "%AS%", "BIT"))

test_that("simulate_mssql() still works", {
  expect_translation(simulate_mssql(), x + 1, "[x] + 1.0")
})

test_that("table_path_components parses SQL Server bracket syntax", {
  components <- function(x) {
    table_path_components(table_path(x), dialect_mssql())
  }

  # Bracket quoting
  expect_equal(components("[df]"), list("df"))
  expect_equal(components("[schema].[table]"), list(c("schema", "table")))
  expect_equal(components("[x.y].[z]"), list(c("x.y", "z")))
  expect_equal(components(c("[a]", "[b].[c]")), list("a", c("b", "c")))

  # Double-quote quoting
  expect_equal(components('"df"'), list("df"))
  expect_equal(components('"schema"."table"'), list(c("schema", "table")))
  expect_equal(components('"x.y"."z"'), list(c("x.y", "z")))
})

# function translation ----------------------------------------------------

test_that("custom scalar translated correctly", {
  con <- dialect_mssql()

  expect_translation(con, as.logical(x), "TRY_CAST([x] AS BIT)")
  expect_translation(con, as.numeric(x), "TRY_CAST([x] AS FLOAT)")
  expect_translation(
    con,
    as.integer(x),
    "TRY_CAST(TRY_CAST([x] AS NUMERIC) AS INT)"
  )
  expect_translation(
    con,
    as.integer64(x),
    "TRY_CAST(TRY_CAST([x] AS NUMERIC(38, 0)) AS BIGINT)"
  )
  expect_translation(con, as.double(x), "TRY_CAST([x] AS FLOAT)")
  expect_translation(con, as.character(x), "TRY_CAST([x] AS VARCHAR(MAX))")
  expect_translation(con, log(x), "LOG([x])")
  expect_translation(con, nchar(x), "LEN([x])")
  expect_translation(con, atan2(x), "ATN2([x])")
  expect_translation(con, ceiling(x), "CEILING([x])")
  expect_translation(con, ceil(x), "CEILING([x])")
  expect_translation(con, substr(x, 1, 2), "SUBSTRING([x], 1, 2)")
  expect_translation(con, trimws(x), "LTRIM(RTRIM([x]))")
  expect_translation(con, paste(x, y), "[x] + ' ' + [y]")

  expect_error(
    translate_sql(bitwShiftL(x, 2L), con = con),
    class = "dbplyr_error_unsupported_fn"
  )
  expect_error(
    translate_sql(bitwShiftR(x, 2L), con = con),
    class = "dbplyr_error_unsupported_fn"
  )
})

test_that("contents of [ have bool context", {
  con <- dialect_mssql()
  local_context(list(clause = "SELECT"))

  expect_translation(con, x[x > y], "CASE WHEN ([x] > [y]) THEN ([x]) END")
})

test_that("custom stringr functions translated correctly", {
  con <- dialect_mssql()

  expect_translation(con, str_length(x), "LEN([x])")
})

test_that("stringr fixed patterns use CHARINDEX", {
  con <- dialect_mssql()

  expect_translation(
    con,
    str_detect(x, fixed("abc")),
    "CHARINDEX('abc', [x]) > 0"
  )
  expect_translation(
    con,
    str_detect(x, fixed("abc"), negate = TRUE),
    "CHARINDEX('abc', [x]) = 0"
  )
  expect_translation(
    con,
    str_starts(x, fixed("abc")),
    "CHARINDEX('abc', [x]) = 1"
  )
  expect_translation(
    con,
    str_ends(x, fixed("abc")),
    "CHARINDEX('abc', [x]) = (LEN([x]) - LEN('abc')) + 1"
  )
})

test_that("stringr regex patterns require SQL Server 2025", {
  con <- dialect_mssql("15.0")
  expect_error(
    translate_sql(str_detect(x, "abc"), con = con),
    "Only fixed patterns are supported"
  )
})

test_that("stringr regex functions work on SQL Server 2025", {
  con <- dialect_mssql("17.0")

  expect_translation(con, str_detect(x, "abc"), "REGEXP_LIKE([x], 'abc')")
  expect_translation(
    con,
    str_detect(x, "abc", negate = TRUE),
    "NOT REGEXP_LIKE([x], 'abc')"
  )
  expect_translation(
    con,
    str_starts(x, "abc"),
    "REGEXP_LIKE([x], '^' + 'abc')"
  )
  expect_translation(
    con,
    str_ends(x, "abc"),
    "REGEXP_LIKE([x], 'abc' + '$')"
  )
  expect_translation(
    con,
    str_replace(x, "abc", "def"),
    "REGEXP_REPLACE([x], 'abc', 'def', 1, 1)"
  )
  expect_translation(
    con,
    str_replace_all(x, "abc", "def"),
    "REGEXP_REPLACE([x], 'abc', 'def')"
  )
  expect_translation(
    con,
    str_remove(x, "abc"),
    "REGEXP_REPLACE([x], 'abc', '', 1, 1)"
  )
  expect_translation(
    con,
    str_remove_all(x, "abc"),
    "REGEXP_REPLACE([x], 'abc', '')"
  )
  expect_translation(con, str_extract(x, "abc"), "REGEXP_SUBSTR([x], 'abc')")
  expect_translation(con, str_count(x, "abc"), "REGEXP_COUNT([x], 'abc')")
})

test_that("str_detect returns bit in SELECT context on SQL Server 2025", {
  lf <- lazy_frame(x = "test", con = dialect_mssql("17.0"))

  expect_snapshot(lf |> mutate(detected = str_detect(x, "abc")))
  expect_snapshot(lf |> filter(str_detect(x, "abc")))
})

test_that("custom aggregators translated correctly", {
  con <- dialect_mssql()

  expect_translation(con, sd(x, na.rm = TRUE), "STDEV([x])", window = FALSE)
  expect_translation(con, var(x, na.rm = TRUE), "VAR([x])", window = FALSE)

  expect_error(
    translate_sql(cor(x), con = con, window = FALSE),
    class = "dbplyr_error_unsupported_fn"
  )
  expect_error(
    translate_sql(cov(x), con = con, window = FALSE),
    class = "dbplyr_error_unsupported_fn"
  )

  expect_translation(con, str_flatten(x), "STRING_AGG([x], '')", window = FALSE)
  expect_snapshot(error = TRUE, {
    translate_sql(quantile(x, 0.5, na.rm = TRUE), con = con, window = FALSE)
    translate_sql(median(x, na.rm = TRUE), con = con, window = FALSE)
  })

  expect_translation(
    con,
    all(x, na.rm = TRUE),
    "CAST(MIN(CAST([x] AS INT)) AS BIT)",
    window = FALSE
  )
  expect_translation(
    con,
    any(x, na.rm = TRUE),
    "CAST(MAX(CAST([x] AS INT)) AS BIT)",
    window = FALSE
  )
})

test_that("custom window functions translated correctly", {
  con <- dialect_mssql()

  expect_translation(con, sd(x, na.rm = TRUE), "STDEV([x]) OVER ()")
  expect_translation(con, var(x, na.rm = TRUE), "VAR([x]) OVER ()")

  expect_translation(con, str_flatten(x), "STRING_AGG([x], '') OVER ()")

  expect_translation(
    con,
    quantile(x, 0.3, na.rm = TRUE),
    "PERCENTILE_CONT(0.3) WITHIN GROUP (ORDER BY [x]) OVER ()",
    window = TRUE
  )
  expect_translation(
    con,
    median(x, na.rm = TRUE),
    "PERCENTILE_CONT(0.5) WITHIN GROUP (ORDER BY [x]) OVER ()",
    window = TRUE
  )

  expect_translation(
    con,
    all(x, na.rm = TRUE),
    "CAST(MIN(CAST([x] AS INT)) OVER () AS BIT)"
  )
  expect_translation(
    con,
    any(x, na.rm = TRUE),
    "CAST(MAX(CAST([x] AS INT)) OVER () AS BIT)"
  )

  expect_snapshot(
    translate_sql(n_distinct(x), con = con, vars_group = "x"),
    error = TRUE
  )
})

test_that("custom lubridate functions translated correctly", {
  con <- dialect_mssql()
  expect_translation(con, as_date(x), "TRY_CAST([x] AS DATE)")
  expect_translation(con, as_datetime(x), "TRY_CAST([x] AS DATETIME2)")
  expect_translation(con, today(), "CAST(SYSDATETIME() AS DATE)")
  expect_translation(con, year(x), "DATEPART(YEAR, [x])")
  expect_translation(con, day(x), "DATEPART(DAY, [x])")
  expect_translation(con, mday(x), "DATEPART(DAY, [x])")
  expect_translation(con, yday(x), "DATEPART(DAYOFYEAR, [x])")
  expect_translation(con, hour(x), "DATEPART(HOUR, [x])")
  expect_translation(con, minute(x), "DATEPART(MINUTE, [x])")
  expect_translation(con, second(x), "DATEPART(SECOND, [x])")
  expect_translation(con, month(x), "DATEPART(MONTH, [x])")
  expect_translation(
    con,
    month(x, label = TRUE, abbr = FALSE),
    "DATENAME(MONTH, [x])"
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(month(x, label = TRUE, abbr = TRUE), con = con)
  )

  expect_translation(con, quarter(x), "DATEPART(QUARTER, [x])")
  expect_translation(
    con,
    quarter(x, with_year = TRUE),
    "(DATENAME(YEAR, [x]) + '.' + DATENAME(QUARTER, [x]))"
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(quarter(x, fiscal_start = 5), con = con)
  )
})

test_that("custom clock functions translated correctly", {
  con <- dialect_mssql()
  expect_translation(con, add_years(x, 1), "DATEADD(YEAR, 1.0, [x])")
  expect_translation(con, add_days(x, 1), "DATEADD(DAY, 1.0, [x])")
  expect_error(
    translate_sql(add_days(x, 1, "dots", "must", "be empty"), con = con),
    class = "rlib_error_dots_nonempty"
  )
  expect_translation(
    con,
    date_build(2020, 1, 1),
    "DATEFROMPARTS(2020.0, 1.0, 1.0)"
  )
  expect_translation(
    con,
    date_build(year_column, 1L, 1L),
    "DATEFROMPARTS([year_column], 1, 1)"
  )
  expect_translation(
    con,
    get_year(date_column),
    "DATEPART(YEAR, [date_column])"
  )
  expect_translation(
    con,
    get_month(date_column),
    "DATEPART(MONTH, [date_column])"
  )
  expect_translation(con, get_day(date_column), "DATEPART(DAY, [date_column])")
  expect_translation(
    con,
    date_count_between(date_column_1, date_column_2, "day"),
    "DATEDIFF_BIG(DAY, [date_column_1], [date_column_2])"
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      date_count_between(date_column_1, date_column_2, "year"),
      con = con
    )
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      date_count_between(
        date_column_1,
        date_column_2,
        "day",
        n = 5
      ),
      con = con
    )
  )
})

test_that("difftime is translated correctly", {
  con <- dialect_mssql()
  expect_translation(
    con,
    difftime(start_date, end_date, units = "days"),
    "DATEDIFF_BIG(DAY, [end_date], [start_date])"
  )
  expect_translation(
    con,
    difftime(start_date, end_date),
    "DATEDIFF_BIG(DAY, [end_date], [start_date])"
  )

  expect_snapshot(
    error = TRUE,
    translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      difftime(
        start_date,
        end_date,
        tz = "UTC",
        units = "days"
      ),
      con = con
    )
  )
})

test_that("last_value_sql() translated correctly", {
  con <- dialect_mssql()
  expect_equal(
    translate_sql(last(x, na_rm = TRUE), vars_order = "a", con = con),
    sql(
      "LAST_VALUE([x]) IGNORE NULLS OVER (ORDER BY [a] ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING)"
    )
  )
})

test_that("between translation respects context", {
  con <- dialect_mssql()

  local_context(list(clause = "WHERE"))
  expect_translation(con, between(a, 1L, 2L), "[a] BETWEEN 1 AND 2")
  local_context(list(clause = "SELECT"))
  expect_translation(
    con,
    between(a, 1L, 2L),
    "CASE WHEN [a] BETWEEN 1 AND 2 THEN 1 ELSE 0 END"
  )
})

# verb translation --------------------------------------------------------

test_that("convert between bit and boolean as needed", {
  mf <- lazy_frame(x = 1, con = dialect_mssql())

  # No conversion
  expect_snapshot(mf |> filter(is.na(x)))
  expect_snapshot(mf |> filter(!is.na(x)))
  expect_snapshot(mf |> filter(x == 1L || x == 2L))
  expect_snapshot(mf |> mutate(z = ifelse(x == 1L, 1L, 2L)))
  expect_snapshot(mf |> mutate(z = case_when(x == 1L ~ 1L)))

  # Single conversion on outer layer
  expect_snapshot(mf |> mutate(z = !is.na(x)))
  expect_snapshot(mf |> mutate(x = x == 1L))
  expect_snapshot(mf |> mutate(x = x == 1L || x == 2L))
  expect_snapshot(mf |> mutate(x = x == 1L || x == 2L || x == 3L))
  expect_snapshot(mf |> mutate(x = !(x == 1L || x == 2L || x == 3L)))
})

test_that("handles ORDER BY in subqueries", {
  expect_snapshot(
    sql_query_select(
      dialect_mssql(),
      sql("[x]"),
      sql("[y]"),
      order_by = "z",
      subquery = TRUE
    )
  )
})

test_that("custom limit translation", {
  expect_snapshot(
    sql_query_select(
      dialect_mssql(),
      sql("[x]"),
      sql("[y]"),
      order_by = sql("[z]"),
      limit = 10
    )
  )
})

test_that("custom escapes translated correctly", {
  mf <- lazy_frame(x = "abc", con = dialect_mssql())

  a <- blob::as_blob("abc")
  b <- blob::as_blob(as.raw(c(0x01, 0x02)))
  L <- c(a, b)

  expect_snapshot(mf |> filter(x == a))
  expect_snapshot(mf |> filter(x %in% L))

  # expect_snapshot() also uses !!
  qry <- mf |> filter(x %in% !!L)
  expect_snapshot(qry)
})

test_that("logical escaping to 0/1 for both filter() and mutate()", {
  mf <- lazy_frame(x = "abc", con = dialect_mssql())
  expect_snapshot(mf |> filter(x == TRUE))
  expect_snapshot(mf |> mutate(x = TRUE))
})

test_that("sql_escape_raw handles NULLs", {
  con <- dialect_mssql()
  expect_equal(sql_escape_raw(con, NULL), "NULL")
})

test_that("generates custom sql", {
  con <- dialect_mssql()

  expect_snapshot(sql_table_analyze(con, in_schema("schema", "tbl")))

  # Creates the same SQL since there's no temporary CLAUSE
  # Automatic renaming is handled upstream by db_collect()/db_copy_to()
  expect_snapshot(sql_query_save(
    con,
    sql("SELECT * FROM foo"),
    in_schema("schema", "tbl")
  ))
  expect_snapshot(sql_query_save(
    con,
    sql("SELECT * FROM foo"),
    in_schema("schema", "tbl"),
    temporary = FALSE
  ))

  lf <- lazy_frame(x = 1:3, con = dialect_mssql())
  expect_snapshot(lf |> slice_sample(n = 1))

  expect_snapshot(
    copy_inline(con, tibble(x = 1:2, y = letters[1:2])) |> remote_query()
  )
  expect_snapshot(copy_inline(con, trees) |> remote_query())
})

test_that("`sql_query_insert()` is correct", {
  con <- dialect_mssql()
  df_y <- lazy_frame(
    a = 2:3,
    b = c(12L, 13L),
    c = -(2:3),
    d = c("y", "z"),
    con = con,
    .name = "df_y"
  ) |>
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_insert(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      insert_cols = colnames(df_y),
      by = c("a", "b"),
      conflict = "ignore",
      returning_cols = c("a", b2 = "b")
    )
  )
})

test_that("`sql_query_append()` is correct", {
  con <- dialect_mssql()
  df_y <- lazy_frame(
    a = 2:3,
    b = c(12L, 13L),
    c = -(2:3),
    d = c("y", "z"),
    con = con,
    .name = "df_y"
  ) |>
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_append(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      insert_cols = colnames(df_y),
      returning_cols = c("a", b2 = "b")
    )
  )
})

test_that("`sql_query_update_from()` is correct", {
  con <- dialect_mssql()
  df_y <- lazy_frame(
    a = 2:3,
    b = c(12L, 13L),
    c = -(2:3),
    d = c("y", "z"),
    con = con,
    .name = "df_y"
  ) |>
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_update_from(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      by = c("a", "b"),
      update_values = sql(
        c = "COALESCE([df_x].[c], [...y].[c])",
        d = "[...y].[d]"
      ),
      returning_cols = c("a", b2 = "b")
    )
  )
})

test_that("`sql_query_delete()` is correct", {
  df_y <- lazy_frame(
    a = 2:3,
    b = c(12L, 13L),
    c = -(2:3),
    d = c("y", "z"),
    con = dialect_mssql(),
    .name = "df_y"
  ) |>
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_delete(
      con = dialect_mssql(),
      table = ident("df_x"),
      from = sql_render(df_y, dialect_mssql(), lvl = 2),
      by = c("a", "b"),
      returning_cols = c("a", b2 = "b")
    )
  )
})

test_that("`sql_query_upsert()` is correct", {
  con <- dialect_mssql()
  df_y <- lazy_frame(
    a = 2:3,
    b = c(12L, 13L),
    c = -(2:3),
    d = c("y", "z"),
    con = con,
    .name = "df_y"
  ) |>
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_upsert(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      by = c("a", "b"),
      update_cols = c("c", "d"),
      returning_cols = c("a", b2 = "b")
    )
  )
})

test_that("atoms and symbols are cast to bit in `filter`", {
  mf <- lazy_frame(x = TRUE, con = dialect_mssql())

  # as simple symbol and atom
  expect_snapshot(mf |> filter(x))
  expect_snapshot(mf |> filter(TRUE))

  # when involved in a (perhaps nested) logical expression
  expect_snapshot(mf |> filter((!x) | FALSE))

  # in a subquery
  expect_snapshot(mf |> filter(x) |> inner_join(mf, by = "x"))

  # on the RHS of a join (#1839)
  lhs <- lazy_frame(a = 1, b1 = TRUE, con = dialect_mssql()) |> filter(b1)
  rhs <- lazy_frame(a = 1, b2 = TRUE, con = dialect_mssql()) |> filter(b2)
  expect_snapshot(left_join(lhs, rhs, by = "a"))
})

test_that("row_number() with and without group_by() and arrange(): unordered defaults to Ordering by NULL (per empty_order)", {
  mf <- lazy_frame(x = c(1:5), y = c(rep("A", 5)), con = dialect_mssql())
  expect_snapshot(mf |> mutate(rown = row_number()))
  expect_snapshot(mf |> group_by(y) |> mutate(rown = row_number()))
  expect_snapshot(mf |> arrange(y) |> mutate(rown = row_number()))
})

test_that("count_big", {
  mf <- lazy_frame(x = c(1:5), y = c(rep("A", 5)), con = dialect_mssql())
  expect_snapshot(count(mf))
})

# Live database -----------------------------------------------------------

test_that("can copy_to() and compute() with temporary tables (#438)", {
  con <- test_mssql()
  df <- tibble(x = 1:3)

  # converts to name automatically with message
  expect_snapshot(
    db <- copy_to(con, df, name = unique_table_name(), temporary = TRUE),
    transform = snap_transform_dbi
  )
  expect_equal(db |> pull(), 1:3)

  expect_snapshot(
    db2 <- db |> mutate(y = x + 1) |> compute(),
    transform = snap_transform_dbi
  )
  expect_equal(db2 |> pull(), 2:4)
})

test_that("add prefix to temporary table", {
  con <- dialect_mssql()
  expect_snapshot(
    out <- sql_table_temporary(con, table_path("foo.bar"), temporary = TRUE)
  )
  expect_equal(out, list(table = table_path("[foo].[#bar]"), temporary = FALSE))

  expect_silent(
    out <- sql_table_temporary(con, table_path("foo.#bar"), temporary = TRUE)
  )
  expect_equal(out, list(table = table_path("foo.#bar"), temporary = FALSE))
})

test_that("bit conversion works for important cases", {
  df <- tibble(x = 1:3, y = 3:1)
  db <- copy_to(test_mssql(), df, name = unique_table_name("#"))
  expect_equal(db |> mutate(z = x == y) |> pull(), c(FALSE, TRUE, FALSE))
  expect_equal(db |> filter(x == y) |> pull(), 2)

  df <- tibble(x = c(TRUE, FALSE, FALSE), y = c(TRUE, FALSE, TRUE))
  db <- copy_to(test_mssql(), df, name = unique_table_name("#"))
  expect_equal(db |> filter(x == 1) |> pull(), TRUE)
  expect_equal(db |> mutate(z = TRUE) |> pull(), c(1, 1, 1))

  # Currently not supported: would need to determine that we have a bit
  # vector in a boolean context, and convert to boolean with x == 1.
  # expect_equal(db |> mutate(z = x) |> pull(), c(TRUE, FALSE, FALSE))
  # expect_equal(db |> mutate(z = !x) |> pull(), c(FALSE, TRUE, TRUE))
  # expect_equal(db |> mutate(z = x & y) |> pull(), c(TRUE, FALSE, FALSE))
})

test_that("as.integer and as.integer64 translations if parsing failures", {
  df <- data.frame(x = c("1.3", "2x"))
  db <- copy_to(test_mssql(), df, name = unique_table_name("#"))

  out <- db |>
    mutate(
      integer = as.integer(x),
      integer64 = as.integer64(x),
      numeric = as.numeric(x),
    ) |>
    collect()

  expect_identical(out$integer, c(1L, NA))
  expect_identical(out$integer64, bit64::as.integer64(c(1L, NA)))
  expect_identical(out$numeric, c(1.3, NA))
})

test_that("can insert", {
  con <- test_mssql()

  df_x <- tibble(a = 1L, b = 11L, c = 1L, d = "a")
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"))
  y <- local_db_table(con, df_y, "df_y", temporary = TRUE, overwrite = TRUE) |>
    mutate(c = c + 1)

  expect_equal(
    rows_insert(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      conflict = "ignore"
    ) |>
      collect(),
    tibble(
      a = 1:3,
      b = 11:13,
      c = c(1L, -1L, -2L),
      d = c("a", "y", "z")
    )
  )
})

test_that("can insert with returning", {
  con <- test_mssql()

  df_x <- tibble(a = 1L, b = 11L, c = 1L, d = "a")
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"))
  y <- local_db_table(con, df_y, "df_y") |>
    mutate(c = c + 1)

  expect_equal(
    rows_insert(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      conflict = "ignore",
      returning = everything()
    ) |>
      get_returned_rows(),
    tibble(
      a = 2:3,
      b = 12:13,
      c = c(-1L, -2L),
      d = c("y", "z")
    )
  )
})

test_that("can append", {
  con <- test_mssql()

  df_x <- tibble(a = 1L, b = 11L, c = 1L, d = "a")
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 1:3, b = 11:13, c = -(2:4), d = c("y", "z", "w"))
  y <- local_db_table(con, df_y, "df_y") |>
    mutate(c = c + 1)

  expect_equal(
    rows_append(
      x,
      y,
      in_place = TRUE
    ) |>
      collect(),
    tibble(
      a = c(1L, 1:3),
      b = c(11L, 11:13),
      c = c(1L, -1L, -2L, -3L),
      d = c("a", "y", "z", "w")
    )
  )
})

test_that("can append with returning", {
  con <- test_mssql()

  df_x <- tibble(a = 1L, b = 11L, c = 1L, d = "a")
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 1:3, b = 11:13, c = -(2:4), d = c("y", "z", "w"))
  y <- local_db_table(con, df_y, "df_y") |>
    mutate(c = c + 1)

  expect_equal(
    rows_append(
      x,
      y,
      in_place = TRUE,
      returning = everything()
    ) |>
      get_returned_rows(),
    tibble(
      a = 1:3,
      b = 11:13,
      c = c(-1L, -2L, -3L),
      d = c("y", "z", "w")
    )
  )
})

test_that("can update", {
  con <- test_mssql()

  df_x <- tibble(a = 1:3, b = 11:13, c = 1:3, d = c("a", "b", "c"))
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"))
  y <- local_db_table(con, df_y, "df_y") |>
    mutate(c = c + 1)

  expect_equal(
    rows_update(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      unmatched = "ignore"
    ) |>
      collect(),
    tibble(
      a = 1:3,
      b = 11:13,
      c = c(1L, -1L, -2L),
      d = c("a", "y", "z")
    )
  )
})

test_that("can update with returning", {
  con <- test_mssql()

  df_x <- tibble(a = 1:3, b = 11:13, c = 1:3, d = c("a", "b", "c"))
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"))
  y <- local_db_table(con, df_y, "df_y") |>
    mutate(c = c + 1)

  expect_equal(
    rows_update(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      unmatched = "ignore",
      returning = everything()
    ) |>
      get_returned_rows(),
    tibble(
      a = 2:3,
      b = 12:13,
      c = c(-1L, -2L),
      d = c("y", "z")
    )
  )
})

test_that("can upsert", {
  con <- test_mssql()

  df_x <- tibble(a = 1:2, b = 11:12, c = 1:2, d = c("a", "b"))
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"))
  y <- local_db_table(con, df_y, "df_y") |>
    mutate(c = c + 1)

  expect_equal(
    rows_upsert(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE
    ) |>
      collect(),
    tibble(
      a = 1:3,
      b = 11:13,
      c = c(1L, -1L, -2L),
      d = c("a", "y", "z")
    )
  )
})

test_that("can upsert with returning", {
  con <- test_mssql()

  df_x <- tibble(a = 1:2, b = 11:12, c = 1:2, d = c("a", "b"))
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"))
  y <- local_db_table(con, df_y, "df_y") |>
    mutate(c = c + 1)

  expect_equal(
    rows_upsert(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      returning = everything()
    ) |>
      get_returned_rows() |>
      arrange(a),
    tibble(
      a = 2:3,
      b = 12:13,
      c = c(-1L, -2L),
      d = c("y", "z")
    )
  )
})

test_that("can delete", {
  con <- test_mssql()

  df_x <- tibble(a = 1:3, b = 11:13, c = 1:3, d = c("a", "b", "c"))
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 2:3, b = c(12L, 13L))
  y <- local_db_table(con, df_y, "df_y")

  expect_equal(
    rows_delete(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      unmatched = "ignore"
    ) |>
      collect(),
    tibble(
      a = 1L,
      b = 11L,
      c = 1L,
      d = "a"
    )
  )
})

test_that("can delete with returning", {
  con <- test_mssql()

  df_x <- tibble(a = 1:3, b = 11:13, c = 1:3, d = c("a", "b", "c"))
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 2:3, b = c(12L, 13L))
  y <- local_db_table(con, df_y, "df_y")

  expect_equal(
    rows_delete(
      x,
      y,
      by = c("a", "b"),
      in_place = TRUE,
      unmatched = "ignore",
      returning = everything()
    ) |>
      get_returned_rows(),
    tibble(
      a = 2:3,
      b = 12:13,
      c = 2:3,
      d = c("b", "c")
    )
  )
})

## -----------------------------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
options(tibble.print_min = 6L, tibble.print_max = 6L, digits = 3)

## ----setup, message = FALSE---------------------------------------------------
library(dplyr)
library(dbplyr)
library(DBI)

## -----------------------------------------------------------------------------
con <- DBI::dbConnect(RSQLite::SQLite(), path = ":memory:")
DBI::dbWriteTable(con, "mtcars", mtcars)

tbl(con, "mtcars")

## -----------------------------------------------------------------------------
#' @importFrom dbplyr dbplyr_edition
#' @export
dbplyr_edition.myConnectionClass <- function(con) 2L

## ----eval = FALSE-------------------------------------------------------------
# #' @export
# sql_dialect.myConnectionClass <- function(con) {
#   new_sql_dialect(
#     "mybackend",
#     quote_identifier = function(x) sql_quote(x, '"')
#   )
# }

## -----------------------------------------------------------------------------
con <- simulate_dbi()

# Create an index
index_name <- "index"
table <- I("schema.table")
columns <- c("column1", "column2")
sql_glue2(con, "CREATE INDEX {.id index_name} ON {.tbl table} {.id columns*}")

# Insert values safely
name <- "O'Brien"
sql_glue2(con, "INSERT INTO students (name) VALUES {name*}")

# Build a query
table <- "my_table"
cols <- c("id", "name", "value")
sql_glue2(con, "SELECT {.id cols} FROM {.tbl table}")

## ----eval = FALSE-------------------------------------------------------------
# # Method on dialect class (recommended)
# sql_translation.sql_dialect_mybackend <- function(con) {
#   sql_variant(
#     scalar = sql_translator(base_scalar, ...), # Functions in SELECT (non-aggregated)
#     aggregate = sql_translator(base_aggregate, ...), # Aggregation functions (mean, sum, etc.)
#     window = sql_translator(base_win, ...) # Window functions (lead, lag, rank, etc.)
#   )
# }

## ----eval = FALSE-------------------------------------------------------------
# sql_translator(
#   base_scalar, # Inherit most translations
#   # Override specific functions for your backend
#   `+` = sql_infix("+"),
#   mean = sql_aggregate("AVG", "mean")
# )

## -----------------------------------------------------------------------------
# sql_translation.sql_dialect_mybackend <- function(con) {
#   sql_variant(
#     scalar = sql_translator(
#       base_scalar,
#       # Standard SQL functions
#       cos = sql_prefix("COS", 1),
#       round = sql_prefix("ROUND", 2),
#       # Infix operators
#       `+` = sql_infix("+"),
#       `*` = sql_infix("*"),
#       `==` = sql_infix("="),
#       # Type casting
#       as.numeric = sql_cast("NUMERIC"),
#       as.character = sql_cast("VARCHAR")
#     ),
#     aggregate = sql_translator(base_agg),
#     window = sql_translator(base_win)
#   )
# }

## -----------------------------------------------------------------------------
# sql_translation.sql_dialect_mybackend <- function(con) {
#   sql_variant(
#     scalar = sql_translator(base_scalar),
#     aggregate = sql_translator(
#       .parent = base_agg,
#       # Single-argument aggregates
#       mean = sql_aggregate("AVG", "mean"),
#       var = sql_aggregate("VAR_SAMP", "var"),
#       # Two-argument aggregates
#       cov = sql_aggregate_2("COVAR_SAMP"),
#       # Variadic aggregates
#       pmin = sql_aggregate_n("LEAST", "pmin"),
#       pmax = sql_aggregate_n("GREATEST", "pmax"),
#       # Unsupported functions
#       median = sql_not_supported("median")
#     ),
#     window = sql_translator(base_win)
#   )
# }

## ----eval = FALSE-------------------------------------------------------------
# window = sql_translator(
#   base_win,
#   # Ranking functions
#   row_number = win_rank("ROW_NUMBER"),
#   rank = win_rank("RANK"),
#   dense_rank = win_rank("DENSE_RANK"),
#   # Aggregate functions as window functions
#   mean = win_aggregate("AVG"),
#   sum = win_aggregate("SUM"),
#   # Cumulative functions
#   cumsum = win_cumulative("SUM"),
#   # Absent functions
#   cume_dist = win_absent("cume_dist")
# )

## ----eval = FALSE-------------------------------------------------------------
# scalar = sql_translator(
#   base_scalar,
# 
#   # Custom log function with change of base
#   log = function(x, base = exp(1)) {
#     if (isTRUE(all.equal(base, exp(1)))) {
#       sql_glue("LN({x})")
#     } else {
#       sql_glue("LOG({x}) / LOG({base})")
#     }
#   },
# 
#   # Custom paste function using CONCAT
#   paste = function(..., sep = " ") {
#     sql_glue("CONCAT_WS({sep}, {...})")
#   }
# )

new_join_by <- function(exprs = list(),
                        condition = character(),
                        filter = character(),
                        x = character(),
                        y = character()) {
  out <- list(
    exprs = exprs,
    condition = condition,
    filter = filter,
    x = x,
    y = y
  )
  structure(out, class = "dplyr_join_by")
}

# ------------------------------------------------------------------------------

# Internal generic
dbplyr_as_join_by <- function(x, error_call = caller_env()) {
  UseMethod("dbplyr_as_join_by")
}

#' @export
dbplyr_as_join_by.default <- function(x, error_call = caller_env()) {
  message <- glue(paste0(
    "`by` must be a (named) character vector, list, `join_by()` result, ",
    "or NULL, not {obj_type_friendly(x)}."
  ))
  abort(message, call = error_call)
}

#' @export
dbplyr_as_join_by.dplyr_join_by <- function(x, error_call = caller_env()) {
  x
}

#' @export
dbplyr_as_join_by.character <- function(x, error_call = caller_env()) {
  x_names <- names(x) %||% x
  y_names <- unname(x)

  # If x partially named, assume unnamed are the same in both tables
  x_names[x_names == ""] <- y_names[x_names == ""]

  finalise_equi_join_by(x_names, y_names)
}

#' @export
dbplyr_as_join_by.list <- function(x, error_call = caller_env()) {
  # TODO: check lengths
  x_names <- x[["x"]]
  y_names <- x[["y"]]

  if (!is_character(x_names)) {
    abort("`by$x` must evaluate to a character vector.")
  }
  if (!is_character(y_names)) {
    abort("`by$y` must evaluate to a character vector.")
  }

  finalise_equi_join_by(x_names, y_names)
}

finalise_equi_join_by <- function(x_names, y_names) {
  n <- length(x_names)

  if (n == 0L) {
    abort(
      "Backwards compatible support for cross joins should have been caught earlier.",
      .internal = TRUE
    )
  }

  exprs <- purrr::map2(x_names, y_names, function(x, y) expr(!!x == !!y))
  condition <- vctrs::vec_rep("==", times = n)
  filter <- vctrs::vec_rep("none", times = n)

  new_join_by(
    exprs = exprs,
    condition = condition,
    filter = filter,
    x = x_names,
    y = y_names
  )
}

# ------------------------------------------------------------------------------

join_by_common <- function(x_names,
                           y_names,
                           ...,
                           error_call = caller_env()) {
  check_dots_empty0(...)

  by <- intersect(x_names, y_names)

  if (length(by) == 0) {
    message <- c(
      "`by` must be supplied when `x` and `y` have no common variables.",
      i = "Use `cross_join()` to perform a cross-join."
    )
    abort(message, call = error_call)
  }

  by_names <- by
  by_names <- glue::glue_collapse(by_names, sep = ", ")
  inform(glue("Joining with `by = join_by({by_names})`"))

  finalise_equi_join_by(by, by)
}

test_that("new_join_by has useful defaults", {
  out <- new_join_by(c("x", "y"))
  expect_equal(out$y, c("x", "y"))
  expect_equal(out$condition, c("==", "=="))
})

test_that("join_by checks inputs", {
  expect_snapshot(error = TRUE, {
    new_join_by("x", c("x", "y"))
    new_join_by("x", "y", c("<", ">"))
  })
})

#' Subset distinct/unique rows
#'
#' This is a method for the dplyr [distinct()] generic. It adds the
#' `DISTINCT` clause to the SQL query.
#'
#' @inheritParams arrange.tbl_lazy
#' @inheritParams dplyr::distinct
#' @inherit arrange.tbl_lazy return
#' @export
#' @importFrom dplyr distinct
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(x = c(1, 1, 2, 2), y = c(1, 2, 1, 1))
#' db %>% distinct() %>% show_query()
#' db %>% distinct(x) %>% show_query()
distinct.tbl_lazy <- function(.data, ..., .keep_all = FALSE) {
  grps <- syms(op_grps(.data))
  empty_dots <- dots_n(...) == 0
  can_use_distinct <- !.keep_all || (empty_dots && is_empty(grps))
  if (!can_use_distinct) {
    needs_dummy_order <- is.null(op_sort(.data))

    if (needs_dummy_order) {
      dummy_order_vars <- colnames(.data)[[1]]
      .data <- .data %>% window_order(!!sym(dummy_order_vars))
    }

    .data <- .data %>%
      group_by(..., .add = TRUE) %>%
      filter(row_number() == 1L) %>%
      group_by(!!!grps)

    if (needs_dummy_order) {
      .data <- .data %>% window_order()
    }

    return(.data)
  }

  if (empty_dots) {
    dots <- quos(!!!syms(colnames(.data)))
  } else {
    dots <- partial_eval_dots(.data, ..., .named = FALSE)
    dots <- quos(!!!dots)
  }
  prep <- distinct_prepare_compat(.data, dots, group_vars = group_vars(.data))
  out <- dplyr::select(prep$data, prep$keep)

  out$lazy_query <- add_distinct(out)
  out
}

# copied from dplyr with minor changes (names -> colnames)
# https://github.com/tidyverse/dplyr/blob/main/R/distinct.R
distinct_prepare_compat <- function(.data,
                             vars,
                             group_vars = character(),
                             .keep_all = FALSE,
                             caller_env = caller_env(2),
                             error_call = caller_env()
                             ) {
  stopifnot(is_quosures(vars), is.character(group_vars))

  # If no input, keep all variables
  if (length(vars) == 0) {
    return(list(
      data = .data,
      vars = seq_along(.data),
      keep = seq_along(.data)
    ))
  }

  # If any calls, use mutate to add new columns, then distinct on those
  computed_columns <- add_computed_columns(.data, vars, error_call = error_call)
  .data <- computed_columns$data
  distinct_vars <- computed_columns$added_names

  # Once we've done the mutate, we no longer need lazy objects, and
  # can instead just use their names
  missing_vars <- setdiff(distinct_vars, colnames(.data))
  if (length(missing_vars) > 0) {
    bullets <- c(
      "Must use existing variables.",
      set_names(glue("`{missing_vars}` not found in `.data`."), rep("x", length(missing_vars)))
    )
    abort(bullets, call = error_call)
  }

  # Only keep unique vars
  distinct_vars <- unique(distinct_vars)
  # Missing grouping variables are added to the front
  new_vars <- c(setdiff(group_vars, distinct_vars), distinct_vars)

  if (.keep_all) {
    keep <- seq_along(.data)
  } else {
    keep <- new_vars
  }

  list(data = .data, vars = new_vars, keep = keep)
}

# copied from dplyr
# https://github.com/tidyverse/dplyr/blob/main/R/group-by.R#L243
add_computed_columns <- function(.data,
                                 vars,
                                 error_call = caller_env()) {
  is_symbol <- purrr::map_lgl(vars, quo_is_variable_reference)
  needs_mutate <- have_name(vars) | !is_symbol

  if (any(needs_mutate)) {
    out <- mutate(.data, !!!vars)
    col_names <- names(exprs_auto_name(vars))
  } else {
    out <- .data
    col_names <- names(exprs_auto_name(vars))
  }

  list(data = out, added_names = col_names)
}

# copied from dplyr
# https://github.com/tidyverse/dplyr/blob/main/R/group-by.R#L276
quo_is_variable_reference <- function(quo) {
  if (quo_is_symbol(quo)) {
    return(TRUE)
  }

  if (quo_is_call(quo, n = 2)) {
    expr <- quo_get_expr(quo)

    if (is_call(expr, c("$", "[["))) {
      if (!identical(expr[[2]], sym(".data"))) {
        return(FALSE)
      }

      param <- expr[[3]]
      if (is_symbol(param) || is_string(param)) {
        return(TRUE)
      }
    }
  }

  FALSE
}


add_distinct <- function(.data) {
  lazy_query <- .data$lazy_query

  out <- lazy_select_query(
    x = lazy_query,
    distinct = TRUE
  )
  # TODO this could also work for joins
  if (!is_lazy_select_query(lazy_query)) {
    return(out)
  }

  # Optimisation overview
  # * `distinct()` adds the `DISTINCT` clause to `SELECT`
  # * `WHERE`, `GROUP BY`, and `HAVING` are executed before `SELECT`
  #   => they do not matter
  # * `ORDER BY`
  #   => but `arrange()` should not have an influence on `distinct()` so it
  #      should not matter
  # * `LIMIT` are executed after `SELECT`
  #   => needs a subquery
  if (!is_null(lazy_query$limit)) {
    return(out)
  }

  lazy_query$distinct <- TRUE
  lazy_query
}

#' Cache and retrieve an `src_sqlite` of the Lahman baseball database.
#'
#' This creates an interesting database using data from the Lahman baseball
#' data source, provided by Sean Lahman at
#' \url{https://www.seanlahman.com/baseball-archive/statistics/}, and
#' made easily available in R through the \pkg{Lahman} package by
#' Michael Friendly, Dennis Murphy and Martin Monkman. See the documentation
#' for that package for documentation of the individual tables.
#'
#' @param ... Other arguments passed to `src` on first
#'   load. For MySQL and PostgreSQL, the defaults assume you have a local
#'   server with `lahman` database already created.
#'   For `lahman_srcs()`, character vector of names giving srcs to generate.
#' @param quiet if `TRUE`, suppress messages about databases failing to
#'   connect.
#' @param type src type.
#' @keywords internal
#' @examples
#' # Connect to a local sqlite database, if already created
#' \donttest{
#' library(dplyr)
#'
#' if (has_lahman("sqlite")) {
#'   lahman_sqlite()
#'   batting <- tbl(lahman_sqlite(), "Batting")
#'   batting
#' }
#'
#' # Connect to a local postgres database with lahman database, if available
#' if (has_lahman("postgres")) {
#'   lahman_postgres()
#'   batting <- tbl(lahman_postgres(), "Batting")
#' }
#' }
#' @name lahman
NULL

# nocov start
#' @export
#' @rdname lahman
lahman_sqlite <- function(path = NULL) {
  path <- db_location(path, "lahman.sqlite")
  con <- DBI::dbConnect(RSQLite::SQLite(), dbname = path)
  copy_lahman(con)
}

#' @export
#' @rdname lahman
lahman_postgres <- function(dbname = "lahman", host = "localhost", ...) {
  con <- DBI::dbConnect(RPostgres::Postgres(), dbname = dbname, host = host, ...)
  copy_lahman(con)
}

#' @export
#' @rdname lahman
lahman_mysql <- function(dbname = "lahman", ...) {
  con <- DBI::dbConnect(RMariaDB::MariaDB(), dbname = dbname, ...)
  copy_lahman(con)
}

#' @rdname lahman
#' @export
copy_lahman <- function(con, ...) {
  # Create missing tables
  tables <- setdiff(lahman_tables(), DBI::dbListTables(con))
  for (table in tables) {
    df <- getExportedValue("Lahman", table)
    message("Creating table: ", table)

    ids <- as.list(names(df)[grepl("ID$", names(df))])
    copy_to(con, df, table, indexes = ids, temporary = FALSE)
  }

  invisible(con)
}
# Get list of all non-label data frames in package
lahman_tables <- function() {
  tables <- utils::data(package = "Lahman")$results[, 3]
  tables[!grepl("Labels", tables)]
}

#' @rdname lahman
#' @export
has_lahman <- function(type, ...) {
  if (!requireNamespace("Lahman", quietly = TRUE)) return(FALSE)
  succeeds(lahman(type, ...), quiet = FALSE)
}

#' @rdname lahman
#' @export
lahman_srcs <- function(..., quiet = NULL) {
  load_srcs(lahman, c(...), quiet = quiet)
}

lahman <- function(type, ...) {
  f <- match.fun(paste0("lahman_", type))
  f(...)
}
# nocov end

#' MySQL/MariaDB backend
#'
#' @description
#' This backend supports MySQL and MariaDB databases, typically accessed via
#' `MySQLConnection` or `MariaDBConnection` created by [DBI::dbConnect()]. Use
#' `dialect_mysql()` with `lazy_frame()` to see simulated SQL without connecting
#' to a live database.
#'
#' Key differences for this backend are:
#'
#' * `paste()` uses `CONCAT_WS()`
#' * String translations for `str_detect()`, `str_locate()`, and
#'   `str_replace_all()`
#' * Clear error message for unsupported full joins
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-mysql
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_mysql())
#' lf |> transmute(x = paste0(d, " times"))
NULL

#' @export
#' @rdname backend-mysql
dialect_mariadb <- function() {
  new_sql_dialect(
    "mariadb",
    quote_identifier = function(x) sql_quote(x, "`"),
    has_window_clause = TRUE
  )
}

#' @export
#' @rdname backend-mysql
dialect_mysql <- function() {
  new_sql_dialect(
    "mysql",
    quote_identifier = function(x) sql_quote(x, "`"),
    has_window_clause = TRUE
  )
}

#' @export
#' @rdname backend-mysql
simulate_mysql <- function() simulate_dbi("MySQLConnection")

#' @export
#' @rdname backend-mysql
simulate_mariadb <- function() simulate_dbi("MariaDBConnection")

#' @export
sql_dialect.MariaDBConnection <- function(con) {
  dialect_mariadb()
}
#' @export
sql_dialect.MySQL <- function(con) {
  dialect_mysql()
}
#' @export
sql_dialect.MySQLConnection <- function(con) {
  dialect_mysql()
}

#' @export
dbplyr_edition.MariaDBConnection <- function(con) {
  2L
}
#' @export
dbplyr_edition.MySQL <- dbplyr_edition.MariaDBConnection
#' @export
dbplyr_edition.MySQLConnection <- dbplyr_edition.MariaDBConnection

#' @export
db_connection_describe.MariaDBConnection <- function(con, ...) {
  info <- DBI::dbGetInfo(con)

  paste0(
    "mysql ",
    info$serverVersion,
    " [",
    info$username,
    "@",
    info$host,
    ":",
    info$port,
    "/",
    info$dbname,
    "]"
  )
}
#' @export
db_connection_describe.MySQL <- db_connection_describe.MariaDBConnection
#' @export
db_connection_describe.MySQLConnection <- db_connection_describe.MariaDBConnection

#' @export
db_col_types.MariaDBConnection <- function(con, table, call = caller_env()) {
  sql <- sql_glue2(con, "SHOW COLUMNS FROM {.tbl table}")
  col_info_df <- DBI::dbGetQuery(con, sql)
  set_names(col_info_df[["Type"]], col_info_df[["Field"]])
}
#' @export
db_col_types.MySQL <- db_col_types.MariaDBConnection
#' @export
db_col_types.MySQLConnection <- db_col_types.MariaDBConnection

#' @export
sql_translation.sql_dialect_mariadb <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_scalar,
      # https://dev.mysql.com/doc/refman/5.7/en/comparison-operators.html#operator_equal-to
      is_distinct_from = \(x, y) sql_glue("NOT (({x}) <=> ({y}))"),
      is_not_distinct_from = \(x, y) sql_glue("({x}) <=> ({y})"),
      # basic type casts as per:
      # https://mariadb.com/kb/en/cast/
      # https://dev.mysql.com/doc/refman/8.0/en/cast-functions.html#function_cast
      # https://cran.r-project.org/doc/manuals/r-release/R-lang.html#Vector-objects
      as.logical = function(x) {
        sql_glue("IF({x}, TRUE, FALSE)")
      },
      as.character = sql_cast("CHAR"),
      as.numeric = sql_cast("DOUBLE"),
      as.double = sql_cast("DOUBLE"),
      as.POSIXct = sql_cast("DATETIME"),
      as_datetime = sql_cast("DATETIME"),
      # Neither MySQL nor MariaDB support CASTing to BIGINT. MariaDB may
      # silently cast an INTEGER into a BIGINT type, MySQL outright fails.
      # https://dba.stackexchange.com/a/205822
      as.integer64 = sql_cast("INTEGER"),

      runif = function(n = n(), min = 0, max = 1) {
        sql_runif("RAND()", n = {{ n }}, min = min, max = max)
      },

      # string functions ------------------------------------------------
      paste = sql_paste(" "),
      paste0 = sql_paste(""),

      # stringr
      str_c = sql_paste(""),
      # https://dev.mysql.com/doc/refman/8.0/en/regexp.html
      # NB: case insensitive by default; could use REGEXP_LIKE for MySQL,
      # but available in MariaDB. A few more details at:
      # https://www.oreilly.com/library/view/mysql-cookbook/0596001452/ch04s11.html
      str_detect = sql_infix("REGEXP"),
      str_like = function(string, pattern, ignore_case = deprecated()) {
        ignore_case <- deprecate_ignore_case(ignore_case)
        if (ignore_case) {
          sql_glue("{string} LIKE {pattern}")
        } else {
          sql_glue("{string} LIKE BINARY {pattern}")
        }
      },
      str_ilike = function(string, pattern) {
        # MySQL's LIKE is case-insensitive by default
        sql_glue("{string} LIKE {pattern}")
      },
      str_locate = function(string, pattern) {
        sql_glue("REGEXP_INSTR({string}, {pattern})")
      },
      str_replace_all = function(string, pattern, replacement) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement})")
      }
    ),
    sql_translator(
      .parent = base_agg,
      sd = sql_aggregate("STDDEV_SAMP", "sd"),
      var = sql_aggregate("VAR_SAMP", "var"),
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        sql_glue("GROUP_CONCAT({x} SEPARATOR {collapse})")
      }
    ),
    sql_translator(
      .parent = base_win,
      sd = win_aggregate("STDDEV_SAMP"),
      var = win_aggregate("VAR_SAMP"),
      # GROUP_CONCAT not currently available as window function
      # https://mariadb.com/kb/en/library/aggregate-functions-as-window-functions/
      str_flatten = win_absent("str_flatten")
    )
  )
}

#' @export
sql_translation.sql_dialect_mysql <- function(con) {
  maria <- unclass(sql_translation(dialect_mariadb()))
  sql_variant(
    sql_translator(
      .parent = maria$scalar,
      # SIGNED INTEGER is a BIGINT; no way to cast smaller
      as.integer = sql_cast("SIGNED INTEGER"),
      as.integer64 = sql_cast("SIGNED INTEGER"),
    ),
    maria$aggregate,
    maria$window
  )
}

#' @export
sql_table_analyze.sql_dialect_mariadb <- function(con, table, ...) {
  sql_glue2(con, "ANALYZE TABLE {.tbl table}")
}
#' @export
sql_table_analyze.sql_dialect_mysql <- sql_table_analyze.sql_dialect_mariadb

#' @export
sql_query_join.sql_dialect_mariadb <- function(
  con,
  x,
  y,
  select,
  type = "inner",
  by = NULL,
  ...
) {
  if (identical(type, "full")) {
    cli_abort("MySQL does not support full joins")
  }
  NextMethod()
}
#' @export
sql_query_join.sql_dialect_mysql <- sql_query_join.sql_dialect_mariadb


# https://modern-sql.com/blog/2018-08/whats-new-in-mariadb-10.3#3.values
# MariaDB doesn't accept `ROW` unlike MySQL
#' @export
sql_values_subquery.sql_dialect_mariadb <- sql_values_subquery.sql_dialect

#' @export
sql_values_subquery.sql_dialect_mysql <- function(
  con,
  df,
  types,
  lvl = 0,
  ...
) {
  # https://dev.mysql.com/doc/refman/8.0/en/values.html
  sql_values_subquery_default(con, df, types = types, lvl = lvl, row = TRUE)
}

#' @export
sql_query_update_from.sql_dialect_mariadb <- function(
  con,
  table,
  from,
  by,
  update_values,
  ...,
  returning_cols = NULL
) {
  table <- as_table_path(table, con)

  if (!is_empty(returning_cols)) {
    check_unsupported_arg(returning_cols, backend = "MariaDB")
  }

  # https://stackoverflow.com/a/19346375/946850
  parts <- rows_prep(con, table, from, by, lvl = 0)
  update_cols <- sql_table_prefix(con, table, names(update_values))

  table_sql <- sql_escape_table_source(con, table)
  clauses <- list(
    sql_clause_update(table_sql),
    sql_clause("INNER JOIN", parts$from),
    sql_clause_on(parts$where, lvl = 1),
    sql_clause_set(update_cols, update_values),
    sql_returning_cols(con, returning_cols, table)
  )
  sql_format_clauses(clauses, lvl = 0)
}
#' @export
sql_query_update_from.sql_dialect_mysql <- sql_query_update_from.sql_dialect_mariadb


#' @export
sql_query_upsert.sql_dialect_mariadb <- function(
  con,
  table,
  from,
  by,
  update_cols,
  ...,
  returning_cols = NULL,
  method = NULL
) {
  cli_abort("{.fun rows_upsert} is not supported for MariaDB.")
}
#' @export
sql_query_upsert.sql_dialect_mysql <- sql_query_upsert.sql_dialect_mariadb

#' @export
sql_escape_datetime.sql_dialect_mariadb <- function(con, x) {
  # DateTime format as per:
  # https://dev.mysql.com/doc/refman/8.0/en/datetime.html
  # https://mariadb.com/kb/en/datetime/
  x <- strftime(x, "%Y-%m-%d %H:%M:%OS", tz = "UTC")
  sql_escape_string(con, x)
}
#' @export
sql_escape_datetime.sql_dialect_mysql <- sql_escape_datetime.sql_dialect_mariadb

test_that("simulate_mysql() and simulate_mariadb() still work", {
  expect_translation(simulate_mysql(), x + 1, "`x` + 1.0")
  expect_translation(simulate_mariadb(), x + 1, "`x` + 1.0")
})

# function translation ----------------------------------------------------

test_that("custom scalar translated correctly", {
  con <- dialect_mysql()
  expect_translation(con, as.logical(1L), "IF(1, TRUE, FALSE)")

  expect_translation(con, str_locate("abc", "b"), "REGEXP_INSTR('abc', 'b')")
  expect_translation(
    con,
    str_replace_all("abc", "b", "d"),
    "REGEXP_REPLACE('abc', 'b', 'd')"
  )
})

test_that("custom aggregators translated correctly", {
  con <- dialect_mysql()

  expect_translation(
    con,
    str_flatten(y, ","),
    "GROUP_CONCAT(`y` SEPARATOR ',')",
    window = FALSE
  )
})

test_that("use CHAR type for as.character", {
  con <- dialect_mysql()
  expect_translation(con, as.character(x), "CAST(`x` AS CHAR)")
})

test_that("custom date escaping works as expected", {
  con <- dialect_mysql()

  expect_equal(escape(as.Date("2020-01-01"), con = con), sql("'2020-01-01'"))
  expect_equal(escape(as.Date(NA), con = con), sql("NULL"))

  expect_equal(
    escape(as.POSIXct("2020-01-01", tz = "UTC"), con = con),
    sql("'2020-01-01 00:00:00'")
  )
  expect_equal(escape(as.POSIXct(NA, tz = "UTC"), con = con), sql("NULL"))
})


test_that("custom stringr functions translated correctly", {
  con <- dialect_mysql()

  expect_translation(con, str_c(x, y), "CONCAT_WS('', `x`, `y`)")
  expect_translation(con, str_detect(x, y), "`x` REGEXP `y`")
  expect_translation(con, str_like(x, y), "`x` LIKE BINARY `y`")
  expect_translation(con, str_ilike(x, y), "`x` LIKE `y`")

  expect_translation(con, str_locate(x, y), "REGEXP_INSTR(`x`, `y`)")
  expect_translation(
    con,
    str_replace_all(x, y, z),
    "REGEXP_REPLACE(`x`, `y`, `z`)"
  )
})

# verbs -------------------------------------------------------------------

test_that("generates custom sql", {
  con_maria <- dialect_mariadb()

  expect_snapshot(sql_table_analyze(con_maria, in_schema("schema", "tbl")))
  expect_snapshot(sql_query_explain(con_maria, sql("SELECT * FROM table")))

  lf <- lazy_frame(x = 1, con = con_maria)
  expect_snapshot(left_join(lf, lf, by = "x", na_matches = "na"))
  expect_snapshot(error = TRUE, full_join(lf, lf, by = "x"))

  expect_snapshot(slice_sample(lf, n = 1))

  expect_snapshot(
    copy_inline(con_maria, tibble(x = 1:2, y = letters[1:2])) |> remote_query()
  )

  con_mysql <- dialect_mysql()
  expect_snapshot(
    copy_inline(con_mysql, tibble(x = 1:2, y = letters[1:2])) |> remote_query()
  )
})

test_that("`sql_query_update_from()` is correct", {
  con <- dialect_mysql()
  df_y <- lazy_frame(
    a = 2:3,
    b = c(12L, 13L),
    c = -(2:3),
    d = c("y", "z"),
    con = con,
    .name = "df_y"
  ) |>
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_update_from(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      by = c("a", "b"),
      update_values = sql(
        c = "COALESCE(`df_x`.`c`, `...y`.`c`)",
        d = "`...y`.`d`"
      )
    )
  )

  expect_snapshot_error(
    sql_query_update_from(
      con = con,
      table = ident("df_x"),
      from = sql_render(df_y, con, lvl = 1),
      by = c("a", "b"),
      update_values = sql(
        c = "COALESCE(`df_x`.`c`, `...y`.`c`)",
        d = "`...y`.`d`"
      ),
      returning_cols = c("a", b2 = "b")
    )
  )
})

# live database -----------------------------------------------------------

test_that("logicals converted to integer correctly", {
  db <- copy_to(
    test_mariadb(),
    data.frame(x = c(TRUE, FALSE, NA)),
    name = "test",
    overwrite = TRUE
  )
  expect_identical(db |> pull(), c(1L, 0L, NA))
})

test_that("can explain", {
  db <- copy_to(
    test_mariadb(),
    data.frame(x = 1:3),
    name = "test",
    overwrite = TRUE
  )
  expect_snapshot(db |> mutate(y = x + 1) |> explain())
})

test_that("can overwrite temp tables", {
  con <- test_mariadb()

  df1 <- tibble(x = 1)
  copy_to(con, df1, "test-df", temporary = TRUE)
  withr::defer(DBI::dbRemoveTable(con, "test-df"))

  df2 <- tibble(x = 2)
  db2 <- copy_to(con, df2, "test-df", temporary = TRUE, overwrite = TRUE)
  expect_equal(collect(db2), df2)
})

test_that("can update", {
  con <- test_mariadb()

  df_x <- tibble(a = 1:3, b = 11:13, c = 1:3, d = c("a", "b", "c"))
  x <- local_db_table(con, df_x, "df_x")
  df_y <- tibble(a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"))
  y <- local_db_table(con, df_y, "df_y") |> mutate(c = c + 1)

  # `RETURNING` in an `UPDATE` clause is not (yet) supported for MariaDB
  # https://jira.mariadb.org/browse/MDEV-5092
  rows_update(x, y, by = c("a", "b"), unmatched = "ignore", in_place = TRUE)
  expect_equal(
    collect(x),
    tibble(
      a = 1:3,
      b = 11:13,
      c = c(1L, -1L, -2L),
      d = c("a", "y", "z")
    )
  )
})

test_that("casts integer, bigint, and longtext columns", {
  con <- test_mariadb()

  df1 <- tibble(id = 1L, val = 10L, ltext = strrep("a", times = 65535 + 2))
  df2 <- tibble(id = "2", val = 20, ltext = strrep("b", times = 65535 + 2))
  out <- tibble(
    # output is a bigint, since that's all MariaDB has
    id = bit64::as.integer64(1:2),
    val = bit64::as.integer64(c(10L, 20L)),
    ltext = c(df1$ltext, df2$ltext)
  )

  types <- c(id = "integer", val = "bigint", ltext = "longtext")

  db1 <- local_db_table(con, df1, "df1", types = types, temporary = FALSE)
  db2 <- local_db_table(con, df2, "df2", types = types, temporary = FALSE)

  # Via query
  expect_equal(collect(rows_append(db1, db2, copy = TRUE)), out)

  # Via modification
  rows_append(db1, db2, copy = TRUE, in_place = TRUE)
  expect_equal(collect(db1), out)

  types_expected <- c(id = "int(11)", val = "bigint(20)", ltext = "longtext")
  expect_equal(db_col_types(con, "df2"), types_expected)
})

#' Snowflake backend
#'
#' @description
#' This backend supports Snowflake databases, typically accessed via odbc. Use
#' `dialect_snowflake()` with `lazy_frame()` to see simulated SQL without
#' connecting to a live database.
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-snowflake
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_snowflake())
#' lf |> transmute(x = paste0(d, " times"))
NULL

#' @export
#' @rdname backend-snowflake
dialect_snowflake <- function() {
  new_sql_dialect(
    "snowflake",
    quote_identifier = function(x) sql_quote(x, '"')
  )
}

#' @export
#' @rdname backend-snowflake
simulate_snowflake <- function() simulate_dbi("Snowflake")

#' @export
sql_dialect.Snowflake <- function(con) {
  dialect_snowflake()
}

#' @export
dbplyr_edition.Snowflake <- function(con) {
  2L
}

#' @export
sql_translation.sql_dialect_snowflake <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_odbc_scalar,
      log10 = \(x) sql_glue("LOG(10, {x})"),
      grepl = snowflake_grepl,
      round = snowflake_round,
      paste = snowflake_paste(" "),
      paste0 = snowflake_paste(""),
      str_c = function(..., sep = "", collapse = NULL) {
        check_string(sep)
        if (!is.null(collapse)) {
          cli_abort(c(
            "{.arg collapse} not supported in DB translation of {.fn str_c}.",
            i = "Please use {.fn str_flatten} instead."
          ))
        }
        sql_glue("CONCAT_WS({sep}, {...})")
      },
      str_like = function(string, pattern, ignore_case = deprecated()) {
        ignore_case <- deprecate_ignore_case(ignore_case)
        if (ignore_case) {
          sql_glue("{string} ILIKE {pattern}")
        } else {
          sql_glue("{string} LIKE {pattern}")
        }
      },
      str_ilike = function(string, pattern) {
        sql_glue("{string} ILIKE {pattern}")
      },
      str_locate = function(string, pattern) {
        sql_glue("POSITION({pattern}, {string})")
      },
      str_detect = function(string, pattern, negate = FALSE) {
        con <- sql_current_con()
        check_bool(negate)

        # Snowflake needs backslashes escaped, so we must increase the level of escaping
        pattern <- gsub("\\", "\\\\", pattern, fixed = TRUE)
        if (negate) {
          sql_glue("REGEXP_INSTR({string}, {pattern}) = 0")
        } else {
          sql_glue("REGEXP_INSTR({string}, {pattern}) != 0")
        }
      },
      str_starts = function(string, pattern, negate = FALSE) {
        con <- sql_current_con()
        check_bool(negate)

        # Snowflake needs backslashes escaped, so we must increase the level of escaping
        pattern <- gsub("\\", "\\\\", pattern, fixed = TRUE)
        if (negate) {
          sql_glue("REGEXP_INSTR({string}, {pattern}) != 1")
        } else {
          sql_glue("REGEXP_INSTR({string}, {pattern}) = 1")
        }
      },
      str_ends = function(string, pattern, negate = FALSE) {
        con <- sql_current_con()
        check_bool(negate)

        # Snowflake needs backslashes escaped, so we must increase the level of escaping
        pattern <- gsub("\\", "\\\\", pattern, fixed = TRUE)
        if (negate) {
          sql_glue(
            "REGEXP_INSTR({string}, {pattern}, 1, 1, 1) != (LENGTH({string}) + 1)"
          )
        } else {
          sql_glue(
            "REGEXP_INSTR({string}, {pattern}, 1, 1, 1) = (LENGTH({string}) + 1)"
          )
        }
      },
      # On Snowflake, REGEXP_REPLACE is used like this:
      # REGEXP_REPLACE( <subject> , <pattern> [ , <replacement> ,
      #                <position> , <occurrence> , <parameters> ] )
      # so we must set <occurrence> to 1 if not replacing all.  See:
      # https://docs.snowflake.com/en/sql-reference/functions/regexp_replace.html
      # Also, Snowflake needs backslashes escaped, so we must increase the
      # level of escaping by 1
      str_replace = function(string, pattern, replacement) {
        pattern <- gsub("\\", "\\\\", pattern, fixed = TRUE)
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement}, 1, 1)")
      },
      str_replace_all = function(string, pattern, replacement) {
        pattern <- gsub("\\", "\\\\", pattern, fixed = TRUE)
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement})")
      },
      str_remove = function(string, pattern) {
        pattern <- gsub("\\", "\\\\", pattern, fixed = TRUE)
        sql_glue("REGEXP_REPLACE({string}, {pattern}, '', 1, 1)")
      },
      str_remove_all = function(string, pattern) {
        pattern <- gsub("\\", "\\\\", pattern, fixed = TRUE)
        sql_glue("REGEXP_REPLACE({string}, {pattern})")
      },
      str_trim = function(string) {
        sql_glue("TRIM({string})")
      },
      str_squish = function(string) {
        sql_glue("REGEXP_REPLACE(TRIM({string}), '\\\\s+', ' ')")
      },

      # lubridate functions
      # https://docs.snowflake.com/en/sql-reference/functions-date-time.html
      day = function(x) {
        sql_glue("EXTRACT(DAY FROM {x})")
      },
      mday = function(x) {
        sql_glue("EXTRACT(DAY FROM {x})")
      },
      wday = function(x, label = FALSE, abbr = TRUE, week_start = NULL) {
        check_bool(label)
        check_bool(abbr)
        check_number_whole(week_start, allow_null = TRUE)
        if (!label) {
          week_start <- week_start %||% getOption("lubridate.week.start", 7)
          offset <- as.integer(7 - week_start)
          sql_glue("EXTRACT('dayofweek', DATE({x}) + {offset}) + 1")
        } else if (label && !abbr) {
          sql_glue(
            "DECODE(EXTRACT('dayofweek', {x}), 1, 'Monday', 2, 'Tuesday', 3, 'Wednesday', 4, 'Thursday', 5, 'Friday', 6, 'Saturday', 0, 'Sunday')"
          )
        } else if (label && abbr) {
          sql_glue("DAYNAME({x})")
        } else {
          cli::cli_abort("Unrecognized arguments", call = quote(wday()))
        }
      },
      yday = \(x) sql_glue("EXTRACT('dayofyear', {x})"),
      week = function(x) {
        sql_glue("FLOOR((EXTRACT('dayofyear', {x}) - 1) / 7) + 1")
      },
      isoweek = \(x) sql_glue("EXTRACT('weekiso', {x})"),
      month = function(x, label = FALSE, abbr = TRUE) {
        check_bool(label)
        check_bool(abbr)
        if (!label) {
          sql_glue("EXTRACT('month', {x})")
        } else {
          if (abbr) {
            sql_glue("MONTHNAME({x})")
          } else {
            sql_glue(
              "DECODE(EXTRACT('month', {x}), 1, 'January', 2, 'February', 3, 'March', 4, 'April', 5, 'May', 6, 'June', 7, 'July', 8, 'August', 9, 'September', 10, 'October', 11, 'November', 12, 'December')"
            )
          }
        }
      },
      quarter = function(x, with_year = FALSE, fiscal_start = 1) {
        check_bool(with_year)
        check_unsupported_arg(fiscal_start, 1)

        if (with_year) {
          sql_glue("(EXTRACT('year', {x}) || '.' || EXTRACT('quarter', {x}))")
        } else {
          sql_glue("EXTRACT('quarter', {x})")
        }
      },
      isoyear = function(x) {
        sql_glue("EXTRACT('year', {x})")
      },
      seconds = function(x) {
        sql_glue("INTERVAL '{x} second'")
      },
      minutes = function(x) {
        sql_glue("INTERVAL '{x} minute'")
      },
      hours = function(x) {
        sql_glue("INTERVAL '{x} hour'")
      },
      days = function(x) {
        sql_glue("INTERVAL '{x} day'")
      },
      weeks = function(x) {
        sql_glue("INTERVAL '{x} week'")
      },
      months = function(x) {
        sql_glue("INTERVAL '{x} month'")
      },
      years = function(x) {
        sql_glue("INTERVAL '{x} year'")
      },
      # https://docs.snowflake.com/en/sql-reference/functions/date_trunc.html
      floor_date = function(x, unit = "seconds") {
        unit <- arg_match(
          unit,
          c(
            "second",
            "minute",
            "hour",
            "day",
            "week",
            "month",
            "quarter",
            "year",
            "seconds",
            "minutes",
            "hours",
            "days",
            "weeks",
            "months",
            "quarters",
            "years"
          )
        )
        sql_glue("DATE_TRUNC({unit}, {x})")
      },
      # clock ---------------------------------------------------------------
      add_days = function(x, n, ...) {
        check_dots_empty()
        sql_glue("DATEADD(DAY, {n}, {x})")
      },
      add_years = function(x, n, ...) {
        check_dots_empty()
        sql_glue("DATEADD(YEAR, {n}, {x})")
      },
      date_build = function(year, month = 1L, day = 1L, ..., invalid = NULL) {
        check_dots_empty()
        check_unsupported_arg(invalid, allowed = NULL)
        # https://docs.snowflake.com/en/sql-reference/functions/date_from_parts
        sql_glue("DATE_FROM_PARTS({year}, {month}, {day})")
      },
      get_year = function(x) {
        sql_glue("DATE_PART(YEAR, {x})")
      },
      get_month = function(x) {
        sql_glue("DATE_PART(MONTH, {x})")
      },
      get_day = function(x) {
        sql_glue("DATE_PART(DAY, {x})")
      },
      date_count_between = function(start, end, precision, ..., n = 1L) {
        check_dots_empty()
        check_unsupported_arg(precision, allowed = "day")
        check_unsupported_arg(n, allowed = 1L)

        sql_glue("DATEDIFF(DAY, {start}, {end})")
      },

      difftime = function(time1, time2, tz, units = "days") {
        check_unsupported_arg(tz)
        check_unsupported_arg(units, allowed = "days")

        sql_glue("DATEDIFF(DAY, {time2}, {time1})")
      },
      # LEAST / GREATEST on Snowflake return NULL if any argument is NULL.
      # The *_IGNORE_NULLS variants skip NULLs, matching `na.rm = TRUE`.
      # https://docs.snowflake.com/en/sql-reference/functions/least
      # https://docs.snowflake.com/en/sql-reference/functions/greatest
      # https://docs.snowflake.com/en/sql-reference/functions/least_ignore_nulls
      # https://docs.snowflake.com/en/sql-reference/functions/greatest_ignore_nulls
      pmin = function(..., na.rm = FALSE) {
        dots <- list(...)
        if (identical(na.rm, TRUE)) {
          sql_glue("LEAST_IGNORE_NULLS({dots})")
        } else {
          sql_glue("LEAST({dots})")
        }
      },
      pmax = function(..., na.rm = FALSE) {
        dots <- list(...)
        if (identical(na.rm, TRUE)) {
          sql_glue("GREATEST_IGNORE_NULLS({dots})")
        } else {
          sql_glue("GREATEST({dots})")
        }
      },
      `$` = function(x, name) {
        if (is.ident(x)) {
          sql_glue("{x}:{name}")
        } else {
          eval(bquote(`$`(x, .(substitute(name)))))
        }
      }
    ),
    sql_translator(
      .parent = base_agg,
      cor = sql_aggregate_2("CORR"),
      cov = sql_aggregate_2("COVAR_SAMP"),
      all = sql_aggregate("BOOLAND_AGG", "all"),
      any = sql_aggregate("BOOLOR_AGG", "any"),
      sd = sql_aggregate("STDDEV", "sd"),
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        sql_glue("LISTAGG({x}, {collapse})")
      }
    ),
    sql_translator(
      .parent = base_win,
      cor = win_aggregate_2("CORR"),
      cov = win_aggregate_2("COVAR_SAMP"),
      all = win_aggregate("BOOLAND_AGG"),
      any = win_aggregate("BOOLOR_AGG"),
      sd = win_aggregate("STDDEV"),
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        win_over(
          sql_glue("LISTAGG({x}, {collapse})"),
          partition = win_current_group(),
          order = win_current_order()
        )
      },
      row_number = win_rank("ROW_NUMBER", empty_order = TRUE)
    )
  )
}

# There seems to be no concept of ANALYZE TABLE in Snowflake.  I searched for
# functions that performed similar operations, and found none.
# Link to full list: https://docs.snowflake.com/en/sql-reference/sql-all.html
#' @export
sql_table_analyze.sql_dialect_snowflake <- function(con, table, ...) {}

snowflake_grepl <- function(
  pattern,
  x,
  ignore.case = FALSE,
  perl = FALSE,
  fixed = FALSE,
  useBytes = FALSE
) {
  con <- sql_current_con()

  check_unsupported_arg(perl, FALSE, backend = "Snowflake")
  check_unsupported_arg(fixed, FALSE, backend = "Snowflake")
  check_unsupported_arg(useBytes, FALSE, backend = "Snowflake")

  # https://docs.snowflake.com/en/sql-reference/functions/regexp_instr.html
  # REGEXP_INSTR optional parameters: position, occurrence, option, regex_parameters
  regexp_parameters <- "c"
  if (ignore.case) {
    regexp_parameters <- "i"
  }
  # Snowflake needs backslashes escaped, so we must increase the level of escaping
  pattern <- gsub("\\", "\\\\", pattern, fixed = TRUE)
  sql_glue("REGEXP_INSTR({x}, {pattern}, 1, 1, 0, {regexp_parameters}) != 0")
}

snowflake_round <- function(x, digits = 0L) {
  digits <- as.integer(digits)
  sql_glue("ROUND(({x})::FLOAT, {digits})")
}

# On Snowflake, CONCAT_WS is null if any of its arguments are null.  Paste
# is implemented here to avoid this behavior.
snowflake_paste <- function(default_sep) {
  function(..., sep = default_sep, collapse = NULL) {
    check_collapse(collapse)
    sql_glue("ARRAY_TO_STRING(ARRAY_CONSTRUCT_COMPACT({...}), {sep})")
  }
}

test_that("simulate_snowflake() still works", {
  expect_translation(simulate_snowflake(), x + 1, '"x" + 1.0')
})

test_that("custom scalar translated correctly", {
  con <- dialect_snowflake()
  expect_translation(con, log10(x), 'LOG(10, "x")')
  expect_translation(
    con,
    round(x, digits = 1.1),
    'ROUND(("x")::FLOAT, 1)'
  )
  expect_translation(
    con,
    grepl("exp", x),
    "REGEXP_INSTR(\"x\", 'exp', 1, 1, 0, 'c') != 0"
  )

  expect_translation(
    con,
    grepl("exp", x, ignore.case = TRUE),
    "REGEXP_INSTR(\"x\", 'exp', 1, 1, 0, 'i') != 0"
  )
})

test_that("pasting translated correctly", {
  con <- dialect_snowflake()

  expect_translation(
    con,
    paste(x, y),
    "ARRAY_TO_STRING(ARRAY_CONSTRUCT_COMPACT(\"x\", \"y\"), ' ')"
  )
  expect_translation(
    con,
    paste0(x, y),
    "ARRAY_TO_STRING(ARRAY_CONSTRUCT_COMPACT(\"x\", \"y\"), '')"
  )
  expect_translation(con, str_c(x, y), "CONCAT_WS('', \"x\", \"y\")")
  expect_translation(
    con,
    str_c(x, y, sep = "|"),
    "CONCAT_WS('|', \"x\", \"y\")"
  )

  expect_snapshot(
    error = TRUE,
    translate_sql(paste0(x, collapse = ""), con = con)
  )

  expect_translation(
    con,
    str_flatten(x),
    "LISTAGG(\"x\", '') OVER ()",
    window = TRUE
  )
  expect_translation(
    con,
    str_flatten(x, collapse = "|"),
    "LISTAGG(\"x\", '|') OVER ()",
    window = TRUE
  )
  expect_translation(
    con,
    str_flatten(x, collapse = "|"),
    "LISTAGG(\"x\", '|')",
    window = FALSE
  )
})

test_that("custom stringr functions translated correctly", {
  con <- dialect_snowflake()

  expect_translation(con, str_locate(x, y), "POSITION(\"y\", \"x\")")
  expect_translation(
    con,
    str_detect(x, y),
    "REGEXP_INSTR(\"x\", \"y\") != 0"
  )
  expect_translation(
    con,
    str_detect(x, y, negate = TRUE),
    "REGEXP_INSTR(\"x\", \"y\") = 0"
  )
  expect_translation(
    con,
    str_replace(x, y, z),
    "REGEXP_REPLACE(\"x\", \"y\", \"z\", 1, 1)"
  )
  expect_translation(
    con,
    str_replace(x, "\\d", z),
    "REGEXP_REPLACE(\"x\", '\\\\d', \"z\", 1, 1)"
  )
  expect_translation(
    con,
    str_replace_all(x, y, z),
    "REGEXP_REPLACE(\"x\", \"y\", \"z\")"
  )
  expect_translation(
    con,
    str_squish(x),
    "REGEXP_REPLACE(TRIM(\"x\"), '\\\\s+', ' ')"
  )
  expect_translation(
    con,
    str_remove(x, y),
    "REGEXP_REPLACE(\"x\", \"y\", '', 1, 1)"
  )
  expect_translation(
    con,
    str_remove_all(x, y),
    "REGEXP_REPLACE(\"x\", \"y\")"
  )
  expect_translation(con, str_trim(x), "TRIM(\"x\")")
  expect_translation(
    con,
    str_starts(x, y),
    "REGEXP_INSTR(\"x\", \"y\") = 1"
  )
  expect_translation(
    con,
    str_starts(x, y, negate = TRUE),
    "REGEXP_INSTR(\"x\", \"y\") != 1"
  )
  expect_translation(
    con,
    str_ends(x, y),
    "REGEXP_INSTR(\"x\", \"y\", 1, 1, 1) = (LENGTH(\"x\") + 1)"
  )
  expect_translation(
    con,
    str_ends(x, y, negate = TRUE),
    "REGEXP_INSTR(\"x\", \"y\", 1, 1, 1) != (LENGTH(\"x\") + 1)"
  )

  expect_translation(con, str_like(x, y), "\"x\" LIKE \"y\"")
  expect_translation(con, str_ilike(x, y), "\"x\" ILIKE \"y\"")
})

test_that("aggregates are translated correctly", {
  con <- dialect_snowflake()

  expect_translation(
    con,
    cor(x, y),
    "CORR(\"x\", \"y\")",
    window = FALSE
  )
  expect_translation(
    con,
    cor(x, y),
    "CORR(\"x\", \"y\") OVER ()",
    window = TRUE
  )

  expect_translation(
    con,
    cov(x, y),
    "COVAR_SAMP(\"x\", \"y\")",
    window = FALSE
  )
  expect_translation(
    con,
    cov(x, y),
    "COVAR_SAMP(\"x\", \"y\") OVER ()",
    window = TRUE
  )

  expect_translation(
    con,
    all(x, na.rm = TRUE),
    "BOOLAND_AGG(\"x\")",
    window = FALSE
  )
  expect_translation(
    con,
    all(x, na.rm = TRUE),
    "BOOLAND_AGG(\"x\") OVER ()",
    window = TRUE
  )

  expect_translation(
    con,
    any(x, na.rm = TRUE),
    "BOOLOR_AGG(\"x\")",
    window = FALSE
  )
  expect_translation(
    con,
    any(x, na.rm = TRUE),
    "BOOLOR_AGG(\"x\") OVER ()",
    window = TRUE
  )

  expect_translation(
    con,
    sd(x, na.rm = TRUE),
    "STDDEV(\"x\")",
    window = FALSE
  )
  expect_translation(
    con,
    sd(x, na.rm = TRUE),
    "STDDEV(\"x\") OVER ()",
    window = TRUE
  )
})

test_that("snowflake mimics two argument log", {
  con <- dialect_snowflake()

  expect_translation(con, log(x), "LN(\"x\")")
  expect_translation(con, log(x, 10), "LOG(10.0, \"x\")")
  expect_translation(con, log(x, 10L), "LOG(10, \"x\")")
})

test_that("custom lubridate functions translated correctly", {
  con <- dialect_snowflake()

  expect_translation(con, day(x), "EXTRACT(DAY FROM \"x\")")
  expect_translation(con, mday(x), "EXTRACT(DAY FROM \"x\")")
  expect_translation(con, yday(x), "EXTRACT('dayofyear', \"x\")")
  expect_translation(
    con,
    wday(x),
    "EXTRACT('dayofweek', DATE(\"x\") + 0) + 1"
  )
  expect_translation(con, wday(x, label = TRUE), "DAYNAME(\"x\")")
  expect_translation(
    con,
    wday(x, label = TRUE, abbr = FALSE),
    "DECODE(EXTRACT('dayofweek', \"x\"), 1, 'Monday', 2, 'Tuesday', 3, 'Wednesday', 4, 'Thursday', 5, 'Friday', 6, 'Saturday', 0, 'Sunday')"
  )
  expect_translation(
    con,
    week(x),
    "FLOOR((EXTRACT('dayofyear', \"x\") - 1) / 7) + 1"
  )
  expect_translation(con, isoweek(x), "EXTRACT('weekiso', \"x\")")
  expect_translation(con, month(x), "EXTRACT('month', \"x\")")
  expect_translation(
    con,
    month(x, label = TRUE),
    "MONTHNAME(\"x\")"
  )
  expect_translation(
    con,
    month(x, label = TRUE, abbr = FALSE),
    "DECODE(EXTRACT('month', \"x\"), 1, 'January', 2, 'February', 3, 'March', 4, 'April', 5, 'May', 6, 'June', 7, 'July', 8, 'August', 9, 'September', 10, 'October', 11, 'November', 12, 'December')"
  )
  expect_translation(con, quarter(x), "EXTRACT('quarter', \"x\")")
  expect_translation(
    con,
    quarter(x, with_year = TRUE),
    "(EXTRACT('year', \"x\") || '.' || EXTRACT('quarter', \"x\"))"
  )
  expect_error(
    translate_sql(quarter(x, fiscal_start = 2), con = con),
    class = "dbplyr_error_unsupported_arg"
  )
  expect_translation(con, isoyear(x), "EXTRACT('year', \"x\")")

  expect_translation(con, seconds(x), "INTERVAL '\"x\" second'")
  expect_translation(con, minutes(x), "INTERVAL '\"x\" minute'")
  expect_translation(con, hours(x), "INTERVAL '\"x\" hour'")
  expect_translation(con, days(x), "INTERVAL '\"x\" day'")
  expect_translation(con, weeks(x), "INTERVAL '\"x\" week'")
  expect_translation(con, months(x), "INTERVAL '\"x\" month'")
  expect_translation(con, years(x), "INTERVAL '\"x\" year'")

  expect_translation(
    con,
    floor_date(x, "month"),
    "DATE_TRUNC('month', \"x\")"
  )
  expect_translation(
    con,
    floor_date(x, "week"),
    "DATE_TRUNC('week', \"x\")"
  )
})

test_that("custom clock functions translated correctly", {
  con <- dialect_snowflake()
  expect_translation(
    con,
    add_years(x, 1),
    "DATEADD(YEAR, 1.0, \"x\")"
  )
  expect_translation(
    con,
    add_days(x, 1),
    "DATEADD(DAY, 1.0, \"x\")"
  )
  expect_error(
    translate_sql(add_days(x, 1, "dots", "must", "be empty"), con = con),
    class = "rlib_error_dots_nonempty"
  )
  expect_translation(
    con,
    date_build(2020, 1, 1),
    "DATE_FROM_PARTS(2020.0, 1.0, 1.0)"
  )
  expect_translation(
    con,
    date_build(year_column, 1L, 1L),
    "DATE_FROM_PARTS(\"year_column\", 1, 1)"
  )
  expect_translation(
    con,
    get_year(date_column),
    "DATE_PART(YEAR, \"date_column\")"
  )
  expect_translation(
    con,
    get_month(date_column),
    "DATE_PART(MONTH, \"date_column\")"
  )
  expect_translation(
    con,
    get_day(date_column),
    "DATE_PART(DAY, \"date_column\")"
  )
  expect_translation(
    con,
    date_count_between(date_column_1, date_column_2, "day"),
    "DATEDIFF(DAY, \"date_column_1\", \"date_column_2\")"
  )
  expect_error(
    translate_sql(
      date_count_between(
        date_column_1,
        date_column_2,
        "year"
      ),
      con = con
    ),
    class = "dbplyr_error_unsupported_arg"
  )
  expect_error(
    translate_sql(
      date_count_between(
        date_column_1,
        date_column_2,
        "day",
        n = 5
      ),
      con = con
    ),
    class = "dbplyr_error_unsupported_arg"
  )
})

test_that("difftime is translated correctly", {
  con <- dialect_snowflake()
  expect_translation(
    con,
    difftime(start_date, end_date, units = "days"),
    "DATEDIFF(DAY, \"end_date\", \"start_date\")"
  )
  expect_translation(
    con,
    difftime(start_date, end_date),
    "DATEDIFF(DAY, \"end_date\", \"start_date\")"
  )

  expect_error(
    translate_sql(difftime(start_date, end_date, units = "auto"), con = con),
    class = "dbplyr_error_unsupported_arg"
  )
  expect_error(
    translate_sql(
      difftime(
        start_date,
        end_date,
        tz = "UTC",
        units = "days"
      ),
      con = con
    ),
    class = "dbplyr_error_unsupported_arg"
  )
})

test_that("min() and max()", {
  con <- dialect_snowflake()

  expect_translation(
    con,
    min(x, na.rm = TRUE),
    "MIN(\"x\") OVER ()"
  )
  expect_translation(
    con,
    max(x, na.rm = TRUE),
    "MAX(\"x\") OVER ()"
  )

  # na.rm = FALSE is ignored
  # https://docs.snowflake.com/en/sql-reference/functions/min
  # https://docs.snowflake.com/en/sql-reference/functions/max
  # NULL values are ignored unless all the records are NULL, in which case a NULL value is returned.
  expect_translation(
    con,
    min(x, na.rm = TRUE),
    "MIN(\"x\") OVER ()"
  )

  expect_translation(
    con,
    max(x, na.rm = TRUE),
    "MAX(\"x\") OVER ()"
  )
})

test_that("pmin() and pmax() respect na.rm", {
  con <- dialect_snowflake()

  expect_translation(
    con,
    pmin(x, y, z, na.rm = FALSE),
    "LEAST(\"x\", \"y\", \"z\")"
  )
  expect_translation(
    con,
    pmin(x, y, z, na.rm = TRUE),
    "LEAST_IGNORE_NULLS(\"x\", \"y\", \"z\")"
  )

  expect_translation(
    con,
    pmax(x, y, z, na.rm = FALSE),
    "GREATEST(\"x\", \"y\", \"z\")"
  )
  expect_translation(
    con,
    pmax(x, y, z, na.rm = TRUE),
    "GREATEST_IGNORE_NULLS(\"x\", \"y\", \"z\")"
  )
})

test_that("row_number() with and without group_by() and arrange(): unordered defaults to Ordering by NULL (per empty_order)", {
  mf <- lazy_frame(x = c(1:5), y = c(rep("A", 5)), con = dialect_snowflake())
  expect_snapshot(mf |> mutate(rown = row_number()))
  expect_snapshot(mf |> group_by(y) |> mutate(rown = row_number()))
  expect_snapshot(mf |> arrange(y) |> mutate(rown = row_number()))
})

test_that("window functions use inline OVER, not named WINDOW clause", {
  mf <- lazy_frame(a = 1, b = 2, con = dialect_snowflake())
  expect_snapshot(mf |> group_by(a) |> mutate(b = mean(b)))
})

test_that("correctly translates $", {
  con <- dialect_snowflake()
  expect_translation(con, x$y, "\"x\":\"y\"")
})

test_that("custom scalar translated correctly", {
  local_con(simulate_snowflake())
  expect_equal(translate_sql(log10(x)), sql("LOG(10.0, `x`)"))
  expect_equal(translate_sql(round(x, digits = 1.1)), sql("ROUND((`x`) :: FLOAT, 1)"))
  expect_equal(translate_sql(grepl("exp", x)),        sql("(`x`) REGEXP ('.*' || '(exp)' || '.*')"))
  expect_error(translate_sql(grepl("exp", x, ignore.case = TRUE)),
               "`perl`, `fixed`, `useBytes`, and `ignore.case` parameters are unsupported.")
})

test_that("pasting translated correctly", {
  local_con(simulate_snowflake())

  expect_equal(translate_sql(paste(x, y)),  sql("ARRAY_TO_STRING(ARRAY_CONSTRUCT_COMPACT(`x`, `y`), ' ')"))
  expect_equal(translate_sql(paste0(x, y)), sql("ARRAY_TO_STRING(ARRAY_CONSTRUCT_COMPACT(`x`, `y`), '')"))
  expect_equal(translate_sql(str_c(x, y)), sql("CONCAT_WS('', `x`, `y`)"))
  expect_equal(translate_sql(str_c(x, y, sep = '|')), sql("CONCAT_WS('|', `x`, `y`)"))

  expect_error(translate_sql(paste0(x, collapse = "")), "`collapse` not supported")

  expect_equal(translate_sql(str_flatten(x), window = TRUE), sql("LISTAGG(`x`, '') OVER ()"))
  expect_equal(translate_sql(str_flatten(x, collapse = "|"), window = TRUE), sql("LISTAGG(`x`, '|') OVER ()"))
  expect_equal(translate_sql(str_flatten(x, collapse = "|"), window = FALSE), sql("LISTAGG(`x`, '|')"))
})

test_that("custom stringr functions translated correctly", {
  local_con(simulate_snowflake())

  expect_equal(translate_sql(str_locate(x, y)), sql("POSITION(`y`, `x`)"))
  expect_equal(translate_sql(str_detect(x, y)), sql("(`x`) REGEXP ('.*' || `y` || '.*')"))
  expect_equal(translate_sql(str_detect(x, y, negate = TRUE)), sql("!((`x`) REGEXP ('.*' || `y` || '.*'))"))
  expect_equal(translate_sql(str_replace(x, y, z)), sql("REGEXP_REPLACE(`x`, `y`, `z`, 1.0, 1.0)"))
  expect_equal(translate_sql(str_replace(x, "\\d", z)), sql("REGEXP_REPLACE(`x`, '\\\\d', `z`, 1.0, 1.0)"))
  expect_equal(translate_sql(str_replace_all(x, y, z)), sql("REGEXP_REPLACE(`x`, `y`, `z`)"))
  expect_equal(translate_sql(str_squish(x)), sql("REGEXP_REPLACE(TRIM(`x`), '\\\\s+', ' ')"))
  expect_equal(translate_sql(str_remove(x, y)), sql("REGEXP_REPLACE(`x`, `y`, '', 1.0, 1.0)"))
  expect_equal(translate_sql(str_remove_all(x, y)), sql("REGEXP_REPLACE(`x`, `y`)"))
  expect_equal(translate_sql(str_trim(x)), sql("TRIM(`x`)"))
})

test_that("aggregates are translated correctly", {
  local_con(simulate_snowflake())

  expect_equal(translate_sql(cor(x, y), window = FALSE), sql("CORR(`x`, `y`)"))
  expect_equal(translate_sql(cor(x, y), window = TRUE),  sql("CORR(`x`, `y`) OVER ()"))

  expect_equal(translate_sql(cov(x, y), window = FALSE), sql("COVAR_SAMP(`x`, `y`)"))
  expect_equal(translate_sql(cov(x, y), window = TRUE),  sql("COVAR_SAMP(`x`, `y`) OVER ()"))

  expect_equal(translate_sql(all(x, na.rm = TRUE), window = FALSE), sql("BOOLAND_AGG(`x`)"))
  expect_equal(translate_sql(all(x, na.rm = TRUE), window = TRUE),  sql("BOOLAND_AGG(`x`) OVER ()"))

  expect_equal(translate_sql(any(x, na.rm = TRUE), window = FALSE), sql("BOOLOR_AGG(`x`)"))
  expect_equal(translate_sql(any(x, na.rm = TRUE), window = TRUE),  sql("BOOLOR_AGG(`x`) OVER ()"))

  expect_equal(translate_sql(sd(x, na.rm = TRUE), window = FALSE), sql("STDDEV(`x`)"))
  expect_equal(translate_sql(sd(x, na.rm = TRUE), window = TRUE),  sql("STDDEV(`x`) OVER ()"))
})

test_that("snowflake mimics two argument log", {
  local_con(simulate_snowflake())

  expect_equal(translate_sql(log(x)), sql('LN(`x`)'))
  expect_equal(translate_sql(log(x, 10)), sql('LOG(10.0, `x`)'))
  expect_equal(translate_sql(log(x, 10L)), sql('LOG(10, `x`)'))
})

test_that("custom lubridate functions translated correctly", {
  local_con(simulate_snowflake())

  expect_equal(translate_sql(day(x)),  sql("EXTRACT(DAY FROM `x`)"))
  expect_equal(translate_sql(mday(x)), sql("EXTRACT(DAY FROM `x`)"))
  expect_equal(translate_sql(yday(x)), sql("EXTRACT('dayofyear', `x`)"))
  expect_equal(translate_sql(wday(x)), sql("EXTRACT('dayofweek', DATE(`x`) + 0) + 1.0"))
  expect_equal(translate_sql(wday(x, label = TRUE)), sql("DAYNAME(`x`)"))
  expect_equal(translate_sql(wday(x, label = TRUE, abbr = FALSE)), sql(
    "DECODE(EXTRACT('dayofweek', `x`), 1.0, 'Monday', 2.0, 'Tuesday', 3.0, 'Wednesday', 4.0, 'Thursday', 5.0, 'Friday', 6.0, 'Saturday', 0.0, 'Sunday')"
  ))
  expect_equal(translate_sql(week(x)), sql("FLOOR((EXTRACT('dayofyear', `x`) - 1) / 7) + 1"))
  expect_equal(translate_sql(isoweek(x)), sql("EXTRACT('weekiso', `x`)"))
  expect_equal(translate_sql(month(x)), sql("EXTRACT('month', `x`)"))
  expect_equal(translate_sql(month(x, label = TRUE)), sql("MONTHNAME(`x`)"))
  expect_equal(translate_sql(month(x, label = TRUE, abbr = FALSE)), sql(
    "DECODE(EXTRACT('month', `x`), 1.0, 'January', 2.0, 'February', 3.0, 'March', 4.0, 'April', 5.0, 'May', 6.0, 'June', 7.0, 'July', 8.0, 'August', 9.0, 'September', 10.0, 'October', 11.0, 'November', 12.0, 'December')"
  ))
  expect_equal(translate_sql(quarter(x)), sql("EXTRACT('quarter', `x`)"))
  expect_equal(translate_sql(quarter(x, with_year = TRUE)), sql("(EXTRACT('year', `x`) || '.' || EXTRACT('quarter', `x`))"))
  expect_error(translate_sql(quarter(x, fiscal_start = 2)))
  expect_equal(translate_sql(isoyear(x)), sql("EXTRACT('year', `x`)"))

  expect_equal(translate_sql(seconds(x)), sql("INTERVAL '`x` second'"))
  expect_equal(translate_sql(minutes(x)), sql("INTERVAL '`x` minute'"))
  expect_equal(translate_sql(hours(x)),   sql("INTERVAL '`x` hour'"))
  expect_equal(translate_sql(days(x)),    sql("INTERVAL '`x` day'"))
  expect_equal(translate_sql(weeks(x)),   sql("INTERVAL '`x` week'"))
  expect_equal(translate_sql(months(x)),  sql("INTERVAL '`x` month'"))
  expect_equal(translate_sql(years(x)),   sql("INTERVAL '`x` year'"))

  expect_equal(translate_sql(floor_date(x, 'month')), sql("DATE_TRUNC('month', `x`)"))
  expect_equal(translate_sql(floor_date(x, 'week')),  sql("DATE_TRUNC('week', `x`)"))
})

#' "Uncount" a database table
#'
#' This is a method for the tidyr `uncount()` generic. It uses a temporary
#' table, so your database user needs permissions to create one.
#'
#' @inheritParams arrange.tbl_lazy
#' @param weights A vector of weights. Evaluated in the context of `data`;
#'   supports quasiquotation.
#' @param .id Supply a string to create a new variable which gives a unique
#'   identifier for each created row.
#' @param .remove If `TRUE`, and `weights` is the name of a column in `data`,
#'   then this column is removed.
#' @export
#' @examples
#' df <- memdb_frame(x = c("a", "b"), n = c(1, 2))
#' dbplyr_uncount(df, n)
#' dbplyr_uncount(df, n, .id = "id")
#'
#' # You can also use constants
#' dbplyr_uncount(df, 2)
#'
#' # Or expressions
#' dbplyr_uncount(df, 2 / n)
dbplyr_uncount <- function(data, weights, .remove = TRUE, .id = NULL) {
  # Overview of the strategy:
  # 1. calculate `n_max`, the max weight
  # 2. create a temporary db table with ids from 1 to `n_max`
  # 3. duplicate data by inner joining with this temporary table under the
  #    condition that data$weight <= tmp$id
  #
  # An alternative approach would be to use a recursive CTE but this requires
  # more code and testing across backends.
  # See https://stackoverflow.com/questions/33327837/repeat-rows-n-times-according-to-column-value


  weights_quo <- enquo(weights)
  weights_is_col <- quo_is_symbol(weights_quo) &&
    quo_name(weights_quo) %in% colnames(data)

  if (weights_is_col) {
    weights_col <- quo_name(weights_quo)
  } else {
    weights_col <- "..dbplyr_weight_col"
    data <- mutate(
      data,
      !!weights_col := !!weights_quo,
    )
  }

  n_max <- pull(summarise(ungroup(data), max(!!sym(weights_col), na.rm = TRUE)))
  n_max <- vctrs::vec_cast(n_max, integer(), x_arg = "weights")

  if (is_null(.id)) {
    row_id_col <- "..dbplyr_row_id"
  } else {
    row_id_col <- .id
  }
  tbl_name <- unclass(query_name(data)) %||% "LHS"
  sql_on_expr <- expr(RHS[[!!row_id_col]] <= (!!sym(tbl_name))[[!!weights_col]])

  con <- remote_con(data)
  # FIXME: Use generate_series() if available on database
  helper_table <- copy_inline(con, tibble(!!row_id_col := seq2(1, n_max)))

  data_uncounted <- inner_join(
    data,
    helper_table,
    by = character(),
    sql_on = translate_sql(!!sql_on_expr, con = con),
    x_as = tbl_name,
    y_as = "RHS"
  )

  cols_to_remove <- character()
  if (is_null(.id)) {
    cols_to_remove <- c(cols_to_remove, row_id_col)
  }

  grps <- group_vars(data_uncounted)
  if (!weights_is_col || (weights_is_col && .remove)) {
    cols_to_remove <- c(cols_to_remove, weights_col)
    # need to regroup to be able to remove weights_col
    grps <- setdiff(grps, weights_col)
  }


  data_uncounted %>%
    ungroup() %>%
    select(-all_of(cols_to_remove)) %>%
    group_by(!!!syms(grps))
}

globalVariables(c("RHS", "LHS", "all_of"))

test_that("symbols weights are dropped in output", {
  df <- copy_to_test("sqlite", tibble(x = 1, w = 1))
  expect_equal(dbplyr_uncount(df, w) %>% collect(), tibble(x = 1))

  expect_snapshot(
    dbplyr_uncount(df, w) %>% show_query(),
    transform = function(lines) {
      lines_to_transform <- grepl("INNER JOIN", lines)
      lines[lines_to_transform] <- gsub(
        "`dbplyr_\\d+`", "`dbplyr_table`",
        lines[lines_to_transform]
      )

      lines
    }
  )

  expect_equal(
    df %>% mutate(w = w + 1) %>% dbplyr_uncount(w) %>% collect(),
    tibble(x = c(1L, 1L))
  )

  expect_snapshot(
    df %>%
      mutate(w = w + 1) %>%
      dbplyr_uncount(w) %>%
      show_query()
  )
})

test_that("can request to preserve symbols", {
  df <- memdb_frame(x = 1, w = 1)

  expect_equal(
    dbplyr_uncount(df, w, .remove = FALSE) %>% colnames(),
    c("x", "w")
  )
})

test_that("unique identifiers created on request", {
  df <- memdb_frame(w = 1:3)
  expect_equal(
    dbplyr_uncount(df, w, .id = "id") %>% collect() %>% arrange(id),
    tibble(id = c(1L, 1:2, 1:3)) %>% arrange(id)
  )
})

test_that("expands constants and expressions", {
  df <- memdb_frame(x = 1, w = 2)

  expect_equal(dbplyr_uncount(df, 2) %>% collect(), collect(df)[c(1, 1), ])
  expect_equal(dbplyr_uncount(df, 1 + 1) %>% collect(), collect(df)[c(1, 1), ])
})


test_that("works with groups", {
  df <- memdb_frame(g = 1, x = 1, w = 1) %>% dplyr::group_by(g)
  expect_equal(group_vars(dbplyr_uncount(df, w)), "g")
})

test_that("grouping variable are removed", {
  df <- memdb_frame(g = 1, x = 1, w = 1) %>% dplyr::group_by(g)

  expect_equal(dbplyr_uncount(df, g) %>% colnames(), c("x", "w"))
})

test_that("must evaluate to integer", {
  df <- memdb_frame(x = 1, w = 1/2)
  expect_error(dbplyr_uncount(df, w), class = "vctrs_error_cast_lossy")

  expect_error(dbplyr_uncount(df, "W"), class = "vctrs_error_incompatible_type")
})

test_that("works with 0 weights", {
  df <- memdb_frame(x = 1:2, w = c(0, 1))
  expect_equal(dbplyr_uncount(df, w) %>% collect(), tibble(x = 2))
})

#' Arrange rows by column values
#'
#' @description
#' This is an method for the dplyr [arrange()] generic. It generates
#' the `ORDER BY` clause of the SQL query. It also affects the
#' [window_order()] of windowed expressions in [mutate.tbl_lazy()].
#'
#' Note that `ORDER BY` clauses can not generally appear in subqueries, which
#' means that you should `arrange()` as late as possible in your pipelines.
#'
#' @section Missing values:
#' Unlike R, most databases sorts `NA` (`NULL`s) at the front. You can
#' can override this behaviour by explicitly sorting on `is.na(x)`.
#'
#' @param .data A lazy data frame backed by a database query.
#' @inheritParams dplyr::arrange
#' @return Another `tbl_lazy`. Use [show_query()] to see the generated
#'   query, and use [`collect()`][collect.tbl_sql] to execute the query
#'   and return data to R.
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(a = c(3, 4, 1, 2), b = c(5, 1, 2, NA))
#' db |> arrange(a) |> show_query()
#'
#' # Note that NAs are sorted first
#' db |> arrange(b)
#' # override by sorting on is.na() first
#' db |> arrange(is.na(b), b)
#' @export
#' @importFrom dplyr arrange
arrange.tbl_lazy <- function(.data, ..., .by_group = FALSE) {
  dots <- partial_eval_dots(.data, ..., .named = FALSE)
  names(dots) <- NULL

  .data$lazy_query <- add_arrange(.data$lazy_query, dots, .by_group)
  .data
}

add_arrange <- function(lazy_query, exprs, .by_group) {
  # Empty arrange() preserves existing ordering (like dplyr)
  if (is_empty(exprs)) {
    return(lazy_query)
  }

  if (.by_group) {
    exprs <- c(syms(op_grps(lazy_query)), exprs)
  }

  # Prepend new ordering to existing ordering (like dplyr)
  order_vars <- c(exprs, op_sort(lazy_query))

  if (can_inline_arrange(lazy_query)) {
    lazy_query$order_vars <- order_vars
    lazy_query$order_by <- order_vars
    lazy_query
  } else {
    lazy_select_query(
      x = lazy_query,
      order_by = order_vars,
      order_vars = order_vars
    )
  }
}

# arrange() adds/modifies the ORDER BY clause
# * ORDER BY is executed after LIMIT
#   => can't inline if LIMIT is set
can_inline_arrange <- function(lazy_query) {
  if (!is_lazy_select_query(lazy_query)) {
    return(FALSE)
  }

  if (!is.null(lazy_query$limit)) {
    return(FALSE)
  }

  TRUE
}

# Used in slice_min/slice_max and the order_by argument to window functions
# convert a single order by expression to a list of expresions
unwrap_order_expr <- function(order_by, f, error_call = caller_env()) {
  if (is_call(order_by, "c")) {
    args <- call_args(order_by)
    tibble_expr <- expr_text(expr(tibble(!!!args)))
    cli_abort(
      c(
        "Can't use `c()` in {.fun {f}}",
        i = "Did you mean to use `{tibble_expr}` instead?"
      ),
      call = error_call
    )
  }

  if (is.null(order_by)) {
    NULL
  } else if (is_call(order_by, c("tibble", "data.frame"))) {
    as.list(order_by[-1])
  } else {
    list(order_by)
  }
}

swap_order_direction <- function(x) {
  if (is_call(x, "desc", n = 1)) {
    x[[2]]
  } else {
    call2("desc", x)
  }
}

sql_quantile <- function(f,
                         style = c("infix", "ordered"),
                         window = FALSE) {
  force(f)
  style <- match.arg(style)
  force(window)

  function(x, probs, na.rm = FALSE) {
    check_probs(probs)
    check_na_rm(na.rm)

    sql <- switch(style,
      infix = sql_call2(f, x, probs),
      ordered = build_sql(
        sql_call2(f, probs), " WITHIN GROUP (ORDER BY ", x, ")"
      )
    )

    if (window) {
      sql <- win_over(sql,
        partition = win_current_group(),
        frame = win_current_frame()
      )
    }
    sql
  }
}

sql_median <- function(f,
                       style = c("infix", "ordered"),
                       window = FALSE) {
  quantile <- sql_quantile(f, style = style, window = window)
  function(x, na.rm = FALSE) {
    quantile(x, 0.5, na.rm = na.rm)
  }
}

check_probs <- function(probs) {
  if (!is.numeric(probs)) {
    cli_abort("{.arg probs} must be numeric")
  }

  if (length(probs) > 1) {
    cli_abort("SQL translation only supports single value for {.arg probs}.")
  }
}

test_that("quantile and median don't change without warning", {
  expect_snapshot(translate_sql(quantile(x, 0.75, na.rm = TRUE), window = FALSE))
  expect_snapshot(translate_sql(quantile(x, 0.75, na.rm = TRUE), vars_group = "g"))
  expect_snapshot(translate_sql(median(x, na.rm = TRUE), window = FALSE))
  expect_snapshot(translate_sql(median(x, na.rm = TRUE), vars_group = "g"))
})

test_that("checks for invalid probs", {
  expect_error(check_probs("a"), "numeric")
  expect_error(check_probs(1:3), "single value")
})

#' Create a lazy query backed by a database
#'
#' @description
#' Use `tbl()` to create a SQL query backed by a database. Manipulating this
#' object with dplyr verbs then builds up a SQL query that will only be executed
#' when you explicitly ask for it, either by printing the object, calling
#' [collect()] to bring the data back to R or calling [compute()] to create a
#' new table in the database. You can see the query without executing it with
#' [show_query()].
#'
#' Learn more in `vignette("dbplyr")`.
#'
#' @param src A `DBIConnection` object produced by [DBI::dbConnect()].
#' @param from Either a table identifier or a literal [sql()] string.
#'
#'   Use a string to identify a table in the current schema/catalog or
#'   `I()` for a table elsewhere, e.g. `I("schema.table")` or
#'   `I("catalog.schema.table")`. For backward compatibility, you can also use
#'   [in_schema()]/[in_catalog()] or [DBI::Id()].
#' @param vars Optionally, provide a character vector of column names. If
#'   not supplied, will be retrieved from the database by running a simple
#'   query. This argument is mainly useful for better performance when creating
#'   many `tbl`s with known variables.
#' @param ... Passed on to [tbl_sql()]
#' @export
#' @examples
#' library(dplyr)
#'
#' # Connect to a temporary in-memory SQLite database and add some data
#' con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#' copy_to(con, mtcars)
#'
#' # To retrieve a single table from a source, use `tbl()`
#' mtcars_db <- con |> tbl("mtcars")
#' mtcars_db
#'
#' # Use `I()` for qualified table names
#' con |> tbl(I("temp.mtcars")) |> head(1)
#'
#' # You can also pass raw SQL if you want a more sophisticated query
#' con |> tbl(sql("SELECT * FROM mtcars WHERE cyl = 8")) |> head(1)
#'
#' # But in most cases, you'll rely on dbplyr to construct the SQL:
#' mtcars_db |>
#'   filter(vs == 1) |>
#'   summarise(mpg = mean(mpg, na.rm = TRUE), .by = cyl) |>
#'   show_query()
#'
#' @importFrom dplyr tbl
#' @aliases tbl_dbi
tbl.src_dbi <- function(src, from, vars = NULL, ...) {
  check_dots_empty()
  db_table(src$con, from, vars = vars)
}

db_table <- function(
  con,
  from,
  vars = NULL,
  subclass = NULL,
  call = caller_env()
) {
  check_con(con)
  source <- as_table_source(from, con = con, error_call = call)

  check_character(vars, allow_null = TRUE, call = call)
  vars <- vars %||% find_variables(con, from, call = call)

  new_tbl_sql(con = con, source = source, vars = vars, subclass = subclass)
}

#' Database src
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' Since can generate a `tbl()` directly from a DBI connection we no longer
#' recommend using `src_dbi()`.
#'
#' @param con An object that inherits from [DBI::DBIConnection-class],
#'   typically generated by [DBI::dbConnect]
#' @param auto_disconnect Should the connection be automatically closed when
#'   the src is deleted? Set to `TRUE` if you initialize the connection
#'   the call to `src_dbi()`. Pass `NA` to auto-disconnect but print a message
#'   when this happens.
#' @return An S3 object with class `src_dbi`, `src_sql`, `src`.
#' @keywords internal
#' @export
src_dbi <- function(con, auto_disconnect = FALSE) {
  # Avoid registering disconnector if con can't be evaluated
  force(con)

  # stopifnot(is(con, "DBIConnection"))
  if (is_false(auto_disconnect)) {
    disco <- NULL
  } else {
    disco <- db_disconnector(con, quiet = is_true(auto_disconnect)) # nocov
  }

  structure(
    list(
      con = con,
      disco = disco
    ),
    class = connection_s3_class(con)
  )
}

connection_s3_class <- function(con) {
  subclass <- setdiff(methods::is(con), methods::extends("DBIConnection"))
  c(paste0("src_", subclass), "src_dbi", "src_sql", "src")
}

methods::setOldClass(c("src_dbi", "src_sql", "src"))

# nocov start
# Creates an environment that disconnects the database when it's GC'd
db_disconnector <- function(con, quiet = FALSE) {
  reg.finalizer(environment(), function(...) {
    if (!quiet) {
      message("Auto-disconnecting ", class(con)[[1]])
    }
    DBI::dbDisconnect(con)
  })
  environment()
}
# nocov end

test_that("tbl and src classes include connection class", {

  mf <- memdb_frame(x = 1, y = 2)
  expect_true(inherits(mf, "tbl_SQLiteConnection"))
  expect_true(inherits(mf$src, "src_SQLiteConnection"))
})

test_that("generates S3 class based on S4 class name", {
  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  expect_equal(
    connection_s3_class(con), 
    c("src_SQLiteConnection", "src_dbi", "src_sql", "src")
 )

  on.exit(removeClass("Foo2"))
  on.exit(removeClass("Foo1"))

  Foo1 <- setClass("Foo1", contains = "DBIConnection")
  Foo2 <- setClass("Foo2", contains = "Foo1")
  expect_equal(
    connection_s3_class(Foo2()), 
    c("src_Foo2", "src_Foo1", "src_dbi", "src_sql", "src")
  )
})

#' Metadata about a remote table
#'
#' `remote_name()` gives the unescaped name of the remote table, or `NULL` if it
#' is a query (created by `sql()`) or already escape (created by `ident_q()`).
#' `remote_table()` gives the remote table or the query.
#' `remote_query()` gives the text of the query, and `remote_query_plan()`
#' the query plan (as computed by the remote database). `remote_src()` and
#' `remote_con()` give the dplyr source and DBI connection respectively.
#'
#' @param x Remote table, currently must be a [`tbl_sql`].
#' @param null_if_local Return `NULL` if the remote table is created via
#'   [tbl_lazy()] or [lazy_frame()]?
#' @param cte `r lifecycle::badge("deprecated")`
#'   Use the `sql_options` argument instead.
#' @param sql_options SQL rendering options generated by [sql_options()].
#' @param ... Additional arguments passed on to methods.
#' @return
#'   * `remote_name()` returns a string, or `NULL` if not a remote table, or not
#'     applicable. For example, computed queries do not have a "name".
#'   * `remote_table()` returns a [sql()] object containing the (correctly
#'     quoted) table identifier, or the custom query if the table was created
#'     from `sql()`. Returns `NULL` if applied to a lazy query that no longer
#'     corresponds to a single table (e.g. after a `filter()` or a join), or if
#'     applied to a local lazy table.
#' @export
#' @examples
#' mf <- memdb_frame(x = 1:5, y = 5:1, .name = "blorp")
#' remote_name(mf)
#' remote_src(mf)
#' remote_con(mf)
#' remote_query(mf)
#'
#' mf2 <- dplyr::filter(mf, x > 3)
#' remote_name(mf2)
#' remote_src(mf2)
#' remote_con(mf2)
#' remote_query(mf2)
remote_name <- function(x, null_if_local = TRUE) {
  table <- remote_table_path(x, null_if_local = null_if_local)

  if (is.sql(table) || is.null(table)) {
    NULL
  } else if (inherits(x, "tbl_lazy")) {
    table_path_name(table, x$con)
  } else {
    table
  }
}

#' @export
#' @rdname remote_name
remote_table <- function(x, null_if_local = TRUE) {
  check_bool(null_if_local)
  table <- remote_table_path(x, null_if_local = null_if_local)
  if (is.null(table) || is.sql(table)) {
    table
  } else {
    sql(unclass(table))
  }
}

# Internal generic that returns the raw table path (or sql for custom queries,
# or NULL when the lazy query no longer corresponds to a single table).
# Use this instead of `remote_table()` when internal code needs a value that can
# be passed to `as_table_path()` / `sql_escape_table_source()`.
remote_table_path <- function(x, null_if_local = TRUE) {
  UseMethod("remote_table_path")
}

#' @export
remote_table_path.tbl_lazy <- function(x, null_if_local = TRUE) {
  remote_table_path(x$lazy_query, null_if_local = null_if_local)
}

#' @export
remote_table_path.lazy_base_remote_query <- function(x, null_if_local = TRUE) {
  x$x
}

#' @export
remote_table_path.lazy_base_local_query <- function(x, null_if_local = TRUE) {
  if (null_if_local) {
    return()
  }

  x$name
}

#' @export
remote_table_path.lazy_query <- function(x, null_if_local = TRUE) {
  NULL
}

#' @export
#' @rdname remote_name
remote_src <- function(x) {
  x$src
}

#' @export
#' @rdname remote_name
remote_con <- function(x) {
  x$con
}

#' @export
#' @rdname remote_name
remote_query <- function(x, sql_options = NULL, cte = deprecated()) {
  rlang::check_exclusive(sql_options, cte, .require = FALSE)
  if (lifecycle::is_present(cte)) {
    lifecycle::deprecate_warn(
      when = "2.4.0",
      what = "remote_query(cte)",
      with = I("remote_query(sql_options = sql_options(cte = TRUE))")
    )
    sql_options <- dbplyr::sql_options(cte = cte)
  }
  db_sql_render(remote_con(x), x, sql_options = sql_options)
}

#' @export
#' @rdname remote_name
remote_query_plan <- function(x, ...) {
  dbplyr_explain(remote_con(x), db_sql_render(remote_con(x), x$lazy_query), ...)
}

#' Retrieve the last SQL query generated
#'
#' This is a helper function that retrieves the most recent SQL query generated
#' by dbplyr, which can be useful for debugging.
#'
#' @return A SQL string, or `NULL` if no query has been generated yet.
#' @export
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' df <- lazy_frame(x = 1:3)
#' df |> filter(x > 1)
#'
#' last_sql()
last_sql <- function() {
  the$last_sql
}

test_that("remote_table returns name when it makes sense", {
  mf <- local_memdb_frame("refxiudlph", x = 5)

  # produces name after `group_by()`
  expect_equal(
    mf |> group_by(x) |> remote_table(),
    sql("`refxiudlph`")
  )

  # produces name after compute()
  expect_false(is_null(
    mf |> mutate(x = x + 1) |> compute() |> remote_table()
  ))
})

test_that("remote_table returns null for computed tables", {
  mf <- local_memdb_frame("refxiudlph", x = 5, y = 1)
  expect_equal(remote_table(mf), sql("`refxiudlph`"))

  expect_null(mf |> filter(x == 3) |> remote_table())
  expect_null(mf |> distinct(x) |> remote_table())
  expect_null(mf |> mutate(x = x + 1) |> remote_table())
  expect_null(mf |> select(x) |> remote_table())
  expect_null(mf |> relocate(y, x) |> remote_table())
  expect_null(mf |> head(3) |> remote_table())

  expect_null(left_join(mf, mf, by = "x") |> remote_table())
  lf <- lazy_frame(x = 1)
  expect_null(lf |> remote_table())
  expect_null(lf |> group_by(x) |> remote_table())

  lf <- lazy_frame(x = 1)
  expect_equal(lf |> remote_table(null_if_local = FALSE), sql('"df"'))
  expect_equal(
    lf |> group_by(x) |> remote_table(null_if_local = FALSE),
    sql('"df"')
  )
})

test_that("remote_name and remote_table can handle different table identifiers", {
  test_remote_table <- function(x, exp_tbl) {
    lf <- lazy_frame(x = 1, .name = x)
    expect_equal(remote_table(lf, null_if_local = FALSE), exp_tbl)
    expect_equal(remote_name(lf, null_if_local = FALSE), "tbl")
  }

  test_remote_table("tbl", sql('"tbl"'))
  test_remote_table(ident("tbl"), sql('"tbl"'))
  test_remote_table(in_schema("schema", "tbl"), sql('"schema"."tbl"'))
  test_remote_table(
    in_catalog("catalog", "schema", "tbl"),
    sql('"catalog"."schema"."tbl"')
  )
  test_remote_table(
    DBI::Id(catalog = "catalog", schema = "schema", table = "tbl"),
    sql('"catalog"."schema"."tbl"')
  )
  test_remote_table(ident_q("schema.tbl"), sql("schema.tbl"))
  test_remote_table(I("schema.tbl"), sql("schema.tbl"))
})

test_that("remote_name() returns NULL when there is no single table name", {
  mf <- local_memdb_frame("ulfxutwbnh", x = 1)

  # NULL when lazy query no longer corresponds to a single table
  expect_null(mf |> filter(x == 1) |> remote_name())
  expect_null(mf |> mutate(y = x + 1) |> remote_name())
  expect_null(left_join(mf, mf, by = "x") |> remote_name())

  # NULL when source is a custom sql() query
  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  withr::defer(DBI::dbDisconnect(con))
  DBI::dbWriteTable(con, "mytbl", data.frame(x = 1))
  custom <- tbl(con, sql("SELECT * FROM mytbl"))
  expect_null(remote_name(custom))
})

test_that("remote_table() result can be inlined into build_sql()", {
  con <- dialect_ansi()
  lf <- lazy_frame(x = 1, .name = in_schema("myschema", "mytbl"), con = con)

  expect_equal(
    build_sql(
      "SELECT * FROM ",
      remote_table(lf, null_if_local = FALSE),
      con = con
    ),
    sql('SELECT * FROM "myschema"."mytbl"')
  )
})

test_that("can retrieve query, src and con metadata", {
  mf <- local_memdb_frame(x = 5)

  expect_s4_class(remote_con(mf), "DBIConnection")
  expect_s3_class(remote_src(mf), "src_sql")
  expect_s3_class(remote_query(mf), "sql")
  expect_type(remote_query_plan(mf), "character")
})

test_that("last_sql() retrieves the most recent query", {
  lf <- lazy_frame(x = 1:3, y = c("a", "b", "c"))

  capture.output(lf |> filter(x > 1) |> show_query())
  expect_match(last_sql(), "WHERE")

  capture.output(lf |> mutate(z = x + 1) |> show_query())
  expect_match(last_sql(), "\\+ 1")
})

# Vectors -----------------------------------------------------------------

check_list <- function(x, ..., allow_null = FALSE, arg = caller_arg(x), call = caller_env()) {
  if (vctrs::vec_is_list(x)) {
    return()
  }
  stop_input_type(
    x,
    c("a list"),
    ...,
    allow_na = FALSE,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_integer <- function(x,
                          ...,
                          allow_null = FALSE,
                          arg = caller_arg(x),
                          call = caller_env()) {
  if (!missing(x)) {
    if (is.integer(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    what = "an integer vector",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}

check_logical <- function(x,
                          ...,
                          allow_null = FALSE,
                          arg = caller_arg(x),
                          call = caller_env()) {
  if (!missing(x)) {
    if (is.logical(x)) {
      return(invisible(NULL))
    }
    if (allow_null && is_null(x)) {
      return(invisible(NULL))
    }
  }

  stop_input_type(
    x,
    what = "a logical vector",
    ...,
    allow_null = allow_null,
    arg = arg,
    call = call
  )
}


# Scalars -----------------------------------------------------------------

check_con <- function(con, ..., arg = caller_arg(con), call = caller_env()) {
  if (is.null(con)) {
    cli_abort("{.arg {arg}} must not be NULL.", call = call)
  }
}

check_lazy_query <- function(x, ..., arg = caller_arg(x), call = caller_env()) {
  if (!inherits(x, "lazy_query")) {
    stop_input_type(
      x,
      what = "a lazy query",
      ...,
      arg = arg,
      call = call
    )
  }
}

check_scalar_sql <- function(x,
                             ...,
                             arg = caller_arg(x),
                             call = caller_env()) {
  if ((is.sql(x) || is_string(x)) && length(x) == 1L) {
    return()
  }

  stop_input_type(
    x,
    what = c("a single SQL query"),
    call = call,
    arg = arg
  )
}


# Other checks ------------------------------------------------------------

check_unsupported_arg <- function(x,
                                  allowed = NULL,
                                  allow_null = FALSE,
                                  ...,
                                  backend = NULL,
                                  arg = caller_arg(x),
                                  call = caller_env()) {
  if (is_missing(x)) {
    return()
  }

  if (allow_null && is_null(x)) {
    return()
  }

  if (identical(x, allowed)) {
    return()
  }

  if (is_null(allowed)) {
    msg <- "Argument {.arg {arg}} isn't supported"
  } else {
    msg <- "{.code {arg} = {.val {x}}} isn't supported"
  }

  if (is.null(backend)) {
    msg <- paste0(msg, " on database backends.")
  } else {
    msg <- paste0(msg, " in {backend} translation.")
  }

  if (!is_null(allowed)) {
    if (allow_null) {
      allow_msg <- "It must be {.val {allowed}} or {.code NULL} instead."
    } else {
      allow_msg <- "It must be {.val {allowed}} instead."
    }

    msg <- c(msg, i = allow_msg)
  }
  cli_abort(msg, call = call)
}

stop_unsupported_function <- function(f, ..., with = NULL, call = caller_env()) {
  cli_abort(c(
    "{.fun {f}} is not supported on database backends.",
    i = if (!is_null(with)) "Please use {.fun {with}} instead."
  ), call = call)
}

check_named <- function(x, ..., arg = caller_arg(x), call = caller_env()) {
  if (!is_named2(x)) {
    cli_abort("All elements of {.arg {arg}} must be named.", call = call)
  }

  if (vctrs::vec_duplicate_any(names2(x))) {
    cli_abort("The names of {.arg {arg}} must be unique.", call = call)
  }
}

check_has_names <- function(x, names, ..., arg = caller_arg(x), call = caller_env()) {
  if (is.data.frame(x)) {
    x_nms <- colnames(x)
  } else {
    x_nms <- names(x)
  }
  if (identical(x_nms, names)) {
    return()
  }

  cli_abort("{.arg {arg}} must have fields {.val {names}}", .internal = TRUE)
}

with_indexed_errors <- function(expr,
                                message,
                                ...,
                                .error_call = caller_env(),
                                .frame = caller_env()) {
  withCallingHandlers(
    expr,
    purrr_error_indexed = function(cnd) {
      message <- message(cnd)
      abort(message, ..., call = .error_call, parent = cnd$parent, .frame = .frame)
    }
  )
}

#' Translate an expression to SQL
#'
#' @description
#' dbplyr translates commonly used base functions including logical
#' (`!`, `&`, `|`), arithmetic (`^`), and comparison (`!=`) operators, as well
#' as common summary (`mean()`, `var()`), and transformation (`log()`)
#' functions.  All other functions will be preserved as is. R's infix functions
#' (e.g. `%like%`) will be converted to their SQL equivalents (e.g. `LIKE`).
#'
#' Learn more in `vignette("translation-function")`.
#'
#' @param ...,dots Expressions to translate. `translate_sql()`
#'   automatically quotes them for you.  `translate_sql_()` expects
#'   a list of already quoted objects.
#' @param con An optional database connection to control the details of
#'   the translation. The default, `NULL`, generates ANSI SQL.
#' @param vars Deprecated. Now call [partial_eval()] directly.
#' @param vars_group,vars_order,vars_frame Parameters used in the `OVER`
#'   expression of windowed functions.
#' @param window Use `FALSE` to suppress generation of the `OVER`
#'   statement used for window functions. This is necessary when generating
#'   SQL for a grouped summary.
#' @param context Use to carry information for special translation cases. For example, MS SQL needs a different conversion for is.na() in WHERE vs. SELECT clauses.  Expects a list.
#' @export
#' @examples
#' # Regular maths is translated in a very straightforward way
#' translate_sql(x + 1)
#' translate_sql(sin(x) + tan(y))
#'
#' # Note that all variable names are escaped
#' translate_sql(like == "x")
#' # In ANSI SQL: "" quotes variable _names_, '' quotes strings
#'
#' # Logical operators are converted to their sql equivalents
#' translate_sql(x < 5 & !(y >= 5))
#' # xor() doesn't have a direct SQL equivalent
#' translate_sql(xor(x, y))
#'
#' # If is translated into case when
#' translate_sql(if (x > 5) "big" else "small")
#'
#' # Infix functions are passed onto SQL with % removed
#' translate_sql(first %like% "Had%")
#' translate_sql(first %is% NA)
#' translate_sql(first %in% c("John", "Roger", "Robert"))
#'
#' # And be careful if you really want integers
#' translate_sql(x == 1)
#' translate_sql(x == 1L)
#'
#' # If you have an already quoted object, use translate_sql_:
#' x <- quote(y + 1 / sin(t))
#' translate_sql_(list(x), con = simulate_dbi())
#'
#' # Windowed translation --------------------------------------------
#' # Known window functions automatically get OVER()
#' translate_sql(mpg > mean(mpg))
#'
#' # Suppress this with window = FALSE
#' translate_sql(mpg > mean(mpg), window = FALSE)
#'
#' # vars_group controls partition:
#' translate_sql(mpg > mean(mpg), vars_group = "cyl")
#'
#' # and vars_order controls ordering for those functions that need it
#' translate_sql(cumsum(mpg))
#' translate_sql(cumsum(mpg), vars_order = "mpg")
translate_sql <- function(...,
                          con = NULL,
                          vars = character(),
                          vars_group = NULL,
                          vars_order = NULL,
                          vars_frame = NULL,
                          window = TRUE) {

  if (!missing(vars)) {
    cli_abort("{.arg vars} is deprecated. Please use {.fun partial_eval} directly.")
  }

  con <- con %||% sql_current_con() %||% simulate_dbi()

  translate_sql_(
    quos(...),
    con = con,
    vars_group = vars_group,
    vars_order = vars_order,
    vars_frame = vars_frame,
    window = window
  )
}

#' @export
#' @rdname translate_sql
translate_sql_ <- function(dots,
                           con = NULL,
                           vars_group = NULL,
                           vars_order = NULL,
                           vars_frame = NULL,
                           window = TRUE,
                           context = list()) {

  if (length(dots) == 0) {
    return(sql())
  }

  stopifnot(is.list(dots))

  if (!any(have_name(dots))) {
    names(dots) <- NULL
  }

  old_con <- set_current_con(con)
  on.exit(set_current_con(old_con), add = TRUE)

  if (length(context) > 0) {
    old_context <- set_current_context(context)
    on.exit(set_current_context(old_context), add = TRUE)
  }

  if (window) {
    old_group <- set_win_current_group(vars_group)
    on.exit(set_win_current_group(old_group), add = TRUE)

    old_order <- set_win_current_order(vars_order)
    on.exit(set_win_current_order(old_order), add = TRUE)

    old_frame <- set_win_current_frame(vars_frame)
    on.exit(set_win_current_frame(old_frame), add = TRUE)
  }

  variant <- dbplyr_sql_translation(con)
  pieces <- lapply(dots, function(x) {
    if (is_null(get_expr(x))) {
      NULL
    } else if (is_atomic(get_expr(x))) {
      escape(get_expr(x), con = con)
    } else {
      mask <- sql_data_mask(x, variant, con = con, window = window)
      escape(eval_tidy(x, mask), con = con)
    }
  })

  sql(unlist(pieces))
}

sql_data_mask <- function(expr, variant, con, window = FALSE,
                          strict = getOption("dplyr.strict_sql", FALSE)) {
  stopifnot(is.sql_variant(variant))

  # Default for unknown functions
  unknown <- setdiff(all_calls(expr), names(variant))
  op <- if (strict) missing_op else default_op
  top_env <- ceply(unknown, op, parent = empty_env(), env = get_env(expr))

  # Known R -> SQL functions
  special_calls <- copy_env(variant$scalar, parent = top_env)
  if (!window) {
    special_calls2 <- copy_env(variant$aggregate, parent = special_calls)
  } else {
    special_calls2 <- copy_env(variant$window, parent = special_calls)
  }
  special_calls2$`::` <- function(pkg, name) {
    pkg <- as.character(substitute(pkg))
    name <- as.character(substitute(name))
    if (!is_installed(pkg)) {
      cli_abort("There is no package called {.pkg {pkg}}")
    }
    if (!env_has(ns_env(pkg), name)) {
      cli_abort("{.val {name}} is not an exported object from {.pkg {pkg}}")
    }

    if (env_has(special_calls2, name) || env_has(special_calls, name)) {
      env_get(special_calls2, name, inherit = TRUE)
    } else {
      # TODO use {.fun dbplyr::{fn}} after https://github.com/r-lib/cli/issues/422 is fixed
      cli_abort("No known translation for `{pkg}::{name}()`")
    }
  }

  special_calls2$sql <- function(...) {
    dots <- exprs(...)
    dots <- purrr::map(dots, eval_tidy, env = top_env)
    exec(sql, !!!dots)
  }

  # Existing symbols in expression
  names <- all_names(expr)
  idents <- lapply(names, ident)
  name_env <- ceply(idents, escape, con = con, parent = special_calls2)

  # Known sql expressions
  symbol_env <- env_clone(base_symbols, parent = name_env)

  new_data_mask(symbol_env, top_env)
}

is_infix_base <- function(x) {
  x %in% c("::", "$", "@", "^", "*", "/", "+", "-", ">", ">=", "<", "<=",
    "==", "!=", "!", "&", "&&", "|", "||", "~", "<-", "<<-")
}
is_infix_user <- function(x) {
  grepl("^%.*%$", x)
}

default_op <- function(x, env) {
  assert_that(is_string(x))

  # Check for shiny reactives; these are zero-arg functions
  # so need special handling to give a useful error
  obj <- env_get(env, x, default = NULL, inherit = TRUE)
  if (inherits(obj, "reactive")) {
    error_embed("a shiny reactive", "foo()")
  }

  if (is_infix_base(x)) {
    sql_infix(x)
  } else if (is_infix_user(x)) {
    x <- substr(x, 2, nchar(x) - 1)
    sql_infix(x)
  } else {
    sql_prefix(x)
  }
}

missing_op <- function(x, env) {
  force(x)

  function(...) {
    needs_parens <- !is_infix_base(x) && !is_infix_user(x)

    if (needs_parens) {
      cli_abort("Don't know how to translate {.fun {x}}")
    } else {
      cli_abort("Don't know how to translate `{x}`")
    }
  }
}

all_calls <- function(x) {
  if (is_quosure(x)) return(all_calls(quo_get_expr(x)))
  if (!is.call(x)) return(NULL)

  fname <- as.character(x[[1]])
  unique(c(fname, unlist(lapply(x[-1], all_calls), use.names = FALSE)))
}

all_names <- function(x) {
  if (is.name(x)) return(as.character(x))
  if (is_quosure(x)) return(all_names(quo_get_expr(x)))
  if (!is.call(x)) return(NULL)

  unique(unlist(lapply(x[-1], all_names), use.names = FALSE))
}

# character vector -> environment
ceply <- function(x, f, ..., parent = parent.frame()) {
  if (length(x) == 0) return(new.env(parent = parent))
  l <- lapply(x, f, ...)
  names(l) <- x
  list2env(l, parent = parent)
}

expect_translation <- function(con, expr, expected, ...) {
  expected <- paste0(expected, collapse = "\n")

  expr <- list(enexpr(expr))
  actual_sql <- translate_sql_(expr, con = con, ...)
  expect_equal(actual_sql, sql(expected))
}

expect_selects <- function(code, n) {
  expect_equal(n_selects(code), n)
}

expect_translation_snapshot <- function(con, expr, ..., error = FALSE) {
  expr <- substitute(expr)
  dots <- exprs(...)

  inject(expect_snapshot(
    translate_sql(!!expr, !!!dots, con = con),
    error = error
  ))
}

scrub_sqlite_version <- \(x) gsub(sqlite_version(), "<version>", x)

local_show_na_rm_warning <- function(frame = caller_env()) {
  reset_warning_verbosity("dbplyr_check_na_rm")
  local_mocked_bindings(is_testing = function() FALSE, .env = frame)
}

#' Join SQL tables
#'
#' @description
#' These are methods for the dplyr [join] generics. They are translated
#' to the following SQL queries:
#'
#' * `inner_join(x, y)`: `SELECT * FROM x JOIN y ON x.a = y.a`
#' * `left_join(x, y)`:  `SELECT * FROM x LEFT JOIN y ON x.a = y.a`
#' * `right_join(x, y)`: `SELECT * FROM x RIGHT JOIN y ON x.a = y.a`
#' * `full_join(x, y)`:  `SELECT * FROM x FULL JOIN y ON x.a = y.a`
#' * `semi_join(x, y)`:  `SELECT * FROM x WHERE EXISTS (SELECT 1 FROM y WHERE x.a = y.a)`
#' * `anti_join(x, y)`:  `SELECT * FROM x WHERE NOT EXISTS (SELECT 1 FROM y WHERE x.a = y.a)`
#'
#' @param x,y A pair of lazy data frames backed by database queries.
#' @inheritParams dplyr::join
#' @param copy If `x` and `y` are not from the same data source,
#'   and `copy` is `TRUE`, then `y` will be copied into a
#'   temporary table in same database as `x`. `*_join()` will automatically
#'   run `ANALYZE` on the created table in the hope that this will make
#'   you queries as efficient as possible by giving more data to the query
#'   planner.
#'
#'   This allows you to join tables across srcs, but it's potentially expensive
#'   operation so you must opt into it.
#' @param auto_index if `copy` is `TRUE`, automatically create
#'   indices for the variables in `by`. This may speed up the join if
#'   there are matching indexes in `x`.
#' @param sql_on A custom join predicate as an SQL expression.
#'   Usually joins use column equality, but you can perform more complex
#'   queries by supply `sql_on` which should be a SQL expression that
#'   uses `LHS` and `RHS` aliases to refer to the left-hand side or
#'   right-hand side of the join respectively.
#' @param na_matches Should NA (NULL) values match one another?
#'   The default, "never", is how databases usually work. `"na"` makes
#'   the joins behave like the dplyr join functions, [merge()], [match()],
#'   and `%in%`.
#' @param multiple,unmatched Unsupported in database backends. As a workaround
#'   for multiple use a unique key and for unmatched a foreign key constraint.
#' @param relationship Unsupported in database backends.
#' @param x_as,y_as Alias to use for `x` resp. `y`. Defaults to `"LHS"` resp.
#'   `"RHS"`
#' @inherit arrange.tbl_lazy return
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' band_db <- tbl_memdb(dplyr::band_members)
#' instrument_db <- tbl_memdb(dplyr::band_instruments)
#' band_db %>% left_join(instrument_db) %>% show_query()
#'
#' # Can join with local data frames by setting copy = TRUE
#' band_db %>%
#'   left_join(dplyr::band_instruments, copy = TRUE)
#'
#' # Unlike R, joins in SQL don't usually match NAs (NULLs)
#' db <- memdb_frame(x = c(1, 2, NA))
#' label <- memdb_frame(x = c(1, NA), label = c("one", "missing"))
#' db %>% left_join(label, by = "x")
#' # But you can activate R's usual behaviour with the na_matches argument
#' db %>% left_join(label, by = "x", na_matches = "na")
#'
#' # By default, joins are equijoins, but you can use `sql_on` to
#' # express richer relationships
#' db1 <- memdb_frame(x = 1:5)
#' db2 <- memdb_frame(x = 1:3, y = letters[1:3])
#' db1 %>% left_join(db2) %>% show_query()
#' db1 %>% left_join(db2, sql_on = "LHS.x < RHS.x") %>% show_query()
#' @name join.tbl_sql
NULL

#' @rdname join.tbl_sql
#' @export
#' @importFrom dplyr inner_join
inner_join.tbl_lazy <- function(x,
                                y,
                                by = NULL,
                                copy = FALSE,
                                suffix = NULL,
                                ...,
                                keep = NULL,
                                na_matches = c("never", "na"),
                                multiple = NULL,
                                unmatched = "drop",
                                relationship = NULL,
                                sql_on = NULL,
                                auto_index = FALSE,
                                x_as = NULL,
                                y_as = NULL) {
  check_dots_empty()
  x$lazy_query <- add_join(
    x,
    y,
    type = "inner",
    by = by,
    copy = copy,
    suffix = suffix,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    unmatched = unmatched,
    relationship = relationship,
    sql_on = sql_on,
    auto_index = auto_index,
    x_as = x_as,
    y_as = y_as
  )

  x
}

#' @rdname join.tbl_sql
#' @export
#' @importFrom dplyr left_join
left_join.tbl_lazy <- function(x,
                               y,
                               by = NULL,
                               copy = FALSE,
                               suffix = NULL,
                               ...,
                               keep = NULL,
                               na_matches = c("never", "na"),
                               multiple = NULL,
                               unmatched = "drop",
                               relationship = NULL,
                               sql_on = NULL,
                               auto_index = FALSE,
                               x_as = NULL,
                               y_as = NULL) {
  check_dots_empty()
  x$lazy_query <- add_join(
    x,
    y,
    type = "left",
    by = by,
    copy = copy,
    suffix = suffix,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    unmatched = unmatched,
    relationship = relationship,
    sql_on = sql_on,
    auto_index = auto_index,
    x_as = x_as,
    y_as = y_as
  )

  x
}

#' @rdname join.tbl_sql
#' @export
#' @importFrom dplyr right_join
right_join.tbl_lazy <- function(x,
                                y,
                                by = NULL,
                                copy = FALSE,
                                suffix = NULL,
                                ...,
                                keep = NULL,
                                na_matches = c("never", "na"),
                                multiple = NULL,
                                unmatched = "drop",
                                relationship = NULL,
                                sql_on = NULL,
                                auto_index = FALSE,
                                x_as = NULL,
                                y_as = NULL) {
  check_dots_empty()
  x$lazy_query <- add_join(
    x,
    y,
    type = "right",
    by = by,
    copy = copy,
    suffix = suffix,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    unmatched = unmatched,
    relationship = relationship,
    sql_on = sql_on,
    auto_index = auto_index,
    x_as = x_as,
    y_as = y_as
  )

  x
}

#' @rdname join.tbl_sql
#' @export
#' @importFrom dplyr full_join
full_join.tbl_lazy <- function(x,
                               y,
                               by = NULL,
                               copy = FALSE,
                               suffix = NULL,
                               ...,
                               keep = NULL,
                               na_matches = c("never", "na"),
                               multiple = NULL,
                               relationship = NULL,
                               sql_on = NULL,
                               auto_index = FALSE,
                               x_as = NULL,
                               y_as = NULL) {
  check_dots_empty()
  x$lazy_query <- add_join(
    x,
    y,
    type = "full",
    by = by,
    copy = copy,
    suffix = suffix,
    keep = keep,
    na_matches = na_matches,
    multiple = multiple,
    relationship = relationship,
    sql_on = sql_on,
    auto_index = auto_index,
    x_as = x_as,
    y_as = y_as
  )

  x
}

#' @rdname join.tbl_sql
#' @export
#' @importFrom dplyr cross_join
cross_join.tbl_lazy <- function(x,
                                y,
                                ...,
                                copy = FALSE,
                                suffix = c(".x", ".y"),
                                x_as = NULL,
                                y_as = NULL) {
  x$lazy_query <- add_join(
    x, y,
    "cross",
    by = character(),
    sql_on = NULL,
    copy = copy,
    suffix = suffix,
    auto_index = FALSE,
    keep = NULL,
    na_matches = "never",
    x_as = x_as,
    y_as = y_as,
    ...
  )

  x
}

#' @rdname join.tbl_sql
#' @export
#' @importFrom dplyr semi_join
semi_join.tbl_lazy <- function(x,
                               y,
                               by = NULL,
                               copy = FALSE,
                               ...,
                               na_matches = c("never", "na"),
                               sql_on = NULL,
                               auto_index = FALSE,
                               x_as = NULL,
                               y_as = NULL) {
  check_dots_empty()
  x$lazy_query <- add_semi_join(
    x,
    y,
    anti = FALSE,
    by = by,
    copy = copy,
    na_matches = na_matches,
    sql_on = sql_on,
    auto_index = auto_index,
    x_as = x_as,
    y_as = y_as
  )

  x
}

#' @rdname join.tbl_sql
#' @export
#' @importFrom dplyr anti_join
anti_join.tbl_lazy <- function(x,
                               y,
                               by = NULL,
                               copy = FALSE,
                               ...,
                               na_matches = c("never", "na"),
                               sql_on = NULL,
                               auto_index = FALSE,
                               x_as = NULL,
                               y_as = NULL) {
  check_dots_empty()
  x$lazy_query <- add_semi_join(
    x,
    y,
    anti = TRUE,
    by = by,
    copy = copy,
    na_matches = na_matches,
    sql_on = sql_on,
    auto_index = auto_index,
    x_as = x_as,
    y_as = y_as
  )

  x
}

add_join <- function(x,
                     y,
                     type,
                     by = NULL,
                     copy = FALSE,
                     suffix = NULL,
                     keep = NULL,
                     na_matches = "never",
                     multiple = NULL,
                     unmatched = "drop",
                     relationship = NULL,
                     sql_on = NULL,
                     auto_index = FALSE,
                     x_as = NULL,
                     y_as = NULL,
                     call = caller_env()) {
  x_names <- tbl_vars(x)
  y_names <- tbl_vars(y)

  by <- join_prepare_by(
    by = by,
    sql_on = sql_on,
    type = type,
    x_names = x_names,
    y_names = y_names,
    error_call = call
  )
  type <- by$type

  check_join_by_supported(by, call = call)

  check_join_multiple(multiple, by, call = call)
  check_join_unmatched(unmatched, by, call = call)
  check_unsupported_arg(
    relationship,
    allowed = "many-to-many",
    allow_null = TRUE,
    call = call
  )

  y <- auto_copy(
    x, y,
    copy = copy,
    indexes = if (auto_index) list(by$y)
  )

  suffix <- suffix %||% sql_join_suffix(x$src$con, suffix)
  vars <- join_cols(
    x_names = x_names,
    y_names = y_names,
    by = by,
    suffix = suffix,
    keep = keep,
    error_call = call
  )

  na_matches <- arg_match(na_matches, c("na", "never"), error_call = call)

  # the table alias can only be determined after `select()` was inlined.
  # This works even though `by` is used in `join_inline_select()` and updated
  # because this does not touch `by$x_as` and `by$y_as`.
  join_alias <- make_join_aliases(x$src$con, x_as, y_as, sql_on, call)

  inline_result <- join_inline_select(x$lazy_query, by$x, by$on)
  x_lq <- inline_result$lq
  x_vars <- inline_result$vars
  by$x <- inline_result$by

  inline_result <- join_inline_select(y$lazy_query, by$y, by$on)
  y_lq <- inline_result$lq
  y_vars <- inline_result$vars
  by$y <- inline_result$by

  if (type %in% c("full", "right")) {
    vars <- rf_join_vars(
      x_vars = x_vars,
      y_vars = y_vars,
      vars_info = vars,
      keep = keep,
      condition = by$condition,
      by_y = by$y,
      type = type
    )

    table_names_x <- make_table_names(join_alias$x, x_lq)
    table_names_y <- make_table_names(join_alias$y, y_lq)
    out <- lazy_rf_join_query(
      x = x_lq,
      y = y_lq,
      type = type,
      by = list(
        x = by$x,
        y = by$y,
        condition = by$condition,
        on = sql(by$on),
        na_matches = na_matches
      ),
      table_names = vctrs::vec_rbind(table_names_x, table_names_y),
      vars = vars
    )

    return(out)
  }

  new_query <- join_needs_new_query(x$lazy_query, join_alias, type)
  vars <- multi_join_vars(
    x_lq = x_lq,
    x_vars = x_vars,
    y_vars = y_vars,
    new_query = new_query,
    vars_info = vars,
    error_call = call
  )

  joins_data <- new_joins_data(
    x_lq,
    y_lq,
    new_query = new_query,
    type = type,
    by = by,
    na_matches = na_matches
  )

  table_names_y <- make_table_names(join_alias$y, y_lq)

  if (new_query) {
    table_names_x <- make_table_names(join_alias$x, x_lq)
    out <- lazy_multi_join_query(
      x = x_lq,
      joins = joins_data,
      table_names = vctrs::vec_rbind(table_names_x, table_names_y),
      vars = vars
    )
    return(out)
  }

  # `x_lq` must be a `lazy_multi_join_query` so it can be modified directly
  if (!is_null(join_alias$x)) {
    x_lq$table_names$name[[1]] <- join_alias$x
    x_lq$table_names$from[[1]] <- "as"
  }

  x_lq$joins <- vctrs::vec_rbind(x_lq$joins, joins_data)
  x_lq$table_names <- vctrs::vec_rbind(x_lq$table_names, table_names_y)
  x_lq$vars <- vars

  x_lq
}

join_inline_select <- function(lq, by, on) {
  if (is_empty(on) && is_lazy_select_query_simple(lq, select = "projection")) {
    vars <- purrr::map_chr(lq$select$expr, as_string)

    idx <- vctrs::vec_match(lq$select$name, by)
    by <- vctrs::vec_assign(by, idx, vars)

    lq_org <- lq
    lq <- lq$x
    lq$group_vars <- op_grps(lq_org)
    lq$order_vars <- op_sort(lq_org)
    lq$frame <- op_frame(lq_org)
  } else {
    vars <- op_vars(lq)
  }

  list(
    lq = lq,
    vars = vars,
    by = by
  )
}

join_needs_new_query <- function(x_lq, join_alias, type) {
  if (!inherits(x_lq, "lazy_multi_join_query")) {
    return(TRUE)
  }

  x_as <- join_alias$x
  y_as <- join_alias$y

  names <- x_lq$table_names$name
  from <- x_lq$table_names$from
  if (!is_null(x_as)) {
    if (from[[1]] == "as" && !identical(x_as, names[[1]])) {
      return(TRUE)
    }

    if (x_as %in% names[-1][from[-1] == "as"]) {
      return(TRUE)
    }
  }

  if (!is_null(y_as)) {
    if (y_as %in% names[from == "as"]) {
      return(TRUE)
    }
  }

  FALSE
}

multi_join_vars <- function(x_lq,
                            x_vars,
                            y_vars,
                            new_query,
                            vars_info,
                            error_call) {
  if (new_query) {
    x_join_vars <- tibble(
      name = names(vars_info$x$out),
      table = 1L,
      var = vctrs::vec_slice(x_vars, vars_info$x$out)
    )
    table_id <- 2L
  } else {
    x_join_vars <- vctrs::vec_slice(x_lq$vars, vars_info$x$out)
    x_join_vars$name <- names(vars_info$x$out)
    table_id <- vctrs::vec_size(x_lq$table_names) + 1L
  }

  y_join_vars <- tibble(
    name = names(vars_info$y$out),
    table = table_id,
    var = vctrs::vec_slice(y_vars, vars_info$y$out)
  )

  vctrs::vec_rbind(x_join_vars, y_join_vars, .error_call = error_call)
}

rf_join_vars <- function(x_vars,
                         y_vars,
                         vars_info,
                         keep,
                         condition,
                         by_y,
                         type,
                         call) {
  x_join_vars <- tibble(
    name = names(vars_info$x$out),
    x = vctrs::vec_slice(x_vars, vars_info$x$out),
    y = NA_character_
  )

  y_join_vars <- tibble(
    name = names(vars_info$y$out),
    x = NA_character_,
    y = vctrs::vec_slice(y_vars, vars_info$y$out)
  )

  keep <- keep %||% (condition != "==")
  idx <- vars_info$x$key[!keep]
  if (type == "right") {
    # `x`: non-join variables; `y`: all variables
    x_join_vars$x[idx] <- NA
  }
  x_join_vars$y[idx] <- by_y[!keep]

  vctrs::vec_rbind(x_join_vars, y_join_vars)
}


join_prepare_by <- function(by,
                            sql_on,
                            type,
                            x_names,
                            y_names,
                            error_call,
                            env = caller_env(2),
                            user_env = caller_env(3)) {
  if (identical(by, character()) && is.null(sql_on)) {
    if (type != "cross") {
      lifecycle::deprecate_warn(
        when = "1.1.0",
        what = I("Using `by = character()` to perform a cross join"),
        with = "cross_join()",
        env = env,
        user_env = user_env
      )
    }
    type <- "cross"
  }

  if (!is.null(sql_on)) {
    by <- list(x = character(0), y = character(0), on = unclass(sql_on))
  } else if (type == "cross") {
    by <- list(x = character(0), y = character(0))
  } else if (is_null(by)) {
    by <- join_by_common(x_names, y_names, error_call = error_call)
  } else {
    by <- dbplyr_as_join_by(by, error_call = error_call)
  }

  by$type <- type
  by
}

new_joins_data <- function(x_lq, y_lq, new_query, type, by, na_matches) {
  if (new_query) {
    by_x_table_id <- rep_along(by$x, 1L)
  } else {
    idx <- vctrs::vec_match(by$x, x_lq$vars$name)

    # need to fix `by$x` in case it was renamed in an inlined select
    by$x <- x_lq$vars$var[idx]
    by_x_table_id <- x_lq$vars$table[idx]
  }

  tibble(
    table = list(y_lq),
    type = type,
    by_x_table_id = list(by_x_table_id),
    by = list(list(
      x = by$x,
      y = by$y,
      condition = by$condition,
      on = sql(by$on),
      na_matches = na_matches
    ))
  )
}

add_semi_join <- function(x,
                          y,
                          anti = FALSE,
                          by = NULL,
                          sql_on = NULL,
                          copy = FALSE,
                          auto_index = FALSE,
                          na_matches = "never",
                          x_as = NULL,
                          y_as = NULL,
                          call = caller_env()) {
  x_names <- tbl_vars(x)
  y_names <- tbl_vars(y)

  if (!is.null(sql_on)) {
    by <- list(x = character(0), y = character(0), on = sql(sql_on))
  } else if (is_null(by)) {
    by <- join_by_common(x_names, y_names, error_call = call)
  } else {
    by <- dbplyr_as_join_by(by, error_call = call)
  }

  check_join_by_supported(by, call = call)
  na_matches <- arg_match(na_matches, c("na", "never"), error_call = call)

  y <- auto_copy(
    x, y, copy,
    indexes = if (auto_index) list(by$y)
  )

  inline_result <- join_inline_select(x$lazy_query, by$x, sql_on)
  x_lq <- inline_result$lq
  x_vars <- inline_result$vars
  by$x <- inline_result$by

  inline_result <- semi_join_inline_select(y$lazy_query, by$y, sql_on)
  y_lq <- inline_result$lq
  by$y <- inline_result$by
  where <- inline_result$where

  vars <- tibble(
    name = op_vars(x),
    var = x_vars
  )
  by$na_matches <- na_matches

  # the table alias can only be determined after `select()` was inlined
  join_alias <- make_join_aliases(x$src$con, x_as, y_as, sql_on, call)

  aliases <- vctrs::vec_rbind(
    make_table_names(join_alias$x, x_lq),
    make_table_names(join_alias$y, y_lq)
  )
  names <- generate_join_table_names(aliases, remote_con(x))
  by$x_as <- names[1]
  by$y_as <- names[2]

  lazy_semi_join_query(
    x_lq,
    y_lq,
    vars = vars,
    anti = anti,
    by = by,
    where = where,
    call = call
  )
}

can_inline_semi_join <- function(x) {
  is_lazy_select_query_simple(
    x,
    ignore_where = TRUE,
    ignore_group_by = TRUE,
    select = "projection"
  )
}

semi_join_inline_select <- function(lq, by, on) {
  if (is_empty(on) && can_inline_semi_join(lq)) {
    vars <- purrr::map_chr(lq$select$expr, as_string)

    idx <- vctrs::vec_match(lq$select$name, by)
    by <- vctrs::vec_assign(by, idx, vars)

    lq_org <- lq
    lq <- lq$x
    lq$group_vars <- op_grps(lq_org)
    lq$order_vars <- op_sort(lq_org)
    lq$frame <- op_frame(lq_org)

    where <- lq_org$where
  } else {
    vars <- op_vars(lq)
    where <- NULL
  }

  list(
    lq = lq,
    vars = vars,
    by = by,
    where = where
  )
}

make_join_aliases <- function(con, x_as, y_as, sql_on, call) {
  x_as <- check_join_as1(x_as, arg = "x_as", sql_on, default = "LHS", call)
  y_as <- check_join_as1(y_as, arg = "y_as", sql_on, default = "RHS", call)

  if (identical(x_as, y_as) && !is.null(x_as)) {
    cli_abort("{.arg y_as} must be different from {.arg x_as}.", call = call)
  }

  list(
    x = if (!is.null(x_as)) as_table_path(x_as, con),
    y = if (!is.null(y_as)) as_table_path(y_as, con)
  )
}

make_table_names <- function(as, lq) {
  name <- remote_name(lq, null_if_local = FALSE)

  if (!is.null(as)) {
    tibble(name = as, from = "as")
  } else if (!is.null(name)) {
    tibble(name = name, from = "name")
  } else {
    tibble(name = table_path(""), from = "")
  }
}

check_join_as1 <- function(as, arg, sql_on, default, call) {
  check_string(as, allow_null = TRUE, arg = arg, call = call)

  if (!is_null(sql_on)) {
    # for backwards compatibility use "LHS"/"RHS" if `sql_on` is used
    # without a table alias
    as <- as %||% default
  }

  as
}

join_vars <- function(x_names, y_names, type, by, suffix = c(".x", ".y"), call = caller_env()) {
  y_names_org <- y_names
  # Remove join keys from y
  y_names <- setdiff(y_names, by$y)

  # Add suffix where needed
  x_new <- add_suffixes(x_names, y_names, suffix$x)
  y_new <- add_suffixes(y_names, x_names, suffix$y)

  # In left and inner joins, return key values only from x
  # In right joins, return key values only from y
  # In full joins, return key values by coalescing values from x and y
  x_x <- x_names
  x_y <- by$y[match(x_names, by$x)]
  x_y[type == "left" | type == "inner"] <- NA
  x_x[type == "right" & !is.na(x_y)] <- NA
  y_x <- rep_len(NA, length(y_names))
  y_y <- y_names

  # Return a list with 3 parallel vectors
  # At each position, values in the 3 vectors represent
  #  alias - name of column in join result
  #  x - name of column from left table or NA if only from right table
  #  y - name of column from right table or NA if only from left table
  list(
    alias = c(x_new, y_new),
    x = c(x_x, y_x),
    y = c(x_y, y_y),
    all_x = x_names,
    all_y = y_names_org
  )
}

# names are bare table names without qualifiers
# only called from generate_join_table_names
join_two_table_alias <- function(names, from) {
  check_character(names)
  check_character(from)
  vctrs::vec_assert(names, size = 2L)

  out <- names
  out[from == ""] <- c("LHS", "RHS")[from == ""]

  if (!identical(out[1], out[2])) {
    return(out)
  }
  # -> must rename

  tables_have_same_name <- from[1] == "name" && from[2] == "name" && identical(names[1], names[2])
  if (tables_have_same_name) {
    return(paste0(names, c("_LHS", "_RHS")))
  }

  out_repaired <- vctrs::vec_as_names(out, repair = "unique", quiet = TRUE)
  may_repair_name <- c(from[1] != "as", from[2] != "as")
  out[may_repair_name] <- out_repaired[may_repair_name]

  out
}

check_join_by_supported <- function(by, call = caller_env()) {
  if (any(by$filter != "none")) {
    cli_abort("Rolling joins aren't supported on database backends.", call = call)
  }
}

check_join_multiple <- function(multiple, by, call = caller_env()) {
  if (is.null(multiple) || identical(multiple, "all")) {
    return()
  }

  msg <- "Argument {.arg multiple} isn't supported on database backends."

  if (all(by$condition == "==")) {
    code <- glue::glue("
      db_create_index(
        con = remote_con(y),
        table = remote_name(y),
        columns = {deparse(by$y)},
        unique = TRUE
      )"
    )

    msg <- c(
      msg,
      i = c(
        "For equi joins you can instead add a unique index for the join columns in {.arg y}.",
        "{.code {code}}"
      )
    )
  }

  cli_abort(msg, call = call)
}

check_join_unmatched <- function(unmatched, by, call = caller_env()) {
  if (is.null(unmatched) || identical(unmatched, "drop")) {
    return()
  }

  cli_abort(c(
    "Argument {.arg unmatched} isn't supported on database backends.",
    i = "For equi joins you can instead add a foreign key from {.arg x} to {.arg y} for the join columns."
  ), call = call)
}

#' Fill in missing values with previous or next value
#'
#' @inheritParams arrange.tbl_lazy
#' @param ... Columns to fill.
#' @param .direction Direction in which to fill missing values. Currently
#'   either "down" (the default) or "up". Note that "up" does not work when
#'   `.data` is sorted by non-numeric columns. As a workaround revert the order
#'   yourself beforehand; for example replace `arrange(x, desc(y))` by
#'   `arrange(desc(x), y)`.
#'
#' @examplesIf rlang::is_installed("tidyr", version = "1.0.0")
#' squirrels <- tibble::tribble(
#'   ~group,    ~name,     ~role,     ~n_squirrels, ~ n_squirrels2,
#'   1,      "Sam",    "Observer",   NA,                 1,
#'   1,     "Mara", "Scorekeeper",    8,                NA,
#'   1,    "Jesse",    "Observer",   NA,                NA,
#'   1,      "Tom",    "Observer",   NA,                 4,
#'   2,     "Mike",    "Observer",   NA,                NA,
#'   2,  "Rachael",    "Observer",   NA,                 6,
#'   2,  "Sydekea", "Scorekeeper",   14,                NA,
#'   2, "Gabriela",    "Observer",   NA,                NA,
#'   3,  "Derrick",    "Observer",   NA,                NA,
#'   3,     "Kara", "Scorekeeper",    9,                 10,
#'   3,    "Emily",    "Observer",   NA,                NA,
#'   3, "Danielle",    "Observer",   NA,                NA
#' )
#' squirrels$id <- 1:12
#'
#' tbl_memdb(squirrels) %>%
#'   window_order(id) %>%
#'   tidyr::fill(
#'     n_squirrels,
#'     n_squirrels2,
#'   )
fill.tbl_lazy <- function(.data, ..., .direction = c("down", "up", "updown", "downup")) {
  cols_to_fill <- tidyselect::eval_select(expr(c(...)), .data)
  cols_to_fill <- syms(names(cols_to_fill))
  order_by_cols <- op_sort(.data)
  .direction <- arg_match0(.direction, c("down", "up", "updown", "downup"))

  if (is_empty(order_by_cols)) {
    cli_abort(
      c(
        x = "{.arg .data} does not have explicit order.",
        i = "Please use {.fun dbplyr::window_order} to make order explicit."
      )
    )
  }

  if (.direction == "updown") {
    .data <- tidyr::fill(.data, !!!cols_to_fill, .direction = "up")
    .data <- tidyr::fill(.data, !!!cols_to_fill, .direction = "down")
    return(.data)
  } else if (.direction == "downup") {
    .data <- tidyr::fill(.data, !!!cols_to_fill, .direction = "down")
    .data <- tidyr::fill(.data, !!!cols_to_fill, .direction = "up")
    return(.data)
  }

  if (.direction == "up") {
    order_by_cols <- purrr::map(order_by_cols, swap_order_direction)
  }

  dbplyr_fill0(
    .con = remote_con(.data),
    .data = .data,
    cols_to_fill = cols_to_fill,
    order_by_cols = order_by_cols,
    .direction = .direction
  )
}

swap_order_direction <- function(x) {
  if (is_quosure(x)) {
    x <- quo_get_expr(x)
  }

  if (is_call(x, "desc", n = 1)) {
    call_args(x)[[1]]
  } else {
    expr(desc(!!x))
  }
}

dbplyr_fill0 <- function(.con, .data, cols_to_fill, order_by_cols, .direction) {
  UseMethod("dbplyr_fill0")
}

# databases with support for `IGNORE NULLS`
# * hive: https://cwiki.apache.org/confluence/display/Hive/LanguageManual+WindowingAndAnalytics
# * impala: https://docs.cloudera.com/documentation/enterprise/5-11-x/topics/impala_analytic_functions.html
# * oracle: https://oracle-base.com/articles/misc/first-value-and-last-value-analytic-functions
# * redshift: https://docs.aws.amazon.com/redshift/latest/dg/r_WF_first_value.html
# * teradata: https://docs.teradata.com/r/756LNiPSFdY~4JcCCcR5Cw/V~t1FC7orR6KCff~6EUeDQ
#' @export
dbplyr_fill0.DBIConnection <- function(.con,
                                          .data,
                                          cols_to_fill,
                                          order_by_cols,
                                          .direction) {
  # strategy:
  # 1. construct a window
  # * from the first row to the current row
  # * ordered by `order_by_cols`
  # * partitioned by groups if any
  # 2. in this window use the last value ignoring nulls; in SQL this is (usually)
  #    `LAST_VALUE(<column> IGNORE NULLS)`
  #
  # Source: https://dzone.com/articles/how-to-fill-sparse-data-with-the-previous-non-empt
  grps <- op_grps(.data)

  fill_sql <- purrr::map(
    cols_to_fill,
    ~ win_over(
      last_value_sql(.con, .x),
      partition = if (!is_empty(grps)) escape(ident(op_grps(.data)), con = .con),
      order = translate_sql(!!!order_by_cols, con = .con),
      con = .con
    )
  ) %>%
    set_names(as.character(cols_to_fill))

  .data %>%
    transmute(
      !!!syms(colnames(.data)),
      !!!fill_sql
    )
}

# databases without support for `IGNORE NULLS`
# * sqlite
# * access: https://www.techonthenet.com/access/functions/
# * hana: https://help.sap.com/viewer/4fe29514fd584807ac9f2a04f6754767/2.0.04/en-US/e7ef7cc478f14a408e1af27fc1a78624.html
# * mysql: https://dev.mysql.com/doc/refman/8.0/en/window-function-descriptions.html
# * mariadb
# * postgres: https://www.postgresql.org/docs/13/functions-window.html
# * mssql: https://docs.microsoft.com/en-us/sql/t-sql/functions/first-value-transact-sql?view=sql-server-ver15
#   -> `IGNORE NULLS` only in Azure SQL Edge
#' @export
dbplyr_fill0.SQLiteConnection <- function(.con,
                                             .data,
                                             cols_to_fill,
                                             order_by_cols,
                                             .direction) {
  # this strategy is used for databases that don't support `IGNORE NULLS` in
  # `LAST_VALUE()`.
  #
  # strategy:
  # for each column to fill:
  # 1. generate a helper column `....dbplyr_partion_x`. It creates one partition
  #    per non-NA value and all following NA (in the order of `order_by_cols`),
  #    i.e. each partition has exactly one non-NA value and any number of NA.
  # 2. use the non-NA value in each partition (`max()` is just the simplest
  #    way to do that reliable among databases).
  # 3. remove the helper column again.
  partition_sql <- purrr::map(
    cols_to_fill,
    ~ translate_sql(
      cumsum(case_when(is.na(!!.x) ~ 0L, TRUE ~ 1L)),
      con = .con,
      vars_order = translate_sql(!!!order_by_cols, con = .con),
      vars_group = op_grps(.data),
    )
  ) %>%
    set_names(paste0("..dbplyr_partion_", seq_along(cols_to_fill)))

  dp <- .data %>%
    mutate(!!!partition_sql)

  fill_sql <- purrr::map2(
    cols_to_fill, names(partition_sql),
    ~ translate_sql(
      max(!!.x, na.rm = TRUE),
      con = .con,
      vars_group = c(op_grps(.data), .y),
    )
  ) %>%
    set_names(purrr::map_chr(cols_to_fill, as_name))

  dp %>%
    transmute(
      !!!syms(colnames(.data)),
      !!!fill_sql
    ) %>%
    select(!!!colnames(.data))
}

#' @export
dbplyr_fill0.PostgreSQL <- dbplyr_fill0.SQLiteConnection
#' @export
dbplyr_fill0.PqConnection <- dbplyr_fill0.SQLiteConnection
#' @export
dbplyr_fill0.PostgreSQLConnection <- dbplyr_fill0.SQLiteConnection

#' @export
dbplyr_fill0.HDB <- dbplyr_fill0.SQLiteConnection

#' @export
dbplyr_fill0.ACCESS <- dbplyr_fill0.SQLiteConnection
#' @export
`dbplyr_fill0.Microsoft SQL Server` <- dbplyr_fill0.SQLiteConnection

#' @export
dbplyr_fill0.MariaDBConnection <- dbplyr_fill0.SQLiteConnection
#' @export
dbplyr_fill0.MySQLConnection <- dbplyr_fill0.SQLiteConnection
#' @export
dbplyr_fill0.MySQL <- dbplyr_fill0.SQLiteConnection


last_value_sql <- function(con, x) {
  UseMethod("last_value_sql")
}

#' @export
last_value_sql.DBIConnection <- function(con, x) {
  build_sql("LAST_VALUE(", ident(as.character(x)), " IGNORE NULLS)", con = con)
}

#' @export
`last_value_sql.Microsoft SQL Server` <- function(con, x) {
  build_sql("LAST_VALUE(", ident(as.character(x)), ") IGNORE NULLS", con = con)
}

globalVariables("last_value")

uses_mutated_vars <- function(dots, select) {
  if (is_null(select)) {
    return(TRUE)
  }

  vars <- set_names(select$expr, select$name)
  mutated_vars <- names(purrr::discard(vars, is_symbol))

  if (is_empty(mutated_vars)) {
    return(FALSE)
  }

  any(purrr::map_lgl(dots, expr_uses_var, mutated_vars))
}

expr_uses_var <- function(x, vars) {
  if (is.sql(x)) {
    return(TRUE)
  }
  if (is_symbol(x)) {
    return(is_symbol(x, vars))
  }
  if (!is_call(x)) {
    return(FALSE)
  }

  any(purrr::map_lgl(as.list(x[-1]), expr_uses_var, vars))
}

any_expr_uses_sql <- function(dots) {
  purrr::some(dots, expr_uses_sql)
}

expr_uses_sql <- function(x) {
  if (is.sql(x)) {
    return(TRUE)
  }
  if (!is_call(x)) {
    return(FALSE)
  }

  any(purrr::map_lgl(as.list(x[-1]), expr_uses_sql))
}

#' @export
#' @rdname sql_build
lazy_query <- function(
  query_type,
  x,
  ...,
  group_vars = op_grps(x),
  order_vars = op_sort(x),
  frame = op_frame(x)
) {
  stopifnot(
    is.null(group_vars) ||
      (is.character(group_vars) && is.null(names(group_vars)))
  )
  stopifnot(is_lazy_sql_part(order_vars), is.null(names(order_vars)))
  check_frame(frame)

  structure(
    list(
      x = x,
      ...,
      group_vars = group_vars,
      order_vars = order_vars,
      frame = frame
    ),
    class = c(paste0("lazy_", query_type, "_query"), "lazy_query")
  )
}


#' @export
print.lazy_query <- function(x, ...) {
  cat_line("-- <", class(x)[1], "> -------------------------")
  cat_line(sql_render(x, simulate_dbi()))
}

# query -----------------------------------------------------------------------

query <- function(query_type, ...) {
  structure(
    list(...),
    class = c(paste0(query_type, "_query"), "query")
  )
}

#' @export
print.query <- function(x, ...) {
  cat_line("-- <", class(x)[1], "> -------------------------")
  cat_line(sql_render(x, simulate_dbi()))
}

#' Infrastructure for testing dplyr
#'
#' Register testing sources, then use `test_load()` to load an existing
#' data frame into each source. To create a new table in each source,
#' use `test_frame()`.
#'
#' @keywords internal
#' @examples
#' \dontrun{
#' test_register_src("sqlite", {
#'   DBI::dbConnect(RSQLite::SQLite(), ":memory:", create = TRUE)
#' })
#'
#' test_frame(x = 1:3, y = 3:1)
#' test_load(mtcars)
#' }
#' @name testing
NULL


#' @export
#' @rdname testing
test_register_src <- function(name, src) {
  message("Registering testing src: ", name, " ", appendLF = FALSE)
  tryCatch(
    {
      test_srcs$add(name, src)
      message("OK")
    },
    error = function(e) message("\n* ", conditionMessage(e))
  )
}

#' @export
#' @rdname testing
test_register_con <- function(name, ...) {
  test_register_src(name, DBI::dbConnect(...))
}

#' @export
#' @rdname testing
src_test <- function(name) {
  srcs <- test_srcs$get()
  if (!name %in% names(srcs)) {
    testthat::skip(paste0("No ", name))
  } else {
    srcs[[name]]
  }
}

#' @export
#' @rdname testing
test_load <- function(df, name = unique_table_name(), srcs = test_srcs$get(),
                      ignore = character()) {
  stopifnot(is.data.frame(df))
  stopifnot(is.character(ignore))

  srcs <- srcs[setdiff(names(srcs), ignore)]
  lapply(srcs, copy_to, df, name = name)
}

#' @export
#' @rdname testing
test_frame <- function(..., srcs = test_srcs$get(), ignore = character()) {
  df <- tibble(...)
  test_load(df, srcs = srcs, ignore = ignore)
}

# Manage cache of testing srcs
test_srcs <- local({
  list(
    get = function() env_get(cache(), "srcs", list()),

    has = function(x) {
      srcs <- env_get(cache(), "srcs", list())
      has_name(srcs, x)
    },

    add = function(name, src) {
      srcs <- env_get(cache(), "srcs", list())
      srcs[[name]] <- src
      env_poke(cache(), "srcs", srcs)
    },

    set = function(...) {
      env_poke(cache(), "src", list(...))
    },

    length = function() {
      length(cache()$srcs)
    }
  )
})


# Modern helpers ----------------------------------------------------------

copy_to_test <- function(src, df, ..., name = "test") {
  copy_to(src_test(src), df, name, ..., overwrite = TRUE)
}

test_that("rendering table wraps in SELECT *", {
  out <- copy_to_test("sqlite", tibble(x = 1), name = "test-sql-build")
  expect_snapshot(out %>% sql_render())
  expect_equal(out %>% collect(), tibble(x = 1))
})

compare_tbl <- function(x, y, label = NULL, expected.label = NULL) {
  testthat::expect_equal(
    arrange(collect(x), dplyr::across(everything())),
    arrange(collect(y), dplyr::across(everything())),
    label = label,
    expected.label = expected.label
  )
}

expect_equal_tbls <- function(results, ref = NULL, ...) {
  stopifnot(is.list(results))

  if (!is_named(results)) {
    result_name <- expr_name(substitute(results)) # nocov
    names(results) <- paste0(result_name, "_", seq_along(results)) # nocov
  }

  # If ref is NULL, use the first result
  if (is.null(ref)) {
    if (length(results) < 2) {
      testthat::skip("Need at least two srcs to compare")
    }

    ref <- results[[1]]
    ref_name <- names(results)[[1]]

    rest <- results[-1]
  } else {
    rest <- results
    ref_name <- "`ref`"
  }

  for (i in seq_along(rest)) {
    compare_tbl(
      rest[[i]], ref,
      label = names(rest)[[i]],
      expected.label = ref_name
    )
  }

  invisible(TRUE)
}

#' Count observations by group
#'
#' These are methods for the dplyr [count()] and [tally()] generics. They
#' wrap up [group_by.tbl_lazy()], [summarise.tbl_lazy()] and, optionally,
#' [arrange.tbl_lazy()].
#'
#' @inheritParams arrange.tbl_lazy
#' @inheritParams dplyr::count
#' @inheritParams args_by
#' @param .drop Not supported for lazy tables.
#' @importFrom dplyr count
#' @export
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(g = c(1, 1, 1, 2, 2), x = c(4, 3, 6, 9, 2))
#' db %>% count(g) %>% show_query()
#' db %>% count(g, wt = x) %>% show_query()
#' db %>% count(g, wt = x, sort = TRUE) %>% show_query()
count.tbl_lazy <- function(x, ..., wt = NULL, sort = FALSE, name = NULL) {
  if (!missing(...)) {
    out <- group_by(x, ..., .add = TRUE)
  } else {
    out <- x
  }
  out <- tally(out, wt = {{ wt }}, sort = sort, name = name)
  out <- group_by(out, !!!syms(group_vars(x)))
  out
}

#' @rdname count.tbl_lazy
#' @importFrom dplyr add_count
#' @export
add_count.tbl_lazy <- function (x, ..., wt = NULL, sort = FALSE, name = NULL, .drop = NULL) {
  if (!missing(.drop)) {
    cli_abort("{.arg .drop} argument not supported for lazy tables.")
  }

  if (!missing(...)) {
    out <- group_by(x, ..., .add = TRUE)
  } else {
    out <- x
  }
  out <- dplyr::add_tally(out, wt = !!enquo(wt), sort = sort, name = name)
  group_by(out, !!!syms(group_vars(x)))
}

#' @rdname count.tbl_lazy
#' @importFrom dplyr tally
#' @export
tally.tbl_lazy <- function(x, wt = NULL, sort = FALSE, name = NULL) {
  wt <- enquo(wt)
  if (quo_is_null(wt)) {
    n <- expr(n())
  } else {
    n <- expr(sum(!!wt, na.rm = TRUE))
  }

  name <- check_name(name, group_vars(x))

  out <- summarise(x, !!name := !!n, .groups = "drop")

  if (sort) {
    arrange(out, desc(!!sym(name)))
  } else {
    out
  }
}

n_name <- function (x) {
  name <- "n"
  while (name %in% x) {
    name <- paste0("n", name)
  }
  name
}

check_name <- function(name, vars, arg = caller_arg(name), call = caller_env()) {
  if (is.null(name)) {
    name <- n_name(vars)
    if (name != "n") {
      cli::cli_inform(c(
        "Storing counts in {.field {name}}, as {.field n} already present in input",
        i = "Use {.code name = \"new_name\"} to pick a new name."
      ))
    }
  } else {
    # TODO use rlang checker functions
    # check_string(name, arg = arg, call = call)
  }
  name
}

#' SQL helpers for string functions
#'
#' @description
#' These functions help you create custom string SQL translations when
#' implementing a new backend. They are typically used within [sql_translator()]
#' to define how R string functions should be translated to SQL.
#'
#' * `sql_substr()` creates a SQL substring function translator that converts
#'   R's `substr(x, start, stop)` to SQL's `SUBSTR(x, start, length)`.
#' * `sql_str_sub()` creates a SQL substring function translator that handles
#'   stringr's `str_sub()` with support for negative indices.
#' * `sql_paste()` creates a SQL paste function using `CONCAT_WS()` or similar.
#' * `sql_paste_infix()` creates a SQL paste function using an infix operator
#'   like `||`.
#'
#' @param f The name of the SQL function as a string.
#' @family SQL translation helpers
#' @name sql_translation_string
NULL

#' @export
#' @rdname sql_translation_string
sql_substr <- function(f = "SUBSTR") {
  # R prefers to specify start / stop or start / end
  # databases usually specify start / length
  # https://www.postgresql.org/docs/current/functions-string.html
  function(x, start, stop) {
    start <- max(cast_number_whole(start), 1L)
    stop <- max(cast_number_whole(stop), 1L)
    length <- max(stop - start + 1L, 0L)

    sql_glue("{.sql f}({x}, {start}, {length})")
  }
}

cast_number_whole <- function(x, arg = caller_arg(x), call = caller_env()) {
  check_number_whole(x, arg = arg, call = call)
  vctrs::vec_cast(x, integer(), x_arg = arg)
}

# str_sub(x, start, end) - start and end can be negative
# SUBSTR(string, start, length) - start can be negative

#' @export
#' @rdname sql_translation_string
#' @param subset_f The name of the SQL substring function.
#' @param length_f The name of the SQL string length function.
#' @param optional_length Whether the length argument is optional in the SQL
#'   substring function.
sql_str_sub <- function(
  subset_f = "SUBSTR",
  length_f = "LENGTH",
  optional_length = TRUE
) {
  function(string, start = 1L, end = -1L) {
    start <- cast_number_whole(start)
    end <- cast_number_whole(end)

    start_sql <- start_pos(string, start, length_f)

    if (optional_length && end == -1L) {
      sql_glue("{.sql subset_f}({string}, {start_sql})")
    } else {
      if (end == 0L) {
        length_sql <- 0L
      } else if (start > 0 && end < 0) {
        n <- start - end - 2L
        if (n == 0) {
          length_sql <- sql_glue("{.sql length_f}({string})")
        } else {
          length_sql <- sql_glue("{.sql length_f}({string}) - {n}")
        }
      } else {
        length_sql <- pmax(end - start + 1L, 0L)
      }
      sql_glue("{.sql subset_f}({string}, {start_sql}, {length_sql})")
    }
  }
}

start_pos <- function(string, start, length_f) {
  if (start == -1) {
    sql_glue("{.sql length_f}({string})")
  } else if (start < 0) {
    sql_glue("{.sql length_f}({string}) - {abs(start + 1L)}")
  } else {
    start
  }
}

sql_str_trim <- function(string, side = c("both", "left", "right")) {
  side <- match.arg(side)
  switch(
    side,
    left = sql_glue("LTRIM({string})"),
    right = sql_glue("RTRIM({string})"),
    both = sql_glue("LTRIM(RTRIM({string}))"),
  )
}


sql_str_pattern_switch <- function(
  string,
  pattern,
  negate = FALSE,
  f_fixed = NULL,
  f_regex = NULL,
  error_call = caller_env()
) {
  pattern_quo <- enquo(pattern)
  is_fixed <- quo_is_call(pattern_quo, "fixed") ||
    inherits(pattern, "stringr_fixed")

  if (is_fixed) {
    f_fixed(string, pattern, negate)
  } else {
    if (is_null(f_regex)) {
      cli_abort(
        "Only fixed patterns are supported on this backend.",
        call = error_call
      )
    } else {
      f_regex(string, pattern, negate)
    }
  }
}

# INSTR
# * SQLite https://www.sqlitetutorial.net/sqlite-functions/sqlite-instr/
# * MySQL https://dev.mysql.com/doc/refman/8.0/en/string-functions.html#function_instr
# * Oracle https://docs.oracle.com/en/database/oracle/oracle-database/19/sqlrf/INSTR.html#GUID-47E3A7C4-ED72-458D-A1FA-25A9AD3BE113
# * Teradata https://docs.teradata.com/r/Teradata-VantageTM-SQL-Functions-Expressions-and-Predicates/March-2019/String-Operators-and-Functions/INSTR
# * Access https://support.microsoft.com/de-de/office/instr-funktion-85d3392c-3b1c-4232-bb18-77cd0cb8a55b
# * Hana https://help.sap.com/docs/SAP_HANA_PLATFORM/e8e6c8142e60469bb401de5fdb6f7c00/f5a9ca3718354a499a98ba61ae3da170.html
# * Hive https://www.revisitclass.com/hadoop/instr-function-in-hive-with-examples/
# * Impala https://impala.apache.org/docs/build/html/topics/impala_string_functions.html#string_functions__instr
# POSITION
# * Snowflake https://docs.snowflake.com/en/sql-reference/functions/position
sql_str_detect_fixed_instr <- function(type = c("detect", "start", "end")) {
  type <- arg_match(type)

  function(string, pattern, negate = FALSE) {
    con <- sql_current_con()
    pattern <- unclass(pattern)
    index_sql <- sql_glue2(con, "INSTR({string}, {pattern})")

    if (negate) {
      switch(
        type,
        detect = translate_sql(!!index_sql == 0L, con = con),
        start = translate_sql(!!index_sql != 1L, con = con),
        end = translate_sql(
          !!index_sql != nchar(!!string) - nchar(!!pattern) + 1L,
          con = con
        )
      )
    } else {
      switch(
        type,
        detect = translate_sql(!!index_sql > 0L, con = con),
        start = translate_sql(!!index_sql == 1L, con = con),
        end = translate_sql(
          !!index_sql == nchar(!!string) - nchar(!!pattern) + 1L,
          con = con
        )
      )
    }
  }
}

sql_str_detect_fixed_position <- function(type = c("detect", "start", "end")) {
  type <- arg_match(type)

  function(string, pattern, negate = FALSE) {
    con <- sql_current_con()
    pattern <- unclass(pattern)
    index_sql <- sql_glue2(con, "POSITION({pattern} in {string})")

    if (negate) {
      switch(
        type,
        detect = translate_sql(!!index_sql == 0L, con = con),
        start = translate_sql(!!index_sql != 1L, con = con),
        end = translate_sql(
          !!index_sql != nchar(!!string) - nchar(!!pattern) + 1L,
          con = con
        )
      )
    } else {
      switch(
        type,
        detect = translate_sql(!!index_sql > 0L, con = con),
        start = translate_sql(!!index_sql == 1L, con = con),
        end = translate_sql(
          !!index_sql == nchar(!!string) - nchar(!!pattern) + 1L,
          con = con
        )
      )
    }
  }
}

#' @export
#' @rdname sql_translation_string
#' @param default_sep The default separator for paste operations.
sql_paste <- function(default_sep, f = "CONCAT_WS") {
  function(..., sep = default_sep, collapse = NULL) {
    check_collapse(collapse)
    sql_glue("{.sql f}({sep}, {...})")
  }
}

#' @export
#' @rdname sql_translation_string
#' @param op The SQL operator to use for infix paste operations.
#' @param cast A function to cast values to strings.
sql_paste_infix <- function(default_sep, op, cast = sql_cast("text")) {
  check_string(default_sep)
  check_string(op)
  check_function(cast)

  function(..., sep = default_sep, collapse = NULL) {
    check_collapse(collapse)

    args <- list(...)
    if (length(args) == 1) {
      return(cast(args[[1]]))
    }

    if (sep == "") {
      infix <- function(x, y) sql_glue("{x} {.sql op} {y}")
    } else {
      infix <- function(x, y) sql_glue("{x} {.sql op} {sep} {.sql op} {y}")
    }

    purrr::reduce(args, infix)
  }
}

check_collapse <- function(collapse) {
  if (is.null(collapse)) {
    return()
  }

  cli_abort(c(
    "{.arg collapse} not supported in DB translation of {.fun paste}.",
    i = "Please use {.fun str_flatten} instead."
  ))
}

deprecate_ignore_case <- function(ignore_case, frame = caller_env()) {
  if (lifecycle::is_present(ignore_case)) {
    lifecycle::deprecate_warn(
      when = "2.6.0",
      what = "str_like(ignore_case)",
      details = c(
        "`str_like()` is always case sensitive.",
        "Use `str_ilike()` for case insensitive string matching."
      ),
      env = frame
    )
    check_bool(ignore_case, call = frame)
    ignore_case
  } else {
    FALSE
  }
}

#' @export
#' @rdname sql_build
lazy_union_query <- function(x, unions, call = caller_env()) {
  check_lazy_query(x, call = call)

  lazy_query(
    query_type = "union",
    x = x,
    unions = unions
  )
}

#' @export
op_vars.lazy_union_query <- function(op) {
  purrr::reduce(
    op$unions$table,
    \(acc, table) union(acc, op_vars(table$lazy_query)),
    .init = op_vars(op$x)
  )
}

#' @export
sql_build.lazy_union_query <- function(op, con, ..., sql_options = NULL) {
  # add_union() ensures that both have same variables
  unions <- list(
    table = purrr::map(
      op$unions$table,
      function(table) {
        sql_build(table, con, sql_options = sql_options)
      }
    ),
    all = op$unions$all
  )

  union_query(
    sql_build(op$x, con, sql_options = sql_options),
    unions
  )
}

# Built query -------------------------------------------------------------

#' @export
#' @rdname sql_build
union_query <- function(x, unions) {
  query("union", x = x, unions = unions)
}

#' @export
sql_render.union_query <- function(
  query,
  con = NULL,
  ...,
  sql_options = NULL,
  subquery = FALSE,
  lvl = 0
) {
  from_x <- sql_render(query$x, con, ..., subquery = FALSE, lvl = lvl)
  unions <- list()
  unions$table <- purrr::map(
    query$unions$table,
    \(table) sql_render(table, con, ..., subquery = FALSE, lvl = lvl)
  )
  unions$all <- query$unions$all

  sql_query_union(con, from_x, unions, lvl = lvl)
}

#' @export
flatten_query.union_query <- function(qry, query_list, con) {
  x <- qry$x
  query_list_new <- flatten_query(x, query_list, con)
  qry$x <- get_subquery_name(x, query_list_new)

  for (i in seq_along(qry$unions$table)) {
    y <- qry$unions$table[[i]]
    query_list_new <- flatten_query(y, query_list_new, con)
    qry$unions$table[[i]] <- get_subquery_name(y, query_list_new)
  }

  # TODO reuse query
  name <- as_table_path(unique_subquery_name(), con)
  wrapped_query <- set_names(list(qry), name)

  query_list$queries <- c(query_list_new$queries, wrapped_query)
  query_list$name <- name
  query_list
}

# SQL generation ----------------------------------------------------------

#' @rdname db-sql
#' @export
sql_query_union <- function(con, x, unions, ..., lvl = 0) {
  UseMethod("sql_query_union", sql_dialect(con))
}
#' @export
sql_query_union.DBIConnection <- function(con, x, unions, ..., lvl = 0) {
  ops <- ifelse(unions$all, "UNION ALL", "UNION")
  ops <- purrr::map_chr(ops, \(op) sql_set_op_method(con, op))
  ops <- indent_lvl(style_kw(ops), lvl)
  tables <- unlist(unions$table)

  union_clauses <- vctrs::vec_interleave(as.character(ops), tables)
  out <- paste0(
    x,
    "\n\n",
    paste0(union_clauses, collapse = "\n\n")
  )

  sql(out)
}

#' @export
sql_query_union.sql_dialect <- sql_query_union.DBIConnection

cache <- function() {
  if (!is_attached("dbplyr_cache")) {
    get("attach")(new_environment(), name = "dbplyr_cache", pos = length(search()) - 1)
  }
  search_env("dbplyr_cache")
}

cache_computation <- function(name, computation) {
  cache <- cache()

  if (env_has(cache, name)) {
    env_get(cache, name)
  } else {
    res <- force(computation)
    env_poke(cache, name, res)
    res
  }
}

# nocov start
load_srcs <- function(f, src_names, quiet = NULL) {
  if (is.null(quiet)) {
    quiet <- !identical(Sys.getenv("NOT_CRAN"), "true")
  }

  srcs <- lapply(src_names, function(x) {
    out <- NULL
    try(out <- f(x), silent = TRUE)
    if (is.null(out) && !quiet) {
      message("Could not instantiate ", x, " src")
    }
    out
  })

  purrr::compact(setNames(srcs, src_names))
}


db_location <- function(path = NULL, filename) {
  if (!is.null(path)) {
    # Check that path is a directory and is writeable
    if (!file.exists(path) || !file.info(path)$isdir) {
      cli_abort("{.path {path}} is not a directory")
    }
    if (!is_writeable(path)) cli_abort("Can not write to {.path {path}}")
    return(file.path(path, filename))
  }

  pkg <- file.path(system.file("db", package = "dplyr"))
  if (is_writeable(pkg)) return(file.path(pkg, filename))

  tmp <- tempdir()
  if (is_writeable(tmp)) return(file.path(tmp, filename))

  cli_abort("Could not find writeable location to cache db")
}

is_writeable <- function(x) {
  unname(file.access(x, 2) == 0)
}
# nocov end

#' Build and render SQL from a sequence of lazy operations
#'
#' `sql_build()` creates a `select_query` S3 object, that is rendered
#' to a SQL string by `sql_render()`. The output from `sql_build()` is
#' designed to be easy to test, as it's database agnostic, and has
#' a hierarchical structure. Outside of testing, however, you should
#' always call `sql_render()`.
#'
#' `sql_build()` is generic over the lazy operations, \link{lazy_ops},
#' and generates an S3 object that represents the query. `sql_render()`
#' takes a query object and then calls a function that is generic
#' over the database. For example, `sql_build.op_mutate()` generates
#' a `select_query`, and `sql_render.select_query()` calls
#' `sql_select()`, which has different methods for different databases.
#' The default methods should generate ANSI 92 SQL where possible, so you
#' backends only need to override the methods if the backend is not ANSI
#' compliant.
#'
#' @export
#' @keywords internal
#' @param op A sequence of lazy operations
#' @param con A database connection. The default `NULL` uses a set of
#'   rules that should be very similar to ANSI 92, and allows for testing
#'   without an active database connection.
#' @param ... Other arguments passed on to the methods. Not currently used.
#' @param sql_options SQL rendering options generated by `sql_options()`.
sql_build <- function(op, con = NULL, ..., sql_options = NULL) {
  if (is.null(sql_options)) {
    sql_options <- as_sql_options(sql_options)
    out <- sql_build(
      op = op,
      con = con,
      ...,
      sql_options = sql_options
    )

    return(out)
  }

  unique_subquery_name_reset()
  unique_column_name_reset()
  check_dots_used()
  UseMethod("sql_build")
}

#' @export
sql_build.tbl_lazy <- function(op, con = op$src$con, ..., sql_options = NULL) {
  con <- con %||% op$src$con
  sql_options <- sql_options %||% sql_options()

  # only used for testing
  qry <- sql_build(
    op$lazy_query,
    con = con,
    ...,
    sql_options = sql_options
  )
  sql_optimise(qry, con = con, ...)
}


# Render ------------------------------------------------------------------

#' @export
#' @rdname sql_build
#' @param subquery Is this SQL going to be used in a subquery?
#'   This is important because you can place a bare table name in a subquery
#'   and  ORDER BY does not work in subqueries.
sql_render <- function(query,
                       con = NULL,
                       ...,
                       sql_options = NULL,
                       subquery = FALSE,
                       lvl = 0) {
  if (is.null(sql_options)) {
    sql_options <- as_sql_options(sql_options)
    out <- sql_render(
      query = query,
      con = con,
      ...,
      sql_options = sql_options,
      subquery = subquery,
      lvl = lvl
    )

    return(out)
  }

  check_dots_used()
  UseMethod("sql_render")
}

#' @export
sql_render.tbl_lazy <- function(query,
                                con = query$src$con,
                                ...,
                                sql_options = NULL,
                                subquery = FALSE,
                                lvl = 0) {
  con <- con %||% query$src$con
  sql_render(
    query$lazy_query,
    con = con,
    ...,
    sql_options = sql_options,
    subquery = subquery,
    lvl = lvl
  )
}

#' @export
sql_render.lazy_query <- function(query,
                                  con = NULL,
                                  ...,
                                  sql_options = NULL,
                                  subquery = FALSE,
                                  lvl = 0) {
  qry <- sql_build(query, con = con, sql_options = sql_options)
  qry <- sql_optimise(qry, con = con, subquery = subquery)

  if (sql_options$cte) {
    query_list <- flatten_query(qry, list(queries = list(), name = NULL), con = con)
    queries <- query_list$queries

    rendered_queries <- purrr::map2(
      queries, seq_along(queries) != length(queries),
      function(query, indent) {
        sql_render(query, con = con, ..., subquery = subquery, lvl = as.integer(indent))
      }
    )

    cte_render(rendered_queries, con)
  } else {
    sql_render(qry, con = con, ..., subquery = subquery, lvl = lvl)
  }
}

cte_render <- function(query_list, con) {
  n <- length(query_list)
  if (n == 1) {
    return(query_list[[1]])
  }

  ctes <- purrr::imap(
    query_list[-n],
    function(query, name) {
      name <- table_path(name)
      glue_sql2(con, "{.name name} {.kw 'AS'} (\n{query}\n)")
    }
  )
  cte_query <- sql_vector(unname(ctes), parens = FALSE, collapse = ",\n", con = con)

  glue_sql2(con, "{.kw 'WITH'} ", cte_query, "\n", query_list[[n]])
}

get_subquery_name <- function(x, query_list) {
  if (inherits(x, "base_query")) return(x)

  base_query(query_list$name)
}

flatten_query <- function(qry, query_list, con) {
  UseMethod("flatten_query")
}

querylist_reuse_query <- function(qry, query_list, con) {
  id <- vctrs::vec_match(list(unclass(qry)), purrr::map(query_list$queries, unclass))

  if (!is.na(id)) {
    query_list$name <- names(query_list$queries)[[id]]
  } else {
    name <- as_table_path(unique_subquery_name(), con)
    wrapped_query <- set_names(list(qry), name)
    query_list$queries <- c(query_list$queries, wrapped_query)
    query_list$name <- name
  }

  query_list
}

flatten_query_2_tables <- function(qry, query_list, con) {
  x <- qry$x
  query_list_x <- flatten_query(x, query_list, con)
  qry$x <- get_subquery_name(x, query_list_x)

  y <- qry$y
  query_list_y <- flatten_query(y, query_list_x, con)
  qry$y <- get_subquery_name(y, query_list_y)

  querylist_reuse_query(qry, query_list_y, con)
}


# Optimise ----------------------------------------------------------------

#' @export
#' @rdname sql_build
sql_optimise <- function(x, con = NULL, ..., subquery = FALSE) {
  UseMethod("sql_optimise")
}

test_that("queries generated by select() don't alias unnecessarily", {
  lf_build <- lazy_frame(x = 1, y = 1) |>
    select(x) |>
    sql_build()

  lf_render <- sql_render(lf_build, con = dialect_ansi())
  expect_snapshot(lf_render)
})

test_that("queries generated by select() don't alias unnecessarily", {
  expect_snapshot(error = TRUE, lazy_frame(x = 1, y = 1) |> select())
})

test_that("basic distinct behaviour is as expected", {
  db <- local_memdb_frame(x = c(1, 1, 1, 1), y = c(1, 1, 2, 2))

  expect_equal(db |> distinct() |> collect(), tibble(x = c(1, 1), y = c(1, 2)))
  expect_equal(db |> distinct(x) |> collect(), tibble(x = 1))
  expect_equal(db |> distinct(y) |> collect(), tibble(y = c(1, 2)))
})

test_that("correctly inlines across all verbs", {
  lf <- lazy_frame(x = 1, y = 2)

  # single table verbs
  expect_selects(lf |> arrange(x) |> distinct(), 1)
  expect_selects(lf |> distinct() |> distinct(), 1)
  expect_selects(lf |> filter(x == 1) |> distinct(), 1)
  expect_selects(lf |> head(1) |> distinct(), 2)
  expect_selects(lf |> mutate(z = x + 1) |> distinct(), 1)
  expect_selects(lf |> select(y = x) |> distinct(), 1)
  expect_selects(lf |> summarise(y = mean(x)) |> distinct(), 1)

  # two table verbs
  lf2 <- lazy_frame(x = 1)
  expect_selects(lf |> left_join(lf2, by = "x") |> distinct(), 1)
  expect_selects(lf |> right_join(lf2, by = "x") |> distinct(), 2)
  expect_selects(lf |> semi_join(lf2, by = "x") |> distinct(), 3)
  expect_selects(lf |> union(lf2) |> distinct(), 3)

  # special cases
  # doesn't inline if .keep_all is TRUE, since that requires a window function
  expect_selects(
    lf |> left_join(lf2, by = "x") |> distinct(y, .keep_all = TRUE),
    3
  )
})

test_that("distinct returns all columns when .keep_all is TRUE", {
  mf <- local_memdb_frame(x = c(1, 1, 2, 2), y = 1:4)

  result <- mf |> distinct(x, .keep_all = TRUE) |> collect()
  expect_named(result, c("x", "y"))
  expect_equal(result$x, c(1, 2))
  expect_equal(group_vars(result), character())
})

test_that("distinct doesn't duplicate column names if grouped (#354)", {
  df <- lazy_frame(a = 1)
  expect_equal(df |> group_by(a) |> distinct() |> op_vars(), "a")
})

test_that("distinct respects groups", {
  df <- local_memdb_frame(a = 1:2, b = 1) |> group_by(a)
  expect_equal(df |> group_by(a) |> distinct() |> op_vars(), c("a", "b"))
})


test_that("distinct respects groups when .keep_all is TRUE", {
  mf <- local_memdb_frame(x = c(1, 1, 2, 2), y = 1:4)

  result <- mf |> group_by(x) |> distinct(.keep_all = TRUE) |> collect()
  expect_named(result, c("x", "y"))
  expect_equal(result$x, c(1, 2))
  expect_equal(group_vars(result), "x")
})

test_that("distinct can select variables via pick() #1125", {
  lf <- lazy_frame(x_1 = 1, x_2 = 1, y = 1)
  expect_equal(
    lf |> distinct(pick(starts_with("x_"))) |> remote_query(),
    sql("SELECT DISTINCT \"x_1\", \"x_2\"\nFROM \"df\"")
  )
})

test_that("distinct() produces optimized SQL", {
  lf <- lazy_frame(x = 1, y = 1)

  # can use renamed variable
  out <- lf |>
    select(a = x, y) |>
    distinct(a)

  expect_equal(
    remote_query(out),
    sql("SELECT DISTINCT \"x\" AS \"a\"\nFROM \"df\"")
  )

  expect_true(out$lazy_query$distinct)
  expect_equal(out$lazy_query$select$name, "a")
  expect_equal(out$lazy_query$select$expr, syms("x"))

  # unnecessary extra variables do not matter
  out <- lf |>
    mutate(z = 1) |>
    group_by(x) |>
    distinct(y)
  expect_s3_class(out$lazy_query$x, "lazy_base_local_query")
  expect_equal(out$lazy_query$select$name, c("x", "y"))
  expect_equal(out$lazy_query$select$expr, syms(c("x", "y")))

  # inlined after `summarise()`
  out <- lf |>
    group_by(x) |>
    summarise(y = mean(y, na.rm = TRUE)) |>
    distinct(x, y)

  expect_equal(
    remote_query(out),
    sql(
      "SELECT DISTINCT \"x\", AVG(\"y\") AS \"y\"\nFROM \"df\"\nGROUP BY \"x\""
    )
  )

  expect_true(out$lazy_query$distinct)
  expect_equal(out$lazy_query$select$name, c("x", "y"))
  expect_equal(
    out$lazy_query$select$expr,
    list(sym("x"), expr(mean(y, na.rm = TRUE)))
  )
  expect_equal(out$lazy_query$group_by, syms("x"))

  # inlined after `filter()`
  out <- lf |>
    filter(x == 1L) |>
    distinct(y)

  expect_equal(
    remote_query(out),
    sql("SELECT DISTINCT \"y\"\nFROM \"df\"\nWHERE (\"x\" = 1)")
  )

  expect_true(out$lazy_query$distinct)
  expect_equal(out$lazy_query$select$name, "y")
  expect_equal(out$lazy_query$select$expr, syms("y"))
  expect_equal(out$lazy_query$where, list(expr(x == 1L)))

  # Note: currently this needs `distinct()` or `distinct(x, y)` because
  # `summarise()` + `select()` is not inlined.
  out <- lf |>
    group_by(x) |>
    summarise(y = mean(y, na.rm = TRUE)) |>
    filter(x == 1) |>
    distinct(x, y)

  expect_equal(
    remote_query(out),
    sql(
      "SELECT DISTINCT \"x\", AVG(\"y\") AS \"y\"\nFROM \"df\"\nGROUP BY \"x\"\nHAVING (\"x\" = 1.0)"
    )
  )
  expect_true(out$lazy_query$distinct)
  expect_equal(out$lazy_query$select$name, c("x", "y"))
  expect_equal(
    out$lazy_query$select$expr,
    list(expr(x), expr(mean(y, na.rm = TRUE)))
  )

  out <- lf |>
    arrange(y) |>
    distinct(x)

  expect_equal(
    remote_query(out),
    sql("SELECT DISTINCT \"x\"\nFROM \"df\"\nORDER BY \"y\"")
  )

  expect_snapshot(
    (out <- lf |>
      head(2) |>
      distinct(x, y))
  )

  expect_s3_class(out$lazy_query$x, "lazy_select_query")
  expect_equal(out$lazy_query$x$limit, 2)
})

test_that("distinct() after join is inlined", {
  lf1 <- lazy_frame(x = 1, y = 1)
  lf2 <- lazy_frame(x = 1, z = 2)

  out <- lf1 |>
    left_join(lf2, by = "x") |>
    distinct()
  expect_s3_class(out$lazy_query, "lazy_multi_join_query")
  expect_snapshot(show_query(out))

  out <- lf1 |>
    left_join(lf2, by = "x") |>
    distinct(x)
  expect_s3_class(out$lazy_query, "lazy_multi_join_query")
  expect_snapshot(show_query(out))
})

# sql-render --------------------------------------------------------------

test_that("distinct adds DISTINCT suffix", {
  out <- local_memdb_frame(x = c(1, 1)) |> distinct()

  expect_match(out |> sql_render(), "SELECT DISTINCT")
  expect_equal(out |> collect(), tibble(x = 1))
})

test_that("distinct can compute variables", {
  out <- local_memdb_frame(x = c(2, 1), y = c(1, 2)) |>
    distinct(z = x + y)
  expect_equal(out |> collect(), tibble(z = 3))
})

test_that("distinct ignores groups when computing variables (#1081)", {
  df <- tibble(g = c("a", "a", "b"), x = c(1, 2, 3))

  db <- local_memdb_frame("df", g = c("a", "a", "b"), x = c(1, 2, 3))
  db_distinct <- db |> group_by(g) |> distinct(n = n()) |> ungroup()

  expect_snapshot(show_query(db_distinct))
  expect_equal(collect(db_distinct), tibble(g = c("a", "b"), n = 3))
})

test_that("distinct can compute variables when .keep_all is TRUE", {
  out <- local_memdb_frame(x = c(2, 1), y = c(1, 2)) |>
    distinct(z = x + y, .keep_all = TRUE) |>
    collect()

  expect_named(out, c("x", "y", "z"))
  expect_equal(out$z, 3)
})

test_that("distinct respects window_order when .keep_all is TRUE", {
  mf <- local_memdb_frame(x = c(1, 1, 2, 2), y = 1:4)
  out <- mf |>
    window_order(desc(y)) |>
    distinct(x, .keep_all = TRUE)

  expect_equal(out |> collect(), tibble(x = 1:2, y = c(2, 4)))

  lf <- lazy_frame(x = c(1, 1, 2, 2), y = 1:4)
  expect_snapshot(
    lf |>
      window_order(desc(y)) |>
      distinct(x, .keep_all = TRUE)
  )
})

test_that("distinct uses dummy window order when .keep_all is TRUE and no order is used", {
  lf <- lazy_frame(x = 1, y = 2)
  expect_snapshot(lf |> distinct(x, .keep_all = TRUE))
})

# sql_build ---------------------------------------------------------------

test_that("distinct sets flagged", {
  out1 <- lazy_frame(x = 1) |>
    select() |>
    sql_build()
  expect_false(out1$distinct)

  out2 <- lazy_frame(x = 1) |>
    distinct() |>
    sql_build()
  expect_true(out2$distinct)
})

# ops ---------------------------------------------------------------------

test_that("distinct produces correct vars", {
  out <- lazy_frame(x = 1, y = 2) |> distinct()
  expect_equal(op_vars(out), c("x", "y"))

  out <- lazy_frame(x = 1, y = 2, z = 3) |> distinct(x, y)
  expect_equal(op_vars(out), c("x", "y"))

  out <- lazy_frame(x = 1, y = 2, z = 3) |> distinct(a = x, b = y)
  expect_equal(op_vars(out), c("a", "b"))

  out <- lazy_frame(x = 1, y = 2, z = 3) |> group_by(x) |> distinct(y)
  expect_equal(op_vars(out), c("x", "y"))
})

test_that("distinct produces correct vars when .keep_all is TRUE", {
  lf <- lazy_frame(x = 1, y = 2)
  out <- lf |> distinct(.keep_all = TRUE)
  expect_equal(op_vars(out), c("x", "y"))

  out <- lf |> distinct(x, .keep_all = TRUE)
  expect_equal(op_vars(out), c("x", "y"))

  out <- lf |> distinct(a = x, .keep_all = TRUE)
  expect_equal(op_vars(out), c("x", "y", "a"))

  out <- lazy_frame(x = 1, y = 2, z = 3) |>
    group_by(x) |>
    distinct(y, .keep_all = TRUE)
  expect_equal(op_vars(out), c("x", "y", "z"))
})

test_that("distinct respects order of the specified variables (#3195, #6156)", {
  d <- lazy_frame(x = 1:2, y = 3:4)
  expect_equal(colnames(distinct(d, y, x)), c("y", "x"))
})

test_that("distinct adds grouping variables to front if missing", {
  d <- lazy_frame(x = 1:2, y = 3:4)
  expect_equal(colnames(distinct(group_by(d, y), x)), c("y", "x"))
  expect_equal(colnames(distinct(group_by(d, y), x, y)), c("x", "y"))
})

#' SQL helpers for aggregate functions
#'
#' @description
#' These functions help you create custom aggregate SQL translations when
#' implementing a new backend. They are typically used within [sql_translator()]
#' to define how R aggregate functions should be translated to SQL.
#'
#' * `sql_aggregate()` creates a SQL aggregate function translator for functions
#'   with a single argument and an optional `na.rm` parameter (e.g., `SUM()`,
#'   `AVG()`).
#' * `sql_aggregate_2()` creates a SQL aggregate function translator for
#'   functions with exactly two arguments (e.g., `CORR()`, `COVAR_SAMP()`).
#' * `sql_aggregate_n()` creates a SQL aggregate function translator for
#'   functions with any number of arguments and an optional `na.rm` parameter
#'   (e.g., `LEAST()`, `GREATEST()`).
#' * `sql_not_supported()` creates a function that throws an informative error
#'   when a function is not supported in SQL.
#' * `sql_check_na_rm()` is a helper that you can use in aggregate functions
#'   to direct the user towards setting `na.rm = TRUE`.
#'
#' @param f The name of the SQL function as a string.
#' @param f_r The name of the R function being translated as a string.
#' @family SQL translation helpers
#' @name sql_translation_agg
NULL

#' @rdname sql_translation_agg
#' @export
sql_aggregate <- function(f, f_r = f) {
  check_string(f)

  function(x, na.rm = FALSE) {
    sql_check_na_rm(na.rm)
    sql_glue("{.sql f}({x})")
  }
}

#' @rdname sql_translation_agg
#' @export
sql_aggregate_2 <- function(f) {
  check_string(f)

  function(x, y) {
    sql_glue("{.sql f}({x}, {y})")
  }
}

#' @rdname sql_translation_agg
#' @export
sql_aggregate_n <- function(f, f_r = f) {
  check_string(f)

  function(..., na.rm = FALSE) {
    sql_check_na_rm(na.rm)
    sql_glue("{.sql f}({...})")
  }
}

sql_aggregate_win <- function(f) {
  force(f)

  function(...) {
    cli_abort(
      "{.fun {f}} is only available in a windowed ({.fun mutate}) context"
    )
  }
}

#' @rdname sql_translation_agg
#' @param na.rm Logical indicating whether missing values should be removed.
#'   In SQL, missing values are always removed in aggregate functions, so this
#'   function will warn if `na.rm` is not `TRUE`.
#' @export
sql_check_na_rm <- function(na.rm) {
  if (identical(na.rm, TRUE) || is_testing()) {
    return()
  }

  cli::cli_warn(
    c(
      "Missing values are always removed in SQL aggregation functions.",
      "Use {.code na.rm = TRUE} to silence this warning"
    ),
    .frequency = "regularly",
    .frequency_id = "dbplyr_check_na_rm"
  )
}

#' @rdname sql_translation_agg
#' @export
sql_not_supported <- function(f) {
  check_string(f)

  function(...) {
    cli_abort(
      "{.fun {f}} is not available in this SQL variant.",
      class = "dbplyr_error_unsupported_fn"
    )
  }
}

sql_agg_not_supported <- function(f, backend) {
  check_string(f)

  msg <- "Translation of {.fun {f}} in {.fun summarise} is not supported"

  if (is.null(backend)) {
    msg <- paste0(msg, " on database backends.")
  } else {
    msg <- paste0(msg, " for {backend}.")
  }

  function(...) {
    dots <- enexprs(...)
    f_call_str <- as_label(call2(f, !!!dots))
    msg <- c(
      msg,
      i = "Use a combination of {.fun distinct} and {.fun mutate} for the same result:",
      " " = "{.code mutate(<col> = {f_call_str}) |> distinct(<col>)}"
    )
    cli_abort(msg)
  }
}

sql_win_not_supported <- function(f, backend) {
  check_string(f)

  msg <- "Translation of {.fun {f}} in {.fun mutate} is not supported"

  if (is.null(backend)) {
    msg <- paste0(msg, " on database backends.")
  } else {
    msg <- paste0(msg, " for {backend}.")
  }

  function(...) {
    dots <- enexprs(...)
    f_call_str <- as_label(call2(f, !!!dots))
    msg <- c(
      msg,
      i = "Use a combination of {.fun summarise} and {.fun left_join} instead:",
      " " = "{.code df |> left_join(summarise(<col> = {f_call_str}))}."
    )
    cli_abort(msg)
  }
}

sql_quantile <- function(f, style = c("infix", "ordered"), window = FALSE) {
  force(f)
  style <- match.arg(style)
  force(window)

  function(x, probs, na.rm = FALSE) {
    check_probs(probs)
    sql_check_na_rm(na.rm)

    sql <- switch(
      style,
      infix = sql_glue("{.sql f}({x}, {probs})"),
      ordered = sql_glue("{.sql f}({probs}) WITHIN GROUP (ORDER BY {x})")
    )

    if (window) {
      sql <- win_over(
        sql,
        partition = win_current_group(),
        frame = win_current_frame()
      )
    }
    sql
  }
}

sql_median <- function(f, style = c("infix", "ordered"), window = FALSE) {
  quantile <- sql_quantile(f, style = style, window = window)
  function(x, na.rm = FALSE) {
    quantile(x, 0.5, na.rm = na.rm)
  }
}

check_probs <- function(probs, call = caller_env()) {
  # TODO min, max? Inf? NA?
  check_number_decimal(probs, call = call)

  if (length(probs) > 1) {
    cli_abort("SQL translation only supports single value for {.arg probs}.")
  }
}

test_that("aggregation functions warn once if na.rm = FALSE", {
  local_show_na_rm_warning()
  con <- dialect_ansi()

  expect_no_warning(translate_sql(mean(x, na.rm = TRUE), con = con))
  expect_snapshot(translate_sql(mean(x, na.rm = FALSE), con = con))
  expect_no_warning(translate_sql(mean(x, na.rm = FALSE), con = con))
})

test_that("warns informatively with unsupported function", {
  expect_snapshot(error = TRUE, sql_not_supported("cor")())
})

test_that("quantile and median don't change without warning", {
  con <- dialect_ansi()
  expect_snapshot(translate_sql(
    quantile(x, 0.75, na.rm = TRUE),
    con = con,
    window = FALSE
  ))
  expect_snapshot(translate_sql(
    quantile(x, 0.75, na.rm = TRUE),
    con = con,
    vars_group = "g"
  ))
  expect_snapshot(translate_sql(
    median(x, na.rm = TRUE),
    con = con,
    window = FALSE
  ))
  expect_snapshot(translate_sql(
    median(x, na.rm = TRUE),
    con = con,
    vars_group = "g"
  ))
})

test_that("checks for invalid probs", {
  expect_error(check_probs("a"), "number")
  expect_error(check_probs(1:3), "vector")
})

#' Extract a single column
#'
#' This is a method for the dplyr [pull()] generic. It evaluates the query
#' retrieving just the specified column.
#'
#' @inheritParams arrange.tbl_lazy
#' @inheritParams dplyr::pull
#' @return A vector of data.
#' @importFrom dplyr pull
#' @export
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(x = 1:5, y = 5:1)
#' db %>%
#'   mutate(z = x + y * 2) %>%
#'   pull()
pull.tbl_sql <- function(.data, var = -1, name = NULL, ...) {
  vars <- tbl_vars(.data)
  var <- tidyselect::vars_pull(vars, !!enquo(var), error_arg = "var")
  name_quo <- enquo(name)
  if (!quo_is_null(name_quo)) {
    name <- tidyselect::vars_pull(vars, !!name_quo, error_arg = "name")
  }

  .data <- ungroup(.data)
  .data <- select(.data, all_of(c(var, name)))

  .data <- collect(.data)
  out <- .data[[var]]
  if (!is.null(name)) {
    out <- set_names(out, nm = .data[[name]])
  }

  out
}

#' @include translate-sql-conditional.R
#' @include translate-sql-window.R
#' @include translate-sql-helpers.R
#' @include translate-sql-scalar.R
#' @include translate-sql-aggregate.R
#' @include translate-sql-string.R
#' @include translate-sql-cut.R
#' @include escape.R
#' @include sql.R
#' @include utils.R
#' @include import-standalone-obj-type.R
#' @include import-standalone-types-check.R
#' @include utils-check.R
#' @include db.R
#' @include db-sql.R
#' @include verb-copy-to.R
#' @include verb-copy-inline.R
NULL

#' ANSI SQL backend
#'
#' @description
#' This is the base dialect for ANSI compliant SQL, forming the foundation
#' of all other dialects. Use `dialect_ansi()` with `lazy_frame()` to see
#' simulated SQL without connecting to a live database.
#'
#' See `vignette("translation-function")` for a list of functions that are
#' translated.
#'
#' @name backend-ansi
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_ansi())
#' lf |> transmute(x = mean(b, na.rm = TRUE))
#' lf |> transmute(x = log(b), y = log(b, base = 2))
NULL

#' @export
#' @rdname backend-ansi
dialect_ansi <- function() {
  structure(
    list(
      quote_identifier = function(x) sql_quote(x, '"'),
      has = list(
        window_clause = FALSE,
        table_alias_with_as = TRUE
      )
    ),
    class = "sql_dialect"
  )
}

#' @export
#' @rdname backend-ansi
#' @param class,... No longer used.
simulate_dbi <- function(class = character(), ...) {
  structure(
    list(),
    ...,
    class = c(class, "TestConnection", "DBIConnection")
  )
}

#' @export
dbplyr_edition.TestConnection <- function(con) 2L

#' @export
sql_escape_ident.TestConnection <- function(con, x) {
  if (inherits(con, "Microsoft SQL Server")) {
    sql_quote(x, c("[", "]"))
  } else if (
    inherits(con, "MySQLConnection") ||
      inherits(con, "MariaDBConnection") ||
      inherits(con, "SQLiteConnection")
  ) {
    sql_quote(x, "`")
  } else {
    sql_quote(x, '"')
  }
}


#' @export
sql_translation.DBIConnection <- function(con) {
  sql_variant(
    base_scalar,
    base_agg,
    base_win
  )
}

#' @export
#' @rdname sql_variant
#' @format NULL
base_scalar <- sql_translator(
  `+` = function(x, y = NULL) {
    x <- escape_infix_expr(enexpr(x), x, escape_unary_minus = TRUE)
    if (is.null(y)) {
      if (is.numeric(x)) {
        x
      } else {
        sql_glue("{x}")
      }
    } else {
      y <- escape_infix_expr(enexpr(y), y)

      sql_glue("{x} + {y}")
    }
  },
  `*` = sql_infix("*"),
  `/` = sql_infix("/"),
  `%/%` = sql_not_supported("%/%"),
  `%%` = sql_infix("%"),
  `^` = sql_prefix("POWER", 2),
  `-` = function(x, y = NULL) {
    x <- escape_infix_expr(enexpr(x), x, escape_unary_minus = TRUE)
    if (is.null(y)) {
      if (is.numeric(x)) {
        -x
      } else {
        sql_glue("-{x}")
      }
    } else {
      y <- escape_infix_expr(enexpr(y), y)

      sql_glue("{x} - {y}")
    }
  },

  `$` = function(x, name) {
    if (is.ident(x)) {
      sql_glue("{x}.{name}")
    } else {
      eval(bquote(`$`(x, .(substitute(name)))))
    }
  },

  `[[` = function(x, i) {
    # `x` can be a table, column or even an expression (e.g. for json)
    i <- enexpr(i)
    if (is.character(i)) {
      i <- ident(i)
      sql_glue("{x}.{i}")
    } else if (is.numeric(i)) {
      i <- as.integer(i)
      sql_glue("{x}[{i}]")
    } else {
      cli_abort("Can only index with strings and numbers")
    }
  },
  `[` = function(x, i) {
    sql_glue("CASE WHEN ({i}) THEN ({x}) END")
  },

  `!=` = sql_infix("!="),
  `==` = sql_infix("="),
  `<` = sql_infix("<"),
  `<=` = sql_infix("<="),
  `>` = sql_infix(">"),
  `>=` = sql_infix(">="),

  `%in%` = function(x, table) {
    if (is.ident(table)) {
      return(sql_glue("{x} IN {table}"))
    }

    table <- unname(table)
    if (length(table) == 0) {
      sql("FALSE")
    } else {
      sql_glue("{x} IN {table*}")
    }
  },
  `%notin%` = function(x, table) {
    if (is.ident(table)) {
      return(sql_glue("{x} NOT IN {table}"))
    }

    table <- unname(table)
    if (length(table) == 0) {
      sql("TRUE")
    } else {
      sql_glue("{x} NOT IN {table*}")
    }
  },

  `!` = sql_prefix("NOT"),
  `&` = sql_infix("AND"),
  `&&` = sql_infix("AND"),
  `|` = sql_infix("OR"),
  `||` = sql_infix("OR"),
  xor = function(x, y) {
    sql_glue("{x} OR {y} AND NOT ({x} AND {y})")
  },

  # bitwise operators
  # SQL Syntax reference links:
  #   Hive: https://cwiki.apache.org/confluence/display/Hive/LanguageManual+UDF#LanguageManualUDF-ArithmeticOperators
  #   Impala: https://www.cloudera.com/documentation/enterprise/5-9-x/topics/impala_bit_functions.html
  #   PostgreSQL: https://www.postgresql.org/docs/7.4/functions-math.html
  #   MS SQL: https://docs.microsoft.com/en-us/sql/t-sql/language-elements/bitwise-operators-transact-sql?view=sql-server-2017
  #   MySQL https://dev.mysql.com/doc/refman/5.7/en/bit-functions.html
  #   Oracle: https://docs.oracle.com/cd/E19253-01/817-6223/chp-typeopexpr-7/index.html
  #   SQLite: https://www.tutorialspoint.com/sqlite/sqlite_bitwise_operators.htm
  #   Teradata: https://docs.teradata.com/reader/1DcoER_KpnGTfgPinRAFUw/h3CS4MuKL1LCMQmnubeSRQ
  bitwNot = \(x) sql_glue("~({x})"),
  bitwAnd = sql_infix("&"),
  bitwOr = sql_infix("|"),
  bitwXor = sql_infix("^"),
  bitwShiftL = sql_infix("<<"),
  bitwShiftR = sql_infix(">>"),

  abs = sql_prefix("ABS", 1),
  acos = sql_prefix("ACOS", 1),
  asin = sql_prefix("ASIN", 1),
  atan = sql_prefix("ATAN", 1),
  atan2 = sql_prefix("ATAN2", 2),
  ceil = sql_prefix("CEIL", 1),
  ceiling = sql_prefix("CEIL", 1),
  cos = sql_prefix("COS", 1),
  cot = sql_prefix("COT", 1),
  exp = sql_prefix("EXP", 1),
  floor = sql_prefix("FLOOR", 1),
  log = function(x, base = exp(1)) {
    if (isTRUE(all.equal(base, exp(1)))) {
      sql_glue("LN({x})")
    } else {
      sql_glue("LOG({base}, {x})")
    }
  },
  log10 = sql_prefix("LOG10", 1),
  round = function(x, digits = 0L) {
    digits <- as.integer(digits)
    sql_glue("ROUND({x}, {digits})")
  },
  sign = sql_prefix("SIGN", 1),
  sin = sql_prefix("SIN", 1),
  sqrt = sql_prefix("SQRT", 1),
  tan = sql_prefix("TAN", 1),
  # cosh, sinh, coth and tanh calculations are based on this article
  # https://en.wikipedia.org/wiki/Hyperbolic_function
  cosh = \(x) sql_glue("({sql_exp(1, x)} + {sql_exp(-1, x)}) / 2"),
  sinh = \(x) sql_glue("({sql_exp(1, x)} - {sql_exp(-1, x)}) / 2"),
  tanh = \(x) sql_glue("({sql_exp(2, x)} - 1) / ({sql_exp(2, x)} + 1)"),
  coth = \(x) sql_glue("({sql_exp(2, x)} + 1) / ({sql_exp(2, x)} - 1)"),

  `if` = function(cond, if_true, if_false = NULL) {
    sql_if(enquo(cond), enquo(if_true), enquo(if_false))
  },
  if_else = function(condition, true, false, missing = NULL) {
    check_required(true)
    check_required(false)

    sql_if(
      enquo(condition),
      enquo(true),
      enquo(false),
      enquo(missing)
    )
  },
  ifelse = function(test, yes, no) {
    check_required(yes)
    check_required(no)

    sql_if(enquo(test), enquo(yes), enquo(no))
  },

  switch = \(x, ...) sql_switch(x, ...),
  case_when = function(..., .default = NULL, .ptype = NULL, .size = NULL) {
    sql_case_when(..., .default = .default, .ptype = .ptype, .size = .size)
  },
  case_match = sql_case_match,

  `(` = function(x) {
    sql_glue("({x})")
  },
  `{` = function(x) {
    sql_glue("({x})")
  },
  desc = function(x) {
    sql_glue("{x} DESC")
  },

  is.null = sql_is_null,
  is.na = sql_is_null,
  na_if = sql_prefix("NULLIF", 2),
  coalesce = sql_prefix("COALESCE"),

  # `is_distinct_from()` and `is_not_distinct_from()` are dbplyr-only
  # translations that emit SQL's `IS DISTINCT FROM` / `IS NOT DISTINCT FROM`.
  # Defaults use approach from https://modern-sql.com/feature/is-distinct-from
  is_distinct_from = function(x, y) {
    sql_glue(
      "CASE WHEN (({x}) = ({y})) OR (({x}) IS NULL AND ({y}) IS NULL) THEN 0 ELSE 1 END = 1"
    )
  },
  is_not_distinct_from = function(x, y) {
    sql_glue(
      "CASE WHEN (({x}) = ({y})) OR (({x}) IS NULL AND ({y}) IS NULL) THEN 0 ELSE 1 END = 0"
    )
  },

  as.numeric = sql_cast("NUMERIC"),
  as.double = sql_cast("NUMERIC"),
  as.integer = sql_cast("INTEGER"),
  as.character = sql_cast("TEXT"),
  as.logical = sql_cast("BOOLEAN"),
  as.Date = sql_cast("DATE"),
  as.POSIXct = sql_cast("TIMESTAMP"),
  # MS SQL - https://docs.microsoft.com/en-us/sql/t-sql/data-types/int-bigint-smallint-and-tinyint-transact-sql
  # Hive - https://cwiki.apache.org/confluence/display/Hive/LanguageManual+Types#LanguageManualTypes-IntegralTypes(TINYINT,SMALLINT,INT/INTEGER,BIGINT)
  # Postgres - https://www.postgresql.org/docs/8.4/static/datatype-numeric.html
  # Impala - https://impala.apache.org/docs/build/html/topics/impala_bigint.html
  as.integer64 = sql_cast("BIGINT"),
  as.blob = sql_cast("BLOB"),
  as = function(x, type) {
    check_string(type)
    sql_glue("CAST({x} AS {.sql type})")
  },

  c = function(...) {
    c(...)
  },
  `:` = \(from, to) from:to,

  between = function(x, left, right) {
    sql_glue("{x} BETWEEN {left} AND {right}")
  },

  pmin = sql_aggregate_n("LEAST", "pmin"),
  pmax = sql_aggregate_n("GREATEST", "pmax"),

  # Copied onLoad
  `%>%` = NULL,

  # lubridate ---------------------------------------------------------------
  # https://en.wikibooks.org/wiki/SQL_Dialects_Reference/Functions_and_expressions/Date_and_time_functions
  as_date = sql_cast("DATE"),
  as_datetime = sql_cast("TIMESTAMP"),

  today = \() sql("CURRENT_DATE"),
  now = \() sql("CURRENT_TIMESTAMP"),

  # https://modern-sql.com/feature/extract
  year = \(x) sql_glue("EXTRACT(year FROM {x})"),
  month = \(x) sql_glue("EXTRACT(month FROM {x})"),
  day = \(x) sql_glue("EXTRACT(day FROM {x})"),
  mday = \(x) sql_glue("EXTRACT(day FROM {x})"),
  yday = sql_not_supported("yday"),
  qday = sql_not_supported("qday"),
  wday = sql_not_supported("wday"),
  hour = \(x) sql_glue("EXTRACT(hour FROM {x})"),
  minute = \(x) sql_glue("EXTRACT(minute FROM {x})"),
  second = \(x) sql_glue("EXTRACT(second FROM {x})"),

  # String functions ------------------------------------------------------
  # SQL Syntax reference links:
  #   MySQL https://dev.mysql.com/doc/refman/5.7/en/string-functions.html
  #   Hive: https://cwiki.apache.org/confluence/display/Hive/LanguageManual+UDF#LanguageManualUDF-StringFunctions
  #   Impala: https://www.cloudera.com/documentation/enterprise/5-9-x/topics/impala_string_functions.html
  #   PostgreSQL: https://www.postgresql.org/docs/9.1/static/functions-string.html
  #   MS SQL: https://docs.microsoft.com/en-us/sql/t-sql/functions/string-functions-transact-sql
  #   Oracle: https://docs.oracle.com/database/121/SQLRF/functions002.htm#SQLRF51180

  # base R
  nchar = sql_prefix("LENGTH", 1),
  nzchar = function(x, keepNA = FALSE) {
    check_bool(keepNA)
    if (keepNA) {
      exp <- expr(!!x != "")
      translate_sql(!!exp, con = sql_current_con())
    } else {
      exp <- expr((is.na(!!x) | !!x != ""))
      translate_sql(!!exp, con = sql_current_con())
    }
  },
  tolower = sql_prefix("LOWER", 1),
  toupper = sql_prefix("UPPER", 1),
  trimws = \(x, which = "both") sql_str_trim(x, side = which),
  paste = sql_paste(" "),
  paste0 = sql_paste(""),
  substr = sql_substr("SUBSTR"),
  substring = sql_substr("SUBSTR"),
  cut = sql_cut,
  runif = function(n = n(), min = 0, max = 1) {
    sql_runif("RANDOM()", n = {{ n }}, min = min, max = max)
  },

  # stringr functions
  str_length = sql_prefix("LENGTH", 1),
  str_to_lower = sql_prefix("LOWER", 1),
  str_to_upper = sql_prefix("UPPER", 1),
  str_to_title = sql_prefix("INITCAP", 1),
  str_trim = sql_str_trim,
  str_c = sql_paste(""),
  str_sub = sql_str_sub("SUBSTR"),
  # https://docs.getdbt.com/sql-reference/like is typically case sensitive (#1490)
  str_like = function(string, pattern, ignore_case = deprecated()) {
    ignore_case <- deprecate_ignore_case(ignore_case)

    if (ignore_case) {
      cli_abort(c(
        "Backend does not support case insensitive {.fn str_like}.",
        i = "Use {.fn tolower} on both arguments to achieve a case insensitive match."
      ))
    } else {
      sql_glue("{string} LIKE {pattern}")
    }
  },

  str_ilike = sql_not_supported("str_ilike"),
  str_conv = sql_not_supported("str_conv"),
  str_count = sql_not_supported("str_count"),

  fixed = function(pattern, ignore_case = FALSE) {
    check_unsupported_arg(ignore_case, allowed = FALSE)
    pattern
  },
  str_detect = function(string, pattern, negate = FALSE) {
    sql_str_pattern_switch(
      string = string,
      pattern = {{ pattern }},
      negate = negate,
      f_fixed = sql_str_detect_fixed_instr("detect")
    )
  },
  str_dup = sql_not_supported("str_dup"),
  str_ends = function(string, pattern, negate = FALSE) {
    sql_str_pattern_switch(
      string = string,
      pattern = {{ pattern }},
      negate = negate,
      f_fixed = sql_str_detect_fixed_instr("end")
    )
  },
  str_extract = sql_not_supported("str_extract"),
  str_extract_all = sql_not_supported("str_extract_all"),
  str_flatten = sql_not_supported("str_flatten"),
  str_glue = sql_not_supported("str_glue"),
  str_glue_data = sql_not_supported("str_glue_data"),
  str_interp = sql_not_supported("str_interp"),
  str_locate = sql_not_supported("str_locate"),
  str_locate_all = sql_not_supported("str_locate_all"),
  str_match = sql_not_supported("str_match"),
  str_match_all = sql_not_supported("str_match_all"),
  str_order = sql_not_supported("str_order"),
  str_pad = sql_not_supported("str_pad"),
  str_remove = sql_not_supported("str_remove"),
  str_remove_all = sql_not_supported("str_remove_all"),
  str_replace = sql_not_supported("str_replace"),
  str_replace_all = sql_not_supported("str_replace_all"),
  str_replace_na = sql_not_supported("str_replace_na"),
  str_sort = sql_not_supported("str_sort"),
  str_split = sql_not_supported("str_split"),
  str_split_fixed = sql_not_supported("str_split_fixed"),
  str_squish = sql_not_supported("str_squish"),
  str_starts = function(string, pattern, negate = FALSE) {
    sql_str_pattern_switch(
      string = string,
      pattern = {{ pattern }},
      negate = negate,
      f_fixed = sql_str_detect_fixed_instr("start")
    )
  },
  str_subset = sql_not_supported("str_subset"),
  str_trunc = sql_not_supported("str_trunc"),
  str_view = sql_not_supported("str_view"),
  str_view_all = sql_not_supported("str_view_all"),
  str_which = sql_not_supported("str_which"),
  str_wrap = sql_not_supported("str_wrap")
)

sql_exp <- function(a, x) {
  a <- as.integer(a)
  if (identical(a, 1L)) {
    sql_glue("EXP({x})")
  } else if (identical(a, -1L)) {
    sql_glue("EXP(-({x}))")
  } else {
    sql_glue("EXP({a} * ({x}))")
  }
}

sql_any_na <- function(x, window = FALSE) {
  exp <- expr(any(is.na(!!x), na.rm = TRUE))
  translate_sql(!!exp, con = sql_current_con(), window = window)
}

#' @export
#' @rdname sql_variant
#' @format NULL
base_agg <- sql_translator(
  # SQL-92 aggregates
  # http://db.apache.org/derby/docs/10.7/ref/rrefsqlj33923.html
  n = \() sql("COUNT(*)"),
  mean = sql_aggregate("AVG", "mean"),
  sum = sql_aggregate("SUM"),
  min = sql_aggregate("MIN"),
  max = sql_aggregate("MAX"),

  # https://blog.jooq.org/a-true-sql-gem-you-didnt-know-yet-the-every-aggregate-function/#comment-344695
  all = sql_aggregate("MIN"),
  any = sql_aggregate("MAX"),
  anyNA = \(x) sql_any_na(x, window = FALSE),

  sd = sql_not_supported("sd"),
  var = sql_not_supported("var"),
  cor = sql_not_supported("cor"),
  cov = sql_not_supported("cov"),

  # Ordered set functions
  quantile = sql_quantile("PERCENTILE_CONT", "ordered"),
  median = sql_median("PERCENTILE_CONT", "ordered"),

  # `first()`, `last()` and `nth()` don't work as aggregate functions in SQL
  # b/c grouping takes places before sorting => it is undefined which value comes
  # first/last. See
  # https://stackoverflow.com/a/46805009/7529482
  # first = sql_prefix("FIRST_VALUE", 1),
  # last = sql_prefix("LAST_VALUE", 1),
  # nth = sql_prefix("NTH_VALUE", 2),

  n_distinct = function(x, na.rm = FALSE) {
    sql_check_na_rm(na.rm)
    sql_glue("COUNT(DISTINCT {x})")
  }
)

#' @export
#' @rdname sql_variant
#' @format NULL
base_win <- sql_translator(
  # rank functions have a single order argument that overrides the default
  row_number = win_rank("ROW_NUMBER"),
  min_rank = win_rank("RANK"),
  rank = win_rank("RANK"),
  dense_rank = win_rank("DENSE_RANK"),
  percent_rank = win_rank("PERCENT_RANK"),
  cume_dist = win_rank("CUME_DIST"),
  ntile = function(x, n) {
    n <- as.integer(n)
    win_over(
      sql_glue("NTILE({n})"),
      win_current_group(),
      x %||% win_current_order()
    )
  },

  # Variants that take more arguments
  first = function(x, order_by = NULL, na_rm = FALSE) {
    sql_nth(
      x = x,
      n = 1L,
      order_by = order_by,
      na_rm = na_rm,
      ignore_nulls = "inside"
    )
  },
  last = function(x, order_by = NULL, na_rm = FALSE) {
    sql_nth(
      x = x,
      n = Inf,
      order_by = order_by,
      na_rm = na_rm,
      ignore_nulls = "inside"
    )
  },
  nth = function(x, n, order_by = NULL, na_rm = FALSE) {
    sql_nth(
      x = x,
      n = n,
      order_by = order_by,
      na_rm = na_rm,
      ignore_nulls = "inside"
    )
  },

  lead = function(x, n = 1L, default = NA, order_by = NULL) {
    n <- as.integer(n)
    win_over(
      sql_glue("LEAD({x}, {n}, {default})"),
      win_current_group(),
      order_by %||% win_current_order(),
      win_current_frame()
    )
  },
  lag = function(x, n = 1L, default = NA, order_by = NULL) {
    n <- as.integer(n)
    win_over(
      sql_glue("LAG({x}, {n}, {default})"),
      win_current_group(),
      order_by %||% win_current_order(),
      win_current_frame()
    )
  },
  # Recycled aggregate fuctions take single argument, don't need order and
  # include entire partition in frame.
  mean = win_aggregate("AVG"),
  sum = win_aggregate("SUM"),
  min = win_aggregate("MIN"),
  max = win_aggregate("MAX"),

  all = win_aggregate("MIN"),
  any = win_aggregate("MAX"),
  anyNA = \(x) sql_any_na(x, window = TRUE),

  sd = sql_not_supported("sd"),
  var = sql_not_supported("var"),
  cor = sql_not_supported("cor"),
  cov = sql_not_supported("cov"),

  # Ordered set functions
  quantile = sql_quantile("PERCENTILE_CONT", "ordered", window = TRUE),
  median = sql_median("PERCENTILE_CONT", "ordered", window = TRUE),

  # Counts
  n = function() {
    frame <- win_current_frame()
    win_over(
      sql("COUNT(*)"),
      partition = win_current_group(),
      order = if (!is.null(frame)) win_current_order(),
      frame = frame
    )
  },
  n_distinct = function(x, na.rm = FALSE) {
    sql_check_na_rm(na.rm)
    win_over(
      sql_glue("COUNT(DISTINCT {x})"),
      win_current_group()
    )
  },

  # Cumulative function are like recycled aggregates except that R names
  # have cum prefix, order_by is inherited and frame goes from -Inf to 0.
  cummean = win_cumulative("AVG"),
  cumsum = win_cumulative("SUM"),
  cummin = win_cumulative("MIN"),
  cummax = win_cumulative("MAX"),

  # Manually override other parameters --------------------------------------
  order_by = function(order_by, expr) {
    old <- set_win_current_order(order_by)
    on.exit(set_win_current_order(old))

    expr
  }
)

#' @export
#' @rdname sql_variant
#' @format NULL
base_no_win <- sql_translator(
  row_number = win_absent("row_number"),
  min_rank = win_absent("min_rank"),
  rank = win_absent("rank"),
  dense_rank = win_absent("dense_rank"),
  percent_rank = win_absent("percent_rank"),
  cume_dist = win_absent("cume_dist"),
  ntile = win_absent("ntile"),
  mean = win_absent("mean"),
  sd = win_absent("sd"),
  var = win_absent("var"),
  cov = win_absent("cov"),
  cor = win_absent("cor"),
  sum = win_absent("sum"),
  min = win_absent("min"),
  max = win_absent("max"),
  all = win_absent("all"),
  any = win_absent("any"),
  anyNA = win_absent("anyNA"),
  median = win_absent("median"),
  quantile = win_absent("quantile"),
  n = win_absent("n"),
  n_distinct = win_absent("n_distinct"),
  cummean = win_absent("cummean"),
  cumsum = win_absent("cumsum"),
  cummin = win_absent("cummin"),
  cummax = win_absent("cummax"),
  nth = win_absent("nth"),
  first = win_absent("first"),
  last = win_absent("last"),
  lead = win_absent("lead"),
  lag = win_absent("lag"),
  order_by = win_absent("order_by"),
  str_flatten = win_absent("str_flatten"),
  count = win_absent("count")
)

test_that("window functions without group have empty over", {
  con <- dialect_ansi()
  expect_translation(con, n(), "COUNT(*) OVER ()")
  expect_translation(con, sum(x, na.rm = TRUE), "SUM(\"x\") OVER ()")
})

test_that("aggregating window functions ignore order_by", {
  con <- dialect_ansi()
  expect_translation(con, n(), "COUNT(*) OVER ()", vars_order = "x")
  expect_translation(
    con,
    sum(x, na.rm = TRUE),
    "SUM(\"x\") OVER ()",
    vars_order = "x"
  )
})

test_that("count uses order_by if frame is used", {
  con <- dialect_ansi()
  expect_translation(
    con,
    n(),
    "COUNT(*) OVER (ORDER BY \"x\" ROWS BETWEEN 2 PRECEDING AND 1 FOLLOWING)",
    vars_order = "x",
    vars_frame = c(-2, 1)
  )
})

test_that("order_by overrides default ordering", {
  con <- dialect_ansi()
  expect_translation(
    con,
    order_by(y, cumsum(x)),
    "SUM(\"x\") OVER (ORDER BY \"y\" ROWS UNBOUNDED PRECEDING)",
    vars_order = "x"
  )
  expect_translation(
    con,
    order_by(y, cummean(x)),
    "AVG(\"x\") OVER (ORDER BY \"y\" ROWS UNBOUNDED PRECEDING)",
    vars_order = "x"
  )
  expect_translation(
    con,
    order_by(y, cummin(x)),
    "MIN(\"x\") OVER (ORDER BY \"y\" ROWS UNBOUNDED PRECEDING)",
    vars_order = "x"
  )
  expect_translation(
    con,
    order_by(y, cummax(x)),
    "MAX(\"x\") OVER (ORDER BY \"y\" ROWS UNBOUNDED PRECEDING)",
    vars_order = "x"
  )
})

test_that("cumulative windows warn if no order", {
  con <- dialect_ansi()
  expect_warning(
    translate_sql(cumsum(x), con = con),
    "does not have explicit order"
  )
  expect_warning(translate_sql(cumsum(x), con = con, vars_order = "x"), NA)
})

test_that("ntile always casts to integer", {
  con <- dialect_ansi()
  expect_translation(con, ntile(x, 10.5), "NTILE(10) OVER (ORDER BY \"x\")")
})

test_that("first, last, and nth translated to _value", {
  con <- dialect_ansi()
  expect_translation(con, first(x), "FIRST_VALUE(\"x\") OVER ()")
  expect_translation(
    con,
    first(x, na_rm = TRUE),
    "FIRST_VALUE(\"x\" IGNORE NULLS) OVER ()"
  )
  # `last()` must default to unbounded preceding and following
  expect_translation(
    con,
    last(x),
    "LAST_VALUE(\"x\") OVER (ORDER BY \"a\" ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING)",
    vars_order = "a"
  )
  expect_translation(
    con,
    last(x),
    "LAST_VALUE(\"x\") OVER (ORDER BY \"a\" ROWS BETWEEN CURRENT ROW AND UNBOUNDED FOLLOWING)",
    vars_order = "a",
    vars_frame = c(0, Inf)
  )
  expect_translation(
    con,
    nth(x, 3),
    "NTH_VALUE(\"x\", 3) OVER (ORDER BY \"a\" ROWS UNBOUNDED PRECEDING)",
    vars_order = "a",
    vars_frame = c(-Inf, 0)
  )

  # can also use a column #1236
  expect_translation(
    con,
    nth(x, n),
    "NTH_VALUE(\"x\", \"n\") OVER (ORDER BY \"a\" ROWS UNBOUNDED PRECEDING)",
    vars_order = "a",
    vars_frame = c(-Inf, 0)
  )
})

test_that("can override frame of recycled functions", {
  con <- dialect_ansi()
  expect_translation(
    con,
    sum(x, na.rm = TRUE),
    "SUM(\"x\") OVER (ORDER BY \"y\" ROWS 1 PRECEDING)",
    vars_frame = c(-1, 0),
    vars_order = "y"
  )
})

test_that("frame is checked", {
  con <- dialect_ansi()
  expect_snapshot(
    error = TRUE,
    translate_sql(sum(x, na.rm = TRUE), con = con, vars_frame = c(1, 0))
  )
})

# win_rank() -------------------------------------------------------------------

test_that("win_rank() correctly handles missing values", {
  db <- local_memdb_frame("df2", x = c(1, 1, 2, NA), id = 1:4)
  out <- db |> mutate(asc = rank(x), desc = rank(desc(x))) |> arrange(id)

  # Need to test both asc and desc because NULLs can end up in different place
  expect_equal(out |> pull(asc), c(1, 1, 3, NA))
  expect_equal(out |> pull(desc), c(2, 2, 1, NA))
})

test_that("win_rank works in both directions", {
  con <- dialect_ansi()
  sql_row_number <- win_rank("ROW_NUMBER")
  expect_translation_snapshot(con, row_number(x))
  expect_translation_snapshot(con, row_number(desc(x)))
})

test_that("win_rank works with multiple variables", {
  con <- dialect_ansi()

  expect_equal(
    translate_sql(row_number(tibble(x)), con = con),
    translate_sql(row_number(x), con = con)
  )
  expect_equal(
    translate_sql(row_number(tibble(desc(x))), con = con),
    translate_sql(row_number(desc(x)), con = con)
  )

  expect_translation_snapshot(con, row_number(tibble(x, desc(y))))
})

test_that("win_rank(c()) gives an informative error", {
  con <- dialect_ansi()
  expect_snapshot(error = TRUE, {
    translate_sql(row_number(c(x)), con = con)
  })
})

test_that("row_number() with and without group_by() and arrange()", {
  mf <- lazy_frame(x = c(1:5), y = c(rep("A", 5)), con = dialect_ansi())
  expect_snapshot(mf |> mutate(rown = row_number()))
  expect_snapshot(mf |> group_by(y) |> mutate(rown = row_number()))
  expect_snapshot(
    mf |> group_by(y) |> arrange(y) |> mutate(rown = row_number())
  )
  expect_snapshot(mf |> arrange(y) |> mutate(rown = row_number()))
})

test_that("win_cumulative works", {
  con <- dialect_ansi()
  expect_translation(
    con,
    cumsum(x, "y"),
    "SUM(\"x\") OVER (ORDER BY \"y\" ROWS UNBOUNDED PRECEDING)"
  )

  # NA values results in NA rank
  db <- local_memdb_frame(x = c(1, 2, NA, 3))
  expect_equal(
    db |> mutate(rank = dense_rank(x)) |> collect() |> arrange(x),
    tibble(x = c(1:3, NA), rank = c(1:3, NA))
  )
})


# win_over ----------------------------------------------------------------

test_that("over() only requires first argument", {
  con <- dialect_ansi()
  expect_equal(win_over("X", con = con), sql("'X' OVER ()"))
})

test_that("multiple group by or order values don't have parens", {
  con <- dialect_ansi()

  expect_equal(
    win_over(ident("x"), order = c("x", "y"), con = con),
    sql("\"x\" OVER (ORDER BY \"x\", \"y\")")
  )
  expect_equal(
    win_over(ident("x"), partition = c("x", "y"), con = con),
    sql("\"x\" OVER (PARTITION BY \"x\", \"y\")")
  )
})

# window_frame ------------------------------------------------------------

test_that("window_frame()", {
  lf <- lazy_frame(x = runif(10), y = 1:10)

  expect_snapshot(
    lf |>
      window_frame(-3, 0) |>
      window_order(x) |>
      mutate(z = sum(y, na.rm = TRUE)) |>
      show_query()
  )

  expect_snapshot(
    lf |>
      window_frame(-3) |>
      window_order(x) |>
      mutate(z = sum(y, na.rm = TRUE)) |>
      show_query()
  )
})

test_that("window_frame() checks arguments", {
  skip_if(getRversion() <= '3.5.0', "R too old")
  lf <- lazy_frame(x = runif(10), y = 1:10)

  expect_snapshot(error = TRUE, window_frame(lf, "a"))
  expect_snapshot(error = TRUE, window_frame(lf, 1:2))

  expect_snapshot(error = TRUE, window_frame(lf, 1, "a"))
  expect_snapshot(error = TRUE, window_frame(lf, 1, 1:2))
})


# named windows -----------------------------------------------------------

test_that("names windows automatically", {
  lf <- lazy_frame(
    col1 = runif(3),
    col2 = runif(3),
    col3 = runif(3),
    col4 = runif(3),
    part = c("a", "a", "b"),
    ord = 3:1,
    con = dialect_sqlite()
  ) |>
    group_by(part) |>
    window_order(ord)

  lf1 <- lf |>
    transmute(
      across(c(col1, col2), ~ sum(.x, na.rm = TRUE)),
      across(c(col3, col4), ~ order_by(desc(ord), cumsum(.x)))
    )

  sql_list <- get_select_sql(
    select = lf1$lazy_query$select,
    select_operation = "mutate",
    in_vars = op_vars(lf),
    table_alias = "df",
    con = dialect_sqlite(),
    use_star = TRUE
  )
  expect_equal(
    sql_list$window_sql,
    sql(
      "`win1` AS (PARTITION BY `part`)",
      "`win2` AS (PARTITION BY `part` ORDER BY `ord` DESC ROWS UNBOUNDED PRECEDING)"
    )
  )
  expect_equal(
    sql_list$select_sql,
    sql(
      "`part`",
      "SUM(`col1`) OVER `win1` AS `col1`",
      "SUM(`col2`) OVER `win1` AS `col2`",
      "SUM(`col3`) OVER `win2` AS `col3`",
      "SUM(`col4`) OVER `win2` AS `col4`"
    )
  )

  # Different order does not confuse naming of windows
  lf2 <- lf |>
    transmute(
      col1 = sum(col1, na.rm = TRUE),
      col3 = order_by(desc(ord), cumsum(col3)),
      col2 = sum(col2, na.rm = TRUE),
      col4 = order_by(desc(ord), cumsum(col4))
    )

  sql_list <- get_select_sql(
    select = lf2$lazy_query$select,
    select_operation = "mutate",
    in_vars = op_vars(lf),
    table_alias = "df",
    con = dialect_sqlite(),
    use_star = TRUE
  )
  expect_equal(
    sql_list$window_sql,
    sql(
      "`win1` AS (PARTITION BY `part`)",
      "`win2` AS (PARTITION BY `part` ORDER BY `ord` DESC ROWS UNBOUNDED PRECEDING)"
    )
  )
  expect_equal(
    sql_list$select_sql,
    sql(
      "`part`",
      "SUM(`col1`) OVER `win1` AS `col1`",
      "SUM(`col3`) OVER `win2` AS `col3`",
      "SUM(`col2`) OVER `win1` AS `col2`",
      "SUM(`col4`) OVER `win2` AS `col4`"
    )
  )
})

test_that("only name windows if they appear multiple times", {
  lf <- lazy_frame(
    col1 = runif(3),
    col2 = runif(3),
    col3 = runif(3),
    part = c("a", "a", "b"),
    ord = 3:1,
    con = dialect_sqlite()
  ) |>
    group_by(part) |>
    window_order(ord) |>
    transmute(
      across(c(col1, col2), ~ sum(.x, na.rm = TRUE)),
      across(c(col3), ~ order_by(desc(ord), cumsum(.x)))
    )

  sql_list <- get_select_sql(
    select = lf$lazy_query$select,
    select_operation = "mutate",
    in_vars = op_vars(lf),
    table_alias = "df",
    con = dialect_sqlite(),
    use_star = TRUE
  )
  expect_equal(sql_list$window_sql, sql("`win1` AS (PARTITION BY `part`)"))
  expect_equal(
    sql_list$select_sql,
    sql(
      "`part`",
      "SUM(`col1`) OVER `win1` AS `col1`",
      "SUM(`col2`) OVER `win1` AS `col2`",
      "SUM(`col3`) OVER (PARTITION BY `part` ORDER BY `ord` DESC ROWS UNBOUNDED PRECEDING) AS `col3`"
    )
  )
})

test_that("name windows only if supported", {
  lf <- lazy_frame(
    col1 = runif(3),
    col2 = runif(3),
    part = c("a", "a", "b"),
    con = dialect_hana()
  ) |>
    group_by(part) |>
    transmute(
      across(c(col1, col2), ~ sum(.x, na.rm = TRUE))
    )

  sql_list <- get_select_sql(
    select = lf$lazy_query$select,
    select_operation = "mutate",
    in_vars = op_vars(lf),
    table_alias = "df",
    con = dialect_hana(),
    use_star = TRUE
  )
  expect_equal(sql_list$window_sql, sql())
  expect_equal(
    sql_list$select_sql,
    sql(
      "\"part\"",
      "SUM(\"col1\") OVER (PARTITION BY \"part\") AS \"col1\"",
      "SUM(\"col2\") OVER (PARTITION BY \"part\") AS \"col2\""
    )
  )
})

test_that("dplyr.strict_sql = TRUE prevents auto conversion", {
  withr::local_options(dplyr.strict_sql = TRUE)
  con <- dialect_ansi()

  expect_translation(con, 1 + 2, "1.0 + 2.0")
  expect_snapshot(error = TRUE, {
    translate_sql(blah(x), con = con)
    translate_sql(x %blah% y, con = con)
  })
})

test_that("namespace calls are translated", {
  con <- dialect_ansi()
  expect_translation(con, dplyr::n(), "COUNT(*)", window = FALSE)
  expect_translation(con, base::ceiling(x), "CEIL(\"x\")")

  expect_snapshot(error = TRUE, {
    translate_sql(NOSUCHPACKAGE::foo(), con = con)
    translate_sql(dbplyr::NOSUCHFUNCTION(), con = con)
    translate_sql(base::abbreviate(x), con = con)
  })

  lz <- lazy_frame(x = 1)
  # Also test full pipeline to ensure that they make it through partial_eval
  expect_snapshot(error = TRUE, {
    lz |> mutate(x = NOSUCHPACKAGE::foo())
    lz |> mutate(x = dbplyr::NOSUCHFUNCTION())
    lz |> mutate(x = base::abbreviate(x))
  })
})

test_that("Wrong number of arguments raises error", {
  con <- dialect_ansi()
  expect_error(
    translate_sql(mean(1, 2, na.rm = TRUE), con = con, window = FALSE),
    "unused argument"
  )
})

test_that("between translated to special form (#503)", {
  con <- dialect_ansi()
  expect_translation(con, between(x, 1, 2), "\"x\" BETWEEN 1.0 AND 2.0")
})

test_that("%in% translation parenthesises when needed", {
  con <- dialect_ansi()
  expect_translation(con, x %in% 1L, "\"x\" IN (1)")
  expect_translation(con, x %in% c(1L), "\"x\" IN (1)")
  expect_translation(con, x %in% 1:2, "\"x\" IN (1, 2)")
  expect_translation(con, x %in% y, "\"x\" IN \"y\"")
})

test_that("%in% strips vector names", {
  con <- dialect_ansi()
  expect_translation(con, x %in% c(a = 1L), "\"x\" IN (1)")
  expect_translation(con, x %in% !!c(a = 1L), "\"x\" IN (1)")
})

test_that("%in% with empty vector", {
  con <- dialect_ansi()
  expect_translation(con, x %in% !!integer(), "FALSE")
})

test_that("%notin% translation parenthesises when needed", {
  con <- dialect_ansi()
  expect_translation(con, x %notin% 1L, "\"x\" NOT IN (1)")
  expect_translation(con, x %notin% c(1L), "\"x\" NOT IN (1)")
  expect_translation(con, x %notin% 1:2, "\"x\" NOT IN (1, 2)")
  expect_translation(con, x %notin% y, "\"x\" NOT IN \"y\"")
})

test_that("%notin% strips vector names", {
  con <- dialect_ansi()
  expect_translation(con, x %notin% c(a = 1L), "\"x\" NOT IN (1)")
  expect_translation(con, x %notin% !!c(a = 1L), "\"x\" NOT IN (1)")
})

test_that("%notin% with empty vector", {
  con <- dialect_ansi()
  expect_translation(con, x %notin% !!integer(), "TRUE")
})

test_that("n_distinct(x) translated to COUNT(distinct, x)", {
  con <- dialect_ansi()
  expect_translation(
    con,
    n_distinct(x),
    "COUNT(DISTINCT \"x\")",
    window = FALSE
  )
  expect_translation(
    con,
    n_distinct(x),
    "COUNT(DISTINCT \"x\") OVER ()",
    window = TRUE
  )
})

test_that("na_if is translated to NULLIF (#211)", {
  con <- dialect_ansi()
  expect_translation(con, na_if(x, 0L), "NULLIF(\"x\", 0)")
})

test_that("anyNA() translates like any(is.na()) (#1814)", {
  con <- dialect_ansi()
  expect_translation(con, anyNA(x), 'MAX(("x" IS NULL))', window = FALSE)
  expect_translation(con, anyNA(x), 'MAX(("x" IS NULL)) OVER ()', window = TRUE)
})

test_that("anyNA() never warns about na.rm", {
  local_show_na_rm_warning()

  con <- dialect_ansi()
  expect_no_warning(translate_sql(anyNA(x), con = con, window = FALSE))
  expect_no_warning(translate_sql(anyNA(x), con = con, window = TRUE))
})

test_that("connection affects quoting character", {
  lf <- lazy_frame(field1 = 1, field2 = 2, con = dialect_sqlite())
  out <- select(lf, field1)
  expect_match(sql_render(out), "^SELECT `field1`\nFROM `df`$")
})

test_that("magrittr pipe is translated", {
  con <- dialect_ansi()
  expect_identical(
    translate_sql(1 |> is.na(), con = con),
    translate_sql(is.na(1), con = con)
  )
})

test_that("user infix functions are translated", {
  con <- dialect_ansi()
  expect_translation(con, x %foo% y, "\"x\" foo \"y\"")

  # keep case and also works with vectors of length > 1 #1299
  expect_translation(con, x %foo% (1:2), "\"x\" foo (1, 2)")
})

test_that("sql() evaluates input locally in across()", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_equal(
    lf |>
      summarise(
        across(x, ~ sql(gsub("x", "y", cur_column())))
      ) |>
      remote_query(),
    sql("SELECT y AS \"x\"\nFROM \"df\"")
  )
})

# casts -------------------------------------------------------------------

test_that("casts as expected", {
  con <- dialect_ansi()
  expect_translation(con, as.integer64(x), "CAST(\"x\" AS BIGINT)")
  expect_translation(con, as.logical(x), "CAST(\"x\" AS BOOLEAN)")
  expect_translation(con, as.Date(x), "CAST(\"x\" AS DATE)")
})

test_that("as() translates to CAST", {
  con <- dialect_ansi()
  expect_translation(con, as(x, "TIME"), "CAST(\"x\" AS TIME)")
  expect_translation(
    con,
    as(x, "DECIMAL(10, 2)"),
    "CAST(\"x\" AS DECIMAL(10, 2))"
  )
  expect_snapshot(error = TRUE, translate_sql(as(x, 1L), con = con))
})

# numeric -----------------------------------------------------------------

test_that("hypergeometric functions use manual calculation", {
  con <- dialect_ansi()
  expect_translation(con, cosh(x), "(EXP(\"x\") + EXP(-(\"x\"))) / 2")
  expect_translation(con, sinh(x), "(EXP(\"x\") - EXP(-(\"x\"))) / 2")
  expect_translation(
    con,
    tanh(x),
    "(EXP(2 * (\"x\")) - 1) / (EXP(2 * (\"x\")) + 1)"
  )
  expect_translation(
    con,
    coth(x),
    "(EXP(2 * (\"x\")) + 1) / (EXP(2 * (\"x\")) - 1)"
  )
})

test_that("pmin and max use GREATEST and LEAST", {
  con <- dialect_ansi()
  expect_translation(
    con,
    pmin(x, y, z, na.rm = TRUE),
    "LEAST(\"x\", \"y\", \"z\")"
  )
  expect_translation(con, pmax(x, y, na.rm = TRUE), "GREATEST(\"x\", \"y\")")
})

test_that("round uses integer digits", {
  con <- dialect_ansi()
  expect_translation(con, round(10.1), "ROUND(10.1, 0)")
  expect_translation(con, round(10.1, digits = 1), "ROUND(10.1, 1)")
})

# string functions --------------------------------------------------------

test_that("paste() translated to CONCAT_WS", {
  con <- dialect_ansi()
  expect_translation(con, paste0(x, y), "CONCAT_WS('', \"x\", \"y\")")
  expect_translation(con, paste(x, y), "CONCAT_WS(' ', \"x\", \"y\")")
  expect_translation(
    con,
    paste(x, y, sep = ","),
    "CONCAT_WS(',', \"x\", \"y\")"
  )
})

# stringr -------------------------------------------

test_that("str_length() translates correctly ", {
  con <- dialect_ansi()
  expect_translation(con, str_length(x), "LENGTH(\"x\")")
})

test_that("lower/upper translates correctly ", {
  con <- dialect_ansi()
  expect_translation(con, str_to_upper(x), "UPPER(\"x\")")
  expect_translation(con, str_to_lower(x), "LOWER(\"x\")")
})

test_that("str_trim() translates correctly ", {
  con <- dialect_ansi()
  expect_translation(con, str_trim(x, "both"), "LTRIM(RTRIM(\"x\"))")
  expect_translation(con, str_trim(x, "left"), "LTRIM(\"x\")")
  expect_translation(con, str_trim(x, "right"), "RTRIM(\"x\")")
})

# subsetting --------------------------------------------------------------

test_that("[ treated as if it is logical subsetting", {
  con <- dialect_ansi()
  expect_translation(con, y[x == 0L], "CASE WHEN (\"x\" = 0) THEN (\"y\") END")
})

test_that("sql_substr works as expected", {
  con <- dialect_ansi()

  expect_translation(con, substr(x, 3, 4), "SUBSTR(\"x\", 3, 2)")
  expect_translation(con, substr(x, 3, 3), "SUBSTR(\"x\", 3, 1)")
  expect_translation(con, substr(x, 3, 2), "SUBSTR(\"x\", 3, 0)")
  expect_translation(con, substr(x, 3, 1), "SUBSTR(\"x\", 3, 0)")

  expect_translation(con, substr(x, 0, 1), "SUBSTR(\"x\", 1, 1)")
  expect_translation(con, substr(x, -1, 1), "SUBSTR(\"x\", 1, 1)")

  # Missing arguments
  expect_translation_snapshot(con, substr("test"), error = TRUE)
  expect_translation_snapshot(con, substr("test", 0), error = TRUE)

  # Wrong types
  expect_translation_snapshot(con, substr("test", "x", 1), error = TRUE)
  expect_translation_snapshot(con, substr("test", 1, "x"), error = TRUE)
})

test_that("substring is also translated", {
  con <- dialect_ansi()
  expect_translation(con, substring(x, 3, 4), "SUBSTR(\"x\", 3, 2)")
})

test_that("sql_str_sub works as expected", {
  con <- dialect_ansi()

  expect_translation(con, str_sub(x), "SUBSTR(\"x\", 1)")
  expect_translation(con, str_sub(x, 1), "SUBSTR(\"x\", 1)")
  expect_translation(con, str_sub(x, -1), "SUBSTR(\"x\", LENGTH(\"x\"))")
  expect_translation(con, str_sub(x, 2, 4), "SUBSTR(\"x\", 2, 3)")
  expect_translation(con, str_sub(x, 2, 2), "SUBSTR(\"x\", 2, 1)")
  expect_translation(con, str_sub(x, 2, 0), "SUBSTR(\"x\", 2, 0)")
  expect_translation(
    con,
    str_sub(x, 1, -2),
    "SUBSTR(\"x\", 1, LENGTH(\"x\") - 1)"
  )
  expect_translation(
    con,
    str_sub(x, 3, -3),
    "SUBSTR(\"x\", 3, LENGTH(\"x\") - 4)"
  )
  expect_translation(
    con,
    str_sub(x, -3, 0),
    "SUBSTR(\"x\", LENGTH(\"x\") - 2, 0)"
  )
  expect_translation(
    con,
    str_sub(x, -3, -3),
    "SUBSTR(\"x\", LENGTH(\"x\") - 2, 1)"
  )
})

test_that("sql_str_sub can require length parameter", {
  local_con(dialect_ansi())
  x <- ident("x")
  str_sub <- sql_str_sub("SUBSTR", optional_length = FALSE)

  expect_equal(str_sub(x), sql("SUBSTR(\"x\", 1, LENGTH(\"x\"))"))
  expect_equal(str_sub(x, 1), sql("SUBSTR(\"x\", 1, LENGTH(\"x\"))"))
  expect_equal(str_sub(x, -1), sql("SUBSTR(\"x\", LENGTH(\"x\"), 1)"))
})

test_that("str_sub() returns consistent results", {
  mf <- local_memdb_frame(t = "abcde")

  expect_equal(mf |> transmute(str_sub(t, -3, -1)) |> pull(1), "cde")
  expect_equal(mf |> transmute(str_sub(t, 0, -1)) |> pull(1), "abcde")
  expect_equal(mf |> transmute(str_sub(t, 1, -3)) |> pull(1), "abc")

  expect_equal(mf |> transmute(str_sub(t, -3, 0)) |> pull(1), "")
  expect_equal(mf |> transmute(str_sub(t, 0, 0)) |> pull(1), "")
  expect_equal(mf |> transmute(str_sub(t, 1, 0)) |> pull(1), "")

  expect_equal(mf |> transmute(str_sub(t, -3, 5)) |> pull(1), "cde")
  expect_equal(mf |> transmute(str_sub(t, 0, 1)) |> pull(1), "a")
  expect_equal(mf |> transmute(str_sub(t, 1, 3)) |> pull(1), "abc")
})

test_that("str_detect(), str_starts(), str_ends() support fixed patterns", {
  mf <- local_memdb_frame(
    x = c("%0 start", "end %0", "detect %0 detect", "no", NA)
  )

  # detects fixed pattern
  expect_equal(
    mf |> transmute(str_starts(x, fixed("%0"))) |> pull(1),
    c(1, 0, 0, 0, NA)
  )
  # hack to avoid check complaining about not declared imports
  pattern <- rlang::parse_expr("stringr::fixed('%0')")
  expect_equal(
    mf |> transmute(str_starts(x, !!pattern)) |> pull(1),
    c(1, 0, 0, 0, NA)
  )

  # also works with ends and detect
  expect_equal(
    mf |> transmute(str_ends(x, fixed("%0"))) |> pull(1),
    c(0, 1, 0, 0, NA)
  )
  expect_equal(
    mf |> transmute(str_detect(x, fixed("%0"))) |> pull(1),
    c(1, 1, 1, 0, NA)
  )

  # negate works
  expect_equal(
    mf |> transmute(str_detect(x, fixed("%0"), negate = TRUE)) |> pull(1),
    c(0, 0, 0, 1, NA)
  )

  expect_error(translate_sql(str_detect(x, "a"), con = dialect_ansi()))
})

test_that("basic prefix paste", {
  con <- dialect_ansi()

  expect_translation(con, paste0(x), "CONCAT_WS('', \"x\")")
  expect_translation(con, paste0(x, y), "CONCAT_WS('', \"x\", \"y\")")
  expect_translation(
    con,
    paste0(x, y, sep = " "),
    "CONCAT_WS(' ', \"x\", \"y\")"
  )
})

test_that("basic infix paste", {
  local_con(dialect_ansi())

  paste <- sql_paste_infix("", "&&")
  x <- ident("x")
  y <- ident("y")

  expect_equal(paste(x), sql("CAST(\"x\" AS text)"))
  expect_equal(paste(x, y), sql("\"x\" && \"y\""))
  expect_equal(paste(x, y, sep = " "), sql("\"x\" && ' ' && \"y\""))
})

test_that("simulate_dbi() still works", {
  expect_translation(simulate_dbi(), x + 1, '"x" + 1.0')
})

test_that("base_no_win includes all aggregates and window functions", {
  # All aggregates must be included in window functions
  expect_equal(setdiff(names(base_agg), names(base_win)), character())

  # All window functions all need to be in no_in
  expect_equal(setdiff(names(base_win), names(base_no_win)), character())
})

test_that("can translate both pipes", {
  con <- dialect_ansi()
  expect_translation(
    con,
    x %>% mean() %>% sum(),
    'SUM(AVG("x"))',
    window = FALSE
  )
  expect_translation(con, x |> mean() |> sum(), 'SUM(AVG("x"))', window = FALSE)
})

# mathematics --------------------------------------------------------

test_that("basic arithmetic is correct", {
  con <- dialect_ansi()
  expect_translation(con, 1 + 2, "1.0 + 2.0")
  expect_translation(con, 2 * 4, "2.0 * 4.0")
  expect_translation(con, 5^2, "POWER(5.0, 2.0)")
  expect_translation(con, 100L %% 3L, "100 % 3")

  expect_translation_snapshot(con, 100L %/% 3L, error = TRUE)
})

test_that("small numbers aren't converted to 0", {
  con <- dialect_ansi()
  expect_translation(con, 1e-9, "1e-09")
})

test_that("unary plus works with numbers", {
  con <- dialect_ansi()
  expect_translation(con, +10L, "10")
  expect_translation(con, x == +10, '"x" = 10.0')
  expect_translation(con, x %in% c(+1L, 0L), '"x" IN (1, 0)')
})

test_that("unary plus works for non-numeric expressions", {
  con <- dialect_ansi()
  expect_translation(con, +(1L + 2L), "(1 + 2)")
  expect_translation(con, mean(x, na.rm = TRUE), 'AVG("x")', window = FALSE)
})

test_that("unary minus flips sign of number", {
  con <- dialect_ansi()
  expect_translation(con, -10L, "-10")
  expect_translation(con, -10L + x, '-10 + "x"')
  expect_translation(con, x == -10, '"x" = -10.0')
  expect_translation(con, x %in% c(-1L, 0L), '"x" IN (-1, 0)')
})

test_that("unary minus wraps non-numeric expressions", {
  con <- dialect_ansi()
  expect_translation(con, -(1L + 2L), "-(1 + 2)")
  expect_translation(con, -mean(x, na.rm = TRUE), '-AVG("x")', window = FALSE)
})

test_that("binary minus subtracts", {
  con <- dialect_ansi()
  expect_translation(con, 1L - 10L, "1 - 10")
})

test_that("log base comes first", {
  con <- dialect_ansi()
  expect_translation(con, log(x, 10), 'LOG(10.0, "x")')
})

test_that("log becomes ln", {
  con <- dialect_ansi()
  expect_translation(con, log(x), 'LN("x")')
})

test_that("can translate subsetting", {
  con <- dialect_ansi()
  expect_translation(con, a$b, '"a"."b"')
  expect_translation(con, a[["b"]], '"a"."b"')
  expect_translation(con, f(a)[["b"]], 'f("a")."b"')
  expect_translation(con, a[["b"]][[1]], '"a"."b"[1]')

  expect_translation_snapshot(con, a[[x]], error = TRUE)
  expect_translation_snapshot(con, a[[TRUE]], error = TRUE)
})

test_that("$ doesn't evaluate second argument", {
  y <- list(id = 1)

  expect_snapshot(lazy_frame(x = 1, y = 1) |> filter(x == y$id))
  expect_snapshot(lazy_frame(x = 1) |> filter(x == y$id))
})

test_that("useful error if $ used with inlined value", {
  y <- 1
  expect_snapshot(lazy_frame(x = 1) |> filter(x == y$id), error = TRUE)
})

test_that("if_else() and ifelse() require true/yes and false/no", {
  con <- dialect_ansi()
  expect_translation_snapshot(con, if_else(x), error = TRUE)
  expect_translation_snapshot(con, if_else(x, 1L), error = TRUE)
  expect_translation_snapshot(con, ifelse(x), error = TRUE)
  expect_translation_snapshot(con, ifelse(x, 1L), error = TRUE)
})

# window ------------------------------------------------------------------

test_that("lead and lag translate n to integers", {
  con <- dialect_ansi()
  expect_translation(con, lead(x, 1), 'LEAD("x", 1, NULL) OVER ()')
  expect_translation(con, lag(x, 1), 'LAG("x", 1, NULL) OVER ()')
})

# strings -----------------------------------------------------------------

test_that("can only translate case sensitive str_like", {
  con <- dialect_ansi()
  expect_translation(con, str_like(x, "abc"), '"x" LIKE \'abc\'')

  expect_translation_snapshot(con, str_like(x, "abc", ignore_case = FALSE))
  expect_translation_snapshot(
    con,
    str_like(x, "abc", ignore_case = TRUE),
    error = TRUE
  )
  expect_translation_snapshot(
    con,
    str_ilike(x, "abc"),
    error = TRUE
  )
})

test_that("can translate nzchar", {
  con <- dialect_ansi()
  expect_translation(con, nzchar(y), '(("y" IS NULL) OR "y" != \'\')')
  expect_translation(con, nzchar(y, TRUE), '"y" != \'\'')

  # Uses individual translations from backend
  con <- dialect_access()
  expect_translation(con, nzchar(y), '(ISNULL("y") OR "y" <> \'\')')
  expect_translation(con, nzchar(y, TRUE), '"y" <> \'\'')
})

# aggregates --------------------------------------------------------------

test_that("all and any translated correctly", {
  db <- local_memdb_frame(
    "df",
    g = c(1, 1, 2, 2, 3, 3),
    x = c(0, 0, 0, 1, 1, 1)
  )

  sum_all_g <- db |>
    group_by(g) |>
    summarise(all = all(x == 1, na.rm = TRUE)) |>
    filter(all) |>
    pull(g)
  expect_equal(sum_all_g, 3)

  sum_any_g <- db |>
    group_by(g) |>
    summarise(any = any(x == 1, na.rm = TRUE)) |>
    filter(any) |>
    pull(g)
  expect_equal(sum_any_g, c(2, 3))

  win_all_g <- db |>
    group_by(g) |>
    filter(all(x == 1, na.rm = TRUE)) |>
    pull(g)
  expect_equal(win_all_g, c(3, 3))

  win_any_g <- db |>
    group_by(g) |>
    filter(any(x == 1, na.rm = TRUE)) |>
    pull(g)
  expect_equal(win_any_g, c(2, 2, 3, 3))
})

# binary/bitwise ---------------------------------------------------------------

test_that("bitwise operations", {
  con <- dialect_ansi()
  expect_translation(con, bitwNot(x), '~("x")')
  expect_translation(con, bitwAnd(x, 128L), '"x" & 128')
  expect_translation(con, bitwOr(x, 128L), '"x" | 128')
  expect_translation(con, bitwXor(x, 128L), '"x" ^ 128')
  expect_translation(con, bitwShiftL(x, 2L), '"x" << 2')
  expect_translation(con, bitwShiftR(x, 2L), '"x" >> 2')
})

test_that("default raw escapes translated correctly", {
  mf <- lazy_frame(x = "abc", con = dialect_sqlite())

  a <- blob::as_blob("abc")
  b <- blob::as_blob(as.raw(c(0x01, 0x02)))
  L <- c(a, b)

  expect_snapshot(mf |> filter(x == a))
  expect_snapshot(mf |> filter(x %in% L))

  qry <- mf |> filter(x %in% !!L)
  expect_snapshot(qry)
})

test_that("case_when converted to CASE WHEN", {
  con <- dialect_ansi()
  expect_snapshot(translate_sql(case_when(x > 1L ~ "a"), con = con))
})

test_that("even inside mutate", {
  out <- lazy_frame(x = 1:5) |>
    mutate(y = case_when(x > 1L ~ "a")) |>
    sql_build()
  expect_snapshot(out$select[[2]])
})

test_that("case_when translates correctly to ELSE when TRUE ~ is used 2", {
  con <- dialect_ansi()
  expect_snapshot(
    translate_sql(
      case_when(
        x == 1L ~ "yes",
        x == 0L ~ "no",
        TRUE ~ "undefined"
      ),
      con = con
    )
  )
})

test_that("case_when uses the .default arg", {
  con <- dialect_ansi()
  expect_snapshot(
    translate_sql(
      case_when(
        x == 1L ~ "yes",
        x == 0L ~ "no",
        .default = "undefined"
      ),
      con = con
    )
  )

  expect_snapshot(
    translate_sql(
      case_when(
        x == 1L ~ "yes",
        x == 0L ~ "no",
        .default = x + 1
      ),
      con = con
    )
  )

  # TRUE ~ has precedence over .default
  expect_snapshot(
    translate_sql(
      case_when(
        x == 1L ~ "yes",
        x == 0L ~ "no",
        TRUE ~ "true",
        .default = "undefined"
      ),
      con = con
    )
  )
})

test_that("case_when does not support .ptype and .size", {
  con <- dialect_ansi()
  expect_snapshot({
    (expect_error(translate_sql(
      case_when(
        x == 1L ~ "yes",
        .ptype = character()
      ),
      con = con
    )))
    (expect_error(translate_sql(
      case_when(x == 1L ~ "yes", .size = 1),
      con = con
    )))
  })
})

test_that("long case_when is on multiple lines", {
  con <- dialect_ansi()
  expect_snapshot(
    translate_sql(
      case_when(
        x == 1L ~ "this is long",
        x == 0L ~ "so it should",
        TRUE ~ "be wrapped"
      ),
      con = con
    )
  )
})

test_that("all forms of if translated to case statement", {
  con <- dialect_ansi()

  expect_translation(
    con,
    if (x) 1L else 2L,
    "CASE WHEN \"x\" THEN 1 WHEN NOT \"x\" THEN 2 END"
  )
  expect_translation(
    con,
    ifelse(x, 1L, 2L),
    "CASE WHEN \"x\" THEN 1 WHEN NOT \"x\" THEN 2 END"
  )
  expect_translation(
    con,
    if_else(x, 1L, 2L),
    "CASE WHEN \"x\" THEN 1 WHEN NOT \"x\" THEN 2 END"
  )
})

test_that("if_else can be simplified", {
  con <- dialect_ansi()
  expect_translation(
    con,
    if_else(x, 1L, 2L, 2L),
    "CASE WHEN \"x\" THEN 1 ELSE 2 END"
  )
})

test_that("if translation adds parens", {
  con <- dialect_ansi()
  expect_translation(con, if (x) y, "CASE WHEN \"x\" THEN \"y\" END")
  expect_translation(
    con,
    if (x > 1L) y + 1L,
    "CASE WHEN (\"x\" > 1) THEN (\"y\" + 1) END"
  )
  expect_translation(
    con,
    if (x) y else z,
    "CASE WHEN \"x\" THEN \"y\" WHEN NOT \"x\" THEN \"z\" END"
  )
  expect_translation(
    con,
    if (x > 1L) y + 1L else z + 1L,
    "CASE WHEN (\"x\" > 1) THEN (\"y\" + 1) WHEN NOT (\"x\" > 1) THEN (\"z\" + 1) END"
  )
})

test_that("if_else can translate missing", {
  con <- dialect_ansi()
  expect_translation(
    con,
    if_else(x, 1L, 2L, 3L),
    "CASE WHEN \"x\" THEN 1 WHEN NOT \"x\" THEN 2 ELSE 3 END"
  )
})

test_that("if and ifelse use correctly named arguments", {
  con <- dialect_ansi()
  exp <- translate_sql(if (x) 1 else 2, con = con)

  expect_equal(translate_sql(ifelse(test = x, yes = 1, no = 2), con = con), exp)
  expect_equal(
    translate_sql(if_else(condition = x, true = 1, false = 2), con = con),
    exp
  )

  expect_translation(
    con,
    if_else(
      condition = x,
      true = 1,
      false = 2,
      missing = 3
    ),
    "CASE WHEN \"x\" THEN 1.0 WHEN NOT \"x\" THEN 2.0 ELSE 3.0 END"
  )
})

test_that("switch translated to CASE WHEN", {
  con <- dialect_ansi()
  expect_translation(
    con,
    switch(x, a = 1L),
    "CASE \"x\" WHEN ('a') THEN (1) END"
  )
  expect_translation(
    con,
    switch(x, a = 1L, 2L),
    "CASE \"x\" WHEN ('a') THEN (1) ELSE (2) END"
  )
})

test_that("is.na and is.null are equivalent", {
  con <- dialect_ansi()
  # Needs to be wrapped in parens to ensure correct precedence
  expect_translation(con, is.na(x + y), "((\"x\" + \"y\") IS NULL)")
  expect_translation(con, is.null(x + y), "((\"x\" + \"y\") IS NULL)")

  expect_translation(con, x + is.na(x), "\"x\" + (\"x\" IS NULL)")
  expect_translation(con, !is.na(x), "NOT((\"x\" IS NULL))")
})

test_that("magrittr pipe is translated in conditionals", {
  con <- dialect_ansi()
  expect_translation(
    con,
    x |> ifelse(1L, 2L),
    "CASE WHEN \"x\" THEN 1 WHEN NOT \"x\" THEN 2 END"
  )
})

test_that("conditionals check arguments", {
  con <- dialect_ansi()
  expect_snapshot(error = TRUE, translate_sql(case_when(), con = con))

  expect_snapshot(error = TRUE, translate_sql(switch(x, 1L, 2L), con = con))
})


# case_match --------------------------------------------------------------

test_that("LHS can handle different types", {
  con <- dialect_ansi()
  expect_translation(
    con,
    case_match(z, 1L ~ "a"),
    "CASE WHEN (\"z\" IN (1)) THEN 'a' END"
  )

  expect_translation(
    con,
    case_match(z, "x" ~ "a"),
    "CASE WHEN (\"z\" IN ('x')) THEN 'a' END"
  )

  expect_translation(
    con,
    case_match(z, y ~ "a"),
    "CASE WHEN (\"z\" IN (\"y\")) THEN 'a' END"
  )

  expect_translation(
    con,
    case_match(z, as.character(y) ~ "a"),
    "CASE WHEN (\"z\" IN (CAST(\"y\" AS TEXT))) THEN 'a' END"
  )
})

test_that("LHS can match multiple values", {
  con <- dialect_ansi()
  expect_translation(
    con,
    case_match(z, 1:2 ~ "a"),
    "CASE WHEN (\"z\" IN ((1, 2))) THEN 'a' END"
  )

  expect_translation(
    con,
    case_match(z, c(1L, 3L) ~ "a"),
    "CASE WHEN (\"z\" IN (1, 3)) THEN 'a' END"
  )

  expect_translation(
    con,
    case_match(z, c("x", "y") ~ "a"),
    "CASE WHEN (\"z\" IN ('x', 'y')) THEN 'a' END"
  )

  expect_translation(
    con,
    case_match(z, c(1L, y) ~ "a"),
    "CASE WHEN (\"z\" IN (1, \"y\")) THEN 'a' END"
  )
})

test_that("LHS can match NA", {
  con <- dialect_ansi()
  expect_translation(
    con,
    case_match(z, NA ~ "a"),
    "CASE WHEN (\"z\" IS NULL) THEN 'a' END"
  )

  expect_translation(
    con,
    case_match(z, c(1L, NA) ~ "a"),
    "CASE WHEN (\"z\" IN (1) OR \"z\" IS NULL) THEN 'a' END"
  )
})

test_that("LHS can handle bang bang", {
  con <- dialect_ansi()
  expect_snapshot({
    translate_sql(case_match(x, !!1L ~ "x"), con = con)
    translate_sql(case_match(x, !!c(1L, 2L) ~ "x"), con = con)
    translate_sql(case_match(x, !!c(NA, 1L) ~ "x"), con = con)
  })
})

test_that("`NULL` values in `...` are dropped", {
  con <- dialect_ansi()
  expect_translation(
    con,
    case_match(x, 1L ~ "a", NULL, 2L ~ "b", NULL),
    "CASE WHEN (\"x\" IN (1)) THEN 'a' WHEN (\"x\" IN (2)) THEN 'b' END"
  )
})

test_that("requires at least one condition", {
  con <- dialect_ansi()
  expect_snapshot(error = TRUE, {
    translate_sql(case_match(x), con = con)
  })
  expect_snapshot(error = TRUE, {
    translate_sql(case_match(x, NULL), con = con)
  })
})

test_that("passes through `.default` correctly", {
  con <- dialect_ansi()
  expect_translation(
    con,
    case_match(z, 3L ~ 1L, .default = 2L),
    "CASE WHEN (\"z\" IN (3)) THEN 1 ELSE 2 END"
  )
})

test_that("can handle multiple cases", {
  con <- dialect_ansi()
  expect_translation(
    con,
    case_match(z, 1L ~ "a", 2L ~ "b"),
    "CASE WHEN (\"z\" IN (1)) THEN 'a' WHEN (\"z\" IN (2)) THEN 'b' END"
  )

  # also with .default
  expect_translation(
    con,
    case_match(z, 1L ~ "a", 2L ~ "b", .default = "default"),
    "CASE WHEN (\"z\" IN (1)) THEN 'a' WHEN (\"z\" IN (2)) THEN 'b' ELSE 'default' END"
  )
})

test_that("`.ptype` not supported", {
  con <- dialect_ansi()
  expect_snapshot({
    (expect_error(translate_sql(
      case_match(x, 1 ~ 1, .ptype = integer()),
      con = con
    )))
  })
})

test_that(".x must be a symbol", {
  con <- dialect_ansi()
  expect_snapshot({
    (expect_error(translate_sql(case_match(1, 1 ~ 1), con = con)))
  })
})

test_that("two arranges equivalent to one", {
  mf <- local_memdb_frame(x = c(2, 2, 1), y = c(1, -1, 1))

  mf1 <- mf |> arrange(x, y)
  mf2 <- mf |> arrange(y) |> arrange(x)

  expect_equal(mf1 |> collect(), mf2 |> collect())
})

test_that("correctly inlines across all verbs", {
  lf <- lazy_frame(x = 1, y = 2)

  # single table verbs
  expect_selects(lf |> arrange(x) |> arrange(y), 1)
  expect_selects(lf |> distinct() |> arrange(y), 1)
  expect_selects(lf |> filter(x) |> arrange(y), 1)
  expect_selects(lf |> head(1) |> arrange(y), 2) # TODO: 1?
  expect_selects(lf |> mutate(y = x + 1) |> arrange(y), 1)
  expect_selects(lf |> select(y = x) |> arrange(y), 1)
  expect_selects(lf |> summarise(y = mean(x)) |> arrange(y), 1)

  # two table verbs
  lf2 <- lazy_frame(x = 1)
  expect_selects(lf |> left_join(lf2, by = "x") |> arrange(y), 2)
  expect_selects(lf |> right_join(lf2, by = "x") |> arrange(y), 2)
  expect_selects(lf |> semi_join(lf2, by = "x") |> arrange(y), 3)
  expect_selects(lf |> union(lf2) |> arrange(y), 3)
})

# sql_render --------------------------------------------------------------

test_that("quoting for rendering ordered grouped table", {
  db <- local_memdb_frame("test-verb-arrange", x = 1, y = 2)
  out <- db |> group_by(x) |> arrange(y) |> ungroup()
  expect_snapshot(sql_render(out))
  expect_equal(collect(out), tibble(x = 1, y = 2))
})

test_that("arrange renders correctly (#373)", {
  expect_snapshot({
    "# arrange renders correctly"
    lf <- lazy_frame(a = 1:3, b = 3:1)

    "basic"
    lf |> arrange(a)

    "double arrange"
    lf |> arrange(a) |> arrange(b)

    "remove ordered by"
    lf |> arrange(a) |> select(-a)
    lf |> arrange(a) |> select(-a) |> arrange(b)

    "un-arrange"
    lf |> arrange(a) |> arrange()
    lf |> arrange(a) |> select(-a) |> arrange()

    "use order"
    lf |> arrange(a) |> select(-a) |> mutate(c = lag(b))
  })
})

test_that("arrange renders correctly for single-table verbs (#373)", {
  expect_snapshot({
    lf <- lazy_frame(a = 1:3, b = 3:1)

    "head"
    lf |> head(1) |> arrange(a)
    lf |> arrange(a) |> head(1)
    lf |> arrange(a) |> head(1) |> arrange(b)

    "mutate"
    lf |> mutate(a = b) |> arrange(a)

    "complex mutate"
    lf |> arrange(a) |> mutate(a = b) |> arrange(a)
    lf |> arrange(a) |> mutate(a = 1) |> arrange(b)
    lf |> mutate(a = -a) |> arrange(a) |> mutate(a = -a)
  })
})

test_that("can combine arrange with dual table verbs", {
  expect_snapshot({
    lf <- lazy_frame(a = 1:3, b = 3:1)
    rf <- lazy_frame(a = 1:3, c = 4:6)

    "warn if arrange before join"
    lf |> arrange(a) |> left_join(rf)
    lf |> arrange(a) |> semi_join(rf)
    lf |> arrange(a) |> union(rf)

    "can arrange after join"
    lf |> left_join(rf) |> arrange(a)
    lf |> semi_join(rf) |> arrange(a)
    lf |> union(rf) |> arrange(a)
  })
})

test_that("only add step if necessary", {
  lf <- lazy_frame(x = 1:3, y = 1:3)
  expect_equal(lf |> arrange(), lf)
})

# sql_build ---------------------------------------------------------------

test_that("arrange generates order_by", {
  out <- lazy_frame(x = 1, y = 1) |>
    arrange(x) |>
    sql_build()

  expect_equal(out$order_by, sql('"x"'))
})

test_that("arrange converts desc", {
  out <- lazy_frame(x = 1, y = 1) |>
    arrange(desc(x)) |>
    sql_build()

  expect_equal(out$order_by, sql('"x" DESC'))
})

test_that("grouped arrange doesn't order by groups", {
  out <- lazy_frame(x = 1, y = 1) |>
    group_by(x) |>
    arrange(y) |>
    sql_build()

  expect_equal(out$order_by, sql('"y"'))
})

test_that("grouped arrange order by groups when .by_group  is set to TRUE", {
  lf <- lazy_frame(x = 1, y = 1, con = dialect_ansi())
  out <- lf |>
    group_by(x) |>
    arrange(y, .by_group = TRUE) |>
    sql_build()
  expect_equal(out$order_by, sql(c('"x"', '"y"')))
})

# ops ---------------------------------------------------------------------

test_that("arranges captures DESC", {
  out <- lazy_frame(x = 1:3, y = 3:1) |> arrange(desc(x))

  sort <- lapply(op_sort(out), get_expr)
  expect_equal(sort, list(quote(desc(x))))
})

test_that("multiple arranges combine", {
  out <- lazy_frame(x = 1:3, y = 3:1) |> arrange(x) |> arrange(y)

  sort <- lapply(op_sort(out), get_expr)
  expect_equal(sort, list(quote(y), quote(x)))
})

# unwrap_order_expr/swap_order_direction ---------------------------------------

test_that("unwrap_order_expr handles NULL", {
  expect_equal(unwrap_order_expr(NULL, "test"), NULL)
})

test_that("unwrap_order_expr wraps single expression into a list", {
  expect_equal(unwrap_order_expr(expr(x), "test"), list(expr(x)))
  expect_equal(unwrap_order_expr(expr(desc(x)), "test"), list(expr(desc(x))))
})

test_that("unwrap_order_expr unwraps tibble/data.frame", {
  expect_equal(
    unwrap_order_expr(expr(tibble(x, y)), "test"),
    list(expr(x), expr(y))
  )
  expect_equal(
    unwrap_order_expr(expr(data.frame(x, desc(y))), "test"),
    list(expr(x), expr(desc(y)))
  )
})

test_that("unwrap_order_expr informatively errors on c()", {
  expect_snapshot(unwrap_order_expr(expr(c(x, y)), "slice_min"), error = TRUE)
})

test_that("swap_order_direction swaps direction", {
  expect_equal(swap_order_direction(expr(x)), expr(desc(x)))
  expect_equal(swap_order_direction(expr(desc(x))), expr(x))
  expect_equal(swap_order_direction(expr(a + b)), expr(desc(a + b)))
})

#' Impala backend
#'
#' @description
#' This backend supports Apache Impala, typically accessed via odbc. Use
#' `dialect_impala()` with `lazy_frame()` to see simulated SQL without
#' connecting to a live database.
#'
#' Key differences for this backend are a scattering of custom translations
#' provided by users, mostly focussed on bitwise operations.
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-impala
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_impala())
#' lf |> transmute(X = bitwNot(bitwOr(b, c)))
NULL

#' @export
#' @rdname backend-impala
dialect_impala <- function() {
  new_sql_dialect(
    "impala",
    quote_identifier = function(x) sql_quote(x, '"'),
    has_window_clause = TRUE
  )
}

#' @export
#' @rdname backend-impala
simulate_impala <- function() simulate_dbi("Impala")

#' @export
sql_dialect.Impala <- function(con) {
  dialect_impala()
}

#' @export
dbplyr_edition.Impala <- function(con) {
  2L
}

#' @export
sql_translation.sql_dialect_impala <- function(con) {
  sql_variant(
    scalar = sql_translator(
      .parent = base_odbc_scalar,
      bitwNot = sql_prefix("BITNOT", 1),
      bitwAnd = sql_prefix("BITAND", 2),
      bitwOr = sql_prefix("BITOR", 2),
      bitwXor = sql_prefix("BITXOR", 2),
      bitwShiftL = sql_prefix("SHIFTLEFT", 2),
      bitwShiftR = sql_prefix("SHIFTRIGHT", 2),

      as.Date = sql_cast("VARCHAR(10)"),
      ceiling = sql_prefix("CEIL")
    ),
    base_odbc_agg,
    base_odbc_win
  )
}

#' @export
sql_table_analyze.sql_dialect_impala <- function(con, table, ...) {
  # Using COMPUTE STATS instead of ANALYZE as recommended in this article
  # https://www.cloudera.com/documentation/enterprise/5-9-x/topics/impala_compute_stats.html
  sql_glue2(con, "COMPUTE STATS {.tbl table}")
}

test_that("simulate_impala() still works", {
  expect_translation(simulate_impala(), x + 1, '"x" + 1.0')
})

test_that("custom scalar functions translated correctly", {
  con <- dialect_impala()

  expect_translation(con, as.Date(x), 'CAST("x" AS VARCHAR(10))')
  expect_translation(con, ceiling(x), 'CEIL("x")')
})

test_that("custom bitwise operations translated correctly", {
  con <- dialect_impala()

  expect_translation(con, bitwNot(x), 'BITNOT("x")')
  expect_translation(con, bitwAnd(x, 128L), 'BITAND("x", 128)')
  expect_translation(con, bitwOr(x, 128L), 'BITOR("x", 128)')
  expect_translation(con, bitwXor(x, 128L), 'BITXOR("x", 128)')
  expect_translation(con, bitwShiftL(x, 2L), 'SHIFTLEFT("x", 2)')
  expect_translation(con, bitwShiftR(x, 2L), 'SHIFTRIGHT("x", 2)')
})

test_that("generates custom sql", {
  con <- dialect_impala()

  expect_snapshot(sql_table_analyze(con, in_schema("schema", "tbl")))
})

new_table_ident <- function(...,
                            table = NA_character_,
                            schema = NA_character_,
                            catalog = NA_character_,
                            quoted = FALSE,
                            error_call = caller_env()) {
  check_dots_empty(call = error_call)

  check_character(table, call = error_call)
  check_character(schema, call = error_call)
  check_character(catalog, call = error_call)
  check_logical(quoted, call = error_call)

  data <- vctrs::vec_recycle_common(
    table = table,
    schema = schema,
    catalog = catalog,
    quoted = quoted,
    alias = NA_character_,
    .call = error_call
  )

  purrr::pmap(
    data,
    function(table, schema, catalog, quoted, alias) {
      check_db_table_args(
        table = table,
        schema = schema,
        catalog = catalog,
        quoted = quoted,
        call = error_call
      )
    }
  )

  vctrs::new_rcrd(data, class = "dbplyr_table_ident")
}

check_db_table_args <- function(table,
                                schema,
                                catalog,
                                quoted,
                                call = caller_env()) {
  if (!is.na(catalog) && is.na(schema)) {
    cli_abort("Must supply {.arg schema} when {.arg catalog} is supplied.", call = call)
  }
  if (!is.na(schema) && is.na(table)) {
    cli_abort("Must supply {.arg table} when {.arg schema} is supplied.", call = call)
  }

  if (quoted && !is.na(schema)) {
    cli_abort("Can't supply a schema when {.arg table} is quoted.", call = call)
  }
}

is_table_ident <- function(x) {
  inherits(x, "dbplyr_table_ident")
}

as_table_ident <- function(x, ..., error_call = caller_env()) {
  check_dots_empty()
  UseMethod("as_table_ident")
}

#' @export
as_table_ident.dbplyr_table_ident <- function(x, ..., error_call = caller_env()) {
  x
}

#' @export
as_table_ident.character <- function(x, ..., error_call = caller_env()) {
  new_table_ident(table = unclass(x))
}

#' @export
as_table_ident.ident <- function(x, ..., error_call = caller_env()) {
  new_table_ident(table = unclass(x))
}

#' @export
as_table_ident.ident_q <- function(x, ..., error_call = caller_env()) {
  inform_unwanted_table_ident("ident_q")
  new_table_ident(table = unclass(x), quoted = TRUE)
}

#' @export
as_table_ident.sql <- function(x, ..., error_call = caller_env()) {
  inform_unwanted_table_ident("sql")
  new_table_ident(table = unclass(x), quoted = TRUE)
}

inform_unwanted_table_ident <- function(f) {
  cli::cli_inform(
    c(
      i = "Using {.fn {f}} for a table identifier is intended as fallback in case of bugs.",
      i = "If you need it to work around a bug please open an issue {.url https://github.com/tidyverse/dbplyr/issues}."
    ),
    .frequency = "regularly",
    .frequency_id = paste0("dbplyr_", f)
  )
}

#' @export
as_table_ident.dbplyr_schema <- function(x, ..., error_call = caller_env()) {
  new_table_ident(
    table = unclass(x$table),
    schema = unclass(x$schema)
  )
}

#' @export
as_table_ident.dbplyr_catalog <- function(x, ..., error_call = caller_env()) {
  new_table_ident(
    table = unclass(x$table),
    schema = unclass(x$schema),
    catalog = unclass(x$catalog)
  )
}

#' @export
as_table_ident.Id <- function(x, ..., error_call = caller_env()) {
  id <- x@name
  nms <- names(id)
  known_names <- c("catalog", "schema", "table")
  unknown_names <- setdiff(nms, known_names)
  if (!is_empty(unknown_names)) {
    cli_abort(c(
      "{.arg table} is an <Id> with unknown names {.val {unknown_names}}.",
      i = "Only {.val {known_names}} are supported."
    ), call = error_call)
  }

  new_table_ident(
    table = if ("table" %in% nms) id[["table"]] else NA_character_,
    schema = if ("schema" %in% nms) id[["schema"]] else NA_character_,
    catalog = if ("catalog" %in% nms) id[["catalog"]] else NA_character_
  )
}

as_table_ident_or_sql <- function(x, ..., error_call = caller_env()) {
  if (is.sql(x)) {
    return(x)
  }

  as_table_ident(x, error_call = error_call)
}

#' @export
format.dbplyr_table_ident <- function(x, ..., sep = ".", con = NULL) {
  quoted <- vctrs::field(x, "quoted")
  con <- con %||% simulate_dbi()
  x <- quote_table_ident(x, con)

  collapse_table_ident(x, sep = sep)
}

is_table_ident <- function(x) {
  inherits(x, "dbplyr_table_ident")
}

#' @export
escape.dbplyr_table_ident <- function(x, parens = FALSE, collapse = ", ", con = NULL) {
  # this ignores `parens` and `collapse`; at least for now
  x_quoted <- format(x, con = con)

  canonical_alias <- purrr::map_chr(x, ~ table_ident_name(.x) %||% "")
  alias <- table_ident_alias(x) %||% vctrs::vec_rep("", vctrs::vec_size(x))

  if (db_supports_table_alias_with_as(con)) {
    as_sql <- style_kw(" AS ")
  } else {
    as_sql <- " "
  }

  alias_esc <- sql_escape_ident(con, alias)
  out <- ifelse(
    alias == "" | alias == canonical_alias,
    x_quoted,
    paste0(x_quoted, as_sql, alias_esc)
  )

  sql_vector(out, parens, collapse, con = con)
}

quote_table_ident <- function(x, con) {
  quoted <- vctrs::field(x, "quoted")

  for (field in c("table", "schema", "catalog")) {
    xf <- vctrs::field(x, field)
    idx <- !is.na(xf) & !quoted
    xf[idx] <- sql_escape_ident(con, xf[idx])
    x <- vctrs::`field<-`(x, field, xf)
  }

  x
}

collapse_table_ident <- function(x, sep = ".") {
  table <- vctrs::field(x, "table")
  schema <- vctrs::field(x, "schema")
  catalog <- vctrs::field(x, "catalog")

  out <- table
  out[!is.na(schema)] <- paste0(schema, sep, table)[!is.na(schema)]
  out[!is.na(catalog)] <- paste0(catalog, sep, schema, sep, table)[!is.na(catalog)]

  out
}

table_ident_name <- function(x) {
  table <- vctrs::field(x, "table")
  quoted <- vctrs::field(x, "quoted")
  if (quoted) {
    NULL
  } else {
    table
  }
}

table_ident_alias <- function(x) {
  alias <- vctrs::field(x, "alias")
  if (all(is.na(alias))) {
    return(NULL)
  }

  alias
}

set_table_ident_alias <- function(x, alias) {
  alias <- alias %||% vctrs::vec_rep(NA_character_, times = vctrs::vec_size(x))
  vctrs::`field<-`(x, "alias", alias)
}

as_from <- function(x, ..., arg = caller_arg(x), error_call = caller_env()) {
  if (is.sql(x)) {
    return(x)
  }

  as_table_ident(x, error_call = error_call)
}

table_ident_to_id <- function(x) {
  vctrs::vec_check_size(x, 1)

  quoted <- vctrs::field(x, "quoted")
  if (quoted) {
    out <- DBI::SQL(sql)
    return(out)
  }

  catalog <- vctrs::field(x, "catalog")
  schema <- vctrs::field(x, "schema")
  table <- vctrs::field(x, "table")

  if (!is.na(catalog)) {
    DBI::Id(catalog = catalog, schema = schema, table = table)
  } else if (!is.na(schema)) {
    DBI::Id(schema = schema, table = table)
  } else {
    DBI::Id(table = table)
  }
}

#' Backend: Hive
#'
#' @description
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology. Key differences for this backend
#' are a scattering of custom translations provided by users.
#'
#' Use `simulate_hive()` with `lazy_frame()` to see simulated SQL without
#' converting to live access database.
#'
#' @name backend-hive
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, d = 2, c = "z", con = simulate_hive())
#' lf %>% transmute(x = cot(b))
#' lf %>% transmute(x = bitwShiftL(c, 1L))
#' lf %>% transmute(x = str_replace_all(c, "a", "b"))
#'
#' lf %>% summarise(x = median(d, na.rm = TRUE))
#' lf %>% summarise(x = var(c, na.rm = TRUE))
NULL

#' @export
#' @rdname backend-hive
simulate_hive <- function() simulate_dbi("Hive")

#' @export
dbplyr_edition.Hive <- function(con) {
  2L
}

#' @export
sql_translation.Hive <- function(con) {
  sql_variant(
    sql_translator(.parent = base_odbc_scalar,
      bitwShiftL    = sql_prefix("SHIFTLEFT", 2),
      bitwShiftR    = sql_prefix("SHIFTRIGHT", 2),

      cot = function(x){
        sql_expr(1 / tan(!!x))
      },

      str_replace_all = function(string, pattern, replacement) {
        sql_expr(regexp_replace(!!string, !!pattern, !!replacement))
      }
    ),
    sql_translator(.parent = base_odbc_agg,
      var = sql_aggregate("VARIANCE", "var"),
      quantile = sql_quantile("PERCENTILE"),
      median = sql_median("PERCENTILE")
    ),
    sql_translator(.parent = base_odbc_win,
      var = win_aggregate("VARIANCE"),
      quantile = sql_quantile("PERCENTILE", window = TRUE),
      median = sql_median("PERCENTILE", window = TRUE),
      first = function(x, order_by = NULL, na_rm = FALSE) {
        sql_nth(
          x = x,
          n = 1L,
          order_by = order_by,
          na_rm = na_rm,
          ignore_nulls = "bool"
        )
      },
      last = function(x, order_by = NULL, na_rm = FALSE) {
        sql_nth(
          x = x,
          n = Inf,
          order_by = order_by,
          na_rm = na_rm,
          ignore_nulls = "bool"
        )
      },
      nth = function(x, n, order_by = NULL, na_rm = FALSE) {
        sql_nth(
          x = x,
          n = n,
          order_by = order_by,
          na_rm = na_rm,
          ignore_nulls = "bool"
        )
      },
    )
  )
}

#' @export
sql_table_analyze.Hive <- function(con, table, ...) {
  # https://cwiki.apache.org/confluence/display/Hive/StatsDev
  glue_sql2(con, "ANALYZE TABLE {.tbl table} COMPUTE STATISTICS")
}

#' @export
sql_query_set_op.Hive <- function(con, x, y, method, ..., all = FALSE, lvl = 0) {
  # parentheses are not allowed
  method <- paste0(method, if (all) " ALL")
  glue_sql2(
    con,
    "\n", # dummy line to protect indent
    "{x}\n",
    lvl_indent(lvl), "{.kw method}\n",
    y
  )
}

#' @export
supports_window_clause.Hive <- function(con) {
  TRUE
}

utils::globalVariables("regexp_replace")

test_that("simulate_hive() still works", {
  expect_translation(simulate_hive(), x + 1, '"x" + 1.0')
})

test_that("custom scalar & string functions translated correctly", {
  con <- dialect_hive()

  expect_translation(con, bitwShiftL(x, 2L), 'SHIFTLEFT("x", 2)')
  expect_translation(con, bitwShiftR(x, 2L), 'SHIFTRIGHT("x", 2)')
  expect_translation(con, cot(x), '1.0 / TAN("x")')
  expect_translation(
    con,
    str_replace_all(x, "old", "new"),
    'REGEXP_REPLACE("x", \'old\', \'new\')'
  )
  expect_translation(
    con,
    median(x, na.rm = TRUE),
    'PERCENTILE("x", 0.5) OVER ()'
  )
})

test_that("generates custom sql", {
  con <- dialect_hive()
  expect_snapshot(sql_table_analyze(con, in_schema("schema", "tbl")))

  expect_equal(
    translate_sql(last(x, na_rm = TRUE), vars_order = "a", con = con),
    sql(
      'LAST_VALUE("x", TRUE) OVER (ORDER BY "a" ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING)'
    )
  )

  lf <- lazy_frame(tibble(x = 1), con = con)
  expect_snapshot(union_all(lf, lf))
})

#' Collect results into a local data frame
#'
#' `collect()` executes the query and retrieves the results into a local tibble.
#' This brings all the data from the database into R's memory, which is useful
#' once you've done as much as possible in the database, and now need to use
#' R functions.
#'
#' @param x A lazy data frame backed by a database query.
#' @param ... Ignored.
#' @param n Number of rows to fetch. Defaults to `Inf`, meaning all rows.
#' @param warn_incomplete Warn if `n` is less than the number of result rows?
#' @param cte `r lifecycle::badge("deprecated")`
#'   Use the `sql_options` argument instead.
#' @param sql_options SQL rendering options generated by [sql_options()].
#' @importFrom dplyr collect
#' @export
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(a = c(3, 4, 1, 2), b = c(5, 1, 2, NA))
#' db |> filter(a <= 2) |> collect()
collect.tbl_sql <- function(
  x,
  ...,
  n = Inf,
  warn_incomplete = TRUE,
  sql_options = NULL,
  cte = deprecated()
) {
  rlang::check_exclusive(sql_options, cte, .require = FALSE)
  if (lifecycle::is_present(cte)) {
    lifecycle::deprecate_warn(
      when = "2.4.0",
      what = "collect(cte)",
      with = I("collect(sql_options = sql_options(cte = TRUE))")
    )
    sql_options <- dbplyr::sql_options(cte = cte)
  }

  if (identical(n, Inf)) {
    n <- -1
  } else {
    # Gives the query planner information that it might be able to take
    # advantage of
    x <- head(x, n)
  }

  sql <- db_sql_render(x$con, x, sql_options = sql_options)
  withCallingHandlers(
    out <- db_collect(
      x$con,
      sql,
      n = n,
      warn_incomplete = warn_incomplete,
      ...
    ),
    error = function(cnd) {
      cli_abort("Failed to collect lazy table.", parent = cnd)
    }
  )
  dplyr::grouped_df(out, intersect(op_grps(x), names(out)))
}

sql_select_clauses <- function(
  select,
  from,
  where,
  group_by,
  having,
  window,
  order_by,
  limit = NULL,
  lvl = 0
) {
  out <- list(select, from, where, group_by, having, window, order_by, limit)
  sql_format_clauses(out, lvl)
}

sql_clause <- function(kw, parts, sep = ",", parens = FALSE, lvl = 0) {
  if (length(parts) == 0) {
    return()
  }

  check_string(kw)
  check_sql(parts, allow_names = FALSE)
  check_string(sep, allow_null = TRUE)
  check_bool(parens)

  clause <- list(
    kw = kw,
    parts = parts,
    sep = sep,
    parens = parens,
    lvl = lvl
  )

  class(clause) <- "sql_clause"
  clause
}

sql_clause_select <- function(select, distinct = FALSE, top = NULL, lvl = 0) {
  check_bool(distinct)
  check_number_whole(top, min = 0, allow_null = TRUE)

  if (is_empty(select)) {
    cli_abort("Query contains no columns")
  }

  clause <- paste0(
    "SELECT",
    if (distinct) " DISTINCT",
    if (!is.null(top)) paste0(" TOP ", as.integer(top))
  )
  sql_clause(clause, select)
}

sql_clause_from <- function(from, lvl = 0) {
  sql_clause("FROM", from, lvl = lvl)
}

sql_clause_where <- function(where, lvl = 0) {
  where <- sql(paste0("(", where, ")", recycle0 = TRUE))
  sql_clause("WHERE", where, sep = " AND", lvl = lvl)
}

sql_clause_group_by <- function(group_by, lvl = 0) {
  sql_clause("GROUP BY", group_by)
}

sql_clause_having <- function(having, lvl = 0, call = caller_env()) {
  having <- sql(paste0("(", having, ")", recycle0 = TRUE))
  sql_clause("HAVING", having, sep = " AND")
}

sql_clause_window <- function(window) {
  sql_clause("WINDOW", window)
}

sql_clause_order_by <- function(
  order_by,
  subquery = FALSE,
  limit = NULL,
  lvl = 0
) {
  if (subquery && length(order_by) > 0 && is.null(limit)) {
    warn_drop_order_by()
    NULL
  } else {
    sql_clause("ORDER BY", order_by)
  }
}

sql_clause_limit <- function(con, limit, lvl = 0) {
  if (!is.null(limit) && !identical(limit, Inf)) {
    sql_clause("LIMIT", sql(format(limit, scientific = FALSE)))
  }
}

sql_clause_update <- function(table) {
  sql_clause("UPDATE", table)
}

sql_clause_set <- function(lhs, rhs) {
  update_clauses <- sql(paste0(lhs, " = ", rhs, recycle0 = TRUE))
  sql_clause("SET", update_clauses)
}

sql_clause_insert <- function(cols, into = NULL, lvl = 0) {
  check_sql(into, allow_null = TRUE, allow_names = FALSE)

  if (is.null(into)) {
    sql_clause("INSERT", cols, parens = TRUE, lvl = lvl)
  } else {
    kw <- paste0("INSERT INTO ", into)
    sql_clause(kw, cols, parens = TRUE, lvl = lvl)
  }
}

sql_clause_on <- function(on, lvl = 0, parens = FALSE) {
  sql_clause("ON", on, sep = " AND", parens = parens, lvl = lvl)
}

sql_clause_where_exists <- function(table, where, not) {
  check_sql(table)
  list(
    sql(paste0("WHERE ", if (not) "NOT ", "EXISTS (")),
    # lvl = 1 because they are basically in a subquery
    sql_clause("SELECT 1 FROM", table, lvl = 1),
    sql_clause_where(where, lvl = 1),
    sql(")")
  )
}

#' @export
print.sql_clause <- function(x, ...) {
  out <- sql_format_clause(x, lvl = 0)
  cat("<sql clause>", out)
}

# helpers -----------------------------------------------------------------

sql_format_clauses <- function(clauses, lvl = 0) {
  clauses <- purrr::compact(clauses)

  formatted_clauses <- purrr::map(clauses, sql_format_clause, lvl = lvl)
  clause_level <- purrr::map_dbl(clauses, "lvl", .default = 0)
  out <- indent_lvl(formatted_clauses, lvl + clause_level)

  sql(paste0(out, collapse = "\n"))
}

sql_format_clause <- function(x, lvl, nchar_max = 80) {
  if (is.sql(x)) {
    return(x)
  }

  lvl <- lvl + x$lvl

  # check length without starting a new line
  if (x$sep == " AND") {
    x$sep <- style_kw(x$sep)
  }

  fields_same_line <- sql_collapse(
    x$parts,
    collapse = paste0(x$sep, " "),
    parens = x$parens
  )

  x$kw <- style_kw(x$kw)
  same_line_clause <- paste0(x$kw, " ", fields_same_line)
  nchar_same_line <- cli::ansi_nchar(lvl_indent(lvl)) +
    cli::ansi_nchar(same_line_clause)

  if (length(x$parts) == 1 || nchar_same_line <= nchar_max) {
    return(sql(same_line_clause))
  }

  indent <- lvl_indent(lvl + 1)
  collapse <- paste0(x$sep, "\n", indent)

  field_string <- paste0(
    x$kw,
    if (x$parens) " (",
    "\n",
    indent,
    sql_collapse(x$parts, collapse = collapse),
    if (x$parens) paste0("\n", indent_lvl(")", lvl))
  )

  sql(field_string)
}

lvl_indent <- function(times, char = "  ") {
  strrep(char, times)
}

indent_lvl <- function(x, lvl) {
  sql(paste0(lvl_indent(lvl), x))
}

test_that("sql_clause_select generates expected SQL", {
  expect_snapshot(sql_clause_select(sql("x", "y")))
  expect_snapshot(sql_clause_select(sql("x"), distinct = TRUE))
  expect_snapshot(sql_clause_select(sql("x"), top = 10))
  expect_snapshot(sql_clause_select(sql("x"), distinct = TRUE, top = 5))
})

test_that("clauses wrap long inputs nicely", {
  vars <- paste0("variable", seq_len(8))
  expect_snapshot(sql_clause_select(sql(vars)))
})

test_that("sql_clause_select requires non-empty input", {
  expect_snapshot(sql_clause_select(sql()), error = TRUE)
})

test_that("sql_clause_where wraps in parens and combines conditions with AND", {
  expect_snapshot(sql_clause_where(sql("x > 1")))
  expect_snapshot(sql_clause_where(sql("x > 1", "y < 2")))
})

test_that("sql_clause_order_by warns when dropping in subquery", {
  expect_snapshot(sql_clause_order_by(sql("x"), subquery = TRUE))
})

test_that("sql_select_clauses generates expected SQL", {
  expect_snapshot(
    sql_select_clauses(
      select = sql_clause_select(sql("x", "y")),
      from = sql_clause_from(sql("foo")),
      where = sql_clause_where(sql("x > 1")),
      group_by = sql_clause_group_by(sql("x")),
      having = sql_clause_having(sql("COUNT(*) > 1")),
      window = sql_clause_window(sql("win1 AS (PARTITION BY x)")),
      order_by = sql_clause_order_by(sql("x")),
      limit = sql_clause_limit(NULL, 10)
    )
  )
})

test_that("update clauses generate expected SQL", {
  expect_snapshot(sql_clause_update(sql("foo")))
  expect_snapshot(sql_clause_set(sql("x", "y"), sql("1", "2")))
})

test_that("insert clauses generate expected SQL", {
  expect_snapshot(sql_clause_insert(sql("x", "y")))
  expect_snapshot(sql_clause_insert(sql("x", "y"), into = sql("foo")))
})

test_that("sql_clause_where_exists generates expected SQL", {
  clause_true <- sql_clause_where_exists(sql("bar"), sql("x = y"), not = FALSE)
  clause_false <- sql_clause_where_exists(sql("bar"), sql("x = y"), not = TRUE)

  expect_snapshot({
    sql_format_clauses(clause_true)
    sql_format_clauses(clause_false)
  })
})

on_gha <- function() identical(Sys.getenv("GITHUB_ACTIONS"), "true")
on_cran <- function() !identical(Sys.getenv("NOT_CRAN"), "true")
on_github_postgres <- function() {
  identical(Sys.getenv("GITHUB_POSTGRES"), "true")
}
on_github_mssql <- function() identical(Sys.getenv("GITHUB_MSSQL"), "true")

con_cache <- new_environment()
reg.finalizer(con_cache, function(e) {
  for (con in env_names(con_cache)) {
    DBI::dbDisconnect(con_cache[[con]])
  }
})

cache_test_con <- function(name, get_args) {
  if (!env_has(con_cache, name)) {
    args <- get_args()
    if (is.null(args)) {
      con <- NULL
    } else {
      con <- tryCatch(exec(DBI::dbConnect, !!!args), error = \(e) NULL)
    }

    env_poke(con_cache, name, con)
  } else {
    con <- env_get(con_cache, name)
  }

  if (is.null(con)) {
    testthat::skip(paste0("No ", name))
  } else {
    con
  }
}

test_postgres <- function() {
  cache_test_con("postgres", function() {
    if (on_github_postgres()) {
      list(
        drv = RPostgres::Postgres(),
        dbname = "test",
        user = "postgres",
        password = "password",
        host = "127.0.0.1"
      )
    } else if (on_gha() || on_cran()) {
      NULL
    } else {
      list(
        drv = RPostgres::Postgres(),
        dbname = "test",
        host = "localhost",
        user = ""
      )
    }
  })
}

test_mariadb <- function() {
  cache_test_con("mariadb", function() {
    if (on_gha() || on_cran()) {
      NULL
    } else {
      list(
        drv = RMariaDB::MariaDB(),
        dbname = "test",
        host = "localhost",
        username = Sys.getenv("USER")
      )
    }
  })
}

test_mssql <- function() {
  cache_test_con("mssql", function() {
    if (on_github_mssql()) {
      list(
        drv = odbc::odbc(),
        driver = "ODBC Driver 17 for SQL Server",
        database = "test",
        uid = "SA",
        pwd = "Password12",
        server = "localhost",
        port = 1433
      )
    } else {
      NULL
    }
  })
}

local_sqlite_con_with_aux <- function(envir = parent.frame()) {
  tmp <- tempfile()

  con <- withr::local_db_connection(
    DBI::dbConnect(RSQLite::SQLite(), ":memory:"),
    .local_envir = envir
  )
  DBI::dbExecute(con, paste0("ATTACH '", tmp, "' AS aux"))

  con
}

snap_transform_dbi <- function(x) {
  x <- gsub("dbplyr_tmp_[a-zA-Z0-9]+", "dbplyr_{tmp}", x)

  # use the last line matching this in case of multiple chained errors
  caused_by <- which(x == "Caused by error:")
  if (length(caused_by) == 0) {
    return(x)
  }

  dbi_line_id <- max(caused_by)

  n <- length(x)
  x <- x[-seq2(dbi_line_id + 1, n)]
  c(x, "! dummy DBI error")
}

#' Teradata backend
#'
#' @description
#' This backend supports Teradata databases, typically accessed via odbc. Use
#' `dialect_teradata()` with `lazy_frame()` to see simulated SQL without
#' connecting to a live database.
#'
#' Key differences for this backend are:
#'
#' * Uses `TOP` instead of `LIMIT`
#' * Selection of user supplied translations
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-teradata
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_teradata())
#' lf |> head()
NULL

#' @export
#' @rdname backend-teradata
dialect_teradata <- function() {
  new_sql_dialect(
    "teradata",
    quote_identifier = function(x) sql_quote(x, '"'),
    has_window_clause = TRUE,
    has_star_table_prefix = TRUE
  )
}

#' @export
#' @rdname backend-teradata
simulate_teradata <- function() simulate_dbi("Teradata")

#' @export
sql_dialect.Teradata <- function(con) {
  dialect_teradata()
}

#' @export
dbplyr_edition.Teradata <- function(con) {
  2L
}

#' @export
sql_query_select.sql_dialect_teradata <- function(
  con,
  select,
  from,
  where = NULL,
  group_by = NULL,
  having = NULL,
  window = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  ...,
  subquery = FALSE,
  lvl = 0
) {
  # #685
  # https://docs.teradata.com/r/2_MC9vCtAJRlKle2Rpb0mA/frQm7Rn09FJZZLQAuaUvJA
  # You cannot specify these options in a SELECT statement that specifies the TOP n operator:
  # * DISTINCT option
  # * QUALIFY clause
  # * SAMPLE clause
  # * WITH clause

  limit_needs_subquery <- is_true(distinct)

  if (!is_null(limit) && limit_needs_subquery) {
    unlimited_query <- sql_select_clauses(
      select = sql_clause_select(select, distinct, top = NULL),
      from = sql_clause_from(from),
      where = sql_clause_where(where),
      group_by = sql_clause_group_by(group_by),
      having = sql_clause_having(having),
      window = sql_clause_window(window),
      order_by = sql_clause_order_by(order_by, subquery, limit = NULL),
      lvl = lvl + 1
    )

    alias <- unique_subquery_name()
    from <- sql_query_wrap(con, unlimited_query, name = alias)
    select_outer <- sql_star(con, alias)
    out <- sql_select_clauses(
      select = sql_clause_select(select_outer, distinct = FALSE, top = limit),
      from = sql_clause_from(from),
      where = NULL,
      group_by = NULL,
      having = NULL,
      window = NULL,
      order_by = NULL,
      lvl = lvl
    )

    return(out)
  }

  sql_select_clauses(
    select = sql_clause_select(select, distinct, top = limit),
    from = sql_clause_from(from),
    where = sql_clause_where(where),
    group_by = sql_clause_group_by(group_by),
    having = sql_clause_having(having),
    window = sql_clause_window(window),
    order_by = sql_clause_order_by(order_by, subquery, limit),
    lvl = lvl
  )
}

#' @export
sql_translation.sql_dialect_teradata <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_odbc_scalar,
      `!=` = sql_infix("<>"),
      bitwNot = sql_prefix("BITNOT", 1),
      bitwAnd = sql_prefix("BITAND", 2),
      bitwOr = sql_prefix("BITOR", 2),
      bitwXor = sql_prefix("BITXOR", 2),
      bitwShiftL = sql_prefix("SHIFTLEFT", 2),
      bitwShiftR = sql_prefix("SHIFTRIGHT", 2),
      as.numeric = function(x, digits = 9L) {
        digits <- vctrs::vec_cast(digits, integer())
        sql_glue("CAST({x} AS DECIMAL(12, {digits}))")
      },
      as.double = sql_cast("FLOAT"),
      as.character = function(x, nchar = 255L) {
        check_number_whole(nchar, min = 0, max = 64000)
        nchar <- vctrs::vec_cast(nchar, integer())
        sql_glue("CAST({x} AS VARCHAR({nchar}))")
      },
      as.Date = teradata_as_date,
      log10 = sql_prefix("LOG"),
      log = sql_log(),
      cot = sql_cot(),
      quantile = sql_quantile("APPROX_PERCENTILE"),
      median = sql_median("APPROX_PERCENTILE"),
      nchar = sql_prefix("CHARACTER_LENGTH"),
      ceil = sql_prefix("CEILING"),
      ceiling = sql_prefix("CEILING"),
      atan2 = function(x, y) {
        sql_glue("ATAN2({y}, {x})")
      },
      substr = function(x, start, stop) {
        len <- stop - start + 1
        sql_glue("SUBSTR({x}, {start}, {len})")
      },
      startsWith = function(string, pattern) {
        sql_glue(
          "CAST(CASE WHEN INSTR({string}, {pattern}) = 1 THEN 1 ELSE 0 END AS INTEGER)"
        )
      },
      paste = sql_paste_infix(" ", "||", function(x) x),
      paste0 = sql_paste_infix("", "||", function(x) x),
      row_number = win_rank("ROW_NUMBER", empty_order = TRUE),

      # lubridate ---------------------------------------------------------------
      # https://en.wikibooks.org/wiki/SQL_Dialects_Reference/Functions_and_expressions/Date_and_time_functions
      as_date = teradata_as_date,
      week = function(x) {
        sql_glue("WEEKNUMBER_OF_YEAR({x}, 'iso')")
      },
      quarter = function(x) {
        sql_glue("TO_CHAR({x}, 'q')")
      }
    ),
    sql_translator(
      .parent = base_odbc_agg,
      var = sql_aggregate("VAR_SAMP", "var"),
      row_number = win_rank("ROW_NUMBER", empty_order = TRUE),
      weighted.mean = function(x, w, na.rm = T) {
        # nocov start
        check_unsupported_arg(na.rm, allowed = TRUE)
        win_over(
          sql_glue("SUM(({x} * {w})) / SUM({w})"),
          win_current_group(),
          win_current_group(),
          win_current_frame()
        )
        # nocov end
      }
    ),
    sql_translator(
      .parent = base_odbc_win,
      var = win_recycled("VAR_SAMP"),
      row_number = win_rank("ROW_NUMBER", empty_order = TRUE),
      lead = function(x, n = 1L, default = NA, order_by = NULL) {
        win_over(
          sql_glue("LEAD({x}, {n}, {default})"),
          win_current_group(),
          order_by %||% win_current_group(),
          win_current_frame()
        )
      },
      lag = function(x, n = 1L, default = NA, order_by = NULL) {
        n <- as.integer(n)
        win_over(
          sql_glue("LAG({x}, {n}, {default})"),
          win_current_group(),
          order_by %||% win_current_group(),
          win_current_frame()
        )
      },
      cumsum = function(x, order_by = NULL, frame = c(-Inf, 0)) {
        win_over(
          sql_glue("SUM({x})"),
          win_current_group(),
          order_by %||% win_current_group(),
          frame %||% win_current_frame()
        )
      },
      weighted.mean = function(x, w, na.rm = T) {
        # nocov start
        check_unsupported_arg(na.rm, allowed = TRUE)
        win_over(
          sql_glue("SUM(({x} * {w})) / SUM({w})"),
          win_current_group(),
          win_current_group(),
          win_current_frame()
        )
        # nocov end
      }
    )
  )
}

teradata_as_date <- function(x) {
  xq <- enquo(x)
  x_expr <- quo_get_expr(xq)
  if (is.character(x_expr) && !is.sql(x_expr)) {
    sql_glue("DATE {x}")
  } else {
    sql_cast("DATE")(x)
  }
}

#' @export
sql_table_analyze.sql_dialect_teradata <- function(con, table, ...) {
  # https://www.tutorialspoint.com/teradata/teradata_statistics.htm
  sql_glue2(con, "COLLECT STATISTICS {.tbl table}")
}

test_that("simulate_teradata() still works", {
  expect_translation(simulate_teradata(), x + 1, '"x" + 1.0')
})

test_that("custom scalar translated correctly", {
  con <- dialect_teradata()

  expect_translation(con, x != y, '"x" <> "y"')
  expect_translation(
    con,
    as.numeric(x),
    'CAST("x" AS DECIMAL(12, 9))'
  )
  expect_translation(
    con,
    as.numeric(x, 8),
    'CAST("x" AS DECIMAL(12, 8))'
  )
  expect_translation(con, as.double(x), 'CAST("x" AS FLOAT)')
  expect_translation(
    con,
    as.character(x),
    'CAST("x" AS VARCHAR(255))'
  )
  expect_translation(
    con,
    as.character(x, 12),
    'CAST("x" AS VARCHAR(12))'
  )
  expect_translation(con, log(x), 'LN("x")')
  expect_translation(con, cot(x), '1 / TAN("x")')
  expect_translation(con, nchar(x), 'CHARACTER_LENGTH("x")')
  expect_translation(con, ceil(x), 'CEILING("x")')
  expect_translation(con, ceiling(x), 'CEILING("x")')
  expect_translation(con, atan2(x, y), 'ATAN2("y", "x")')
  expect_translation(
    con,
    substr(x, 1, 2),
    'SUBSTR("x", 1.0, 2.0)'
  )
  expect_translation(
    con,
    startsWith(x, "A"),
    'CAST(CASE WHEN INSTR("x", \'A\') = 1 THEN 1 ELSE 0 END AS INTEGER)'
  )
  expect_translation(con, paste0(x, y), '"x" || "y"')
  expect_translation(con, paste(x, y), '"x" || \' \' || "y"')
  expect_translation(
    con,
    as.Date("2020-01-01"),
    "DATE '2020-01-01'"
  )
  expect_translation(con, as.Date(x), 'CAST("x" AS DATE)')
  expect_translation(
    con,
    week(as.Date("2020-01-01")),
    "WEEKNUMBER_OF_YEAR(DATE '2020-01-01', 'iso')"
  )
  expect_translation(
    con,
    quarter(as.Date("2020-01-01")),
    "TO_CHAR(DATE '2020-01-01', 'q')"
  )
})

test_that("custom bitwise operations translated correctly", {
  con <- dialect_teradata()

  expect_translation(con, bitwNot(x), 'BITNOT("x")')
  expect_translation(con, bitwAnd(x, 128L), 'BITAND("x", 128)')
  expect_translation(con, bitwOr(x, 128L), 'BITOR("x", 128)')
  expect_translation(con, bitwXor(x, 128L), 'BITXOR("x", 128)')
  expect_translation(con, bitwShiftL(x, 2L), 'SHIFTLEFT("x", 2)')
  expect_translation(con, bitwShiftR(x, 2L), 'SHIFTRIGHT("x", 2)')
})

test_that("custom aggregators translated correctly", {
  con <- dialect_teradata()

  expect_translation(
    con,
    var(x, na.rm = TRUE),
    'VAR_SAMP("x")',
    window = FALSE
  )
  expect_translation(
    con,
    cumsum(x, order_by = c(A, B)),
    'SUM("x") OVER (ORDER BY "A", "B" ROWS UNBOUNDED PRECEDING)'
  )
})

test_that("custom window functions translated correctly", {
  con <- dialect_teradata()

  expect_translation(
    con,
    var(x, na.rm = TRUE),
    'VAR_SAMP("x") OVER ()'
  )
  expect_translation(
    con,
    cumsum(x, order_by = c(A, B)),
    'SUM("x") OVER (ORDER BY "A", "B" ROWS UNBOUNDED PRECEDING)'
  )
})

test_that("generates custom sql", {
  con <- dialect_teradata()

  expect_snapshot(sql_table_analyze(con, in_schema("schema", "tbl")))
})

# verb translation --------------------------------------------------------

test_that("head translated to TOP", {
  mf <- lazy_frame(x = 1, con = dialect_teradata())
  expect_snapshot(mf |> head() |> sql_render())
})

test_that("lead, lag work", {
  mf <- lazy_frame(x = c(1:5), y = c(rep("A", 5)), con = dialect_teradata())

  expect_snapshot(
    mf |> group_by(y) |> mutate(val2 = lead(x, order_by = x)) |> sql_render()
  )
  expect_snapshot(
    mf |> group_by(y) |> mutate(val2 = lag(x, order_by = x)) |> sql_render()
  )
})


test_that("weighted.mean", {
  mf <- lazy_frame(x = c(1:5), y = c(6:10), con = dialect_teradata())

  expect_snapshot(mf |> summarise(wt_mean = weighted.mean(x, y)))
})

test_that("row_number() with and without group_by() and arrange(): unordered defaults to Ordering by NULL (per empty_order)", {
  mf <- lazy_frame(x = c(1:5), y = c(rep("A", 5)), con = dialect_teradata())
  expect_snapshot(mf |> mutate(rown = row_number()))
  expect_snapshot(mf |> group_by(y) |> mutate(rown = row_number()))
  expect_snapshot(mf |> arrange(y) |> mutate(rown = row_number()))
})

test_that("head after distinct() produces subquery", {
  lf <- lazy_frame(x = 1, y = 2, con = dialect_teradata())
  expect_snapshot(lf |> distinct() |> head())
})

test_that("simulate_odbc() still works", {
  expect_translation(simulate_odbc(), x + 1, '"x" + 1.0')
})

test_that("custom scalar translated correctly", {
  con <- dialect_odbc()

  expect_translation(con, as.numeric(x), 'CAST("x" AS DOUBLE)')
  expect_translation(con, as.double(x), 'CAST("x" AS DOUBLE)')
  expect_translation(con, as.integer(x), 'CAST("x" AS INT)')
  expect_translation(con, as.character(x), 'CAST("x" AS STRING)')
})

test_that("custom aggregators translated correctly", {
  con <- dialect_odbc()

  expect_translation(
    con,
    sd(x, na.rm = TRUE),
    'STDDEV_SAMP("x")',
    window = FALSE
  )
})

test_that("custom window functions translated correctly", {
  con <- dialect_odbc()

  expect_translation(
    con,
    sd(x, na.rm = TRUE),
    'STDDEV_SAMP("x") OVER ()'
  )
})

#' Summarise each group to one row
#'
#' This is a method for the dplyr [summarise()] generic. It generates the
#' `SELECT` clause of the SQL query, and generally needs to be combined with
#' `group_by()`.
#'
#' @inheritParams arrange.tbl_lazy
#' @inheritParams dplyr::summarise
#' @param .groups \Sexpr[results=rd]{lifecycle::badge("experimental")} Grouping structure of the result.
#'
#'   * "drop_last": dropping the last level of grouping. This was the
#'   only supported option before version 1.0.0.
#'   * "drop": All levels of grouping are dropped.
#'   * "keep": Same grouping structure as `.data`.
#'
#'   When `.groups` is not specified, it defaults to "drop_last".
#'
#'   In addition, a message informs you of that choice, unless the result is ungrouped,
#'   the option "dplyr.summarise.inform" is set to `FALSE`,
#'   or when `summarise()` is called from a function in a package.
#' @inherit arrange.tbl_lazy return
#' @importFrom dplyr summarise
#' @export
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(g = c(1, 1, 1, 2, 2), x = c(4, 3, 6, 9, 2))
#' db |>
#'   summarise(n()) |>
#'   show_query()
#'
#' db |>
#'   group_by(g) |>
#'   summarise(n()) |>
#'   show_query()
summarise.tbl_lazy <- function(.data, ..., .by = NULL, .groups = NULL) {
  check_groups(.groups)

  by <- compute_by(
    {{ .by }},
    .data,
    by_arg = ".by",
    data_arg = ".data",
    error_call = caller_env()
  )

  if (by$from_by) {
    .groups <- "drop"
  }
  .data <- set_by_groups(.data, by)

  dots <- summarise_eval_dots(.data, ...)
  .data$lazy_query <- add_summarise(
    .data$lazy_query,
    dots,
    .groups = .groups,
    env_caller = caller_env()
  )

  clear_by_groups(.data, by)
}

#' @export
#' @importFrom dplyr reframe
reframe.tbl_lazy <- function(.data, ..., .by = NULL) {
  stop_unsupported_function("reframe")
}

summarise_eval_dots <- function(.data, ..., error_call = caller_env()) {
  dots <- as.list(enquos(..., .named = TRUE))
  dot_names <- names2(exprs(...))
  was_named <- have_name(exprs(...))
  cur_data <- .data
  error_env <- new_environment()

  for (i in seq_along(dots)) {
    dot <- dots[[i]]
    dot_name <- dot_names[[i]]
    parent.env(error_env) <- quo_get_env(dot)
    dot <- quo_set_env(dot, error_env)

    dots[[i]] <- partial_eval_quo(
      dot,
      .data,
      dot_name,
      error_call = error_call,
      was_named = was_named[[i]]
    )
    cur_data <- summarise_bind_error(cur_data, dots, i, error_env)
  }

  # Remove names from any list elements
  is_list <- purrr::map_lgl(dots, is.list)
  names2(dots)[is_list] <- ""

  # Auto-splice list results from partial_eval_quo()
  dots[!is_list] <- lapply(dots[!is_list], list)
  unlist(dots, recursive = FALSE)
}

summarise_bind_error <- function(cur_data, dots, i, error_env) {
  quos <- dots[[i]]
  if (!is.list(quos)) {
    quos <- set_names(list(quos), names(dots)[[i]])
  }

  for (j in seq_along(quos)) {
    dot_name <- names(quos)[[j]]
    summarise_bind_error1(error_env, dot_name)
    # remove variable from `cur_data` so that `partial_eval_sym()` evaluates
    # variable in `error_env`
    cur_data$lazy_query <- add_select(
      cur_data$lazy_query,
      set_names(setdiff(colnames(cur_data), dot_name))
    )
  }

  cur_data
}

summarise_bind_error1 <- function(error_env, dot_name) {
  msg <- c(
    "In {.pkg dbplyr} you cannot use a variable created in the same {.fun summarise}.",
    x = "{.var {dot_name}} was created earlier in this {.fun summarise}.",
    i = "You need an extra {.fun mutate} step to use it."
  )
  env_bind_lazy(
    error_env,
    !!dot_name := cli::cli_abort(msg, call = NULL)
  )
}

check_groups <- function(.groups) {
  if (is_null(.groups)) {
    return()
  }

  if (.groups %in% c("drop_last", "drop", "keep")) {
    return()
  }

  cli_abort(
    c(
      paste0(
        "{.arg .groups} can't be {as_label(.groups)}",
        if (.groups == "rowwise") " in dbplyr"
      ),
      i = 'Possible values are NULL (default), "drop_last", "drop", and "keep"'
    ),
    call = caller_env()
  )
}

add_summarise <- function(lazy_query, exprs, .groups, env_caller) {
  cur_grps <- op_grps(lazy_query)
  summarise_message(cur_grps, .groups, env_caller)

  new_grps <- switch(
    .groups %||% "drop_last",
    drop_last = cur_grps[-length(cur_grps)],
    keep = cur_grps,
    drop = character()
  )

  # ensure grouping variables are listed first
  vars <- c(cur_grps, setdiff(names(exprs), cur_grps))
  select <- syms(set_names(vars))
  select[names(exprs)] <- exprs

  if (can_inline_summarise(lazy_query)) {
    lazy_query$select <- new_lazy_select(select, group_vars = new_grps)
    lazy_query$select_operation <- "summarise"
    lazy_query$group_by <- syms(cur_grps)
    lazy_query$group_vars <- new_grps
    lazy_query
  } else {
    lazy_select_query(
      x = lazy_query,
      select = select,
      group_by = syms(cur_grps),
      group_vars = new_grps,
      select_operation = "summarise"
    )
  }
}

# summarise() adds GROUP BY and modifies SELECT
# * GROUP BY is executed after WHERE but before SELECT, DISTINCT, ORDER BY, LIMIT
#   => can inline if previous query only has WHERE or ORDER BY
# * Can't inline after another summarise (would need to re-aggregate)
# * Can only inline if previous SELECT is a pure projection
can_inline_summarise <- function(lazy_query) {
  if (!is_lazy_select_query(lazy_query)) {
    return(FALSE)
  }

  if (lazy_query$select_operation == "summarise") {
    return(FALSE)
  }

  if (is_true(lazy_query$distinct)) {
    return(FALSE)
  }
  if (!is_null(lazy_query$limit)) {
    return(FALSE)
  }

  if (!is_pure_projection(lazy_query$select$expr, lazy_query$select$name)) {
    return(FALSE)
  }

  TRUE
}

summarise_message <- function(grps, .groups, env_caller) {
  verbose <- summarise_verbose(.groups, env_caller)
  if (!verbose) {
    return(invisible())
  }

  n <- length(grps)
  if (n <= 1) {
    return(invisible())
  }

  cli::cli_inform(c(
    "!" = "Grouped output by {.val {grps[-n]}}.",
    i = "Override behaviour and silence this message with the {.arg .groups} argument.",
    i = "Or use {.arg .by} instead of {.fun group_by}."
  ))
  invisible()
}

summarise_verbose <- function(.groups, .env) {
  is.null(.groups) &&
    is_reference(topenv(.env), global_env()) &&
    !identical(getOption("dplyr.summarise.inform"), FALSE)
}

compute_by <- function(
  by,
  data,
  ...,
  by_arg = "by",
  data_arg = "data",
  error_call = caller_env()
) {
  check_dots_empty0(...)

  by <- enquo(by)
  check_by(
    by,
    data,
    by_arg = by_arg,
    data_arg = data_arg,
    error_call = error_call
  )

  if (is_grouped_lf(data)) {
    names <- group_vars(data)
    from_by <- FALSE
  } else {
    by <- eval_select_by(by, data, error_call = error_call)
    names <- by
    from_by <- TRUE
  }

  new_by(from_by = from_by, names = names)
}

is_grouped_lf <- function(data) {
  !is_empty(group_vars(data))
}

check_by <- function(
  by,
  data,
  ...,
  by_arg = "by",
  data_arg = "data",
  error_call = caller_env()
) {
  check_dots_empty0(...)

  if (quo_is_null(by)) {
    return(invisible(NULL))
  }

  if (is_grouped_lf(data)) {
    message <- paste0(
      "Can't supply {.arg {by_arg}} when ",
      "{.arg {data_arg}} is a grouped data frame."
    )
    cli::cli_abort(message, call = error_call)
  }

  invisible(NULL)
}

eval_select_by <- function(by, data, error_call = caller_env()) {
  out <- tidyselect::eval_select(
    expr = by,
    data = data,
    allow_rename = FALSE,
    error_call = error_call
  )
  names(out)
}

new_by <- function(from_by, names) {
  structure(list(from_by = from_by, names = names), class = "dbplyr_by")
}

# Apply `.by` groups for a single verb call. `set_by_groups()` installs the
# groups before the operation so window functions and HAVING clauses see them;
# `clear_by_groups()` removes them after so they don't leak into subsequent
# operations (matching dplyr's `.by` semantics).
set_by_groups <- function(.data, by) {
  if (by$from_by) {
    .data$lazy_query$group_vars <- by$names
  }
  .data
}

clear_by_groups <- function(.data, by) {
  if (by$from_by) {
    .data$lazy_query$group_vars <- character()
  }
  .data
}

#' Save results into a new remote table
#'
#' `compute()` executes the query and stores the results in a new remote table.
#' This is useful when you want to cache intermediate results for reuse or to
#' improve performance by avoiding repeated computation of complex queries.
#'
#' @inheritParams collect.tbl_sql
#' @inheritParams copy_to.src_sql
#' @export
#' @importFrom dplyr compute
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(a = c(3, 4, 1, 2), b = c(5, 1, 2, NA))
#' db |> filter(a <= 2) |> show_query()
#' db |> filter(a <= 2) |> compute() |> show_query()
compute.tbl_sql <- function(
  x,
  name = NULL,
  temporary = TRUE,
  overwrite = FALSE,
  unique_indexes = list(),
  indexes = list(),
  analyze = TRUE,
  ...,
  sql_options = NULL,
  cte = deprecated()
) {
  check_bool(temporary)
  rlang::check_exclusive(sql_options, cte, .require = FALSE)
  if (lifecycle::is_present(cte)) {
    lifecycle::deprecate_warn(
      when = "2.4.0",
      what = "compute(cte)",
      with = I("compute(sql_options = sql_options(cte = TRUE))")
    )
    sql_options <- dbplyr::sql_options(cte = cte)
  }

  if (is.null(name)) {
    if (!temporary) {
      lifecycle::deprecate_stop(
        "2.3.3",
        what = "compute(name = 'must be provided when `temporary = FALSE`')"
      )
    }
    name <- unique_table_name()
  }

  name <- as_table_path(name, x$con)
  vars <- op_vars(x)

  compute_check_indexes(x, indexes)
  compute_check_indexes(x, unique_indexes)

  x_aliased <- select(x, !!!syms(vars)) # avoids problems with SQLite quoting (#1754)
  sql <- db_sql_render(x$con, x_aliased$lazy_query, sql_options = sql_options)

  name <- db_compute(
    x$con,
    name,
    sql,
    temporary = temporary,
    overwrite = overwrite,
    unique_indexes = unique_indexes,
    indexes = indexes,
    analyze = analyze,
    ...
  )

  dplyr::tbl(x$con, from = name, vars = colnames(x)) |>
    group_by(!!!syms(op_grps(x))) |>
    window_order(!!!op_sort(x))
}

compute_check_indexes <- function(
  x,
  indexes,
  ...,
  arg = caller_arg(indexes),
  error_call = caller_env()
) {
  if (is.null(indexes)) {
    return()
  }

  check_dots_empty()
  force(arg)
  if (!is.list(indexes) && !is.character(indexes)) {
    stop_input_type(
      indexes,
      c("a character vector", "a list of characters"),
      arg = arg,
      call = error_call
    )
  }
  if (is.list(indexes)) {
    indexes <- unique(unlist(indexes))
  }

  x_nms <- colnames(x)

  missing <- setdiff(indexes, x_nms)
  if (!is_empty(missing)) {
    message <- c(
      "All columns specified through {.arg {arg}} must exist in {.arg x}.",
      i = "The following columns are missing from {.arg {arg}}: {.field {missing}}."
    )

    cli_abort(message, call = error_call)
  }
}

#' @export
#' @rdname sql_build
lazy_rf_join_query <- function(
  x,
  y,
  type,
  by,
  table_names,
  vars,
  group_vars = op_grps(x),
  order_vars = op_sort(x),
  frame = op_frame(x),
  call = caller_env()
) {
  check_lazy_query(x, call = call)
  check_lazy_query(y, call = call)

  check_character(type, call = call)

  check_has_names(table_names, c("name", "from"), call = call)
  check_character(table_names$name, call = call)
  check_character(table_names$from, call = call)

  check_has_names(vars, c("name", "x", "y"), call = call)
  check_character(vars$name, call = call)
  check_character(vars$x, call = call)
  check_character(vars$y, call = call)

  lazy_query(
    query_type = "rf_join",
    x = x,
    y = y,
    type = type,
    by = by,
    table_names = table_names,
    vars = vars,
    group_vars = group_vars,
    order_vars = order_vars,
    frame = frame
  )
}

#' @export
op_vars.lazy_rf_join_query <- function(op) {
  op$vars$name
}

#' @export
sql_build.lazy_rf_join_query <- function(op, con, ..., sql_options = NULL) {
  table_names_out <- generate_join_table_names(op$table_names, con)

  vars_classic <- as.list(op$vars)
  vars_classic$all_x <- op_vars(op$x)
  vars_classic$all_y <- op_vars(op$y)

  by <- op$by
  by$x_as <- table_names_out[[1]]
  by$y_as <- table_names_out[[2]]

  select <- sql_rf_join_vars(
    con,
    type = op$type,
    vars = vars_classic,
    x_as = by$x_as,
    y_as = by$y_as,
    use_star = sql_options$use_star,
    qualify_all_columns = sql_options$qualify_all_columns
  )

  rf_join_query(
    sql_build(op$x, con, sql_options = sql_options),
    sql_build(op$y, con, sql_options = sql_options),
    select = select,
    type = op$type,
    by = by,
    na_matches = by$na_matches
  )
}

# Built query -------------------------------------------------------------

#' @export
#' @rdname sql_build
rf_join_query <- function(
  x,
  y,
  select,
  ...,
  type = "inner",
  by = NULL,
  suffix = c(".x", ".y"),
  na_matches = FALSE
) {
  query(
    "rf_join",
    x = x,
    y = y,
    select = select,
    type = type,
    by = by,
    na_matches = na_matches
  )
}

#' @export
sql_render.rf_join_query <- function(
  query,
  con = NULL,
  ...,
  sql_options = NULL,
  subquery = FALSE,
  lvl = 0
) {
  from_x <- sql_render(query$x, con, ..., subquery = TRUE, lvl = lvl + 1)
  from_y <- sql_render(query$y, con, ..., subquery = TRUE, lvl = lvl + 1)

  dbplyr_query_join(
    con,
    from_x,
    from_y,
    vars = query$vars,
    type = query$type,
    by = query$by,
    na_matches = query$na_matches,
    select = query$select,
    lvl = lvl
  )
}

#' @export
flatten_query.rf_join_query <- function(qry, query_list, con) {
  flatten_query_2_tables(qry, query_list, con)
}

# SQL generation ----------------------------------------------------------

#' @rdname db-sql
#' @export
sql_query_join <- function(
  con,
  x,
  y,
  select,
  type = "inner",
  by = NULL,
  na_matches = FALSE,
  ...,
  lvl = 0
) {
  check_dots_used()
  UseMethod("sql_query_join", sql_dialect(con))
}
#' @export
sql_query_join.DBIConnection <- function(
  con,
  x,
  y,
  select,
  type = "inner",
  by = NULL,
  na_matches = FALSE,
  ...,
  lvl = 0
) {
  JOIN <- switch(
    type,
    left = sql("LEFT JOIN"),
    inner = sql("INNER JOIN"),
    right = sql("RIGHT JOIN"),
    full = sql("FULL JOIN"),
    cross = sql("CROSS JOIN"),
    cli_abort("Unknown join type: {.val {type}}")
  )

  x <- dbplyr_sql_subquery(con, x, name = by$x_as, lvl = lvl)
  y <- dbplyr_sql_subquery(con, y, name = by$y_as, lvl = lvl)

  on <- sql_join_tbls(con, by, na_matches = na_matches)

  # Wrap with SELECT since callers assume a valid query is returned
  clauses <- list(
    sql_clause_select(select),
    sql_clause_from(x),
    sql_clause(JOIN, y),
    sql_clause("ON", on, sep = " AND", parens = TRUE, lvl = 1)
  )
  sql_format_clauses(clauses, lvl)
}

#' @export
sql_query_join.sql_dialect <- sql_query_join.DBIConnection
dbplyr_query_join <- function(
  con,
  x,
  y,
  vars,
  type = "inner",
  by = NULL,
  na_matches = FALSE,
  ...,
  select = NULL,
  lvl = 0
) {
  check_2ed(con)
  sql_query_join(
    con,
    x,
    y,
    select,
    type = type,
    by = by,
    na_matches = na_matches,
    ...,
    lvl = lvl
  )
}

# Helpers ----------------------------------------------------------------------

sql_rf_join_vars <- function(
  con,
  type,
  vars,
  x_as = "LHS",
  y_as = "RHS",
  use_star,
  qualify_all_columns
) {
  type <- arg_match0(type, c("right", "full"))

  check_table_path(x_as)
  check_table_path(y_as)
  table_names <- c(x_as, y_as)

  if (type == "full") {
    duplicated_vars <- intersect(tolower(vars$all_x), tolower(vars$all_y))
    out <- purrr::map2(
      vars$x,
      vars$y,
      ~ {
        if (!is.na(.x) && !is.na(.y)) {
          x_prefix <- sql_table_prefix(con, x_as, .x)
          y_prefix <- sql_table_prefix(con, y_as, .y)
          sql_glue2(con, "COALESCE({x_prefix}, {y_prefix})")
        } else if (!is.na(.x)) {
          sql_multi_join_var(con, .x, table_names[[1L]], duplicated_vars)
        } else {
          sql_multi_join_var(con, .y, table_names[[2L]], duplicated_vars)
        }
      }
    )

    out <- set_names(out, vars$name)
    return(names_to_as(con, unlist(out)))
  }

  multi_join_vars <- purrr::map2_dfr(
    vars$x,
    vars$y,
    ~ {
      if (!is.na(.x)) {
        table <- 1L
        var <- .x
      } else {
        table <- 2L
        var <- .y
      }

      vctrs::new_data_frame(list(table = table, var = var), n = 1L)
    }
  )

  multi_join_vars <- vctrs::vec_cbind(name = vars$name, multi_join_vars)
  table_vars <- set_names(vars[c("all_x", "all_y")], table_names)

  sql_multi_join_select(
    con = con,
    vars = multi_join_vars,
    table_vars = table_vars,
    use_star = use_star,
    qualify_all_columns = qualify_all_columns
  )
}

#' Redshift backend
#'
#' @description
#' This backend supports Amazon Redshift databases, typically accessed via
#' a `RedshiftConnection` created by [DBI::dbConnect()]. Use `dialect_redshift()`
#' with `lazy_frame()` to see simulated SQL without connecting to a live
#' database.
#'
#' Base translations come from [PostgreSQL backend][dialect_postgres]. There
#' are generally few differences, apart from string manipulation.
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-redshift
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_redshift())
#' lf |> transmute(x = paste(c, " times"))
#' lf |> transmute(x = substr(c, 2, 3))
#' lf |> transmute(x = str_replace_all(c, "a", "z"))
NULL

#' @export
#' @rdname backend-redshift
dialect_redshift <- function() {
  new_sql_dialect(
    "redshift",
    quote_identifier = function(x) sql_quote(x, '"')
  )
}

#' @export
#' @rdname backend-redshift
simulate_redshift <- function() simulate_dbi("RedshiftConnection")

#' @export
sql_dialect.RedshiftConnection <- function(con) {
  dialect_redshift()
}

#' @export
sql_dialect.Redshift <- sql_dialect.RedshiftConnection

#' @export
dbplyr_edition.RedshiftConnection <- function(con) {
  2L
}
#' @export
dbplyr_edition.Redshift <- dbplyr_edition.RedshiftConnection

redshift_round <- function(x, digits = 0L) {
  digits <- as.integer(digits)
  sql_glue("ROUND(({x})::float, {digits})")
}

#' @export
sql_translation.sql_dialect_redshift <- function(con) {
  postgres <- sql_translation(dialect_postgres())

  sql_variant(
    sql_translator(
      .parent = postgres$scalar,

      # https://docs.aws.amazon.com/redshift/latest/dg/r_Numeric_types201.html#r_Numeric_types201-floating-point-types
      as.numeric = sql_cast("FLOAT"),
      as.double = sql_cast("FLOAT"),
      round = redshift_round,

      # https://stackoverflow.com/questions/56708136
      paste = sql_paste_infix(" ", "||"),
      paste0 = sql_paste_infix("", "||"),
      str_c = sql_paste_infix("", "||"),

      str_ilike = function(string, pattern) {
        sql_glue("{string} ILIKE {pattern}")
      },

      # https://docs.aws.amazon.com/redshift/latest/dg/r_SUBSTRING.html
      substr = sql_substr("SUBSTRING"),
      substring = sql_substr("SUBSTRING"),
      str_sub = sql_str_sub("SUBSTRING", "LEN"),

      # https://docs.aws.amazon.com/redshift/latest/dg/REGEXP_REPLACE.html
      str_replace = sql_not_supported("str_replace"),
      str_replace_all = function(string, pattern, replacement) {
        sql_glue("REGEXP_REPLACE({string}, {pattern}, {replacement})")
      },

      # clock ---------------------------------------------------------------
      add_days = function(x, n, ...) {
        check_dots_empty()
        sql_glue("DATEADD(DAY, {n}, {x})")
      },
      add_years = function(x, n, ...) {
        check_dots_empty()
        sql_glue("DATEADD(YEAR, {n}, {x})")
      },
      date_build = function(year, month = 1L, day = 1L, ..., invalid = NULL) {
        check_unsupported_arg(invalid, allow_null = TRUE)
        sql_glue(
          "TO_DATE(CAST({year} AS TEXT) || '-' || CAST({month} AS TEXT) || '-' || CAST({day} AS TEXT), 'YYYY-MM-DD')"
        )
      },
      get_year = function(x) {
        sql_glue("DATE_PART('year', {x})")
      },
      get_month = function(x) {
        sql_glue("DATE_PART('month', {x})")
      },
      get_day = function(x) {
        sql_glue("DATE_PART('day', {x})")
      },
      date_count_between = function(start, end, precision, ..., n = 1L) {
        check_dots_empty()
        check_unsupported_arg(precision, allowed = "day")
        check_unsupported_arg(n, allowed = 1L)

        sql_glue("DATEDIFF(DAY, {start}, {end})")
      },

      difftime = function(time1, time2, tz, units = "days") {
        check_unsupported_arg(tz)
        check_unsupported_arg(units, allowed = "days")

        sql_glue("DATEDIFF(DAY, {time2}, {time1})")
      }
    ),
    sql_translator(
      .parent = postgres$aggregate,
      # https://docs.aws.amazon.com/redshift/latest/dg/r_LISTAGG.html
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        sql_glue("LISTAGG({x}, {collapse})")
      }
    ),
    sql_translator(
      .parent = postgres$window,
      median = sql_win_not_supported("median", "Redshift"),
      quantile = sql_win_not_supported("quantile", "Redshift"),
      # https://docs.aws.amazon.com/redshift/latest/dg/r_WF_LAG.html
      lag = function(x, n = 1L, order_by = NULL) {
        n <- as.integer(n)
        win_over(
          sql_glue("LAG({x}, {n})"),
          win_current_group(),
          order_by %||% win_current_order(),
          win_current_frame()
        )
      },
      # https://docs.aws.amazon.com/redshift/latest/dg/r_WF_LEAD.html
      lead = function(x, n = 1L, order_by = NULL) {
        n <- as.integer(n)
        win_over(
          sql_glue("LEAD({x}, {n})"),
          win_current_group(),
          order_by %||% win_current_order(),
          win_current_frame()
        )
      },
      # https://docs.aws.amazon.com/redshift/latest/dg/r_LISTAGG.html
      str_flatten = function(x, collapse = "", na.rm = FALSE) {
        sql_check_na_rm(na.rm)
        order <- win_current_order()
        listagg_sql <- sql_glue("LISTAGG({x}, {collapse})")

        if (length(order) > 0) {
          sql <- sql_glue("{listagg_sql} WITHIN GROUP (ORDER BY {order})")
        } else {
          sql <- listagg_sql
        }

        win_over(
          sql,
          partition = win_current_group()
          # Cannot use the order here because LISTAGG requires the ordering in the
          # WITHIN GROUP (ORDER BY ...) clause
        )
      }
    )
  )
}

# https://docs.aws.amazon.com/redshift/latest/dg/r_EXPLAIN.html
#' @export
sql_query_explain.sql_dialect_redshift <- function(con, sql, ...) {
  sql_glue2(con, "EXPLAIN {sql}")
}

#' @export
sql_values_subquery.sql_dialect_redshift <- function(
  con,
  df,
  types,
  lvl = 0,
  ...
) {
  sql_values_subquery_union(con, df, types = types, lvl = lvl)
}

test_that("simulate_redshift() still works", {
  expect_translation(simulate_redshift(), x + 1, '"x" + 1.0')
})

test_that("defaults to postgres translations", {
  con <- dialect_redshift()
  expect_translation(con, log10(x), 'LOG("x")')
})

test_that("string translations", {
  con <- dialect_redshift()

  expect_error(
    translate_sql(str_replace("xx", ".", "a"), con = con),
    class = "dbplyr_error_unsupported_fn"
  )
  expect_translation(
    con,
    str_replace_all("xx", ".", "a"),
    "REGEXP_REPLACE('xx', '.', 'a')"
  )

  expect_translation(con, substr(x, 2, 2), 'SUBSTRING("x", 2, 1)')
  expect_translation(
    con,
    str_sub(x, 2, -2),
    'SUBSTRING("x", 2, LEN("x") - 2)'
  )

  expect_translation(con, paste("x", "y"), "'x' || ' ' || 'y'")
  expect_translation(con, paste0("x", "y"), "'x' || 'y'")
  expect_translation(con, str_c("x", "y"), "'x' || 'y'")
  expect_translation(con, str_ilike(x, y), '"x" ILIKE "y"')
})

test_that("numeric translations", {
  con <- dialect_redshift()

  expect_translation(con, as.numeric(x), 'CAST("x" AS FLOAT)')
  expect_translation(con, as.double(x), 'CAST("x" AS FLOAT)')
  expect_translation(con, round(1.234, 1), "ROUND((1.234)::float, 1)")
})

test_that("aggregate functions", {
  con <- dialect_redshift()

  expect_translation(
    con,
    str_flatten(x, y),
    'LISTAGG("x", "y")',
    window = FALSE
  )
  expect_translation(
    con,
    str_flatten(x, y),
    'LISTAGG("x", "y") OVER ()',
    window = TRUE
  )
  expect_translation(
    con,
    order_by(z, str_flatten(x, y)),
    'LISTAGG("x", "y") WITHIN GROUP (ORDER BY "z") OVER ()'
  )
})

test_that("lag and lead translation", {
  con <- dialect_redshift()

  expect_translation(con, lead(x), 'LEAD("x", 1) OVER ()')
  expect_translation(con, lag(x), 'LAG("x", 1) OVER ()')
})

test_that("copy_inline uses UNION ALL", {
  con <- dialect_redshift()
  y <- tibble::tibble(id = 1L, arr = "{1,2,3}")

  types <- c(id = "bigint", arr = "integer[]")
  expect_snapshot({
    copy_inline(con, y |> slice(0)) |> remote_query()
    copy_inline(con, y) |> remote_query()

    # with `types`
    copy_inline(con, y |> slice(0), types = types) |> remote_query()
    copy_inline(con, y, types = types) |> remote_query()
  })
})

test_that("custom clock functions translated correctly", {
  con <- dialect_redshift()
  expect_translation(
    con,
    add_years(x, 1),
    'DATEADD(YEAR, 1.0, "x")'
  )
  expect_translation(
    con,
    add_days(x, 1),
    'DATEADD(DAY, 1.0, "x")'
  )
  expect_error(
    translate_sql(add_days(x, 1, "dots", "must", "be empty"), con = con),
    class = "rlib_error_dots_nonempty"
  )
  expect_translation(
    con,
    date_build(2020, 1, 1),
    "TO_DATE(CAST(2020.0 AS TEXT) || '-' || CAST(1.0 AS TEXT) || '-' || CAST(1.0 AS TEXT), 'YYYY-MM-DD')"
  )
  expect_translation(
    con,
    date_build(year_column, 1L, 1L),
    'TO_DATE(CAST("year_column" AS TEXT) || \'-\' || CAST(1 AS TEXT) || \'-\' || CAST(1 AS TEXT), \'YYYY-MM-DD\')'
  )
  expect_translation(
    con,
    get_year(date_column),
    'DATE_PART(\'year\', "date_column")'
  )
  expect_translation(
    con,
    get_month(date_column),
    'DATE_PART(\'month\', "date_column")'
  )
  expect_translation(
    con,
    get_day(date_column),
    'DATE_PART(\'day\', "date_column")'
  )
  expect_translation(
    con,
    date_count_between(date_column_1, date_column_2, "day"),
    'DATEDIFF(DAY, "date_column_1", "date_column_2")'
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      date_count_between(date_column_1, date_column_2, "year"),
      con = con
    )
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      date_count_between(
        date_column_1,
        date_column_2,
        "day",
        n = 5
      ),
      con = con
    )
  )
})

test_that("difftime is translated correctly", {
  con <- dialect_redshift()
  expect_translation(
    con,
    difftime(start_date, end_date, units = "days"),
    'DATEDIFF(DAY, "end_date", "start_date")'
  )
  expect_translation(
    con,
    difftime(start_date, end_date),
    'DATEDIFF(DAY, "end_date", "start_date")'
  )

  expect_snapshot(
    error = TRUE,
    translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      difftime(
        start_date,
        end_date,
        tz = "UTC",
        units = "days"
      ),
      con = con
    )
  )
})

test_that("window functions with redshift specific error message", {
  con <- dialect_redshift()
  expect_error(
    translate_sql(quantile(x, 0.3, na.rm = TRUE), window = TRUE, con = con),
    "Redshift"
  )
  expect_error(
    translate_sql(median(x, na.rm = TRUE), window = TRUE, con = con),
    "Redshift"
  )
})

capture_across <- function(data, x) {
  x <- enquo(x)
  partial_eval_across(get_expr(x), data, get_env(x))
}

partial_eval_pick <- function(call, data, env, error_call = caller_env()) {
  args <- call_args(call)
  if (length(args) == 0) {
    cli_abort(
      "Must supply at least one input to {.fn pick}.",
      error_call = error_call
    )
  }

  locs <- tidyselect::eval_select(
    expr(c(!!!args)),
    data,
    env = env,
    allow_rename = FALSE,
    error_call = call("pick()")
  )

  syms(names(locs))
}

partial_eval_across <- function(call, data, env, error_call = caller_env()) {
  across_dummy <- function(
    .cols = everything(),
    .fns = NULL,
    ...,
    .names = NULL,
    .unpack = FALSE
  ) {}

  call <- call_match(
    call,
    fn = across_dummy,
    defaults = TRUE,
    dots_expand = FALSE,
    dots_env = env
  )
  deprecate_across_dots(call)
  check_unsupported_arg(call$.unpack, FALSE, arg = ".unpack", call = error_call)

  across_setup(
    data,
    call,
    env,
    allow_rename = TRUE,
    fn = "across()",
    error_call = error_call
  )
}

capture_if_all <- function(data, x) {
  x <- enquo(x)
  partial_eval_if(get_expr(x), data, get_env(x))
}

partial_eval_if <- function(
  call,
  data,
  env,
  reduce = "&",
  error_call = caller_env()
) {
  call <- match.call(dplyr::if_any, call, expand.dots = FALSE, envir = env)
  deprecate_across_dots(call)

  if (reduce == "&") {
    fn <- "if_all()"
  } else {
    fn <- "if_any()"
  }
  conditions <- across_setup(
    data,
    call,
    env,
    allow_rename = FALSE,
    fn = fn,
    error_call = error_call
  )
  if (is_empty(conditions)) {
    return(TRUE)
  }
  Reduce(\(x, y) call2(reduce, x, y), conditions)
}

deprecate_across_dots <- function(call) {
  if (!is_empty(call$...)) {
    details <- paste(
      c(
        "Supply arguments directly to `.fns` through a lambda instead.",
        "",
        "  # Previously",
        "  across(a:b, mean, na.rm = TRUE)",
        "",
        "  # Now",
        "  across(a:b, ~mean(.x, na.rm = TRUE))"
      ),
      collapse = "\n"
    )
    lifecycle::deprecate_stop(
      when = "2.3.0",
      what = "across(...)",
      details = details
    )
  }
}

across_funs <- function(funs, env, names_spec, fn, evaluated = FALSE) {
  if (is.null(funs)) {
    fns <- list(`1` = \(x, ...) x)
    names_spec <- names_spec %||% "{.col}"
  } else if (is_quosure(funs)) {
    return(across_funs(
      quo_squash(funs),
      env,
      names_spec,
      fn,
      evaluated = evaluated
    ))
  } else if (is_call(funs, "::")) {
    return(across_funs(
      funs[[3]],
      env,
      names_spec,
      fn,
      evaluated = evaluated
    ))
  } else if (
    is_symbol(funs) ||
      is_function(funs) ||
      is_call(funs, "~") ||
      is_call(funs, "function")
  ) {
    is_local_list <- function(funs) {
      if (!is_symbol(funs)) {
        return(FALSE)
      }

      funs <- as_name(funs)
      exists(funs, env) && is.list(get(funs, envir = env))
    }

    if (is_local_list(funs)) {
      funs <- eval(funs, env)
      return(across_funs(
        funs,
        env,
        names_spec,
        fn,
        evaluated = evaluated
      ))
    }

    fns <- list(`1` = across_fun(funs, env, fn = fn))
    names_spec <- names_spec %||% "{.col}"
  } else if (is_call(funs, "list")) {
    args <- call_args(funs)
    fns <- lapply(args, across_fun, env, fn = fn)
    names_spec <- names_spec %||% "{.col}_{.fn}"
  } else if (is.list(funs)) {
    fns <- lapply(funs, across_fun, env, fn = fn)
    names_spec <- names_spec %||% "{.col}_{.fn}"
  } else if (!is.null(env) && !evaluated) {
    # Try evaluating once, just in case
    funs <- eval(funs, env)
    return(across_funs(
      funs,
      env,
      names_spec = NULL,
      fn = fn,
      evaluated = TRUE
    ))
  } else {
    cli_abort(
      "{.arg .fns} must be a function, a formula, or list of functions/formulas.",
      call = call2(fn, .ns = "dbplyr")
    )
  }

  list(fns = fns, names = names_spec)
}

across_fun <- function(fun, env, fn) {
  if (is_function(fun)) {
    fn_name <- find_fun(fun)
    if (!is_null(fn_name)) {
      return(\(x, cur_col) call2(fn_name, x))
    }
    partial_eval_fun(fun, env, fn)
  } else if (is_symbol(fun) || is_string(fun)) {
    \(x, cur_col) call2(fun, x)
  } else if (is_call(fun, "~")) {
    partial_eval_prepare_fun(f_rhs(fun), c(".", ".x", "..1"), env)
  } else if (is_call(fun, "function")) {
    fun <- eval(fun, env)
    partial_eval_fun(fun, env, fn)
  } else {
    cli_abort(
      c(
        "{.arg .fns} must contain a function or a formula.",
        x = "Problem with {expr_deparse(fun)}"
      ),
      call = call2(fn, .ns = "dbplyr")
    )
  }
}

partial_eval_fun <- function(fun, env, fn) {
  body <- fn_body(fun)
  if (length(body) > 2) {
    cli_abort(
      "Cannot translate functions consisting of more than one statement.",
      call = call2(fn, .ns = "dbplyr")
    )
  }
  args <- fn_fmls_names(fun)

  partial_eval_prepare_fun(body[[2]], args[[1]], fn_env(fun))
}

partial_eval_prepare_fun <- function(call, sym, env) {
  # First resolve any .data/.env pronouns before symbol replacement
  call <- resolve_mask_pronouns(call, env)
  call <- replace_sym1(call, sym, replace = quote(!!.x))
  call <- replace_call(call, replace = quote(!!.cur_col))
  function(x, .cur_col) {
    inject(
      expr(!!call),
      child_env(empty_env(), .x = x, expr = rlang::expr, .cur_col = .cur_col)
    )
  }
}

resolve_mask_pronouns <- function(call, env) {
  if (is_mask_pronoun(call)) {
    var <- call[[3]]

    if (is_symbol(call[[2]], ".data")) {
      if (is_call(call, "[[")) {
        sym(eval(var, env))
      } else {
        var
      }
    } else {
      if (is_call(call, "[[")) {
        env_get(env, var)
      } else {
        env_get(env, as.character(var))
      }
    }
  } else if (is_call(call)) {
    call[] <- lapply(call, resolve_mask_pronouns, env = env)
    call
  } else {
    call
  }
}

across_setup <- function(data, call, env, allow_rename, fn, error_call) {
  grps <- group_vars(data)
  tbl <- ungroup(data)
  tbl <- select(tbl, -all_of(grps))

  .cols <- call$.cols %||% expr(everything())
  locs <- tidyselect::eval_select(
    .cols,
    tbl,
    env = env,
    allow_rename = allow_rename,
    error_call = call(fn)
  )

  vars <- syms(colnames(tbl))[locs]
  if (allow_rename) {
    names_vars <- names(locs)
  } else {
    names_vars <- colnames(tbl)[locs]
  }

  names_spec <- eval(call$.names, env)
  funs_across_data <- across_funs(
    funs = call$.fns,
    env = env,
    names_spec = names_spec,
    fn = fn
  )

  fns_is_null <- funs_across_data$fns_is_null
  fns <- funs_across_data$fns
  names_spec <- funs_across_data$names

  # make sure fns has names, use number to replace unnamed
  names_fns <- names2(fns)
  empties <- which(names_fns == "")
  if (length(empties)) {
    names_fns[empties] <- empties
  }

  glue_mask <- across_glue_mask(
    env,
    .col = rep(names_vars, each = length(fns)),
    .fn = rep(names_fns, length(vars))
  )
  names_out <- vctrs::vec_as_names(
    glue(names_spec, .envir = glue_mask),
    repair = "check_unique"
  )

  out <- across_apply_fns(vars, fns, names_out, env)
  # Partially evaluate each expression (e.g., to evaluate sql() calls)
  lapply(out, partial_eval, data = data, env = env, error_call = error_call)
}

expr_as_label <- function(expr, name) {
  label <- as_label(expr)
  if (name != "") {
    label <- paste0(name, " = ", label)
  }

  label
}

across_apply_fns <- function(vars, fns, names, env) {
  out <- vector("list", length(vars) * length(fns))
  out <- set_names(out, names)
  k <- 1
  for (i in seq_along(vars)) {
    for (j in seq_along(fns)) {
      out[[k]] <- exec(fns[[j]], vars[[i]], as_name(vars[[i]]))
      k <- k + 1
    }
  }
  out
}

replace_call <- function(call, replace) {
  if (is_call(call)) {
    if (is_call(call, "cur_column")) {
      replace
    } else {
      call[] <- lapply(call, replace_call, replace = replace)
      call
    }
  } else {
    call
  }
}

across_glue_mask <- function(.col, .fn, .caller_env) {
  glue_mask <- env(.caller_env, .col = .col, .fn = .fn)
  # TODO: we can make these bindings louder later
  env_bind_active(
    glue_mask,
    col = \() glue_mask$.col,
    fn = \() glue_mask$.fn
  )
  glue_mask
}

# across() ----------------------------------------------------------------
# test partial_eval_across() indirectly via SQL generation

test_that("across() translates NULL", {
  lf <- lazy_frame(a = 1, b = 2)
  expect_equal(
    capture_across(lf, across(a:b)),
    list(a = expr(a), b = expr(b))
  )
})

test_that("across() drops groups", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_across(group_by(lf, a), across(everything())),
    list(b = expr(b))
  )
  expect_equal(
    capture_across(group_by(lf, b), across(everything())),
    list(a = expr(a))
  )
})


test_that("across() translates functions", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_across(lf, across(a:b, log)),
    exprs(a = log(a), b = log(b))
  )

  expect_equal(
    capture_across(lf, across(a, list(log, exp))),
    exprs(a_1 = log(a), a_2 = exp(a))
  )
})

test_that("across() translates functions in namespace #1231", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_across(lf, across(a:b, dplyr::dense_rank)),
    exprs(a = dense_rank(a), b = dense_rank(b))
  )
})

test_that("across() captures anonymous functions", {
  lf <- lazy_frame(a = 1)

  expect_equal(
    capture_across(lf, across(a, function(x) log(x))),
    list(a = expr(log(a)))
  )

  expect_snapshot(
    (expect_error(capture_across(
      lf,
      across(a, function(x) {
        x <- x + 2
        log(x)
      })
    )))
  )
})

test_that("across() translates formulas", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_across(lf, across(a:b, ~ log(.x))),
    exprs(a = log(a), b = log(b))
  )

  expect_equal(
    capture_across(lf, across(a:b, ~ log(..1))),
    exprs(a = log(a), b = log(b))
  )

  expect_equal(
    capture_across(lf, across(a:b, ~2)),
    exprs(a = 2, b = 2)
  )

  expect_equal(
    capture_across(lf, across(a:b, list(~ log(.x)))),
    exprs(a_1 = log(a), b_1 = log(b))
  )
})

test_that("across() translates evaluated functions", {
  lf <- lazy_frame(x = 1)

  expect_equal(
    capture_across(lf, across(.fns = !!sum)),
    exprs(x = sum(x))
  )

  expect_equal(
    capture_across(lf, across(.fns = !!function(x) sum(x + 2))),
    exprs(x = sum(x + 2))
  )
})

test_that("across() gives informative errors", {
  lf <- lazy_frame(a = 1, b = 2)
  expect_snapshot(error = TRUE, {
    capture_across(lf, across(a, 1))
    capture_across(lf, across(a, list(1)))

    capture_across(lf, across(a:b, "log"))

    capture_across(lf, across(c, mean))
  })
})

test_that("across() can use named selections", {
  lf <- lazy_frame(x = 1, y = 2)

  # no fns
  expect_equal(
    capture_across(lf, across(c(a = x, b = y))),
    list(a = quote(x), b = quote(y))
  )
  expect_equal(
    capture_across(lf, across(all_of(c(a = "x", b = "y")))),
    list(a = quote(x), b = quote(y))
  )

  # one fn
  expect_equal(
    capture_across(lf, across(c(a = x, b = y), mean)),
    list(a = quote(mean(x)), b = quote(mean(y)))
  )
  expect_equal(
    capture_across(lf, across(all_of(c(a = "x", b = "y")), mean)),
    list(a = quote(mean(x)), b = quote(mean(y)))
  )

  # multiple fns
  expect_equal(
    capture_across(lf, across(c(a = x, b = y), list(mean, nm = sum))),
    list(
      a_1 = quote(mean(x)),
      a_nm = quote(sum(x)),
      b_1 = quote(mean(y)),
      b_nm = quote(sum(y))
    )
  )
  expect_equal(
    capture_across(
      lf,
      across(all_of(c(a = "x", b = "y")), list(mean, nm = sum))
    ),
    list(
      a_1 = quote(mean(x)),
      a_nm = quote(sum(x)),
      b_1 = quote(mean(y)),
      b_nm = quote(sum(y))
    )
  )
})

test_that("across() correctly names output columns", {
  gf <- lazy_frame(x = 1, y = 2, z = 3, s = "") |> group_by(x)

  expect_equal(
    summarise(gf, across()) |> op_vars(),
    c("x", "y", "z", "s")
  )
  expect_equal(
    summarise(gf, across(.names = "id_{.col}")) |> op_vars(),
    c("x", "id_y", "id_z", "id_s")
  )
  expect_equal(
    summarise(gf, across(1:2, mean)) |> op_vars(),
    c("x", "y", "z")
  )
  expect_equal(
    summarise(gf, across(1:2, mean, .names = "mean_{.col}")) |> op_vars(),
    c("x", "mean_y", "mean_z")
  )
  expect_equal(
    summarise(gf, across(1:2, list(mean = mean, sum = sum))) |> op_vars(),
    c("x", "y_mean", "y_sum", "z_mean", "z_sum")
  )
  expect_equal(
    summarise(
      gf,
      across(1:2, list(mean = mean, sum = sum), .names = "{.fn}_{.col}")
    ) |>
      op_vars(),
    c("x", "mean_y", "sum_y", "mean_z", "sum_z")
  )

  expect_equal(
    summarise(gf, across(1:2, list(mean = mean, sum))) |> op_vars(),
    c("x", "y_mean", "y_2", "z_mean", "z_2")
  )
  expect_equal(
    summarise(gf, across(1:2, list(mean, sum = sum))) |> op_vars(),
    c("x", "y_1", "y_sum", "z_1", "z_sum")
  )
  expect_equal(
    summarise(gf, across(1:2, list(mean, sum))) |> op_vars(),
    c("x", "y_1", "y_2", "z_1", "z_2")
  )
})

test_that("across(.names=) can use local variables in addition to {col} and {fn}", {
  res <- local({
    prefix <- "MEAN"
    lazy_frame(x = 42) |>
      summarise(across(
        everything(),
        ~ mean(.x, na.rm = TRUE),
        .names = "{prefix}_{.col}"
      ))
  })
  expect_equal(op_vars(res), "MEAN_x")
})

test_that("across() .cols is evaluated in across()'s calling environment", {
  lf <- lazy_frame(y = 1)
  fun <- function(x) capture_across(lf, across(all_of(x)))
  expect_equal(
    fun("y"),
    list(y = expr(y))
  )
})

test_that("across() can handle empty selection", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_equal(
    lf |> mutate(across(character(), c)) |> remote_query(),
    sql("SELECT *\nFROM \"df\"")
  )
})

test_that("across() defaults to everything()", {
  # SELECT `x` + 1.0 AS `x`, `y` + 1.0 AS `y`
  expect_snapshot(
    lazy_frame(x = 1, y = 1) |> summarise(across(.fns = ~ . + 1))
  )
})

test_that("untranslatable functions are preserved", {
  lf <- lazy_frame(a = 1, b = 2)
  expect_snapshot(lf |> summarise(across(a:b, SQL_LOG)))
})

test_that("old _at functions continue to work", {
  withr::local_options(lifecycle_verbosity = "quiet")
  lf <- lazy_frame(a = 1, b = 2)

  expect_snapshot(lf |> dplyr::summarise_at(dplyr::vars(a:b), "sum"))
  expect_snapshot(lf |> dplyr::summarise_at(dplyr::vars(a:b), sum))
  expect_snapshot(lf |> dplyr::summarise_at(dplyr::vars(a:b), ~ sum(.)))
})

test_that("across() uses environment from the current quosure (dplyr#5460)", {
  # If the data frame `y` is selected, causes a subscript conversion
  # error since it is fractional
  lf <- lazy_frame(x = 1, y = 2.4)
  y <- "x"

  expect_equal(
    partial_eval_dots(lf, across(all_of(y), mean)),
    list(x = expr(mean(x)))
  )

  expect_equal(
    partial_eval_dots(lf, if_all(all_of(y), ~ .x < 2)),
    list(expr((x < 2))),
    ignore_attr = "names"
  )
})

test_that("lambdas in across() can use columns", {
  db <- local_memdb_frame("across", x = 2, y = 4, z = 8)

  expect_equal(
    partial_eval_dots(db, across(everything(), ~ .x / y)),
    list(
      x = expr(x / y),
      y = expr(y / y),
      z = expr(z / y)
    )
  )

  db_across <- db |> mutate(across(everything(), ~ .x / y))
  expect_snapshot(db_across |> show_query())

  # z should be 2 because the value of .data$y is only transformed
  # _after_ across() is complete, the same as
  # db |> collect() |> mutate(across(everything(), ~ .x / .data$y))
  expect_equal(collect(db_across), tibble(x = 0.5, y = 1, z = 2))
})

test_that("can use .data and .env pronouns(#1520)", {
  lf <- lazy_frame(x = 1, y = 2)

  my_col <- "y"
  expect_equal(
    capture_across(lf, across(x:y, !!quo(~ .x / .data$y))),
    exprs(x = x / y, y = y / y)
  )
  expect_equal(
    capture_across(lf, across(x:y, !!quo(~ .x / .data[[my_col]]))),
    exprs(x = x / y, y = y / y)
  )

  y <- 10
  expect_equal(
    capture_across(lf, across(x:y, !!quo(~ .x / .env$y))),
    exprs(x = x / 10, y = y / 10)
  )
  expect_equal(
    capture_across(lf, across(x:y, !!quo(~ .x / .env[["y"]]))),
    exprs(x = x / 10, y = y / 10)
  )
})

test_that("can pass quosure through `across()`", {
  summarise_mean <- function(data, vars) {
    data |> summarise(across({{ vars }}, ~ mean(.x, na.rm = TRUE)))
  }
  gdf <- lazy_frame(g = c(1, 1, 2), x = 1:3) |> group_by(g)

  expect_equal(
    gdf |> summarise_mean(x) |> remote_query(),
    summarise(gdf, x = mean(x, na.rm = TRUE)) |> remote_query()
  )
})

test_that("across() translates evaluated lists", {
  lf <- lazy_frame(x = 1)
  fun_list <- list(mean, ~ mean(.x + 1, na.rm = TRUE), function(x) mean(x + 2))

  expect_equal(
    capture_across(lf, across(.fns = !!fun_list)),
    exprs(
      x_1 = mean(x),
      x_2 = mean(x + 1, na.rm = TRUE),
      x_3 = mean(x + 2),
    )
  )
})

test_that("across() translates evaluated quosures", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_across(lf, across(a:b, !!quo(list(log)))),
    exprs(
      a_1 = log(a),
      b_1 = log(b)
    )
  )

  expect_equal(
    capture_across(lf, across(a:b, !!quo(~ log(.x, 2)))),
    exprs(
      a = log(a, 2),
      b = log(b, 2)
    )
  )

  expect_equal(
    capture_across(lf, across(a:b, !!quo(list(log2 = ~ log(.x, 2))))),
    exprs(
      a_log2 = log(a, 2),
      b_log2 = log(b, 2)
    )
  )
})

test_that("across() searches for list in environment", {
  lf <- lazy_frame(a = 1, b = 2)
  list_formula <- list(~ log(.x), sum)

  expect_equal(
    capture_across(lf, across(a:b, list_formula)),
    exprs(
      a_1 = log(a),
      a_2 = sum(a),
      b_1 = log(b),
      b_2 = sum(b)
    )
  )
})

test_that("across() handles cur_column()", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_across(lf, across(a:b, ~ paste0(.x, cur_column()))),
    exprs(a = paste0(a, "a"), b = paste0(b, "b"))
  )

  expect_equal(
    capture_across(lf, across(a:b, function(x) paste0(x, cur_column()))),
    exprs(a = paste0(a, "a"), b = paste0(b, "b"))
  )
})

test_that("across() errors if named", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_snapshot({
    (expect_error(mutate(lf, x = across())))
    (expect_error(group_by(lf, x = across())))
  })
})

test_that("across() throws error if unpack = TRUE", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_snapshot(
    (expect_error(lf |> mutate(across(x, .unpack = TRUE))))
  )
})

test_that("where() isn't suppored", {
  lf <- lazy_frame(x = 1)
  expect_snapshot(error = TRUE, {
    capture_across(lf, across(where(is.integer), as.character))
  })
})


# if_all ------------------------------------------------------------------

test_that("if_all() translates functions", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_if_all(lf, if_all(a:b, log)),
    expr(log(a) & log(b))
  )

  expect_equal(
    capture_if_all(lf, if_all(a, list(log, exp))),
    expr(log(a) & exp(a))
  )
})

test_that("if_all() translates formulas", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_if_all(lf, if_all(a:b, ~ log(.x))),
    expr(log(a) & log(b))
  )

  expect_equal(
    capture_if_all(lf, if_all(a:b, ~2)),
    expr(2 & 2)
  )

  expect_equal(
    capture_if_all(lf, if_all(a:b, list(~ log(.x)))),
    expr(log(a) & log(b))
  )
})

test_that("if_all() gives informative errors", {
  lf <- lazy_frame(a = 1, b = 2)
  expect_snapshot(error = TRUE, {
    capture_if_all(lf, if_all(a, 1))
    capture_if_all(lf, if_all(a, list(1)))
  })
})

test_that("if_all collapses multiple expressions", {
  lf <- lazy_frame(a = 1, b = 2)
  expect_equal(
    capture_if_all(lf, if_all(everything(), is.na)),
    expr(is.na(a) & is.na(b))
  )
})

test_that("if_all/any works in filter()", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_snapshot(lf |> filter(if_all(a:b, ~ . > 0)))
  expect_snapshot(lf |> filter(if_any(a:b, ~ . > 0)))
})

test_that("if_all/any is wrapped in parentheses #1153", {
  lf <- lazy_frame(a = 1, b = 2, c = 3)

  expect_equal(
    lf |> filter(if_any(c(a, b)) & c == 3) |> remote_query(),
    sql(
      "SELECT *\nFROM \"df\"\nWHERE ((\"a\" OR \"b\") AND \"c\" = 3.0)"
    )
  )
})

test_that("if_all/any works in mutate()", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_snapshot(lf |> mutate(c = if_all(a:b, ~ . > 0)))
  expect_snapshot(lf |> mutate(c = if_any(a:b, ~ . > 0)))
})

test_that("if_all/any uses every column as default", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_snapshot(lf |> filter(if_all(.fns = ~ . > 0)))
  expect_snapshot(lf |> filter(if_any(.fns = ~ . > 0)))
})

test_that("if_all/any works without `.fns` argument", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_snapshot(lf |> filter(if_all(a:b)))
  expect_snapshot(lf |> filter(if_any(a:b)))
})

test_that("if_all() drops groups", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_if_all(group_by(lf, a), if_all(everything())),
    sym("b")
  )
  expect_equal(
    capture_if_all(group_by(lf, b), if_all(everything())),
    sym("a")
  )
})

test_that("if_any() and if_all() expansions deal with single inputs", {
  d <- lazy_frame(x = 1)

  # Single inputs
  expect_equal(
    filter(d, if_any(x, ~FALSE)) |> remote_query(),
    sql("SELECT *\nFROM \"df\"\nWHERE ((FALSE))")
  )
  expect_equal(
    filter(d, if_all(x, ~FALSE)) |> remote_query(),
    sql("SELECT *\nFROM \"df\"\nWHERE ((FALSE))")
  )
})

test_that("if_all() cannot rename variables", {
  lf <- lazy_frame(x = 1, y = 2)

  # no fns
  expect_snapshot(
    (expect_error(capture_if_all(lf, if_all(c(a = x, b = y)))))
  )
})

test_that("if_all() can handle empty selection", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_equal(
    capture_if_all(lf, if_all(character(), c)),
    TRUE
  )
})

test_that("across() .cols is evaluated in across()'s calling environment", {
  lf <- lazy_frame(y = 1)
  fun <- function(x) capture_if_all(lf, if_all(all_of(x)))
  expect_equal(
    fun("y"),
    expr(y)
  )
})


# dots --------------------------------------------------------------------

test_that("across(...) is deprecated", {
  lf <- lazy_frame(x = c(1, NA))
  expect_snapshot(
    error = TRUE,
    summarise(lf, across(everything(), mean, na.rm = TRUE))
  )
})

test_that("if_all() translates functions", {
  lf <- lazy_frame(a = 1, b = 2)

  expect_equal(
    capture_if_all(lf, if_all(a:b, log)),
    expr(log(a) & log(b))
  )

  expect_equal(
    capture_if_all(lf, if_all(a, list(log, exp))),
    expr(log(a) & exp(a))
  )
})


# pick --------------------------------------------------------------------

# pick() + arrange()

test_that("can `arrange()` with `pick()` selection", {
  df <- lazy_frame(x = c(2, 2, 1), y = c(3, 1, 3))

  expect_identical(
    arrange(df, pick(x, y)) |> remote_query(),
    sql("SELECT *\nFROM \"df\"\nORDER BY \"x\", \"y\"")
  )

  expect_identical(
    arrange(df, pick(x), y) |> remote_query(),
    sql("SELECT *\nFROM \"df\"\nORDER BY \"x\", \"y\"")
  )
})

test_that("`pick()` errors in `arrange()` are useful", {
  df <- lazy_frame(x = 1)

  expect_snapshot(error = TRUE, {
    arrange(df, pick(y))
  })
})

test_that("doesn't allow renaming", {
  expect_snapshot(error = TRUE, {
    arrange(lazy_frame(x = 1), pick(y = x))
  })
})

test_that("requires at least one input", {
  expect_snapshot(error = TRUE, {
    arrange(lazy_frame(x = 1), pick())
  })
})

# pick() + filter()

test_that("`filter()` with `pick()` that uses invalid tidy-selection errors", {
  df <- lazy_frame(x = c(1, 2, NA, 3), y = c(2, NA, 5, 3))

  expect_snapshot(error = TRUE, {
    filter(df, pick(x, a))
  })
})

# pick() + group_by()

test_that("`pick()` can be used inside `group_by()` wrappers", {
  df <- lazy_frame(a = 1:3, b = 2:4, c = 3:5)

  tidyselect_group_by <- function(data, groups) {
    group_by(data, pick({{ groups }}))
  }
  expect_identical(
    tidyselect_group_by(df, c(a, c)),
    group_by(df, a, c)
  )
})

#' SQL dialects
#'
#' @description
#' The dialect system allows multiple database connection classes to share
#' SQL generation code. A dialect object encapsulates the SQL syntax rules
#' for a particular database, independent of the connection mechanism.
#'
#' * `sql_dialect()` returns the dialect for a connection. For connections
#'   that haven't implemented a dialect method, returns the connection itself
#'   for backward compatibility.
#'
#' * `new_sql_dialect()` creates a new dialect object. This is primarily
#'   intended for dbplyr backend authors.
#'
#' # Dispatching on dialect
#'
#' For backward compatibility, all `sql_` generics (and a handful of others)
#' call `sql_dialect()` on the `con` argument in order to dispatch further on
#' the dialect object, if possible:
#'
#' ```R
#' sql_generic <- function(con, arg1, arg2, ...) {
#'   UseMethod("sql_generic", sql_dialect(con))
#' }
#' ```
#'
#' Unfortunately, due to the way that `UseMethod()` works, this uses
#' `sql_dialect(con)` to control which method is selected, but still passes
#' the original `con` to the method. This means that if you are implementing
#' a method for a dialect and need to access dialect properties, you must call
#' `sql_dialect(con)` again inside the method.
#'
#' @param con A database connection.
#' @param dialect A string giving the dialect name (e.g., "postgres", "mysql").
#' @param quote_identifier A function that quotes identifiers. Should accept
#'   a character vector and return a [sql] vector.
#' @param has_window_clause Does the backend support named window
#'   definitions (the `WINDOW` clause)?
#' @param has_table_alias_with_as Does the backend support using `AS`
#'   when aliasing a table in a subquery?
#' @param has_star_table_prefix Does the backend require table prefixes when
#'   selecting all columns in single-table queries (e.g., `"table".*` vs `*`)?
#'
#' @return
#' * `sql_dialect()` returns a dialect object (class `sql_dialect`) or
#'   the connection itself for backward compatibility.
#' * `new_sql_dialect()` returns a dialect object with class
#'   `c("sql_dialect_{name}", "sql_dialect")`.
#'
#' @export
#' @examples
#' # Create a custom dialect
#' my_dialect <- new_sql_dialect(
#'   "custom",
#'   quote_identifier = function(x) sql_quote(x, "`"),
#'   has_window_clause = TRUE
#' )
#' class(my_dialect)
sql_dialect <- function(con) {
  # Check for dialect override from with_dialect()
  override <- attr(con, "dbplyr_dialect", exact = TRUE)
  if (!is.null(override)) {
    return(override)
  }
  UseMethod("sql_dialect")
}

#' @export
sql_dialect.sql_dialect <- function(con) {
  con
}

#' @export
sql_dialect.default <- function(con) {
  # For backward compatibility: connections without a dialect method
  # pass through unchanged
  con
}

#' @export
#' @rdname sql_dialect
new_sql_dialect <- function(
  dialect,
  quote_identifier,
  has_window_clause = FALSE,
  has_table_alias_with_as = TRUE,
  has_star_table_prefix = FALSE
) {
  check_string(dialect)
  check_function(quote_identifier)
  check_bool(has_window_clause)
  check_bool(has_table_alias_with_as)
  check_bool(has_star_table_prefix)

  structure(
    list(
      quote_identifier = quote_identifier,
      has = list(
        window_clause = has_window_clause,
        table_alias_with_as = has_table_alias_with_as,
        star_table_prefix = has_star_table_prefix
      )
    ),
    class = c(paste0("sql_dialect_", dialect), "sql_dialect")
  )
}

#' @export
sql_translation.sql_dialect <- function(con) {
  sql_variant(
    base_scalar,
    base_agg,
    base_win
  )
}

#' @export
print.sql_dialect <- function(x, ...) {
  cat("<", class(x)[[1]], ">\n", sep = "")
  invisible(x)
}

is_sql_dialect <- function(x) {
  inherits(x, "sql_dialect")
}

sql_has_table_alias_with_as <- function(con) {
  dialect <- sql_dialect(con)

  if (is_sql_dialect(dialect)) {
    dialect$has$table_alias_with_as
  } else {
    # For backward comptaibility
    TRUE
  }
}

sql_has_window_clause <- function(con) {
  dialect <- sql_dialect(con)

  if (is_sql_dialect(dialect)) {
    dialect$has$window_clause
  } else {
    # For backward comptaibility
    FALSE
  }
}

sql_has_star_table_prefix <- function(con) {
  dialect <- sql_dialect(con)

  if (is_sql_dialect(dialect)) {
    dialect$has$star_table_prefix %||% FALSE
  } else {
    # For backward compatibility
    FALSE
  }
}

test_that("new_sql_dialect() creates subclass", {
  d <- new_sql_dialect(
    "test",
    quote_identifier = function(x) sql_quote(x, "`")
  )
  expect_s3_class(d, c("sql_dialect_test", "sql_dialect"))
})

test_that("new_sql_dialect() validates inputs", {
  expect_snapshot(error = TRUE, {
    new_sql_dialect(1)
    new_sql_dialect("x", quote_identifier = "a function")
    new_sql_dialect("x", quote_identifier = \(x) x, has_window_clause = "yes")
  })
})

test_that("sql_dialect() is idempotent for dialects and connections", {
  d <- new_sql_dialect("test", quote_identifier = identity)
  expect_identical(sql_dialect(d), d)

  # for backward compatibility
  con <- dialect_ansi()
  expect_identical(sql_dialect(con), con)
})

test_that("sql_has_window_clause() returns correct values", {
  d1 <- new_sql_dialect(
    "test",
    quote_identifier = identity,
    has_window_clause = FALSE
  )
  d2 <- new_sql_dialect(
    "test",
    quote_identifier = identity,
    has_window_clause = TRUE
  )

  expect_false(sql_has_window_clause(d1))
  expect_true(sql_has_window_clause(d2))

  # DBIConnection defaults to FALSE
  expect_false(sql_has_window_clause(dialect_ansi()))
})

test_that("sql_has_table_alias_with_as() returns correct values", {
  d1 <- new_sql_dialect(
    "test",
    quote_identifier = identity,
    has_table_alias_with_as = TRUE
  )
  d2 <- new_sql_dialect(
    "test",
    quote_identifier = identity,
    has_table_alias_with_as = FALSE
  )

  expect_true(sql_has_table_alias_with_as(d1))
  expect_false(sql_has_table_alias_with_as(d2))

  # DBIConnection defaults to TRUE
  expect_true(sql_has_table_alias_with_as(dialect_ansi()))
})

test_that("sql_has_star_table_prefix() returns correct values", {
  d1 <- new_sql_dialect(
    "test",
    quote_identifier = identity,
    has_star_table_prefix = FALSE
  )
  d2 <- new_sql_dialect(
    "test",
    quote_identifier = identity,
    has_star_table_prefix = TRUE
  )

  expect_false(sql_has_star_table_prefix(d1))
  expect_true(sql_has_star_table_prefix(d2))

  # DBIConnection defaults to FALSE
  expect_false(sql_has_star_table_prefix(dialect_ansi()))
})

test_that("print method shows class", {
  d <- new_sql_dialect("test", quote_identifier = identity)
  expect_snapshot(d)
})

#' Use a local data frame in a dbplyr query
#'
#' This is an alternative to [copy_to()] that does not need write access and
#' is faster for small data.
#'
#' It writes the data directly in the SQL query via the `VALUES` clause.
#'
#' @seealso [copy_to()] to copy the data into a new database table.
#' @export
#' @param con A database connection.
#' @param df A local data frame. The data is written directly in the SQL query
#'   so it should be small.
#' @param types A named character vector of SQL data types to use for the columns.
#'    The data types are backend specific. For example for Postgres this could
#'    be `c(id = "bigint", created_at = "timestamp", values = "integer[]")`.
#'    If `NULL`, the default, the types are determined from `df`.
#' @return A `tbl_lazy`.
#'
#' @examples
#' df <- data.frame(x = 1:3, y = c("a", "b", "c"))
#' con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#'
#' copy_inline(con, df)
#'
#' copy_inline(con, df) |> dplyr::show_query()
copy_inline <- function(con, df, types = NULL) {
  if (!inherits(df, "data.frame")) {
    cli_abort("{.var df} needs to be a data.frame.")
  }

  if (ncol(df) == 0) {
    cli_abort("{.var df} needs at least one column.")
  }

  if (!is_null(types)) {
    check_character(types)

    if (!setequal(colnames(df), names(types))) {
      cli_abort("Names of {.arg df} and {.arg types} must be the same.")
    }
  }

  lazy_query <- lazy_values_query(df, types)
  new_tbl_lazy(con, lazy_query)
}

lazy_values_query <- function(df, types) {
  lazy_base_query(
    class = "values",
    x = df,
    vars = colnames(df),
    col_types = types
  )
}

#' @export
sql_build.lazy_base_values_query <- function(op, con, ..., sql_options = NULL) {
  class(op) <- c("values_query", "query")
  op
}

#' @export
sql_render.values_query <- function(
  query,
  con = query$src$con,
  ...,
  sql_options = NULL,
  subquery = FALSE,
  lvl = 0
) {
  sql_values_subquery(con, query$x, types = query$col_types, lvl = lvl)
}

#' @export
flatten_query.values_query <- function(qry, query_list, con) {
  querylist_reuse_query(qry, query_list, con)
}

#' @export
op_vars.lazy_base_values_query <- function(op) {
  colnames(op$x)
}

sql_values_subquery <- function(con, df, types, lvl = 0, ...) {
  check_dots_used()
  UseMethod("sql_values_subquery", sql_dialect(con))
}

#' @export
sql_values_subquery.DBIConnection <- function(con, df, types, lvl = 0, ...) {
  sql_values_subquery_default(con, df, types = types, lvl = lvl, row = FALSE)
}

#' @export
sql_values_subquery.sql_dialect <- sql_values_subquery.DBIConnection

sql_values_subquery_default <- function(con, df, types, lvl, row) {
  df <- values_prepare(con, df)
  if (nrow(df) == 0L) {
    return(sql_values_zero_rows(con, df, types, lvl))
  }

  # The query consists of two parts:
  # 1) An outer select which converts the values to the correct types. This needs
  # to use the translation of `as.<column type>(<column name>)` (e.g. `as.numeric(mpg)`)
  # because some backends need a special translation for some types e.g. casting
  # to logical/bool in MySQL
  #   `IF(<column name>, TRUE, FALSE)`
  # This is done with the help of `sql_cast_dispatch()` via dispatch on the
  # column type. The explicit cast is required so that joins work e.g. on date
  # columns in Postgres.
  # 2) A subquery which is the union of:
  #   a) a zero row table which is just required to name the columns. This is
  #      necessary as e.g. SQLite cannot name `VALUES`.
  #   b) `VALUES` clause
  cols_clause <- names_to_as(con, rep_named(colnames(df), "NULL"))

  null_row_clauses <- list(
    select = sql_clause_select(cols_clause),
    where = sql_clause_where(sql("0 = 1"))
  )

  rows_clauses <- sql_values_clause(con, df, row = row)
  rows_query <- sql_format_clauses(rows_clauses, lvl = lvl + 1)

  subquery <- sql_query_union(
    con,
    x = sql_format_clauses(null_row_clauses, lvl + 1),
    unions = list(table = as.character(rows_query), all = TRUE),
    lvl = lvl + 1
  )

  sql_query_select(
    con,
    select = sql_values_cast_clauses(con, df, types, na = FALSE),
    from = sql_query_wrap(con, subquery, name = "values_table", lvl = lvl),
    lvl = lvl
  )
}

sql_values_subquery_column_alias <- function(con, df, types, lvl, ...) {
  df <- values_prepare(con, df)
  if (nrow(df) == 0L) {
    return(sql_values_zero_rows(con, df, types, lvl))
  }

  # The `SELECT` clause converts the values to the correct types. This needs
  # to use the translation of `as.<column type>(<column name>)` (e.g. `as.numeric(mpg)`)
  # because some backends need a special translation for some types e.g. casting
  # to logical/bool in MySQL
  #   `IF(<column name>, TRUE, FALSE)`
  # This is done with the help of `sql_cast_dispatch()` via dispatch on the
  # column type. The explicit cast is required so that joins work e.g. on date
  # columns in Postgres.
  # The `FROM` clause is simply the `VALUES` clause with table and column alias
  rows_clauses <- sql_values_clause(con, df, row = FALSE)
  rows_query <- sql_format_clauses(rows_clauses, lvl = lvl + 1)

  table_alias_sql <- sql_glue2(con, "drvd({.id colnames(df)})")

  if (grepl("\\n", rows_query)) {
    rows_query <- sql(paste0(
      "(\n",
      rows_query,
      "\n",
      indent_lvl(") AS ", lvl),
      table_alias_sql
    ))
  } else {
    # indent is not perfect but okay
    rows_query <- sql(paste0("(", rows_query, ") AS ", table_alias_sql))
  }

  sql_query_select(
    con,
    select = sql_values_cast_clauses(con, df, types, na = FALSE),
    from = rows_query,
    lvl = lvl
  )
}

sql_values_subquery_union <- function(con, df, types, lvl, row, from = NULL) {
  df <- values_prepare(con, df)
  if (nrow(df) == 0L) {
    return(sql_values_zero_rows(con, df, types, lvl, from))
  }

  # The query consists of two parts:
  # 1) An outer select which converts the values to the correct types. This needs
  # to use the translation of `as.<column type>(<column name>)` (e.g. `as.numeric(mpg)`)
  # because some backends need a special translation for some types e.g. casting
  # to logical/bool in MySQL
  #   `IF(<column name>, TRUE, FALSE)`
  # This is done with the help of `sql_cast_dispatch()` via dispatch on the
  # column type. The explicit cast is required so that joins work e.g. on date
  # columns in Postgres.
  # 2) A subquery which is the union of:
  #   a) a zero row table which is just required to name the columns. This is
  #      necessary as e.g. SQLite cannot name `VALUES`.
  #   b) `UNION ALL` of one row `SELECT` statements
  cols_clause <- names_to_as(con, rep_named(colnames(df), "NULL"))

  clauses <- list(
    select = sql_clause_select(cols_clause),
    from = if (!is.null(from)) sql_clause_from(sql_escape_ident(con, from)),
    where = sql_clause_where(sql("0 = 1"))
  )
  null_row_query <- sql_format_clauses(clauses, lvl + 1)

  escaped_values <- purrr::map(
    df,
    escape,
    con = con,
    collapse = NULL,
    parens = FALSE
  )

  rows <- rlang::exec(paste, !!!escaped_values, sep = ", ")
  select_kw <- style_kw("SELECT ")
  tables <- paste0(lvl_indent(lvl + 1), select_kw, rows)
  if (!is_null(from)) {
    from_kw <- style_kw("FROM ")
    tables <- paste0(tables, " ", from_kw, from)
  }

  subquery <- sql_query_union(
    con,
    x = null_row_query,
    unions = list(all = TRUE, table = tables),
    lvl = lvl + 1
  )

  sql_query_select(
    con,
    select = sql_values_cast_clauses(con, df, types, na = FALSE),
    from = sql_query_wrap(con, subquery, name = "values_table", lvl = lvl),
    lvl = lvl
  )
}

sql_values_clause <- function(con, df, row = FALSE) {
  escaped_values <- purrr::map(
    df,
    escape,
    con = con,
    collapse = NULL,
    parens = FALSE
  )
  rows <- rlang::exec(paste, !!!escaped_values, sep = ", ")
  rows_sql <- sql(paste0(if (row) "ROW", "(", rows, ")"))

  list(sql_clause("VALUES", rows_sql))
}

sql_values_zero_rows <- function(con, df, types, lvl, from = NULL) {
  if (nrow(df) != 0L) {
    cli_abort("{.arg df} does not have 0 rows", .internal = TRUE)
  }

  typed_cols <- sql_values_cast_clauses(con, df, types, na = TRUE)

  clauses <- list(
    select = sql_clause_select(typed_cols),
    from = if (!is.null(from)) sql_clause_from(sql_escape_ident(con, from)),
    where = sql_clause_where(sql("0 = 1"))
  )
  sql_format_clauses(clauses, lvl)
}

sql_values_cast_clauses <- function(con, df, types, na) {
  if (is_null(types)) {
    typed_cols <- purrr::map2_chr(
      df,
      colnames(df),
      ~ {
        val <- if (na) NA else ident(.y)
        cast_expr <- call2(sql_cast_dispatch(.x), val)
        translate_sql(!!cast_expr, con = con)
      }
    )
  } else {
    typed_cols <- purrr::imap_chr(types, function(type, val) {
      val <- if (na) NA else ident(val)
      sql_glue2(con, "CAST({val} AS {.sql type})")
    })
  }

  names_to_as(con, typed_cols)
}

values_prepare <- function(con, df) {
  dialect <- sql_dialect(con)
  return(values_prepare_(dialect, df))

  UseMethod("values_prepare")
}
values_prepare_ <- function(dialect, df) {
  UseMethod("values_prepare")
}

#' @export
values_prepare.DBIConnection <- function(con, df) {
  df
}

#' @export
values_prepare.sql_dialect <- values_prepare.DBIConnection

# This
sql_cast_dispatch <- function(x) {
  UseMethod("sql_cast_dispatch")
}

#' @export
sql_cast_dispatch.sql <- function(x) {
  expr(as.character)
}

#' @export
sql_cast_dispatch.logical <- function(x) {
  expr(as.logical)
}

#' @export
sql_cast_dispatch.integer <- function(x) {
  expr(as.integer)
}

#' @export
sql_cast_dispatch.numeric <- function(x) {
  expr(as.numeric)
}

#' @export
sql_cast_dispatch.character <- function(x) {
  expr(as.character)
}

#' @export
sql_cast_dispatch.factor <- function(x) {
  expr(as.character)
}

#' @export
sql_cast_dispatch.Date <- function(x) {
  expr(as.Date)
}

#' @export
sql_cast_dispatch.POSIXct <- function(x) {
  expr(as.POSIXct)
}

#' @export
sql_cast_dispatch.integer64 <- function(x) {
  expr(as.integer64)
}

#' @export
sql_cast_dispatch.blob <- function(x) {
  expr(as.blob)
}

#' @importFrom dplyr coalesce
NULL
utils::globalVariables(c("as.integer64", "as.blob"))

#' Flag a character vector as SQL identifiers
#'
#' @description
#' `ident()` marks strings as database identifiers (e.g. table or column names)
#' quoting them using the identifier rules for your database. It is used
#' primarily in [translate_sql()] to label variables as identifiers; use
#' elsewhere should be regarded with suspicion.
#'
#' `ident()` is for internal use only; if you need to supply an table name that
#' is qualified with schema or catalog use `I()`.
#'
#' @param ... A character vector, or name-value pairs.
#' @param x An object.
#' @keywords internal
#' @export
#' @examples
#' con <- dialect_ansi()
#'
#' # SQL92 quotes strings with '
#' escape("x", con = con)
#'
#' # And identifiers with "
#' escape(ident("x"), con = con)
ident <- function(...) {
  x <- c_character(...)
  structure(x, class = c("ident", "character"))
}

#' @export
print.ident <- function(x, ...) cat(format(x, ...), sep = "\n")
#' @export
format.ident <- function(x, ...) {
  if (length(x) == 0) {
    paste0("<IDENT> [empty]")
  } else {
    paste0("<IDENT> ", x)
  }
}

#' @rdname ident
#' @export
is.ident <- function(x) inherits(x, "ident")

# Needed in in_catalog() and in_schema() for backward compatibility since
# those functions recommended that you use sql() to escape table names.
# Also needed in win_over() since the user might be generating their own
# syntax
as_ident_or_sql <- function(x, call = caller_env()) {
  if (is.sql(x)) {
    x
  } else if (is.ident(x)) {
    x
  } else if (is.character(x)) {
    ident(x)
  } else {
    cli::cli_abort(
      "Invalid identifier: expecting a character vector.",
      call = call
    )
  }
}

test_that("zero length inputs return correct clases", {
  expect_s3_class(ident(), "ident")
})

test_that("ident quotes", {
  con <- dialect_ansi()
  x1 <- ident("x")

  expect_equal(escape(x1, con = con), sql('"x"'))
})

test_that("can format ident", {
  expect_snapshot(ident())
})

test_that("as_ident_or_sql() accepts character subclasses like glue", {
  expect_equal(as_ident_or_sql(glue::glue("x")), ident("x"))
})

test_that("as_ident_or_sql() gives informative error for non-character", {
  expect_snapshot(as_ident_or_sql(1), error = TRUE)
})

#' SAP HANA backend
#'
#' @description
#' This backend supports SAP HANA databases, typically accessed via
#' `HDBConnection` created by [DBI::dbConnect()]. Use `dialect_hana()` with
#' `lazy_frame()` to see simulated SQL without connecting to a live database.
#'
#' Key differences for this backend are:
#'
#' * Temporary tables get `#` prefix and use `LOCAL TEMPORARY COLUMN`.
#' * No table analysis performed in [copy_to()].
#' * `paste()` uses `||`
#' * Note that you can't create new boolean columns from logical expressions;
#'   you need to wrap with explicit `ifelse`: `ifelse(x > y, TRUE, FALSE)`.
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-hana
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_hana())
#' lf |> transmute(x = paste0(d, " times"))
NULL

#' @export
#' @rdname backend-hana
dialect_hana <- function() {
  new_sql_dialect(
    "hana",
    quote_identifier = function(x) sql_quote(x, '"')
  )
}

#' @export
#' @rdname backend-hana
simulate_hana <- function() simulate_dbi("HDB")

#' @export
sql_dialect.HDB <- function(con) {
  dialect_hana()
}

#' @export
dbplyr_edition.HDB <- function(con) {
  2L
}

#' @export
sql_translation.sql_dialect_hana <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_scalar,
      as.character = sql_cast("VARCHAR"),
      as.numeric = sql_cast("DOUBLE"),
      as.double = sql_cast("DOUBLE"),

      # string functions ------------------------------------------------
      paste = sql_paste_infix(" ", "||"),
      paste0 = sql_paste_infix("", "||"),
      str_c = sql_paste_infix("", "||"),

      # https://help.sap.com/viewer/7c78579ce9b14a669c1f3295b0d8ca16/Cloud/en-US/20e8341275191014a4cfdcd3c830fc98.html
      substr = sql_substr("SUBSTRING"),
      substring = sql_substr("SUBSTRING"),
      str_sub = sql_str_sub("SUBSTRING"),
    ),
    base_agg,
    base_win
  )
}

# nocov start
#' @export
sql_table_temporary.sql_dialect_hana <- function(con, table, temporary, ...) {
  list(
    table = add_temporary_prefix(con, table, temporary = temporary),
    temporary = FALSE
  )
}
# nocov end

#' @export
sql_table_analyze.sql_dialect_hana <- function(con, table, ...) {
  # CREATE STATISTICS doesn't work for temporary tables, so
  # don't do anything at all
}

#' @export
sql_values_subquery.sql_dialect_hana <- function(con, df, types, lvl = 0, ...) {
  sql_values_subquery_union(con, df, types = types, lvl = lvl, from = "DUMMY")
}

test_that("simulate_hana() still works", {
  expect_translation(simulate_hana(), x + 1, '"x" + 1.0')
})

test_that("custom string translations", {
  con <- dialect_hana()

  expect_snapshot(translate_sql(paste0("a", "b"), con = con))
  expect_snapshot(translate_sql(paste("a", "b"), con = con))

  expect_snapshot(translate_sql(substr(x, 2, 4), con = con))
  expect_snapshot(translate_sql(substring(x, 2, 4), con = con))
  expect_snapshot(translate_sql(str_sub(x, 2, -2), con = con))
})

test_that("copy_inline uses UNION ALL", {
  con <- dialect_hana()
  y <- tibble::tibble(id = 1L, arr = "{1,2,3}")

  types <- c(id = "bigint", arr = "integer[]")
  expect_snapshot({
    copy_inline(con, y |> slice(0)) |> remote_query()
    copy_inline(con, y) |> remote_query()

    # with `types`
    copy_inline(con, y |> slice(0), types = types) |> remote_query()
    copy_inline(con, y, types = types) |> remote_query()
  })
})

#' A temporary in-memory database
#'
#' @description
#' `memdb()` creates a temporary in-memory database that disappears when the
#' R session ends. It's a convenient way to learn about and experiment with
#' dbplyr without having to connect to a "real" database.
#'
#' `memdb_frame()` works like [tibble::tibble()], but instead of creating a new
#' data frame in R, it creates a table in `memdb()`. `local_memdb_frame()`
#' is like `memdb_frame()` but the table will be automatically deleted when
#' the current scope ends. It's useful for tests. But beware: this function
#' will overwrite an existing table of the same name.
#'
#' @inheritParams tibble::tibble
#' @param .name Name of table in database: defaults to a random name that's
#'   unlikely to conflict with an existing table.
#' @export
#' @examples
#' library(dplyr)
#'
#' # use memdb_frame() to create a new database table
#' df <- memdb_frame(x = runif(100), y = runif(100))
#' df |> arrange(x)
#' df |> arrange(x) |> show_query()
#'
#' # Use memdb() + copy_to() to copy an existing data frame
#' iris_db <- copy_to(memdb(), iris)
#' iris_db
memdb <- function() {
  cache_computation("memdb", DBI::dbConnect(RSQLite::SQLite(), ":memory:"))
}

#' @rdname memdb
#' @export
memdb_frame <- function(.name = unique_table_name(), ...) {
  if (is.data.frame(.name)) {
    lifecycle::deprecate_warn(
      "2.6.0",
      I("memdb_frame(data.frame(...))"),
      details = "Use `copy_to(memdb(), df)` instead."
    )
    df <- .name
    .name <- unique_table_name()
  } else {
    check_string(.name)
    df <- tibble::tibble(...)
  }

  copy_to(memdb(), df, name = .name)
}

#' @rdname memdb
#' @param frame The created table is bound to this execution frame and will
#'   be deleted when it ends. For expert use only.
#' @export
local_memdb_frame <- function(
  .name = unique_table_name(),
  ...,
  frame = caller_env()
) {
  tbl <- copy_to(memdb(), tibble(...), .name, overwrite = TRUE)
  withr::defer(DBI::dbRemoveTable(memdb(), .name), envir = frame)
  tbl
}

#' Deprecated
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' * `src_memdb()` is deprecated; use `memdb()` instead.
#' * `tbl_memdb(df)` is deprecated; use `copy_to(memdb(), df)` instead.
#'
#' @param df Data frame to copy.
#' @param name Name of table in database.
#' @export
#' @keywords internal
src_memdb <- function() {
  lifecycle::deprecate_warn("2.6.0", "src_memdb()", "memdb()")
  src_dbi(memdb())
}

#' @rdname src_memdb
#' @export
tbl_memdb <- function(df, name = deparse(substitute(df))) {
  lifecycle::deprecate_warn(
    "2.6.0",
    "tbl_memdb()",
    details = "Use `copy_to(memdb(), df)` instead"
  )
  copy_to(memdb(), df, name = name)
}

test_that("default extracts last var from data frame", {
  df <- memdb_frame(x = 1:3, z = 4:6)
  expect_equal(pull(df), 4:6)
})

test_that("can extract by name, or positive/negative position", {
  x <- 1:3
  df <- memdb_frame(x = x, y = 4:6)

  expect_equal(pull(df, x), x)
  expect_equal(pull(df, 1L), x)
  expect_equal(pull(df, 1), x)
  expect_equal(pull(df, -2), x)
  expect_equal(pull(df, -2L), x)
})

test_that("extracts correct column from grouped tbl", {
  mf <- memdb_frame(id = "a", value = 42)
  gf <- mf %>% group_by(id)

  expect_equal(pull(mf, value), 42)
})

test_that("doesn't unnecessarily select", {
  mf <- memdb_frame(x = c(3, 1, 2))
  # no warning about select after arrange
  expect_warning(out <- mf %>% arrange(x) %>% pull(), NA)
  expect_equal(out, 1:3)
})

test_that("can extract named vectors", {
  x <- 1:3
  y <- letters[x]
  df <- memdb_frame(x = x, y = y)
  xn <- set_names(x, y)

  expect_equal(pull(df, x), x)
  expect_equal(pull(df, x, y), xn)
  expect_equal(pull(df, 1, 2), xn)
  expect_equal(names(pull(df, x, y)), y)
})


test_that("ungroup() produces nice error messages", {
  expect_snapshot(error = TRUE, {
    memdb_frame(x = 1) %>% pull(non_existent)
    memdb_frame(x = 1) %>% pull("non_existent")
    memdb_frame(x = 1) %>% pull(1000)

    memdb_frame(x = 1) %>% pull(x, "name_non_existent")
  })
})

test_that("custom header", {
  expect_snapshot({
    "Number of rows is shown"
    x <- memdb_frame(a = 1:3) %>% filter(a > 0)
    setup <- pillar::tbl_format_setup(x)
    tbl_format_header(x, setup)[[1]]

    "Number of rows still can't be shown"
    x <- memdb_frame(a = 1:11) %>% filter(a > 0)
    setup <- pillar::tbl_format_setup(x)
    tbl_format_header(x, setup)[[1]]
  })
})

test_that("tbl_sql() works with string argument", {
  db <- local_memdb_frame("db", a = 1)
  expect_equal(collect(tbl_sql("sqlite", src_dbi(memdb()), "db")), collect(db))
})

test_that("tbl_sql() respects subclass argument", {
  x <- tbl_sql("foo", src_dbi(memdb()), "abc", vars = letters)
  expect_s3_class(x, "tbl_foo")

  # It always includes tbl_sql and dbi
  expect_s3_class(x, "tbl_dbi")
  expect_s3_class(x, "tbl_sql")
})

test_that("same_src distinguishes srcs", {
  con1 <- local_sqlite_connection()
  con2 <- local_sqlite_connection()

  db1 <- copy_to(con1, iris[1:3, ], 'data1', temporary = FALSE)
  db2 <- copy_to(con2, iris[1:3, ], 'data2', temporary = FALSE)
  expect_true(same_src(db1, db1))
  expect_false(same_src(db1, db2))

  expect_false(same_src(db1, mtcars))
})

test_that("has nice print method", {
  mf <- local_memdb_frame("tbl_sum_test", x = 1, y = 1)
  expect_snapshot(mf, transform = scrub_sqlite_version)

  out2 <- mf |> group_by(x, y) |> arrange(x) |> mutate(z = x + y)
  expect_snapshot(out2, transform = scrub_sqlite_version)
})


# tbl ---------------------------------------------------------------------

test_that("can generate sql tbls with raw sql", {
  mf1 <- local_memdb_frame(x = 1:3, y = 3:1)
  mf2 <- tbl(mf1$src, sql(glue("SELECT * FROM {remote_name(mf1)}")))

  expect_equal(collect(mf1), collect(mf2))
})

test_that("can refer to default schema explicitly", {
  con <- local_sqlite_con_with_aux()
  DBI::dbExecute(con, "CREATE TABLE t1 (x)")

  expect_equal(as.character(tbl_vars(tbl(con, "t1"))), "x")
  expect_equal(as.character(tbl_vars(tbl(con, in_schema("main", "t1")))), "x")
})


test_that("can distinguish 'schema.table' from 'schema'.'table'", {
  con <- local_sqlite_con_with_aux()
  DBI::dbExecute(con, "CREATE TABLE aux.t1 (x, y, z)")
  DBI::dbExecute(con, "CREATE TABLE 'aux.t1' (a, b, c)")

  expect_equal(
    as.character(tbl_vars(tbl(con, in_schema("aux", "t1")))),
    c("x", "y", "z")
  )
  df <- tbl(con, ident("aux.t1"))
  expect_equal(as.character(tbl_vars(df)), c("a", "b", "c"))
})

test_that("useful error if missing I()", {
  withr::local_options(lifecycle_verbosity = "quiet")

  expect_snapshot(
    tbl(src_memdb(), "foo.bar"),
    error = TRUE
  )
})

test_that("check_from is deprecated", {
  con <- local_sqlite_connection()
  DBI::dbExecute(con, "CREATE TABLE x (y)")

  expect_snapshot(
    error = TRUE,
    tbl_sql("foo", src_dbi(con), "x", check_from = FALSE)
  )
})

# n_groups ----------------------------------------------------------------

test_that("check basic group size implementation", {
  db <- local_memdb_frame(x = rep(1:3, each = 10), y = rep(1:6, each = 5))
  expect_equal(n_groups(db), 1L)
  expect_equal(group_size(db), 30)

  gb <- group_by(db, x)
  expect_equal(n_groups(gb), 3L)
  expect_equal(group_size(gb), rep(10, 3))
})

test_that("fill returns same results as tidyr", {
  df <- tibble::tribble(
    ~id , ~group , ~letter , ~n1 , ~n2 ,
      1 ,      1 , NA      , NA  ,   1 ,
      2 ,      1 , "a"     ,   2 , NA  ,
      3 ,      1 , NA      , NA  , NA  ,
      4 ,      1 , "a"     , NA  ,   4 ,
      5 ,      2 , "a"     ,   5 , NA  ,
      6 ,      2 , NA      , NA  ,   6 ,
  )
  df_db <- copy_to(memdb(), df, overwrite = TRUE)
  withr::defer(DBI::dbRemoveTable(memdb(), "df"))

  expect_equal(
    df_db |>
      window_order(id) |>
      group_by(group) |>
      tidyr::fill(n1, n2) |>
      collect(),
    df |> group_by(group) |> tidyr::fill(n1, n2)
  )
})

test_that("generates valid sql for all directions", {
  lf <- lazy_frame(id = 1, group = 1, n1 = 1)

  lf_asc <- lf |> window_order(id)
  lf_desc <- lf |> window_order(desc(id))

  expect_snapshot({
    lf_asc |> tidyr::fill(n1, .direction = "up")
    lf_desc |> tidyr::fill(n1, .direction = "up")

    lf_asc |> tidyr::fill(n1, .direction = "updown")
    lf_asc |> tidyr::fill(n1, .direction = "downup")
  })
})

test_that("fill() respects grouping", {
  lf <- lazy_frame(id = 1, group = 1, n1 = 1)
  expect_snapshot(
    lf |> group_by(group) |> window_order(id) |> tidyr::fill(n1)
  )
})

test_that("can generate variant SQL", {
  lf <- lazy_frame(id = 1, group = 1, n1 = 1, con = dialect_sqlite())

  expect_snapshot({
    lf |> window_order(id) |> tidyr::fill(n1)
    lf |> window_order(desc(id)) |> tidyr::fill(n1)
    lf |> group_by(group) |> window_order(id) |> tidyr::fill(n1)
  })
})

test_that("fill errors on unsorted data", {
  df <- local_memdb_frame(x = 1)

  expect_snapshot(tidyr::fill(df), error = TRUE)
})

test_that("fill() errors on attempted rename", {
  lf <- lazy_frame(x = 1)
  expect_snapshot(tidyr::fill(lf, y = x), error = TRUE)
})

test_that("fill() produces nice error messages", {
  lf <- lazy_frame(x = 1)
  expect_snapshot(tidyr::fill(lf, non_existent), error = TRUE)
})

test_that("src_memdb() is deprecated", {
  expect_snapshot(. <- src_memdb())
})

test_that("tbl_memdb() is deprecated", {
  expect_snapshot(. <- tbl_memdb(data.frame(x = 1)))
})

test_that("memdb_frame() with data frame is deprecated", {
  expect_snapshot(. <- memdb_frame(data.frame(x = 1)))
})

sql_case_match <- function(.x, ..., .default = NULL, .ptype = NULL) {
  error_call <- current_call()
  check_unsupported_arg(.ptype)

  x_expr <- enexpr(.x)
  if (!is_symbol(x_expr) && !is_call(x_expr)) {
    msg <- "{.arg .x} must be a variable or function call, not {.obj_type_friendly {.x}}."
    cli_abort(msg, call = error_call)
  }

  formulas <- list2(...)
  formulas <- purrr::compact(formulas)

  n <- length(formulas)

  if (n == 0) {
    cli_abort("No cases provided", call = error_call)
  }

  con <- sql_current_con()
  query <- vector("list", n)
  value <- vector("list", n)

  for (i in seq_len(n)) {
    f <- formulas[[i]]
    env <- environment(f)

    query[[i]] <- sql_case_match_clause(f, .x, con)
    value[[i]] <- escape(
      enpar(quo(!!f[[3]]), tidy = FALSE, env = env),
      con = con
    )
  }

  clauses <- purrr::map2_chr(
    query,
    value,
    \(lhs, rhs) paste0("WHEN (", lhs, ") THEN ", rhs)
  )
  if (!is_null(.default)) {
    .default <- escape(enpar(quo(.default), tidy = FALSE, env = env), con = con)
    clauses[[n + 1]] <- paste0("ELSE ", .default)
  }

  same_line_sql <- sql(paste0("CASE ", paste0(clauses, collapse = " "), " END"))
  if (nchar(same_line_sql) <= 80) {
    return(same_line_sql)
  }

  sql(paste0(
    "CASE\n",
    paste0(clauses, collapse = "\n"),
    "\nEND"
  ))
}

sql_case_match_clause <- function(f, x, con) {
  env <- environment(f)
  f_query <- f[[2]]
  if (is_call(f_query, "c")) {
    # need to handle `c(...)` specially because on `expr(c(1, y))` it returns
    # `sql('1', '`y`')`
    f_query <- call_args(f_query)
  } else {
    if (!is_vector(f_query) || length(f_query) == 1) {
      f_query <- list(f_query)
    }
  }

  f_query <- purrr::map_if(
    f_query,
    \(clause) !is_vector(clause),
    \(clause) translate_sql(!!clause, con = con)
  )
  missing_loc <- purrr::map_lgl(f_query, function(clause) {
    is.null(clause) || is.na(clause)
  })

  f_query <- vctrs::vec_slice(f_query, !missing_loc)
  has_na <- any(missing_loc)

  query <- NULL
  if (!is_empty(f_query)) {
    query <- sql_glue("{x} IN {f_query*}")
  }

  if (has_na) {
    query <- paste(c(query, sql_glue2(con, "{x} IS NULL")), collapse = " OR ")
  }

  query
}


sql_if <- function(cond, if_true, if_false = quo(NULL), missing = quo(NULL)) {
  enpared_cond <- enpar(cond)
  enpared_if_true <- enpar(if_true)
  enpared_if_false <- enpar(if_false)
  enpared_missing <- enpar(missing)
  con <- sql_current_con()

  out <- "CASE WHEN {enpared_cond} THEN {enpared_if_true}"

  # `ifelse()` and `if_else()` have a three value logic: they return `NA` resp.
  # `missing` if `cond` is `NA`. To get the same in SQL it is necessary to
  # translate to
  # CASE
  #   WHEN <cond> THEN `if_true`
  #   WHEN NOT <cond> THEN `if_false`
  #   ELSE `missing`
  # END
  if (!quo_is_null(if_false) && !identical(if_false, missing)) {
    out <- paste0(out, " WHEN NOT {enpared_cond} THEN {enpared_if_false}")
  }
  if (!quo_is_null(missing)) {
    out <- paste0(out, " ELSE {enpared_missing}")
  }
  out <- paste0(out, " END")

  sql_glue2(con, out)
}

sql_case_when <- function(
  ...,
  .default = NULL,
  .ptype = NULL,
  .size = NULL,
  error_call = caller_env()
) {
  # TODO: switch to dplyr::case_when_prepare when available
  check_unsupported_arg(.ptype, call = error_call)
  check_unsupported_arg(.size, call = error_call)
  con <- sql_current_con()

  formulas <- list2(...)
  n <- length(formulas)

  if (n == 0) {
    cli_abort("No cases provided")
  }

  query <- vector("list", n)
  value <- vector("list", n)

  for (i in seq_len(n)) {
    f <- formulas[[i]]

    env <- environment(f)
    query[[i]] <- escape(
      enpar(quo(!!f[[2]]), tidy = FALSE, env = env),
      con = con
    )
    value[[i]] <- escape(
      enpar(quo(!!f[[3]]), tidy = FALSE, env = env),
      con = con
    )
  }

  clauses <- purrr::map2_chr(query, value, function(cond, val) {
    paste0("WHEN ", cond, " THEN ", val)
  })
  # if a formula like TRUE ~ "other" is at the end of a sequence, use ELSE statement
  # TRUE has precedence over `.default`
  if (is_true(formulas[[n]][[2]])) {
    clauses[[n]] <- paste0("ELSE ", value[[n]])
  } else if (!is_null(.default)) {
    .default <- escape(enpar(quo(.default), tidy = FALSE, env = env), con = con)
    clauses[[n + 1]] <- paste0("ELSE ", .default)
  }

  same_line_sql <- sql(paste0("CASE ", paste0(clauses, collapse = " "), " END"))
  if (nchar(same_line_sql) <= 80) {
    return(same_line_sql)
  }

  sql(paste0(
    "CASE\n",
    paste0(clauses, collapse = "\n"),
    "\nEND"
  ))
}

sql_switch <- function(x, ...) {
  con <- sql_current_con()
  input <- list2(...)

  named <- names2(input) != ""

  clauses <- purrr::map2_chr(names(input)[named], input[named], function(x, y) {
    sql_glue2(con, "WHEN ({x}) THEN ({y})")
  })

  n_unnamed <- sum(!named)
  if (n_unnamed == 0) {
    # do nothing
  } else if (n_unnamed == 1) {
    idx <- which(!named)
    clauses <- c(clauses, sql_glue2(con, "ELSE ({input[[idx]]})"))
  } else {
    cli_abort("Can only have one unnamed (ELSE) input")
  }

  clauses_collapsed <- paste0(clauses, collapse = " ")
  sql_glue2(con, "CASE {x} {.sql clauses_collapsed} END")
}

sql_is_null <- function(x) {
  x_sql <- enpar(enquo(x))
  sql_glue("({x_sql} IS NULL)")
}

enpar <- function(x, tidy = TRUE, env = NULL) {
  if (!is_quosure(x)) {
    cli_abort("Internal error: `x` must be a quosure.") # nocov
  }

  if (tidy) {
    x_sql <- eval_tidy(x, env = env)
  } else {
    x_sql <- eval_bare(x, env = env)
  }
  if (quo_is_call(x)) {
    sql_glue("({x_sql})")
  } else {
    x_sql
  }
}

#' Create a base lazy query
#'
#' `lazy_base_query()` is a constructor for base lazy query objects. A base
#' lazy query represents the root of a lazy query tree, i.e. a database table or
#' query.
#'
#' @param x A data source, typically a table identifier created by [in_schema()]
#'   or [in_catalog()], or a literal [sql()] string.
#' @param vars A character vector of column names.
#' @param class A character vector of additional subclasses to add. The
#'   resulting object will have class `c("lazy_base_{class}_query",
#'   "lazy_base_query", "lazy_query")`.
#' @param ... Additional arguments passed to [lazy_query()].
#' @returns A lazy query object.
#' @keywords internal
#' @export
lazy_base_query <- function(x, vars, class = character(), ...) {
  check_character(vars)

  lazy_query(
    query_type = c(paste0("base_", class), "base"),
    x = x,
    vars = vars,
    ...,
    group_vars = character(),
    order_vars = NULL,
    frame = NULL
  )
}

lazy_query_remote <- function(x, vars) {
  lazy_base_query(x, vars, class = "remote")
}

#' @export
sql_build.lazy_base_remote_query <- function(op, con, ...) {
  base_query(op$x)
}

#' @export
sql_build.lazy_base_local_query <- function(op, con, ...) {
  base_query(op$name)
}

# base_query() ----------------------------------------------------------------

base_query <- function(from) {
  check_table_source(from)
  query("base", from = from)
}

#' @export
sql_render.base_query <- function(
  query,
  con = NULL,
  ...,
  sql_options = NULL,
  subquery = FALSE,
  lvl = 0
) {
  from <- query$from
  if (subquery || is.sql(from)) {
    from
  } else {
    from <- sql_escape_table_source(con, from)
    sql_query_select(con, sql("*"), from, lvl = lvl)
  }
}

#' @export
flatten_query.base_query <- function(qry, query_list, con) {
  query_list
}

#' Keep or drop rows that match a condition
#'
#' @description
#' These are methods for the dplyr [dplyr::filter()] and [dplyr::filter_out()]
#' generics. They generate the `WHERE` clause of the SQL query.
#'
#' `filter()` is translated directly to `WHERE`, which already matches dplyr's
#' behaviour of treating `NA` like `FALSE` (SQL's three-valued logic drops
#' `NULL` rows from `WHERE`).
#'
#' `filter_out()` requires an additional step, where the combined condition
#' is wrapped in `is_distinct_from(., TRUE)`, which is then translated using
#' the backend (e.g. to `IS DISTINCT FROM` on PostgreSQL, `IS NOT` on SQLite).
#' This ensures that the SQL translation matches dplyr's semantics.
#'
#' @inheritParams arrange.tbl_lazy
#' @inheritParams dplyr::filter
#' @param .preserve Not supported by this method.
#' @inherit arrange.tbl_lazy return
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(x = c(2, NA, 5, NA, 10), y = 1:5)
#' db |> filter(x < 5) |> show_query()
#' db |> filter_out(x < 5) |> show_query()
#' db |> filter(is.na(x)) |> show_query()
#' @importFrom dplyr filter
#' @exportS3Method NULL
# Registered onLoad
filter.tbl_lazy <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  check_unsupported_arg(.preserve, FALSE)
  check_filter(..., .fn = "filter")
  by <- compute_by(
    {{ .by }},
    .data,
    by_arg = ".by",
    data_arg = ".data",
    error_call = caller_env()
  )

  if (missing(...)) {
    return(.data)
  }

  .data <- set_by_groups(.data, by)
  dots <- partial_eval_dots(.data, ..., .named = FALSE)
  .data$lazy_query <- add_filter(.data$lazy_query, remote_con(.data), dots)
  clear_by_groups(.data, by)
}

#' @rdname filter.tbl_lazy
#' @importFrom dplyr filter_out
#' @export
filter_out.tbl_lazy <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  check_unsupported_arg(.preserve, FALSE)
  check_filter(..., .fn = "filter_out")
  by <- compute_by(
    {{ .by }},
    .data,
    by_arg = ".by",
    data_arg = ".data",
    error_call = caller_env()
  )

  if (missing(...)) {
    return(filter(.data, FALSE))
  }

  .data <- set_by_groups(.data, by)

  # Multiple conditions are AND-combined to match dplyr's semantics, then
  # wrapped in `is_distinct_from(., TRUE)`. The backend translation of
  # `is_distinct_from()` provides the dialect-specific SQL.
  dots <- partial_eval_dots(.data, ..., .named = FALSE)
  dots_and <- Reduce(function(a, b) expr((!!a) & (!!b)), dots)
  dots2 <- list(expr(is_distinct_from(!!dots_and, TRUE)))

  .data$lazy_query <- add_filter(.data$lazy_query, remote_con(.data), dots2)

  clear_by_groups(.data, by)
}

add_filter <- function(lazy_query, con, exprs) {
  exprs <- unname(exprs)

  if (uses_window_fun(exprs, con)) {
    # Extract out window functions into new mutate layer
    where <- translate_window_where_all(exprs, window_funs(con))

    # add_mutate() always creates a subquery, so we need to bring all
    # existing variables along for the ride
    original_vars <- op_vars(lazy_query)
    new_exprs <- c(syms(set_names(original_vars)), where$comp)
    mutated <- add_mutate(lazy_query, new_exprs)

    # filter with the modified `where` using the new columns
    lazy_select_query(
      x = mutated,
      select = syms(set_names(original_vars)),
      where = where$expr
    )
  } else if (filter_can_use_having(lazy_query)) {
    filter_via_having(lazy_query, exprs)
  } else if (inherits(lazy_query, "lazy_multi_join_query")) {
    # Due to way JOINs generate their SELECT clauses much later we have to
    # do the backtransformation of variable names in `sql_build.lazy_multi_join_query`
    lazy_query$where <- c(lazy_query$where, exprs)
    lazy_query
  } else if (can_inline_filter(exprs, lazy_query, con)) {
    # WHERE processed before SELECT
    exprs <- replace_sym(exprs, lazy_query$select$name, lazy_query$select$expr)

    lazy_query$where <- c(lazy_query$where, exprs)
    lazy_query
  } else {
    lazy_select_query(x = lazy_query, where = exprs)
  }
}

# filter() adds the WHERE clause
# * WHERE is executed before SELECT
#   => can't reference columns computed in SELECT
#   => can't inline if SELECT uses window functions (evaluated after WHERE)
#   => can't inline if SELECT uses sql() (can't extract column references)
can_inline_filter <- function(exprs, lazy_query, con) {
  if (!is_lazy_select_query(lazy_query)) {
    return(FALSE)
  }

  if (uses_mutated_vars(exprs, lazy_query$select)) {
    return(FALSE)
  }

  if (uses_window_fun(lazy_query$select$expr, con)) {
    return(FALSE)
  }

  if (any_expr_uses_sql(lazy_query$select$expr)) {
    return(FALSE)
  }

  TRUE
}

filter_can_use_having <- function(lazy_query) {
  # From the Postgres documentation: https://www.postgresql.org/docs/current/sql-select.html#SQL-HAVING
  # Each column referenced in condition must unambiguously reference a grouping
  # column, unless the reference appears within an aggregate function or the
  # ungrouped column is functionally dependent on the grouping columns.

  # After `summarise()` every column is either
  # * a grouping column
  # * or an aggregated column
  # (this is not the case for data frames but valid for SQL tables)
  #
  # Therefore, if `filter()` does not use a window function, then we only use
  # grouping or aggregated columns
  if (!is_lazy_select_query(lazy_query)) {
    FALSE
  } else {
    lazy_query$select_operation == "summarise"
  }
}

filter_via_having <- function(lazy_query, exprs) {
  # ANSI SQL processes HAVING before SELECT
  exprs <- replace_sym(exprs, lazy_query$select$name, lazy_query$select$expr)

  lazy_query$having <- c(lazy_query$having, exprs)
  lazy_query
}

check_filter <- function(..., .fn) {
  dots <- enquos(...)
  named <- have_name(dots)

  for (i in which(named)) {
    quo <- dots[[i]]

    # Unlike in `dplyr` named logical vectors do not make sense so they are
    # also not allowed
    expr <- quo_get_expr(quo)
    cli_abort(
      c(
        "Problem with {.fun {(.fn)}} input `..{i}`.",
        x = "Input `..{i}` is named.",
        i = "This usually means that you've used {.code =} instead of {.code ==}.",
        i = "Did you mean `{names(dots)[i]} == {as_label(expr)}`?"
      ),
      call = caller_env()
    )
  }
}

#' Subset rows using their positions
#'
#' @description
#' These are methods for the dplyr generics [slice_min()], [slice_max()], and
#' [slice_sample()]. They are translated to SQL using [dplyr::filter()] and
#' window functions (`ROWNUMBER`, `MIN_RANK`, or `CUME_DIST` depending on
#' arguments). `slice()`, `slice_head()`, and `slice_tail()` are not supported
#' since database tables have no intrinsic order.
#'
#' If data is grouped, the operation will be performed on each group so that
#' (e.g.) `slice_min(db, x, n = 3)` will select the three rows with the smallest
#' value of `x` in each group.
#'
#' @inheritParams arrange.tbl_lazy
#' @inheritParams dplyr::slice
#' @param ... Not used.
#' @param n,prop Provide either `n`, the number of rows, or `prop`, the
#'   proportion of rows to select. If neither are supplied, `n = 1` will be
#'   used.
#'
#'   If `n` is greater than the number of rows in the group (or `prop` > 1),
#'   the result will be silently truncated to the group size. If the proportion
#'   of a group size is not an integer, it is rounded down.
#' @param order_by Variable or function of variables to order by.
#' @param with_ties Should ties be kept together? The default, `TRUE`, may
#'   return more rows than you request. Use FALSE to ignore ties, and return
#'   the first n rows.
#' @param weight_by,replace Not supported for database backends.
#' @name dbplyr-slice
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(x = 1:3, y = c(1, 1, 2))
#' db |> slice_min(x) |> show_query()
#' db |> slice_max(x) |> show_query()
#' db |> slice_sample() |> show_query()
#'
#' db |> group_by(y) |> slice_min(x) |> show_query()
#'
#' # By default, ties are includes so you may get more rows
#' # than you expect
#' db |> slice_min(y, n = 1)
#' db |> slice_min(y, n = 1, with_ties = FALSE)
#'
#' # Non-integer group sizes are rounded down
#' db |> slice_min(x, prop = 0.5)
NULL

#' @importFrom dplyr slice
#' @export
slice.tbl_lazy <- function(.data, ...) {
  stop_unsupported_function("slice")
}

#' @importFrom dplyr slice_head
#' @export
slice_head.tbl_lazy <- function(.data, ..., n, prop, by = NULL) {
  stop_unsupported_function("slice_head", with = "slice_min")
}

#' @importFrom dplyr slice_tail
#' @export
slice_tail.tbl_lazy <- function(.data, ..., n, prop, by = NULL) {
  stop_unsupported_function("slice_tail", with = "slice_max")
}

#' @rdname dbplyr-slice
#' @importFrom dplyr slice_min
#' @export
slice_min.tbl_lazy <- function(
  .data,
  order_by,
  ...,
  n,
  prop,
  by = NULL,
  with_ties = TRUE,
  na_rm = TRUE
) {
  size <- check_slice_size(n, prop)
  check_unsupported_arg(na_rm, allowed = TRUE)
  order_by_expr <- partial_eval_quo(enquo(order_by), .data, "order_by")
  order_by <- unwrap_order_expr(order_by_expr, f = "slice_min")
  slice_by(.data, order_by, size, {{ by }}, with_ties = with_ties)
}

#' @rdname dbplyr-slice
#' @importFrom dplyr slice_max
#' @export
slice_max.tbl_lazy <- function(
  .data,
  order_by,
  ...,
  n,
  by = NULL,
  prop,
  with_ties = TRUE,
  na_rm = TRUE
) {
  size <- check_slice_size(n, prop)
  check_unsupported_arg(na_rm, allowed = TRUE)
  order_by_expr <- partial_eval_quo(enquo(order_by), .data, "order_by")
  order_by <- unwrap_order_expr(order_by_expr, f = "slice_max")
  order_by <- purrr::map(order_by, swap_order_direction)
  slice_by(.data, order_by, size, {{ by }}, with_ties = with_ties)
}

#' @rdname dbplyr-slice
#' @importFrom dplyr slice_sample
#' @export
slice_sample.tbl_lazy <- function(
  .data,
  ...,
  n,
  prop,
  by = NULL,
  weight_by = NULL,
  replace = FALSE
) {
  size <- check_slice_size(n, prop)
  weight_by <- enquo(weight_by)
  if (size$type == "prop") {
    cli_abort("Sampling by {.arg prop} is not supported on database backends")
  }

  if (!quo_is_null(weight_by)) {
    cli_abort("Weighted resampling is not supported on database backends")
  }
  if (replace) {
    cli_abort("Sampling with replacement is not supported on database backends")
  }

  order_by <- exprs(runif(n()))
  slice_by(.data, order_by, size, {{ by }}, with_ties = FALSE)
}

slice_by <- function(
  .data,
  order_by,
  size,
  .by,
  with_ties = FALSE,
  call = caller_env()
) {
  by <- compute_by(
    {{ .by }},
    .data,
    by_arg = "by",
    data_arg = "data",
    error_call = call
  )
  .data <- set_by_groups(.data, by)

  value <- switch(size$type, n = size$n, prop = size$prop)
  if (with_ties) {
    fun_name <- switch(size$type, n = "min_rank", prop = "cume_dist")
  } else {
    if (size$type == "prop") {
      cli_abort(
        "Can only use {.arg prop} when {.code with_ties = TRUE}",
        call = call
      )
    }
    fun_name <- "row_number"
  }

  # We generate a call to min_rank() and friends because we already do
  # a lot work in win_rank() to correctly handle NULLs. Currently this
  # layer of indirection isn't really needed since window functions have
  # same translations for all backends, but in the future we might add
  # more capability info to the dialect object so that we could generate
  # simpler SQL.
  if (length(order_by) == 1) {
    window_fun <- call2(fun_name, order_by[[1]])
  } else {
    window_fun <- call2(fun_name, call2("tibble", !!!order_by))
  }
  out <- filter(.data, !!window_fun <= !!value)
  clear_by_groups(out, by)
}


# helpers -----------------------------------------------------------------

check_slice_size <- function(n, prop) {
  if (missing(n) && missing(prop)) {
    list(type = "n", n = 1L)
  } else if (!missing(n) && missing(prop)) {
    if (!is.numeric(n) || length(n) != 1) {
      cli_abort("{.arg n} must be a single number.", call = caller_env())
    }
    if (is.na(n) || n < 0) {
      cli_abort(
        "{.arg n} must be a non-missing positive number.",
        call = caller_env()
      )
    }

    list(type = "n", n = as.integer(n))
  } else if (!missing(prop) && missing(n)) {
    if (!is.numeric(prop) || length(prop) != 1) {
      cli_abort("{.arg prop} must be a single number", call = caller_env())
    }
    if (is.na(prop) || prop < 0) {
      cli_abort(
        "{.arg prop} must be a non-missing positive number.",
        call = caller_env()
      )
    }
    list(type = "prop", prop = prop)
  } else {
    cli_abort(
      "Must supply exactly one of {.arg n} and {.arg prop} arguments.",
      call = caller_env()
    )
  }
}

utils::globalVariables(c(
  "min_rank",
  "cume_dist",
  "row_number",
  "desc",
  "runif"
))

#' Create an SQL translator
#'
#' @description
#' `sql_variant()` creates a SQL variant, a list of translators for scalar,
#' aggregate, and window functions. `sql_translator()` creates a translator,
#' an environment containing R to SQL translations. When creating a backend,
#' you'll use these functions to customize how R functions are converted to
#' SQL.
#'
#' Learn more in `vignette("new-backend")`.
#'
#' @param scalar,aggregate,window The three families of functions that an SQL
#'   variant can supply.
#' @param ...,.funs Named functions, used to add custom converters from standard
#'   R functions to SQL functions. Specify individually in `...`, or provide a
#'   list of `.funs`.
#' @param .parent The SQL variant that this variant should inherit from.
#'   Defaults to `base_agg` which provides a standard set of mappings for the
#'   most common operators and functions.
#'
#' @section Base translators:
#'
#' dbplyr provides the following base translators that implement standard SQL
#' semantics:
#'
#' * `base_scalar` - scalar functions and operators
#' * `base_agg` - aggregate functions
#' * `base_win` - window functions
#' * `base_no_win` - versions of window functions that throw errors
#'
#' @family SQL translation helpers
#' @export
#' @examples
#' # An example of adding some mappings for the statistical functions that
#' # postgresql provides: http://bit.ly/K5EdTn
#' postgres_agg <- sql_translator(.parent = base_agg,
#'   cor = sql_aggregate_2("CORR"),
#'   cov = sql_aggregate_2("COVAR_SAMP"),
#'   sd =  sql_aggregate("STDDEV_SAMP", "sd"),
#'   var = sql_aggregate("VAR_SAMP", "var")
#' )
#'
#' # Next we have to simulate a connection that uses this variant
#' con <- new_sql_dialect("test", quote = \(x) sql_quote(x, '"'))
#' sql_translation.sql_dialect_test <- function(x) {
#'   sql_variant(
#'     base_scalar,
#'     postgres_agg,
#'     base_no_win
#'   )
#' }
#'
#' translate_sql(cor(x, y), con = con, window = FALSE)
#' translate_sql(sd(income / years, na.rm = TRUE), con = con, window = FALSE)
sql_variant <- function(
  scalar = sql_translator(),
  aggregate = sql_translator(),
  window = sql_translator()
) {
  check_environment(scalar)
  check_environment(aggregate)
  check_environment(window)

  # Need to check that every function in aggregate also occurs in window
  missing <- setdiff(env_names(aggregate), env_names(window))
  if (length(missing) > 0) {
    warn(paste0(
      "Translator is missing window variants of the following aggregate functions:\n",
      paste0("* ", missing, "\n", collapse = "")
    ))
  }

  aggregate_fns <- env_names(aggregate)

  # An ensure that every window function is flagged in aggregate context
  missing <- setdiff(env_names(window), env_names(aggregate))
  missing_funs <- lapply(missing, sql_aggregate_win)
  env_bind(aggregate, !!!set_names(missing_funs, missing))

  structure(
    list(
      scalar = scalar,
      aggregate = aggregate,
      window = window,
      aggregate_fns = aggregate_fns
    ),
    class = "sql_variant"
  )
}

is.sql_variant <- function(x) inherits(x, "sql_variant")

#' @export
print.sql_variant <- function(x, ...) {
  wrap_ls <- function(x, ...) {
    vars <- sort(env_names(x))
    wrapped <- strwrap(paste0(vars, collapse = ", "), ...)
    if (identical(wrapped, "")) {
      return()
    }
    paste0(wrapped, "\n", collapse = "")
  }

  cat("<sql_variant>\n")
  cat(wrap_ls(
    x$scalar,
    prefix = "scalar:    "
  ))
  cat(wrap_ls(
    x$aggregate,
    prefix = "aggregate: "
  ))
  cat(wrap_ls(
    x$window,
    prefix = "window:    "
  ))
}

#' @export
names.sql_variant <- function(x) {
  c(env_names(x$scalar), env_names(x$aggregate), env_names(x$window))
}

#' @export
#' @rdname sql_variant
sql_translator <- function(
  ...,
  .funs = list(),
  .parent = new.env(parent = emptyenv())
) {
  funs <- c(list2(...), .funs)
  if (length(funs) == 0) {
    return(.parent)
  }

  if (anyDuplicated(names(funs))) {
    bullets <- unique(names(funs)[duplicated(names(funs))])
    cli_abort(c(
      "Duplicate names in {.fun sql_translator}",
      set_names(bullets, "*")
    ))
  }

  list2env(funs, copy_env(.parent))
}

copy_env <- function(from, to = NULL, parent = parent.env(from)) {
  list2env(as.list(from), envir = to, parent = parent)
}

test_that("missing window functions create a warning", {
  con <- dialect_ansi()
  sim_scalar <- sql_translator()
  sim_agg <- sql_translator(`+` = sql_infix("+"))
  sim_win <- sql_translator()

  expect_warning(
    sql_variant(sim_scalar, sim_agg, sim_win),
    "Translator is missing"
  )
})

test_that("duplicates throw an error", {
  expect_snapshot(error = TRUE, {
    sql_translator(round = \(x) x, round = \(y) y)
  })
})

test_that("missing aggregate functions filled in", {
  sim_scalar <- sql_translator()
  sim_agg <- sql_translator()
  sim_win <- sql_translator(mean = \() {})

  trans <- sql_variant(sim_scalar, sim_agg, sim_win)
  expect_error(trans$aggregate$mean(), "only available in a window")
})

test_that("output of print method for sql_variant is correct", {
  sim_trans <- sql_translator(`+` = sql_infix("+"))
  expect_snapshot(sql_variant(sim_trans, sim_trans, sim_trans))
})

test_that("win_rank() is accepted by the sql_translator", {
  expect_snapshot(
    sql_variant(
      sql_translator(
        test = win_rank("test")
      )
    )
  )
})

#' Partially evaluate an expression.
#'
#' @description
#' This function partially evaluates a quosure yielding an expression. It
#' uses information from the current `tbl` and the local environment to yield
#' a standalone expression. This simplifies SQL translation because we can just
#' pass around expressions rather than expressions + environments
#' (i.e. quosures).
#'
#' @section Symbol substitution:
#'
#' `partial_eval()` needs to guess if you're referring to a variable on the
#' server (remote), or in the current environment (local). It's not possible to
#' do this 100% perfectly. `partial_eval()` uses the following heuristic:
#'
#' \itemize{
#'   \item If the tbl variables are known, and the symbol matches a tbl
#'     variable, then remote.
#'   \item If the symbol is defined locally, local.
#'   \item Otherwise, remote.
#' }
#'
#' You can override the guesses using `local()` and `remote()` to force
#' computation, by using the `.data` and `.env` pronouns of tidy evaluation,
#' or by using dbplyr's own `.sql` pronoun.
#'
#' @param call an unevaluated expression, as produced by [quote()]
#' @param data A lazy data frame backed by a database query.
#' @param env environment in which to search for local values
#' @export
#' @keywords internal
#' @examples
#' lf <- lazy_frame(year = 1980, id = 1)
#' partial_eval(quote(year > 1980), data = lf)
#'
#' ids <- c("ansonca01", "forceda01", "mathebo01")
#' partial_eval(quote(id %in% ids), lf)
#'
#' # cf.
#' partial_eval(quote(id == .data$id), lf)
#'
#' # You can use local() or .env to disambiguate between local and remote
#' # variables: otherwise remote is always preferred
#' year <- 1980
#' partial_eval(quote(year > year), lf)
#' partial_eval(quote(year > local(year)), lf)
#' partial_eval(quote(year > .env$year), lf)
#'
#' # Functions are always assumed to be remote. Use local to force evaluation
#' # in R.
#' f <- function(x) x + 1
#' partial_eval(quote(year > f(1980)), lf)
#' partial_eval(quote(year > local(f(1980))), lf)
#'
#' # You can use `.sql` to make it clear that the function comes from SQL,
#' # and inside a package, reduce the number of globalVariables() directives
#' # needed
#' partial_eval(quote(.sql$EXTRACT_YEAR(year)), lf)
partial_eval <- function(
  call,
  data,
  env = caller_env(),
  error_call = caller_env()
) {
  if (!inherits(data, "tbl_lazy")) {
    cli::cli_abort("`data` must be a lazy data frame", call = error_call)
  }

  if (is_sql_literal(call)) {
    call
  } else if (is_symbol(call)) {
    partial_eval_sym(call, data, env)
  } else if (is_quosure(call)) {
    partial_eval(get_expr(call), data, get_env(call), error_call = error_call)
  } else if (is_call(call, "if_any")) {
    out <- partial_eval_if(
      call,
      data,
      env,
      reduce = "|",
      error_call = error_call
    )
    expr(((!!out)))
  } else if (is_call(call, "if_all")) {
    out <- partial_eval_if(
      call,
      data,
      env,
      reduce = "&",
      error_call = error_call
    )
    expr(((!!out)))
  } else if (is_call(call, "across")) {
    partial_eval_across(call, data, env, error_call)
  } else if (is_call(call, "pick")) {
    partial_eval_pick(call, data, env, error_call)
  } else if (is_call(call)) {
    partial_eval_call(call, data, env)
  } else {
    cli_abort("Unknown input type: {typeof(call)}")
  }
}

is_sql_literal <- function(x) {
  is_atomic(x) || is_null(x) || blob::is_blob(x)
}

capture_dot <- function(.data, x) {
  partial_eval(enquo(x), data = .data)
}

partial_eval_dots <- function(
  .data,
  ...,
  .named = TRUE,
  error_call = caller_env()
) {
  # corresponds to `capture_dots()`
  # browser()
  dots <- as.list(enquos(..., .named = .named))
  dot_names <- names2(exprs(...))
  was_named <- have_name(exprs(...))

  for (i in seq_along(dots)) {
    dots[[i]] <- partial_eval_quo(
      dots[[i]],
      .data,
      dot_names[[i]],
      error_call = error_call,
      was_named = was_named[[i]]
    )
  }

  # Remove names from any list elements
  is_list <- purrr::map_lgl(dots, is.list)
  names2(dots)[is_list] <- ""

  # Auto-splice list results from partial_eval_quo()
  dots[!is_list] <- lapply(dots[!is_list], list)
  unlist(dots, recursive = FALSE)
}

partial_eval_quo <- function(
  x,
  data,
  arg_name,
  error_call = caller_env(),
  was_named = FALSE
) {
  # no direct equivalent in `dtplyr`, mostly handled in `dt_squash()`
  withCallingHandlers(
    expr <- partial_eval(
      get_expr(x),
      data,
      get_env(x),
      error_call = error_call
    ),
    error = function(cnd) {
      label <- expr_as_label(x, arg_name)
      msg <- c(i = "In argument: {.code {label}}")
      cli_abort(msg, call = error_call, parent = cnd)
    }
  )

  if (is.list(expr)) {
    if (was_named) {
      msg <- c(
        "In dbplyr, the result of `across()` must be unnamed.",
        i = "`{arg_name} = {as_label(x)}` is named."
      )
      cli_abort(msg, call = error_call)
    }
  }

  expr
}

partial_eval_sym <- function(sym, data, env) {
  vars <- op_vars(data)
  name <- as_string(sym)
  if (name %in% vars) {
    sym
  } else if (env_has(env, name, inherit = TRUE)) {
    # Inline the value so that the translation function can choose what to do

    val <- eval_bare(sym, env)
    if (is_atomic(val)) {
      val <- unname(val)
    }
    val
  } else {
    cli::cli_abort(
      "Object {.var {name}} not found.",
      call = NULL
    )
  }
}

is_mask_pronoun <- function(call) {
  is_call(call, c("$", "[["), n = 2) && is_symbol(call[[2]], c(".data", ".env"))
}
is_sql_pronoun <- function(call) {
  is_call(call, "$", n = 2) && is_symbol(call[[2]], ".sql")
}

partial_eval_call <- function(call, data, env) {
  fun <- call[[1]]

  # Try to find the name of inlined functions
  if (inherits(fun, "inline_colwise_function")) {
    vars <- colnames(tidyselect_data_proxy(data))
    dot_var <- vars[[attr(call, "position")]]
    call <- replace_sym1(attr(fun, "formula")[[2]], c(".", ".x"), sym(dot_var))
    env <- get_env(attr(fun, "formula"))
  } else if (is.function(fun)) {
    fun_name <- find_fun(fun)
    if (is.null(fun_name)) {
      # This probably won't work, but it seems like it's worth a shot.
      return(eval_bare(call, env))
    }

    call[[1]] <- fun <- sym(fun_name)
  }

  # Compound calls, apart from pronouns and `::` aren't translatable
  if (is_call(fun) && !is_call(fun, "::")) {
    if (is_mask_pronoun(fun)) {
      cli::cli_abort(
        "Use local() or remote() to force evaluation of functions",
        call = NULL
      )
    } else if (is_sql_pronoun(fun)) {
      call[[1]] <- fun[[3]]
    } else {
      return(eval_bare(call, env))
    }
  }

  if (is_mask_pronoun(call)) {
    # special handling for .data$, .data[[]], .env$, .env[[]]
    var <- call[[3]]
    if (is_call(call, "[[")) {
      var <- sym(eval(var, env))
    }

    if (is_symbol(call[[2]], ".data")) {
      var
    } else {
      eval_bare(var, env)
    }
  } else if (is_sql_pronoun(call)) {
    # special handling for .sql$
    call[[3]]
  } else {
    # Process call arguments recursively, unless user has manually called
    # remote/local
    if (is_call(call, "local")) {
      eval_bare(call[[2]], env)
    } else if (is_call(call, "remote")) {
      call[[2]]
    } else if (is_call(call, "sql")) {
      eval_bare(call, env = env)
    } else if (is_call(call, "$")) {
      # Only the 1st argument is evaluated
      call[[2]] <- partial_eval(call[[2]], data = data, env = env)
      call
    } else {
      # Check for shiny reactives before processing unknown function calls
      if (is_symbol(fun)) {
        fun_name <- as_string(fun)
        obj <- env_get(env, fun_name, default = NULL, inherit = TRUE)
        if (inherits(obj, "reactive")) {
          error_embed("a shiny reactive", "foo()")
        }
      }

      call[-1] <- lapply(call[-1], partial_eval, data = data, env = env)
      call
    }
  }
}

find_fun <- function(fun) {
  if (is_lambda(fun)) {
    body <- body(fun)
    if (!is_call(body)) {
      return(NULL)
    }

    fun_name <- body[[1]]
    if (!is_symbol(fun_name)) {
      return(NULL)
    }

    as.character(fun_name)
  } else if (is.function(fun)) {
    fun_name(fun)
  }
}

fun_name <- function(fun) {
  # `dtplyr` uses the same idea but needs different environments
  pkg_env <- env_parent(global_env())
  known <- c(env_names(base_agg), env_names(base_scalar))

  for (x in known) {
    if (!env_has(pkg_env, x, inherit = TRUE)) {
      next
    }

    fun_x <- env_get(pkg_env, x, inherit = TRUE)
    if (identical(fun, fun_x)) {
      return(sym(x))
    }
  }

  NULL
}


replace_sym <- function(exprs, old, new) {
  check_list(exprs, allow_null = TRUE)
  check_character(old)
  check_list(new)
  # Allow new to be a list of quosures too
  new <- purrr::map_if(new, is_quosure, quo_get_expr)

  purrr::map(exprs, \(expr) replace_sym1(expr, old, new))
}

replace_sym1 <- function(call, sym, replace) {
  if (is_symbol(call, sym)) {
    if (is_list(replace)) {
      replace[[match(as_string(call), sym)]]
    } else {
      replace
    }
  } else if (is_call(call, c("$", "@"))) {
    # The 2nd argument is a field/slot name, not a variable reference
    call[[2]] <- replace_sym1(call[[2]], sym, replace)
    call
  } else if (is_call(call)) {
    call[] <- lapply(call, replace_sym1, sym = sym, replace = replace)
    call
  } else {
    call
  }
}


#' Flag SQL function usage
#'
#' @description
#' Use `.sql$foo(x, y)` to make it clear that you're calling the SQL
#' `foo()` function, not the R `foo()` function. This also makes it easier to
#' reduce `R CMD check` notes in packages; just import `.sql` from dbplyr with
#' e.g. `@importFrom dbplyr .sql`.
#'
#' Note that `.sql` itself does nothing and is just `NULL`; it is automatically
#' removed when dbplyr translates your R code to SQL.
#'
#' @export
#' @format NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- lazy_frame(x = 1, y = 2)
#' db |> mutate(z = .sql$CUMULATIVE_SUM(x, 1))
.sql <- NULL

test_that("simple expressions left as is", {
  lf <- lazy_frame(x = 1:10, y = 1:10)

  expect_equal(capture_dot(lf, NULL), NULL)
  expect_equal(capture_dot(lf, 10), 10)
  expect_equal(capture_dot(lf, x), sym("x"))
  expect_equal(capture_dot(lf, x + y), expr(x + y))
  expect_equal(capture_dot(lf, x[[1]]), expr(x[[1]]))

  # logicals
  expect_equal(eval(capture_dot(lf, T), globalenv()), TRUE)
  expect_equal(eval(capture_dot(lf, F), globalenv()), FALSE)
  expect_equal(capture_dot(lf, TRUE), TRUE)
  expect_equal(capture_dot(lf, FALSE), FALSE)
})

test_that("existing atomic non-variables get inlined", {
  lf <- lazy_frame(x = 1:10, y = 1:10)

  n <- 10
  expect_equal(capture_dot(lf, x + n), expr(x + 10))
})

test_that("names are stripped", {
  lf <- lazy_frame(x = "a")
  y <- c(x = "a", "b")

  expect_equal(partial_eval(quote(x %in% y), lf), expr(x %in% !!c("a", "b")))
})

test_that("using environment of inlined quosures", {
  lf <- lazy_frame(x = 1:10, y = 1:10)

  n <- 10
  quo <- new_quosure(quote(x + n), env(n = 20))

  expect_equal(capture_dot(lf, f(!!quo)), quote(f(x + 20)))
})

test_that("use quosure environment for unevaluted formulas", {
  lf <- lazy_frame(x = 1, y = 2)

  z <- 1
  expect_equal(partial_eval(expr(~z), lf), quote(~1))
})

test_that("can look up inlined function", {
  lf <- lazy_frame(x = 1, y = 2)

  expect_equal(
    partial_eval(expr((!!mean)(x)), data = lf),
    expr(mean(x))
  )
})

test_that("respects tidy evaluation pronouns", {
  lf <- lazy_frame(x = 1, y = 2)

  x <- "X"
  X <- "XX"

  expect_equal(partial_eval(expr(.data$x), lf), expr(x))
  expect_equal(partial_eval(expr(.data[["x"]]), lf), expr(x))
  expect_equal(partial_eval(expr(.data[[x]]), lf), expr(X))

  expect_equal(partial_eval(expr(.env$x), lf), "X")
  expect_equal(partial_eval(expr(.env[["x"]]), lf), "X")
  expect_equal(partial_eval(expr(.env[[x]]), lf), "XX")
})

test_that("respects .sql pronoun", {
  lf <- lazy_frame(x = 1)

  expect_equal(partial_eval(expr(.sql$foo), lf), expr(foo))
  expect_equal(partial_eval(expr(.sql$foo(x, "y")), lf), expr(foo(x, "y")))
})

test_that("sql() is inlined", {
  lf <- lazy_frame(x = 1)

  foo <- "x"
  expect_equal(partial_eval(expr(sql(foo)), lf), sql("x"))
})

test_that("fails with multi-classes", {
  lf <- lazy_frame(x = 1, y = 2)
  x <- structure(list(), class = c('a', 'b'))
  expect_error(partial_eval(x, lf), "Unknown input type", fixed = TRUE)
})

# replace_sym1 -----------------------------------------------------------------

test_that("replace_sym1() replaces matching symbols", {
  expect_equal(replace_sym1(sym("x"), "x", sym("y")), sym("y"))
  expect_equal(replace_sym1(sym("z"), "x", sym("y")), sym("z"))
  expect_equal(replace_sym1(expr(x + z), "x", sym("y")), expr(y + z))
  expect_equal(replace_sym1(1, "x", sym("y")), 1)
})

test_that("replace_sym1() supports list of replacements", {
  expect_equal(
    replace_sym1(expr(x + y), c("x", "y"), list(sym("a"), sym("b"))),
    expr(a + b)
  )
})

test_that("replace_sym1() does not rewrite field names of `$` and `@` (#1812)", {
  expect_equal(replace_sym1(expr(ltr$x), "x", sym("y")), expr(ltr$x))
  expect_equal(replace_sym1(expr(ltr@x), "x", sym("y")), expr(ltr@x))

  # but still recurses into the object
  expect_equal(replace_sym1(expr(x$foo), "x", sym("y")), expr(y$foo))
  expect_equal(replace_sym1(expr(x@foo), "x", sym("y")), expr(y@foo))
})

#' Helper function for quoting sql elements
#'
#' If the quote character is present in the string, it will be doubled.
#' `NA`s will be replaced with NULL.
#'
#' @param x Character vector to escape.
#' @param quote Quote character. Either a length 1 character vector for
#'   symmetric quotes (e.g., `"'"` or `'"'`), or a length 2 character vector
#'   for asymmetric quotes (e.g., `c("[", "]")`).
#' @export
#' @return A vector of [sql].
#' @keywords internal
#' @examples
#' sql_quote("abc", "'")
#' sql_quote("I've had a good day", "'")
#' sql_quote(c("abc", NA), "'")
#'
#' sql_quote(c("abc", NA), c("[", "]"))
sql_quote <- function(x, quote) {
  check_character(x)
  check_character(quote)

  if (length(x) == 0) {
    return(sql())
  }

  if (length(quote) == 1) {
    y <- gsub(quote, paste0(quote, quote), x, fixed = TRUE)
    y <- paste0(quote, y, quote)
  } else if (length(quote) == 2) {
    y <- gsub(quote[[2]], paste0(quote[[2]], quote[[2]]), x, fixed = TRUE)
    y <- paste0(quote[[1]], y, quote[[2]])
  } else {
    cli::cli_abort("{.arg quote} must be length 1 or 2.")
  }
  y[is.na(x)] <- "NULL"
  names(y) <- names(x)

  sql(y)
}

#' Group by one or more variables
#'
#' This is a method for the dplyr [group_by()] generic. It is translated to
#' the `GROUP BY` clause of the SQL query when used with
#' [`summarise()`][summarise.tbl_lazy] and to the `PARTITION BY` clause of
#' window functions when used with [`mutate()`][mutate.tbl_lazy].
#'
#' @inheritParams arrange.tbl_lazy
#' @inheritParams dplyr::group_by
#' @param .drop Not supported by this method.
#' @export
#' @importFrom dplyr group_by
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' db <- memdb_frame(g = c(1, 1, 1, 2, 2), x = c(4, 3, 6, 9, 2))
#' db |>
#'   group_by(g) |>
#'   summarise(n()) |>
#'   show_query()
#'
#' db |>
#'   group_by(g) |>
#'   mutate(x2 = x / sum(x, na.rm = TRUE)) |>
#'   show_query()
group_by.tbl_lazy <- function(
  .data,
  ...,
  .add = FALSE,
  .drop = TRUE
) {
  dots <- partial_eval_dots(.data, ..., .named = FALSE)

  check_unsupported_arg(.drop, TRUE)
  groups <- dplyr::group_by_prepare(
    .data,
    !!!dots,
    .add = .add,
    error_call = current_call()
  )
  names <- purrr::map_chr(groups$groups, as_string)

  same_groups <- setequal(groups$group_names, group_vars(.data))
  if (same_groups) {
    return(groups$data)
  }

  groups$data$lazy_query <- add_group_by(groups$data$lazy_query, names)
  groups$data
}

# ungroup -----------------------------------------------------------------

#' @importFrom dplyr ungroup
#' @export
ungroup.tbl_lazy <- function(x, ...) {
  if (missing(...)) {
    group_by(x)
  } else {
    old_groups <- group_vars(x)
    to_remove <- tidyselect::vars_select(op_vars(x), ...)
    new_groups <- setdiff(old_groups, to_remove)
    group_by(x, !!!syms(new_groups))
  }
}

add_group_by <- function(lazy_query, vars) {
  lazy_query$group_vars <- unname(vars)
  lazy_query
}

#' Databricks Spark SQL backend
#'
#' @description
#' This backend supports Databricks Spark SQL, typically accessed via the
#' Databricks ODBC or JDBC connector. Use `dialect_spark_sql()` with
#' `lazy_frame()` to see simulated SQL without connecting to a live database.
#'
#' Key differences for this backend are better translation of statistical
#' aggregate functions (e.g. `var()`, `median()`) and use of temporary views
#' instead of temporary tables when copying data.
#'
#' See `vignette("translation-function")` and `vignette("translation-verb")` for
#' details of overall translation technology.
#'
#' @name backend-spark-sql
#' @aliases NULL
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' lf <- lazy_frame(a = TRUE, b = 1, d = 2, c = "z", con = dialect_spark_sql())
#'
#' lf |> summarise(x = median(d, na.rm = TRUE))
#' lf |> summarise(x = var(c, na.rm = TRUE), .by = d)
#'
#' lf |> mutate(x = first(c))
#' lf |> mutate(x = first(c), .by = d)
NULL

#' @export
#' @rdname backend-spark-sql
dialect_spark_sql <- function() {
  new_sql_dialect(
    "spark_sql",
    quote_identifier = function(x) sql_quote(x, "`"),
    has_window_clause = TRUE
  )
}

#' @export
#' @rdname backend-spark-sql
simulate_spark_sql <- function() simulate_dbi("Spark SQL")

#' @export
`sql_dialect.Spark SQL` <- function(con) {
  dialect_spark_sql()
}

#' @export
`dbplyr_edition.Spark SQL` <- function(con) {
  2L
}

#' @export
sql_translation.sql_dialect_spark_sql <- function(con) {
  sql_variant(
    sql_translator(
      .parent = base_odbc_scalar,
      # stringr ---------------------------------------------------------------
      str_like = function(string, pattern, ignore_case = deprecated()) {
        ignore_case <- deprecate_ignore_case(ignore_case)
        if (ignore_case) {
          sql_glue("{string} ILIKE {pattern}")
        } else {
          sql_glue("{string} LIKE {pattern}")
        }
      },
      str_ilike = function(string, pattern) {
        sql_glue("{string} ILIKE {pattern}")
      },
      # clock ---------------------------------------------------------------
      add_days = function(x, n, ...) {
        check_dots_empty()
        sql_glue("DATE_ADD({x}, {n})")
      },
      add_years = function(x, n, ...) {
        check_dots_empty()
        sql_glue("ADD_MONTHS({x}, {n} * 12)")
      },
      date_build = function(year, month = 1L, day = 1L, ..., invalid = NULL) {
        check_unsupported_arg(invalid, allow_null = TRUE)
        sql_glue("MAKE_DATE({year}, {month}, {day})")
      },
      get_year = function(x) {
        sql_glue("DATE_PART('YEAR', {x})")
      },
      get_month = function(x) {
        sql_glue("DATE_PART('MONTH', {x})")
      },
      get_day = function(x) {
        sql_glue("DATE_PART('DAY', {x})")
      },
      date_count_between = function(start, end, precision, ..., n = 1L) {
        check_dots_empty()
        check_unsupported_arg(precision, allowed = "day")
        check_unsupported_arg(n, allowed = 1L)

        sql_glue("DATEDIFF({end}, {start})")
      },

      difftime = function(time1, time2, tz, units = "days") {
        check_unsupported_arg(tz)
        check_unsupported_arg(units, allowed = "days")

        sql_glue("DATEDIFF({time2}, {time1})")
      }
    ),
    sql_translator(
      .parent = base_odbc_agg,
      var = sql_aggregate("VARIANCE", "var"),
      quantile = sql_quantile("PERCENTILE"),
      median = sql_aggregate("MEDIAN"),
      first = function(x, na_rm = FALSE) {
        sql_check_na_rm(na_rm)
        sql_glue("FIRST({x})")
      },
      last = function(x, na_rm = FALSE) {
        sql_check_na_rm(na_rm)
        sql_glue("LAST({x})")
      },
    ),
    sql_translator(
      .parent = base_odbc_win,
      var = win_aggregate("VARIANCE"),
      quantile = sql_quantile("PERCENTILE", window = TRUE),
      median = win_aggregate("MEDIAN"),
      first = function(x, order_by = NULL, na_rm = FALSE) {
        sql_nth(
          x,
          1L,
          order_by = order_by,
          na_rm = na_rm,
          ignore_nulls = "bool"
        )
      },
      last = function(x, order_by = NULL, na_rm = FALSE) {
        sql_nth(
          x,
          Inf,
          order_by = order_by,
          na_rm = na_rm,
          ignore_nulls = "bool"
        )
      },
      nth = function(x, n, order_by = NULL, na_rm = FALSE) {
        sql_nth(x, n, order_by = order_by, na_rm = na_rm, ignore_nulls = "bool")
      },
    )
  )
}

#' @export
sql_table_analyze.sql_dialect_spark_sql <- function(con, table, ...) {
  # https://docs.databricks.com/en/sql/language-manual/sql-ref-syntax-aux-analyze-table.html
  sql_glue2(con, "ANALYZE TABLE {.tbl table} COMPUTE STATISTICS")
}

#' @export
`db_copy_to.Spark SQL` <- function(
  con,
  table,
  values,
  ...,
  overwrite = FALSE,
  types = NULL,
  temporary = TRUE,
  unique_indexes = NULL,
  indexes = NULL,
  analyze = TRUE,
  in_transaction = FALSE
) {
  sql <- sql_values_subquery(con, values, types = types, lvl = 1)
  db_compute(con, table, sql, overwrite = overwrite, temporary = temporary)
}

#' @export
`db_compute.Spark SQL` <- function(
  con,
  table,
  sql,
  ...,
  overwrite = FALSE,
  temporary = TRUE,
  unique_indexes = list(),
  indexes = list(),
  analyze = TRUE,
  in_transaction = FALSE
) {
  check_bool(overwrite)
  check_bool(temporary)

  action <- if (overwrite) "CREATE OR REPLACE" else "CREATE"
  type <- if (temporary) "TEMPORARY VIEW" else "TABLE"

  sql <- as_table_source(sql)
  sql <- sql_glue2(con, "{.sql action} {.sql type} {.tbl table} AS \n{sql}")
  DBI::dbExecute(con, sql)

  table
}

test_that("simulate_spark_sql() still works", {
  expect_translation(simulate_spark_sql(), x + 1, "`x` + 1.0")
})

test_that("custom clock functions translated correctly", {
  con <- dialect_spark_sql()
  expect_translation(con, add_years(x, 1), "ADD_MONTHS(`x`, 1.0 * 12)")
  expect_translation(con, add_days(x, 1), "DATE_ADD(`x`, 1.0)")
  expect_error(
    translate_sql(add_days(x, 1, "dots", "must", "be empty"), con = con),
    class = "rlib_error_dots_nonempty"
  )
  expect_translation(
    con,
    date_build(2020, 1, 1),
    "MAKE_DATE(2020.0, 1.0, 1.0)"
  )
  expect_translation(
    con,
    date_build(year_column, 1L, 1L),
    "MAKE_DATE(`year_column`, 1, 1)"
  )
  expect_translation(
    con,
    get_year(date_column),
    "DATE_PART('YEAR', `date_column`)"
  )
  expect_translation(
    con,
    get_month(date_column),
    "DATE_PART('MONTH', `date_column`)"
  )
  expect_translation(
    con,
    get_day(date_column),
    "DATE_PART('DAY', `date_column`)"
  )
  expect_translation(
    con,
    date_count_between(date_column_1, date_column_2, "day"),
    "DATEDIFF(`date_column_2`, `date_column_1`)"
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      date_count_between(date_column_1, date_column_2, "year"),
      con = con
    )
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      date_count_between(
        date_column_1,
        date_column_2,
        "day",
        n = 5
      ),
      con = con
    )
  )
})

test_that("difftime is translated correctly", {
  con <- dialect_spark_sql()
  expect_translation(
    con,
    difftime(start_date, end_date, units = "days"),
    "DATEDIFF(`end_date`, `start_date`)"
  )
  expect_translation(
    con,
    difftime(start_date, end_date),
    "DATEDIFF(`end_date`, `start_date`)"
  )

  expect_snapshot(
    error = TRUE,
    translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
  )
  expect_snapshot(
    error = TRUE,
    translate_sql(
      difftime(
        start_date,
        end_date,
        tz = "UTC",
        units = "days"
      ),
      con = con
    )
  )
})

#' Copy a local data frame to a remote database
#'
#' @description
#' This is an implementation of the dplyr [copy_to()] generic and it mostly
#' a wrapper around [DBI::dbWriteTable()].
#'
#' It is useful for copying small amounts of data to a database for examples,
#' experiments, and joins. By default, it creates temporary tables which are
#' only visible within the current connection to the database.
#'
#' @export
#' @param df A local data frame, a `tbl_sql` from same source, or a `tbl_sql`
#'   from another source. If from another source, all data must transition
#'   through R in one pass, so it is only suitable for transferring small
#'   amounts of data.
#' @param name Name of new remote table. Use a string to create the table
#'   in the current catalog/schema. Use `I()` to create the table in a specific
#'   catalog/schema, e.g. `I("schema.table")`.
#' @param types a character vector giving variable types to use for the columns.
#'    See <https://www.sqlite.org/datatype3.html> for available types.
#' @param temporary if `TRUE`, will create a temporary table that is
#'   local to this connection and will be automatically deleted when the
#'   connection expires
#' @param unique_indexes a list of character vectors. Each element of the list
#'   will create a new unique index over the specified column(s). Duplicate rows
#'   will result in failure.
#' @param indexes a list of character vectors. Each element of the list
#'   will create a new index.
#' @param analyze if `TRUE` (the default), will automatically ANALYZE the
#'   new table so that the query optimiser has useful information.
#' @param in_transaction Should the table creation be wrapped in a transaction?
#'   This typically makes things faster, but you may want to suppress if the
#'   database doesn't support transactions, or you're wrapping in a transaction
#'   higher up (and your database doesn't support nested transactions.)
#' @inheritParams dplyr::copy_to
#' @inherit arrange.tbl_lazy return
#' @seealso [copy_inline()] to use small data in an SQL query without actually
#'   writing to a table.
#' @examples
#' library(dplyr, warn.conflicts = FALSE)
#'
#' df <- data.frame(x = 1:5, y = letters[5:1])
#' db <- copy_to(memdb(), df)
#' db
#'
#' df2 <- data.frame(y = c("a", "d"), fruit = c("apple", "date"))
#' # copy_to() is called automatically if you set copy = TRUE
#' # in the join functions
#' db |> left_join(df2, copy = TRUE)
#' @importFrom dplyr copy_to
copy_to.src_sql <- function(
  dest,
  df,
  name = deparse(substitute(df)),
  overwrite = FALSE,
  types = NULL,
  temporary = TRUE,
  unique_indexes = NULL,
  indexes = NULL,
  analyze = TRUE,
  ...,
  in_transaction = TRUE
) {
  check_bool(temporary)

  if (!is.data.frame(df) && !inherits(df, "tbl_sql")) {
    cli_abort("{.var df} must be a local dataframe or a remote tbl_sql")
  }

  name <- as_table_path(name, dest$con)

  if (inherits(df, "tbl_sql") && same_src(df$src, dest)) {
    out <- compute(
      df,
      name = name,
      temporary = temporary,
      overwrite = overwrite,
      unique_indexes = unique_indexes,
      indexes = indexes,
      analyze = analyze,
      ...
    )
  } else {
    # avoid S4 dispatch problem in dbSendPreparedQuery
    df <- as.data.frame(collect(df))

    name <- db_copy_to(
      dest$con,
      name,
      df,
      overwrite = overwrite,
      types = types,
      temporary = temporary,
      unique_indexes = unique_indexes,
      indexes = indexes,
      analyze = analyze,
      in_transaction = in_transaction,
      ...
    )
    out <- new_tbl_sql(dest$con, name, vars = names(df))
  }

  invisible(out)
}

#' @importFrom dplyr auto_copy
#' @export
auto_copy.tbl_sql <- function(x, y, copy = FALSE, ..., types = NULL) {
  # NB: copy is always TRUE here - dplyr's generic errors on copy = FALSE
  dbplyr_auto_copy(x, y, copy = "temp-table", ..., types = types)
}

# Use this instead of calling auto_copy() directly since auto_copy() expects
# TRUE or FALSE
dbplyr_auto_copy <- function(
  x,
  y,
  copy,
  ...,
  types = NULL,
  call = caller_env()
) {
  if (same_src(x, y)) {
    return(y)
  }

  copy <- as_copy(copy, error_call = call)
  if (copy == "none") {
    cli_abort(
      c(
        "{.arg x} and {.arg y} must share the same source.",
        i = "Use {.code copy = \"temp-table\"} to copy {.arg y} to a temporary table.",
        i = "USe {.code copy = \"inline\"} to convert {.arg y} to inline SQL."
      ),
      call = call
    )
  } else if (copy == "inline") {
    copy_inline(x$con, as.data.frame(y), types = types)
  } else if (copy == "temp-table") {
    y <- as.data.frame(y)
    table <- unique_table_name()
    table <- db_copy_to(x$con, table = table, values = y, ..., types = types)
    new_tbl_sql(con = x$con, source = table, vars = colnames(y))
  }
}

as_copy <- function(copy, error_call = caller_env()) {
  if (isTRUE(copy)) {
    "temp-table"
  } else if (isFALSE(copy)) {
    "none"
  } else {
    arg_match(copy, c("none", "temp-table", "inline"), error_call = error_call)
  }
}

#' Combine multiple lazy queries
#'
#' Combine multiple lazy queries into a single query using `UNION ALL`.
#' This is a convenient wrapper around `purrr::reduce(tables, union_all)`.
#' Like [dplyr::bind_rows()] (and unlike `UNION ALL`), `bind_queries()` will
#' automatically align columns based on their name, and fill in any missing
#' columns with missing values.
#'
#' @param ... <[dynamic-dots][rlang::dyn-dots]> Lazy tables to combine.
#' @returns A lazy query.
#' @export
#' @examples
#' lf1 <- lazy_frame(x = 1, y = "a")
#' lf2 <- lazy_frame(x = 2, y = "b")
#' bind_queries(lf1, lf2)
#'
#' lf3 <- lazy_frame(y = "c", x = 3, z = 10)
#' bind_queries(lf2, lf3)
#'
#' # If you already have a list, you can use splice operator
#' queries <- list(lf1, lf2)
#' bind_queries(!!!queries)
bind_queries <- function(...) {
  tables <- list2(...)
  if (length(tables) == 0) {
    cli_abort("{.fn bind_queries} requires at least one input.")
  }
  for (i in seq_along(tables)) {
    if (!is_tbl_lazy(tables[[i]])) {
      stop_input_type(tables[[i]], "a lazy query", arg = paste0("..", i))
    }
  }

  purrr::reduce(tables, union_all)
}

#' SQL generation generics
#'
#' @description
#'
#' SQL translation:
#'
#' * `sql_translation(con)` generates a SQL translation environment.
#'
#' * Deprecated: `sql_expr_matches(con, x, y)` previously generated an
#'   alternative to `x = y` when a pair of `NULL`s should match. This is now
#'   handled by the `is_not_distinct_from()` translation; provide a backend
#'   override there instead.
#'
#' * Deprecated: `sql_random(con)` generates SQL to get a random number which can be used
#'   to select random rows in `slice_sample()`. This is now replaced by adding
#'   a translation for `runif(n())`.
#'
#' Tables:
#'
#' * `sql_table_analyze(con, table)` generates SQL that "analyzes" the table,
#'   ensuring that the database has up-to-date statistics for use in the query
#'   planner. It called from [copy_to()] when `analyze = TRUE`.
#'
#' * `sql_table_index()` generates SQL for adding an index to table.
#'
#' Query manipulation:
#'
#' * `sql_query_explain(con, sql)` generates SQL that "explains" a query,
#'   i.e. generates a query plan describing what indexes etc that the
#'   database will use. It can return character vector, in which case the
#'   first n elements are used to generate the plan and the final element
#'   is used to return the query plan.
#'
#' * `sql_query_fields()` generates SQL for a 0-row result that is used to
#'   capture field names in [tbl_sql()]
#'
#' * `sql_query_save(con, sql)` generates SQL for saving a query into a
#'   (temporary) table.
#'
#' * `sql_query_wrap(con, from)` generates SQL for wrapping a query into a
#'   subquery.
#'
#' Query indentation:
#'
#' * `sql_indent_subquery(from, con, lvl)` helps indenting a subquery.
#'
#' Query generation:
#'
#' * `sql_query_select()` generates SQL for a `SELECT` query
#' * `sql_query_join()` generates SQL for joins
#' * `sql_query_semi_join()` generates SQL for semi- and anti-joins
#' * `sql_query_set_op()` generates SQL for `UNION`, `INTERSECT`, and `EXCEPT`
#'   queries.
#'
#' Query generation for manipulation:
#'
#' * `sql_query_insert()` and `sql_query_append()` generate SQL for an `INSERT FROM` query.
#' * `sql_query_update_from()` generates SQL for an `UPDATE FROM` query.
#' * `sql_query_upsert()` generates SQL for an `UPSERT` query.
#' * `sql_query_delete()` generates SQL for an `DELETE FROM` query
#' * `sql_returning_cols()` generates SQL for a `RETURNING` clause
#'
#' @keywords internal
#' @family generic
#' @name db-sql
NULL

#' @export
#' @rdname db-sql
sql_expr_matches <- function(con, x, y, ...) {
  lifecycle::deprecate_warn(
    "2.6.0",
    "sql_expr_matches()",
    with = I(
      "a translation for `is_not_distinct_from()` in `sql_translation()`"
    )
  )
  check_dots_used()
  UseMethod("sql_expr_matches", sql_dialect(con))
}

#' @export
sql_expr_matches.default <- function(con, x, y, ...) {
  NULL
}

#' @export
#' @rdname db-sql
sql_translation <- function(con) {
  UseMethod("sql_translation", sql_dialect(con))
}
# sql_translation.DBIConnection lives in backend-.R
dbplyr_sql_translation <- function(con) {
  check_2ed(con)
  sql_translation(con)
}

#' @export
#' @rdname db-sql
sql_random <- function(con) {
  lifecycle::deprecate_stop(
    "2.3.2",
    "sql_random()",
    with = I("Please add a translation for `runif(n())` instead.")
  )
}


# Tables ------------------------------------------------------------------

#' @rdname db-sql
#' @export
sql_table_analyze <- function(con, table, ...) {
  check_dots_used()
  UseMethod("sql_table_analyze", sql_dialect(con))
}
#' @export
sql_table_analyze.DBIConnection <- function(con, table, ...) {
  sql_glue2(con, "ANALYZE {.tbl {table}}")
}
#' @export
sql_table_analyze.sql_dialect <- sql_table_analyze.DBIConnection

#' @rdname db-sql
#' @export
sql_table_index <- function(
  con,
  table,
  columns,
  name = NULL,
  unique = FALSE,
  ...,
  call = caller_env()
) {
  check_table_id(table, call = call)
  check_character(columns, call = call)
  check_name(name, allow_null = TRUE, call = call)
  check_bool(unique, call = call)

  UseMethod("sql_table_index", sql_dialect(con))
}
#' @export
sql_table_index.DBIConnection <- function(
  con,
  table,
  columns,
  name = NULL,
  unique = FALSE,
  ...,
  call = caller_env()
) {
  table <- as_table_path(table, con)

  if (is.null(name)) {
    table_name <- table_path_name(table, con)
    name <- name %||% paste0(c(table_name, columns), collapse = "_")
  }

  index <- if (unique) "UNIQUE INDEX" else "INDEX"
  sql_glue2(
    con,
    "CREATE {.sql index} {.id name} ON {.tbl table} {.id columns*}"
  )
}

#' @export
sql_table_index.sql_dialect <- sql_table_index.DBIConnection

# Query manipulation ------------------------------------------------------

#' @rdname db-sql
#' @export
sql_query_explain <- function(con, sql, ...) {
  check_scalar_sql(sql)
  check_dots_used()
  UseMethod("sql_query_explain", sql_dialect(con))
}
#' @export
sql_query_explain.DBIConnection <- function(con, sql, ...) {
  sql_glue2(con, "EXPLAIN {sql}")
}

#' @export
sql_query_explain.sql_dialect <- sql_query_explain.DBIConnection

#' @rdname db-sql
#' @export
sql_query_fields <- function(con, sql, ...) {
  check_table_source(sql)
  check_dots_used()

  UseMethod("sql_query_fields", sql_dialect(con))
}
#' @export
sql_query_fields.DBIConnection <- function(con, sql, ...) {
  sql <- as_table_source(sql, con)
  sql_query_select(
    con,
    sql("*"),
    dbplyr_sql_subquery(con, sql),
    where = sql("0 = 1")
  )
}
#' @export
sql_query_fields.sql_dialect <- sql_query_fields.DBIConnection

#' @rdname db-sql
#' @export
sql_query_save <- function(con, sql, name, temporary = TRUE, ...) {
  check_table_id(name)
  check_bool(temporary)
  check_dots_used()

  UseMethod("sql_query_save", sql_dialect(con))
}
#' @export
sql_query_save.DBIConnection <- function(
  con,
  sql,
  name,
  temporary = TRUE,
  ...
) {
  sql <- as_table_source(sql)

  table <- if (temporary) "TEMPORARY TABLE" else "TABLE"
  sql_glue2(con, "CREATE {.sql table}\n{.tbl {name}} AS\n{sql}")
}

#' @export
sql_query_save.sql_dialect <- sql_query_save.DBIConnection
#' @export
#' @rdname db-sql
sql_query_wrap <- function(con, from, name = NULL, ..., lvl = 0) {
  check_name(name, allow_null = TRUE)
  check_dots_used()

  UseMethod("sql_query_wrap", sql_dialect(con))
}
#' @export
sql_query_wrap.DBIConnection <- function(con, from, name = NULL, ..., lvl = 0) {
  from <- as_table_source(from, con)

  if (is.sql(from)) {
    if (sql_has_table_alias_with_as(con)) {
      as_sql <- style_kw("AS ")
    } else {
      as_sql <- ""
    }

    from <- sql_indent_subquery(from, con, lvl)
    # some backends, e.g. Postgres, require an alias for a subquery
    name <- name %||% unique_subquery_name()
    sql_glue2(con, "{from} {.sql as_sql}{.tbl name}")
  } else {
    # must be a table_path
    if (!is.null(name)) {
      table <- table_path_name(name, con)
      names(from) <- as_table_path(table, con)
    }
    sql_escape_table_source(con, from)
  }
}

#' @export
sql_query_wrap.sql_dialect <- sql_query_wrap.DBIConnection

#' @export
#' @rdname db-sql
sql_indent_subquery <- function(from, con, lvl = 0) {
  multi_line <- grepl(x = from, pattern = "\\r\\n|\\r|\\n")
  if (multi_line) {
    out <- paste0(
      "(\n",
      from,
      "\n",
      indent_lvl(")", lvl)
    )
  } else {
    # Strip indent
    from <- gsub("^ +", "", from)
    out <- paste0("(", from, ")")
  }

  sql(out)
}

#' @rdname db-sql
#' @export
sql_query_rows <- function(con, sql, ...) {
  check_table_source(sql)
  check_dots_used()

  UseMethod("sql_query_rows", sql_dialect(con))
}
#' @export
sql_query_rows.DBIConnection <- function(con, sql, ...) {
  sql <- as_table_source(sql, con)
  from <- dbplyr_sql_subquery(con, sql, "master")
  sql_glue2(con, "SELECT COUNT(*) FROM {.tbl from}")
}

#' @export
sql_query_rows.sql_dialect <- sql_query_rows.DBIConnection

#' Generate SQL for Insert, Update, Upsert, and Delete
#'
#' These functions generate the SQL used in `rows_*(in_place = TRUE)`.
#'
#' @param con A [sql_dialect] object or database connection. Connections are
#'   supported for backward compatibility.
#' @param table Table to update. Must be a table identifier.
#'   Use a string to refer to tables in the current schema/catalog or
#'   `I()` to refer to tables in other schemas/catalogs.
#' @param from Table or query that contains the new data. Either a table
#'   identifier or SQL.
#' @inheritParams dplyr::rows_upsert
#' @param insert_cols Names of columns to insert.
#' @param update_cols Names of columns to update.
#' @param update_values A named SQL vector that specify how to update the columns.
#' @param ... Other parameters passed onto methods.
#' @param returning_cols Optional. Names of columns to return.
#' @param method Optional. The method to use.
#'
#' @details Insert Methods
#' ## `"where_not_exists"`
#' The default for most databases.
#'
#' ```
#' INSERT INTO x_name
#' SELECT *
#' FROM y
#' WHERE NOT EXISTS <match on by columns>
#' ```
#'
#' ## `"on_conflict"`
#' Supported by:
#' * Postgres
#' * SQLite
#'
#' This method uses the `ON CONFLICT` clause and therefore requires a unique
#' index on the columns specified in `by`.
#'
#' @details Upsert Methods
#'
#' ## `"merge"`
#' The upsert method according to the SQL standard. It uses the `MERGE` statement
#'
#' ```
#' MERGE INTO x_name
#' USING y
#'   ON <match on by columns>
#' WHEN MATCHED THEN
#'   UPDATE SET ...
#' WHEN NOT MATCHED THEN
#'   INSERT ...
#' ```
#'
#' ## `"on_conflict"`
#' Supported by:
#' * Postgres
#' * SQLite
#'
#' This method uses the `ON CONFLICT` clause and therefore requires a unique
#' index on the columns specified in `by`.
#'
#' ## `"cte_update"`
#' Supported by:
#' * Postgres
#' * SQLite
#' * Oracle
#'
#' The classical way to upsert in Postgres and SQLite before support for
#' `ON CONFLICT` was added. The update is done in a CTE clause and the unmatched
#' values are then inserted outside of the CTE.
#'
#' @return A SQL query.
#' @export
#'
#' @examples
#' sql_query_upsert(
#'   con = dialect_postgres(),
#'   table = "airlines",
#'   from = "df",
#'   by = "carrier",
#'   update_cols = "name"
#' )
sql_query_insert <- function(
  con,
  table,
  from,
  insert_cols,
  by,
  ...,
  conflict = c("error", "ignore"),
  returning_cols = NULL,
  method = NULL
) {
  check_table_id(table)
  check_table_source(from)
  check_character(insert_cols)
  check_character(by)
  check_character(returning_cols, allow_null = TRUE)

  check_dots_used()
  UseMethod("sql_query_insert", sql_dialect(con))
}

#' @export
sql_query_insert.DBIConnection <- function(
  con,
  table,
  from,
  insert_cols,
  by,
  ...,
  conflict = c("error", "ignore"),
  returning_cols = NULL,
  method = NULL
) {
  table <- as_table_path(table, con)
  from <- as_table_source(from, con)

  method <- method %||% "where_not_exists"
  arg_match(method, "where_not_exists", error_arg = "method")
  # https://stackoverflow.com/questions/25969/insert-into-values-select-from
  conflict <- rows_check_conflict(conflict)

  parts <- rows_insert_prep(con, table, from, insert_cols, by, lvl = 0)

  clauses <- list2(
    parts$insert_clause,
    sql_clause_select(sql("*")),
    sql_clause_from(parts$from),
    !!!parts$conflict_clauses,
    sql_returning_cols(con, returning_cols, table)
  )

  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_insert.sql_dialect <- sql_query_insert.DBIConnection

#' @export
#' @rdname sql_query_insert
sql_query_append <- function(
  con,
  table,
  from,
  insert_cols,
  ...,
  returning_cols = NULL
) {
  if (is_tbl_lazy(from)) {
    lifecycle::deprecate_stop(
      when = "2.3.2",
      what = "sql_query_append(from = 'must be a table identifier or an SQL query, not a lazy table.')"
    )
  }

  check_table_id(table)
  check_table_source(from)
  check_character(insert_cols)
  check_character(returning_cols, allow_null = TRUE)

  check_dots_used()
  UseMethod("sql_query_append", sql_dialect(con))
}

#' @export
sql_query_append.DBIConnection <- function(
  con,
  table,
  from,
  insert_cols,
  ...,
  returning_cols = NULL
) {
  table <- as_table_path(table, con)
  from <- as_table_source(from, con)

  # https://stackoverflow.com/questions/25969/insert-into-values-select-from
  parts <- rows_prep(con, table, from, by = list(), lvl = 0)

  insert_cols_sql <- sql_escape_ident(con, insert_cols)
  table_sql <- sql_escape_table_source(con, table)
  clauses <- list2(
    sql_clause_insert(insert_cols_sql, table_sql),
    sql_clause_select(sql("*")),
    sql_clause_from(parts$from),
    sql_returning_cols(con, returning_cols, table)
  )

  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_append.sql_dialect <- sql_query_append.DBIConnection

#' @export
#' @rdname sql_query_insert
sql_query_update_from <- function(
  con,
  table,
  from,
  by,
  update_values,
  ...,
  returning_cols = NULL
) {
  check_table_id(table)
  check_table_source(from)
  check_character(by)
  check_character(update_values)
  check_named(update_values)
  check_character(returning_cols, allow_null = TRUE)

  check_dots_used()
  UseMethod("sql_query_update_from", sql_dialect(con))
}

#' @export
sql_query_update_from.DBIConnection <- function(
  con,
  table,
  from,
  by,
  update_values,
  ...,
  returning_cols = NULL
) {
  table <- as_table_path(table, con)
  from <- as_table_source(from, con)

  # https://stackoverflow.com/questions/2334712/how-do-i-update-from-a-select-in-sql-server
  parts <- rows_prep(con, table, from, by, lvl = 0)
  update_cols <- sql_escape_ident(con, names(update_values))

  table_sql <- sql_escape_table_source(con, table)
  # avoid CTEs for the general case as they do not work everywhere
  clauses <- list(
    sql_clause_update(table_sql),
    sql_clause_set(update_cols, update_values),
    sql_clause_from(parts$from),
    sql_clause_where(parts$where),
    sql_returning_cols(con, returning_cols, table)
  )
  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_update_from.sql_dialect <- sql_query_update_from.DBIConnection


#' @export
#' @rdname sql_query_insert
sql_query_upsert <- function(
  con,
  table,
  from,
  by,
  update_cols,
  ...,
  returning_cols = NULL,
  method = NULL
) {
  check_table_id(table)
  check_table_source(from)
  check_character(by)
  check_character(update_cols)
  check_character(returning_cols, allow_null = TRUE)

  # https://wiki.postgresql.org/wiki/UPSERT#SQL_MERGE_syntax
  # https://github.com/cynkra/dm/pull/616#issuecomment-920613435
  check_dots_used()
  UseMethod("sql_query_upsert", sql_dialect(con))
}

#' @export
sql_query_upsert.DBIConnection <- function(
  con,
  table,
  from,
  by,
  update_cols,
  ...,
  returning_cols = NULL,
  method = NULL
) {
  table <- as_table_path(table, con)
  from <- as_table_source(from, con)

  method <- method %||% "cte_update"
  arg_match(method, "cte_update", error_arg = "method")

  parts <- rows_prep(con, table, from, by, lvl = 0)

  insert_cols <- c(by, update_cols)
  insert_cols_sql <- sql_escape_ident(con, insert_cols)

  update_values <- sql_table_prefix(con, "...y", update_cols)
  update_cols <- sql_escape_ident(con, update_cols)

  table_sql <- sql_escape_table_source(con, table)
  updated_cte <- list(
    sql_clause_update(table_sql),
    sql_clause_set(update_cols, update_values),
    sql_clause_from(parts$from),
    sql_clause_where(parts$where),
    sql(paste0("RETURNING ", sql_star(con, table)))
  )
  updated_sql <- sql_format_clauses(updated_cte, lvl = 1)
  update_name <- sql_escape_ident(con, "updated")

  join_by <- new_join_by(by, x_as = "updated", y_as = "...y")
  where <- sql_join_tbls(con, by = join_by, na_matches = "never")

  clauses <- list2(
    sql(paste0("WITH ", update_name, " AS (")),
    updated_sql,
    sql(")"),
    sql_clause_insert(insert_cols_sql, table_sql),
    sql_clause_select(sql("*")),
    sql_clause_from(parts$from),
    !!!sql_clause_where_exists(update_name, where, not = TRUE),
    sql_returning_cols(con, returning_cols, table)
  )

  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_upsert.sql_dialect <- sql_query_upsert.DBIConnection

#' @export
#' @rdname sql_query_insert
sql_query_delete <- function(con, table, from, by, ..., returning_cols = NULL) {
  check_table_id(table)
  check_table_source(from)
  check_character(by)
  check_character(returning_cols, allow_null = TRUE)

  check_dots_used()
  UseMethod("sql_query_delete", sql_dialect(con))
}

#' @export
sql_query_delete.DBIConnection <- function(
  con,
  table,
  from,
  by,
  ...,
  returning_cols = NULL
) {
  table <- as_table_path(table, con)
  from <- as_table_source(from, con)
  parts <- rows_prep(con, table, from, by, lvl = 1)

  table_sql <- sql_escape_table_source(con, table)
  clauses <- list2(
    sql_clause("DELETE FROM", table_sql),
    !!!sql_clause_where_exists(parts$from, parts$where, not = FALSE),
    sql_returning_cols(con, returning_cols, table)
  )
  sql_format_clauses(clauses, lvl = 0)
}

#' @export
sql_query_delete.sql_dialect <- sql_query_delete.DBIConnection

#' @export
#' @rdname db-sql
sql_returning_cols <- function(con, cols, table, ...) {
  if (is_empty(cols)) {
    return(NULL)
  }

  check_dots_used()
  UseMethod("sql_returning_cols", sql_dialect(con))
}

#' @export
sql_returning_cols.DBIConnection <- function(con, cols, table, ...) {
  returning_cols <- sql_named_cols(con, cols, table = table)

  sql_clause("RETURNING", returning_cols)
}

#' @export
sql_returning_cols.sql_dialect <- sql_returning_cols.DBIConnection

sql_named_cols <- function(con, cols, table = NULL) {
  nms <- names2(cols)
  nms[nms == cols] <- ""

  cols <- sql_table_prefix(con, table, cols)
  names_to_as(con, cols, nms)
}

# dplyr fallbacks ---------------------------------------------------------

dbplyr_analyze <- function(con, table, ...) {
  check_2ed(con)

  sql <- sql_table_analyze(con, table, ...)
  if (is.null(sql)) {
    return() # nocov
  }

  db_execute(
    con,
    sql,
    "Can't analyze table {.field {format(table, con = con)}}."
  )
}

dbplyr_create_index <- function(
  con,
  table,
  columns,
  name = NULL,
  unique = FALSE,
  ...
) {
  check_2ed(con)
  sql <- sql_table_index(con, table, columns, name = name, unique = unique, ...)
  db_execute(
    con,
    sql,
    "Can't create index on table {.field {format(table, con = con)}}."
  )
}

dbplyr_explain <- function(con, sql, ...) {
  check_2ed(con)

  sql <- sql_query_explain(con, sql, ...)
  n <- length(sql)
  for (i in seq_len(n - 1)) {
    db_execute(con, sql[[i]], "Can't explain query.")
  }
  expl <- db_get_query(con, sql[[n]], "Can't explain query.")

  out <- utils::capture.output(print(expl))
  paste(out, collapse = "\n")
}


dbplyr_query_fields <- function(con, sql, ...) {
  check_2ed(con)

  sql <- sql_query_fields(con, sql, ...)
  df <- db_get_query(con, sql, "Can't query fields.")
  names(df)
}

dbplyr_save_query <- function(
  con,
  sql,
  name,
  temporary = TRUE,
  ...,
  overwrite = FALSE
) {
  check_2ed(con)

  name <- as_table_path(name, con)
  sql <- sql_query_save(con, sql, name, temporary = temporary, ...)

  if (overwrite) {
    db_table_drop_if_exists(con, name)
  }

  db_execute(
    con,
    sql,
    "Can't save query to table {.table {format(name, con = con)}}."
  )

  name
}

dbplyr_sql_subquery <- function(
  con,
  from,
  name = unique_subquery_name(),
  ...,
  lvl = 0
) {
  check_2ed(con)
  sql_query_wrap(con, from = from, name = name, ..., lvl = lvl)
}

# Helpers -------------------------------------------------------------------

db_execute <- function(con, sql, msg, call = caller_env(), env = caller_env()) {
  dbi_wrap(
    DBI::dbExecute(con, sql, immediate = TRUE),
    sql = sql,
    msg = msg,
    call = call,
    env = env
  )
  invisible()
}

db_get_query <- function(
  con,
  sql,
  msg,
  call = caller_env(),
  env = caller_env()
) {
  dbi_wrap(DBI::dbGetQuery(con, sql), sql, msg, call = call, env = env)
}

dbi_wrap <- function(code, sql, msg, call = caller_env(), env = caller_env()) {
  withCallingHandlers(
    code,
    error = function(cnd) {
      msg <- c(msg, i = paste0("Using SQL: ", sql))
      cli_abort(msg, parent = cnd, call = call, .envir = env)
    }
  )
}

# rows_insert -------------------------------------------------------------

test_that("rows_insert() checks arguments", {
  lf <- lazy_frame(x = 1:3, y = 11:13, .name = "df_x")

  expect_snapshot(error = TRUE, {
    (rows_insert(lf, lf, by = "x", conflict = "error"))
    (rows_insert(lf, lf, by = "x"))
  })

  expect_snapshot(error = TRUE,
    (rows_insert(
      lf,
      lazy_frame(x = 1, y = 2, z = 3),
      by = "x",
      conflict = "ignore"
    ))
  )

  expect_snapshot(error = TRUE,
    (rows_insert(lf, lf, by = "x", conflict = "ignore", returning = everything()))
  )

  df <- memdb_frame(x = 1)
  expect_snapshot(error = TRUE,
    (df %>%
      mutate(x = x + 1) %>%
      rows_insert(df, by = "x", conflict = "ignore", in_place = TRUE))
  )

  expect_snapshot(error = TRUE,
    (df %>%
      rows_insert(df, by = "x", conflict = "ignore", returning = c(y)))
  )
})


test_that("`rows_insert()` works with empty `by`", {
  expect_message(
    rows_insert(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 1:3, .name = "df_y"),
      in_place = FALSE,
      conflict = "ignore"
    ),
    regexp = 'Matching, by = "x"'
  )
})

test_that("`rows_insert()` errors for `conflict = 'error'` and `in_place = FALSE`", {
  lf <- lazy_frame(x = 1:3, y = 11:13, .name = "df_x")
  expect_snapshot(
    (expect_error(
      rows_insert(
        lf, lf,
        by = "x",
        conflict = "error",
        in_place = FALSE
      )
    ))
  )
})

test_that("`rows_insert()` works with `in_place = FALSE`", {
  expect_snapshot(
    rows_insert(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 3:4, y = 23:24, .name = "df_y"),
      by = "x",
      conflict = "ignore",
      in_place = FALSE
    )
  )

  df <- memdb_frame(x = 1:3, y = 11:13)
  expect_equal(
    rows_insert(
      df, memdb_frame(x = 3:4, y = 23:24),
      by = "x",
      conflict = "ignore",
      in_place = FALSE
    ) %>%
      collect(),
    tibble(x = 1:4, y = c(11:13, 24))
  )

  expect_equal(df %>% collect(), tibble(x = 1:3, y = 11:13))
})

test_that("`rows_insert()` works with `in_place = FALSE` and `returning`", {
  expect_equal(
    rows_insert(
      memdb_frame(x = 1:3, y = 11:13),
      memdb_frame(x = 3:4, y = 23:24),
      by = "x",
      conflict = "ignore",
      in_place = FALSE,
      returning = everything()
    ) %>%
      get_returned_rows(),
    tibble(x = 4, y = 24)
  )

  # all `x` columns are present
  expect_equal(
    rows_insert(
      memdb_frame(x = 1:3, y = 11:13),
      memdb_frame(x = 3:4),
      by = "x",
      conflict = "ignore",
      in_place = FALSE,
      returning = everything()
    ) %>%
      get_returned_rows(),
    tibble(x = 4, y = NA)
  )
})

test_that("`rows_insert()` works with `in_place = TRUE`", {
  expect_snapshot(error = TRUE,
    (rows_insert(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 2:3, y = 22:23, .name = "df_y"),
      by = "x",
      in_place = TRUE
    ))
  )

  df <- memdb_frame(x = 1:3, y = 11:13)
  rows_insert(
    df, memdb_frame(x = 3:4, y = 23:24),
    by = "x",
    conflict = "ignore",
    in_place = TRUE
  )
  expect_equal(df %>% collect(), tibble(x = 1:4, y = c(11:13, 24)))
})

test_that("`rows_insert()` with `in_place = TRUE` and `returning`", {
  skip_if_not_installed("RSQLite", "2.2.8")

  df <- memdb_frame(x = 1:3, y = 11:13)
  df_inserted <- rows_insert(
    df, memdb_frame(x = 3:4, y = 23:24),
    by = "x",
    conflict = "ignore",
    in_place = TRUE,
    returning = everything()
  )

  expect_equal(get_returned_rows(df_inserted), tibble(x = 4L, y = 24L))

  expect_equal(df_inserted %>% collect(), tibble(x = 1:4, y = c(11:13, 24L)))
})

test_that("rows_get_or_execute() gives error context", {
  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(con))
  DBI::dbWriteTable(con, "mtcars", tibble(x = 1, y = 1), overwrite = TRUE, temporary = TRUE)
  DBI::dbExecute(con, "CREATE UNIQUE INDEX `mtcars_x` ON `mtcars` (`x`)")

  expect_snapshot({
    (expect_error(
      rows_append(
        tbl(con, "mtcars"),
        tibble(x = 1),
        copy = TRUE,
        in_place = TRUE
      )
    ))

    (expect_error(
      rows_append(
        tbl(con, "mtcars"),
        tibble(x = 1),
        copy = TRUE,
        in_place = TRUE,
        returning = x
      )
    ))
  }, transform = snap_transform_dbi)
})

test_that("`sql_query_insert()` works", {
  df_y <- lazy_frame(
    a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"),
    con = simulate_dbi(),
    .name = "df_y"
  ) %>%
    mutate(c = c + 1)

  expect_snapshot(error = TRUE,
    (sql_query_insert(
      con = simulate_dbi(),
      x_name = ident("df_x"),
      y = df_y,
      by = c("a", "b"),
      conflict = "error",
      returning_cols = c("a", b2 = "b")
    ))
  )

  expect_snapshot(
    sql_query_insert(
      con = simulate_dbi(),
      x_name = ident("df_x"),
      y = df_y,
      by = c("a", "b"),
      conflict = "ignore",
      returning_cols = c("a", b2 = "b")
    )
  )
})

# rows_append -------------------------------------------------------------

test_that("rows_append() checks arguments", {
  lf <- lazy_frame(x = 1:3, y = 11:13, .name = "df_x")
  expect_snapshot(error = TRUE, {
    (lf %>% rows_append(df, by = "x"))
    (lf %>% rows_append(df, conflict = "error"))
  })
})


test_that("`rows_append()` works with `in_place = FALSE`", {
  expect_snapshot(
    rows_append(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 3:4, y = 23:24, .name = "df_y"),
      in_place = FALSE
    )
  )

  df <- memdb_frame(x = 1:3, y = 11:13)
  expect_equal(
    rows_append(
      df, memdb_frame(x = 3:4, y = 23:24),
      in_place = FALSE
    ) %>%
      collect(),
    tibble(x = c(1:3, 3:4), y = c(11:13, 23:24))
  )

  expect_equal(df %>% collect(), tibble(x = 1:3, y = 11:13))
})

test_that("`rows_append()` works with `in_place = FALSE` and `returning`", {
  expect_equal(
    rows_append(
      memdb_frame(x = 1:3, y = 11:13),
      memdb_frame(x = 3:4, y = 23:24),
      in_place = FALSE,
      returning = everything()
    ) %>%
      get_returned_rows(),
    tibble(x = 3:4, y = 23:24)
  )

  # all `x` columns are present
  expect_equal(
    rows_append(
      memdb_frame(x = 1:3, y = 11:13),
      memdb_frame(x = 3:4),
      in_place = FALSE,
      returning = everything()
    ) %>%
      get_returned_rows(),
    tibble(x = 3:4, y = NA)
  )
})

test_that("`rows_append()` works with `in_place = TRUE`", {
  expect_snapshot(error = TRUE,
    (rows_append(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 2:3, y = 22:23, .name = "df_y"),
      in_place = TRUE
    ))
  )

  df <- memdb_frame(x = 1:3, y = 11:13)
  rows_append(
    df, memdb_frame(x = 3:4, y = 23:24),
    in_place = TRUE
  )
  expect_equal(df %>% collect(), tibble(x = c(1:3, 3:4), y = c(11:13, 23:24)))
})

test_that("`rows_append()` with `in_place = TRUE` and `returning`", {
  skip_if_not_installed("RSQLite", "2.2.8")

  df <- memdb_frame(x = 1:3, y = 11:13)
  df_inserted <- rows_append(
    df, memdb_frame(x = 3:4, y = 23:24),
    in_place = TRUE,
    returning = everything()
  )

  expect_equal(get_returned_rows(df_inserted), tibble(x = 3:4, y = 23:24))

  expect_equal(df_inserted %>% collect(), tibble(x = c(1:3, 3:4), y = c(11:13, 23:24)))
})

test_that("`sql_query_append()` works", {
  df_y <- lazy_frame(
    a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"),
    con = simulate_dbi(),
    .name = "df_y"
  ) %>%
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_append(
      con = simulate_dbi(),
      x_name = ident("df_x"),
      y = df_y,
      returning_cols = c("a", b2 = "b")
    )
  )
})

# rows_update -------------------------------------------------------------

test_that("arguments are checked", {
  lf <- lazy_frame(x = 1:3, y = 11:13, .name = "df_x")

  expect_snapshot(error = TRUE, (rows_update(lf, lf, by = 1, unmatched = "ignore")))

  expect_snapshot(error = TRUE, (rows_update(lf, lf, by = c(y = "x"), unmatched = "ignore")))

  expect_snapshot(error = TRUE, (rows_update(lf, lf, by = "z", unmatched = "ignore")))

  expect_snapshot(error = TRUE, {
    (rows_update(lf, lf, by = "x", unmatched = "error"))
    (rows_update(lf, lf, by = "x"))
  })

  expect_snapshot(error = TRUE,
    (rows_update(
      lf,
      lazy_frame(x = 1, y = 2, z = 3),
      by = "x",
      unmatched = "ignore"
    ))
  )

  expect_snapshot(error = TRUE,
    (rows_update(lf, lf, by = "x", unmatched = "ignore", returning = everything()))
  )

  df <- memdb_frame(x = 1)
  expect_snapshot(error = TRUE,
    (df %>%
      mutate(x = x + 1) %>%
      rows_update(df, by = "x", unmatched = "ignore", in_place = TRUE))
  )
})

test_that("`rows_update()` returns early if no column to update", {
  lf <- lazy_frame(x = 1:3, y = 11:13, .name = "df_x")
  expect_equal(
    rows_update(
      lf,
      lazy_frame(x = 1:3, .name = "df_y"),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    ),
    lf
  )

  db <- memdb_frame(x = 1:3, y = 11:13)
  expect_equal(
    rows_update(
      db,
      memdb_frame(x = 1:3),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    ) %>%
      collect(),
    collect(db)
  )

  expect_equal(
    rows_update(
      db,
      memdb_frame(x = 1:3),
      by = "x",
      unmatched = "ignore",
      in_place = TRUE
    ) %>%
      collect(),
    collect(db)
  )
})

test_that("`rows_update()` works with empty `by`", {
  expect_message(
    rows_update(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 1:3, .name = "df_y"),
      unmatched = "ignore",
      in_place = FALSE
    ),
    regexp = 'Matching, by = "x"'
  )
})

test_that("`rows_update()` works with `in_place = FALSE`", {
  expect_snapshot(
    rows_update(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 2:3, y = 22:23, .name = "df_y"),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    )
  )

  df <- memdb_frame(x = 1:3, y = 11:13)
  expect_equal(
    rows_update(
      df, memdb_frame(x = 2:3, y = 22:23),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    ) %>%
      collect(),
    tibble(x = 1:3, y = c(11L, 22:23))
  )

  expect_equal(df %>% collect(), tibble(x = 1:3, y = 11:13))
})

test_that("`rows_update()` works with `in_place = FALSE` and `returning`", {
  expect_equal(
    rows_update(
      memdb_frame(x = 1:3, y = 11:13),
      memdb_frame(x = 2:3, y = 22:23),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE,
      returning = everything()
    ) %>%
      get_returned_rows(),
    tibble(x = 2:3, y = 22:23)
  )
})

test_that("`rows_update()` works with `in_place = TRUE`", {
  expect_snapshot(error = TRUE,
    (rows_update(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 2:3, y = 22:23, .name = "df_y"),
      by = "x",
      unmatched = "ignore",
      in_place = TRUE
    ))
  )

  df <- memdb_frame(x = 1:3, y = 11:13)
  rows_update(
    df, memdb_frame(x = 2:3, y = 22:23),
    by = "x",
    unmatched = "ignore",
    in_place = TRUE
  )

  expect_equal(df %>% collect(), tibble(x = 1:3, y = c(11L, 22:23)))
})

test_that("`rows_update()` with `in_place = TRUE` and `returning`", {
  skip_if_not_installed("RSQLite", "2.2.8")

  df <- memdb_frame(x = 1:3, y = 11:13)
  df_updated <- rows_update(
    df, memdb_frame(x = 2:4, y = 22:24),
    by = "x",
    unmatched = "ignore",
    in_place = TRUE,
    returning = everything()
  )

  expect_equal(get_returned_rows(df_updated), tibble(x = 2:3, y = 22:23))

  expect_equal(df_updated %>% collect(), tibble(x = 1:3, y = c(11L, 22:23)))
})

test_that("`sql_query_update_from()` works", {
  df_y <- lazy_frame(
    a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"),
    con = simulate_dbi(),
    .name = "df_y"
  ) %>%
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_update_from(
      con = simulate_dbi(),
      x_name = ident("df_x"),
      y = df_y,
      by = c("a", "b"),
      update_values = sql(
        c = "COALESCE(`df_x`.`c`, `...y`.`c`)",
        d = "`...y`.`d`"
      ),
      returning_cols = c("a", b2 = "b")
    )
  )
})

# rows_patch --------------------------------------------------------------

test_that("`rows_patch()` returns early if no column to update", {
  expect_snapshot(
    rows_patch(
      lazy_frame(x = 1:3, y = c(11, 12, NA), .name = "df_x"),
      lazy_frame(x = 1:3, .name = "df_y"),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    )
  )

  expect_equal(
    rows_patch(
      memdb_frame(x = 1:3, y = c(11, 12, NA)),
      memdb_frame(x = 1:3),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    ) %>%
      collect(),
    tibble(x = 1:3, y = c(11, 12, NA))
  )

  expect_equal(
    rows_patch(
      memdb_frame(x = 1:3, y = c(11, 12, NA)),
      memdb_frame(x = 1:3),
      by = "x",
      unmatched = "ignore",
      in_place = TRUE
    ) %>%
      collect(),
    tibble(x = 1:3, y = c(11, 12, NA))
  )
})

test_that("`rows_patch()` works with empty `by`", {
  expect_message(
    rows_patch(
      lazy_frame(x = 1:3, y = c(11, 12, NA), .name = "df_x"),
      lazy_frame(x = 1:3, .name = "df_y"),
      unmatched = "ignore",
      in_place = FALSE
    ),
    regexp = 'Matching, by = "x"'
  )
})

test_that("`rows_patch()` works with `in_place = FALSE`", {
  expect_snapshot(
    rows_patch(
      lazy_frame(x = 1:3, y = c(11, 12, NA), .name = "df_x"),
      lazy_frame(x = 2:3, y = 22:23, .name = "df_y"),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    )
  )

  df <- memdb_frame(x = 1:3, y = c(11, 12, NA))
  expect_equal(
    rows_patch(
      df, memdb_frame(x = 2:3, y = 22:23),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    ) %>%
      collect(),
    tibble(x = 1:3, y = c(11, 12, 23))
  )

  expect_equal(df %>% collect(), tibble(x = 1:3, y = c(11, 12, NA)))
})

test_that("`rows_patch()` works with `in_place = FALSE` and `returning`", {
  expect_equal(
    rows_patch(
      memdb_frame(x = 1:3, y = c(11, 12, NA)),
      memdb_frame(x = 2:3, y = 22:23),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE,
      returning = everything()
    ) %>%
      get_returned_rows(),
    tibble(x = 2:3, y = c(12, 23))
  )
})

test_that("`rows_patch()` works with `in_place = TRUE`", {
  expect_snapshot(error = TRUE,
    (rows_patch(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 2:3, y = 22:23, .name = "df_y"),
      by = "x",
      unmatched = "ignore",
      in_place = TRUE
    ))
  )

  df <- memdb_frame(x = 1:3, y = c(11, 12, NA))
  rows_patch(
    df, memdb_frame(x = 2:3, y = 22:23),
    by = "x",
    unmatched = "ignore",
    in_place = TRUE
  )

  expect_equal(df %>% collect(), tibble(x = 1:3, y = c(11, 12, 23)))
})

test_that("`rows_patch()` works with `in_place = TRUE` and `returning`", {
  skip_if_not_installed("RSQLite", "2.2.8")

  df <- memdb_frame(x = 1:3, y = c(11, 12, NA))
  df_patched <- rows_patch(
    df, memdb_frame(x = 2:4, y = 22:24),
    by = "x",
    unmatched = "ignore",
    in_place = TRUE,
    returning = everything()
  )

  expect_equal(get_returned_rows(df_patched), tibble(x = 2:3, y = c(12, 23)))

  expect_equal(df_patched %>% collect(), tibble(x = 1:3, y = c(11, 12, 23)))
})


# rows_upsert -------------------------------------------------------------

test_that("`rows_upsert()` returns early if no column to update", {
  expect_snapshot(
    rows_upsert(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 1:3, .name = "df_y"),
      by = "x",
      in_place = FALSE
    )
  )

  expect_equal(
    rows_upsert(
      memdb_frame(x = 1:3, y = 11:13),
      memdb_frame(x = 2:4),
      by = "x",
      in_place = FALSE
    ) %>%
      collect(),
    tibble(x = 1:4, y = c(11:13, NA))
  )

  expect_equal(
    rows_upsert(
      memdb_frame(x = 1:3, y = 11:13),
      memdb_frame(x = 2:4),
      by = "x",
      in_place = TRUE
    ) %>%
      collect(),
    tibble(x = 1:3, y = 11:13)
  )
})

test_that("`rows_upsert()` works with empty `by`", {
  expect_message(
    rows_upsert(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 1:3, .name = "df_y"),
      in_place = FALSE
    ),
    regexp = 'Matching, by = "x"'
  )
})

test_that("`rows_upsert()` works with `in_place = FALSE`", {
  expect_snapshot(
    rows_upsert(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 2:3, y = 22:23, .name = "df_y"),
      by = "x",
      in_place = FALSE
    )
  )

  df <- memdb_frame(x = 1:3, y = 11:13)
  expect_equal(
    rows_upsert(
      df, memdb_frame(x = 2:4, y = 22:24),
      by = "x",
      in_place = FALSE
    ) %>%
      collect(),
    tibble(x = 1:4, y = c(11L, 22:24))
  )

  expect_equal(df %>% collect(), tibble(x = 1:3, y = 11:13))
})

test_that("`rows_upsert()` works with `in_place = FALSE` and `returning`", {
  expect_equal(
    rows_upsert(
      memdb_frame(x = 1:3, y = 11:13),
      memdb_frame(x = 2:4, y = 22:24),
      by = "x",
      in_place = FALSE,
      returning = everything()
    ) %>%
      get_returned_rows(),
    tibble(x = 2:4, y = 22:24)
  )
})

test_that("`rows_upsert()` works with `in_place = TRUE`", {
  expect_snapshot(error = TRUE,
    (rows_upsert(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 2:3, y = 22:23, .name = "df_y"),
      by = "x",
      in_place = TRUE
    ))
  )

  skip_if_not_installed("RSQLite", "2.2.8")
  con <- src_memdb()
  x <- copy_to(con, tibble(x = 1:3, y = 11:13), name = "df_x", overwrite = TRUE, unique_indexes = "x")
  y <- copy_to(con, tibble(x = 2:4, y = 22:24), name = "df_y", overwrite = TRUE)

  rows_upsert(
    x, y,
    by = "x",
    in_place = TRUE
  )

  expect_equal(x %>% collect(), tibble(x = 1:4, y = c(11L, 22:24)))
})

test_that("`rows_upsert()` works with `in_place = TRUE` and `returning`", {
  skip_if_not_installed("RSQLite", "2.2.8")

  con <- src_memdb()
  x <- copy_to(con, tibble(x = 1:3, y = 11:13), name = "df_x", overwrite = TRUE, unique_indexes = "x")
  y <- copy_to(con, tibble(x = 2:4, y = 22:24), name = "df_y", overwrite = TRUE)
  df_upserted <- rows_upsert(
    x, y,
    by = "x",
    in_place = TRUE,
    returning = everything()
  )

  expect_equal(get_returned_rows(df_upserted), tibble(x = 4, y = 24))

  expect_equal(df_upserted %>% collect(), tibble(x = 1:4, y = c(11L, 22:24)))
})

test_that("`sql_query_upsert()` is correct", {
  df_y <- lazy_frame(
    a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"),
    con = simulate_dbi(),
    .name = "df_y"
  ) %>%
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_upsert(
      con = simulate_dbi(),
      x_name = ident("df_x"),
      y = df_y,
      by = c("a", "b"),
      update_cols = c("c", "d"),
      returning_cols = c("a", b2 = "b")
    )
  )
})

# rows_delete -------------------------------------------------------------

test_that("`rows_delete()` works with empty `by`", {
  expect_message(
    rows_delete(
      lazy_frame(x = 1:3, y = c(11, 12, NA), .name = "df_x"),
      lazy_frame(x = 1:3, .name = "df_y"),
      unmatched = "ignore",
      in_place = FALSE
    ),
    regexp = 'Matching, by = "x"'
  )
})

test_that("`rows_delete()` ignores extra `y` columns", {
  expect_message(
    rows_delete(
      lazy_frame(x = 1:3, y = c(11, 12, NA), .name = "df_x"),
      lazy_frame(x = 1:3, y = 11, .name = "df_y"),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    ),
    regexp = 'Ignoring extra `y` columns: `y`'
  )
})

test_that("`rows_delete()` works with `in_place = FALSE`", {
  expect_snapshot(
    rows_delete(
      lazy_frame(x = 1:3, y = c(11, 12, NA), .name = "df_x"),
      lazy_frame(x = 2:3, .name = "df_y"),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    )
  )

  df <- memdb_frame(x = 1:3, y = c(11, 12, NA))
  expect_equal(
    rows_delete(
      df, memdb_frame(x = 2:3),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE
    ) %>%
      collect(),
    tibble(x = 1, y = 11)
  )

  expect_equal(df %>% collect(), tibble(x = 1:3, y = c(11, 12, NA)))
})

test_that("`rows_delete()` works with `in_place = FALSE` with `returning`", {
  expect_equal(
    rows_delete(
      memdb_frame(x = 1:3, y = c(11, 12, 13)),
      memdb_frame(x = 2:3),
      by = "x",
      unmatched = "ignore",
      in_place = FALSE,
      returning = everything()
    ) %>%
      get_returned_rows(),
    tibble(x = 2:3, y = 12:13)
  )
})

test_that("`rows_delete()` works with `in_place = TRUE`", {
  expect_snapshot(error = TRUE,
    (rows_delete(
      lazy_frame(x = 1:3, y = 11:13, .name = "df_x"),
      lazy_frame(x = 2:3, .name = "df_y"),
      by = "x",
      unmatched = "ignore",
      in_place = TRUE
    ))
  )

  df <- memdb_frame(x = 1:3, y = 11:13)
  rows_delete(
    df, memdb_frame(x = 2:3),
    by = "x",
    unmatched = "ignore",
    in_place = TRUE
  )

  expect_equal(df %>% collect(), tibble(x = 1L, y = 11L))
})

test_that("`rows_delete()` works with `in_place = TRUE` and `returning`", {
  skip_if_not_installed("RSQLite", "2.2.8")

  df <- memdb_frame(x = 1:3, y = 11:13)
  df_deleted <- rows_delete(
    df, memdb_frame(x = 2:4),
    by = "x",
    unmatched = "ignore",
    in_place = TRUE,
    returning = everything()
  )

  expect_equal(get_returned_rows(df_deleted), tibble(x = 2:3, y = 12:13))

  expect_equal(df_deleted %>% collect(), tibble(x = 1L, y = 11L))
})

test_that("`sql_query_delete()` is correct", {
  df_y <- lazy_frame(
    a = 2:3, b = c(12L, 13L), c = -(2:3), d = c("y", "z"),
    con = simulate_dbi(),
    .name = "df_y"
  ) %>%
    mutate(c = c + 1)

  expect_snapshot(
    sql_query_delete(
      con = simulate_dbi(),
      x_name = ident("df_x"),
      y = df_y,
      by = c("a", "b"),
      returning_cols = c("a", b2 = "b")
    )
  )
})

# returned_rows -----------------------------------------------------------

test_that("`get_returned_rows()` works", {
  df <- tibble(x = 1)
  expect_false(has_returned_rows(df))
  expect_snapshot(error = TRUE, (get_returned_rows(df)))

  df2 <- set_returned_rows(df, mtcars)
  expect_true(has_returned_rows(df2))
  expect_equal(get_returned_rows(df2), as_tibble(mtcars))
})

test_that("2nd edition uses sql methods", {
  con <- structure(list(), class = c("Test", "DBIConnection"))

  local_methods(
    dbplyr_edition.Test = \(con) 2,
    sql_table_analyze.Test = \(con, ...) abort("sql_method")
  )
  expect_error(dbplyr_analyze(con), "sql_method")
})

test_that("sql_query_rows() works", {
  expect_equal(
    sql_query_rows(dialect_ansi(), ident("abc")),
    sql('SELECT COUNT(*) FROM "abc" AS "master"')
  )
})

test_that("analyse/explain sql generates expected SQL", {
  con <- dialect_ansi()
  expect_snapshot(sql_table_analyze(con, "tbl"))
  expect_snapshot(sql_query_explain(con, sql("SELECT * FROM foo")))
})

test_that("sql_query_wrap generates expected SQL", {
  unique_subquery_name_reset()

  con <- dialect_ansi()
  expect_snapshot(sql_query_wrap(con, ident("table")))
  expect_snapshot(sql_query_wrap(con, in_schema("schema", "tbl")))
  expect_snapshot(sql_query_wrap(con, sql("SELECT * FROM foo")))
})

test_that("sql_table_index generates expected SQL", {
  con <- dialect_ansi()
  expect_snapshot(sql_table_index(con, "tbl", c("a", "b")))
  expect_snapshot(sql_table_index(con, "tbl", "c", unique = TRUE))
})

test_that("sql_query_save generates expected SQL", {
  con <- dialect_ansi()

  sql <- sql("SELECT * FROM foo")
  expect_snapshot(sql_query_save(con, sql, "tbl"))
  expect_snapshot(sql_query_save(con, sql, "tbl", temporary = FALSE))
})

test_that("RPostgreSQL backend", {
  skip_if_not(identical(Sys.getenv("GITHUB_POSTGRES"), "true"))

  src <- withr::local_db_connection(
    DBI::dbConnect(
      RPostgreSQL::PostgreSQL(),
      dbname = "test",
      user = "postgres",
      password = "password",
      host = "127.0.0.1"
    )
  )

  suppressWarnings(
    copy_to(src, mtcars, "mtcars", overwrite = TRUE, temporary = FALSE)
  )
  withr::defer(DBI::dbRemoveTable(src, "mtcars"))

  expect_identical(colnames(tbl(src, "mtcars")), colnames(mtcars))

  src_cyl <- tbl(src, "mtcars") %>% select(cyl) %>% collect()
  expect_identical(src_cyl$cyl, mtcars$cyl)
})

---
title: "Reprexes for dbplyr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reprexes for dbplyr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
options(tibble.print_min = 6L, tibble.print_max = 6L, digits = 3)
```

If you're reporting a bug in dbplyr, it is much easier for me to help you if you can supply a [reprex](https://reprex.tidyverse.org) that I can run on my computer. Creating reprexes for dbplyr is particularly challenging because you are probably using a database that you can't share with me. Fortunately, in many cases you can still demonstrate the problem even if I don't have the complete dataset, or even access to the database system that you're using.

This vignette outlines three approaches for creating reprexes that will work anywhere:

* Use `memdb_frame()`/`tbl_memdb()` to easily create datasets that live in an 
  in-memory SQLite database.
  
* Use `lazy_frame()`/`tbl_lazy()` to simulate SQL generation of dplyr pipelines.

* Use `translate_sql()` to simulate SQL generation of columnar expression.

```{r setup, message = FALSE}
library(dplyr)
library(dbplyr)
```

## Using `memdb_frame()`

The first place to start is with SQLite. SQLite is particularly appealing because it's completely embedded instead an R package so doesn't have any external dependencies. SQLite is designed to be small and simple, so it can't demonstrate all problems, but it's easy to try out and a great place to start.

You can easily create a SQLite in-memory database table using `memdb_frame()`:

```{r}
mf <- memdb_frame(g = c(1, 1, 2, 2, 2), x = 1:5, y = 5:1)
mf

mf |> 
  group_by(g) |> 
  summarise_all(mean, na.rm = TRUE)
```

Reprexes are easiest to understand if you create very small custom data, but if you do want to use an existing data frame you can use `copy_to()` + `memdb()`:

```{r}
mtcars_db <- copy_to(memdb(), mtcars)
mtcars_db |> 
  group_by(cyl) |> 
  summarise(n = n()) |> 
  show_query()
```

## Translating verbs

Many problems with dbplyr come down to incorrect SQL generation. Fortunately, it's possible to generate SQL without a database using `lazy_frame()` and `tbl_lazy()`. Both take an `con` argument which takes a database "simulator" like `simulate_postgres()`, `simulate_sqlite()`, etc.

```{r}
x <- c("abc", "def", "ghif")

lazy_frame(x = x, con = simulate_postgres()) |> 
  head(5) |> 
  show_query()

lazy_frame(x = x, con = simulate_mssql()) |> 
  head(5) |> 
  show_query()
```

If you isolate the problem to incorrect SQL generation, it would be very helpful if you could also suggest more appropriate SQL.

## Translating individual expressions

In some cases, you might be able to track the problem down to incorrect translation for a single column expression. In that case, you can make your reprex even simpler with `translate_sql()`:

```{r}
translate_sql(substr(x, 1, 2), con = simulate_postgres())
translate_sql(substr(x, 1, 2), con = simulate_sqlite())
```

---
title: "Introduction to dbplyr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to dbplyr}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
options(tibble.print_min = 6L, tibble.print_max = 6L, digits = 3)
```

As well as working with local in-memory data stored in data frames, dplyr also works with remote on-disk data stored in databases. This is particularly useful in two scenarios:

* Your data is already in a database.
  
* You have so much data that it does not all fit into memory simultaneously
  and you need to use some external storage engine.
  
(If your data fits in memory there is no advantage to putting it in a database: it will only be slower and more frustrating.)

This vignette focuses on the first scenario because it's the most common. If you're using R to do data analysis inside a company, most of the data you need probably already lives in a database (it's just a matter of figuring out which one!). However, you will learn how to load data in to a local database in order to demonstrate dplyr's database tools. At  the end, I'll also give you a few pointers if you do need to set up your own database.

## Getting started

To use databases with dplyr you need to first install dbplyr:

```{r, eval = FALSE}
install.packages("dbplyr")
```

You'll also need to install a DBI backend package. The DBI package provides a common interface that allows dplyr to work with many different databases using the same code. DBI is automatically installed with dbplyr, but you need to install a specific backend for the database that you want to connect to.

Five commonly used backends are:

*   [RMariaDB](https://CRAN.R-project.org/package=RMariaDB) 
    connects to MySQL and MariaDB

*   [RPostgres](https://CRAN.R-project.org/package=RPostgres) 
    connects to Postgres and Redshift.

*   [RSQLite](https://github.com/r-dbi/RSQLite) embeds a SQLite database.

*   [odbc](https://github.com/r-dbi/odbc#odbc) connects to many commercial
    databases via the open database connectivity protocol.

*   [bigrquery](https://github.com/r-dbi/bigrquery) connects to Google's
    BigQuery.

If the database you need to connect to is not listed here, you'll need to do some investigation (i.e. googling) yourself.

In this vignette, we're going to use the RSQLite backend which is automatically installed when you install dbplyr. SQLite is a great way to get started with databases because it's completely embedded inside an R package. Unlike most other systems, you don't need to setup a separate database server. SQLite is great for demos, but is surprisingly powerful, and with a little practice you can use it to easily work with many gigabytes of data.

## Connecting to the database

To work with a database in dplyr, you must first connect to it, using `DBI::dbConnect()`. We're not going to go into the details of the DBI package here, but it's the foundation upon which dbplyr is built. You'll need to learn more about if you need to do things to the database that are beyond the scope of dplyr.

```{r setup, message = FALSE}
library(dplyr)
con <- DBI::dbConnect(RSQLite::SQLite(), dbname = ":memory:")
```

The arguments to `DBI::dbConnect()` vary from database to database, but the first argument is always the database backend. It's `RSQLite::SQLite()` for RSQLite, `RMariaDB::MariaDB()` for RMariaDB, `RPostgres::Postgres()` for RPostgres, `odbc::odbc()` for odbc, and `bigrquery::bigquery()` for BigQuery. SQLite only needs one other argument: the path to the database. Here we use the special string `":memory:"` which causes SQLite to make a temporary in-memory database.

Most existing databases don't live in a file, but instead live on another server. That means in real-life that your code will look more like this:

```{r, eval = FALSE}
con <- DBI::dbConnect(RMariaDB::MariaDB(), 
  host = "database.rstudio.com",
  user = "hadley",
  password = rstudioapi::askForPassword("Database password")
)
```

(If you're not using RStudio, you'll need some other way to securely retrieve your password. You should never record it in your analysis scripts or type it into the console. [Securing Credentials](https://db.rstudio.com/best-practices/managing-credentials) provides some best practices.)

Our temporary database has no data in it, so we'll start by copying over `nycflights13::flights` using the convenient `copy_to()` function. This is a quick and dirty way of getting data into a database and is useful primarily for demos and other small jobs.

```{r}
copy_to(con, nycflights13::flights, "flights",
  temporary = FALSE, 
  indexes = list(
    c("year", "month", "day"), 
    "carrier", 
    "tailnum",
    "dest"
  )
)
```

As you can see, the `copy_to()` operation has an additional argument that allows you to supply indexes for the table. Here we set up indexes that will allow us to quickly process the data by day, carrier, plane, and destination. Creating the right indices is key to good database performance, but is unfortunately beyond the scope of this article.

Now that we've copied the data, we can use `tbl()` to take a reference to it:

```{r}
flights_db <- tbl(con, "flights")
```

When you print it out, you'll notice that it mostly looks like a regular tibble:

```{r}
flights_db 
```

The main difference is that you can see that it's a remote source in a SQLite database.

## Generating queries

To interact with a database you usually use SQL, the Structured Query Language. SQL is over 40 years old, and is used by pretty much every database in existence. The goal of dbplyr is to automatically generate SQL for you so that you're not forced to use it. However, SQL is a very large language and dbplyr doesn't do everything. It focusses on `SELECT` statements, the SQL you write most often as an analyst.

Most of the time you don't need to know anything about SQL, and you can continue to use the dplyr verbs that you're already familiar with:

```{r}
flights_db %>% select(year:day, dep_delay, arr_delay)

flights_db %>% filter(dep_delay > 240)

flights_db %>% 
  group_by(dest) %>%
  summarise(delay = mean(dep_delay))
```

However, in the long-run, I highly recommend you at least learn the basics of SQL. It's a valuable skill for any data scientist, and it will help you debug problems if you run into problems with dplyr's automatic translation. If you're completely new to SQL you might start with this [codeacademy tutorial](https://www.codecademy.com/learn/learn-sql). If you have some familiarity with SQL and you'd like to learn more, I found [how indexes work in SQLite](https://www.sqlite.org/queryplanner.html) and [10 easy steps to a complete understanding of SQL](https://blog.jooq.org/2016/03/17/10-easy-steps-to-a-complete-understanding-of-sql/) to be particularly helpful.

The most important difference between ordinary data frames and remote database queries is that your R code is translated into SQL and executed in the database on the remote server, not in R on your local machine. When working with databases, dplyr tries to be as lazy as possible:

* It never pulls data into R unless you explicitly ask for it.

* It delays doing any work until the last possible moment: it collects together
  everything you want to do and then sends it to the database in one step.

For example, take the following code:

```{r}
tailnum_delay_db <- flights_db %>% 
  group_by(tailnum) %>%
  summarise(
    delay = mean(arr_delay),
    n = n()
  ) %>% 
  arrange(desc(delay)) %>%
  filter(n > 100)
```

Surprisingly, this sequence of operations never touches the database. It's not until you ask for the data (e.g. by printing `tailnum_delay`) that dplyr generates the SQL and requests the results from the database. Even then it tries to do as little work as possible and only pulls down a few rows.

```{r}
tailnum_delay_db
```

Behind the scenes, dplyr is translating your R code into SQL. You can see the SQL it's generating with `show_query()`:

```{r}
tailnum_delay_db %>% show_query()
```

If you're familiar with SQL, this probably isn't exactly what you'd write by hand, but it does the job. You can learn more about the SQL translation in `vignette("translation-verb")` and `vignette("translation-function")`.

Typically, you'll iterate a few times before you figure out what data you need from the database. Once you've figured it out, use `collect()` to pull all the data down into a local tibble:

```{r}
tailnum_delay <- tailnum_delay_db %>% collect()
tailnum_delay
```

`collect()` requires that database does some work, so it may take a long time to complete. Otherwise, dplyr tries to prevent you from accidentally performing expensive query operations:

* Because there's generally no way to determine how many rows a query will 
  return unless you actually run it, `nrow()` is always `NA`.

* Because you can't find the last few rows without executing the whole 
  query, you can't use `tail()`.

```{r, error = TRUE}
nrow(tailnum_delay_db)

tail(tailnum_delay_db)
```

You can also ask the database how it plans to execute the query with `explain()`. The output is database dependent, and can be esoteric, but learning a bit about it can be very useful because it helps you understand if the database can execute the query efficiently, or if you need to create new indices.

## Creating your own database

If you don't already have a database, here's some advice from my experiences setting up and running all of them. SQLite is by far the easiest to get started with. PostgreSQL is not too much harder to use and has a wide range of built-in functions. In my opinion, you shouldn't bother with MySQL/MariaDB: it's a pain to set up, the documentation is subpar, and it's less featureful than Postgres. Google BigQuery might be a good fit if you have very large data, or if you're willing to pay (a small amount of) money to someone who'll look after your database.

All of these databases follow a client-server model - a computer that connects to the database and the computer that is running the database (the two may be one and the same but usually isn't). Getting one of these databases up and running is beyond the scope of this article, but there are plenty of tutorials available on the web.

### MySQL/MariaDB

In terms of functionality, MySQL lies somewhere between SQLite and PostgreSQL. It provides a wider range of [built-in functions](http://dev.mysql.com/doc/refman/5.0/en/functions.html). It gained support for window functions in 2018.

### PostgreSQL

PostgreSQL is a considerably more powerful database than SQLite. It has a much wider range of [built-in functions](http://www.postgresql.org/docs/current/static/functions.html), and is generally a more featureful database.

### BigQuery

BigQuery is a hosted database server provided by Google. To connect, you need to provide your `project`, `dataset` and optionally a project for `billing` (if billing for `project` isn't enabled). 

It provides a similar set of functions to Postgres and is designed specifically for analytic workflows. Because it's a hosted solution, there's no setup involved, but if you have a lot of data, getting it to Google can be an ordeal (especially because upload support from R is not great currently). (If you have lots of data, you can [ship hard drives](https://cloud.google.com/storage-transfer-service)!)

---
title: "Developing a dbplyr backend"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Adding a new DBI backend}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r}
#| include: false
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
options(tibble.print_min = 6L, tibble.print_max = 6L, digits = 3)
```

This document describes how to add a new SQL backend to dbplyr. To begin:

* Ensure that you have a DBI compliant database backend. If not, you'll need
  to first create it by following the instructions in 
  `vignette("backend", package = "DBI")`.
  
* You'll need a working knowledge of S3. Make sure that you're 
  [familiar with the basics](https://adv-r.hadley.nz/s3.html) 
  before you start.
 
This document is still a work in progress, but it will hopefully get you started. I'd also strongly recommend reading the bundled source code for [SQLite](https://github.com/tidyverse/dbplyr/blob/master/R/backend-sqlite.R), [MySQL](https://github.com/tidyverse/dbplyr/blob/master/R/backend-mysql.R), and [PostgreSQL](https://github.com/tidyverse/dbplyr/blob/master/R/backend-postgres.R).

## First steps

For interactive exploitation, attach dplyr and DBI. If you're creating a package, you'll need to import dplyr and DBI.

```{r setup, message = FALSE}
library(dplyr)
library(dbplyr)
library(DBI)
```

Check that you can create a tbl from a connection, like:

```{r}
con <- DBI::dbConnect(RSQLite::SQLite(), path = ":memory:")
DBI::dbWriteTable(con, "mtcars", mtcars)

tbl(con, "mtcars")
```

If you can't, this likely indicates some problem with the DBI methods. Use [DBItest](https://github.com/r-dbi/DBItest) to narrow down the problem.

## Write your first method

The first method of your dbplyr backend should always be for the `dbplyr_edition()` generic:

```{r}
#' @importFrom dbplyr dbplyr_edition
#' @export
dbplyr_edition.myConnectionClass <- function(con) 2L
```

This declares that your package uses version 2 of the API, which is the version that this vignette documents.

## Define a dialect

Next, define a dialect for your backend using `sql_dialect()` and `new_sql_dialect()`. A dialect encapsulates the SQL syntax rules for your database:

```{r, eval = FALSE}
#' @export
sql_dialect.myConnectionClass <- function(con) {
  new_sql_dialect(
    "mybackend",
    quote_identifier = function(x) sql_quote(x, '"')
  )
}
```

The `new_sql_dialect()` function requires two arguments:

* `dialect`: A unique name for your backend (used to create the class name).
* `quote_identifier`: A function that quotes identifiers. This will typically be either a call to `sql_quote()` or to `DBI::dbQuoteIdentifier()`

`new_sql_dialect()` has a number of additional arguments that describe the capabilities of the SQL dialect; see `?new_sql_dialect` for details.

The dialect system allows you to write methods that dispatch on the dialect class (e.g., `sql_translation.sql_dialect_mybackend`) rather than the connection class. This makes it easier to share SQL generation code between different connection types that use the same database (e.g. direct, ODBC, JDBC, ADBC). 

In general, generics that start with `sql_` should have a sql_dialect method, and generics that start with `db_` should use a DBI connection method. It should be relatively rare to need a method for `db_` generic, but occasionally dbplyr's SQL generation system is not quite flexible enough. If you find this, it's worth filing an issue so I can look into it.

## Copying, computing, collecting and collapsing

Next, check that `copy_to()`, `collapse()`, `compute()`, and `collect()` work:

*   If `copy_to()` fails, you probably need a method for `sql_table_analyze()` 
    or `sql_table_index()`. If `copy_to()` fails during creation of the tbl, 
    you may need a method for `sql_query_fields()`.

*   If `collapse()` fails, your database has a non-standard way of constructing 
    subqueries. Add a method for `sql_subquery()`.
  
*   If `compute()` fails, your database has a non-standard way of saving queries
    in temporary tables. Add a method for `db_save_query()`.

## SQL translation: verbs

Make sure you've read `vignette("translation-verb")` so you have the lay of the land. First check that SQL translation for the key verbs work:

* `summarise()`, `mutate()`, `filter()` etc: powered by `sql_query_select()`
* `left_join()`, `inner_join()`: powered by `sql_query_join()`
* `semi_join()`, `anti_join()`: powered by `sql_query_semi_join()`
* `union()`, `intersect()`, `setdiff()`: powered by `sql_query_set_op()`

### Building SQL strings

If you need to generate your own SQL, we recommend using `sql_glue2()`. It uses glue syntax with type markers for safe SQL generation:

```{r}
con <- simulate_dbi()

# Create an index
index_name <- "index"
table <- I("schema.table")
columns <- c("column1", "column2")
sql_glue2(con, "CREATE INDEX {.id index_name} ON {.tbl table} {.id columns*}")

# Insert values safely
name <- "O'Brien"
sql_glue2(con, "INSERT INTO students (name) VALUES {name*}")

# Build a query
table <- "my_table"
cols <- c("id", "name", "value")
sql_glue2(con, "SELECT {.id cols} FROM {.tbl table}")
```

### Behind the scenes

The high-level translation from R code to SQL takes place in a few steps, using two intermediate representations: a "lazy query" and a "query". A lazy query is closer to R and a query is closer to SQL. This continuum of representation allows us to work with expressions when trying to reason about the transformation pipeline, then later switch to strings of SQL when we get closer to generating a query.

All dbplyr pipelines are built on top of the `tbl_lazy` data structure, which is the only data structure that the user will ever see. It has two fields:

* A database `con`nection which is used to specialise SQL translation.
* A `lazy_query` which represents the lazy computation described by the pipeline. It always starts off as a `lazy_base_query` which represents a remote table or SQL query.

Each verb takes a `tbl_lazy` as input and returns a modified `tbl_lazy`, modifying the `lazy_query` with the new requirements. At the simplest level, this involves wrapping the existing lazy query within a new lazy query, such as `lazy_query_select()` or `lazy_query_multi_join()`. These represent what later will become a subquery so that (e.g.) `filter()`, `mutate()`, `select()`, `arrange()`, and `summarise()` all use `lazy_select_query()`, whereas `left_join()` and `inner_join()` use `lazy_query_multi_join()`. 

These lazy queries are typically created by a helper function like `add_select()`, `add_mutate()`, and `add_filter()`. These helpers have two roles:

1. They abstract away repeated code for substantially similar dplyr verbs. For example, `select()`, `rename()`, `rename_with()` and `relocate()` all use `add_select()`. 

2. They are responsible for figuring out if we can re-use the existing query or we need to wrap in a subquery. Typically this work is done by a function called `can_inline_select()` which will use what we know about SQL and the verb to determine whether or not we have to introduce a subquery.

The lazy query gives us a representation of the data manipulation pipeline with enough detail that we can reason about when we need a subquery. Then when the user calls `collect()`, we need to generate an SQL query. This happens in two steps: building and rendering.

First, `sql_build()` recurses over the lazy query, building up query objects like `select_query()` and `multi_join_query()` that represent the different subtypes of `SELECT` queries. Compared to the previous "lazy" query representation, these are much closer to SQL than R. They can be much simpler because we no longer need to reason about subqueries, but we still need some structure to facilitate backend-specific translations.

Next, `sql_render()` is called on the result of `sql_build()` and dispatches to generics like `sql_query_select()`, `sql_query_join()`, `sql_query_semi_join()`, and `sql_query_set_op()`. These generics have methods for each backend that are responsible for creating the needed SQL.

## SQL translation: vectors

Finally, you may have to provide custom R -> SQL translation for functions that work with vectors within verbs. You can do this by providing a method for `sql_translation()`, which returns an object created by `sql_variant()`.

You can define translation methods either on your connection class or on your dialect class. Using the dialect class is recommended as it allows sharing translations between different connection types:

```{r, eval = FALSE}
# Method on dialect class (recommended)
sql_translation.sql_dialect_mybackend <- function(con) {
  sql_variant(
    scalar = sql_translator(base_scalar, ...), # Functions in SELECT (non-aggregated)
    aggregate = sql_translator(base_aggregate, ...), # Aggregation functions (mean, sum, etc.)
    window = sql_translator(base_win, ...) # Window functions (lead, lag, rank, etc.)
  )
}
```

Each translator will inherit from the base (ANSI SQL) translator and override only what's different for your backend:

```{r, eval = FALSE}
sql_translator(
  base_scalar, # Inherit most translations
  # Override specific functions for your backend
  `+` = sql_infix("+"),
  mean = sql_aggregate("AVG", "mean")
)
```

### Scalar function helpers

dbplyr provides several helper functions to make it easier to translate R functions to SQL:

*   **`sql_prefix(f, n = NULL)`**: For standard SQL functions. The `n` argument
    optionally specifies the number of arguments.
*   **`sql_infix(f)`**: For infix operators like `+`, `*`, or `==`.
*   **`sql_cast(type)`**: For type casting functions.
*   **`sql_not_supported(f)`**: For functions with no SQL translation.

Here's an example showing all of these helpers in use:

```{r}
#| eval: false
sql_translation.sql_dialect_mybackend <- function(con) {
  sql_variant(
    scalar = sql_translator(
      base_scalar,
      # Standard SQL functions
      cos = sql_prefix("COS", 1),
      round = sql_prefix("ROUND", 2),
      # Infix operators
      `+` = sql_infix("+"),
      `*` = sql_infix("*"),
      `==` = sql_infix("="),
      # Type casting
      as.numeric = sql_cast("NUMERIC"),
      as.character = sql_cast("VARCHAR")
    ),
    aggregate = sql_translator(base_agg),
    window = sql_translator(base_win)
  )
}
```

### Aggregation function helpers

*   **`sql_aggregate(f, f_r = f)`**: For single-argument SQL aggregate
    functions. The `f_r` argument gives the name of the R function.
*   **`sql_aggregate_2(f)`**: For two-argument SQL aggregate functions.
*   **`sql_aggregate_n(f, f_r = f)`**: For variadic SQL aggregate functions.

```{r}
#| eval: false
sql_translation.sql_dialect_mybackend <- function(con) {
  sql_variant(
    scalar = sql_translator(base_scalar),
    aggregate = sql_translator(
      .parent = base_agg,
      # Single-argument aggregates
      mean = sql_aggregate("AVG", "mean"),
      var = sql_aggregate("VAR_SAMP", "var"),
      # Two-argument aggregates
      cov = sql_aggregate_2("COVAR_SAMP"),
      # Variadic aggregates
      pmin = sql_aggregate_n("LEAST", "pmin"),
      pmax = sql_aggregate_n("GREATEST", "pmax"),
      # Unsupported functions
      median = sql_not_supported("median")
    ),
    window = sql_translator(base_win)
  )
}
```

### Window function helpers

Window functions have their own set of helpers:

*   **`win_rank(f)`**: For ranking functions.
*   **`win_aggregate(f)`**: For aggregate functions used as window functions.
*   **`win_cumulative(f)`**: For cumulative functions.
*   **`win_absent(f)`**: For backends that don't support certain window
    functions.

Here's an example showing all of these helpers in use:

```{r, eval = FALSE}
window = sql_translator(
  base_win,
  # Ranking functions
  row_number = win_rank("ROW_NUMBER"),
  rank = win_rank("RANK"),
  dense_rank = win_rank("DENSE_RANK"),
  # Aggregate functions as window functions
  mean = win_aggregate("AVG"),
  sum = win_aggregate("SUM"),
  # Cumulative functions
  cumsum = win_cumulative("SUM"),
  # Absent functions
  cume_dist = win_absent("cume_dist")
)
```

### Custom translation functions

For more complex translations, you can write custom functions that return SQL expressions using `sql_glue()`. This uses glue syntax for string interpolation with automatic escaping.

```{r, eval = FALSE}
scalar = sql_translator(
  base_scalar,

  # Custom log function with change of base
  log = function(x, base = exp(1)) {
    if (isTRUE(all.equal(base, exp(1)))) {
      sql_glue("LN({x})")
    } else {
      sql_glue("LOG({x}) / LOG({base})")
    }
  },

  # Custom paste function using CONCAT
  paste = function(..., sep = " ") {
    sql_glue("CONCAT_WS({sep}, {...})")
  }
)
```

---
title: "Function translation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Function translation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
options(tibble.print_min = 4L, tibble.print_max = 4L)
```

There are two parts to dbplyr SQL translation: translating dplyr verbs, and translating expressions within those verbs. This vignette describes how individual expressions (function calls) are translated; `vignette("translation-verb")` describes how entire verbs are translated.

```{r, message = FALSE}
library(dbplyr)
library(dplyr)
```

`dbplyr::translate_sql()` powers translation of individual function calls, and I'll use it extensively in this vignette to show what's happening. You shouldn't need to use it ordinary code as dbplyr takes care of the translation automatically.

```{r}
translate_sql((x + y) / 2)
```

`translate_sql()` takes an optional `con` parameter. If not supplied, this causes dplyr to generate (approximately) SQL-92 compliant SQL. If supplied, dplyr uses `sql_translation()` to look up a custom environment which makes it possible for different databases to generate slightly different SQL: see `vignette("new-backend")` for more details.  You can use the various simulate helpers to see the translations used by different backends:

```{r}
translate_sql(x ^ 2L)
translate_sql(x ^ 2L, con = simulate_sqlite())
translate_sql(x ^ 2L, con = simulate_access())
```

Perfect translation is not possible because databases don't have all the functions that R does. The goal of dplyr is to provide a semantic rather than a literal translation: what you mean, rather than precisely what is done. In fact, even for functions that exist both in databases and R, you shouldn't expect results to be identical; database programmers have different priorities than R core programmers. For example, in R in order to get a higher level of numerical accuracy, `mean()` loops through the data twice. R's `mean()` also provides a `trim` option for computing trimmed means; this is something that databases do not provide. 

If you're interested in how `translate_sql()` is implemented, the basic techniques that underlie the implementation of `translate_sql()` are described in ["Advanced R"](https://adv-r.hadley.nz/translation.html). 

## Basic differences

The following examples work through some of the basic differences between R and SQL.

*   `"` and `'` mean different things

    ```{r}
    # In SQLite variable names are escaped by double quotes:
    translate_sql(x)
    # And strings are escaped by single quotes
    translate_sql("x")
    ```
    
*   And some functions have different argument orders:
    
    ```{r}
    translate_sql(substr(x, 5, 10))
    translate_sql(log(x, 10))
    ```

*   R and SQL have different defaults for integers and reals. 
    In R, 1 is a real, and 1L is an integer. In SQL, 1 is an integer, and 1.0 is a real
    
    ```{r}
    translate_sql(1)
    translate_sql(1L)
    ```

## Known functions

### Mathematics

* basic math operators: `+`, `-`, `*`, `/`, `^`
* trigonometry: `acos()`, `asin()`, `atan()`, `atan2()`, `cos()`, `cot()`, `tan()`, `sin()`
* hypergeometric: `cosh()`, `coth()`, `sinh()`, `tanh()`
* logarithmic: `log()`, `log10()`, `exp()`
* misc: `abs()`, `ceiling()`, `sqrt()`, `sign()`, `round()`

## Modulo arithmetic

dbplyr translates `%%` to the SQL equivalents but note that it's not precisely the same: most databases use truncated division where the modulo operator takes the sign of the dividend, where R using the mathematically preferred floored division with the modulo sign taking the sign of the divisor.

```{r}
df <- tibble(
  x = c(10L, 10L, -10L, -10L), 
  y = c(3L, -3L, 3L, -3L)
)
mf <- tbl_memdb(df)

df %>% mutate(x %% y)
mf %>% mutate(x %% y)
```

dbplyr no longer translates `%/%` because there's no robust cross-database translation available.

### Logical comparisons

* logical comparisons: `<`, `<=`, `!=`, `>=`, `>`, `==`, `%in%`
* boolean operations: `&`, `&&`, `|`, `||`, `!`, `xor()`

### Aggregation

All database provide translation for the basic aggregations: `mean()`, `sum()`, `min()`, `max()`, `sd()`, `var()`. Databases automatically drop NULLs (their equivalent of missing values) whereas in R you have to ask nicely. The aggregation functions warn you about this important difference:

```{r}
translate_sql(mean(x))
translate_sql(mean(x, na.rm = TRUE))
```

Note that, by default, `translate()` assumes that the call is inside a `mutate()` or `filter()` and generates a window translation. If you want to see the equivalent `summarise()`/aggregation translation, use `window = FALSE`:

```{r}
translate_sql(mean(x, na.rm = TRUE), window = FALSE)
```

### Conditional evaluation

`if` and `switch()` are translate to `CASE WHEN`:

```{r}
translate_sql(if (x > 5) "big" else "small")
translate_sql(switch(x, a = 1L, b = 2L, 3L))
```
  
### String manipulation

### Date/time

* string functions: `tolower`, `toupper`, `trimws`, `nchar`, `substr`
* coerce types: `as.numeric`, `as.integer`, `as.character`

## Unknown functions

Any function that dplyr doesn't know how to convert is left as is. This means that database functions that are not covered by dplyr can often be used directly via `translate_sql()`. 

### Prefix functions

Any function that dbplyr doesn't know about will be left as is:

```{r}
translate_sql(foofify(x, y))
```

Because SQL functions are general case insensitive, I recommend using upper case when you're using SQL functions in R code. That makes it easier to spot that you're doing something unusual:

```{r}
translate_sql(FOOFIFY(x, y))
```

### Infix functions

As well as prefix functions (where the name of the function comes before the arguments), dbplyr also translates infix functions. That allows you to use expressions like `LIKE` which does a limited form of pattern matching:

```{r}
translate_sql(x %LIKE% "%foo%")
```

Or use `||` for string concatenation (although most backends will translate `paste()` and `paste0()` for you):

```{r}
translate_sql(x %||% y)
```

### Special forms

SQL functions tend to have a greater variety of syntax than R. That means there are a number of expressions that can't be translated directly from R code. To insert these in your own queries, you can use literal SQL inside `sql()`:

```{r}
translate_sql(sql("x!"))
translate_sql(x == sql("ANY VALUES(1, 2, 3)"))
```

This gives you a lot of freedom to generate the SQL you need: 

```{r}
mf <- memdb_frame(x = 1, y = 2)

mf %>% 
  transmute(factorial = sql("x!")) %>% 
  show_query()

mf %>% 
  transmute(factorial = sql("CAST(x AS FLOAT)")) %>% 
  show_query()
```

### Error for unknown translations

If needed, you can also force dbplyr to error if it doesn't know how to translate a function with the `dplyr.strict_sql` option:

```{r, error = TRUE}
options(dplyr.strict_sql = TRUE)
translate_sql(glob(x, y))
```

## Window functions

Things get a little trickier with window functions, because SQL's window functions are considerably more expressive than the specific variants provided by base R or dplyr. They have the form `[expression] OVER ([partition clause] [order clause] [frame_clause])`:

* The __expression__ is a combination of variable names and window functions.
  Support for window functions varies from database to database, but most
  support the ranking functions, `lead`, `lag`, `nth`, `first`,
  `last`, `count`, `min`, `max`, `sum`, `avg` and `stddev`.

* The __partition clause__ specifies how the window function is broken down
  over groups. It plays an analogous role to `GROUP BY` for aggregate functions,
  and `group_by()` in dplyr. It is possible for different window functions to 
  be partitioned into different groups, but not all databases support it, and
  neither does dplyr.
  
* The __order clause__ controls the ordering (when it makes a difference).
  This is important for the ranking functions since it specifies which 
  variables to rank by, but it's also needed for cumulative functions and lead.
  Whenever you're thinking about before and after in SQL, you must always tell 
  it which variable defines the order. If the order clause is missing when 
  needed, some databases fail with an error message while others return 
  non-deterministic results.
  
* The __frame clause__ defines which rows, or __frame__, that are passed 
  to the window function, describing which rows (relative to the current row)
  should be included. The frame clause provides two offsets which determine
  the start and end of frame. There are three special values: -Inf means
  to include all preceding rows (in SQL, "unbounded preceding"), 0 means the
  current row ("current row"), and Inf means all following rows ("unbounded
  following"). The complete set of options is comprehensive, but fairly 
  confusing, and is summarised visually below.

    ```{r echo = FALSE, out.width = "100%"}
    knitr::include_graphics("windows.png", dpi = 300)
    ```

    Of the many possible specifications, there are only three that commonly
    used. They select between aggregation variants:

    * Recycled: `BETWEEN UNBOUND PRECEEDING AND UNBOUND FOLLOWING`
    
    * Cumulative: `BETWEEN UNBOUND PRECEEDING AND CURRENT ROW`
    
    * Rolling: `BETWEEN 2 PRECEEDING AND 2 FOLLOWING`
    
    dplyr generates the frame clause based on whether your using a recycled
    aggregate or a cumulative aggregate.
    
To see how individual window functions are translated to SQL, we can again use `translate_sql()`:

```{r}
translate_sql(mean(G))
translate_sql(rank(G))
translate_sql(ntile(G, 2))
translate_sql(lag(G))
```

If the tbl has been grouped or arranged previously in the pipeline, then dplyr will use that information to set the "partition by" and "order by" clauses. For interactive exploration, you can achieve the same effect by setting the `vars_group` and `vars_order` arguments to `translate_sql()`

```{r}
translate_sql(cummean(G), vars_order = "year")
translate_sql(rank(), vars_group = "ID")
```

There are some challenges when translating window functions between R and SQL, because dplyr tries to keep the window functions as similar as possible to both the existing R analogues and to the SQL functions. This means that there are three ways to control the order clause depending on which window function you're using:

* For ranking functions, the ordering variable is the first argument: `rank(x)`,
  `ntile(y, 2)`. If omitted or `NULL`, will use the default ordering associated 
  with the tbl (as set by `arrange()`).

* Accumulating aggregates only take a single argument (the vector to aggregate).
  To control ordering, use `order_by()`.
  
* Aggregates implemented in dplyr (`lead`, `lag`, `nth_value`, `first_value`, 
  `last_value`) have an `order_by` argument. Supply it to override the
  default ordering.

The three options are illustrated in the snippet below:

```{r, eval = FALSE}
mutate(players,
  min_rank(yearID),
  order_by(yearID, cumsum(G)),
  lead(G, order_by = yearID)
)
```

Currently there is no way to order by multiple variables, except by setting the default ordering with `arrange()`. This will be added in a future release. 

---
title: "Function translation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Function translation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
options(tibble.print_min = 6L, tibble.print_max = 6L, digits = 3)
```

There are two parts to dbplyr SQL translation: translating dplyr verbs, and translating expressions within those verbs. This vignette describes how individual expressions (function calls) are translated; `vignette("translation-verb")` describes how entire verbs are translated.

```{r}
library(dbplyr)
library(dplyr, warn.conflicts = FALSE)
```

## Getting started with translations

In this vignette, I'll use `lazy_frame()` to create a toy lazy table that allows us to see the translation without needing to connect to a real database:

```{r}
lf <- lazy_frame(x = 1, y = 2, g = "a")
lf |> mutate(z = (x + y) / 2)
```

The default `lazy_frame()` uses a generic database that generates (approximately) SQL-92 compliant SQL. You can use `simulate_*()` connections to see the translations used by different backends. Different databases generate slightly different SQL; see `vignette("new-backend")` for more details.

```{r}
lf_sqlite <- lazy_frame(x = 1, con = simulate_sqlite())
lf_access <- lazy_frame(x = 1, con = simulate_access())

lf_sqlite |> transmute(z = x^2)
lf_access |> transmute(z = x^2)
```

One key difference between dbplyr-generated SQL and hand-written SQL is that dbplyr always quotes all table and column names. This is verbose but necessary because column names in database tables can be any string, including SQL reserved words like `select` or `if`. Quoting all names ensures that dbplyr-generated SQL always works regardless of the table and column names involved.

In general, perfect translation is not possible because databases don't have all the functions that R does. The goal of dbplyr is to provide a semantic rather than a literal translation: what you mean, rather than precisely what is done. In fact, even for functions that exist both in databases and in R, you shouldn't expect results to be identical; database programmers have different priorities than R core programmers. For example, in R in order to get a higher level of numerical accuracy, `mean()` loops through the data twice. R's `mean()` also provides a `trim` option for computing trimmed means; this is something that databases do not provide. 

If you're interested in how `translate_sql()` is implemented, the basic techniques that underlie the implementation of `translate_sql()` are described in ["Advanced R"](https://adv-r.hadley.nz/translation.html). 

## Basic differences

There are two fundamental differences between R and SQL:

*   `"` and `'` mean different things. 
    R can use either `"` or `'` for strings, but in ANSI SQL, must be `"` used for names and must be `'` used for strings.

    ```{r}
    lf |> filter(x == "x")
    ```

*   R and SQL have different defaults for integers and reals.
    In R, 1 is a real, and 1L is an integer. In SQL, 1 is an integer, and 1.0 is a real.

    ```{r}
    lf |> transmute(z = 1)
    lf |> transmute(z = 1L)
    ```

## Known functions

### Mathematics

* basic math operators: `+`, `-`, `*`, `/`, `^`
* trigonometry: `acos()`, `asin()`, `atan()`, `atan2()`, `cos()`, `cot()`, `tan()`, `sin()`
* hyperbolic: `cosh()`, `coth()`, `sinh()`, `tanh()`
* logarithmic: `log()`, `log10()`, `exp()`
* misc: `abs()`, `ceiling()`, `floor()`, `sqrt()`, `sign()`, `round()`

```{r}
lf |> transmute(x = x / 2, y = x^2 + y^2)

lf |> transmute(x = log(x), y = round(y, 1))
```

### Modulo arithmetic

dbplyr translates `%%` to the SQL equivalents but note that it's not precisely the same: most databases use truncated division where the modulo operator takes the sign of the dividend, where R using the mathematically preferred floored division with the modulo sign taking the sign of the divisor.

```{r}
df <- tibble(
  x = c(10L, 10L, -10L, -10L),
  y = c(3L, -3L, 3L, -3L)
)
db <- copy_to(memdb(), df)

df |> mutate(x %% y)
db |> mutate(x %% y)
```

dbplyr no longer translates `%/%` because there's no robust cross-database translation available.

### Logical comparisons and boolean operations

* logical comparisons: `<`, `<=`, `!=`, `>=`, `>`, `==`, `%in%`, `between()`
* boolean operations: `&`, `&&`, `|`, `||`, `!`, `xor()`

```{r}
lf |> filter(x > 5 | y == 2)

lf |> filter(x %in% c(1, 2, 3))

lf |> filter(between(x, 1, 5))
```

### Bitwise operations

`bitwNot()`, `bitwAnd()`, `bitwOr()`, `bitwXor()`, `bitwShiftL()`, and `bitwShiftR()` are all supported:

```{r}
lf |> transmute(x = bitwAnd(x, 3L), y = bitwShiftL(x, 2L))
```

### Type coercion

Type coercion functions use the corresponding SQL `CAST()` call:

```{r}
lf |> transmute(x = as.integer(y), y = as.character(x))
```

* integer types: `as.integer()`, `as.integer64()`
* floating point: `as.numeric()`, `as.double()`
* character: `as.character()`
* logical: `as.logical()`
* date/time: `as.Date()`, `as.POSIXct()`

For database-specific types not covered by these functions, use `as()`:

```{r}
lf |> transmute(x = as(x, "TIME"), y = as(y, "DECIMAL(10, 2)"))
```

### `NULL`/`NA` handling

* `is.na()`, `is.null()`: test for `NULL`.
* `na_if()`: replace a value with `NULL`.
* `coalesce()`: replace `NULL` with a default value.

```{r}
lf |> filter(!is.na(x))

lf |> transmute(x = coalesce(x, 0L))

lf |> transmute(x = na_if(x, 0L))
```

### Aggregation

All databases provide translation for the basic aggregations: `mean()`, `sum()`, `min()`, `max()`. Databases automatically drop NULLs (their equivalent of missing values) whereas in R you have to ask nicely. The aggregation functions warn you about this important difference:

```{r}
lf |> summarise(z = mean(x))
lf |> summarise(z = mean(x, na.rm = TRUE))
```

Note that aggregation functions used inside `mutate()` or `filter()` generate a window translation:

```{r}
lf |> mutate(z = mean(x, na.rm = TRUE))
lf |> filter(mean(x, na.rm = TRUE) > 0)
```

Most backends also support:

* `sd()`, `var()`, `cor()`, `cov()`
* `median()`, `quantile()`
* `n()`, `n_distinct()`
* `all()`, `any()`
* `str_flatten()`

### Conditional evaluation

`if`, `ifelse()`, and `if_else()` are translated to `CASE WHEN`:

```{r}
lf |> transmute(z = ifelse(x > 5, "big", "small"))
```

`case_when()`, `case_match()`, and `switch()` are also supported:

```{r}
lf |> 
  mutate(z = case_when(
    x > 10 ~ "medium",
    x > 30 ~ "big", 
    .default = "small"
  ))

lf |> mutate(z = switch(g, a = 1L, b = 2L, 3L))
```

### String functions

Base R string functions and their stringr equivalents are widely supported:

* `nchar()`, `str_length()`
* `tolower()`, `toupper()`, `str_to_lower()`, `str_to_upper()`, `str_to_title()`
* `trimws()`, `str_trim()`
* `paste()`, `paste0()`, `str_c()`
* `substr()`, `substring()`, `str_sub()`

```{r}
lf |> transmute(x = paste0(g, " dog"))

lf |> transmute(x = substr(g, 1L, 2L))
```

Many backends also support regular expression functions like `str_detect()`, `str_replace()`, `str_replace_all()`, `str_remove()`, `str_remove_all()`, `str_squish()`, and `str_like()`. Support varies by backend; see the individual backend documentation for details.

### Date/time functions

dbplyr supports many lubridate functions for extracting date components:

* `today()`, `now()`
* `year()`, `month()`, `day()`, `mday()`, `hour()`, `minute()`, `second()`

```{r}
lf_dt <- lazy_frame(dt = Sys.time())

lf_dt |> transmute(
  year = year(dt),
  month = month(dt),
  day = day(dt)
)
```

Some backends also support additional lubridate functions including `yday()`, `wday()`, `week()`, `isoweek()`, `quarter()`, `isoyear()`, `floor_date()`, and period functions like `seconds()`, `minutes()`, `hours()`, `days()`, `weeks()`, `months()`, `years()`.

Several backends (including PostgreSQL, Snowflake, SQL Server, Redshift, and Spark SQL) support [clock](https://clock.r-lib.org) functions for date arithmetic. 

* `add_days()`, `add_years()`
* `date_build()`
* `get_year()`, `get_month()`, `get_day()`
* `date_count_between()`
* `difftime()`

clock functions tend to be easier to translate than lubridate functions because they are more specific.

### Other functions

* `pmin()`, `pmax()` for parallel min/max
* `desc()` for descending order
* `cut()` for binning numeric values into categories

## Unknown functions

Any function that dbplyr doesn't know how to convert is left as is. This means that database functions that are not covered by dbplyr can often be used directly.

### Prefix functions

Any function that dbplyr doesn't know about will be left as is:

```{r}
lf |> mutate(z = foofify(x, y))
```

But to make it clear that you're deliberately calling a SQL function, we recommend using the `.sql` pronoun:

```{r}
lf |> transmute(z = .sql$foofify(x, y))
```

If you're working inside a package, this also makes it easier to avoid `R CMD CHECK` notes. Just import `.sql` from dbplyr using a roxygen2 tag like `@importFrom dbplyr .sql`

### Infix functions

As well as prefix functions (where the name of the function comes before the arguments), dbplyr also translates infix functions. That allows you to use expressions like `LIKE`, which does a limited form of pattern matching:

```{r}
lf |> filter(x %LIKE% "%foo%")
```

You can also use `str_like()` for this common case:

```{r}
lf |> filter(str_like(x, "%foo%"))
```

You could use `%||%` for string concatenation, but in most cases it's more R-like to use `paste()` or `paste0()`:

```{r}
lf |> transmute(z = x %||% y)
lf |> transmute(z = paste0(x, y))
lf |> transmute(z = paste(x, y))
```

### Special forms

SQL functions tend to have a greater variety of syntax than R. That means there are a number of expressions that can't be translated directly from R code. To insert these in your own queries, you can use literal SQL inside `sql()`:

```{r}
lf |> transmute(z = sql("x!"))
lf |> transmute(z = x == sql("ANY VALUES(1, 2, 3)"))
```

This gives you a lot of freedom to generate the SQL you need:

```{r}
lf |> transmute(factorial = sql("x!"))
lf |> transmute(factorial = sql("CAST(x AS FLOAT)"))
```

### Error for unknown translations

If needed, you can also use the `dplyr.strict_sql` option to force dbplyr to error if it doesn't know how to translate a function:

```{r}
#| error = TRUE
options(dplyr.strict_sql = TRUE)
lf |> mutate(z = glob(x, y))
```

## Window functions

Things get a little trickier with window functions, because SQL's window functions are considerably more expressive than the specific variants provided by base R or dplyr. They have the form `[expression] OVER ([partition clause] [order clause] [frame_clause])`:

* The __expression__ is a combination of variable names and window functions.
  Support for window functions varies from database to database, but most
  support:

  - ranking: `row_number()`, `min_rank()`, `rank()`, `dense_rank()`,
    `percent_rank()`, `cume_dist()`, `ntile()`;
  - offsets: `lead()`, `lag()`, `first()`, `last()`, `nth()`;
  - aggregates: `mean()`, `sum()`, `min()`, `max()`, `n()`, `n_distinct()`;
  - cumulative: `cummean()`, `cumsum()`, `cummin()`, `cummax()`.

* The __partition clause__ specifies how the window function is broken down
  over groups. It plays an analogous role to `GROUP BY` for aggregate functions,
  and `group_by()` in dplyr. It is possible for different window functions to 
  be partitioned into different groups, but not all databases support it, and
  neither does dplyr.
  
*   The __order clause__ controls the ordering (when it makes a difference).
    This is important for the ranking functions since it specifies which 
    variables to rank by, but it's also needed for cumulative functions and lead.
    Whenever you're thinking about before and after in SQL, you must always tell 
    it which variable defines the order. If the order clause is missing when 
    needed, some databases fail with an error message while others return 
    non-deterministic results.
  
*   The __frame clause__ defines which rows, or __frame__, that are passed 
    to the window function, describing which rows (relative to the current row)
    should be included. The frame clause provides two offsets which determine
    the start and end of frame. There are three special values: -Inf means
    to include all preceding rows (in SQL, "unbounded preceding"), 0 means the
    current row ("current row"), and Inf means all following rows ("unbounded
    following"). The complete set of options is comprehensive, but fairly 
    confusing, and is summarised visually below.

    ```{r}
    #| echo: false
    #| out-width: 100%
    #| fig-alt: >
    #|   A visual summary of the frame clause using the real line labelled
    #|   with negative infinity, -3, -2, -1, 0, 1, 2, 3, infinity. The most
    #|   important clauses are rolling, cumulative, and recycling. 
    #|   Rolling, e.g. between 1 preceding and 1, following, run from 
    #|   -1 to -1. Cumulative, between unbounded preceding and 
    #|   current row, runs from negative infinity to 0. Recycled, 
    #|   between unbound preceeding and unbound following, runs from 
    #|   negative infinity to positive infinity.
    knitr::include_graphics("windows.png", dpi = 300)
    ```

    Of the many possible specifications, only three are commonly
    used. They select between aggregation variants:

    * Recycled: `BETWEEN UNBOUND PRECEDING AND UNBOUND FOLLOWING`
    
    * Cumulative: `BETWEEN UNBOUND PRECEDING AND CURRENT ROW`
    
    * Rolling: `BETWEEN 2 PRECEDING AND 2 FOLLOWING`
    
    dbplyr generates the frame clause based on whether you're using a recycled
    aggregate or a cumulative aggregate.
    
To see how individual window functions are translated to SQL, we can use `transmute()`:

```{r}
lf <- lazy_frame(g = 1, year = 2020, id = 3, con = simulate_dbi())

lf |> transmute(
  mean = mean(g), 
  rank = min_rank(g), 
  cumsum = cumsum(g),
  lag = lag(g)
)
```

If the lazy frame has been grouped or arranged previously in the pipeline, then dbplyr will use that information to set the "partition by" and "order by" clauses:

```{r}
lf |> arrange(year) |> mutate(z = cummean(g))
lf |> group_by(id) |> mutate(z = rank())
```

There are some challenges when translating window functions between R and SQL, because dbplyr tries to keep the window functions as similar as possible to both the existing R analogues and to the SQL functions. This means that there are three ways to control the order clause depending on which window function you're using:

* For ranking functions, the ordering variable is the first argument: `rank(x)`,
  `ntile(y, 2)`. If omitted or `NULL`, will use the default ordering associated 
  with the tbl (as set by `arrange()`).

* Accumulating aggregates only take a single argument (the vector to aggregate).
  To control ordering, use `order_by()`.
  
* Aggregates implemented in dplyr (`lead()`, `lag()`, `nth()`, `first()`,
  `last()`) have an `order_by` argument. Supply it to override the
  default ordering.

The three options are illustrated in the snippet below:

```{r}
lf |> transmute(
  x1 = min_rank(g),
  x2 = order_by(year, cumsum(g)),
  x3 = lead(g, order_by = year)
)
```

Currently there is no way to order by multiple variables, except by setting the default ordering with `arrange()`. This will be added in a future release. 

# PostgreSQL

## Install

First install PostgreSQL, create a data directory, and create a default database.

```
brew install postgresql
export PGDATA=~/db/postgres # set this globally somewhere
initdb
createdb test

psql -c "CREATE USER postgres WITH PASSWORD 'password';"
psql -c "ALTER USER postgres WITH SUPERUSER;"
```

## Start

```
pg_ctl start
```

## Connect

```{r, eval = FALSE}
library(DBI)
con <- dbConnect(RPostgres::Postgres(), dbname = "hadley")
dbListTables(con)

con <- dbConnect(
  RPostgres::Postgres(), 
  dbname = "test", 
  user = "postgres", 
  password = "passowrd"
)
```

# MariaDB

## Install

```
brew install mariadb

mariadb-install-db --verbose --user=hadley --basedir=/usr/local \
  --datadir=/User/hadley/db/mariadb --tmpdir=/tmp

mysqld --datadir='/Users/hadley/db/mariadb'

mariadb -e "CREATE DATABASE lahman;"
mariadb -e "CREATE DATABASE nycflights13;"
```


## Connect

```{r, eval = FALSE}
install.packages("RMariaDB")

con <- dbConnect(RMariaDB::MariaDB(), dbname = "lahman", password = "")
dbListTables(con)
```

# Shut down

```
mysqladmin shutdown -u root -p
```

---
title: "Verb translation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Verb translation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
options(tibble.print_min = 6L, tibble.print_max = 6L, digits = 3)
```

There are two parts to dbplyr SQL translation: translating dplyr verbs, and translating expressions within those verbs. This vignette describes how entire verbs are translated; `vignette("translation-function")` describes how individual expressions within those verbs are translated.

All dplyr verbs generate a `SELECT` statement. To demonstrate we'll make a temporary database with a couple of tables

```{r, message = FALSE}
library(dbplyr)
library(dplyr)

con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
flights <- copy_to(con, nycflights13::flights)
airports <- copy_to(con, nycflights13::airports)
```

## Single table verbs

*  `select()` and `mutate()` modify the `SELECT` clause:

    ```{r}
    flights |>
      select(contains("delay")) |>
      show_query()
    
    flights |>
      select(distance, air_time) |>  
      mutate(speed = distance / (air_time / 60)) |>
      show_query()
    ```
    
*  `filter()` generates a `WHERE` clause:
    
    ```{r}
    flights |> 
      filter(month == 1, day == 1) |>
      show_query()
    ```

*  `arrange()` generates an `ORDER BY` clause:

    ```{r}
    flights |> 
      arrange(carrier, desc(arr_delay)) |>
      show_query()
    ```

*  `summarise()` and `group_by()` work together to generate a `GROUP BY` clause:

    ```{r}
    flights |>
      group_by(month, day) |>
      summarise(delay = mean(dep_delay, na.rm = TRUE)) |>
      show_query()
    ```

## Subqueries

It's not always possible to translate a single dplyr verb into a single SQL query. For example, in SQL, variables in the `SELECT` clause have to come from another table; you can't refer to a variable that you just created. For that reason, dbplyr will create subqueries where needed:

```{r}
flights |>
  select(distance, air_time) |>  
  mutate(
    air_time_h = air_time / 60,
    speed = distance / air_time_h) |>
  show_query()
```

It's also possible to use a CTE if you so desire:

```{r}
flights |>
  select(distance, air_time) |>  
  mutate(
    air_time_h = air_time / 60,
    speed = distance / air_time_h) |>
  show_query(sql_options = sql_options(cte = TRUE))
```

Sometimes dbplyr will create a subquery where it's not strictly necessary. We strive to avoid this as much as possible, but our analysis of the generated SQL is not always complete, so we'll typically err on the side of safety (creating more subqueries) rather than performance.

It's important to know that most SQL dialects either error or include `LIMIT` and `ORDER BY` statements inside of subqueries. For that reason, you should always put `head()` and `arrange()` as late as possible in your pipeline.

## Dual table verbs

| R                | SQL
|------------------|------------------------------------------------------------
| `inner_join()`   | `SELECT * FROM x JOIN y ON x.a = y.a`
| `left_join()`    | `SELECT * FROM x LEFT JOIN y ON x.a = y.a`
| `right_join()`   | `SELECT * FROM x RIGHT JOIN y ON x.a = y.a`
| `full_join()`    | `SELECT * FROM x FULL JOIN y ON x.a = y.a`
| `semi_join()`    | `SELECT * FROM x WHERE EXISTS (SELECT 1 FROM y WHERE x.a = y.a)`
| `anti_join()`    | `SELECT * FROM x WHERE NOT EXISTS (SELECT 1 FROM y WHERE x.a = y.a)`
| `intersect(x, y)`| `SELECT * FROM x INTERSECT SELECT * FROM y`
| `union(x, y)`    | `SELECT * FROM x UNION SELECT * FROM y`
| `setdiff(x, y)`  | `SELECT * FROM x EXCEPT SELECT * FROM y`

`x` and `y` don't have to be tables in the same database. If you specify `copy = TRUE`, dplyr will copy the `y` table into the same location as the `x` variable. This is useful if you've downloaded a summarised dataset and determined a subset of interest that you now want the full data for. You can use `semi_join(x, y, copy = TRUE)` to upload the indices of interest to a temporary table in the same database as `x`, and then perform a efficient semi join in the database. 

If you're working with large data, it maybe also be helpful to set `auto_index = TRUE`. That will automatically add an index on the join variables to the temporary table.

---
title: "Writing SQL with dbplyr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Writing SQL with dbplyr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette discusses why you might use dbplyr instead of writing SQL yourself, and what to do when dbplyr's built-in translations can't create the SQL that you need.

```{r setup, message = FALSE}
library(dplyr)
library(dbplyr)

mf <- memdb_frame(x = 1, y = 2)
```

## Why use dbplyr?

One simple nicety of dplyr is that it will automatically generate subqueries if you want to use a freshly created variable in `mutate()`: 

```{r}
mf %>% 
  mutate(
    a = y * x, 
    b = a ^ 2,
  ) %>% 
  show_query()
```

In general, it's much easier to work iteratively in dbplyr. You can easily give intermediate queries names, and reuse them in multiple places. Or if you have a common operation that you want to do to many queries, you can easily wrap it up in a function. It's also easy to chain `count()` to the end of any query to check the results are about what you expect.

## What happens when dbplyr fails?

dbplyr aims to translate the most common R functions to their SQL equivalents, allowing you to ignore the vagaries of the SQL dialect that you're working with, so you can focus on the data analysis problem at hand. But different backends have different capabilities, and sometimes there are SQL functions that don't have exact equivalents in R. In those cases, you'll need to write SQL code directly. 

Any function that dbplyr doesn't know about will be left as is:

```{r}
mf %>% 
  mutate(z = foofify(x, y)) %>% 
  show_query()

mf %>% 
  filter(x %LIKE% "%foo%") %>% 
  show_query()
```

SQL functions tend to have a greater variety of syntax than R. That means there are a number of expressions that can't be translated directly from R code. To insert these in your own queries, you can use literal SQL inside `sql()`:

```{r}
mf %>% 
  transmute(factorial = sql("x!")) %>% 
  show_query()

mf %>% 
  transmute(factorial = sql("CAST(x AS FLOAT)")) %>% 
  show_query()
```

Learn more in `vignette("translation-function")`.

# MySQL

## Install drivers

```
brew tap microsoft/mssql-release https://github.com/Microsoft/homebrew-mssql-release
brew update
brew install msodbcsql17 mssql-tools
```

Check available:

```{r}
odbc::odbcListDrivers()
edit_file("/usr/local/etc/odbcinst.ini")
Sys.setenv("GITHUB_MSSQL" = "true")
```

## Start

```
docker run -e 'ACCEPT_EULA=Y' -e 'SA_PASSWORD=Password12' -t -i -p 1433:1433 microsoft/mssql-server-linux

sqlcmd -U SA -P 'Password12' -Q 'CREATE DATABASE test;'
```

## Connect

```{r, eval = FALSE}
con <- DBI::dbConnect(odbc::odbc(),
  driver = "ODBC Driver 17 for SQL Server",
  database = "test",
  uid = "SA",
  pwd = "Password12",
  server = "localhost",
  port = 1433
)
```

# SAP HANA

## Sign up for free cloud trial

* <https://saphanajourney.com/hana-cloud/plans/>
* Go to <https://account.hanatrial.ondemand.com/cockpit>
* Click trial, then "dev" space.
* Then follow <https://saphanajourney.com/hana-cloud/learning-article/how-to-create-your-trial-sap-hana-cloud-instance/>
* Allow access from all IP addresses.
* Need to restart service every day; it's automatically stopped each night.
* Need to create `test` database using database explorer

## Install drivers

* Download SAP HANA installer from <https://www.sap.com/cmp/td/sap-hana-express-edition.html>
* Uncheck virtual machine; check "clients (mac)"
* Untar download
* Untar `hdb_client_mac.tgz`
* `./hdb_inst`

Configure odbc.init: `edit_file("/usr/local/etc/odbcinst.ini")`

```
[SAP HANA]
Description=SAP HANA
Driver=/Applications/sap/hdbclient/libodbcHDB.dylib
```

Check:

```{r}
odbc::odbcListDrivers()
```

## Connect

```{r, eval = FALSE}
con <- DBI::dbConnect(odbc::odbc(),
  driver = "SAP HANA",
  uid = "DBADMIN",
  pwd = "Password12",
  servernode = "a9441502-7166-4a5a-b5d8-23abe3b5009c.hana.trial-us10.hanacloud.ondemand.com:443",
  encrypt="true",
  sslValidateCertificate = "false"
)
class(con)
#> [1] "HDB"
```

# index fails if columns are missing

    Code
      (expect_error(compute(mf, indexes = list(c("y", "x", "z"), "a"))))
    Output
      <error/rlang_error>
      Error in `compute()`:
      ! All columns specified through `indexes` must exist in `x`.
      i The following columns are missing from `indexes`: y, z, and a.
    Code
      (expect_error(compute(mf, unique_indexes = list(c("y", "x", "z"), "a"))))
    Output
      <error/rlang_error>
      Error in `compute()`:
      ! All columns specified through `unique_indexes` must exist in `x`.
      i The following columns are missing from `unique_indexes`: y, z, and a.

# compute can handle schema

    Code
      df %>% compute(name = in_schema("main", "db1"), temporary = FALSE)
    Condition
      Error in `db_compute()`:
      ! Can't copy query to table `main`.`db1`.
      Caused by error in `db_save_query.DBIConnection()`:
      ! Can't save query to table `main`.`db1`.
      i Using SQL: CREATE TABLE `main`.`db1` AS SELECT * FROM `dbplyr_{tmp}`
      Caused by error:
      ! dummy DBI error

# collect() handles DBI error

    Code
      (expect_error(mf %>% mutate(a = sql("invalid sql")) %>% collect()))
    Output
      <error/rlang_error>
      Error in `collect()`:
      ! Failed to collect lazy table.
      Caused by error:
      ! dummy DBI error

# compute(temporary = FALSE) without a name is deprecated

    The `name` argument of `compute()` must be provided when `temporary = FALSE` as of dbplyr 2.3.3.

# custom header

    Code
      # Number of rows is shown
      x <- memdb_frame(a = 1:3) %>% filter(a > 0)
      setup <- pillar::tbl_format_setup(x)
      tbl_format_header(x, setup)[[1]]
    Output
      [1] "# Source:   SQL [3 x 1]"
    Code
      # Number of rows still can't be shown
      x <- memdb_frame(a = 1:11) %>% filter(a > 0)
      setup <- pillar::tbl_format_setup(x)
      tbl_format_header(x, setup)[[1]]
    Output
      [1] "# Source:   SQL [?? x 1]"

# con must not be NULL

    Code
      escape("a")
    Condition
      Error in `escape()`:
      ! `con` must not be NULL.

---

    Code
      sql_vector("a")
    Condition
      Error in `sql_vector()`:
      ! `con` must not be NULL.

# other objects get informative error

    Code
      lf %>% filter(x == input)
    Condition
      Error:
      ! Cannot translate shiny inputs to SQL.
      i Do you want to force evaluation in R with (e.g.) `!!input$x` or `local(input$x)`?
    Code
      lf %>% filter(x == x())
    Condition
      Error:
      ! Cannot translate a shiny reactive to SQL.
      i Do you want to force evaluation in R with (e.g.) `!!foo()` or `local(foo())`?
    Code
      lf %>% filter(x == df)
    Condition
      Error:
      ! Cannot translate a data.frame to SQL.
      i Do you want to force evaluation in R with (e.g.) `!!df$x` or `local(df$x)`?
    Code
      lf %>% filter(x == mean)
    Condition
      Error:
      ! Cannot translate a function to SQL.
      i Do you want to force evaluation in R with (e.g.) `!!x` or `local(x)`?

# generates custom sql

    Code
      sql_table_analyze(con_maria, in_schema("schema", "tbl"))
    Output
      <SQL> ANALYZE TABLE `schema`.`tbl`

---

    Code
      sql_query_explain(con_maria, sql("SELECT * FROM table"))
    Output
      <SQL> EXPLAIN SELECT * FROM table

---

    Code
      left_join(lf, lf, by = "x", na_matches = "na")
    Output
      <SQL>
      SELECT `df_LHS`.`x` AS `x`
      FROM `df` AS `df_LHS`
      LEFT JOIN `df` AS `df_RHS`
        ON (`df_LHS`.`x` <=> `df_RHS`.`x`)

---

    Code
      full_join(lf, lf, by = "x")
    Condition
      Error in `sql_query_join()`:
      ! MySQL does not support full joins

---

    Code
      slice_sample(lf, n = 1)
    Output
      <SQL>
      SELECT `x`
      FROM (
        SELECT `df`.*, ROW_NUMBER() OVER (ORDER BY RAND()) AS `col01`
        FROM `df`
      ) AS `q01`
      WHERE (`col01` <= 1)

---

    Code
      copy_inline(con_maria, tibble(x = 1:2, y = letters[1:2])) %>% remote_query()
    Output
      <SQL> SELECT CAST(`x` AS INTEGER) AS `x`, CAST(`y` AS CHAR) AS `y`
      FROM (
        SELECT NULL AS `x`, NULL AS `y`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES (1, 'a'), (2, 'b')
      ) AS `values_table`

---

    Code
      copy_inline(con_mysql, tibble(x = 1:2, y = letters[1:2])) %>% remote_query()
    Output
      <SQL> SELECT TRUNCATE(CAST(`x` AS DOUBLE), 0) AS `x`, CAST(`y` AS CHAR) AS `y`
      FROM (
        SELECT NULL AS `x`, NULL AS `y`
        WHERE (0 = 1)
      
        UNION ALL
      
        VALUES ROW(1, 'a'), ROW(2, 'b')
      ) AS `values_table`

# `sql_query_update_from()` is correct

    Code
      sql_query_update_from(con = con, table = ident("df_x"), from = sql_render(df_y,
        con, lvl = 1), by = c("a", "b"), update_values = sql(c = "COALESCE(`df_x`.`c`, `...y`.`c`)",
        d = "`...y`.`d`"))
    Output
      <SQL> UPDATE `df_x`
      INNER JOIN (
        SELECT `a`, `b`, `c` + 1.0 AS `c`, `d`
        FROM `df_y`
      ) AS `...y`
        ON `...y`.`a` = `df_x`.`a` AND `...y`.`b` = `df_x`.`b`
      SET `df_x`.`c` = COALESCE(`df_x`.`c`, `...y`.`c`), `df_x`.`d` = `...y`.`d`

---

    Argument `returning_cols` isn't supported in MariaDB translation.

# can explain

    Code
      db %>% mutate(y = x + 1) %>% explain()
    Output
      <SQL>
      SELECT `test`.*, `x` + 1.0 AS `y`
      FROM `test`
      
      <PLAN>
        id select_type table type possible_keys  key key_len  ref rows Extra
      1  1      SIMPLE  test  ALL          <NA> <NA>    <NA> <NA>    3      

# expand completes all values

    Code
      tidyr::expand(lazy_frame(x = 1, y = 1), x, y)
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT DISTINCT "x"
        FROM "df"
      ) AS "LHS"
      CROSS JOIN (
        SELECT DISTINCT "y"
        FROM "df"
      ) AS "RHS"

# nesting doesn't expand values

    Code
      tidyr::expand(df_lazy, nesting(x, y))
    Output
      <SQL>
      SELECT DISTINCT *
      FROM "df"

# expand accepts expressions

    Code
      tidyr::expand(df, round(x / 2))
    Output
      <SQL>
      SELECT DISTINCT ROUND("x" / 2.0, 0) AS "round(x/2)"
      FROM "df"

---

    Code
      tidyr::expand(df, nesting(x_half = round(x / 2), x1 = x + 1))
    Output
      <SQL>
      SELECT DISTINCT ROUND("x" / 2.0, 0) AS "x_half", "x" + 1.0 AS "x1"
      FROM "df"

# works with tidyr::nesting

    Code
      tidyr::expand(df_lazy, tidyr::nesting(x, y))
    Output
      <SQL>
      SELECT DISTINCT *
      FROM "df"

# expand respects groups

    Code
      tidyr::expand(group_by(df_lazy, a), b, c)
    Output
      <SQL>
      SELECT "LHS".*, "c"
      FROM (
        SELECT DISTINCT "a", "b"
        FROM "df"
      ) AS "LHS"
      LEFT JOIN (
        SELECT DISTINCT "a", "c"
        FROM "df"
      ) AS "RHS"
        ON ("LHS"."a" = "RHS"."a")

# NULL inputs

    Code
      tidyr::expand(lazy_frame(x = 1), x, y = NULL)
    Output
      <SQL>
      SELECT DISTINCT *
      FROM "df"

# expand() errors when expected

    Code
      tidyr::expand(local_memdb_frame(x = 1))
    Condition
      Error in `tidyr::expand()`:
      ! Must supply variables in `...`

---

    Code
      tidyr::expand(local_memdb_frame(x = 1), x = NULL)
    Condition
      Error in `tidyr::expand()`:
      ! Must supply variables in `...`

# expand() errors for non-column expressions

    Code
      tidyr::expand(lf, x, 1:3)
    Condition
      Error in `tidyr::expand()`:
      ! In expression `1:3`:
      Caused by error:
      ! Every expression must use at least one data column
      i `1:3` doesn't use any columns.
    Code
      tidyr::expand(lazy_frame(x = 1, y = 1), nesting(x, 1))
    Condition
      Error in `tidyr::expand()`:
      ! In expression `nesting(x, 1)`:
      Caused by error:
      ! Every expression must use at least one data column
      i `1` doesn't use any columns.

# nesting() respects .name_repair

    Code
      tidyr::expand(local_memdb_frame(x = 1, y = 1), nesting(x, x = x + 1))
    Condition
      Error in `tidyr::expand()`:
      ! In expression `nesting(x, x = x + 1)`:
      Caused by error:
      ! Names must be unique.
      x These names are duplicated:
        * "x" at locations 1 and 2.

# replace_na replaces missing values

    Code
      tidyr::replace_na(lazy_frame(x = 1, y = "a"), list(x = 0, y = "unknown"))
    Output
      <SQL>
      SELECT COALESCE("x", 0.0) AS "x", COALESCE("y", 'unknown') AS "y"
      FROM "df"

# replace_na ignores missing columns

    Code
      tidyr::replace_na(lazy_frame(x = 1), list(not_there = 0))
    Output
      <SQL>
      SELECT *
      FROM "df"

# complete completes missing combinations

    Code
      tidyr::complete(df_lazy, x, y, fill = list(z = "c"))
    Output
      <SQL>
      SELECT "x", "y", COALESCE("z", 'c') AS "z"
      FROM (
        SELECT
          COALESCE("LHS"."x", "df"."x") AS "x",
          COALESCE("LHS"."y", "df"."y") AS "y",
          "z"
        FROM (
          SELECT "x", "y"
          FROM (
            SELECT DISTINCT "x"
            FROM "df"
          ) AS "LHS"
          CROSS JOIN (
            SELECT DISTINCT "y"
            FROM "df"
          ) AS "RHS"
        ) AS "LHS"
        FULL JOIN "df"
          ON ("LHS"."x" = "df"."x" AND "LHS"."y" = "df"."y")
      ) AS "q01"

# can print lazy_select_query

    Code
      lazy_select_query(x = lf$lazy_query, select = quos(x_mean = mean(x), y2 = y),
      where = quos(y > 1, x == y - 2), group_by = quos("x"))
    Output
      <SQL SELECT>
      From:
        <table_path> `df`
      Select:   x_mean = mean(x), y2 = y
      Where:    y > 1, x == y - 2
      Group by: "x"

# unnamed results must be data frames

    Code
      do(mf, nrow(.))
    Condition
      Error in `label_output_dataframe()`:
      ! Results must be data frames
      Problems at positions 1 and 2

# do() argument checking works

    Code
      do(mf, nrow = nrow(.), ncol(.))
    Condition
      Error in `named_args()`:
      ! Arguments to `do()` must either be all named or all unnamed

---

    Code
      do(mf, nrow(.), ncol(.))
    Condition
      Error in `named_args()`:
      ! Can only supply single unnamed argument to `do()`

---

    Code
      do(mf, .f = nrow)
    Condition
      Error in `named_args()`:
      ! `do()` syntax changed in dplyr 0.2. Please see documentation for details

# do() is deprecated

    Code
      . <- do(local_memdb_frame(x = 1), head(.))
    Condition
      Warning:
      `do()` was deprecated in dbplyr 2.6.0.

# generates custom sql

    Code
      sql_table_analyze(con, in_schema("schema", "tbl"))
    Output
      <SQL> ANALYZE TABLE `schema`.`tbl` COMPUTE STATISTICS

---

    Code
      union_all(lf, lf)
    Output
      <SQL>
      SELECT *
      FROM `df`
      
      UNION ALL
      
      SELECT *
      FROM `df`

# is properly vectorised

    Code
      new_table_ident(table = c("A", "B", "c"), schema = c("schema1", "schema2"))
    Condition
      Error:
      ! Can't recycle `table` (size 3) to match `schema` (size 2).

# can't supply table and sql

    Code
      new_table_ident(schema = "my schema", table = "my table", quoted = TRUE)
    Condition
      Error in `purrr::pmap()`:
      i In index: 1.
      Caused by error:
      ! Can't supply a schema when `table` is quoted.

# must supply table and schema when catalog is used

    Code
      new_table_ident(table = "my table", catalog = "cat")
    Condition
      Error in `purrr::pmap()`:
      i In index: 1.
      Caused by error:
      ! Must supply `schema` when `catalog` is supplied.
    Code
      new_table_ident(schema = "schema", catalog = "cat")
    Condition
      Error in `purrr::pmap()`:
      i In index: 1.
      Caused by error:
      ! Must supply `table` when `schema` is supplied.

---

    Code
      new_table_ident(table = "my table", schema = c("my schema", NA), catalog = "cat")
    Condition
      Error in `purrr::pmap()`:
      i In index: 2.
      Caused by error:
      ! Must supply `schema` when `catalog` is supplied.

# can't coerce or cast to character

    Code
      c(table, "character")
    Condition
      Error in `vec_c()`:
      ! Can't combine `..1` <dbplyr_table_ident> and `..2` <character>.
    Code
      as.character(table)
    Condition
      Error in `as.character()`:
      ! Can't convert `x` <dbplyr_table_ident> to <character>.

# can print

    Code
      new_table_ident(table = "table")
    Output
      <dbplyr_table_ident[1]>
      [1] `table`
    Code
      new_table_ident(schema = "schema", table = "table")
    Output
      <dbplyr_table_ident[1]>
      [1] `schema`.`table`
    Code
      new_table_ident(catalog = "catalog", schema = "schema", table = "table")
    Output
      <dbplyr_table_ident[1]>
      [1] `catalog`.`schema`.`table`
    Code
      new_table_ident(table = "`my schema`.`my table`", quoted = TRUE)
    Output
      <dbplyr_table_ident[1]>
      [1] `my schema`.`my table`

---

    Code
      new_table_ident(table = c("`my schema`.`my table`", "table1", "table2",
        "table3"), schema = c(NA, NA, "schema2", "schema3"), catalog = c(NA, NA, NA,
        "catalog3"), quoted = c(TRUE, FALSE, FALSE, FALSE))
    Output
      <dbplyr_table_ident[4]>
      [1] `my schema`.`my table`        `table1`                     
      [3] `schema2`.`table2`            `catalog3`.`schema3`.`table3`

# as_table_ident works

    Code
      expect_equal(as_table_ident(ident_q("my schema.my table")), new_table_ident(
        table = "my schema.my table", quoted = TRUE))
    Message
      i Using `ident_q()` for a table identifier is intended as fallback in case of bugs.
      i If you need it to work around a bug please open an issue <https://github.com/tidyverse/dbplyr/issues>.
      This message is displayed once every 8 hours.
    Code
      expect_equal(as_table_ident(sql("my schema.my table")), new_table_ident(table = "my schema.my table",
        quoted = TRUE))
    Message
      i Using `sql()` for a table identifier is intended as fallback in case of bugs.
      i If you need it to work around a bug please open an issue <https://github.com/tidyverse/dbplyr/issues>.
      This message is displayed once every 8 hours.

# can pivot all cols to wide

    Code
      dbplyr_pivot_wider_spec(lazy_frame(key = c("x", "y", "z"), val = 1:3), spec)
    Output
      <SQL>
      SELECT
        MAX(CASE WHEN ("key" = 'x') THEN "val" END) AS "x",
        MAX(CASE WHEN ("key" = 'y') THEN "val" END) AS "y",
        MAX(CASE WHEN ("key" = 'z') THEN "val" END) AS "z"
      FROM "df"

# implicit missings turn into explicit missings

    Code
      show_query(dbplyr_pivot_wider_spec(df, spec))
    Output
      <SQL>
      SELECT
        `a`,
        MAX(CASE WHEN (`key` = 'x') THEN `val` END) AS `x`,
        MAX(CASE WHEN (`key` = 'y') THEN `val` END) AS `y`
      FROM `df`
      GROUP BY `a`

# error when overwriting existing column

    Code
      tidyr::pivot_wider(df, names_from = key, values_from = val)
    Condition
      Error in `dbplyr_pivot_wider_spec()`:
      ! Names must be unique.
      x These names are duplicated:
        * "a" at locations 1 and 2.

# values_fn can be a single function

    Code
      suppressWarnings(dbplyr_pivot_wider_spec(df, spec1, values_fn = sum))
    Output
      <SQL>
      SELECT "a", SUM(CASE WHEN ("key" = 'x') THEN "val" END) AS "x"
      FROM "df"
      GROUP BY "a"

# values_fn can be a formula or function

    Code
      dbplyr_pivot_wider_spec(df, spec1, values_fn = ~ sum(.x, na.rm = TRUE))
    Output
      <SQL>
      SELECT "a", SUM(CASE WHEN ("key" = 'x') THEN "val" END) AS "x"
      FROM "df"
      GROUP BY "a"

---

    Code
      dbplyr_pivot_wider_spec(df, spec1, values_fn = function(x) sum(x, na.rm = TRUE))
    Output
      <SQL>
      SELECT "a", SUM(CASE WHEN ("key" = 'x') THEN "val" END) AS "x"
      FROM "df"
      GROUP BY "a"

# values_fn can be a named list

    `values_fn` must specify a function for each col in `values_from`

---

    Can't pivot column b:
    Caused by error:
    ! Can't convert `values_fn$b`, `NULL`, to a function.

# values_fn cannot be NULL

    Code
      dbplyr_pivot_wider_spec(df, spec1, values_fn = NULL)
    Condition
      Error in `dbplyr_pivot_wider_spec()`:
      ! `values_fn` must specify a function for each col in `values_from`

# `unused_fn` is validated

    Code
      (expect_error(tidyr::pivot_wider(df, id_cols = id, unused_fn = 1)))
    Output
      <error/rlang_error>
      Error in `tidyr::pivot_wider()`:
      ! `unused_fn` must be `NULL`, a function, or a named list of functions.

# can fill in missing cells

    Code
      dbplyr_pivot_wider_spec(df_lazy, spec, values_fill = 0)
    Output
      <SQL>
      SELECT
        "g",
        MAX(CASE WHEN ("name" = 'x') THEN "value" WHEN NOT ("name" = 'x') THEN 0.0 END) AS "x",
        MAX(CASE WHEN ("name" = 'y') THEN "value" WHEN NOT ("name" = 'y') THEN 0.0 END) AS "y"
      FROM "df"
      GROUP BY "g"

---

    Code
      dbplyr_pivot_wider_spec(df_lazy, spec, values_fill = list(value = 0))
    Output
      <SQL>
      SELECT
        "g",
        MAX(CASE WHEN ("name" = 'x') THEN "value" WHEN NOT ("name" = 'x') THEN 0.0 END) AS "x",
        MAX(CASE WHEN ("name" = 'y') THEN "value" WHEN NOT ("name" = 'y') THEN 0.0 END) AS "y"
      FROM "df"
      GROUP BY "g"

# values_fill is checked

    Code
      dbplyr_pivot_wider_spec(lf, spec, values_fill = 1:2)
    Condition
      Error in `dbplyr_pivot_wider_spec()`:
      ! `values_fill` must be `NULL`, a scalar, or a named list, not an integer vector.

# cannot pivot lazy frames

    Code
      tidyr::pivot_wider(lazy_frame(name = "x", value = 1))
    Condition
      Error in `dbplyr_build_wider_spec()`:
      ! `dbplyr_build_wider_spec()` doesn't work with local lazy tibbles.
      i Use `memdb_frame()` together with `show_query()` to see the SQL code.

# `names_from` must be supplied if `name` isn't in `data` (#1240)

    Code
      (expect_error(tidyr::pivot_wider(df, values_from = val)))
    Output
      <error/vctrs_error_subscript_oob>
      Error in `dbplyr_build_wider_spec()`:
      ! Can't select columns that don't exist.
      x Column `name` doesn't exist.

# `values_from` must be supplied if `value` isn't in `data` (#1240)

    Code
      (expect_error(tidyr::pivot_wider(df, names_from = key)))
    Output
      <error/vctrs_error_subscript_oob>
      Error in `dbplyr_build_wider_spec()`:
      ! Can't select columns that don't exist.
      x Column `value` doesn't exist.

# `names_from` must identify at least 1 column (#1240)

    Code
      (expect_error(tidyr::pivot_wider(df, names_from = starts_with("foo"),
      values_from = val)))
    Output
      <error/rlang_error>
      Error in `dbplyr_build_wider_spec()`:
      ! `names_from` must select at least one column.

# `values_from` must identify at least 1 column (#1240)

    Code
      (expect_error(tidyr::pivot_wider(df, names_from = key, values_from = starts_with(
        "foo"))))
    Output
      <error/rlang_error>
      Error in `dbplyr_build_wider_spec()`:
      ! `values_from` must select at least one column.

# `id_expand` must be FALSE

    Code
      (expect_error(tidyr::pivot_wider(df, id_expand = TRUE)))
    Output
      <error/dbplyr_error_unsupported_arg>
      Error in `tidyr::pivot_wider()`:
      ! `id_expand = TRUE` isn't supported on database backends.
      i It must be FALSE instead.

# useful error if bad identifier

    Code
      sql_glue2(con, "{.id 1}")
    Condition
      Error in `sql_glue2()`:
      ! {.id 1} must be passed a character vector.
    Code
      sql_glue2(con, "{.tbl 1}")
    Condition
      Error in `sql_glue2()`:
      ! Failed to interpolate {.tbl 1}.
      Caused by error in `glue_transformer()`:
      ! `x` must be a table source (SQL or a table identifier), not the number 1.
    Code
      sql_glue2(con, "{.sql 1}")
    Condition
      Error in `sql_glue2()`:
      ! {.sql 1} must be passed a string.

# gives informative errors

    Code
      sql_glue2(con, "{y*}")
    Condition
      Error in `sql_glue2()`:
      ! Failed to interpolate {y*}.
      Caused by error:
      ! object 'y' not found
    Code
      sql_glue2(con, "{1 + }")
    Condition
      Error in `sql_glue2()`:
      ! Failed to interpolate {1 + }.
      Caused by error in `parse()`:
      ! <text>:2:0: unexpected end of input
      1: 1 + 
         ^
    Code
      sql_glue2(con, "{.bar x}")
    Condition
      Error in `sql_glue2()`:
      ! Unknown marker "bar" in {.bar x}.

# join_by checks inputs

    Code
      new_join_by("x", c("x", "y"))
    Condition
      Error in `new_join_by()`:
      ! `x` and `y` must have the same length.
      i This is an internal error that was detected in the dbplyr package.
        Please report it at <https://github.com/tidyverse/dbplyr/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.
    Code
      new_join_by("x", "y", c("<", ">"))
    Condition
      Error in `new_join_by()`:
      ! Can't recycle `condition` (size 2) to size 1.

# sql_quote() checks its inputs

    Code
      sql_quote(1)
    Condition
      Error in `sql_quote()`:
      ! `x` must be a character vector, not the number 1.
    Code
      sql_quote("x", 1)
    Condition
      Error in `sql_quote()`:
      ! `quote` must be a character vector, not the number 1.
    Code
      sql_quote("x", character())
    Condition
      Error in `sql_quote()`:
      ! `quote` must be length 1 or 2.

# bind_queries combines multiple lazy tables

    Code
      bind_queries(lf1, lf2, lf3)
    Output
      <SQL>
      SELECT *
      FROM "lf1"
      
      UNION ALL
      
      SELECT *
      FROM "lf2"
      
      UNION ALL
      
      SELECT *
      FROM "lf3"

# bind_queries works with splice operator

    Code
      bind_queries(!!!queries)
    Output
      <SQL>
      SELECT *
      FROM "lf1"
      
      UNION ALL
      
      SELECT *
      FROM "lf2"

# bind_queries checks its inputs

    Code
      bind_queries()
    Condition
      Error in `bind_queries()`:
      ! `bind_queries()` requires at least one input.
    Code
      bind_queries(lf, 1)
    Condition
      Error in `bind_queries()`:
      ! `..2` must be a lazy query, not the number 1.

# copy_inline uses UNION ALL

    Code
      remote_query(copy_inline(con, slice(y, 0)))
    Output
      <SQL> SELECT CAST(NULL AS INTEGER) AS "id", CAST(NULL AS TEXT) AS "arr"
      WHERE (0 = 1)
    Code
      remote_query(copy_inline(con, y))
    Output
      <SQL> SELECT CAST("id" AS INTEGER) AS "id", CAST("arr" AS TEXT) AS "arr"
      FROM (
        SELECT NULL AS "id", NULL AS "arr"
        WHERE (0 = 1)
      
        UNION ALL
      
        SELECT 1, '{1,2,3}'
      ) AS "values_table"
    Code
      remote_query(copy_inline(con, slice(y, 0), types = types))
    Output
      <SQL> SELECT CAST(NULL AS bigint) AS "id", CAST(NULL AS integer[]) AS "arr"
      WHERE (0 = 1)
    Code
      remote_query(copy_inline(con, y, types = types))
    Output
      <SQL> SELECT CAST("id" AS bigint) AS "id", CAST("arr" AS integer[]) AS "arr"
      FROM (
        SELECT NULL AS "id", NULL AS "arr"
        WHERE (0 = 1)
      
        UNION ALL
      
        SELECT 1, '{1,2,3}'
      ) AS "values_table"

# custom clock functions translated correctly

    Code
      translate_sql(date_count_between(date_column_1, date_column_2, "year"), con = con)
    Condition
      Error in `date_count_between()`:
      ! `precision = "year"` isn't supported on database backends.
      i It must be "day" instead.

---

    Code
      translate_sql(date_count_between(date_column_1, date_column_2, "day", n = 5),
      con = con)
    Condition
      Error in `date_count_between()`:
      ! `n = 5` isn't supported on database backends.
      i It must be 1 instead.

# difftime is translated correctly

    Code
      translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
    Condition
      Error in `difftime()`:
      ! `units = "auto"` isn't supported on database backends.
      i It must be "days" instead.

---

    Code
      translate_sql(difftime(start_date, end_date, tz = "UTC", units = "days"), con = con)
    Condition
      Error in `difftime()`:
      ! Argument `tz` isn't supported on database backends.

# quantile and median don't change without warning

    Code
      translate_sql(quantile(x, 0.75, na.rm = TRUE), window = FALSE)
    Output
      <SQL> PERCENTILE_CONT(0.75) WITHIN GROUP (ORDER BY `x`)

---

    Code
      translate_sql(quantile(x, 0.75, na.rm = TRUE), vars_group = "g")
    Output
      <SQL> PERCENTILE_CONT(0.75) WITHIN GROUP (ORDER BY `x`) OVER (PARTITION BY `g`)

---

    Code
      translate_sql(median(x, na.rm = TRUE), window = FALSE)
    Output
      <SQL> PERCENTILE_CONT(0.5) WITHIN GROUP (ORDER BY `x`)

---

    Code
      translate_sql(median(x, na.rm = TRUE), vars_group = "g")
    Output
      <SQL> PERCENTILE_CONT(0.5) WITHIN GROUP (ORDER BY `x`) OVER (PARTITION BY `g`)

# generated sql doesn't change unexpectedly

    Code
      union(lf, lf)
    Output
      <SQL>
      SELECT *
      FROM "df"
      
      UNION
      
      SELECT *
      FROM "df"

---

    Code
      union_all(lf, lf)
    Output
      <SQL>
      SELECT *
      FROM "df"
      
      UNION ALL
      
      SELECT *
      FROM "df"

# has nice print method

    Code
      mf
    Output
      # A query:  ?? x 2
      # Database: sqlite <version> [:memory:]
            x     y
        <dbl> <dbl>
      1     1     1

---

    Code
      out2
    Output
      # A query:    ?? x 3
      # Database:   sqlite <version> [:memory:]
      # Groups:     x, y
      # Ordered by: x
            x     y     z
        <dbl> <dbl> <dbl>
      1     1     1     2

# useful error if missing I()

    Code
      tbl(src_memdb(), "foo.bar")
    Condition
      Error in `tbl()`:
      ! Failed to find table `foo.bar`.
      i Did you mean `from = I("foo.bar")`?
      Caused by error in `dbplyr_query_fields()`:
      ! Can't query fields.
      i Using SQL: SELECT * FROM `foo.bar` AS `q05` WHERE (0 = 1)
      Caused by error:
      ! no such table: foo.bar

# check_from is deprecated

    Code
      tbl_sql("foo", src_dbi(con), "x", check_from = FALSE)
    Condition
      Error:
      ! The `check_from` argument of `tbl_sql()` was deprecated in dbplyr 2.5.0 and is now defunct.

# build_sql() requires connection

    Code
      build_sql("SELECT * FROM ", x)
    Condition
      Error in `build_sql()`:
      ! `con` must not be NULL.

# glue_sql() checks size

    Code
      glue_sql2("{.col x}", .con = con)
    Condition
      Error in `.transformer()`:
      ! `value` must have size 1, not 2.
    Code
      glue_sql2("{.col character()}", .con = con)
    Condition
      Error in `.transformer()`:
      ! `value` must have size 1, not 0.

# glue_sql() can collapse

    Code
      glue_sql2("{.tbl x*}", .con = con)
    Condition
      Error in `glue_check_collapse()`:
      ! Collapsing is only allowed for "col" and "val", not for "tbl".
    Code
      glue_sql2("{.name x*}", .con = con)
    Condition
      Error in `glue_check_collapse()`:
      ! Collapsing is only allowed for "col" and "val", not for "name".
    Code
      glue_sql2("{.from x*}", .con = con)
    Condition
      Error in `glue_check_collapse()`:
      ! Collapsing is only allowed for "col" and "val", not for "from".

# can format ident

    Code
      ident()
    Output
      <IDENT> [empty]

# as_ident_or_sql() gives informative error for non-character

    Code
      as_ident_or_sql(1)
    Condition
      Error:
      ! Invalid identifier: expecting a character vector.

# sql_clause_select generates expected SQL

    Code
      sql_clause_select(sql("x", "y"))
    Output
      <sql clause> SELECT x, y

---

    Code
      sql_clause_select(sql("x"), distinct = TRUE)
    Output
      <sql clause> SELECT DISTINCT x

---

    Code
      sql_clause_select(sql("x"), top = 10)
    Output
      <sql clause> SELECT TOP 10 x

---

    Code
      sql_clause_select(sql("x"), distinct = TRUE, top = 5)
    Output
      <sql clause> SELECT DISTINCT TOP 5 x

# clauses wrap long inputs nicely

    Code
      sql_clause_select(sql(vars))
    Output
      <sql clause> SELECT
        variable1,
        variable2,
        variable3,
        variable4,
        variable5,
        variable6,
        variable7,
        variable8

# sql_clause_select requires non-empty input

    Code
      sql_clause_select(sql())
    Condition
      Error in `sql_clause_select()`:
      ! Query contains no columns

# sql_clause_where wraps in parens and combines conditions with AND

    Code
      sql_clause_where(sql("x > 1"))
    Output
      <sql clause> WHERE (x > 1)

---

    Code
      sql_clause_where(sql("x > 1", "y < 2"))
    Output
      <sql clause> WHERE (x > 1) AND (y < 2)

# sql_clause_order_by warns when dropping in subquery

    Code
      sql_clause_order_by(sql("x"), subquery = TRUE)
    Condition
      Warning:
      ORDER BY is ignored in subqueries without LIMIT
      i Do you need to move arrange() later in the pipeline or use window_order() instead?
    Output
      NULL

# sql_select_clauses generates expected SQL

    Code
      sql_select_clauses(select = sql_clause_select(sql("x", "y")), from = sql_clause_from(
        sql("foo")), where = sql_clause_where(sql("x > 1")), group_by = sql_clause_group_by(
        sql("x")), having = sql_clause_having(sql("COUNT(*) > 1")), window = sql_clause_window(
        sql("win1 AS (PARTITION BY x)")), order_by = sql_clause_order_by(sql("x")),
      limit = sql_clause_limit(NULL, 10))
    Output
      <SQL> SELECT x, y
      FROM foo
      WHERE (x > 1)
      GROUP BY x
      HAVING (COUNT(*) > 1)
      WINDOW win1 AS (PARTITION BY x)
      ORDER BY x
      LIMIT 10

# update clauses generate expected SQL

    Code
      sql_clause_update(sql("foo"))
    Output
      <sql clause> UPDATE foo

---

    Code
      sql_clause_set(sql("x", "y"), sql("1", "2"))
    Output
      <sql clause> SET x = 1, y = 2

# insert clauses generate expected SQL

    Code
      sql_clause_insert(sql("x", "y"))
    Output
      <sql clause> INSERT (x, y)

---

    Code
      sql_clause_insert(sql("x", "y"), into = sql("foo"))
    Output
      <sql clause> INSERT INTO foo (x, y)

# sql_clause_where_exists generates expected SQL

    Code
      sql_format_clauses(clause_true)
    Output
      <SQL> WHERE EXISTS (
        SELECT 1 FROM bar
        WHERE (x = y)
      )
    Code
      sql_format_clauses(clause_false)
    Output
      <SQL> WHERE NOT EXISTS (
        SELECT 1 FROM bar
        WHERE (x = y)
      )

# quoting for rendering ordered grouped table

    Code
      sql_render(out)
    Output
      <SQL> SELECT *
      FROM `test-verb-arrange`
      ORDER BY `y`

# arrange renders correctly (#373)

    Code
      # # arrange renders correctly
      lf <- lazy_frame(a = 1:3, b = 3:1)
      # basic
      arrange(lf, a)
    Output
      <SQL>
      SELECT *
      FROM "df"
      ORDER BY "a"
    Code
      # double arrange
      arrange(arrange(lf, a), b)
    Output
      <SQL>
      SELECT *
      FROM "df"
      ORDER BY "b", "a"
    Code
      # remove ordered by
      select(arrange(lf, a), -a)
    Output
      <SQL>
      SELECT "b"
      FROM "df"
      ORDER BY "a"
    Code
      arrange(select(arrange(lf, a), -a), b)
    Output
      <SQL>
      SELECT "b"
      FROM "df"
      ORDER BY "b"
    Code
      # un-arrange
      arrange(arrange(lf, a))
    Output
      <SQL>
      SELECT *
      FROM "df"
      ORDER BY "a"
    Code
      arrange(select(arrange(lf, a), -a))
    Output
      <SQL>
      SELECT "b"
      FROM "df"
      ORDER BY "a"
    Code
      # use order
      mutate(select(arrange(lf, a), -a), c = lag(b))
    Output
      <SQL>
      SELECT "b", LAG("b", 1, NULL) OVER () AS "c"
      FROM "df"
      ORDER BY "a"

# arrange renders correctly for single-table verbs (#373)

    Code
      lf <- lazy_frame(a = 1:3, b = 3:1)
      # head
      arrange(head(lf, 1), a)
    Output
      <SQL>
      SELECT *
      FROM (
        SELECT *
        FROM "df"
        LIMIT 1
      ) AS "q01"
      ORDER BY "a"
    Code
      head(arrange(lf, a), 1)
    Output
      <SQL>
      SELECT *
      FROM "df"
      ORDER BY "a"
      LIMIT 1
    Code
      arrange(head(arrange(lf, a), 1), b)
    Output
      <SQL>
      SELECT *
      FROM (
        SELECT *
        FROM "df"
        ORDER BY "a"
        LIMIT 1
      ) AS "q01"
      ORDER BY "b", "a"
    Code
      # mutate
      arrange(mutate(lf, a = b), a)
    Output
      <SQL>
      SELECT "b" AS "a", "b"
      FROM "df"
      ORDER BY "a"
    Code
      # complex mutate
      arrange(mutate(arrange(lf, a), a = b), a)
    Output
      <SQL>
      SELECT "b" AS "a", "b"
      FROM "df"
      ORDER BY "a"
    Code
      arrange(mutate(arrange(lf, a), a = 1), b)
    Output
      <SQL>
      SELECT 1.0 AS "a", "b"
      FROM "df"
      ORDER BY "b", "a"
    Code
      mutate(arrange(mutate(lf, a = -a), a), a = -a)
    Condition
      Warning:
      ORDER BY is ignored in subqueries without LIMIT
      i Do you need to move arrange() later in the pipeline or use window_order() instead?
    Output
      <SQL>
      SELECT -"a" AS "a", "b"
      FROM (
        SELECT -"a" AS "a", "b"
        FROM "df"
      ) AS "q01"

# can combine arrange with dual table verbs

    Code
      lf <- lazy_frame(a = 1:3, b = 3:1)
      rf <- lazy_frame(a = 1:3, c = 4:6)
      # warn if arrange before join
      left_join(arrange(lf, a), rf)
    Message
      Joining with `by = join_by(a)`
    Condition
      Warning:
      ORDER BY is ignored in subqueries without LIMIT
      i Do you need to move arrange() later in the pipeline or use window_order() instead?
    Output
      <SQL>
      SELECT "LHS".*, "c"
      FROM (
        SELECT *
        FROM "df"
      ) AS "LHS"
      LEFT JOIN "df"
        ON ("LHS"."a" = "df"."a")
    Code
      semi_join(arrange(lf, a), rf)
    Message
      Joining with `by = join_by(a)`
    Condition
      Warning:
      ORDER BY is ignored in subqueries without LIMIT
      i Do you need to move arrange() later in the pipeline or use window_order() instead?
    Output
      <SQL>
      SELECT "LHS".*
      FROM (
        SELECT *
        FROM "df"
      ) AS "LHS"
      WHERE EXISTS (
        SELECT 1 FROM "df"
        WHERE ("LHS"."a" = "df"."a")
      )
    Code
      union(arrange(lf, a), rf)
    Output
      <SQL>
      SELECT *, NULL AS "c"
      FROM "df"
      ORDER BY "a"
      
      UNION
      
      SELECT "a", NULL AS "b", "c"
      FROM "df"
    Code
      # can arrange after join
      arrange(left_join(lf, rf), a)
    Message
      Joining with `by = join_by(a)`
    Output
      <SQL>
      SELECT *
      FROM (
        SELECT "df_LHS".*, "c"
        FROM "df" AS "df_LHS"
        LEFT JOIN "df" AS "df_RHS"
          ON ("df_LHS"."a" = "df_RHS"."a")
      ) AS "q01"
      ORDER BY "a"
    Code
      arrange(semi_join(lf, rf), a)
    Message
      Joining with `by = join_by(a)`
    Output
      <SQL>
      SELECT *
      FROM (
        SELECT "df_LHS".*
        FROM "df" AS "df_LHS"
        WHERE EXISTS (
          SELECT 1 FROM "df" AS "df_RHS"
          WHERE ("df_LHS"."a" = "df_RHS"."a")
        )
      ) AS "q01"
      ORDER BY "a"
    Code
      arrange(union(lf, rf), a)
    Output
      <SQL>
      SELECT *
      FROM (
        SELECT *, NULL AS "c"
        FROM "df"
      
        UNION
      
        SELECT "a", NULL AS "b", "c"
        FROM "df"
      ) AS "q01"
      ORDER BY "a"

# unwrap_order_expr informatively errors on c()

    Code
      unwrap_order_expr(expr(c(x, y)), "slice_min")
    Condition
      Error:
      ! Can't use `c()` in `slice_min()`
      i Did you mean to use `tibble(x, y)` instead?

# across() captures anonymous functions

    Code
      (expect_error(capture_across(lf, across(a, function(x) {
        x <- x + 2
        log(x)
      }))))
    Output
      <error/rlang_error>
      Error in `across()`:
      ! Cannot translate functions consisting of more than one statement.

# across() gives informative errors

    Code
      capture_across(lf, across(a, 1))
    Condition
      Error in `across()`:
      ! `.fns` must be a function, a formula, or list of functions/formulas.
    Code
      capture_across(lf, across(a, list(1)))
    Condition
      Error in `across()`:
      ! `.fns` must contain a function or a formula.
      x Problem with 1
    Code
      capture_across(lf, across(a:b, "log"))
    Condition
      Error in `across()`:
      ! `.fns` must be a function, a formula, or list of functions/formulas.
    Code
      capture_across(lf, across(c, mean))
    Condition
      Error in `across()`:
      ! Can't subset columns that don't exist.
      x Column `c` doesn't exist.

# across() defaults to everything()

    Code
      lazy_frame(x = 1, y = 1) %>% summarise(across(.fns = ~ . + 1))
    Output
      <SQL>
      SELECT `x` + 1.0 AS `x`, `y` + 1.0 AS `y`
      FROM `df`

# untranslatable functions are preserved

    Code
      lf %>% summarise(across(a:b, SQL_LOG))
    Output
      <SQL>
      SELECT SQL_LOG(`a`) AS `a`, SQL_LOG(`b`) AS `b`
      FROM `df`

# old _at functions continue to work

    Code
      lf %>% dplyr::summarise_at(dplyr::vars(a:b), "sum")
    Condition
      Warning:
      Missing values are always removed in SQL aggregation functions.
      Use `na.rm = TRUE` to silence this warning
      This warning is displayed once every 8 hours.
    Output
      <SQL>
      SELECT SUM(`a`) AS `a`, SUM(`b`) AS `b`
      FROM `df`

---

    Code
      lf %>% dplyr::summarise_at(dplyr::vars(a:b), sum)
    Output
      <SQL>
      SELECT SUM(`a`) AS `a`, SUM(`b`) AS `b`
      FROM `df`

---

    Code
      lf %>% dplyr::summarise_at(dplyr::vars(a:b), ~ sum(.))
    Output
      <SQL>
      SELECT SUM(`a`) AS `a`, SUM(`b`) AS `b`
      FROM `df`

# across() errors if named

    Code
      (expect_error(mutate(lf, x = across())))
    Output
      <error/rlang_error>
      Error in `mutate()`:
      ! In dbplyr, the result of `across()` must be unnamed.
      i `x = across()` is named.
    Code
      (expect_error(group_by(lf, x = across())))
    Output
      <error/rlang_error>
      Error in `group_by()`:
      ! In dbplyr, the result of `across()` must be unnamed.
      i `x = across()` is named.

# across() throws error if unpack = TRUE

    Code
      (expect_error(lf %>% mutate(across(x, .unpack = TRUE))))
    Output
      <error/rlang_error>
      Error in `mutate()`:
      i In argument: `across(x, .unpack = TRUE)`
      Caused by error in `mutate()`:
      ! `.unpack = TRUE` is not supported in SQL translations.

# if_all() gives informative errors

    Code
      capture_if_all(lf, if_all(a, 1))
    Condition
      Error in `if_all()`:
      ! `.fns` must be a function, a formula, or list of functions/formulas.
    Code
      capture_if_all(lf, if_all(a, list(1)))
    Condition
      Error in `if_all()`:
      ! `.fns` must contain a function or a formula.
      x Problem with 1

# if_all/any works in filter()

    Code
      lf %>% filter(if_all(a:b, ~ . > 0))
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`a` > 0.0 AND `b` > 0.0)

---

    Code
      lf %>% filter(if_any(a:b, ~ . > 0))
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`a` > 0.0 OR `b` > 0.0)

# if_all/any works in mutate()

    Code
      lf %>% mutate(c = if_all(a:b, ~ . > 0))
    Output
      <SQL>
      SELECT *, `a` > 0.0 AND `b` > 0.0 AS `c`
      FROM `df`

---

    Code
      lf %>% mutate(c = if_any(a:b, ~ . > 0))
    Output
      <SQL>
      SELECT *, `a` > 0.0 OR `b` > 0.0 AS `c`
      FROM `df`

# if_all/any uses every colum as default

    Code
      lf %>% filter(if_all(.fns = ~ . > 0))
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`a` > 0.0 AND `b` > 0.0)

---

    Code
      lf %>% filter(if_any(.fns = ~ . > 0))
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`a` > 0.0 OR `b` > 0.0)

# if_all/any works without `.fns` argument

    Code
      lf %>% filter(if_all(a:b))
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`a` AND `b`)

---

    Code
      lf %>% filter(if_any(a:b))
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`a` OR `b`)

# if_all() cannot rename variables

    Code
      (expect_error(capture_if_all(lf, if_all(c(a = x, b = y)))))
    Output
      <error/tidyselect:::error_disallowed_rename>
      Error in `if_all()`:
      ! Can't rename variables in this context.

# across(...) is deprecated

    Code
      summarise(lf, across(everything(), mean, na.rm = TRUE))
    Condition
      Warning:
      The `...` argument of `across()` is deprecated as of dbplyr 2.3.0.
      i Supply arguments directly to `.fns` through a lambda instead.
      
      # Previously across(a:b, mean, na.rm = TRUE)
      
      # Now across(a:b, ~mean(.x, na.rm = TRUE))
    Output
      <SQL>
      SELECT AVG(`x`) AS `x`
      FROM `df`

# across() does not support formulas with dots

    Code
      (expect_error(capture_across(lf, across(a:b, ~ log(.x, base = .y), base = 2))))
    Output
      <error/rlang_error>
      Error in `across()`:
      ! Can't use `...` when a purrr-style lambda is used in `.fns`.
      i Use a lambda instead.
      i Or inline them via a purrr-style lambda.
    Code
      (expect_error(capture_across(lf, across(a:b, list(~ log(.x, base = .y)), base = 2)))
      )
    Output
      <error/rlang_error>
      Error in `across()`:
      ! Can't use `...` when a purrr-style lambda is used in `.fns`.
      i Use a lambda instead.
      i Or inline them via a purrr-style lambda.

# `pick()` errors in `arrange()` are useful

    Code
      arrange(df, pick(y))
    Condition
      Error in `arrange()`:
      i In argument: `pick(y)`
      Caused by error in `pick()`:
      ! Can't subset columns that don't exist.
      x Column `y` doesn't exist.

# doesn't allow renaming

    Code
      arrange(lazy_frame(x = 1), pick(y = x))
    Condition
      Error in `arrange()`:
      i In argument: `pick(y = x)`
      Caused by error in `pick()`:
      ! Can't rename variables in this context.

# requires at least one input

    Code
      arrange(lazy_frame(x = 1), pick())
    Condition
      Error in `arrange()`:
      i In argument: `pick()`
      Caused by error in `partial_eval_pick()`:
      ! Must supply at least one input to `pick()`.

# `filter()` with `pick()` that uses invalid tidy-selection errors

    Code
      filter(df, pick(x, a))
    Condition
      Error in `filter()`:
      i In argument: `pick(x, a)`
      Caused by error in `pick()`:
      ! Can't subset columns that don't exist.
      x Column `a` doesn't exist.

# custom string translations

    Code
      translate_sql(paste0("a", "b"), con = con)
    Output
      <SQL> 'a' || 'b'

---

    Code
      translate_sql(paste("a", "b"), con = con)
    Output
      <SQL> 'a' || ' ' || 'b'

---

    Code
      translate_sql(substr(x, 2, 4), con = con)
    Output
      <SQL> SUBSTRING("x", 2, 3)

---

    Code
      translate_sql(substring(x, 2, 4), con = con)
    Output
      <SQL> SUBSTRING("x", 2, 3)

---

    Code
      translate_sql(str_sub(x, 2, -2), con = con)
    Output
      <SQL> SUBSTRING("x", 2, LENGTH("x") - 2)

# copy_inline uses UNION ALL

    Code
      remote_query(copy_inline(con, slice(y, 0)))
    Output
      <SQL> SELECT CAST(NULL AS INTEGER) AS "id", CAST(NULL AS VARCHAR) AS "arr"
      FROM "DUMMY"
      WHERE (0 = 1)
    Code
      remote_query(copy_inline(con, y))
    Output
      <SQL> SELECT CAST("id" AS INTEGER) AS "id", CAST("arr" AS VARCHAR) AS "arr"
      FROM (
        SELECT NULL AS "id", NULL AS "arr"
        FROM "DUMMY"
        WHERE (0 = 1)
      
        UNION ALL
      
        SELECT 1, '{1,2,3}' FROM DUMMY
      ) AS "values_table"
    Code
      remote_query(copy_inline(con, slice(y, 0), types = types))
    Output
      <SQL> SELECT CAST(NULL AS bigint) AS "id", CAST(NULL AS integer[]) AS "arr"
      FROM "DUMMY"
      WHERE (0 = 1)
    Code
      remote_query(copy_inline(con, y, types = types))
    Output
      <SQL> SELECT CAST("id" AS bigint) AS "id", CAST("arr" AS integer[]) AS "arr"
      FROM (
        SELECT NULL AS "id", NULL AS "arr"
        FROM "DUMMY"
        WHERE (0 = 1)
      
        UNION ALL
      
        SELECT 1, '{1,2,3}' FROM DUMMY
      ) AS "values_table"

# can construct and print

    Code
      in_schema("schema", "table")
    Output
      <SCHEMA> `schema`.`table`

---

    Code
      in_catalog("catalog", "schema", "table")
    Output
      <CATALOG> `catalog`.`schema`.`table`

# rows_insert() checks arguments

    Code
      (rows_insert(lf, lf, by = "x", conflict = "error"))
    Condition
      Error in `rows_insert()`:
      ! `conflict = "error"` isn't supported on database backends.
      i It must be "ignore" instead.
    Code
      (rows_insert(lf, lf, by = "x"))
    Condition
      Error in `rows_insert()`:
      ! `conflict = "error"` isn't supported on database backends.
      i It must be "ignore" instead.

---

    Code
      (rows_insert(lf, lazy_frame(x = 1, y = 2, z = 3), by = "x", conflict = "ignore")
      )
    Condition
      Error in `rows_insert()`:
      ! All columns in `y` must exist in `x`.
      i The following columns only exist in `y`: `z`.

---

    Code
      (rows_insert(lf, lf, by = "x", conflict = "ignore", returning = everything()))
    Condition
      Error in `rows_insert()`:
      ! `returning` does not work for simulated connections.

---

    Code
      (rows_insert(mutate(df, x = x + 1), df, by = "x", conflict = "ignore",
      in_place = TRUE))
    Condition
      Error in `target_table()`:
      ! Can't determine name for target table. Set `in_place = FALSE` to return a lazy table.

---

    Code
      (rows_insert(df, df, by = "x", conflict = "ignore", returning = c(y)))
    Condition
      Error in `rows_insert()`:
      ! Can't select columns that don't exist.
      x Column `y` doesn't exist.

# `rows_insert()` errors for `conflict = 'error'` and `in_place = FALSE`

    Code
      (expect_error(rows_insert(lf, lf, by = "x", conflict = "error", in_place = FALSE))
      )
    Output
      <error/dbplyr_error_unsupported_arg>
      Error in `rows_insert()`:
      ! `conflict = "error"` isn't supported on database backends.
      i It must be "ignore" instead.

# `rows_insert()` works with `in_place = FALSE`

    Code
      rows_insert(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 3:4,
      y = 23:24, .name = "df_y"), by = "x", conflict = "ignore", in_place = FALSE)
    Output
      <SQL>
      SELECT *
      FROM "df_x"
      
      UNION ALL
      
      SELECT "df_y".*
      FROM "df_y"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df_x"
        WHERE ("df_y"."x" = "df_x"."x")
      )

# `rows_insert()` works with `in_place = TRUE`

    Code
      (rows_insert(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 2:3,
      y = 22:23, .name = "df_y"), by = "x", in_place = TRUE))
    Condition
      Error in `rows_insert()`:
      ! `in_place = TRUE` does not work for simulated connections.

# rows_get_or_execute() gives error context

    Code
      (expect_error(rows_append(tbl(con, "mtcars"), tibble(x = 1), copy = TRUE,
      in_place = TRUE)))
    Output
      <error/rlang_error>
      Error in `rows_append()`:
      ! Can't modify database table "mtcars".
      i Using SQL: INSERT INTO `mtcars` (`x`) SELECT * FROM ( SELECT * FROM `dbplyr_{tmp}` ) AS `...y`
      Caused by error:
      ! dummy DBI error
    Code
      (expect_error(rows_append(tbl(con, "mtcars"), tibble(x = 1), copy = TRUE,
      in_place = TRUE, returning = x)))
    Output
      <error/rlang_error>
      Error in `rows_append()`:
      ! Can't modify database table "mtcars".
      i Using SQL: INSERT INTO `mtcars` (`x`) SELECT * FROM ( SELECT * FROM `dbplyr_{tmp}` ) AS `...y` RETURNING `mtcars`.`x`
      Caused by error:
      ! dummy DBI error

# `sql_query_insert()` works

    Code
      (sql_query_insert(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), insert_cols = colnames(df_y), by = c("a", "b"), conflict = "error",
      returning_cols = c("a", b2 = "b")))
    Condition
      Error in `sql_query_insert()`:
      ! `conflict = "error"` isn't supported on database backends.
      i It must be "ignore" instead.

---

    Code
      sql_query_insert(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), insert_cols = colnames(df_y), by = c("a", "b"), conflict = "ignore",
      returning_cols = c("a", b2 = "b"))
    Output
      <SQL> INSERT INTO "df_x" ("a", "b", "c", "d")
      SELECT *
      FROM (
        SELECT "a", "b", "c" + 1.0 AS "c", "d"
        FROM "df_y"
      ) AS "...y"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df_x"
        WHERE ("df_x"."a" = "...y"."a") AND ("df_x"."b" = "...y"."b")
      )
      RETURNING "df_x"."a", "df_x"."b" AS "b2"

# rows_append() checks arguments

    Code
      (rows_append(lf, df, by = "x"))
    Condition
      Error in `rows_append()`:
      ! `...` must be empty.
      x Problematic argument:
      * by = "x"
    Code
      (rows_append(lf, df, conflict = "error"))
    Condition
      Error in `rows_append()`:
      ! `...` must be empty.
      x Problematic argument:
      * conflict = "error"

# `rows_append()` works with `in_place = FALSE`

    Code
      rows_append(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 3:4,
      y = 23:24, .name = "df_y"), in_place = FALSE)
    Output
      <SQL>
      SELECT *
      FROM "df_x"
      
      UNION ALL
      
      SELECT *
      FROM "df_y"

# `rows_append()` works with `in_place = TRUE`

    Code
      (rows_append(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 2:3,
      y = 22:23, .name = "df_y"), in_place = TRUE))
    Condition
      Error in `rows_append()`:
      ! `in_place = TRUE` does not work for simulated connections.

# `sql_query_append()` works

    Code
      sql_query_append(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), insert_cols = colnames(df_y), returning_cols = c("a", b2 = "b"))
    Output
      <SQL> INSERT INTO "df_x" ("a", "b", "c", "d")
      SELECT *
      FROM (
        SELECT "a", "b", "c" + 1.0 AS "c", "d"
        FROM "df_y"
      ) AS "...y"
      RETURNING "df_x"."a", "df_x"."b" AS "b2"

# sql_query_append old interface is defunct

    Code
      sql_query_append(con = con, table = ident("df_x"), from = df_y, returning_cols = c(
        "a", b2 = "b"))
    Condition
      Error:
      ! The `from` argument of `sql_query_append()` must be a table identifier or an SQL query, not a lazy table. as of dbplyr 2.3.2.

# arguments are checked

    Code
      (rows_update(lf, lf, by = 1, unmatched = "ignore"))
    Condition
      Error in `rows_update()`:
      ! `by` must be a character vector.

---

    Code
      (rows_update(lf, lf, by = c(y = "x"), unmatched = "ignore"))
    Condition
      Error in `rows_update()`:
      ! `by` must be unnamed.

---

    Code
      (rows_update(lf, lf, by = "z", unmatched = "ignore"))
    Condition
      Error in `rows_update()`:
      ! All columns specified through `by` must exist in `x` and `y`.
      i The following columns are missing from `x`: `z`.

---

    Code
      (rows_update(lf, lf, by = "x", unmatched = "error"))
    Condition
      Error in `rows_update()`:
      ! `unmatched = "error"` isn't supported on database backends.
      i It must be "ignore" instead.
    Code
      (rows_update(lf, lf, by = "x"))
    Condition
      Error in `rows_update()`:
      ! `unmatched = "error"` isn't supported on database backends.
      i It must be "ignore" instead.

---

    Code
      (rows_update(lf, lazy_frame(x = 1, y = 2, z = 3), by = "x", unmatched = "ignore")
      )
    Condition
      Error in `rows_update()`:
      ! All columns in `y` must exist in `x`.
      i The following columns only exist in `y`: `z`.

---

    Code
      (rows_update(lf, lf, by = "x", unmatched = "ignore", returning = everything()))
    Condition
      Error in `rows_update()`:
      ! `returning` does not work for simulated connections.

---

    Code
      (rows_update(mutate(df, x = x + 1), df, by = "x", unmatched = "ignore",
      in_place = TRUE))
    Condition
      Error in `target_table()`:
      ! Can't determine name for target table. Set `in_place = FALSE` to return a lazy table.

# `rows_update()` works with `in_place = FALSE`

    Code
      rows_update(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 2:3,
      y = 22:23, .name = "df_y"), by = "x", unmatched = "ignore", in_place = FALSE)
    Output
      <SQL>
      SELECT "df_x".*
      FROM "df_x"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df_y"
        WHERE ("df_x"."x" = "df_y"."x")
      )
      
      UNION ALL
      
      SELECT "df_x"."x" AS "x", "df_y"."y" AS "y"
      FROM "df_x"
      INNER JOIN "df_y"
        ON ("df_x"."x" = "df_y"."x")

# `rows_update()` works with `in_place = TRUE`

    Code
      (rows_update(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 2:3,
      y = 22:23, .name = "df_y"), by = "x", unmatched = "ignore", in_place = TRUE))
    Condition
      Error in `rows_update()`:
      ! `in_place = TRUE` does not work for simulated connections.

# `sql_query_update_from()` works

    Code
      sql_query_update_from(con = con, table = ident("df_x"), from = sql_render(df_y,
        con, lvl = 1), by = c("a", "b"), update_values = sql(c = "COALESCE(`df_x`.`c`, `...y`.`c`)",
        d = "`...y`.`d`"), returning_cols = c("a", b2 = "b"))
    Output
      <SQL> UPDATE "df_x"
      SET "c" = COALESCE(`df_x`.`c`, `...y`.`c`), "d" = `...y`.`d`
      FROM (
        SELECT "a", "b", "c" + 1.0 AS "c", "d"
        FROM "df_y"
      ) AS "...y"
      WHERE ("...y"."a" = "df_x"."a") AND ("...y"."b" = "df_x"."b")
      RETURNING "df_x"."a", "df_x"."b" AS "b2"

# `rows_patch()` returns early if no column to update

    Code
      rows_patch(lazy_frame(x = 1:3, y = c(11, 12, NA), .name = "df_x"), lazy_frame(
        x = 1:3, .name = "df_y"), by = "x", unmatched = "ignore", in_place = FALSE)
    Output
      <SQL>
      SELECT *
      FROM "df_x"

# `rows_patch()` works with `in_place = FALSE`

    Code
      rows_patch(lazy_frame(x = 1:3, y = c(11, 12, NA), .name = "df_x"), lazy_frame(
        x = 2:3, y = 22:23, .name = "df_y"), by = "x", unmatched = "ignore",
      in_place = FALSE)
    Output
      <SQL>
      SELECT "df_x".*
      FROM "df_x"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df_y"
        WHERE ("df_x"."x" = "df_y"."x")
      )
      
      UNION ALL
      
      SELECT "x", COALESCE("y", "y...y") AS "y"
      FROM (
        SELECT "df_x".*, "df_y"."y" AS "y...y"
        FROM "df_x"
        INNER JOIN "df_y"
          ON ("df_x"."x" = "df_y"."x")
      ) AS "q01"

# `rows_patch()` works with multiple columns to update

    Code
      rows_patch(lazy_frame(x = 1:3, y = c(11, 12, NA), z = c(31, NA, 33), .name = "df_x"),
      lazy_frame(x = 2:3, y = 22:23, z = 42:43, .name = "df_y"), by = "x", unmatched = "ignore",
      in_place = FALSE)
    Output
      <SQL>
      SELECT "df_x".*
      FROM "df_x"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df_y"
        WHERE ("df_x"."x" = "df_y"."x")
      )
      
      UNION ALL
      
      SELECT "x", COALESCE("y", "y...y") AS "y", COALESCE("z", "z...y") AS "z"
      FROM (
        SELECT "df_x".*, "df_y"."y" AS "y...y", "df_y"."z" AS "z...y"
        FROM "df_x"
        INNER JOIN "df_y"
          ON ("df_x"."x" = "df_y"."x")
      ) AS "q01"

# `rows_patch()` works with `in_place = TRUE`

    Code
      (rows_patch(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 2:3,
      y = 22:23, .name = "df_y"), by = "x", unmatched = "ignore", in_place = TRUE))
    Condition
      Error in `rows_patch()`:
      ! `in_place = TRUE` does not work for simulated connections.

# `rows_upsert()` returns early if no column to update

    Code
      rows_upsert(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 1:3,
      .name = "df_y"), by = "x", in_place = FALSE)
    Output
      <SQL>
      SELECT *
      FROM "df_x"
      
      UNION ALL
      
      SELECT *, NULL AS "y"
      FROM (
        SELECT "df_y".*
        FROM "df_y"
        WHERE NOT EXISTS (
          SELECT 1 FROM "df_x"
          WHERE ("df_y"."x" = "df_x"."x")
        )
      ) AS "q01"

# `rows_upsert()` works with `in_place = FALSE`

    Code
      rows_upsert(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 2:3,
      y = 22:23, .name = "df_y"), by = "x", in_place = FALSE)
    Output
      <SQL>
      SELECT "df_x".*
      FROM "df_x"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df_y"
        WHERE ("df_x"."x" = "df_y"."x")
      )
      
      UNION ALL
      
      SELECT "df_x"."x" AS "x", "df_y"."y" AS "y"
      FROM "df_x"
      INNER JOIN "df_y"
        ON ("df_x"."x" = "df_y"."x")
      
      UNION ALL
      
      SELECT "df_y".*
      FROM "df_y"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df_x"
        WHERE ("df_y"."x" = "df_x"."x")
      )

# `rows_upsert()` works with `in_place = TRUE`

    Code
      (rows_upsert(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 2:3,
      y = 22:23, .name = "df_y"), by = "x", in_place = TRUE))
    Condition
      Error in `rows_upsert()`:
      ! `in_place = TRUE` does not work for simulated connections.

# `sql_query_upsert()` is correct

    Code
      sql_query_upsert(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), by = c("a", "b"), update_cols = c("c", "d"), returning_cols = c("a",
        b2 = "b"))
    Output
      <SQL> WITH "updated" AS (
        UPDATE "df_x"
        SET "c" = "...y"."c", "d" = "...y"."d"
        FROM (
        SELECT "a", "b", "c" + 1.0 AS "c", "d"
        FROM "df_y"
      ) AS "...y"
        WHERE ("...y"."a" = "df_x"."a") AND ("...y"."b" = "df_x"."b")
        RETURNING "df_x".*
      )
      INSERT INTO "df_x" ("a", "b", "c", "d")
      SELECT *
      FROM (
        SELECT "a", "b", "c" + 1.0 AS "c", "d"
        FROM "df_y"
      ) AS "...y"
      WHERE NOT EXISTS (
        SELECT 1 FROM "updated"
        WHERE ("updated"."a" = "...y"."a") AND ("updated"."b" = "...y"."b")
      )
      RETURNING "df_x"."a", "df_x"."b" AS "b2"

# `rows_delete()` works with `in_place = FALSE`

    Code
      rows_delete(lazy_frame(x = 1:3, y = c(11, 12, NA), .name = "df_x"), lazy_frame(
        x = 2:3, .name = "df_y"), by = "x", unmatched = "ignore", in_place = FALSE)
    Output
      <SQL>
      SELECT "df_x".*
      FROM "df_x"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df_y"
        WHERE ("df_x"."x" = "df_y"."x")
      )

# `rows_delete()` works with `in_place = TRUE`

    Code
      (rows_delete(lazy_frame(x = 1:3, y = 11:13, .name = "df_x"), lazy_frame(x = 2:3,
      .name = "df_y"), by = "x", unmatched = "ignore", in_place = TRUE))
    Condition
      Error in `rows_delete()`:
      ! `in_place = TRUE` does not work for simulated connections.

# `sql_query_delete()` is correct

    Code
      sql_query_delete(con = dialect_ansi(), table = ident("df_x"), from = sql_render(
        df_y, dialect_ansi(), lvl = 2), by = c("a", "b"), returning_cols = c("a", b2 = "b"))
    Output
      <SQL> DELETE FROM "df_x"
      WHERE EXISTS (
        SELECT 1 FROM (
          SELECT "a", "b", "c" + 1.0 AS "c", "d"
          FROM "df_y"
        ) AS "...y"
        WHERE ("...y"."a" = "df_x"."a") AND ("...y"."b" = "df_x"."b")
      )
      RETURNING "df_x"."a", "df_x"."b" AS "b2"

# `get_returned_rows()` works

    Code
      (get_returned_rows(df))
    Condition
      Error in `get_returned_rows()`:
      ! No returned rows available.

# with_dialect() validates inputs

    Code
      with_dialect(NULL, dialect_postgres())
    Condition
      Error in `with_dialect()`:
      ! `con` must not be NULL.
    Code
      with_dialect(con, "postgres")
    Condition
      Error in `with_dialect()`:
      ! `dialect` must be a dialect object.

# SQL generation uses specified dialect

    Code
      show_query(summarise(mtcars, mpg = sd(mpg), .by = cyl))
    Output
      <SQL>
      SELECT "cyl", STDDEV_SAMP("mpg") AS "mpg"
      FROM "mtcars"
      GROUP BY "cyl"

# string functions translate correctly

    Code
      translate_sql(str_replace(col, "pattern", "replacement"), con = con)
    Output
      <SQL> REGEXP_REPLACE("col", 'pattern', 'replacement', 1, 1)
    Code
      translate_sql(str_replace_all(col, "pattern", "replacement"), con = con)
    Output
      <SQL> REGEXP_REPLACE("col", 'pattern', 'replacement')

# queries translate correctly

    Code
      head(mf)
    Output
      <SQL>
      SELECT "df".*
      FROM "df"
      FETCH FIRST 6 ROWS ONLY

# `sql_query_upsert()` is correct

    Code
      sql_query_upsert(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), by = c("a", "b"), update_cols = c("c", "d"), returning_cols = c("a",
        b2 = "b"), method = "merge")
    Output
      <SQL> MERGE INTO "df_x"
      USING (
        SELECT "a", "b", "c" + 1.0 AS "c", "d"
        FROM "df_y"
      ) "...y"
        ON ("...y"."a" = "df_x"."a" AND "...y"."b" = "df_x"."b")
      WHEN MATCHED THEN
        UPDATE SET "c" = "...y"."c", "d" = "...y"."d"
      WHEN NOT MATCHED THEN
        INSERT ("a", "b", "c", "d")
        VALUES ("...y"."a", "...y"."b", "...y"."c", "...y"."d")
      RETURNING "df_x"."a", "df_x"."b" AS "b2"
      ;

# db_table_temporary adds ORA$PTT_ prefix

    Code
      result <- sql_table_temporary(con, table_path("tbl"), temporary = TRUE)
    Message
      Created a temporary table named ORA$PTT_tbl

# generates custom sql

    Code
      sql_table_analyze(con, in_schema("schema", "tbl"))
    Output
      <SQL> ANALYZE TABLE "schema"."tbl" COMPUTE STATISTICS

---

    Code
      sql_query_explain(con, sql("SELECT * FROM foo"))
    Output
      <SQL> EXPLAIN PLAN FOR SELECT * FROM foo
      <SQL> SELECT PLAN_TABLE_OUTPUT FROM TABLE(DBMS_XPLAN.DISPLAY())

---

    Code
      left_join(lf, lf, by = "x", na_matches = "na")
    Output
      <SQL>
      SELECT "df_LHS"."x" AS "x"
      FROM "df" "df_LHS"
      LEFT JOIN "df" "df_RHS"
        ON ((decode("df_LHS"."x", "df_RHS"."x", 0, 1) = 0))

---

    Code
      sql_query_save(con, sql("SELECT * FROM foo"), "ORA$PTT_tbl")
    Output
      <SQL> CREATE PRIVATE TEMPORARY TABLE "ORA$PTT_tbl"
      ON COMMIT PRESERVE DEFINITION
      AS
      SELECT * FROM foo

---

    Code
      sql_query_save(con, sql("SELECT * FROM foo"), "tbl", temporary = FALSE)
    Output
      <SQL> CREATE TABLE "tbl" AS
      SELECT * FROM foo

---

    Code
      slice_sample(lf, n = 1)
    Output
      <SQL>
      SELECT "x"
      FROM (
        SELECT
          "df".*,
          CASE
      WHEN (NOT(((DBMS_RANDOM.VALUE()) IS NULL))) THEN ROW_NUMBER() OVER (PARTITION BY (CASE WHEN (((DBMS_RANDOM.VALUE()) IS NULL)) THEN 1 ELSE 0 END) ORDER BY DBMS_RANDOM.VALUE())
      END AS "col01"
        FROM "df"
      ) "q01"
      WHERE ("col01" <= 1)

# oracle_sql_table_create generates correct SQL

    Code
      oracle_sql_table_create(con, table_path("ORA$PTT_test"), c(x = "INTEGER", y = "TEXT"))
    Output
      <SQL> CREATE PRIVATE TEMPORARY TABLE ORA$PTT_test
      ("x" INTEGER, "y" TEXT)
      ON COMMIT PRESERVE DEFINITION
    Code
      oracle_sql_table_create(con, table_path("test"), c(x = "INTEGER", y = "TEXT"))
    Output
      <SQL> CREATE TABLE test 
      ("x" INTEGER, "y" TEXT)

# copy_inline uses UNION ALL

    Code
      remote_query(copy_inline(con, slice(y, 0)))
    Output
      <SQL> SELECT CAST(NULL AS INT) AS "id", CAST(NULL AS VARCHAR2(255)) AS "arr"
      FROM "DUAL"
      WHERE (0 = 1)
    Code
      remote_query(copy_inline(con, y))
    Output
      <SQL> SELECT CAST("id" AS INT) AS "id", CAST("arr" AS VARCHAR2(255)) AS "arr"
      FROM (
        SELECT NULL AS "id", NULL AS "arr"
        FROM "DUAL"
        WHERE (0 = 1)
      
        UNION ALL
      
        SELECT 1, '{1,2,3}' FROM DUAL
      ) "values_table"
    Code
      remote_query(copy_inline(con, slice(y, 0), types = types))
    Output
      <SQL> SELECT CAST(NULL AS bigint) AS "id", CAST(NULL AS integer[]) AS "arr"
      FROM "DUAL"
      WHERE (0 = 1)
    Code
      remote_query(copy_inline(con, y, types = types))
    Output
      <SQL> SELECT CAST("id" AS bigint) AS "id", CAST("arr" AS integer[]) AS "arr"
      FROM (
        SELECT NULL AS "id", NULL AS "arr"
        FROM "DUAL"
        WHERE (0 = 1)
      
        UNION ALL
      
        SELECT 1, '{1,2,3}' FROM DUAL
      ) "values_table"

# difftime is translated correctly

    Code
      translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
    Condition
      Error in `difftime()`:
      ! The only supported value for `units` on SQL backends is "days"

---

    Code
      translate_sql(difftime(start_date, end_date, tz = "UTC", units = "days"), con = con)
    Condition
      Error in `difftime()`:
      ! The `tz` argument is not supported for SQL backends.

# db_copy_to() wraps DBI errors

    Code
      (expect_error(db_copy_to(con = con, table = "tmp2", values = data.frame(x = c(1,
        1)), unique_indexes = list("x"))))
    Output
      <error/rlang_error>
      Error in `db_copy_to()`:
      ! Can't copy data to table `tmp2`.
      Caused by error in `db_create_index.DBIConnection()`:
      ! Can't create index on table `tmp2`.
      i Using SQL: CREATE UNIQUE INDEX `tmp2_x` ON `tmp2` (`x`)
      Caused by error:
      ! dummy DBI error

# db_copy_to() can overwrite a table

    Code
      (expect_error(db_copy_to(con = con, table = "tmp", values = data.frame(x = c(1,
        1)))))
    Output
      <error/rlang_error>
      Error in `db_copy_to()`:
      ! Can't copy data to table `tmp`.
      Caused by error in `dplyr::db_write_table()`:
      ! Can't write table table `tmp`.
      Caused by error:
      ! dummy DBI error

# db_save_query() can overwrite a table

    Code
      (expect_error(db_save_query(con = con, sql = "SELECT 2 FROM tmp", name = "tmp"))
      )
    Output
      <error/rlang_error>
      Error in `db_save_query()`:
      ! Can't save query to table `tmp`.
      i Using SQL: CREATE TEMPORARY TABLE `tmp` AS `SELECT 2 FROM tmp`
      Caused by error:
      ! dummy DBI error

# case_when converted to CASE WHEN

    Code
      translate_sql(case_when(x > 1L ~ "a"))
    Output
      <SQL> CASE WHEN (`x` > 1) THEN 'a' END

# even inside mutate

    Code
      out$select[[2]]
    Output
      [1] "CASE WHEN (`x` > 1) THEN 'a' END"

# case_when translates correctly to ELSE when TRUE ~ is used 2

    Code
      translate_sql(case_when(x == 1L ~ "yes", x == 0L ~ "no", TRUE ~ "undefined"))
    Output
      <SQL> CASE WHEN (`x` = 1) THEN 'yes' WHEN (`x` = 0) THEN 'no' ELSE 'undefined' END

# case_when uses the .default arg

    Code
      translate_sql(case_when(x == 1L ~ "yes", x == 0L ~ "no", .default = "undefined"))
    Output
      <SQL> CASE WHEN (`x` = 1) THEN 'yes' WHEN (`x` = 0) THEN 'no' ELSE 'undefined' END

---

    Code
      translate_sql(case_when(x == 1L ~ "yes", x == 0L ~ "no", .default = x + 1))
    Output
      <SQL> CASE WHEN (`x` = 1) THEN 'yes' WHEN (`x` = 0) THEN 'no' ELSE `x` + 1.0 END

---

    Code
      translate_sql(case_when(x == 1L ~ "yes", x == 0L ~ "no", TRUE ~ "true",
      .default = "undefined"))
    Output
      <SQL> CASE WHEN (`x` = 1) THEN 'yes' WHEN (`x` = 0) THEN 'no' ELSE 'true' END

# case_when does not support .ptype and .size

    Code
      (expect_error(translate_sql(case_when(x == 1L ~ "yes", .ptype = character()))))
    Output
      <error/rlang_error>
      Error in `case_when()`:
      ! `.ptype` is not supported in SQL translations.
    Code
      (expect_error(translate_sql(case_when(x == 1L ~ "yes", .size = 1))))
    Output
      <error/rlang_error>
      Error in `case_when()`:
      ! `.size` is not supported in SQL translations.

# long case_when is on multiple lines

    Code
      translate_sql(case_when(x == 1L ~ "this is long", x == 0L ~ "so it should",
      TRUE ~ "be wrapped"))
    Output
      <SQL> CASE
      WHEN (`x` = 1) THEN 'this is long'
      WHEN (`x` = 0) THEN 'so it should'
      ELSE 'be wrapped'
      END

# conditionals check arguments

    Code
      translate_sql(case_when())
    Condition
      Error in `sql_case_when()`:
      ! No cases provided

---

    Code
      translate_sql(switch(x, 1L, 2L))
    Condition
      Error in `sql_switch()`:
      ! Can only have one unnamed (ELSE) input

# LHS can handle bang bang

    Code
      translate_sql(case_match(x, !!1L ~ "x"))
    Output
      <SQL> CASE WHEN (`x` IN (1)) THEN 'x' END
    Code
      translate_sql(case_match(x, !!c(1L, 2L) ~ "x"))
    Output
      <SQL> CASE WHEN (`x` IN (1, 2)) THEN 'x' END
    Code
      translate_sql(case_match(x, !!c(NA, 1L) ~ "x"))
    Output
      <SQL> CASE WHEN (`x` IN (1) OR `x` IS NULL) THEN 'x' END

# requires at least one condition

    Code
      translate_sql(case_match(x))
    Condition
      Error in `case_match()`:
      ! No cases provided

---

    Code
      translate_sql(case_match(x, NULL))
    Condition
      Error in `case_match()`:
      ! No cases provided

# `.ptype` not supported

    Code
      (expect_error(translate_sql(case_match(x, 1 ~ 1, .ptype = integer()))))
    Output
      <error/rlang_error>
      Error in `case_match()`:
      ! `.ptype` is not supported in SQL translations.

# .x must be a symbol

    Code
      (expect_error(translate_sql(case_match(1, 1 ~ 1))))
    Output
      <error/rlang_error>
      Error in `case_match()`:
      ! `.x` must be a variable or function call, not a number.

# errors about add argument 

    Code
      gf <- mf %>% group_by(x) %>% group_by(y, add = TRUE)
    Condition
      Error:
      ! The `add` argument of `group_by()` was deprecated in dplyr 1.0.0 and is now defunct.
      i Please use the `.add` argument instead.

# errors for .drop = FALSE

    Code
      lazy_frame(x = 1:3, y = 1:3) %>% group_by(y, .drop = FALSE)
    Condition
      Error in `group_by()`:
      ! `.drop = FALSE` isn't supported on database backends.
      i It must be TRUE instead.

# informative errors for missing variables

    Code
      (expect_error(lazy_frame(x = 1:3) %>% group_by(y)))
    Output
      <error/rlang_error>
      Error in `group_by()`:
      i In argument: `y`
      Caused by error:
      ! Object `y` not found.

# group_by() produces nice error messages

    Code
      lf %>% group_by(z = non_existent + 1)
    Condition
      Error in `group_by()`:
      i In argument: `z = non_existent + 1`
      Caused by error:
      ! Object `non_existent` not found.
    Code
      lf %>% group_by(across(non_existent))
    Condition
      Error in `group_by()`:
      i In argument: `across(non_existent)`
      Caused by error in `across()`:
      ! Can't select columns that don't exist.
      x Column `non_existent` doesn't exist.

# ungroup() produces nice error messages

    Code
      lazy_frame(x = 1) %>% ungroup(non_existent)
    Condition
      Error in `ungroup()`:
      ! Can't select columns that don't exist.
      x Column `non_existent` doesn't exist.

# distinct() produces optimized SQL

    Code
      (out <- distinct(head(lf, 2), x, y))
    Output
      <SQL>
      SELECT DISTINCT *
      FROM (
        SELECT *
        FROM "df"
        LIMIT 2
      ) AS "q01"

# distinct() after join is inlined

    Code
      show_query(out)
    Output
      <SQL>
      SELECT DISTINCT "df_LHS".*, "z"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x")

---

    Code
      show_query(out)
    Output
      <SQL>
      SELECT DISTINCT "df_LHS"."x" AS "x"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x")

# distinct ignores groups when computing variables (#1081)

    Code
      show_query(db_distinct)
    Output
      <SQL>
      SELECT DISTINCT `g`, COUNT(*) OVER () AS `n`
      FROM `df`

# distinct respects window_order when .keep_all is TRUE

    Code
      distinct(window_order(lf, desc(y)), x, .keep_all = TRUE)
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT *, ROW_NUMBER() OVER (PARTITION BY "x" ORDER BY "y" DESC) AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" = 1)

# distinct uses dummy window order when .keep_all is TRUE and no order is used

    Code
      distinct(lf, x, .keep_all = TRUE)
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT *, ROW_NUMBER() OVER (PARTITION BY "x" ORDER BY "x") AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" = 1)

# symbols weights are dropped in output

    Code
      dbplyr_uncount(df, w) %>% show_query()
    Output
      <SQL>
      SELECT `x`
      FROM `test`
      INNER JOIN (
        SELECT CAST(`..dbplyr_row_id` AS INTEGER) AS `..dbplyr_row_id`
        FROM (
          SELECT NULL AS `..dbplyr_row_id`
          WHERE (0 = 1)
          UNION ALL
          VALUES (1)
        ) AS `values_table`
      ) AS `RHS`
        ON (`RHS`.`..dbplyr_row_id` <= `test`.`w`)

---

    Code
      df %>% mutate(w = w + 1) %>% dbplyr_uncount(w) %>% show_query()
    Output
      <SQL>
      SELECT `x`
      FROM (
        SELECT `x`, `w` + 1.0 AS `w`
        FROM `test`
      ) AS `LHS`
      INNER JOIN (
        SELECT CAST(`..dbplyr_row_id` AS INTEGER) AS `..dbplyr_row_id`
        FROM (
          SELECT NULL AS `..dbplyr_row_id`
          WHERE (0 = 1)
          UNION ALL
          VALUES (1), (2)
        ) AS `values_table`
      ) AS `RHS`
        ON (`RHS`.`..dbplyr_row_id` <= `LHS`.`w`)

# frame is checked

    Code
      translate_sql(sum(x, na.rm = TRUE), con = con, vars_frame = c(1, 0))
    Condition
      Warning:
      Windowed expression `SUM("x")` does not have explicit order.
      i Please use `arrange()`, `window_order()`, or `.order` to make deterministic.
      Error in `rows()`:
      ! `from` (1) must be less than `to` (0)

# win_rank works in both directions

    Code
      translate_sql(row_number(x), con = con)
    Output
      <SQL> CASE
      WHEN (NOT(("x" IS NULL))) THEN ROW_NUMBER() OVER (PARTITION BY (CASE WHEN (("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x")
      END

---

    Code
      translate_sql(row_number(desc(x)), con = con)
    Output
      <SQL> CASE
      WHEN (NOT(("x" IS NULL))) THEN ROW_NUMBER() OVER (PARTITION BY (CASE WHEN (("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x" DESC)
      END

# win_rank works with multiple variables

    Code
      translate_sql(row_number(tibble(x, desc(y))), con = con)
    Output
      <SQL> CASE
      WHEN (NOT(("x" IS NULL) OR ("y" IS NULL))) THEN ROW_NUMBER() OVER (PARTITION BY (CASE WHEN (("x" IS NULL) OR ("y" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x", "y" DESC)
      END

# win_rank(c()) gives an informative error

    Code
      translate_sql(row_number(c(x)), con = con)
    Condition
      Error in `row_number()`:
      ! Can't use `c()` in `ROW_NUMBER()`
      i Did you mean to use `tibble(x)` instead?

# row_number() with and without group_by() and arrange()

    Code
      mutate(mf, rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER () AS "rown"
      FROM "df"

---

    Code
      mutate(group_by(mf, y), rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER (PARTITION BY "y") AS "rown"
      FROM "df"

---

    Code
      mutate(arrange(group_by(mf, y), y), rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER (PARTITION BY "y" ORDER BY "y") AS "rown"
      FROM "df"
      ORDER BY "y"

---

    Code
      mutate(arrange(mf, y), rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER (ORDER BY "y") AS "rown"
      FROM "df"
      ORDER BY "y"

# window_frame()

    Code
      show_query(mutate(window_order(window_frame(lf, -3, 0), x), z = sum(y, na.rm = TRUE)))
    Output
      <SQL>
      SELECT *, SUM("y") OVER (ORDER BY "x" ROWS 3 PRECEDING) AS "z"
      FROM "df"

---

    Code
      show_query(mutate(window_order(window_frame(lf, -3), x), z = sum(y, na.rm = TRUE)))
    Output
      <SQL>
      SELECT
        *,
        SUM("y") OVER (ORDER BY "x" ROWS BETWEEN 3 PRECEDING AND UNBOUNDED FOLLOWING) AS "z"
      FROM "df"

# window_frame() checks arguments

    Code
      window_frame(lf, "a")
    Condition
      Error in `window_frame()`:
      ! `from` must be a whole number, not the string "a".

---

    Code
      window_frame(lf, 1:2)
    Condition
      Error in `window_frame()`:
      ! `from` must be a whole number, not an integer vector.

---

    Code
      window_frame(lf, 1, "a")
    Condition
      Error in `window_frame()`:
      ! `to` must be a whole number, not the string "a".

---

    Code
      window_frame(lf, 1, 1:2)
    Condition
      Error in `window_frame()`:
      ! `to` must be a whole number, not an integer vector.

# useful error for 1st edition

    Code
      check_2ed(con)
    Condition
      Error:
      ! <Test> uses dbplyr's 1st edition interface, which is no longer supported.
      i Please contact the maintainer of the package for a solution.

# custom clock functions translated correctly

    Code
      translate_sql(date_count_between(date_column_1, date_column_2, "year"), con = con)
    Condition
      Error in `date_count_between()`:
      ! `precision = "year"` isn't supported on database backends.
      i It must be "day" instead.

---

    Code
      translate_sql(date_count_between(date_column_1, date_column_2, "day", n = 5),
      con = con)
    Condition
      Error in `date_count_between()`:
      ! `n = 5` isn't supported on database backends.
      i It must be 1 instead.

# difftime is translated correctly

    Code
      translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
    Condition
      Error in `difftime()`:
      ! `units = "auto"` isn't supported on database backends.
      i It must be "days" instead.

---

    Code
      translate_sql(difftime(start_date, end_date, tz = "UTC", units = "days"), con = con)
    Condition
      Error in `difftime()`:
      ! Argument `tz` isn't supported on database backends.

# table_path possess key methods

    Code
      name <- table_path(c("x", "y", "z"))
      name
    Output
      <table_path> x, y, z

# can check for table name

    Code
      foo(1)
    Condition
      Error in `foo()`:
      ! `y` must be a <table_path>, not a string.
      i This is an internal error that was detected in the dbplyr package.
        Please report it at <https://github.com/tidyverse/dbplyr/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

# as_table_path validates its inputs

    Code
      as_table_path("x")
    Condition
      Error in `as_table_path()`:
      ! `con` is absent but must be supplied.
    Code
      as_table_path(c("x", "y"), con)
    Condition
      Error:
      ! `c("x", "y")` must be a single string, not a character vector.
    Code
      as_table_path(1, con)
    Condition
      Error in `as_table_path()`:
      ! `1` uses unknown specification for table name
    Code
      as_table_path(I(1), con)
    Condition
      Error:
      ! `I(1)` must be a single string, not the number 1.

# as_table_path warns when using sql

    Code
      as_table_path(sql("x"), con)
    Condition
      Warning:
      `sql("x")` uses SQL where a table identifier is expected.
      i If you want to use a literal (unquoted) identifier use `I()` instead.
    Output
      <table_path> x

# generated sql doesn't change unexpectedly

    Code
      setdiff(lf, lf)
    Output
      <SQL>
      (
        SELECT *
        FROM "df"
      )
      EXCEPT
      (
        SELECT *
        FROM "df"
      )

---

    Code
      intersect(lf, lf)
    Output
      <SQL>
      (
        SELECT *
        FROM "df"
      )
      INTERSECT
      (
        SELECT *
        FROM "df"
      )

---

    Code
      setdiff(lf, lf, all = TRUE)
    Output
      <SQL>
      (
        SELECT *
        FROM "df"
      )
      EXCEPT ALL
      (
        SELECT *
        FROM "df"
      )

---

    Code
      intersect(lf, lf, all = TRUE)
    Output
      <SQL>
      (
        SELECT *
        FROM "df"
      )
      INTERSECT ALL
      (
        SELECT *
        FROM "df"
      )

# print method renders sql

    Code
      lf$lazy_query
    Output
      -- <lazy_select_query> -------------------------
      SELECT *, "x" + 1.0 AS "y"
      FROM "df"
    Code
      sql_build(lf)
    Output
      -- <select_query> -------------------------
      SELECT *, "x" + 1.0 AS "y"
      FROM "df"

# slice, head and tail aren't available

    Code
      slice(lf)
    Condition
      Error in `slice()`:
      ! `slice()` is not supported on database backends.

---

    Code
      slice_head(lf)
    Condition
      Error in `slice_head()`:
      ! `slice_head()` is not supported on database backends.
      i Please use `slice_min()` instead.

---

    Code
      slice_tail(lf)
    Condition
      Error in `slice_tail()`:
      ! `slice_tail()` is not supported on database backends.
      i Please use `slice_max()` instead.

# min, max, and sample generate useful sql

    Code
      slice_min(lf, x, n = 1)
    Output
      <SQL>
      SELECT "x"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("x" IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN (("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x")
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)
    Code
      slice_max(lf, x, prop = 0.5)
    Output
      <SQL>
      SELECT "x"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("x" IS NULL))) THEN CUME_DIST() OVER (PARTITION BY (CASE WHEN (("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x" DESC)
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 0.5)
    Code
      slice_sample(lf, x, n = 1)
    Output
      <SQL>
      SELECT "x"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(((RANDOM()) IS NULL))) THEN ROW_NUMBER() OVER (PARTITION BY (CASE WHEN (((RANDOM()) IS NULL)) THEN 1 ELSE 0 END) ORDER BY RANDOM())
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)

# slice_min handles arguments

    Code
      slice_min(db)
    Condition
      Error in `slice_min()`:
      ! `order_by` is absent but must be supplied.

---

    Code
      slice_min(db, id, prop = 0.5, with_ties = FALSE)
    Condition
      Error in `slice_min()`:
      ! Can only use `prop` when `with_ties = TRUE`

---

    Code
      slice_min(db, id, n = 1, na_rm = FALSE)
    Condition
      Error in `slice_min()`:
      ! `na_rm = FALSE` isn't supported on database backends.
      i It must be TRUE instead.

# slice_max orders in opposite order

    Code
      slice_max(db)
    Condition
      Error in `slice_max()`:
      ! `order_by` is absent but must be supplied.

---

    Code
      slice_max(db, id, n = 1, na_rm = FALSE)
    Condition
      Error in `slice_max()`:
      ! `na_rm = FALSE` isn't supported on database backends.
      i It must be TRUE instead.

# slice_* can use data masking pronouns

    Code
      slice_max(lf, x)
    Output
      <SQL>
      SELECT "x", "id"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("x" IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN (("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x" DESC)
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)
    Code
      slice_max(lf, .data$x)
    Output
      <SQL>
      SELECT "x", "id"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("x" IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN (("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x" DESC)
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)
    Code
      slice_max(lf, .data$x * .env$x)
    Output
      <SQL>
      SELECT "x", "id"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT((("x" * -1) IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN ((("x" * -1) IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x" * -1 DESC)
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)

# slice_min and slice_max give clear errors for bad order_by

    Code
      slice_min(lf, doesntexist)
    Condition
      Error in `slice_min()`:
      i In argument: `order_by = doesntexist`
      Caused by error:
      ! Object `doesntexist` not found.
    Code
      slice_max(lf, doesntexist)
    Condition
      Error in `slice_max()`:
      i In argument: `order_by = doesntexist`
      Caused by error:
      ! Object `doesntexist` not found.

# slice_sample errors when expected

    Code
      slice_sample(db, replace = TRUE)
    Condition
      Error in `slice_sample()`:
      ! Sampling with replacement is not supported on database backends

---

    Code
      slice_sample(db, weight_by = x)
    Condition
      Error in `slice_sample()`:
      ! Weighted resampling is not supported on database backends

---

    Code
      slice_sample(db, prop = 0.5)
    Condition
      Error in `slice_sample()`:
      ! Sampling by `prop` is not supported on database backends

# check_slice_size checks for common issues

    Code
      slice_sample(lf, n = 1, prop = 1)
    Condition
      Error in `slice_sample()`:
      ! Must supply exactly one of `n` and `prop` arguments.

---

    Code
      slice_sample(lf, n = "a")
    Condition
      Error in `slice_sample()`:
      ! `n` must be a single number.

---

    Code
      slice_sample(lf, prop = "a")
    Condition
      Error in `slice_sample()`:
      ! `prop` must be a single number

---

    Code
      slice_sample(lf, n = -1)
    Condition
      Error in `slice_sample()`:
      ! `n` must be a non-missing positive number.

---

    Code
      slice_sample(lf, prop = -1)
    Condition
      Error in `slice_sample()`:
      ! `prop` must be a non-missing positive number.

# slice_helper `by` errors use correct error context and correct `by_arg`

    Code
      slice_min(gdf, order_by = x, by = x)
    Condition
      Error in `slice_min()`:
      ! Can't supply `by` when `data` is a grouped data frame.
    Code
      slice_max(gdf, order_by = x, by = x)
    Condition
      Error in `slice_max()`:
      ! Can't supply `by` when `data` is a grouped data frame.
    Code
      slice_sample(gdf, n = 1, by = x)
    Condition
      Error in `slice_sample()`:
      ! Can't supply `by` when `data` is a grouped data frame.

# slice_min/max can order by multiple columns

    Code
      slice_min(lf, tibble(x))
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("x" IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN (("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x")
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)
    Code
      slice_min(lf, tibble::tibble(x, y))
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("x" IS NULL) OR ("y" IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN (("x" IS NULL) OR ("y" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x", "y")
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)
    Code
      slice_min(lf, data.frame(y, x))
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("y" IS NULL) OR ("x" IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN (("y" IS NULL) OR ("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "y", "x")
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)

---

    Code
      slice_max(lf, tibble(x))
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("x" IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN (("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x" DESC)
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)
    Code
      slice_max(lf, tibble::tibble(x, y))
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("x" IS NULL) OR ("y" IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN (("x" IS NULL) OR ("y" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "x" DESC, "y" DESC)
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)
    Code
      slice_max(lf, data.frame(y, x))
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(("y" IS NULL) OR ("x" IS NULL))) THEN RANK() OVER (PARTITION BY (CASE WHEN (("y" IS NULL) OR ("x" IS NULL)) THEN 1 ELSE 0 END) ORDER BY "y" DESC, "x" DESC)
      END AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("col01" <= 1)

# slice_min/max informs if order_by uses c()

    Can't use `c()` in `slice_min()`
    i Did you mean to use `tibble(x, y)` instead?

# lazy_semi_join_query() checks arguments

    Code
      (my_lazy_semi_join_query(x = lazy_frame(x = 1)))
    Condition
      Error in `my_lazy_semi_join_query()`:
      ! `x` must be a lazy query, not a <tbl_TestConnection> object.
    Code
      (my_lazy_semi_join_query(y = lazy_frame(x = 1)))
    Condition
      Error in `my_lazy_semi_join_query()`:
      ! `y` must be a lazy query, not a <tbl_TestConnection> object.

---

    Code
      (my_lazy_semi_join_query(by = lmod(by0, x = 1)))
    Condition
      Error in `my_lazy_semi_join_query()`:
      ! `by$x` must be a character vector, not the number 1.

---

    Code
      (my_lazy_semi_join_query(anti = NA))
    Condition
      Error in `my_lazy_semi_join_query()`:
      ! `anti` must be `TRUE` or `FALSE`, not `NA`.

# generated sql doesn't change unexpectedly

    Code
      right_join(lf, lf)
    Message
      Joining with `by = join_by(x, y)`
    Output
      <SQL>
      SELECT "df_RHS".*
      FROM "df" AS "df_LHS"
      RIGHT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x" AND "df_LHS"."y" = "df_RHS"."y")

---

    Code
      full_join(lf, lf)
    Message
      Joining with `by = join_by(x, y)`
    Output
      <SQL>
      SELECT
        COALESCE("df_LHS"."x", "df_RHS"."x") AS "x",
        COALESCE("df_LHS"."y", "df_RHS"."y") AS "y"
      FROM "df" AS "df_LHS"
      FULL JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x" AND "df_LHS"."y" = "df_RHS"."y")

# disambiguate variables that only differ in case

    Code
      full_join(lf1, lf2, by = "y")
    Output
      <SQL>
      SELECT
        "df_LHS"."x" AS "x",
        COALESCE("df_LHS"."y", "df_RHS"."y") AS "y",
        "df_RHS"."X" AS "X"
      FROM "df" AS "df_LHS"
      FULL JOIN "df" AS "df_RHS"
        ON ("df_LHS"."y" = "df_RHS"."y")

# sql_on query doesn't change unexpectedly

    Code
      right_join(lf1, lf2, sql_on = "LHS.y < RHS.z")
    Output
      <SQL>
      SELECT "LHS"."x" AS "x.x", "y", "RHS"."x" AS "x.y", "z"
      FROM "df" AS "LHS"
      RIGHT JOIN "df" AS "RHS"
        ON (LHS.y < RHS.z)

---

    Code
      full_join(lf1, lf2, sql_on = "LHS.y < RHS.z")
    Output
      <SQL>
      SELECT "LHS"."x" AS "x.x", "y", "RHS"."x" AS "x.y", "z"
      FROM "df" AS "LHS"
      FULL JOIN "df" AS "RHS"
        ON (LHS.y < RHS.z)

# can pivot all cols to long

    Code
      tidyr::pivot_longer(lazy_frame(x = 1:2, y = 3:4), x:y)
    Output
      <SQL>
      SELECT 'x' AS "name", "x" AS "value"
      FROM "df"
      
      UNION ALL
      
      SELECT 'y' AS "name", "y" AS "value"
      FROM "df"

# can add multiple columns from spec

    Code
      pv
    Output
      <SQL>
      SELECT 11 AS "a", 13 AS "b", "x" AS "v"
      FROM "df"
      
      UNION ALL
      
      SELECT 12 AS "a", 14 AS "b", "y" AS "v"
      FROM "df"

# preserves original keys

    Code
      pv
    Output
      <SQL>
      SELECT "x", 'y' AS "name", "y" AS "value"
      FROM "df"
      
      UNION ALL
      
      SELECT "x", 'z' AS "name", "z" AS "value"
      FROM "df"

# can drop missing values

    Code
      tidyr::pivot_longer(lazy_frame(x = c(1, NA), y = c(NA, 2)), x:y,
      values_drop_na = TRUE)
    Output
      <SQL>
      SELECT *
      FROM (
        SELECT 'x' AS "name", "x" AS "value"
        FROM "df"
      
        UNION ALL
      
        SELECT 'y' AS "name", "y" AS "value"
        FROM "df"
      ) AS "q01"
      WHERE (NOT(("value" IS NULL)))

# can handle missing combinations

    Code
      show_query(sql)
    Output
      <SQL>
      SELECT *, NULL AS `y`
      FROM (
        SELECT `id`, '1' AS `n`, `x_1` AS `x`
        FROM `df`
      ) AS `q01`
      
      UNION ALL
      
      SELECT `id`, '2' AS `n`, `x_2` AS `x`, `y_2` AS `y`
      FROM `df`

# can override default output column type

    Code
      tidyr::pivot_longer(lazy_frame(x = 1), x, values_transform = list(value = as.character))
    Output
      <SQL>
      SELECT 'x' AS "name", CAST("x" AS TEXT) AS "value"
      FROM "df"

# values_transform can be a formula

    Code
      tidyr::pivot_longer(lazy_frame(x = 1), x, values_transform = list(value = ~
        as.character(.x)))
    Output
      <SQL>
      SELECT 'x' AS "name", CAST("x" AS TEXT) AS "value"
      FROM "df"

# `values_transform` is validated

    Code
      (expect_error(tidyr::pivot_longer(df, x, values_transform = 1)))
    Output
      <error/rlang_error>
      Error in `dbplyr_pivot_longer_spec()`:
      ! `values_transform` must be `NULL`, a function, or a named list of functions.
    Code
      (expect_error(tidyr::pivot_longer(df, x, values_transform = list(~.x))))
    Output
      <error/rlang_error>
      Error in `dbplyr_pivot_longer_spec()`:
      ! All elements of `values_transform` must be named.

# can pivot to multiple measure cols

    Code
      pv
    Output
      <SQL>
      SELECT 1.0 AS "row", "x" AS "X", "y" AS "Y"
      FROM "df"

# .value can be at any position in `names_to`

    Code
      value_first
    Output
      <SQL>
      SELECT "i", 't1' AS "time", "y_t1" AS "y", "z_t1" AS "z"
      FROM "df"
      
      UNION ALL
      
      SELECT "i", 't2' AS "time", "y_t2" AS "y", "z_t2" AS "z"
      FROM "df"

---

    Code
      value_second
    Output
      <SQL>
      SELECT "i", 't1' AS "time", "t1_y" AS "y", "t1_z" AS "z"
      FROM "df"
      
      UNION ALL
      
      SELECT "i", 't2' AS "time", "t2_y" AS "y", "t2_z" AS "z"
      FROM "df"

# can repair names

    Code
      out <- tidyr::pivot_longer(df, c(x, y), names_repair = "unique")
    Message
      New names:
      * `name` -> `name...2`
      * `value` -> `value...3`
      * `name` -> `name...4`
      * `value` -> `value...5`

# values_ptype is not supported

    Code
      tidyr::pivot_longer(lazy_frame(x = 1:2, y = 3:4), x:y, values_ptypes = character())
    Condition
      Error in `tidyr::pivot_longer()`:
      ! Argument `values_ptypes` isn't supported on database backends.

# cols_vary is not supported

    Code
      tidyr::pivot_longer(lazy_frame(x = 1:2, y = 3:4), x:y, cols_vary = "fastest")
    Condition
      Error in `tidyr::pivot_longer()`:
      ! Argument `cols_vary` isn't supported on database backends.

# works with labels a character vector

    Code
      translate_sql(cut(x, 1:3, labels = c("a", "b", "c")), con = con)
    Condition
      Error in `cut()`:
      ! Can't recycle `labels` (size 3) to size 2.

# cut checks arguments

    Code
      translate_sql(cut(x, 1), con = con)
    Condition
      Error in `cut()`:
      ! `breaks` must have at least two values.

---

    Code
      translate_sql(cut(x, c(1, 1)), con = con)
    Condition
      Error in `cut()`:
      ! `breaks` are not unique.

---

    Code
      translate_sql(cut(x, c(1, 2, NA)), con = con)
    Condition
      Error in `cut()`:
      ! `breaks` values must not be missing.

# mutate() isn't inlined after distinct() #1119

    Code
      mutate(distinct(lf, x), x = 0)
    Output
      <SQL>
      SELECT 0.0 AS "x"
      FROM (
        SELECT DISTINCT *
        FROM "df"
      ) AS "q01"

# can use window function after summarise and pure projection #1104

    Code
      (expect_no_error(mutate(lf, r = row_number())))
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER () AS "r"
      FROM (
        SELECT "g"
        FROM "df"
        GROUP BY "g"
      ) AS "q01"

# can refer to fresly created values

    Code
      show_query(out2)
    Output
      <SQL>
      SELECT `x` + 4.0 AS `x`
      FROM (
        SELECT `x` + 2.0 AS `x`
        FROM (
          SELECT `x` + 1.0 AS `x`
          FROM `multi_mutate`
        ) AS `q01`
      ) AS `q01`

# transmute includes all needed variables

    Code
      out
    Output
      <SQL>
      SELECT "x", "x" + "y" AS "x2"
      FROM (
        SELECT "x" / 2.0 AS "x", "y"
        FROM "df"
      ) AS "q01"

# across() does not select grouping variables

    Code
      mutate(group_by(df, g), across(.fns = ~0))
    Output
      <SQL>
      SELECT "g", 0.0 AS "x"
      FROM "df"

---

    Code
      transmute(group_by(df, g), across(.fns = ~0))
    Output
      <SQL>
      SELECT "g", 0.0 AS "x"
      FROM "df"

# across() can access previously created variables

    Code
      remote_query(lf)
    Output
      <SQL> SELECT "x", SQRT("y") AS "y"
      FROM (
        SELECT *, 2.0 AS "y"
        FROM "df"
      ) AS "q01"

# across() uses original column rather than overridden one

    Code
      mutate(lf, x = -x, across(everything(), ~ .x / x), y = y + x)
    Output
      <SQL>
      SELECT "x", "y" + "x" AS "y", "z"
      FROM (
        SELECT "x" / "x" AS "x", "y" / "x" AS "y", "z" / "x" AS "z"
        FROM (
          SELECT -"x" AS "x", "y", "z"
          FROM "df"
        ) AS "q01"
      ) AS "q01"

# new columns take precedence over global variables

    Code
      remote_query(lf)
    Output
      <SQL> SELECT *, "y" + 1.0 AS "z"
      FROM (
        SELECT *, 2.0 AS "y"
        FROM "df"
      ) AS "q01"

# mutate() produces nice error messages

    Code
      mutate(lazy_frame(x = 1), z = non_existent + 1)
    Condition
      Error in `mutate()`:
      i In argument: `z = non_existent + 1`
      Caused by error:
      ! Object `non_existent` not found.
    Code
      mutate(lazy_frame(x = 1), across(x, .fns = "a"))
    Condition
      Error in `mutate()`:
      i In argument: `across(x, .fns = "a")`
      Caused by error in `across()`:
      ! `.fns` must be a function, a formula, or list of functions/formulas.

# .order and .frame are scoped to mutate call

    Code
      out
    Output
      <SQL>
      SELECT *, AVG("x") OVER (PARTITION BY "g") AS "r2"
      FROM (
        SELECT
          *,
          SUM("x") OVER (PARTITION BY "g" ORDER BY "y" ROWS UNBOUNDED PRECEDING) AS "r1"
        FROM "df"
      ) AS "q01"

# .order generates correct SQL

    Code
      mutate(lf, r = cumsum(x), .by = g, .order = y)
    Output
      <SQL>
      SELECT
        *,
        SUM("x") OVER (PARTITION BY "g" ORDER BY "y" ROWS UNBOUNDED PRECEDING) AS "r"
      FROM "df"
    Code
      mutate(lf, r = cumsum(x), .by = g, .order = c(x, desc(y)))
    Output
      <SQL>
      SELECT
        *,
        SUM("x") OVER (PARTITION BY "g" ORDER BY "x", "y" DESC ROWS UNBOUNDED PRECEDING) AS "r"
      FROM "df"

# .frame generates correct SQL

    Code
      mutate(lf, r = sum(x), .by = g, .order = y, .frame = c(-Inf, 0))
    Output
      <SQL>
      SELECT
        *,
        SUM("x") OVER (PARTITION BY "g" ORDER BY "y" ROWS UNBOUNDED PRECEDING) AS "r"
      FROM "df"
    Code
      mutate(lf, r = sum(x), .by = g, .order = y, .frame = c(-1, 1))
    Output
      <SQL>
      SELECT
        *,
        SUM("x") OVER (PARTITION BY "g" ORDER BY "y" ROWS BETWEEN 1 PRECEDING AND 1 FOLLOWING) AS "r"
      FROM "df"

# .order validates inputs

    Code
      mutate(lf, r = cumsum(x), .order = x + y)
    Condition
      Error in `mutate()`:
      ! Every element of `order` must be a single column name or a column wrapped in `desc()`.
      x Element 1 is `x + y`.
    Code
      mutate(lf, r = cumsum(x), .order = foo())
    Condition
      Error in `mutate()`:
      ! Every element of `order` must be a single column name or a column wrapped in `desc()`.
      x Element 1 is `foo()`.

# .frame validates inputs

    Code
      mutate(lf, r = sum(x), .frame = c(1, 2, 3))
    Condition
      Error in `mutate()`:
      ! `frame` must have size 2, not size 3.
    Code
      mutate(lf, r = sum(x), .frame = c("a", "b"))
    Condition
      Error in `mutate()`:
      ! `frame` must be a whole number, not the string "a".

# mutate generates subqueries as needed

    Code
      mutate(lf, x = x + 1, x = x + 1)
    Output
      <SQL>
      SELECT `x` + 1.0 AS `x`
      FROM (
        SELECT `x` + 1.0 AS `x`
        FROM `df`
      ) AS `q01`

---

    Code
      mutate(lf, x1 = x + 1, x2 = x1 + 1)
    Output
      <SQL>
      SELECT *, `x1` + 1.0 AS `x2`
      FROM (
        SELECT *, `x` + 1.0 AS `x1`
        FROM `df`
      ) AS `q01`

# mutate collapses over nested select

    Code
      mutate(select(lf, x:y), x = x * 2, y = y * 2)
    Output
      <SQL>
      SELECT "x" * 2.0 AS "x", "y" * 2.0 AS "y"
      FROM "df"

---

    Code
      mutate(select(lf, y:x), x = x * 2, y = y * 2)
    Output
      <SQL>
      SELECT "y" * 2.0 AS "y", "x" * 2.0 AS "x"
      FROM "df"

# var = NULL works when var is in original data

    Code
      remote_query(lf)
    Output
      <SQL> SELECT "x" * 2.0 AS "z"
      FROM (
        SELECT 2.0 AS "x"
        FROM "df"
      ) AS "q01"

# var = NULL when var is in final output

    Code
      remote_query(lf)
    Output
      <SQL> SELECT *, 3.0 AS "y"
      FROM "df"

# temp var with nested arguments

    Code
      remote_query(lf)
    Output
      <SQL> SELECT "x", "y" * 2.0 AS "z"
      FROM (
        SELECT *, 2.0 AS "y"
        FROM "df"
      ) AS "q01"

# ungroup() produces nice error messages

    Code
      pull(local_memdb_frame("df", x = 1), non_existent)
    Condition
      Error in `pull()`:
      Caused by error:
      ! object 'non_existent' not found
    Code
      pull(local_memdb_frame("df", x = 1), "non_existent")
    Condition
      Error in `pull()`:
      ! Can't extract columns that don't exist.
      x Column `non_existent` doesn't exist.
    Code
      pull(local_memdb_frame("df", x = 1), 1000)
    Condition
      Error in `pull()`:
      ! Can't extract columns past the end.
      i Location 1000 doesn't exist.
      i There is only 1 column.
    Code
      pull(local_memdb_frame("df", x = 1), x, "name_non_existent")
    Condition
      Error in `pull()`:
      ! Can't extract columns that don't exist.
      x Column `name_non_existent` doesn't exist.

# df must be a local or remote table

    Code
      copy_to(con, list(x = 1), name = "df")
    Condition
      Error in `copy_to()`:
      ! `df` must be a local dataframe or a remote tbl_sql

# as_copy() errors on invalid values

    Code
      as_copy("other")
    Condition
      Error:
      ! `copy` must be one of "none", "temp-table", or "inline", not "other".

# dbplyr_auto_copy() errors when copy = FALSE and different sources

    Code
      dbplyr_auto_copy(df, local_df, copy = FALSE)
    Condition
      Error:
      ! `x` and `y` must share the same source.
      i Use `copy = "temp-table"` to copy `y` to a temporary table.
      i USe `copy = "inline"` to convert `y` to inline SQL.

---

    Code
      dbplyr_auto_copy(df, local_df, copy = "none")
    Condition
      Error:
      ! `x` and `y` must share the same source.
      i Use `copy = "temp-table"` to copy `y` to a temporary table.
      i USe `copy = "inline"` to convert `y` to inline SQL.

# complete join pipeline works with SQLite and table alias

    Code
      left_join(lf1, lf2, by = "x", x_as = "df1", y_as = "df2")
    Output
      <SQL>
      SELECT "df1"."x" AS "x", "y"
      FROM "lf1" AS "df1"
      LEFT JOIN "lf2" AS "df2"
        ON ("df1"."x" = "df2"."x")

# complete semi join works with SQLite and table alias

    Code
      inner_join(lf1, lf2, by = "x", x_as = "df1", y_as = "df2")
    Output
      <SQL>
      SELECT "df1".*
      FROM "df" AS "df1"
      INNER JOIN "df" AS "df2"
        ON ("df1"."x" = "df2"."x")

# join works with in_schema

    Code
      remote_query(left_join(df1, df2, by = "x"))
    Output
      <SQL> SELECT `df`.*, `z`
      FROM `foo`.`df`
      LEFT JOIN `foo`.`df2`
        ON (`df`.`x` = `df2`.`x`)

---

    Code
      remote_query(left_join(df1, df3, by = "x"))
    Output
      <SQL> SELECT `df_LHS`.*, `z`
      FROM `foo`.`df` AS `df_LHS`
      LEFT JOIN `foo2`.`df` AS `df_RHS`
        ON (`df_LHS`.`x` = `df_RHS`.`x`)

---

    Code
      remote_query(left_join(df4, df5, by = "x"))
    Output
      <SQL> SELECT `df_LHS`.*, `z`
      FROM foo.df AS `df_LHS`
      LEFT JOIN foo2.df AS `df_RHS`
        ON (`df_LHS`.`x` = `df_RHS`.`x`)

# alias truncates long table names at database limit

    Code
      remote_query(self_join2)
    Output
      <SQL> SELECT "a01234567890123456789012345678901234567890123456789012345678901".*
      FROM "a01234567890123456789012345678901234567890123456789012345678901"
      LEFT JOIN "a01234567890123456789012345678901234567890123456789012345678901" AS "34567890123456789012345678901234567890123456789012345678901_RHS"
        ON (
          "a01234567890123456789012345678901234567890123456789012345678901"."x" = "34567890123456789012345678901234567890123456789012345678901_RHS"."x" AND
          "a01234567890123456789012345678901234567890123456789012345678901"."y" = "34567890123456789012345678901234567890123456789012345678901_RHS"."y"
        )

---

    Code
      remote_query(self_join3)
    Output
      <SQL> SELECT
        "a01234567890123456789012345678901234567890123456789012345678901"."x" AS "x",
        "a01234567890123456789012345678901234567890123456789012345678901"."y" AS "y.x",
        "b01234567890123456789012345678901234567890123456789012345678901"."y" AS "y.y"
      FROM "a01234567890123456789012345678901234567890123456789012345678901"
      LEFT JOIN "a01234567890123456789012345678901234567890123456789012345678901" AS "34567890123456789012345678901234567890123456789012345678901...2"
        ON (
          "a01234567890123456789012345678901234567890123456789012345678901"."x" = "34567890123456789012345678901234567890123456789012345678901...2"."x" AND
          "a01234567890123456789012345678901234567890123456789012345678901"."y" = "34567890123456789012345678901234567890123456789012345678901...2"."y"
        )
      INNER JOIN "b01234567890123456789012345678901234567890123456789012345678901"
        ON ("a01234567890123456789012345678901234567890123456789012345678901"."x" = "b01234567890123456789012345678901234567890123456789012345678901"."x")

# cross join via by = character() is deprecated

    Code
      inner_join(df1, df2, by = character())
    Condition
      Error:
      ! Using `by = character()` to perform a cross join was deprecated in dbplyr 1.1.0 and is now defunct.
      i Please use `cross_join()` instead.
    Code
      full_join(df1, df2, by = character())
    Condition
      Error:
      ! Using `by = character()` to perform a cross join was deprecated in dbplyr 1.1.0 and is now defunct.
      i Please use `cross_join()` instead.

# join check `x_as` and `y_as`

    Code
      left_join(x, x, by = "x", y_as = c("A", "B"))
    Condition
      Error in `left_join()`:
      ! `y_as` must be a single string or `NULL`, not a character vector.

---

    Code
      left_join(x, x, by = "x", x_as = "LHS", y_as = "LHS")
    Condition
      Error in `left_join()`:
      ! `y_as` must be different from `x_as`.

# select() before join is inlined

    Code
      out_left
    Output
      <SQL>
      SELECT "a" AS "a2", "x1" AS "x", "b"
      FROM "lf1"
      LEFT JOIN "lf2"
        ON ("lf1"."x1" = "lf2"."x2")

# rename works with duplicate column names in join_by (#1572)

    Code
      out
    Output
      <SQL>
      SELECT "x", "y" AS "z"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" >= "df_RHS"."y" AND "df_LHS"."x" <= "df_RHS"."y")

# select() before semi_join is inlined

    Code
      out_semi
    Output
      <SQL>
      SELECT "a" AS "a2", "x1" AS "x"
      FROM "lf1"
      WHERE EXISTS (
        SELECT 1 FROM "lf2"
        WHERE ("lf1"."x1" = "lf2"."x2")
      )

# can combine full_join with other joins #1178

    Code
      left_join(full_join(lf1, lf2, by = "x"), lf3, by = "x")
    Output
      <SQL>
      SELECT "LHS".*, "z"
      FROM (
        SELECT COALESCE("df_LHS"."x", "df_RHS"."x") AS "x", "y"
        FROM "df" AS "df_LHS"
        FULL JOIN "df" AS "df_RHS"
          ON ("df_LHS"."x" = "df_RHS"."x")
      ) AS "LHS"
      LEFT JOIN "df"
        ON ("LHS"."x" = "df"."x")

---

    Code
      full_join(left_join(lf1, lf2, by = "x"), lf3, by = "x")
    Output
      <SQL>
      SELECT COALESCE("LHS"."x", "df"."x") AS "x", "y", "z"
      FROM (
        SELECT "df_LHS"."x" AS "x", "y"
        FROM "df" AS "df_LHS"
        LEFT JOIN "df" AS "df_RHS"
          ON ("df_LHS"."x" = "df_RHS"."x")
      ) AS "LHS"
      FULL JOIN "df"
        ON ("LHS"."x" = "df"."x")

---

    Code
      full_join(full_join(lf1, lf2, by = "x"), lf3, by = "x")
    Output
      <SQL>
      SELECT COALESCE("LHS"."x", "df"."x") AS "x", "y", "z"
      FROM (
        SELECT COALESCE("df_LHS"."x", "df_RHS"."x") AS "x", "y"
        FROM "df" AS "df_LHS"
        FULL JOIN "df" AS "df_RHS"
          ON ("df_LHS"."x" = "df_RHS"."x")
      ) AS "LHS"
      FULL JOIN "df"
        ON ("LHS"."x" = "df"."x")

# filter() before semi join is inlined

    Code
      out
    Output
      <SQL>
      SELECT "df_LHS".*
      FROM "df" AS "df_LHS"
      WHERE EXISTS (
        SELECT 1 FROM "df" AS "df_RHS"
        WHERE
          ("df_LHS"."x" = "df_RHS"."x2") AND
          ("df_RHS"."a" = 1) AND
          ("df_RHS"."b" = 2)
      )

# filtered aggregates with subsequent select are not inlined away in semi_join (#1474)

    Code
      out
    Output
      <SQL>
      SELECT "df".*
      FROM "df"
      WHERE EXISTS (
        SELECT 1 FROM (
        SELECT "x"
        FROM "df"
        GROUP BY "x"
        HAVING (COUNT(*) = 1.0)
      ) AS "RHS"
        WHERE ("df"."x" = "RHS"."x")
      )

# filtered window joins work in a semi_join

    Code
      show_query(out)
    Output
      <SQL>
      SELECT `df1`.*
      FROM `df1`
      WHERE NOT EXISTS (
        SELECT 1 FROM (
        SELECT *, ROW_NUMBER() OVER () AS `col01`
        FROM `df2`
      ) AS `RHS`
        WHERE (`df1`.`id` = `RHS`.`id`) AND (`RHS`.`col01` <= 3.0)
      )

# multiple joins create a single query

    Code
      out
    Output
      <SQL>
      SELECT "df1".*, "df2"."b" AS "b.x", "df3"."b" AS "b.y"
      FROM "df1"
      LEFT JOIN "df2"
        ON ("df1"."x" = "df2"."x")
      INNER JOIN "df3"
        ON ("df1"."x" = "df3"."x")

# can join 4 tables with same column #1101

    Code
      remote_query(out)
    Output
      <SQL> SELECT "lf1".*, "b", "c", "lf4"."a" AS "a4"
      FROM "lf1"
      INNER JOIN "lf2"
        ON ("lf1"."x" = "lf2"."x")
      INNER JOIN "lf3"
        ON ("lf1"."x" = "lf3"."x")
      INNER JOIN "lf4"
        ON ("lf1"."x" = "lf4"."x")

# multiple joins produce separate queries if using right/full join

    Code
      remote_query(out)
    Output
      <SQL> SELECT "df3"."x" AS "x", "a", "LHS"."b" AS "b.x", "df3"."b" AS "b.y"
      FROM (
        SELECT "df1".*, "b"
        FROM "df1"
        LEFT JOIN "df2"
          ON ("df1"."x" = "df2"."x")
      ) AS "LHS"
      RIGHT JOIN "df3"
        ON ("LHS"."x" = "df3"."x")

# can't use `keep = FALSE` with non-equi conditions (#6499)

    Code
      left_join(df1, df2, join_by(overlaps(xl, xu, yl, yu)), keep = FALSE)
    Condition
      Error in `left_join()`:
      ! Can't set `keep = FALSE` when using an inequality, rolling, or overlap join.

---

    Code
      full_join(df1, df2, join_by(overlaps(xl, xu, yl, yu)), keep = FALSE)
    Condition
      Error in `full_join()`:
      ! Can't set `keep = FALSE` when using an inequality, rolling, or overlap join.

# can translate join conditions

    Code
      left_join(lf1, lf1, by = join_by(a == a, b >= b, c < c), keep = TRUE)
    Output
      <SQL>
      SELECT
        "df_LHS"."a" AS "a.x",
        "df_LHS"."b" AS "b.x",
        "df_LHS"."c" AS "c.x",
        "df_RHS"."a" AS "a.y",
        "df_RHS"."b" AS "b.y",
        "df_RHS"."c" AS "c.y"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON (
          "df_LHS"."a" = "df_RHS"."a" AND
          "df_LHS"."b" >= "df_RHS"."b" AND
          "df_LHS"."c" < "df_RHS"."c"
        )

# rolling joins aren't supported

    Code
      (expect_error(left_join(lf, lf, join_by(closest(x >= y)))))
    Output
      <error/rlang_error>
      Error in `left_join()`:
      ! Rolling joins aren't supported on database backends.
    Code
      (expect_error(semi_join(lf, lf, join_by(closest(x >= y)))))
    Output
      <error/rlang_error>
      Error in `semi_join()`:
      ! Rolling joins aren't supported on database backends.

# `na_matches` is validated

    Code
      left_join(df, df, by = "x", na_matches = 1)
    Condition
      Error in `left_join()`:
      ! `na_matches` must be a character vector, not the number 1.

---

    Code
      left_join(df, df, by = "x", na_matches = "foo")
    Condition
      Error in `left_join()`:
      ! `na_matches` must be one of "na" or "never", not "foo".

---

    Code
      semi_join(df, df, by = "x", na_matches = 1)
    Condition
      Error in `semi_join()`:
      ! `na_matches` must be a character vector, not the number 1.

---

    Code
      semi_join(df, df, by = "x", na_matches = "foo")
    Condition
      Error in `semi_join()`:
      ! `na_matches` must be one of "na" or "never", not "foo".

# using multiple gives an informative error

    Code
      left_join(lf, lf, by = "x", multiple = "first")
    Condition
      Error in `left_join()`:
      ! Argument `multiple` isn't supported on database backends.
      For equi joins you can instead add a unique index for the join columns in `y`.
      `db_create_index( con = remote_con(y), table = remote_name(y), columns = "x", unique = TRUE )`

# using unmatched gives an informative error

    Code
      left_join(lf, lf, by = "x", unmatched = "error")
    Condition
      Error in `left_join()`:
      ! Argument `unmatched` isn't supported on database backends.
      i For equi joins you can instead add a foreign key from `x` to `y` for the join columns.

# using relationship gives an informative error

    Code
      left_join(lf, lf, by = "x", relationship = "one-to-one")
    Condition
      Error in `left_join()`:
      ! `relationship = "one-to-one"` isn't supported on database backends.
      i It must be "many-to-many" or `NULL` instead.

# can optionally match NA values

    Code
      left_join(lf1, lf2, by = "x", na_matches = "na")
    Output
      <SQL>
      SELECT "lf1"."x" AS "x"
      FROM "lf1"
      LEFT JOIN "lf2"
        ON (("lf1"."x") IS NOT DISTINCT FROM ("lf2"."x"))

---

    Code
      semi_join(lf1, lf2, by = "x", na_matches = "na")
    Output
      <SQL>
      SELECT "lf1".*
      FROM "lf1"
      WHERE EXISTS (
        SELECT 1 FROM "lf2"
        WHERE (("lf1"."x") IS NOT DISTINCT FROM ("lf2"."x"))
      )

---

    Code
      left_join(lf1, lf3, by = by, na_matches = "na")
    Output
      <SQL>
      SELECT "x", "lf3".*
      FROM "lf1"
      LEFT JOIN "lf3"
        ON (
          ("lf1"."x" >= "lf3"."lower" OR ("lf1"."x" IS NULL AND "lf3"."lower" IS NULL)) AND
          ("lf1"."x" <= "lf3"."upper" OR ("lf1"."x" IS NULL AND "lf3"."upper" IS NULL))
        )

# suffix arg is checked

    Code
      (expect_error(inner_join(lf1, lf2, by = "x", suffix = "a")))
    Output
      <error/rlang_error>
      Error in `inner_join()`:
      ! `suffix` must be a character vector of length 2, not a string of length 1.
    Code
      (expect_error(inner_join(lf1, lf2, by = "x", suffix = 1L)))
    Output
      <error/rlang_error>
      Error in `inner_join()`:
      ! `suffix` must be a character vector of length 2, not an integer of length 1.

# cross_join() errors if `...` is not empty (#1792)

    Code
      cross_join(lf1, lf2, lf3)
    Condition
      Error in `cross_join()`:
      ! `...` must be empty.
      x Problematic argument:
      * ..1 = lf3
      i Did you forget to name an argument?

# joins reuse queries in cte mode

    Code
      remote_query(left_join(lf, lf, by = "x"), sql_options = sql_options(cte = TRUE))
    Output
      <SQL> WITH "q01" AS (
        SELECT "lf1_LHS"."x" AS "x"
        FROM "lf1" AS "lf1_LHS"
        INNER JOIN "lf1" AS "lf1_RHS"
          ON ("lf1_LHS"."x" = "lf1_RHS"."x")
      )
      SELECT "lf1...1"."x" AS "x"
      FROM "lf1" AS "lf1...1"
      INNER JOIN "lf1" AS "lf1...2"
        ON ("lf1...1"."x" = "lf1...2"."x")
      LEFT JOIN "q01" AS "...3"
        ON ("lf1...1"."x" = "...3"."x")

# joins correctly quote reused queries in CTEs

    Code
      show_query(left_join(lf, lf, by = join_by(z)), sql_options = sql_options(cte = TRUE))
    Output
      <SQL>
      WITH "q01" AS (
        SELECT *, "x" * 2.0 AS "z"
        FROM "lf"
      )
      SELECT "LHS"."x" AS "x.x", "LHS"."z" AS "z", "RHS"."x" AS "x.y"
      FROM "q01" AS "LHS"
      LEFT JOIN "q01" AS "RHS"
        ON ("LHS"."z" = "RHS"."z")

# src_memdb() is deprecated

    Code
      . <- src_memdb()
    Condition
      Warning:
      `src_memdb()` was deprecated in dbplyr 2.6.0.
      i Please use `memdb()` instead.

# tbl_memdb() is deprecated

    Code
      . <- tbl_memdb(data.frame(x = 1))
    Condition
      Warning:
      `tbl_memdb()` was deprecated in dbplyr 2.6.0.
      i Use `copy_to(memdb(), df)` instead

# memdb_frame() with data frame is deprecated

    Code
      . <- memdb_frame(data.frame(x = 1))
    Condition
      Warning:
      memdb_frame(data.frame(...)) was deprecated in dbplyr 2.6.0.
      i Use `copy_to(memdb(), df)` instead.

# argument src is deprecated

    Code
      tbl_lazy(mtcars, src = simulate_sqlite())
    Condition
      Error in `tbl_lazy()`:
      ! `...` must be empty.
      x Problematic argument:
      * src = simulate_sqlite()

# cannot convert tbl_lazy to data.frame

    Code
      as.data.frame(tbl_lazy(mtcars, con = simulate_sqlite()))
    Condition
      Error in `as.data.frame()`:
      ! Can not coerce <tbl_lazy> to <data.frame>

# has print method

    Code
      tbl_lazy(mtcars)
    Output
      <SQL>
      SELECT *
      FROM `df`

# names() inform that they aren't meant to be used

    Code
      names(lazy_frame(x = 1))
    Message
      ! The `names()` method of <tbl_lazy> is for internal use only.
      i Did you mean `colnames()`?
    Output
      [1] "lazy_query" "src"       

# $ aborts when not used with src or lazy_query

    The `$` method of <tbl_lazy> is for internal use only.
    i Use `dplyr::pull()` to get the values in a column.

# sql_substr works as expected

    Code
      translate_sql(substr("test"), con = con)
    Condition
      Error in `substr()`:
      ! `start` must be a whole number, not absent.

---

    Code
      translate_sql(substr("test", 0), con = con)
    Condition
      Error in `substr()`:
      ! `stop` must be a whole number, not absent.

---

    Code
      translate_sql(substr("test", "x", 1), con = con)
    Condition
      Error in `substr()`:
      ! `start` must be a whole number, not the string "x".

---

    Code
      translate_sql(substr("test", 1, "x"), con = con)
    Condition
      Error in `substr()`:
      ! `stop` must be a whole number, not the string "x".

# generates valid sql for all directions

    Code
      tidyr::fill(lf_asc, n1, .direction = "up")
    Output
      <SQL>
      SELECT
        "id",
        "group",
        LAST_VALUE("n1" IGNORE NULLS) OVER (ORDER BY "id" DESC ROWS UNBOUNDED PRECEDING) AS "n1"
      FROM "df"
    Code
      tidyr::fill(lf_desc, n1, .direction = "up")
    Output
      <SQL>
      SELECT
        "id",
        "group",
        LAST_VALUE("n1" IGNORE NULLS) OVER (ORDER BY "id" ROWS UNBOUNDED PRECEDING) AS "n1"
      FROM "df"
    Code
      tidyr::fill(lf_asc, n1, .direction = "updown")
    Output
      <SQL>
      SELECT
        "id",
        "group",
        LAST_VALUE("n1" IGNORE NULLS) OVER (ORDER BY "id" ROWS UNBOUNDED PRECEDING) AS "n1"
      FROM (
        SELECT
          "id",
          "group",
          LAST_VALUE("n1" IGNORE NULLS) OVER (ORDER BY "id" DESC ROWS UNBOUNDED PRECEDING) AS "n1"
        FROM "df"
      ) AS "q01"
    Code
      tidyr::fill(lf_asc, n1, .direction = "downup")
    Output
      <SQL>
      SELECT
        "id",
        "group",
        LAST_VALUE("n1" IGNORE NULLS) OVER (ORDER BY "id" DESC ROWS UNBOUNDED PRECEDING) AS "n1"
      FROM (
        SELECT
          "id",
          "group",
          LAST_VALUE("n1" IGNORE NULLS) OVER (ORDER BY "id" ROWS UNBOUNDED PRECEDING) AS "n1"
        FROM "df"
      ) AS "q01"

# fill() respects grouping

    Code
      tidyr::fill(window_order(group_by(lf, group), id), n1)
    Output
      <SQL>
      SELECT
        "id",
        "group",
        LAST_VALUE("n1" IGNORE NULLS) OVER (PARTITION BY "group" ORDER BY "id" ROWS UNBOUNDED PRECEDING) AS "n1"
      FROM "df"

# can generate variant SQL

    Code
      tidyr::fill(window_order(lf, id), n1)
    Output
      <SQL>
      SELECT
        `id`,
        `group`,
        MAX(`n1`) OVER (PARTITION BY `..dbplyr_partition_1`) AS `n1`
      FROM (
        SELECT
          *,
          SUM(CASE WHEN ((`n1` IS NULL)) THEN 0 ELSE 1 END) OVER (ORDER BY `id` ROWS UNBOUNDED PRECEDING) AS `..dbplyr_partition_1`
        FROM `df`
      ) AS `q01`
    Code
      tidyr::fill(window_order(lf, desc(id)), n1)
    Output
      <SQL>
      SELECT
        `id`,
        `group`,
        MAX(`n1`) OVER (PARTITION BY `..dbplyr_partition_1`) AS `n1`
      FROM (
        SELECT
          *,
          SUM(CASE WHEN ((`n1` IS NULL)) THEN 0 ELSE 1 END) OVER (ORDER BY `id` DESC ROWS UNBOUNDED PRECEDING) AS `..dbplyr_partition_1`
        FROM `df`
      ) AS `q01`
    Code
      tidyr::fill(window_order(group_by(lf, group), id), n1)
    Output
      <SQL>
      SELECT
        `id`,
        `group`,
        MAX(`n1`) OVER (PARTITION BY `group`, `..dbplyr_partition_1`) AS `n1`
      FROM (
        SELECT
          *,
          SUM(CASE WHEN ((`n1` IS NULL)) THEN 0 ELSE 1 END) OVER (PARTITION BY `group` ORDER BY `id` ROWS UNBOUNDED PRECEDING) AS `..dbplyr_partition_1`
        FROM `df`
      ) AS `q01`

# fill errors on unsorted data

    Code
      tidyr::fill(df)
    Condition
      Error in `tidyr::fill()`:
      x `.data` does not have explicit order.
      i Please use `dbplyr::window_order()` to make order explicit.

# fill() errors on attempted rename

    Code
      tidyr::fill(lf, y = x)
    Condition
      Error in `tidyr::fill()`:
      ! Arguments in `...` must be passed by position, not name.
      x Problematic argument:
      * y = x

# fill() produces nice error messages

    Code
      tidyr::fill(lf, non_existent)
    Condition
      Error in `tidyr::fill()`:
      ! Can't select columns that don't exist.
      x Column `non_existent` doesn't exist.

# dplyr.strict_sql = TRUE prevents auto conversion

    Code
      translate_sql(blah(x), con = con)
    Condition
      Error in `blah()`:
      ! Don't know how to translate `blah()`
    Code
      translate_sql(x %blah% y, con = con)
    Condition
      Error in `x %blah% y`:
      ! Don't know how to translate `%blah%`

# namespace calls are translated

    Code
      translate_sql(NOSUCHPACKAGE::foo(), con = con)
    Condition
      Error:
      ! There is no package called NOSUCHPACKAGE
    Code
      translate_sql(dbplyr::NOSUCHFUNCTION(), con = con)
    Condition
      Error:
      ! "NOSUCHFUNCTION" is not an exported object from dbplyr
    Code
      translate_sql(base::abbreviate(x), con = con)
    Condition
      Error in `base::abbreviate()`:
      ! No known SQL translation

---

    Code
      mutate(lz, x = NOSUCHPACKAGE::foo())
    Condition
      Error:
      ! There is no package called NOSUCHPACKAGE
    Code
      mutate(lz, x = dbplyr::NOSUCHFUNCTION())
    Condition
      Error:
      ! "NOSUCHFUNCTION" is not an exported object from dbplyr
    Code
      mutate(lz, x = base::abbreviate(x))
    Condition
      Error in `base::abbreviate()`:
      ! No known SQL translation

# as() translates to CAST

    Code
      translate_sql(as(x, 1L), con = con)
    Condition
      Error in `as()`:
      ! `type` must be a single string, not the number 1.

# check_sql() checks inputs

    Code
      f("x")
    Condition
      Error in `f()`:
      ! `x` must be a <sql>, not the string "x".

# check_sql() optionally checks names

    Code
      f(sql(a = "x"))
    Condition
      Error in `f()`:
      ! `x` must be an unnamed <sql>, not a <sql> object.

# check_sql() optionally allows NULL

    Code
      f("x")
    Condition
      Error in `f()`:
      ! `x` must be a <sql> or `NULL`, not the string "x".

# old arguments are defunct

    Code
      partial_eval(quote(x), vars = c("x", "y"))
    Condition
      Error:
      ! The `vars` argument of `partial_eval()` was deprecated in dbplyr 2.1.2 and is now defunct.
    Code
      partial_eval(quote(x), data = c("x", "y"))
    Condition
      Error:
      ! The `data` argument of `partial_eval()` must be a lazy frame as of dbplyr 2.1.2.

# new_sql_dialect() validates inputs

    Code
      new_sql_dialect(1)
    Condition
      Error in `new_sql_dialect()`:
      ! `dialect` must be a single string, not the number 1.
    Code
      new_sql_dialect("x", quote_identifier = "a function")
    Condition
      Error in `new_sql_dialect()`:
      ! `quote_identifier` must be a function, not the string "a function".
    Code
      new_sql_dialect("x", quote_identifier = function(x) x, has_window_clause = "yes")
    Condition
      Error in `new_sql_dialect()`:
      ! `has_window_clause` must be `TRUE` or `FALSE`, not the string "yes".

# print method shows class

    Code
      d
    Output
      <sql_dialect_test>

# uses FETCH FIRST instead of LIMIT

    Code
      head(mf)
    Output
      <SQL>
      SELECT *
      FROM "df"
      FETCH FIRST 6 ROWS ONLY

---

    Code
      head(mf, 10)
    Output
      <SQL>
      SELECT *
      FROM "df"
      FETCH FIRST 10 ROWS ONLY

# queries with WHERE and ORDER BY work

    Code
      head(arrange(filter(mf, x > 0), y), 5)
    Output
      <SQL>
      SELECT *
      FROM "df"
      WHERE ("x" > 0.0)
      ORDER BY "y"
      FETCH FIRST 5 ROWS ONLY

# custom lubridate functions translated correctly

    Code
      translate_sql(wday(x), con = con)
    Output
      <SQL> MOD(DAYOFWEEK("x") + 0 - 1, 7) + 1

---

    Code
      translate_sql(quarter(x, fiscal_start = 2), con = con)
    Condition
      Error in `quarter()`:
      ! `fiscal_start = 2` isn't supported in DB2 translation.
      i It must be 1 instead.

# difftime is translated correctly

    Code
      translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
    Condition
      Error in `difftime()`:
      ! `units = "auto"` isn't supported on database backends.
      i It must be "days" instead.

# sql_table_analyze uses RUNSTATS

    Code
      sql_table_analyze(con, ident("mytable"))
    Output
      <SQL> RUNSTATS ON TABLE "mytable" WITH DISTRIBUTION AND DETAILED INDEXES ALL

# copy_inline uses VALUES with column aliases

    Code
      remote_query(copy_inline(con, df))
    Output
      <SQL> SELECT CAST("x" AS INTEGER) AS "x", CAST("y" AS VARCHAR(255)) AS "y"
      FROM (  VALUES (1, 'a'), (2, 'b')) AS drvd("x", "y")

# pasting translated correctly

    Code
      translate_sql(paste0(x, collapse = ""), con = con, window = FALSE)
    Condition
      Error in `check_collapse()`:
      ! `collapse` not supported in DB translation of `paste()`.
      i Please use `str_flatten()` instead.

# custom lubridate functions translated correctly

    Code
      translate_sql(quarter(x, fiscal_start = 2), con = con)
    Condition
      Error in `quarter()`:
      ! `fiscal_start = 2` isn't supported in PostgreSQL translation.
      i It must be 1 instead.

# custom clock functions translated correctly

    Code
      translate_sql(date_count_between(date_column_1, date_column_2, "year"), con = con)
    Condition
      Error in `date_count_between()`:
      ! `precision = "year"` isn't supported on database backends.
      i It must be "day" instead.

---

    Code
      translate_sql(date_count_between(date_column_1, date_column_2, "day", n = 5),
      con = con)
    Condition
      Error in `date_count_between()`:
      ! `n = 5` isn't supported on database backends.
      i It must be 1 instead.

# difftime is translated correctly

    Code
      translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
    Condition
      Error in `difftime()`:
      ! `units = "auto"` isn't supported on database backends.
      i It must be "days" instead.

---

    Code
      translate_sql(difftime(start_date, end_date, tz = "UTC", units = "days"), con = con)
    Condition
      Error in `difftime()`:
      ! Argument `tz` isn't supported on database backends.

# custom window functions translated correctly

    Code
      translate_sql(quantile(x, 0.3, na.rm = TRUE), con = con, window = TRUE)
    Condition
      Error in `quantile()`:
      ! Translation of `quantile()` in `mutate()` is not supported for PostgreSQL.
      i Use a combination of `summarise()` and `left_join()` instead:
        `df |> left_join(summarise(<col> = quantile(x, 0.3, na.rm = TRUE)))`.

---

    Code
      translate_sql(median(x, na.rm = TRUE), con = con, window = TRUE)
    Condition
      Error in `median()`:
      ! Translation of `median()` in `mutate()` is not supported for PostgreSQL.
      i Use a combination of `summarise()` and `left_join()` instead:
        `df |> left_join(summarise(<col> = median(x, na.rm = TRUE)))`.

# custom SQL translation

    Code
      left_join(lf, lf, by = "x", na_matches = "na")
    Output
      <SQL>
      SELECT "df_LHS"."x" AS "x"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON (("df_LHS"."x") IS NOT DISTINCT FROM ("df_RHS"."x"))

---

    Code
      remote_query(copy_inline(con, tibble(x = integer(), y = character())))
    Output
      <SQL> SELECT CAST(NULL AS INTEGER) AS "x", CAST(NULL AS TEXT) AS "y"
      WHERE (0 = 1)

---

    Code
      remote_query(copy_inline(con, tibble(x = 1:2, y = letters[1:2])))
    Output
      <SQL> SELECT CAST("x" AS INTEGER) AS "x", CAST("y" AS TEXT) AS "y"
      FROM (  VALUES (1, 'a'), (2, 'b')) AS drvd("x", "y")

# `sql_query_insert()` works

    Code
      (sql_query_insert(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), insert_cols = colnames(df_y), by = c("a", "b"), conflict = "error",
      returning_cols = c("a", b2 = "b")))
    Condition
      Error in `sql_query_insert()`:
      ! `conflict = "error"` isn't supported on database backends.
      i It must be "ignore" instead.

---

    Code
      sql_query_insert(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), insert_cols = colnames(df_y), by = c("a", "b"), conflict = "ignore",
      returning_cols = c("a", b2 = "b"))
    Output
      <SQL> INSERT INTO "df_x" ("a", "b", "c", "d")
      SELECT *
      FROM (
        SELECT "a", "b", "c" + 1.0 AS "c", "d"
        FROM "df_y"
      ) AS "...y"
      ON CONFLICT ("a", "b")
      DO NOTHING
      RETURNING "df_x"."a", "df_x"."b" AS "b2"

# `sql_query_upsert()` with method = 'on_conflict' is correct

    Code
      sql_query_upsert(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), by = c("c", "d"), update_cols = c("a", "b"), returning_cols = c("a",
        b2 = "b"), method = "on_conflict")
    Output
      <SQL> INSERT INTO "df_x" ("c", "d", "a", "b")
      SELECT "c", "d", "a", "b"
      FROM (
        SELECT "a", "b", "c" + 1.0 AS "c", "d"
        FROM "df_y"
      ) AS "...y"
      WHERE true
      ON CONFLICT ("c", "d")
      DO UPDATE
      SET "a" = "excluded"."a", "b" = "excluded"."b"
      RETURNING "df_x"."a", "df_x"."b" AS "b2"

# can explain

    Code
      explain(mutate(db, y = x + 1))
    Output
      <SQL>
      SELECT *, "x" + 1.0 AS "y"
      FROM "test"
      
      <PLAN>
                                                 QUERY PLAN
      1 Seq Scan on test  (cost=0.00..1.04 rows=3 width=36)

# can insert with returning

    Code
      rows_insert(x, y, by = c("a", "b"), in_place = TRUE, conflict = "ignore",
      returning = everything(), method = "on_conflict")
    Condition
      Error in `rows_insert()`:
      ! Can't modify database table "df_x".
      i Using SQL: INSERT INTO "df_x" ("a", "b", "c", "d") SELECT * FROM ( SELECT "a", "b", "c" + 1.0 AS "c", "d" FROM "df_y" ) AS "...y" ON CONFLICT ("a", "b") DO NOTHING RETURNING "df_x"."a", "df_x"."b", "df_x"."c", "df_x"."d"
      Caused by error:
      ! dummy DBI error

# can upsert with returning

    Code
      rows_upsert(x, y, by = c("a", "b"), in_place = TRUE, returning = everything(),
      method = "on_conflict")
    Condition
      Error in `rows_upsert()`:
      ! Can't modify database table "df_x".
      i Using SQL: INSERT INTO "df_x" ("a", "b", "c", "d") SELECT "a", "b", "c", "d" FROM ( SELECT "a", "b", "c" + 1.0 AS "c", "d" FROM "df_y" ) AS "...y" WHERE true ON CONFLICT ("a", "b") DO UPDATE SET "c" = "excluded"."c", "d" = "excluded"."d" RETURNING "df_x"."a", "df_x"."b", "df_x"."c", "df_x"."d"
      Caused by error:
      ! dummy DBI error

# generates custom sql

    Code
      sql_table_analyze(con, in_schema("schema", "tbl"))
    Output
      <SQL> COLLECT STATISTICS "schema"."tbl"

# head translated to TOP

    Code
      sql_render(head(mf))
    Output
      <SQL> SELECT TOP 6 "df".*
      FROM "df"

# lead, lag work

    Code
      sql_render(mutate(group_by(mf, y), val2 = lead(x, order_by = x)))
    Output
      <SQL> SELECT "df".*, LEAD("x", 1, NULL) OVER (PARTITION BY "y" ORDER BY "x") AS "val2"
      FROM "df"

---

    Code
      sql_render(mutate(group_by(mf, y), val2 = lag(x, order_by = x)))
    Output
      <SQL> SELECT "df".*, LAG("x", 1, NULL) OVER (PARTITION BY "y" ORDER BY "x") AS "val2"
      FROM "df"

# weighted.mean

    Code
      summarise(mf, wt_mean = weighted.mean(x, y))
    Output
      <SQL>
      SELECT SUM(("x" * "y")) / SUM("y") OVER () AS "wt_mean"
      FROM "df"

# row_number() with and without group_by() and arrange(): unordered defaults to Ordering by NULL (per empty_order)

    Code
      mutate(mf, rown = row_number())
    Output
      <SQL>
      SELECT "df".*, ROW_NUMBER() OVER (ORDER BY (SELECT NULL)) AS "rown"
      FROM "df"

---

    Code
      mutate(group_by(mf, y), rown = row_number())
    Output
      <SQL>
      SELECT
        "df".*,
        ROW_NUMBER() OVER (PARTITION BY "y" ORDER BY (SELECT NULL)) AS "rown"
      FROM "df"

---

    Code
      mutate(arrange(mf, y), rown = row_number())
    Output
      <SQL>
      SELECT "df".*, ROW_NUMBER() OVER (ORDER BY "y") AS "rown"
      FROM "df"
      ORDER BY "y"

# head after distinct() produces subquery

    Code
      head(distinct(lf))
    Output
      <SQL>
      SELECT TOP 6 "q01".*
      FROM (
        SELECT DISTINCT "df".*
        FROM "df"
      ) AS "q01"

# custom scalar translated correctly

    Code
      translate_sql(paste(x, collapse = "-"), con = con)
    Condition
      Error in `check_collapse()`:
      ! `collapse` not supported in DB translation of `paste()`.
      i Please use `str_flatten()` instead.

# queries translate correctly

    Code
      head(mf)
    Output
      <SQL>
      SELECT TOP 6 *
      FROM "df"

# multiple joins use parens #1576

    Code
      inner_join(left_join(lf1, lf2, by = "x"), lf3, by = "x")
    Output
      <SQL>
      SELECT "lf1".*, "b", "c"
      FROM (("lf1"
      LEFT JOIN "lf2"
      ON "lf1"."x" = "lf2"."x")
      INNER JOIN "lf3"
      ON "lf1"."x" = "lf3"."x")

---

    Code
      left_join(inner_join(left_join(lf1, lf2, by = "x"), lf3, by = "x"), lf4, by = "x")
    Output
      <SQL>
      SELECT "lf1".*, "b", "c", "d"
      FROM ((("lf1"
      LEFT JOIN "lf2"
      ON "lf1"."x" = "lf2"."x")
      INNER JOIN "lf3"
      ON "lf1"."x" = "lf3"."x")
      LEFT JOIN "lf4"
      ON "lf1"."x" = "lf4"."x")

# sql_prefix checks arguments

    Code
      sin_db(sin(1, 2))
    Condition
      Error in `sin()`:
      ! 2 arguments passed to 'sin' which requires 1

---

    Code
      sin_db(sin(a = 1))
    Condition
      Error in `sin()`:
      ! supplied argument name 'a' does not match 'x'

# runif is translated

    Code
      translate_sql(runif(2), con = con)
    Condition
      Error in `sql_runif()`:
      ! Only `n = n()` is supported.

# lazy_semi_join_query() checks arguments

    Code
      (my_lazy_semi_join_query(x = lazy_frame(x = 1)))
    Condition
      Error in `my_lazy_semi_join_query()`:
      ! `x` must be a lazy query, not a <tbl_TestConnection> object.
    Code
      (my_lazy_semi_join_query(y = lazy_frame(x = 1)))
    Condition
      Error in `my_lazy_semi_join_query()`:
      ! `y` must be a lazy query, not a <tbl_TestConnection> object.

---

    Code
      (my_lazy_semi_join_query(by = lmod(by0, x = 1)))
    Condition
      Error in `my_lazy_semi_join_query()`:
      ! `by$x` must be a character vector, not the number 1.

---

    Code
      (my_lazy_semi_join_query(anti = NA))
    Condition
      Error in `my_lazy_semi_join_query()`:
      ! `anti` must be `TRUE` or `FALSE`, not `NA`.

# generated sql doesn't change unexpectedly

    Code
      semi_join(lf, lf)
    Message
      Joining with `by = join_by(x, y)`
    Output
      <SQL>
      SELECT "df_LHS".*
      FROM "df" AS "df_LHS"
      WHERE EXISTS (
        SELECT 1 FROM "df" AS "df_RHS"
        WHERE ("df_LHS"."x" = "df_RHS"."x") AND ("df_LHS"."y" = "df_RHS"."y")
      )

---

    Code
      anti_join(lf, lf)
    Message
      Joining with `by = join_by(x, y)`
    Output
      <SQL>
      SELECT "df_LHS".*
      FROM "df" AS "df_LHS"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df" AS "df_RHS"
        WHERE ("df_LHS"."x" = "df_RHS"."x") AND ("df_LHS"."y" = "df_RHS"."y")
      )

# sql_on query doesn't change unexpectedly

    Code
      semi_join(lf1, lf2, sql_on = "LHS.y < RHS.z")
    Output
      <SQL>
      SELECT "LHS".*
      FROM "df" AS "LHS"
      WHERE EXISTS (
        SELECT 1 FROM "df" AS "RHS"
        WHERE (LHS.y < RHS.z)
      )

---

    Code
      anti_join(lf1, lf2, sql_on = "LHS.y < RHS.z")
    Output
      <SQL>
      SELECT "LHS".*
      FROM "df" AS "LHS"
      WHERE NOT EXISTS (
        SELECT 1 FROM "df" AS "RHS"
        WHERE (LHS.y < RHS.z)
      )

# reframe is not supported

    Code
      reframe(lazy_frame(x = 1))
    Condition
      Error in `reframe()`:
      ! `reframe()` is not supported on database backends.

# can performed grouped summarise with no inputs

    Code
      show_query(out)
    Output
      <SQL>
      SELECT `x`
      FROM `empty-summarise`
      GROUP BY `x`

# generates minimal sql when possible

    Code
      summarise(arrange(lf, x), y = mean(x))
    Output
      <SQL>
      SELECT AVG("x") AS "y"
      FROM "df"
      ORDER BY "x"
    Code
      summarise(filter(lf, x < 1), y = mean(x))
    Output
      <SQL>
      SELECT AVG("x") AS "y"
      FROM "df"
      WHERE ("x" < 1.0)

# can't refer to freshly created variables

    Code
      (expect_error(summarise(lf1, a_sum = sum(a), issue_col = sum(a_sum))))
    Output
      <error/rlang_error>
      Error in `summarise()`:
      i In argument: `issue_col = sum(a_sum)`
      Caused by error:
      ! In dbplyr you cannot use a variable created in the same `summarise()`.
      x `a_sum` was created earlier in this `summarise()`.
      i You need an extra `mutate()` step to use it.
    Code
      (expect_error(summarise(lf1, across(c(a, b), list(sum = sum)), issue_col = sum(
        a_sum))))
    Output
      <error/rlang_error>
      Error in `summarise()`:
      i In argument: `issue_col = sum(a_sum)`
      Caused by error:
      ! In dbplyr you cannot use a variable created in the same `summarise()`.
      x `a_sum` was created earlier in this `summarise()`.
      i You need an extra `mutate()` step to use it.
    Code
      (expect_error(summarise(lf1, across(c(a, b), list(sum = sum)), issue_col = sum(
        b_sum))))
    Output
      <error/rlang_error>
      Error in `summarise()`:
      i In argument: `issue_col = sum(b_sum)`
      Caused by error:
      ! In dbplyr you cannot use a variable created in the same `summarise()`.
      x `b_sum` was created earlier in this `summarise()`.
      i You need an extra `mutate()` step to use it.
    Code
      (expect_error(summarise(lf1, a_sum = sum(a), issue_col = across(a_sum, sum))))
    Output
      <error/rlang_error>
      Error in `summarise()`:
      i In argument: `issue_col = across(a_sum, sum)`
      Caused by error:
      ! In dbplyr you cannot use a variable created in the same `summarise()`.
      x `a_sum` was created earlier in this `summarise()`.
      i You need an extra `mutate()` step to use it.

# summarise(.groups=)

    Code
      summarise(df, .groups = "rowwise")
    Condition
      Error in `summarise()`:
      ! `.groups` can't be "rowwise" in dbplyr
      i Possible values are NULL (default), "drop_last", "drop", and "keep"

# summarise produces informative message about grouping

    Code
      . <- summarise(lf)
    Message
      ! Grouped output by "x".
      i Override behaviour and silence this message with the `.groups` argument.
      i Or use `.by` instead of `group_by()`.

# summarise can modify grouping variables

    Code
      (result1 <- summarise(group_by(lf, g), g = g + 1))
    Output
      <SQL>
      SELECT "g" + 1.0 AS "g"
      FROM "df"
      GROUP BY "g"

---

    Code
      (result2 <- summarise(group_by(lf, g), x = x + 1, g = g + 1))
    Output
      <SQL>
      SELECT "g" + 1.0 AS "g", "x" + 1.0 AS "x"
      FROM "df"
      GROUP BY "g"

# across() does not select grouping variables

    Code
      summarise(group_by(df, g), across(.fns = ~0))
    Output
      <SQL>
      SELECT "g", 0.0 AS "x"
      FROM "df"
      GROUP BY "g"

# across doesn't select columns from `.by` #1493

    Code
      out
    Output
      <SQL>
      SELECT "g", SUM("x") AS "x"
      FROM "df"
      GROUP BY "g"

# can't use `.by` with `.groups`

    Code
      summarise(df, .by = x, .groups = "drop")
    Condition
      Error in `summarise()`:
      ! Can't supply both `.by` and `.groups`.

# catches `.by` with grouped-df

    Code
      summarise(gdf, .by = x)
    Condition
      Error:
      ! Can't supply `.by` when `.data` is a grouped data frame.

# quoting for rendering summarized grouped table

    Code
      sql_render(out)
    Output
      <SQL> SELECT `x`, COUNT(*) AS `n`
      FROM `verb-summarise`
      GROUP BY `x`

# two filters equivalent to one

    Code
      show_query(df1)
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`x` > 3.0) AND (`y` < 3.0)

---

    Code
      show_query(df2)
    Output
      <SQL>
      SELECT `x`, `y`, `id`
      FROM (
        SELECT *, AVG(`x`) OVER () AS `col01`
        FROM `df`
      ) AS `q01`
      WHERE (`col01` > 2.0) AND (`y` < 3.0)

# errors for named input

    Code
      filter(lf, x = 1)
    Condition
      Error in `filter()`:
      ! Problem with `filter()` input `..1`.
      x Input `..1` is named.
      i This usually means that you've used `=` instead of `==`.
      i Did you mean `x == 1`?

---

    Code
      filter(lf, y > 1, x = 1)
    Condition
      Error in `filter()`:
      ! Problem with `filter()` input `..2`.
      x Input `..2` is named.
      i This usually means that you've used `=` instead of `==`.
      i Did you mean `x == 1`?

# .preserve is not supported

    Code
      filter(lf, x == 1, .preserve = TRUE)
    Condition
      Error in `filter()`:
      ! `.preserve = TRUE` isn't supported on database backends.
      i It must be FALSE instead.

# catches `.by` with grouped-df

    Code
      filter(gdf, .by = x)
    Condition
      Error:
      ! Can't supply `.by` when `.data` is a grouped data frame.

# filter() can use window function and external vector - #1048

    Code
      filter(lazy_frame(x = 1L), x == max(x, na.rm = T), x %in% to_filter)
    Output
      <SQL>
      SELECT "x"
      FROM (
        SELECT *, MAX("x") OVER () AS "col01"
        FROM "df"
      ) AS "q01"
      WHERE ("x" = "col01") AND ("x" IN (1, 2))

# filter() after summarise() uses `HAVING`

    Code
      (out <- filter(lf, g == 1))
    Output
      <SQL>
      SELECT "g", "h", AVG("x") AS "x_mean"
      FROM "df"
      GROUP BY "g", "h"
      HAVING ("g" = 1.0)

---

    Code
      (out <- filter(lf, x_mean > 1))
    Output
      <SQL>
      SELECT "g", "h", AVG("x") AS "x_mean"
      FROM "df"
      GROUP BY "g", "h"
      HAVING (AVG("x") > 1.0)

---

    Code
      (out <- filter(filter(lf, g == 1), g == 2))
    Output
      <SQL>
      SELECT "g", "h", AVG("x") AS "x_mean"
      FROM "df"
      GROUP BY "g", "h"
      HAVING ("g" = 1.0) AND ("g" = 2.0)

---

    Code
      (out <- filter(filter(lf, g == 1), h == 2))
    Output
      <SQL>
      SELECT "g", "h", AVG("x") AS "x_mean"
      FROM "df"
      GROUP BY "g", "h"
      HAVING ("g" = 1.0) AND ("h" = 2.0)

# `HAVING` supports expressions #1128

    Code
      filter(summarise(lf, x_sum = sum(x, na.rm = TRUE)), !is.na(x_sum))
    Output
      <SQL>
      SELECT SUM("x") AS "x_sum"
      FROM "df"
      HAVING (NOT(((SUM("x")) IS NULL)))

# filter() after mutate() does not use `HAVING`

    Code
      (out <- filter(lf, x_mean > 1))
    Output
      <SQL>
      SELECT *
      FROM (
        SELECT *, AVG("x") OVER (PARTITION BY "g", "h") AS "x_mean"
        FROM "df"
      ) AS "q01"
      WHERE ("x_mean" > 1.0)

# filter() using a window function after summarise() does not use `HAVING`

    Code
      (out <- filter(lf, cumsum(x_mean) == 1))
    Condition
      Warning:
      Windowed expression `SUM("x_mean")` does not have explicit order.
      i Please use `arrange()`, `window_order()`, or `.order` to make deterministic.
    Output
      <SQL>
      SELECT "g", "h", "x_mean"
      FROM (
        SELECT
          *,
          SUM("x_mean") OVER (PARTITION BY "g" ROWS UNBOUNDED PRECEDING) AS "col01"
        FROM (
          SELECT "g", "h", AVG("x") AS "x_mean"
          FROM "df"
          GROUP BY "g", "h"
        ) AS "q01"
      ) AS "q01"
      WHERE ("col01" = 1.0)

# filter_out uses backend-aware IS DISTINCT FROM TRUE

    Code
      filter_out_backend(simulate_postgres())
    Output
      <SQL>
      SELECT *
      FROM "df"
      WHERE (("x" = 1.0 AND "y" = 3.0) IS DISTINCT FROM (TRUE))
    Code
      filter_out_backend(simulate_sqlite())
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE ((`x` = 1.0 AND `y` = 3.0) IS NOT (1))
    Code
      filter_out_backend(simulate_dbi())
    Output
      <SQL>
      SELECT *
      FROM "df"
      WHERE (CASE WHEN (("x" = 1.0 AND "y" = 3.0) = (TRUE)) OR (("x" = 1.0 AND "y" = 3.0) IS NULL AND (TRUE) IS NULL) THEN 0 ELSE 1 END = 1)

# filter_out errors for named input

    Code
      filter_out(lf, x = 1)
    Condition
      Error in `filter_out()`:
      ! Problem with `filter_out()` input `..1`.
      x Input `..1` is named.
      i This usually means that you've used `=` instead of `==`.
      i Did you mean `x == 1`?

# filter_out errors for .preserve = TRUE

    Code
      filter_out(lf, x == 1, .preserve = TRUE)
    Condition
      Error in `filter_out()`:
      ! `.preserve = TRUE` isn't supported on database backends.
      i It must be FALSE instead.

# build_sql() requires connection

    Code
      build_sql("SELECT * FROM ", x)
    Condition
      Error in `build_sql()`:
      ! `con` must not be NULL.

# as.sql() is deprecated

    Code
      as.sql(ident("x"))
    Condition
      Warning:
      `as.sql()` was deprecated in dbplyr 2.6.0.
      i Please use `as_table_path()` instead.
    Output
      <IDENT> x

# str_detect returns bit in SELECT context on SQL Server 2025

    Code
      mutate(lf, detected = str_detect(x, "abc"))
    Output
      <SQL>
      SELECT
        *,
        CAST(CASE WHEN REGEXP_LIKE([x], 'abc') THEN 1 ELSE 0 END AS BIT) AS [detected]
      FROM [df]

---

    Code
      filter(lf, str_detect(x, "abc"))
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE (REGEXP_LIKE([x], 'abc'))

# custom aggregators translated correctly

    Code
      translate_sql(quantile(x, 0.5, na.rm = TRUE), con = con, window = FALSE)
    Condition
      Error in `quantile()`:
      ! Translation of `quantile()` in `summarise()` is not supported for SQL Server.
      i Use a combination of `distinct()` and `mutate()` for the same result:
        `mutate(<col> = quantile(x, 0.5, na.rm = TRUE)) |> distinct(<col>)`
    Code
      translate_sql(median(x, na.rm = TRUE), con = con, window = FALSE)
    Condition
      Error in `median()`:
      ! Translation of `median()` in `summarise()` is not supported for SQL Server.
      i Use a combination of `distinct()` and `mutate()` for the same result:
        `mutate(<col> = median(x, na.rm = TRUE)) |> distinct(<col>)`

# custom window functions translated correctly

    Code
      translate_sql(n_distinct(x), con = con, vars_group = "x")
    Condition
      Error in `n_distinct()`:
      ! No translation available in `mutate()`/`filter()` for SQL server.

# custom lubridate functions translated correctly

    Code
      translate_sql(month(x, label = TRUE, abbr = TRUE), con = con)
    Condition
      Error in `month()`:
      ! `abbr = TRUE` isn't supported in SQL Server translation.
      i It must be FALSE instead.

---

    Code
      translate_sql(quarter(x, fiscal_start = 5), con = con)
    Condition
      Error in `quarter()`:
      ! `fiscal_start = 5` isn't supported in SQL Server translation.
      i It must be 1 instead.

# custom clock functions translated correctly

    Code
      translate_sql(date_count_between(date_column_1, date_column_2, "year"), con = con)
    Condition
      Error in `date_count_between()`:
      ! `precision = "year"` isn't supported on database backends.
      i It must be "day" instead.

---

    Code
      translate_sql(date_count_between(date_column_1, date_column_2, "day", n = 5),
      con = con)
    Condition
      Error in `date_count_between()`:
      ! `n = 5` isn't supported on database backends.
      i It must be 1 instead.

# difftime is translated correctly

    Code
      translate_sql(difftime(start_date, end_date, units = "auto"), con = con)
    Condition
      Error in `difftime()`:
      ! `units = "auto"` isn't supported on database backends.
      i It must be "days" instead.

---

    Code
      translate_sql(difftime(start_date, end_date, tz = "UTC", units = "days"), con = con)
    Condition
      Error in `difftime()`:
      ! Argument `tz` isn't supported on database backends.

# convert between bit and boolean as needed

    Code
      filter(mf, is.na(x))
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE (([x] IS NULL))

---

    Code
      filter(mf, !is.na(x))
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE (NOT(([x] IS NULL)))

---

    Code
      filter(mf, x == 1L || x == 2L)
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE ([x] = 1 OR [x] = 2)

---

    Code
      mutate(mf, z = ifelse(x == 1L, 1L, 2L))
    Output
      <SQL>
      SELECT *, CASE WHEN ([x] = 1) THEN 1 WHEN NOT ([x] = 1) THEN 2 END AS [z]
      FROM [df]

---

    Code
      mutate(mf, z = case_when(x == 1L ~ 1L))
    Output
      <SQL>
      SELECT *, CASE WHEN ([x] = 1) THEN 1 END AS [z]
      FROM [df]

---

    Code
      mutate(mf, z = !is.na(x))
    Output
      <SQL>
      SELECT *, ~CAST(CASE WHEN ([x] IS NULL) THEN 1 ELSE 0 END AS BIT) AS [z]
      FROM [df]

---

    Code
      mutate(mf, x = x == 1L)
    Output
      <SQL>
      SELECT CAST(CASE WHEN [x] = 1 THEN 1 ELSE 0 END AS BIT) AS [x]
      FROM [df]

---

    Code
      mutate(mf, x = x == 1L || x == 2L)
    Output
      <SQL>
      SELECT CAST(CASE WHEN [x] = 1 OR [x] = 2 THEN 1 ELSE 0 END AS BIT) AS [x]
      FROM [df]

---

    Code
      mutate(mf, x = x == 1L || x == 2L || x == 3L)
    Output
      <SQL>
      SELECT CAST(CASE WHEN [x] = 1 OR [x] = 2 OR [x] = 3 THEN 1 ELSE 0 END AS BIT) AS [x]
      FROM [df]

---

    Code
      mutate(mf, x = !(x == 1L || x == 2L || x == 3L))
    Output
      <SQL>
      SELECT ~CAST(CASE WHEN ([x] = 1 OR [x] = 2 OR [x] = 3) THEN 1 ELSE 0 END AS BIT) AS [x]
      FROM [df]

# handles ORDER BY in subqueries

    Code
      sql_query_select(dialect_mssql(), sql("[x]"), sql("[y]"), order_by = "z",
      subquery = TRUE)
    Condition
      Warning:
      ORDER BY is ignored in subqueries without LIMIT
      i Do you need to move arrange() later in the pipeline or use window_order() instead?
    Output
      <SQL> SELECT [x]
      FROM [y]

# custom limit translation

    Code
      sql_query_select(dialect_mssql(), sql("[x]"), sql("[y]"), order_by = sql("[z]"),
      limit = 10)
    Output
      <SQL> SELECT TOP 10 [x]
      FROM [y]
      ORDER BY [z]

# custom escapes translated correctly

    Code
      filter(mf, x == a)
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE ([x] = 0x616263)

---

    Code
      filter(mf, x %in% L)
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE ([x] IN (0x616263, 0x0102))

---

    Code
      qry
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE ([x] IN (0x616263, 0x0102))

# logical escaping to 0/1 for both filter() and mutate()

    Code
      filter(mf, x == TRUE)
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE ([x] = 1)

---

    Code
      mutate(mf, x = TRUE)
    Output
      <SQL>
      SELECT 1 AS [x]
      FROM [df]

# generates custom sql

    Code
      sql_table_analyze(con, in_schema("schema", "tbl"))
    Output
      <SQL> UPDATE STATISTICS [schema].[tbl]

---

    Code
      sql_query_save(con, sql("SELECT * FROM foo"), in_schema("schema", "tbl"))
    Output
      <SQL> SELECT * INTO [schema].[tbl] FROM (
        SELECT * FROM foo
      ) AS temp

---

    Code
      sql_query_save(con, sql("SELECT * FROM foo"), in_schema("schema", "tbl"),
      temporary = FALSE)
    Output
      <SQL> SELECT * INTO [schema].[tbl] FROM (
        SELECT * FROM foo
      ) AS temp

---

    Code
      slice_sample(lf, n = 1)
    Output
      <SQL>
      SELECT [x]
      FROM (
        SELECT
          *,
          CASE
      WHEN (NOT(((RAND(CHECKSUM(NEWID()))) IS NULL))) THEN ROW_NUMBER() OVER (PARTITION BY (CASE WHEN (((RAND(CHECKSUM(NEWID()))) IS NULL)) THEN 1 ELSE 0 END) ORDER BY RAND(CHECKSUM(NEWID())))
      END AS [col01]
        FROM [df]
      ) AS [q01]
      WHERE ([col01] <= 1)

---

    Code
      remote_query(copy_inline(con, tibble(x = 1:2, y = letters[1:2])))
    Output
      <SQL> SELECT
        TRY_CAST(TRY_CAST([x] AS NUMERIC) AS INT) AS [x],
        TRY_CAST([y] AS VARCHAR(MAX)) AS [y]
      FROM (  VALUES (1, 'a'), (2, 'b')) AS drvd([x], [y])

---

    Code
      remote_query(copy_inline(con, trees))
    Output
      <SQL> SELECT
        TRY_CAST([Girth] AS FLOAT) AS [Girth],
        TRY_CAST([Height] AS FLOAT) AS [Height],
        TRY_CAST([Volume] AS FLOAT) AS [Volume]
      FROM (
        VALUES
          (8.3, 70.0, 10.3),
          (8.6, 65.0, 10.3),
          (8.8, 63.0, 10.2),
          (10.5, 72.0, 16.4),
          (10.7, 81.0, 18.8),
          (10.8, 83.0, 19.7),
          (11.0, 66.0, 15.6),
          (11.0, 75.0, 18.2),
          (11.1, 80.0, 22.6),
          (11.2, 75.0, 19.9),
          (11.3, 79.0, 24.2),
          (11.4, 76.0, 21.0),
          (11.4, 76.0, 21.4),
          (11.7, 69.0, 21.3),
          (12.0, 75.0, 19.1),
          (12.9, 74.0, 22.2),
          (12.9, 85.0, 33.8),
          (13.3, 86.0, 27.4),
          (13.7, 71.0, 25.7),
          (13.8, 64.0, 24.9),
          (14.0, 78.0, 34.5),
          (14.2, 80.0, 31.7),
          (14.5, 74.0, 36.3),
          (16.0, 72.0, 38.3),
          (16.3, 77.0, 42.6),
          (17.3, 81.0, 55.4),
          (17.5, 82.0, 55.7),
          (17.9, 80.0, 58.3),
          (18.0, 80.0, 51.5),
          (18.0, 80.0, 51.0),
          (20.6, 87.0, 77.0)
      ) AS drvd([Girth], [Height], [Volume])

# `sql_query_insert()` is correct

    Code
      sql_query_insert(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), insert_cols = colnames(df_y), by = c("a", "b"), conflict = "ignore",
      returning_cols = c("a", b2 = "b"))
    Output
      <SQL> INSERT INTO [df_x] ([a], [b], [c], [d])
      OUTPUT [INSERTED].[a], [INSERTED].[b] AS [b2]
      SELECT *
      FROM (
        SELECT [a], [b], [c] + 1.0 AS [c], [d]
        FROM [df_y]
      ) AS [...y]
      WHERE NOT EXISTS (
        SELECT 1 FROM [df_x]
        WHERE ([df_x].[a] = [...y].[a]) AND ([df_x].[b] = [...y].[b])
      )

# `sql_query_append()` is correct

    Code
      sql_query_append(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), insert_cols = colnames(df_y), returning_cols = c("a", b2 = "b"))
    Output
      <SQL> INSERT INTO [df_x] ([a], [b], [c], [d])
      OUTPUT [INSERTED].[a], [INSERTED].[b] AS [b2]
      SELECT *
      FROM (
        SELECT [a], [b], [c] + 1.0 AS [c], [d]
        FROM [df_y]
      ) AS [...y]

# `sql_query_update_from()` is correct

    Code
      sql_query_update_from(con = con, table = ident("df_x"), from = sql_render(df_y,
        con, lvl = 1), by = c("a", "b"), update_values = sql(c = "COALESCE([df_x].[c], [...y].[c])",
        d = "[...y].[d]"), returning_cols = c("a", b2 = "b"))
    Output
      <SQL> UPDATE [df_x]
      SET [c] = COALESCE([df_x].[c], [...y].[c]), [d] = [...y].[d]
      OUTPUT [INSERTED].[a], [INSERTED].[b] AS [b2]
      FROM [df_x]
      INNER JOIN (
        SELECT [a], [b], [c] + 1.0 AS [c], [d]
        FROM [df_y]
      ) AS [...y]
        ON [...y].[a] = [df_x].[a] AND [...y].[b] = [df_x].[b]

# `sql_query_delete()` is correct

    Code
      sql_query_delete(con = dialect_mssql(), table = ident("df_x"), from = sql_render(
        df_y, dialect_mssql(), lvl = 2), by = c("a", "b"), returning_cols = c("a",
        b2 = "b"))
    Output
      <SQL> DELETE FROM [df_x]
      OUTPUT [DELETED].[a], [DELETED].[b] AS [b2]
      WHERE EXISTS (
        SELECT 1 FROM (
          SELECT [a], [b], [c] + 1.0 AS [c], [d]
          FROM [df_y]
      ) AS [...y]
        WHERE ([...y].[a] = [df_x].[a]) AND ([...y].[b] = [df_x].[b])
      )

# `sql_query_upsert()` is correct

    Code
      sql_query_upsert(con = con, table = ident("df_x"), from = sql_render(df_y, con,
        lvl = 1), by = c("a", "b"), update_cols = c("c", "d"), returning_cols = c("a",
        b2 = "b"))
    Output
      <SQL> MERGE INTO [df_x]
      USING (
        SELECT [a], [b], [c] + 1.0 AS [c], [d]
        FROM [df_y]
      ) AS [...y]
        ON [...y].[a] = [df_x].[a] AND [...y].[b] = [df_x].[b]
      WHEN MATCHED THEN
        UPDATE SET [c] = [...y].[c], [d] = [...y].[d]
      WHEN NOT MATCHED THEN
        INSERT ([a], [b], [c], [d])
        VALUES ([...y].[a], [...y].[b], [...y].[c], [...y].[d])
      OUTPUT [INSERTED].[a], [INSERTED].[b] AS [b2]
      ;

# atoms and symbols are cast to bit in `filter`

    Code
      filter(mf, x)
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE (cast([x] AS [BIT]) = 1)

---

    Code
      filter(mf, TRUE)
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE (cast(1 AS [BIT]) = 1)

---

    Code
      filter(mf, (!x) | FALSE)
    Output
      <SQL>
      SELECT *
      FROM [df]
      WHERE ((NOT(cast([x] AS [BIT]) = 1)) OR cast(0 AS [BIT]) = 1)

---

    Code
      inner_join(filter(mf, x), mf, by = "x")
    Output
      <SQL>
      SELECT [LHS].[x] AS [x]
      FROM (
        SELECT *
        FROM [df]
        WHERE (cast([x] AS [BIT]) = 1)
      ) AS [LHS]
      INNER JOIN [df]
        ON ([LHS].[x] = [df].[x])

---

    Code
      left_join(lhs, rhs, by = "a")
    Output
      <SQL>
      SELECT [LHS].*, [b2]
      FROM (
        SELECT *
        FROM [df]
        WHERE (cast([b1] AS [BIT]) = 1)
      ) AS [LHS]
      LEFT JOIN (
        SELECT *
        FROM [df]
        WHERE (cast([b2] AS [BIT]) = 1)
      ) AS [RHS]
        ON ([LHS].[a] = [RHS].[a])

# row_number() with and without group_by() and arrange(): unordered defaults to Ordering by NULL (per empty_order)

    Code
      mutate(mf, rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER (ORDER BY (SELECT NULL)) AS [rown]
      FROM [df]

---

    Code
      mutate(group_by(mf, y), rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER (PARTITION BY [y] ORDER BY (SELECT NULL)) AS [rown]
      FROM [df]

---

    Code
      mutate(arrange(mf, y), rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER (ORDER BY [y]) AS [rown]
      FROM [df]
      ORDER BY [y]

# count_big

    Code
      count(mf)
    Output
      <SQL>
      SELECT COUNT_BIG(*) AS [n]
      FROM [df]

# add prefix to temporary table

    Code
      out <- sql_table_temporary(con, table_path("foo.bar"), temporary = TRUE)
    Message
      Created a temporary table named #bar

# basic arithmetic is correct

    Code
      translate_sql(100L %/% 3L, con = con)
    Condition
      Error in `100L %/% 3L`:
      ! `%/%()` is not available in this SQL variant.

# can translate subsetting

    Code
      translate_sql(a[[x]], con = con)
    Condition
      Error in `a[[x]]`:
      ! Can only index with strings and numbers

---

    Code
      translate_sql(a[[TRUE]], con = con)
    Condition
      Error in `a[[TRUE]]`:
      ! Can only index with strings and numbers

# $ doesn't evaluate second argument

    Code
      filter(lazy_frame(x = 1, y = 1), x == y$id)
    Output
      <SQL>
      SELECT *
      FROM "df"
      WHERE ("x" = "y"."id")

---

    Code
      filter(lazy_frame(x = 1), x == y$id)
    Output
      <SQL>
      SELECT *
      FROM "df"
      WHERE ("x" = 1.0)

# useful error if $ used with inlined value

    Code
      filter(lazy_frame(x = 1), x == y$id)
    Condition
      Error in `x$id`:
      ! $ operator is invalid for atomic vectors

# if_else() and ifelse() require true/yes and false/no

    Code
      translate_sql(if_else(x), con = con)
    Condition
      Error in `if_else()`:
      ! `true` is absent but must be supplied.

---

    Code
      translate_sql(if_else(x, 1L), con = con)
    Condition
      Error in `if_else()`:
      ! `false` is absent but must be supplied.

---

    Code
      translate_sql(ifelse(x), con = con)
    Condition
      Error in `ifelse()`:
      ! `yes` is absent but must be supplied.

---

    Code
      translate_sql(ifelse(x, 1L), con = con)
    Condition
      Error in `ifelse()`:
      ! `no` is absent but must be supplied.

# can only translate case sensitive str_like

    Code
      translate_sql(str_like(x, "abc", ignore_case = FALSE), con = con)
    Condition
      Warning:
      The `ignore_case` argument of `str_like()` is deprecated as of dbplyr 2.6.0.
      i `str_like()` is always case sensitive.
      i Use `str_ilike()` for case insensitive string matching.
    Output
      <SQL> "x" LIKE 'abc'

---

    Code
      translate_sql(str_like(x, "abc", ignore_case = TRUE), con = con)
    Condition
      Warning:
      The `ignore_case` argument of `str_like()` is deprecated as of dbplyr 2.6.0.
      i `str_like()` is always case sensitive.
      i Use `str_ilike()` for case insensitive string matching.
      Error in `str_like()`:
      ! Backend does not support case insensitive `str_like()`.
      i Use `tolower()` on both arguments to achieve a case insensitive match.

---

    Code
      translate_sql(str_ilike(x, "abc"), con = con)
    Condition
      Error in `str_ilike()`:
      ! `str_ilike()` is not available in this SQL variant.

# default raw escapes translated correctly

    Code
      filter(mf, x == a)
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`x` = X'616263')

---

    Code
      filter(mf, x %in% L)
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`x` IN (X'616263', X'0102'))

---

    Code
      qry
    Output
      <SQL>
      SELECT *
      FROM `df`
      WHERE (`x` IN (X'616263', X'0102'))

# queries generated by select() don't alias unnecessarily

    Code
      lf_render
    Output
      <SQL> SELECT "x"
      FROM "df"

---

    Code
      select(lazy_frame(x = 1, y = 1))
    Condition
      Error in `sql_clause_select()`:
      ! Query contains no columns

# pasting translated correctly

    Code
      translate_sql(paste0(x, collapse = ""), con = con)
    Condition
      Error in `check_collapse()`:
      ! `collapse` not supported in DB translation of `paste()`.
      i Please use `str_flatten()` instead.

# row_number() with and without group_by() and arrange(): unordered defaults to Ordering by NULL (per empty_order)

    Code
      mutate(mf, rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER (ORDER BY (SELECT NULL)) AS "rown"
      FROM "df"

---

    Code
      mutate(group_by(mf, y), rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER (PARTITION BY "y" ORDER BY (SELECT NULL)) AS "rown"
      FROM "df"

---

    Code
      mutate(arrange(mf, y), rown = row_number())
    Output
      <SQL>
      SELECT *, ROW_NUMBER() OVER (ORDER BY "y") AS "rown"
      FROM "df"
      ORDER BY "y"

# window functions use inline OVER, not named WINDOW clause

    Code
      mutate(group_by(mf, a), b = mean(b))
    Output
      <SQL>
      SELECT "a", AVG("b") OVER (PARTITION BY "a") AS "b"
      FROM "df"

# aggregation functions warn once if na.rm = FALSE

    Code
      translate_sql(mean(x, na.rm = FALSE), con = con)
    Condition
      Warning:
      Missing values are always removed in SQL aggregation functions.
      Use `na.rm = TRUE` to silence this warning
      This warning is displayed once every 8 hours.
    Output
      <SQL> AVG("x") OVER ()

# warns informatively with unsupported function

    Code
      sql_not_supported("cor")()
    Condition
      Error:
      ! `cor()` is not available in this SQL variant.

# quantile and median don't change without warning

    Code
      translate_sql(quantile(x, 0.75, na.rm = TRUE), con = con, window = FALSE)
    Output
      <SQL> PERCENTILE_CONT(0.75) WITHIN GROUP (ORDER BY "x")

---

    Code
      translate_sql(quantile(x, 0.75, na.rm = TRUE), con = con, vars_group = "g")
    Output
      <SQL> PERCENTILE_CONT(0.75) WITHIN GROUP (ORDER BY "x") OVER (PARTITION BY "g")

---

    Code
      translate_sql(median(x, na.rm = TRUE), con = con, window = FALSE)
    Output
      <SQL> PERCENTILE_CONT(0.5) WITHIN GROUP (ORDER BY "x")

---

    Code
      translate_sql(median(x, na.rm = TRUE), con = con, vars_group = "g")
    Output
      <SQL> PERCENTILE_CONT(0.5) WITHIN GROUP (ORDER BY "x") OVER (PARTITION BY "g")

# select after distinct produces subquery

    Code
      select(distinct(lf), x)
    Output
      <SQL>
      SELECT "x"
      FROM (
        SELECT DISTINCT *
        FROM "df"
      ) AS "q01"

# rename/relocate after distinct is inlined #1141

    Code
      rename(distinct(lf), z = y)
    Output
      <SQL>
      SELECT DISTINCT "x", "y" AS "z"
      FROM "df"
    Code
      relocate(distinct(lf), y)
    Output
      <SQL>
      SELECT DISTINCT "y", "x"
      FROM "df"

# select preserves grouping vars

    Code
      out <- collect(select(mf, a))
    Message
      Adding missing grouping variables: `b`

# select() after left_join() is inlined

    Code
      (out <- select(left_join(lf1, lf2, by = "x"), b, x))
    Output
      <SQL>
      SELECT "b", "lf1"."x" AS "x"
      FROM "lf1"
      LEFT JOIN "lf2"
        ON ("lf1"."x" = "lf2"."x")

---

    Code
      (out <- relocate(left_join(lf1, lf2, by = "x"), b))
    Output
      <SQL>
      SELECT "b", "lf1".*
      FROM "lf1"
      LEFT JOIN "lf2"
        ON ("lf1"."x" = "lf2"."x")

# select() after filter() on join preserves qualified where

    Code
      out
    Output
      <SQL>
      SELECT "x", "y"
      FROM (
        SELECT "t1".*, "z"
        FROM "t1"
        LEFT JOIN "t2"
          ON ("t1"."x" = "t2"."x")
        WHERE ("t2"."z" > 0.0)
      ) AS "q01"

---

    Code
      out
    Output
      <SQL>
      SELECT "x", "y", "z" AS "zz"
      FROM (
        SELECT "t1".*, "z"
        FROM "t1"
        LEFT JOIN "t2"
          ON ("t1"."x" = "t2"."x")
        WHERE ("t2"."z" > 0.0)
      ) AS "q01"

# select() after semi_join() is inlined

    Code
      (out <- select(semi_join(lf1, lf2, by = "x"), x, a2 = a))
    Output
      <SQL>
      SELECT "x", "a" AS "a2"
      FROM "lf1"
      WHERE EXISTS (
        SELECT 1 FROM "lf2"
        WHERE ("lf1"."x" = "lf2"."x")
      )

---

    Code
      (out <- relocate(anti_join(lf1, lf2, by = "x"), a))
    Output
      <SQL>
      SELECT "a", "x"
      FROM "lf1"
      WHERE NOT EXISTS (
        SELECT 1 FROM "lf2"
        WHERE ("lf1"."x" = "lf2"."x")
      )

# select() after join handles previous select

    Code
      print(lf)
    Output
      <SQL>
      SELECT "x" AS "x2", "y" AS "y3", "z"
      FROM "df" AS "df_LHS"
      WHERE EXISTS (
        SELECT 1 FROM "df" AS "df_RHS"
        WHERE ("df_LHS"."x" = "df_RHS"."x")
      )

---

    Code
      print(lf2)
    Output
      <SQL>
      SELECT "df_LHS"."x" AS "x2", "df_LHS"."y" AS "y3", "z"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x")

# select() produces nice error messages

    Code
      select(lf, non_existent)
    Condition
      Error in `select()`:
      ! Can't select columns that don't exist.
      x Column `non_existent` doesn't exist.
    Code
      select(lf, non_existent + 1)
    Condition
      Error in `select()`:
      i In argument: `non_existent + 1`.
      Caused by error:
      ! object 'non_existent' not found

---

    Code
      relocate(lf, non_existent)
    Condition
      Error in `relocate()`:
      ! Can't relocate columns that don't exist.
      x Column `non_existent` doesn't exist.
    Code
      relocate(lf, non_existent + 1)
    Condition
      Error in `relocate()`:
      i In argument: `non_existent + 1`.
      Caused by error:
      ! object 'non_existent' not found

---

    Code
      rename(lf, x)
    Condition
      Error in `rename()`:
      ! All renaming inputs must be named.
    Code
      rename(lf, y = non_existent)
    Condition
      Error in `rename()`:
      ! Can't rename columns that don't exist.
      x Column `non_existent` doesn't exist.
    Code
      rename(lf, y = non_existent + 1)
    Condition
      Error in `rename()`:
      i In argument: `non_existent + 1`.
      Caused by error:
      ! object 'non_existent' not found

---

    Code
      rename_with(lf, toupper, .cols = non_existent)
    Condition
      Error in `rename_with()`:
      ! Can't select columns that don't exist.
      x Column `non_existent` doesn't exist.
    Code
      rename_with(lf, toupper, .cols = non_existent + 1)
    Condition
      Error in `rename_with()`:
      i In argument: `non_existent + 1`.
      Caused by error:
      ! object 'non_existent' not found

# where() isn't suppored

    Code
      select(lf, where(is.integer))
    Condition
      Error in `select()`:
      ! This tidyselect interface doesn't support predicates.

# arranged computed columns are not inlined away

    Code
      select(arrange(mutate(lf, z = 1), x, z), x)
    Condition
      Warning:
      ORDER BY is ignored in subqueries without LIMIT
      i Do you need to move arrange() later in the pipeline or use window_order() instead?
    Output
      <SQL>
      SELECT "x"
      FROM (
        SELECT *, 1.0 AS "z"
        FROM "df"
      ) AS "q01"

# multiple selects are collapsed

    Code
      select(select(lf, 2:1), 2:1)
    Output
      <SQL>
      SELECT *
      FROM "df"

---

    Code
      select(select(select(lf, 2:1), 2:1), 2:1)
    Output
      <SQL>
      SELECT "y", "x"
      FROM "df"

---

    Code
      select(select(lf, x1 = x), x2 = x1)
    Output
      <SQL>
      SELECT "x" AS "x2"
      FROM "df"

# mutate collapses over nested select

    Code
      select(mutate(lf, a = 1, b = 2), a)
    Output
      <SQL>
      SELECT 1.0 AS "a"
      FROM "df"

---

    Code
      select(mutate(lf, a = 1, b = 2), x)
    Output
      <SQL>
      SELECT "x"
      FROM "df"

# output is styled

    Code
      show_query(out, sql_options = sql_options(cte = TRUE))
    Output
      <SQL>
      WITH "q01" AS (
        [34mSELECT[39m "x", AVG("y") OVER (PARTITION BY "x")[34m AS [39m"y", "z" + 1.0[34m AS [39m"z"
        [34mFROM[39m "df"
      ),
      "q02" AS (
        [34mSELECT[39m *
        [34mFROM[39m "q01"
        [34mWHERE[39m ("z" = 1.0)
      )
      [34mSELECT[39m
        "LHS"."x"[34m AS [39m"x",
        "LHS"."y"[34m AS [39m"y.x",
        "LHS"."z"[34m AS [39m"z.x",
        "df"."y"[34m AS [39m"y.y",
        "df"."z"[34m AS [39m"z.y"
      [34mFROM[39m "q02"[34m AS [39m"LHS"
      [34mLEFT JOIN[39m "df"
        [34mON[39m ("LHS"."x" = "df"."x")

# duplicates throw an error

    Code
      sql_translator(round = function(x) x, round = function(y) y)
    Condition
      Error in `sql_translator()`:
      ! Duplicate names in `sql_translator()`
      * round

# output of print method for sql_variant is correct

    Code
      sql_variant(sim_trans, sim_trans, sim_trans)
    Output
      <sql_variant>
      scalar:    +
      aggregate: +
      window:    +

# win_rank() is accepted by the sql_translator

    Code
      sql_variant(sql_translator(test = win_rank("test")))
    Output
      <sql_variant>
      scalar:    test

# custom SQL translation

    Code
      left_join(lf, lf, by = "x", na_matches = "na")
    Output
      <SQL>
      SELECT `df_LHS`.`x` AS `x`
      FROM `df` AS `df_LHS`
      LEFT JOIN `df` AS `df_RHS`
        ON ((`df_LHS`.`x`) IS (`df_RHS`.`x`))

---

    Code
      translate_sql(runif(n()), con = con)
    Output
      <SQL> (0.5 + RANDOM() / 18446744073709551616.0)

# case_when translates correctly to ELSE when TRUE ~ is used

    Code
      translate_sql(case_when(x == 1L ~ "yes", x == 0L ~ "no", TRUE ~ "undefined"),
      con = con)
    Output
      <SQL> CASE WHEN (`x` = 1) THEN 'yes' WHEN (`x` = 0) THEN 'no' ELSE 'undefined' END

# can explain a query

    Code
      explain(filter(db, x > 2))
    Output
      <SQL>
      SELECT *
      FROM `test`
      WHERE (`x` > 2.0)
      
      <PLAN>
        id parent notused                                        detail
      1  2      0      35 SEARCH test USING COVERING INDEX test_x (x>?)

# basic set ops generate expected SQL

    Code
      union(db1, db2)
    Output
      <SQL>
      SELECT *
      FROM "db1"
      
      UNION
      
      SELECT *
      FROM "db2"
    Code
      union_all(db1, db2)
    Output
      <SQL>
      SELECT *
      FROM "db1"
      
      UNION ALL
      
      SELECT *
      FROM "db2"
    Code
      intersect(db1, db2)
    Output
      <SQL>
      (
        SELECT *
        FROM "db1"
      )
      INTERSECT
      (
        SELECT *
        FROM "db2"
      )
    Code
      setdiff(db1, db2)
    Output
      <SQL>
      (
        SELECT *
        FROM "db1"
      )
      EXCEPT
      (
        SELECT *
        FROM "db2"
      )

# can combine multiple unions in one query

    Code
      show_query(lf_union)
    Output
      <SQL>
      SELECT *, NULL AS "z"
      FROM "lf1"
      
      UNION ALL
      
      SELECT *, NULL AS "z"
      FROM (
        SELECT NULL AS "x", *
        FROM "lf2"
      ) AS "q01"
      
      UNION
      
      SELECT NULL AS "x", NULL AS "y", *
      FROM "lf3"

---

    Code
      show_query(lf_union, sql_options = with_cte)
    Output
      <SQL>
      WITH "q01" AS (
        SELECT *, NULL AS "z"
        FROM "lf1"
      ),
      "q02" AS (
        SELECT NULL AS "x", *
        FROM "lf2"
      ),
      "q03" AS (
        SELECT *, NULL AS "z"
        FROM "q02" AS "q01"
      ),
      "q04" AS (
        SELECT NULL AS "x", NULL AS "y", *
        FROM "lf3"
      )
      SELECT *
      FROM "q01"
      
      UNION ALL
      
      SELECT *
      FROM "q03"
      
      UNION
      
      SELECT *
      FROM "q04"

# set ops correctly quote reused queries in CTEs

    Code
      show_query(union_all(lf, lf), sql_options = sql_options(cte = TRUE))
    Output
      <SQL>
      WITH "q01" AS (
        SELECT *, "x" + 1.0 AS "y"
        FROM "lf"
      )
      SELECT *
      FROM "q01"
      
      UNION ALL
      
      SELECT *
      FROM "q01"

# multi_join where clause uses qualified column names

    Code
      filter(left_join(lf1, lf2, by = "x"), y > 1)
    Output
      <SQL>
      SELECT "df_LHS".*, "z"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x")
      WHERE ("df_LHS"."y" > 1.0)

---

    Code
      filter(left_join(lf1, lf2, by = "x"), z > 1)
    Output
      <SQL>
      SELECT "df_LHS".*, "z"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x")
      WHERE ("df_RHS"."z" > 1.0)

---

    Code
      filter(left_join(lf1, lf2, by = "x"), y > 1, z < 5)
    Output
      <SQL>
      SELECT "df_LHS".*, "z"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x")
      WHERE ("df_LHS"."y" > 1.0) AND ("df_RHS"."z" < 5.0)

---

    Code
      filter(left_join(lf1, lf3, by = "x"), y.x > 1)
    Output
      <SQL>
      SELECT "df_LHS"."x" AS "x", "df_LHS"."y" AS "y.x", "df_RHS"."y" AS "y.y"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x")
      WHERE ("df_LHS"."y" > 1.0)

# generated sql doesn't change unexpectedly

    Code
      inner_join(lf, lf)
    Message
      Joining with `by = join_by(x, y)`
    Output
      <SQL>
      SELECT "df_LHS".*
      FROM "df" AS "df_LHS"
      INNER JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x" AND "df_LHS"."y" = "df_RHS"."y")

---

    Code
      left_join(lf, lf)
    Message
      Joining with `by = join_by(x, y)`
    Output
      <SQL>
      SELECT "df_LHS".*
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x" AND "df_LHS"."y" = "df_RHS"."y")

# only disambiguates shared variables

    Code
      left_join(lf1, lf2, by = c(x = "x"))
    Output
      <SQL>
      SELECT "df_LHS".*, "z"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."x" = "df_RHS"."x")

---

    Code
      left_join(lf1, lf2, by = c(y = "z"))
    Output
      <SQL>
      SELECT "df_LHS"."x" AS "x.x", "y", "df_RHS"."x" AS "x.y"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."y" = "df_RHS"."z")

# disambiguate variables that only differ in case

    Code
      left_join(lf1, lf2, by = "y")
    Output
      <SQL>
      SELECT "df_LHS".*, "df_RHS"."X" AS "X"
      FROM "df" AS "df_LHS"
      LEFT JOIN "df" AS "df_RHS"
        ON ("df_LHS"."y" = "df_RHS"."y")

# sql_on query doesn't change unexpectedly

    Code
      inner_join(lf1, lf2, sql_on = "LHS.y < RHS.z")
    Output
      <SQL>
      SELECT "LHS"."x" AS "x.x", "y", "RHS"."x" AS "x.y", "z"
      FROM "df" AS "LHS"
      INNER JOIN "df" AS "RHS"
        ON (LHS.y < RHS.z)

---

    Code
      left_join(lf1, lf2, sql_on = "LHS.y < RHS.z")
    Output
      <SQL>
      SELECT "LHS"."x" AS "x.x", "y", "RHS"."x" AS "x.y", "z"
      FROM "df" AS "LHS"
      LEFT JOIN "df" AS "RHS"
        ON (LHS.y < RHS.z)

# generates expected SQL for common situations

    Code
      count(db, g)
    Output
      <SQL>
      SELECT "g", COUNT(*) AS "n"
      FROM "df"
      GROUP BY "g"

---

    Code
      count(db, g, wt = x)
    Output
      <SQL>
      SELECT "g", SUM("x") AS "n"
      FROM "df"
      GROUP BY "g"

---

    Code
      count(db, g, sort = TRUE)
    Output
      <SQL>
      SELECT "g", COUNT(*) AS "n"
      FROM "df"
      GROUP BY "g"
      ORDER BY "n" DESC

---

    Code
      add_count(db, g, sort = TRUE)
    Output
      <SQL>
      SELECT *, COUNT(*) OVER (PARTITION BY "g") AS "n"
      FROM "df"
      ORDER BY "n" DESC

---

    Code
      add_count(group_by(db, g))
    Output
      <SQL>
      SELECT *, COUNT(*) OVER (PARTITION BY "g") AS "n"
      FROM "df"

# .drop is not supported

    Code
      add_count(lazy_frame(g = 1), .drop = TRUE)
    Condition
      Error in `add_count()`:
      ! Argument `.drop` isn't supported on database backends.

# analyse/explain sql generates expected SQL

    Code
      sql_table_analyze(con, "tbl")
    Output
      <SQL> ANALYZE "tbl"

---

    Code
      sql_query_explain(con, sql("SELECT * FROM foo"))
    Output
      <SQL> EXPLAIN SELECT * FROM foo

# sql_query_wrap generates expected SQL

    Code
      sql_query_wrap(con, ident("table"))
    Output
      <SQL> "table"

---

    Code
      sql_query_wrap(con, in_schema("schema", "tbl"))
    Output
      <SQL> "schema"."tbl"

---

    Code
      sql_query_wrap(con, sql("SELECT * FROM foo"))
    Output
      <SQL> (SELECT * FROM foo) AS "q01"

# sql_table_index generates expected SQL

    Code
      sql_table_index(con, "tbl", c("a", "b"))
    Output
      <SQL> CREATE INDEX "tbl_a_b" ON "tbl" ("a", "b")

---

    Code
      sql_table_index(con, "tbl", "c", unique = TRUE)
    Output
      <SQL> CREATE UNIQUE INDEX "tbl_c" ON "tbl" ("c")

# sql_query_save generates expected SQL

    Code
      sql_query_save(con, sql, "tbl")
    Output
      <SQL> CREATE TEMPORARY TABLE
      "tbl" AS
      SELECT * FROM foo

---

    Code
      sql_query_save(con, sql, "tbl", temporary = FALSE)
    Output
      <SQL> CREATE TABLE
      "tbl" AS
      SELECT * FROM foo

# window_order errors for data frame

    Code
      (expect_error(window_order(data.frame(x = 1))))
    Output
      <error/rlang_error>
      Error in `window_order()`:
      ! `.data` must be a <tbl_lazy>, not a data frame.
      i Did you mean to use `arrange()` instead?
    Code
      (expect_error(window_order("a")))
    Output
      <error/rlang_error>
      Error in `window_order()`:
      ! `.data` must be a <tbl_lazy>, not a string.

# window_order only accepts variables

    Code
      window_order(lf, x + y)
    Condition
      Error in `window_order()`:
      ! Every element of `...` must be a single column name or a column wrapped in `desc()`.
      x Element 1 is `x + y`.
    Code
      window_order(lf, foo())
    Condition
      Error in `window_order()`:
      ! Every element of `...` must be a single column name or a column wrapped in `desc()`.
      x Element 1 is `foo()`.
    Code
      window_order(lf, desc(x + y))
    Condition
      Error in `window_order()`:
      ! Every element of `...` must be a single column name or a column wrapped in `desc()`.
      x Element 1 is `desc(x + y)`.

# window order works afer renaming variable

    Code
      mutate(rename(window_order(lazy_frame(x = 1, y = 1), y), y2 = y), x_cum = cumsum(
        x))
    Output
      <SQL>
      SELECT *, SUM("x") OVER (ORDER BY "y2" ROWS UNBOUNDED PRECEDING) AS "x_cum"
      FROM (
        SELECT "x", "y" AS "y2"
        FROM "df"
      ) AS "q01"
    Code
      mutate(window_order(rename(lazy_frame(x = 1, y = 1), y2 = y), y2), x_cum = cumsum(
        x))
    Output
      <SQL>
      SELECT *, SUM("x") OVER (ORDER BY "y2" ROWS UNBOUNDED PRECEDING) AS "x_cum"
      FROM (
        SELECT "x", "y" AS "y2"
        FROM "df"
      ) AS "q01"

# window_frame errors for data frame

    Code
      (expect_error(window_frame(data.frame(x = 1))))
    Output
      <error/rlang_error>
      Error in `window_frame()`:
      ! `.data` must be a <tbl_lazy>, not a <data.frame>.

# window_order works with qualified desc()

    Code
      show_query(out)
    Output
      <SQL>
      SELECT "y", SUM("y") OVER (ORDER BY "y" DESC ROWS UNBOUNDED PRECEDING) AS "z"
      FROM "df"

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remote.R
\name{remote_name}
\alias{remote_name}
\alias{remote_table}
\alias{remote_src}
\alias{remote_con}
\alias{remote_query}
\alias{remote_query_plan}
\title{Metadata about a remote table}
\usage{
remote_name(x, null_if_local = TRUE)

remote_table(x, null_if_local = TRUE)

remote_src(x)

remote_con(x)

remote_query(x, sql_options = NULL, cte = deprecated())

remote_query_plan(x, ...)
}
\arguments{
\item{x}{Remote table, currently must be a \code{\link{tbl_sql}}.}

\item{null_if_local}{Return \code{NULL} if the remote table is created via
\code{\link[=tbl_lazy]{tbl_lazy()}} or \code{\link[=lazy_frame]{lazy_frame()}}?}

\item{sql_options}{SQL rendering options generated by \code{\link[=sql_options]{sql_options()}}.}

\item{cte}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Use the \code{sql_options} argument instead.}

\item{...}{Additional arguments passed on to methods.}
}
\value{
\itemize{
\item \code{remote_name()} returns a string, or \code{NULL} if not a remote table, or not
applicable. For example, computed queries do not have a "name".
\item \code{remote_table()} returns a \code{\link[=sql]{sql()}} object containing the (correctly
quoted) table identifier, or the custom query if the table was created
from \code{sql()}. Returns \code{NULL} if applied to a lazy query that no longer
corresponds to a single table (e.g. after a \code{filter()} or a join), or if
applied to a local lazy table.
}
}
\description{
\code{remote_name()} gives the unescaped name of the remote table, or \code{NULL} if it
is a query (created by \code{sql()}) or already escape (created by \code{ident_q()}).
\code{remote_table()} gives the remote table or the query.
\code{remote_query()} gives the text of the query, and \code{remote_query_plan()}
the query plan (as computed by the remote database). \code{remote_src()} and
\code{remote_con()} give the dplyr source and DBI connection respectively.
}
\examples{
mf <- memdb_frame(x = 1:5, y = 5:1, .name = "blorp")
remote_name(mf)
remote_src(mf)
remote_con(mf)
remote_query(mf)

mf2 <- dplyr::filter(mf, x > 3)
remote_name(mf2)
remote_src(mf2)
remote_con(mf2)
remote_query(mf2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sql.R
\name{sql}
\alias{sql}
\alias{is.sql}
\title{Literal SQL escaping}
\usage{
sql(...)

is.sql(x)
}
\arguments{
\item{...}{Character vectors that will be combined into a single SQL vector.}

\item{x}{Object to check if it is an sql object.}
}
\description{
Use \code{sql()} to declare that a string is literal SQL and should be used
as is, without quoting.
}
\examples{
library(dplyr, warn.conflicts = FALSE)
# sql() just adds a class
sql("x + 1")
is.sql(sql("x + 1"))

# You can use it when you need to insert some literal SQL in a query
db <- memdb_frame(x = 1:3)
db |> mutate(y = sql("CAST(x as VARCHAR)"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db-io.R
\name{db-io}
\alias{db_copy_to}
\alias{db_compute}
\alias{db_collect}
\alias{db_table_temporary}
\title{Database I/O generics}
\usage{
db_copy_to(
  con,
  table,
  values,
  overwrite = FALSE,
  types = NULL,
  temporary = TRUE,
  unique_indexes = NULL,
  indexes = NULL,
  analyze = TRUE,
  ...,
  in_transaction = TRUE
)

db_compute(
  con,
  table,
  sql,
  temporary = TRUE,
  unique_indexes = list(),
  indexes = list(),
  analyze = TRUE,
  ...
)

db_collect(con, sql, n = -1, warn_incomplete = TRUE, ...)

db_table_temporary(con, table, temporary)
}
\description{
These generics are responsible for getting data into and out of the
database. They should be used a last resort - only use them when you can't
make a backend work by providing methods for DBI generics, or for dbplyr's
SQL generation generics. They tend to be most needed when a backend has
special handling of temporary tables.
\itemize{
\item \code{db_copy_to()} implements \code{\link[=copy_to.src_sql]{copy_to.src_sql()}} by calling
\code{db_write_table()} (which calls \code{\link[DBI:dbWriteTable]{DBI::dbWriteTable()}}) to transfer the
data, then optionally adds indexes (via \code{\link[=sql_table_index]{sql_table_index()}}) and
analyses (via \code{\link[=sql_table_analyze]{sql_table_analyze()}}).
\item \code{db_compute()} implements \code{\link[=compute.tbl_sql]{compute.tbl_sql()}} by calling
\code{\link[=sql_query_save]{sql_query_save()}} to create the table, then optionally adds indexes
(via \code{\link[=sql_table_index]{sql_table_index()}}) and analyses (via \code{\link[=sql_table_analyze]{sql_table_analyze()}}).
\item \code{db_collect()} implements \code{\link[=collect.tbl_sql]{collect.tbl_sql()}} using \code{\link[DBI:dbSendQuery]{DBI::dbSendQuery()}}
and \code{\link[DBI:dbFetch]{DBI::dbFetch()}}.
\item \code{db_table_temporary()} is used for databases that have special naming
schemes for temporary tables (e.g. SQL server and SAP HANA require
temporary tables to start with \verb{#})
}
}
\seealso{
Other generic: 
\code{\link{db-sql}},
\code{\link{db_connection_describe}()},
\code{\link{sql_escape_logical}()}
}
\concept{generic}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-copy-to.R
\name{copy_to.src_sql}
\alias{copy_to.src_sql}
\title{Copy a local data frame to a remote database}
\usage{
\method{copy_to}{src_sql}(
  dest,
  df,
  name = deparse(substitute(df)),
  overwrite = FALSE,
  types = NULL,
  temporary = TRUE,
  unique_indexes = NULL,
  indexes = NULL,
  analyze = TRUE,
  ...,
  in_transaction = TRUE
)
}
\arguments{
\item{dest}{remote data source}

\item{df}{A local data frame, a \code{tbl_sql} from same source, or a \code{tbl_sql}
from another source. If from another source, all data must transition
through R in one pass, so it is only suitable for transferring small
amounts of data.}

\item{name}{name for new remote table.}

\item{overwrite}{If \code{TRUE}, will overwrite an existing table with
name \code{name}. If \code{FALSE}, will throw an error if \code{name} already
exists.}

\item{types}{a character vector giving variable types to use for the columns.
See \url{https://www.sqlite.org/datatype3.html} for available types.}

\item{temporary}{if \code{TRUE}, will create a temporary table that is
local to this connection and will be automatically deleted when the
connection expires}

\item{unique_indexes}{a list of character vectors. Each element of the list
will create a new unique index over the specified column(s). Duplicate rows
will result in failure.}

\item{indexes}{a list of character vectors. Each element of the list
will create a new index.}

\item{analyze}{if \code{TRUE} (the default), will automatically ANALYZE the
new table so that the query optimiser has useful information.}

\item{...}{other parameters passed to methods.}

\item{in_transaction}{Should the table creation be wrapped in a transaction?
This typically makes things faster, but you may want to suppress if the
database doesn't support transactions, or you're wrapping in a transaction
higher up (and your database doesn't support nested transactions.)}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
This is an implementation of the dplyr \code{\link[=copy_to]{copy_to()}} generic and it mostly
a wrapper around \code{\link[DBI:dbWriteTable]{DBI::dbWriteTable()}}.

It is useful for copying small amounts of data to a database for examples,
experiments, and joins. By default, it creates temporary tables which are
only visible within the current connection to the database.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

df <- data.frame(x = 1:5, y = letters[5:1])
db <- copy_to(src_memdb(), df)
db

df2 <- data.frame(y = c("a", "d"), fruit = c("apple", "date"))
# copy_to() is called automatically if you set copy = TRUE
# in the join functions
db \%>\% left_join(df2, copy = TRUE)
}
\seealso{
\code{\link[=copy_inline]{copy_inline()}} to use small data in an SQL query without actually
writing to a table.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-arrange.R
\name{arrange.tbl_lazy}
\alias{arrange.tbl_lazy}
\title{Arrange rows by column values}
\usage{
\method{arrange}{tbl_lazy}(.data, ..., .by_group = FALSE)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{...}{<\code{\link[dplyr:dplyr_data_masking]{data-masking}}> Variables, or functions of
variables. Use \code{\link[dplyr:desc]{desc()}} to sort a variable in descending order.}

\item{.by_group}{If \code{TRUE}, will sort first by grouping variable. Applies to
grouped data frames only.}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
This is an method for the dplyr \code{\link[=arrange]{arrange()}} generic. It generates
the \verb{ORDER BY} clause of the SQL query. It also affects the
\code{\link[=window_order]{window_order()}} of windowed expressions in \code{\link[=mutate.tbl_lazy]{mutate.tbl_lazy()}}.

Note that \verb{ORDER BY} clauses can not generally appear in subqueries, which
means that you should \code{arrange()} as late as possible in your pipelines.
}
\section{Missing values}{

Unlike R, most databases sorts \code{NA} (\code{NULL}s) at the front. You can
can override this behaviour by explicitly sorting on \code{is.na(x)}.
}

\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(a = c(3, 4, 1, 2), b = c(5, 1, 2, NA))
db \%>\% arrange(a) \%>\% show_query()

# Note that NAs are sorted first
db \%>\% arrange(b)
# override by sorting on is.na() first
db \%>\% arrange(is.na(b), b)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query-base.R
\name{lazy_base_query}
\alias{lazy_base_query}
\title{Create a base lazy query}
\usage{
lazy_base_query(x, vars, class = character(), ...)
}
\arguments{
\item{x}{A data source, typically a table identifier created by \code{\link[=in_schema]{in_schema()}}
or \code{\link[=in_catalog]{in_catalog()}}, or a literal \code{\link[=sql]{sql()}} string.}

\item{vars}{A character vector of column names.}

\item{class}{A character vector of additional subclasses to add. The
resulting object will have class \code{c("lazy_base_{class}_query", "lazy_base_query", "lazy_query")}.}

\item{...}{Additional arguments passed to \code{\link[=lazy_query]{lazy_query()}}.}
}
\value{
A lazy query object.
}
\description{
\code{lazy_base_query()} is a constructor for base lazy query objects. A base
lazy query represents the root of a lazy query tree, i.e. a database table or
query.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-do.R
\name{do.tbl_sql}
\alias{do.tbl_sql}
\title{Perform arbitrary computation on remote backend}
\usage{
\method{do}{tbl_sql}(.data, ..., .chunk_size = 10000L)
}
\arguments{
\item{.data}{a tbl}

\item{...}{Expressions to apply to each group. If named, results will be
stored in a new column. If unnamed, must return a data frame. You can
use \code{.} to refer to the current group. You can not mix named and
unnamed arguments.}

\item{.chunk_size}{The size of each chunk to pull into R. If this number is
too big, the process will be slow because R has to allocate and free a lot
of memory. If it's too small, it will be slow, because of the overhead of
talking to the database.}
}
\description{
Perform arbitrary computation on remote backend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-window.R
\name{window_order}
\alias{window_order}
\alias{window_frame}
\title{Override window order and frame}
\usage{
window_order(.data, ...)

window_frame(.data, from = -Inf, to = Inf)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{...}{Variables to order by}

\item{from, to}{Bounds of the frame.}
}
\description{
These allow you to override the \verb{PARTITION BY} and \verb{ORDER BY} clauses
of window functions generated by grouped mutates.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(g = rep(1:2, each = 5), y = runif(10), z = 1:10)
db \%>\%
  window_order(y) \%>\%
  mutate(z = cumsum(y)) \%>\%
  show_query()

db \%>\%
  group_by(g) \%>\%
  window_frame(-3, 0) \%>\%
  window_order(z) \%>\%
  mutate(z = sum(y)) \%>\%
  show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-joins.R
\name{join.tbl_sql}
\alias{join.tbl_sql}
\alias{inner_join.tbl_lazy}
\alias{left_join.tbl_lazy}
\alias{right_join.tbl_lazy}
\alias{full_join.tbl_lazy}
\alias{cross_join.tbl_lazy}
\alias{semi_join.tbl_lazy}
\alias{anti_join.tbl_lazy}
\title{Join SQL tables}
\usage{
\method{inner_join}{tbl_lazy}(
  x,
  y,
  by = NULL,
  copy = "none",
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("never", "na"),
  multiple = NULL,
  unmatched = "drop",
  relationship = NULL,
  sql_on = NULL,
  auto_index = FALSE,
  x_as = NULL,
  y_as = NULL
)

\method{left_join}{tbl_lazy}(
  x,
  y,
  by = NULL,
  copy = "none",
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("never", "na"),
  multiple = NULL,
  unmatched = "drop",
  relationship = NULL,
  sql_on = NULL,
  auto_index = FALSE,
  x_as = NULL,
  y_as = NULL
)

\method{right_join}{tbl_lazy}(
  x,
  y,
  by = NULL,
  copy = "none",
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("never", "na"),
  multiple = NULL,
  unmatched = "drop",
  relationship = NULL,
  sql_on = NULL,
  auto_index = FALSE,
  x_as = NULL,
  y_as = NULL
)

\method{full_join}{tbl_lazy}(
  x,
  y,
  by = NULL,
  copy = "none",
  suffix = NULL,
  ...,
  keep = NULL,
  na_matches = c("never", "na"),
  multiple = NULL,
  relationship = NULL,
  sql_on = NULL,
  auto_index = FALSE,
  x_as = NULL,
  y_as = NULL
)

\method{cross_join}{tbl_lazy}(
  x,
  y,
  ...,
  copy = "none",
  suffix = c(".x", ".y"),
  x_as = NULL,
  y_as = NULL
)

\method{semi_join}{tbl_lazy}(
  x,
  y,
  by = NULL,
  copy = "none",
  ...,
  na_matches = c("never", "na"),
  sql_on = NULL,
  auto_index = FALSE,
  x_as = NULL,
  y_as = NULL
)

\method{anti_join}{tbl_lazy}(
  x,
  y,
  by = NULL,
  copy = "none",
  ...,
  na_matches = c("never", "na"),
  sql_on = NULL,
  auto_index = FALSE,
  x_as = NULL,
  y_as = NULL
)
}
\arguments{
\item{x, y}{A pair of lazy data frames backed by database queries.}

\item{by}{A join specification created with \code{\link[dplyr:join_by]{join_by()}}, or a character
vector of variables to join by.

If \code{NULL}, the default, \verb{*_join()} will perform a natural join, using all
variables in common across \code{x} and \code{y}. A message lists the variables so
that you can check they're correct; suppress the message by supplying \code{by}
explicitly.

To join on different variables between \code{x} and \code{y}, use a \code{\link[dplyr:join_by]{join_by()}}
specification. For example, \code{join_by(a == b)} will match \code{x$a} to \code{y$b}.

To join by multiple variables, use a \code{\link[dplyr:join_by]{join_by()}} specification with
multiple expressions. For example, \code{join_by(a == b, c == d)} will match
\code{x$a} to \code{y$b} and \code{x$c} to \code{y$d}. If the column names are the same between
\code{x} and \code{y}, you can shorten this by listing only the variable names, like
\code{join_by(a, c)}.

\code{\link[dplyr:join_by]{join_by()}} can also be used to perform inequality, rolling, and overlap
joins. See the documentation at \link[dplyr:join_by]{?join_by} for details on
these types of joins.

For simple equality joins, you can alternatively specify a character vector
of variable names to join by. For example, \code{by = c("a", "b")} joins \code{x$a}
to \code{y$a} and \code{x$b} to \code{y$b}. If variable names differ between \code{x} and \code{y},
use a named character vector like \code{by = c("x_a" = "y_a", "x_b" = "y_b")}.

To perform a cross-join, generating all combinations of \code{x} and \code{y}, see
\code{\link[dplyr:cross_join]{cross_join()}}.}

\item{copy}{If \code{x} and \code{y} are not from the same data source, \code{copy}
controls how \code{y} is copied into the same source as \code{x}. There are three
options:
\itemize{
\item \code{"none"}, the default, will error if \code{y} needs to be copied. This ensures
that you don't accidentally copy large datasets from R to the database.
\item \code{"temp-table"}: copies \code{y} into a temporary table in the same
database as \code{x}. \verb{*_join()} will automatically run \code{ANALYZE} on the
created table in the hope that this will make your queries as efficient
as possible by giving more data to the query planner.
\item \code{"inline"}: \code{y} will be inlined into the query using \code{\link[=copy_inline]{copy_inline()}}.
This should be faster for small datasets and doesn't require write
access.
}

\code{TRUE} (\code{"temp-table"}) and \code{FALSE} (\code{"none"}) are also accepted for
backward compatibility.}

\item{suffix}{If there are non-joined duplicate variables in \code{x} and
\code{y}, these suffixes will be added to the output to disambiguate them.
Should be a character vector of length 2.}

\item{...}{Other parameters passed onto methods.}

\item{keep}{Should the join keys from both \code{x} and \code{y} be preserved in the
output?
\itemize{
\item If \code{NULL}, the default, joins on equality retain only the keys from \code{x},
while joins on inequality retain the keys from both inputs.
\item If \code{TRUE}, all keys from both inputs are retained.
\item If \code{FALSE}, only keys from \code{x} are retained. For right and full joins,
the data in key columns corresponding to rows that only exist in \code{y} are
merged into the key columns from \code{x}. Can't be used when joining on
inequality conditions.
}}

\item{na_matches}{Should NA (NULL) values match one another?
The default, "never", is how databases usually work. \code{"na"} makes
the joins behave like the dplyr join functions, \code{\link[=merge]{merge()}}, \code{\link[base:match]{base::match()}},
and \code{\%in\%}.}

\item{multiple, unmatched}{Unsupported in database backends. As a workaround
for multiple use a unique key and for unmatched a foreign key constraint.}

\item{relationship}{Unsupported in database backends.}

\item{sql_on}{A custom join predicate as an SQL expression.
Usually joins use column equality, but you can perform more complex
queries by supplying \code{sql_on} which should be a SQL expression that
uses \code{LHS} and \code{RHS} aliases to refer to the left-hand side or
right-hand side of the join respectively.}

\item{auto_index}{if \code{copy} is \code{TRUE}, automatically create
indices for the variables in \code{by}. This may speed up the join if
there are matching indexes in \code{x}.}

\item{x_as, y_as}{Alias to use for \code{x} resp. \code{y}. Defaults to \code{"LHS"} resp.
\code{"RHS"}}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
These are methods for the dplyr \link[dplyr:join]{join} generics. They are translated
to the following SQL queries:
\itemize{
\item \code{inner_join(x, y)}: \verb{SELECT * FROM x JOIN y ON x.a = y.a}
\item \code{left_join(x, y)}:  \verb{SELECT * FROM x LEFT JOIN y ON x.a = y.a}
\item \code{right_join(x, y)}: \verb{SELECT * FROM x RIGHT JOIN y ON x.a = y.a}
\item \code{full_join(x, y)}:  \verb{SELECT * FROM x FULL JOIN y ON x.a = y.a}
\item \code{semi_join(x, y)}:  \verb{SELECT * FROM x WHERE EXISTS (SELECT 1 FROM y WHERE x.a = y.a)}
\item \code{anti_join(x, y)}:  \verb{SELECT * FROM x WHERE NOT EXISTS (SELECT 1 FROM y WHERE x.a = y.a)}
}
}
\examples{
library(dplyr, warn.conflicts = FALSE)

band_db <- copy_to(memdb(), dplyr::band_members)
instrument_db <- copy_to(memdb(), dplyr::band_instruments)
band_db |> left_join(instrument_db) |> show_query()

# Can join with local data frames by setting copy = TRUE
band_db |>
  left_join(dplyr::band_instruments, copy = TRUE)

# Unlike R, joins in SQL don't usually match NAs (NULLs)
db <- memdb_frame(x = c(1, 2, NA))
label <- memdb_frame(x = c(1, NA), label = c("one", "missing"))
db |> left_join(label, by = "x")
# But you can activate R's usual behaviour with the na_matches argument
db |> left_join(label, by = "x", na_matches = "na")

# By default, joins are equijoins, but you can use `sql_on` to
# express richer relationships
db1 <- memdb_frame(x = 1:5)
db2 <- memdb_frame(x = 1:3, y = letters[1:3])
db1 |> left_join(db2) |> show_query()
db1 |> left_join(db2, sql_on = "LHS.x < RHS.x") |> show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/src_dbi.R
\name{src_dbi}
\alias{src_dbi}
\title{Database src}
\usage{
src_dbi(con, auto_disconnect = FALSE)
}
\arguments{
\item{con}{An object that inherits from \link[DBI:DBIConnection-class]{DBI::DBIConnection},
typically generated by \link[DBI:dbConnect]{DBI::dbConnect}}

\item{auto_disconnect}{Should the connection be automatically closed when
the src is deleted? Set to \code{TRUE} if you initialize the connection
the call to \code{src_dbi()}. Pass \code{NA} to auto-disconnect but print a message
when this happens.}
}
\value{
An S3 object with class \code{src_dbi}, \code{src_sql}, \code{src}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

Since can generate a \code{tbl()} directly from a DBI connection we no longer
recommend using \code{src_dbi()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-pull.R
\name{pull.tbl_sql}
\alias{pull.tbl_sql}
\title{Extract a single column}
\usage{
\method{pull}{tbl_sql}(.data, var = -1)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{var}{A variable specified as:
\itemize{
\item a literal variable name
\item a positive integer, giving the position counting from the left
\item a negative integer, giving the position counting from the right.
}

The default returns the last column (on the assumption that's the
column you've created most recently).

This argument is taken by expression and supports
\link[rlang:topic-inject]{quasiquotation} (you can unquote column
names and column locations).}
}
\value{
A vector of data.
}
\description{
This is a method for the dplyr \code{\link[=pull]{pull()}} generic. It evaluates the query
retrieving just the specified column.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(x = 1:5, y = 5:1)
db \%>\%
  mutate(z = x + y * 2) \%>\%
  pull()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memdb.R
\name{src_memdb}
\alias{src_memdb}
\alias{tbl_memdb}
\title{Deprecated}
\usage{
src_memdb()

tbl_memdb(df, name = deparse(substitute(df)))
}
\arguments{
\item{df}{Data frame to copy.}

\item{name}{Name of table in database.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
\itemize{
\item \code{src_memdb()} is deprecated; use \code{memdb()} instead.
\item \code{tbl_memdb(df)} is deprecated; use \code{copy_to(memdb(), df)} instead.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/translate-sql.R
\name{translate_sql}
\alias{translate_sql}
\alias{translate_sql_}
\title{Translate an expression to SQL}
\usage{
translate_sql(
  ...,
  con = NULL,
  vars = character(),
  vars_group = NULL,
  vars_order = NULL,
  vars_frame = NULL,
  window = TRUE
)

translate_sql_(
  dots,
  con = NULL,
  vars_group = NULL,
  vars_order = NULL,
  vars_frame = NULL,
  window = TRUE,
  context = list()
)
}
\arguments{
\item{..., dots}{Expressions to translate. \code{translate_sql()}
automatically quotes them for you.  \code{translate_sql_()} expects
a list of already quoted objects.}

\item{con}{An optional database connection to control the details of
the translation. The default, \code{NULL}, generates ANSI SQL.}

\item{vars}{Deprecated. Now call \code{\link[=partial_eval]{partial_eval()}} directly.}

\item{vars_group, vars_order, vars_frame}{Parameters used in the \code{OVER}
expression of windowed functions.}

\item{window}{Use \code{FALSE} to suppress generation of the \code{OVER}
statement used for window functions. This is necessary when generating
SQL for a grouped summary.}

\item{context}{Use to carry information for special translation cases. For example, MS SQL needs a different conversion for is.na() in WHERE vs. SELECT clauses.  Expects a list.}
}
\description{
dbplyr translates commonly used base functions including logical
(\code{!}, \code{&}, \code{|}), arithmetic (\code{^}), and comparison (\code{!=}) operators, as well
as common summary (\code{mean()}, \code{var()}), and transformation (\code{log()})
functions.  All other functions will be preserved as is. R's infix functions
(e.g. \verb{\%like\%}) will be converted to their SQL equivalents (e.g. \code{LIKE}).

Learn more in \code{vignette("translation-function")}.
}
\examples{
# Regular maths is translated in a very straightforward way
translate_sql(x + 1)
translate_sql(sin(x) + tan(y))

# Note that all variable names are escaped
translate_sql(like == "x")
# In ANSI SQL: "" quotes variable _names_, '' quotes strings

# Logical operators are converted to their sql equivalents
translate_sql(x < 5 & !(y >= 5))
# xor() doesn't have a direct SQL equivalent
translate_sql(xor(x, y))

# If is translated into case when
translate_sql(if (x > 5) "big" else "small")

# Infix functions are passed onto SQL with \% removed
translate_sql(first \%like\% "Had\%")
translate_sql(first \%is\% NA)
translate_sql(first \%in\% c("John", "Roger", "Robert"))

# And be careful if you really want integers
translate_sql(x == 1)
translate_sql(x == 1L)

# If you have an already quoted object, use translate_sql_:
x <- quote(y + 1 / sin(t))
translate_sql_(list(x), con = simulate_dbi())

# Windowed translation --------------------------------------------
# Known window functions automatically get OVER()
translate_sql(mpg > mean(mpg))

# Suppress this with window = FALSE
translate_sql(mpg > mean(mpg), window = FALSE)

# vars_group controls partition:
translate_sql(mpg > mean(mpg), vars_group = "cyl")

# and vars_order controls ordering for those functions that need it
translate_sql(cumsum(mpg))
translate_sql(cumsum(mpg), vars_order = "mpg")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/escape.R
\name{sql_quote}
\alias{sql_quote}
\title{Helper function for quoting sql elements.}
\usage{
sql_quote(x, quote)
}
\arguments{
\item{x}{Character vector to escape.}

\item{quote}{Single quoting character.}
}
\description{
If the quote character is present in the string, it will be doubled.
\code{NA}s will be replaced with NULL.
}
\examples{
sql_quote("abc", "'")
sql_quote("I've had a good day", "'")
sql_quote(c("abc", NA), "'")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-hana.R
\name{backend-hana}
\alias{simulate_hana}
\title{Backend: SAP HANA}
\usage{
simulate_hana()
}
\description{
See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology. Key differences for this backend
are:
\itemize{
\item Temporary tables get \verb{#} prefix and use \verb{LOCAL TEMPORARY COLUMN}.
\item No table analysis performed in \code{\link[=copy_to]{copy_to()}}.
\item \code{paste()} uses \code{||}
\item Note that you can't create new boolean columns from logical expressions;
you need to wrap with explicit \code{ifelse}: \code{ifelse(x > y, TRUE, FALSE)}.
}

Use \code{simulate_hana()} with \code{lazy_frame()} to see simulated SQL without
converting to live access database.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = simulate_hana())
lf \%>\% transmute(x = paste0(d, " times"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-snowflake.R
\name{backend-snowflake}
\alias{simulate_snowflake}
\title{Backend: Snowflake}
\usage{
simulate_snowflake()
}
\description{
See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.

Use \code{simulate_snowflake()} with \code{lazy_frame()} to see simulated SQL without
converting to live access database.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = simulate_snowflake())
lf \%>\% transmute(x = paste0(d, " times"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-impala.R
\name{backend-impala}
\alias{dialect_impala}
\alias{simulate_impala}
\title{Impala backend}
\usage{
dialect_impala()

simulate_impala()
}
\description{
This backend supports Apache Impala, typically accessed via odbc. Use
\code{dialect_impala()} with \code{lazy_frame()} to see simulated SQL without
connecting to a live database.

Key differences for this backend are a scattering of custom translations
provided by users, mostly focussed on bitwise operations.

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_impala())
lf |> transmute(X = bitwNot(bitwOr(b, c)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyeval.R
\name{partial_eval}
\alias{partial_eval}
\title{Partially evaluate an expression.}
\usage{
partial_eval(call, data, env = caller_env(), error_call = caller_env())
}
\arguments{
\item{call}{an unevaluated expression, as produced by \code{\link[=quote]{quote()}}}

\item{data}{A lazy data frame backed by a database query.}

\item{env}{environment in which to search for local values}
}
\description{
This function partially evaluates a quosure yielding an expression. It
uses information from the current \code{tbl} and the local environment to yield
a standalone expression. This simplifies SQL translation because we can just
pass around expressions rather than expressions + environments
(i.e. quosures).
}
\section{Symbol substitution}{


\code{partial_eval()} needs to guess if you're referring to a variable on the
server (remote), or in the current environment (local). It's not possible to
do this 100\% perfectly. \code{partial_eval()} uses the following heuristic:

\itemize{
\item If the tbl variables are known, and the symbol matches a tbl
variable, then remote.
\item If the symbol is defined locally, local.
\item Otherwise, remote.
}

You can override the guesses using \code{local()} and \code{remote()} to force
computation, by using the \code{.data} and \code{.env} pronouns of tidy evaluation,
or by using dbplyr's own \code{.sql} pronoun.
}

\examples{
lf <- lazy_frame(year = 1980, id = 1)
partial_eval(quote(year > 1980), data = lf)

ids <- c("ansonca01", "forceda01", "mathebo01")
partial_eval(quote(id \%in\% ids), lf)

# cf.
partial_eval(quote(id == .data$id), lf)

# You can use local() or .env to disambiguate between local and remote
# variables: otherwise remote is always preferred
year <- 1980
partial_eval(quote(year > year), lf)
partial_eval(quote(year > local(year)), lf)
partial_eval(quote(year > .env$year), lf)

# Functions are always assumed to be remote. Use local to force evaluation
# in R.
f <- function(x) x + 1
partial_eval(quote(year > f(1980)), lf)
partial_eval(quote(year > local(f(1980))), lf)

# You can use `.sql` to make it clear that the function comes from SQL,
# and inside a package, reduce the number of globalVariables() directives
# needed
partial_eval(quote(.sql$EXTRACT_YEAR(year)), lf)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-nycflights13.R
\name{nycflights13}
\alias{nycflights13}
\alias{nycflights13_sqlite}
\alias{nycflights13_postgres}
\alias{has_nycflights13}
\alias{copy_nycflights13}
\title{Database versions of the nycflights13 data}
\usage{
nycflights13_sqlite(path = NULL)

nycflights13_postgres(dbname = "nycflights13", ...)

has_nycflights13(type = c("sqlite", "postgresql"), ...)

copy_nycflights13(con, ...)
}
\arguments{
\item{path}{location of SQLite database file}

\item{dbname, ...}{Arguments passed on to \code{\link[=src_postgres]{src_postgres()}}}
}
\description{
These functions cache the data from the \code{nycflights13} database in
a local database, for use in examples and vignettes. Indexes are created
to making joining tables on natural keys efficient.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-expand.R
\name{replace_na.tbl_lazy}
\alias{replace_na.tbl_lazy}
\title{Replace NAs with specified values}
\usage{
\method{replace_na}{tbl_lazy}(data, replace = list(), ...)
}
\arguments{
\item{data}{A pair of lazy data frame backed by database queries.}

\item{replace}{A named list of values, with one value for each column that
has NA values to be replaced.}

\item{...}{Unused; included for compatibility with generic.}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
This is a method for the \code{\link[tidyr:replace_na]{tidyr::replace_na()}} generic.
}
\examples{
\dontshow{if (rlang::is_installed("tidyr", version = "1.0.0")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
df <- memdb_frame(x = c(1, 2, NA), y = c("a", NA, "b"))
df \%>\% tidyr::replace_na(list(x = 0, y = "unknown"))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-slice.R
\name{dbplyr-slice}
\alias{slice_min.tbl_lazy}
\alias{slice_max.tbl_lazy}
\alias{slice_sample.tbl_lazy}
\title{Subset rows using their positions}
\usage{
\method{slice_min}{tbl_lazy}(.data, order_by, ..., n, prop, by = NULL, with_ties = TRUE)

\method{slice_max}{tbl_lazy}(.data, order_by, ..., n, by = NULL, prop, with_ties = TRUE)

\method{slice_sample}{tbl_lazy}(.data, ..., n, prop, by = NULL, weight_by = NULL, replace = FALSE)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{order_by}{Variable or function of variables to order by.}

\item{...}{Not used.}

\item{n, prop}{Provide either \code{n}, the number of rows, or \code{prop}, the
proportion of rows to select. If neither are supplied, \code{n = 1} will be
used.

If \code{n} is greater than the number of rows in the group (or \code{prop} > 1),
the result will be silently truncated to the group size. If the proportion
of a group size is not an integer, it is rounded down.}

\item{by}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

<\code{\link[=dplyr_tidy_select]{tidy-select}}> Optionally, a selection of columns to
temporarily group by using an inline alternative to \code{\link[=group_by]{group_by()}}.}

\item{with_ties}{Should ties be kept together? The default, \code{TRUE}, may
return more rows than you request. Use FALSE to ignore ties, and return
the first n rows.}

\item{weight_by, replace}{Not supported for database backends.}
}
\description{
These are methods for the dplyr generics \code{\link[=slice_min]{slice_min()}}, \code{\link[=slice_max]{slice_max()}}, and
\code{\link[=slice_sample]{slice_sample()}}. They are translated to SQL using \code{\link[=filter]{filter()}} and
window functions (\code{ROWNUMBER}, \code{MIN_RANK}, or \code{CUME_DIST} depending on
arguments). \code{slice()}, \code{slice_head()}, and \code{slice_tail()} are not supported
since database tables have no intrinsic order.

If data is grouped, the operation will be performed on each group so that
(e.g.) \code{slice_min(db, x, n = 3)} will select the three rows with the smallest
value of \code{x} in each group.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(x = 1:3, y = c(1, 1, 2))
db \%>\% slice_min(x) \%>\% show_query()
db \%>\% slice_max(x) \%>\% show_query()
db \%>\% slice_sample() \%>\% show_query()

db \%>\% group_by(y) \%>\% slice_min(x) \%>\% show_query()

# By default, ties are includes so you may get more rows
# than you expect
db \%>\% slice_min(y, n = 1)
db \%>\% slice_min(y, n = 1, with_ties = FALSE)

# Non-integer group sizes are rounded down
db \%>\% slice_min(x, prop = 0.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-explain.R
\name{show_query}
\alias{show_query}
\alias{show_query.tbl_lazy}
\alias{explain.tbl_sql}
\title{Show generated SQL and query plan}
\usage{
\method{show_query}{tbl_lazy}(x, ..., use_colour = TRUE, sql_options = NULL, cte = deprecated())

\method{explain}{tbl_sql}(x, ...)
}
\arguments{
\item{x}{An object to explain}

\item{...}{For \code{explain()}, further arguments to \code{\link[=remote_query_plan]{remote_query_plan()}}.
For \code{show_query()}, ignored.}

\item{use_colour}{Should the output be coloured?}

\item{sql_options}{SQL rendering options generated by \code{\link[=sql_options]{sql_options()}}.}

\item{cte}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Use the \code{sql_options} argument instead.}
}
\description{
\code{show_sql()} displays the SQL query that will be dispatched to the database;
\code{explain()} displays both the SQL query and the query plan.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-mssql.R
\name{backend-mssql}
\title{Backend: SQL server}
\arguments{
\item{version}{Version of MS SQL to simulate. Currently only, difference is
that 15.0 and above will use \code{TRY_CAST()} instead of \code{CAST()}.}
}
\description{
See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology. Key differences for this backend
are:
\itemize{
\item \code{SELECT} uses \code{TOP} not \code{LIMIT}
\item Automatically prefixes \verb{#} to create temporary tables. Add the prefix
yourself to avoid the message.
\item String basics: \code{paste()}, \code{substr()}, \code{nchar()}
\item Custom types for \verb{as.*} functions
\item Lubridate extraction functions, \code{year()}, \code{month()}, \code{day()} etc
\item Semi-automated bit <-> boolean translation (see below)
}

Use \code{simulate_mssql()} with \code{lazy_frame()} to see simulated SQL without
converting to live access database.
}
\section{Bit vs boolean}{

SQL server uses two incompatible types to represent \code{TRUE} and \code{FALSE}
values:
\itemize{
\item The \code{BOOLEAN} type is the result of logical comparisons (e.g. \code{x > y})
and can be used \code{WHERE} but not to create new columns in \code{SELECT}.
\url{https://docs.microsoft.com/en-us/sql/t-sql/language-elements/comparison-operators-transact-sql}
\item The \code{BIT} type is a special type of numeric column used to store
\code{TRUE} and \code{FALSE} values, but can't be used in \code{WHERE} clauses.
\url{https://learn.microsoft.com/en-us/sql/t-sql/data-types/bit-transact-sql?view=sql-server-ver15}
}

dbplyr does its best to automatically create the correct type when needed,
but can't do it 100\% correctly because it does not have a full type
inference system. This means that you many need to manually do conversions
from time to time.
\itemize{
\item To convert from bit to boolean use \code{x == 1}
\item To convert from boolean to bit use \verb{as.logical(if(x, 0, 1))}
}
}

\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = simulate_mssql())
lf \%>\% head()
lf \%>\% transmute(x = paste(b, c, d))

# Can use boolean as is:
lf \%>\% filter(c > d)
# Need to convert from boolean to bit:
lf \%>\% transmute(x = c > d)
# Can use boolean as is:
lf \%>\% transmute(x = ifelse(c > d, "c", "d"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lazy-join-query.R, R/lazy-query.R,
%   R/lazy-select-query.R, R/lazy-set-op-query.R, R/query-join.R,
%   R/query-select.R, R/query-semi-join.R, R/query-set-op.R, R/sql-build.R
\name{lazy_multi_join_query}
\alias{lazy_multi_join_query}
\alias{lazy_semi_join_query}
\alias{lazy_query}
\alias{lazy_select_query}
\alias{lazy_set_op_query}
\alias{join_query}
\alias{select_query}
\alias{semi_join_query}
\alias{set_op_query}
\alias{sql_build}
\alias{sql_render}
\alias{sql_optimise}
\title{Build and render SQL from a sequence of lazy operations}
\usage{
lazy_multi_join_query(
  x,
  joins,
  table_names,
  vars,
  group_vars = op_grps(x),
  order_vars = op_sort(x),
  frame = op_frame(x),
  call = caller_env()
)

lazy_semi_join_query(
  x,
  y,
  vars,
  anti,
  by,
  na_matches = c("never", "na"),
  group_vars = op_grps(x),
  order_vars = op_sort(x),
  frame = op_frame(x),
  call = caller_env()
)

lazy_query(
  query_type,
  x,
  ...,
  group_vars = op_grps(x),
  order_vars = op_sort(x),
  frame = op_frame(x)
)

lazy_select_query(
  x,
  select = NULL,
  where = NULL,
  group_by = NULL,
  having = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  group_vars = NULL,
  order_vars = NULL,
  frame = NULL,
  select_operation = c("select", "mutate", "summarise"),
  message_summarise = NULL
)

lazy_set_op_query(x, y, type, all, call = caller_env())

join_query(
  x,
  y,
  vars,
  type = "inner",
  by = NULL,
  suffix = c(".x", ".y"),
  na_matches = FALSE
)

select_query(
  from,
  select = sql("*"),
  where = character(),
  group_by = character(),
  having = character(),
  window = character(),
  order_by = character(),
  limit = NULL,
  distinct = FALSE
)

semi_join_query(x, y, vars, anti = FALSE, by = NULL, na_matches = FALSE)

set_op_query(x, y, type, all = FALSE)

sql_build(op, con = NULL, ...)

sql_render(query, con = NULL, ..., subquery = FALSE, lvl = 0, cte = FALSE)

sql_optimise(x, con = NULL, ..., subquery = FALSE)
}
\arguments{
\item{...}{Other arguments passed on to the methods. Not currently used.}

\item{op}{A sequence of lazy operations}

\item{con}{A database connection. The default \code{NULL} uses a set of
rules that should be very similar to ANSI 92, and allows for testing
without an active database connection.}

\item{subquery}{Is this SQL going to be used in a subquery?
This is important because you can place a bare table name in a subquery
and  ORDER BY does not work in subqueries.}
}
\description{
\code{sql_build()} creates a \code{select_query} S3 object, that is rendered
to a SQL string by \code{sql_render()}. The output from \code{sql_build()} is
designed to be easy to test, as it's database agnostic, and has
a hierarchical structure. Outside of testing, however, you should
always call \code{sql_render()}.
}
\details{
\code{sql_build()} is generic over the lazy operations, \link{lazy_ops},
and generates an S3 object that represents the query. \code{sql_render()}
takes a query object and then calls a function that is generic
over the database. For example, \code{sql_build.op_mutate()} generates
a \code{select_query}, and \code{sql_render.select_query()} calls
\code{sql_select()}, which has different methods for different databases.
The default methods should generate ANSI 92 SQL where possible, so you
backends only need to override the methods if the backend is not ANSI
compliant.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-mutate.R
\name{mutate.tbl_lazy}
\alias{mutate.tbl_lazy}
\title{Create, modify, and delete columns}
\usage{
\method{mutate}{tbl_lazy}(
  .data,
  ...,
  .by = NULL,
  .order = NULL,
  .frame = NULL,
  .keep = c("all", "used", "unused", "none"),
  .before = NULL,
  .after = NULL
)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{...}{<\code{\link[rlang:args_data_masking]{data-masking}}> Variables, or
functions of variables. Use \code{\link[dplyr:desc]{desc()}} to sort a variable in descending
order.}

\item{.by}{<\code{\link[dplyr:dplyr_tidy_select]{tidy-select}}> Optionally, a selection of columns to
group by for just this operation, functioning as an alternative to \code{\link[dplyr:group_by]{group_by()}}. For
details and examples, see \link[dplyr:dplyr_by]{?dplyr_by}.}

\item{.order}{<\code{\link[rlang:args_data_masking]{data-masking}}> A selection of
columns to control ordering for window functions within this
\code{\link[dplyr:mutate]{mutate()}} call. Use \code{c()} to order by multiple columns, e.g.
\code{.order = c(x, y)}. Each column can be wrapped in \code{\link[dplyr:desc]{desc()}} to specify
descending order. Equivalent to calling \code{\link[=window_order]{window_order()}} before and
clearing it after the \code{\link[dplyr:mutate]{mutate()}}.}

\item{.frame}{A length-2 numeric vector specifying the bounds for
window function frames. The first element is the lower bound (use \code{-Inf}
for "unbounded preceding") and the second is the upper bound (use \code{Inf}
for "unbounded following", \code{0} for "current row"). Equivalent to calling
\code{\link[=window_frame]{window_frame()}} before and clearing it after the \code{\link[dplyr:mutate]{mutate()}}.}

\item{.keep}{Control which columns from \code{.data} are retained in the output. Grouping
columns and columns created by \code{...} are always kept.
\itemize{
\item \code{"all"} retains all columns from \code{.data}. This is the default.
\item \code{"used"} retains only the columns used in \code{...} to create new
columns. This is useful for checking your work, as it displays inputs
and outputs side-by-side.
\item \code{"unused"} retains only the columns \emph{not} used in \code{...} to create new
columns. This is useful if you generate new columns, but no longer need
the columns used to generate them.
\item \code{"none"} doesn't retain any extra columns from \code{.data}. Only the grouping
variables and columns created by \code{...} are kept.
}}

\item{.before, .after}{<\code{\link[dplyr:dplyr_tidy_select]{tidy-select}}> Optionally, control where new columns
should appear (the default is to add to the right hand side). See
\code{\link[dplyr:relocate]{relocate()}} for more details.}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
These are methods for the dplyr \code{\link[dplyr:mutate]{mutate()}} and \code{\link[dplyr:transmute]{transmute()}} generics.
They are translated to computed expressions in the \code{SELECT} clause of
the SQL query.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(x = 1:5, y = 5:1)
db |>
  mutate(a = (x + y) / 2, b = sqrt(x^2L + y^2L)) |>
  show_query()

# dbplyr automatically creates subqueries as needed
db |>
  mutate(x1 = x + 1, x2 = x1 * 2) |>
  show_query()

# `.order` and `.frame` control window functions
db <- memdb_frame(g = c(1, 1, 2, 2, 2), x = c(5, 3, 1, 4, 2))
db |>
  mutate(rolling_sum = sum(x), .by = g, .order = x, .frame = c(-2, 2)) |>
  show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sql-glue.R
\name{sql_glue}
\alias{sql_glue}
\alias{sql_glue2}
\title{Build SQL strings with glue syntax}
\usage{
sql_glue(sql, envir = parent.frame())

sql_glue2(con, sql, envir = parent.frame())
}
\arguments{
\item{sql}{A string to interpolate.}

\item{envir}{Environment to evaluate \code{sql} in.}

\item{con}{A \link{sql_dialect} object or database connection. Connections are
supported for backward compatibility.}
}
\value{
An SQL string.
}
\description{
\code{sql_glue()} and \code{sql_glue2()} are designed to help dbplyr extenders
generate custom SQL. They differ only in whether or not they require
a connection. \code{sql_glue()} retrieves the ambient connection, making it
suitable for use inside \code{\link[=sql_translation]{sql_translation()}} methods; \code{sql_glue2()} requires
a connection, making it suitable for use inside all other \code{sql_} methods.

As the name suggests, these functions use \pkg{glue} syntax to make it
easy to mix fixed SQL with varying user inputs. The default glue syntax, \code{{x}},
will escape \code{x} using the database connection. If there are multiple values
in \code{x}, they'll be collapsed into a single string with \verb{,}. If you want them
to be wrapped in \verb{()}, use a \code{*} suffix, e.g. \verb{\{x*\}}.

You can also use type markers to control how the value is treated:
\itemize{
\item \verb{\{.sql x\}}: \code{x} is literal SQL that should be interpolated as
is, without additional escaping. \code{x} must be a string.
\item \verb{\{.tbl x\}}: \code{x} is a table identifier like a string, \code{I()}, or one of
the older forms like \code{DBI::Id()} or \code{in_schema()}.
\item \verb{\{.id x\}}: \code{x} is a generic identifier, e.g. for a column or index.
\code{x} must be a character vector
}
}
\examples{
con <- dialect_ansi()

tbl <- "my_table"
sql_glue2(con, "SELECT * FROM {.tbl tbl}")

# Values are properly escaped
name <- "Robert'); DROP TABLE students;--"
sql_glue2(con, "INSERT INTO students (name) VALUES ({name})")

# Control wrapping with *
x <- c("name", "age", "grade")
sql_glue2(con, "SELECT {.id x} FROM students")
sql_glue2(con, "SELECT * WHERE variable IN {x*}")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-group_by.R
\name{group_by.tbl_lazy}
\alias{group_by.tbl_lazy}
\title{Group by one or more variables}
\usage{
\method{group_by}{tbl_lazy}(.data, ..., .add = FALSE, add = NULL, .drop = TRUE)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{...}{<\code{\link[dplyr:dplyr_data_masking]{data-masking}}> Variables, or functions of
variables. Use \code{\link[dplyr:desc]{desc()}} to sort a variable in descending order.}

\item{.add}{When \code{FALSE}, the default, \code{group_by()} will
override existing groups. To add to the existing groups, use
\code{.add = TRUE}.

This argument was previously called \code{add}, but that prevented
creating a new grouping variable called \code{add}, and conflicts with
our naming conventions.}

\item{add}{Deprecated. Please use \code{.add} instead.}

\item{.drop}{Not supported by this method.}
}
\description{
This is a method for the dplyr \code{\link[=group_by]{group_by()}} generic. It is translated to
the \verb{GROUP BY} clause of the SQL query when used with
\code{\link[=summarise.tbl_lazy]{summarise()}} and to the \verb{PARTITION BY} clause of
window functions when used with \code{\link[=mutate.tbl_lazy]{mutate()}}.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(g = c(1, 1, 1, 2, 2), x = c(4, 3, 6, 9, 2))
db \%>\%
  group_by(g) \%>\%
  summarise(n()) \%>\%
  show_query()

db \%>\%
  group_by(g) \%>\%
  mutate(x2 = x / sum(x, na.rm = TRUE)) \%>\%
  show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-compute.R
\name{compute.tbl_sql}
\alias{compute.tbl_sql}
\title{Save results into a new remote table}
\usage{
\method{compute}{tbl_sql}(
  x,
  name = NULL,
  temporary = TRUE,
  overwrite = FALSE,
  unique_indexes = list(),
  indexes = list(),
  analyze = TRUE,
  ...,
  sql_options = NULL,
  cte = deprecated()
)
}
\arguments{
\item{x}{A lazy data frame backed by a database query.}

\item{name}{Name of new remote table. Use a string to create the table
in the current catalog/schema. Use \code{I()} to create the table in a specific
catalog/schema, e.g. \code{I("schema.table")}.}

\item{temporary}{if \code{TRUE}, will create a temporary table that is
local to this connection and will be automatically deleted when the
connection expires}

\item{overwrite}{If \code{TRUE}, will overwrite an existing table with
name \code{name}. If \code{FALSE}, will throw an error if \code{name} already
exists.}

\item{unique_indexes}{a list of character vectors. Each element of the list
will create a new unique index over the specified column(s). Duplicate rows
will result in failure.}

\item{indexes}{a list of character vectors. Each element of the list
will create a new index.}

\item{analyze}{if \code{TRUE} (the default), will automatically ANALYZE the
new table so that the query optimiser has useful information.}

\item{...}{Ignored.}

\item{sql_options}{SQL rendering options generated by \code{\link[=sql_options]{sql_options()}}.}

\item{cte}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Use the \code{sql_options} argument instead.}
}
\description{
\code{compute()} executes the query and stores the results in a new remote table.
This is useful when you want to cache intermediate results for reuse or to
improve performance by avoiding repeated computation of complex queries.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(a = c(3, 4, 1, 2), b = c(5, 1, 2, NA))
db |> filter(a <= 2) |> show_query()
db |> filter(a <= 2) |> compute() |> show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/src_dbi.R
\name{tbl.src_dbi}
\alias{tbl.src_dbi}
\alias{tbl_dbi}
\title{Create a lazy query backed by a database}
\usage{
\method{tbl}{src_dbi}(src, from, vars = NULL, ...)
}
\arguments{
\item{src}{A \code{DBIConnection} object produced by \code{\link[DBI:dbConnect]{DBI::dbConnect()}}.}

\item{from}{Either a table identifier or a literal \code{\link[=sql]{sql()}} string.

Use a string to identify a table in the current schema/catalog or
\code{I()} for a table elsewhere, e.g. \code{I("schema.table")} or
\code{I("catalog.schema.table")}. For backward compatibility, you can also use
\code{\link[=in_schema]{in_schema()}}/\code{\link[=in_catalog]{in_catalog()}} or \code{\link[DBI:Id]{DBI::Id()}}.}

\item{vars}{Optionally, provide a character vector of column names. If
not supplied, will be retrieved from the database by running a simple
query. This argument is mainly useful for better performance when creating
many \code{tbl}s with known variables.}

\item{...}{Passed on to \code{\link[=tbl_sql]{tbl_sql()}}}
}
\description{
Use \code{tbl()} to create a SQL query backed by a database. Manipulating this
object with dplyr verbs then builds up a SQL query that will only be executed
when you explicitly ask for it, either by printing the object, calling
\code{\link[dplyr:collect]{collect()}} to bring the data back to R or calling \code{\link[dplyr:compute]{compute()}} to create a
new table in the database. You can see the query without executing it with
\code{\link[=show_query]{show_query()}}.

Learn more in \code{vignette("dbplyr")}.
}
\examples{
library(dplyr)

# Connect to a temporary in-memory SQLite database and add some data
con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
copy_to(con, mtcars)

# To retrieve a single table from a source, use `tbl()`
mtcars_db <- con |> tbl("mtcars")
mtcars_db

# Use `I()` for qualified table names
con |> tbl(I("temp.mtcars")) |> head(1)

# You can also pass raw SQL if you want a more sophisticated query
con |> tbl(sql("SELECT * FROM mtcars WHERE cyl = 8")) |> head(1)

# But in most cases, you'll rely on dbplyr to construct the SQL:
mtcars_db |>
  filter(vs == 1) |>
  summarise(mpg = mean(mpg, na.rm = TRUE), .by = cyl) |>
  show_query()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-uncount.R
\name{dbplyr_uncount}
\alias{dbplyr_uncount}
\title{"Uncount" a database table}
\usage{
dbplyr_uncount(data, weights, .remove = TRUE, .id = NULL)
}
\arguments{
\item{data}{A lazy data frame backed by a database query.}

\item{weights}{A vector of weights. Evaluated in the context of \code{data};
supports quasiquotation.}

\item{.remove}{If \code{TRUE}, and \code{weights} is the name of a column in \code{data},
then this column is removed.}

\item{.id}{Supply a string to create a new variable which gives a unique
identifier for each created row.}
}
\description{
This is a method for the tidyr \code{uncount()} generic. It uses a temporary
table, so your database user needs permissions to create one.
}
\examples{
df <- memdb_frame(x = c("a", "b"), n = c(1, 2))
dbplyr_uncount(df, n)
dbplyr_uncount(df, n, .id = "id")

# You can also use constants
dbplyr_uncount(df, 2)

# Or expressions
dbplyr_uncount(df, 2 / n)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backward-compatibility.R
\name{sql_optimise}
\alias{sql_optimise}
\title{Optimise a SQL query}
\usage{
sql_optimise(x, con = NULL, ..., subquery = FALSE)
}
\description{
No longer used.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-hive.R
\name{backend-hive}
\alias{dialect_hive}
\alias{simulate_hive}
\title{Hive backend}
\usage{
dialect_hive()

simulate_hive()
}
\description{
This backend supports Apache Hive, typically accessed via odbc. Use
\code{dialect_hive()} with \code{lazy_frame()} to see simulated SQL without connecting
to a live database.

Key differences for this backend are a scattering of custom translations
provided by users.

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, d = 2, c = "z", con = dialect_hive())
lf |> transmute(x = cot(b))
lf |> transmute(x = bitwShiftL(c, 1L))
lf |> transmute(x = str_replace_all(c, "a", "b"))

lf |> summarise(x = median(d, na.rm = TRUE))
lf |> summarise(x = var(c, na.rm = TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memdb.R
\name{memdb_frame}
\alias{memdb_frame}
\alias{tbl_memdb}
\alias{src_memdb}
\title{Create a database table in temporary in-memory database.}
\usage{
memdb_frame(..., .name = unique_table_name())

tbl_memdb(df, name = deparse(substitute(df)))

src_memdb()
}
\arguments{
\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}>
A set of name-value pairs. These arguments are
processed with \code{\link[rlang:defusing-advanced]{rlang::quos()}} and support unquote via \code{\link{!!}} and
unquote-splice via \code{\link{!!!}}. Use \verb{:=} to create columns that start with a dot.

Arguments are evaluated sequentially.
You can refer to previously created elements directly or using the \link{.data}
pronoun.
To refer explicitly to objects in the calling environment, use \code{\link{!!}} or
\link{.env}, e.g. \code{!!.data} or \code{.env$.data} for the special case of an object
named \code{.data}.}

\item{df}{Data frame to copy}

\item{name, .name}{Name of table in database: defaults to a random name that's
unlikely to conflict with an existing table.}
}
\description{
\code{memdb_frame()} works like \code{\link[tibble:tibble]{tibble::tibble()}}, but instead of creating a new
data frame in R, it creates a table in \code{\link[=src_memdb]{src_memdb()}}.
}
\examples{
library(dplyr)
df <- memdb_frame(x = runif(100), y = runif(100))
df \%>\% arrange(x)
df \%>\% arrange(x) \%>\% show_query()

mtcars_db <- tbl_memdb(mtcars)
mtcars_db \%>\% group_by(cyl) \%>\% summarise(n = n()) \%>\% show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-summarise.R
\name{args_by}
\alias{args_by}
\title{Helper for consistent documentation of \code{.by}}
\arguments{
\item{.by}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

<\code{\link[=dplyr_tidy_select]{tidy-select}}> Optionally, a selection of columns to
temporarily group by using an inline alternative to \code{\link[=group_by]{group_by()}}.}
}
\description{
Use \verb{@inheritParams args_by} to consistently document \code{.by}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-odbc.R
\name{backend-odbc}
\alias{dialect_odbc}
\alias{simulate_odbc}
\title{ODBC backend}
\usage{
dialect_odbc()

simulate_odbc()
}
\description{
This backend supports databases accessed via \code{OdbcConnection} created by
\code{\link[DBI:dbConnect]{DBI::dbConnect()}}. Use \code{dialect_odbc()} with \code{lazy_frame()} to see simulated
SQL without connecting to a live database.

Key differences for this backend are minor translations for common data
types.

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, d = 2, c = "z", con = dialect_odbc())
lf |> transmute(x = as.numeric(b))
lf |> transmute(x = as.integer(b))
lf |> transmute(x = as.character(b))
}
\seealso{
\code{\link[=with_dialect]{with_dialect()}} to use a different dialect if dbplyr guesses
incorrectly, or a more specific translation is available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/translate-sql-aggregate.R
\name{sql_translation_agg}
\alias{sql_translation_agg}
\alias{sql_aggregate}
\alias{sql_aggregate_2}
\alias{sql_aggregate_n}
\alias{sql_check_na_rm}
\alias{sql_not_supported}
\title{SQL helpers for aggregate functions}
\usage{
sql_aggregate(f, f_r = f)

sql_aggregate_2(f)

sql_aggregate_n(f, f_r = f)

sql_check_na_rm(na.rm)

sql_not_supported(f)
}
\arguments{
\item{f}{The name of the SQL function as a string.}

\item{f_r}{The name of the R function being translated as a string.}

\item{na.rm}{Logical indicating whether missing values should be removed.
In SQL, missing values are always removed in aggregate functions, so this
function will warn if \code{na.rm} is not \code{TRUE}.}
}
\description{
These functions help you create custom aggregate SQL translations when
implementing a new backend. They are typically used within \code{\link[=sql_translator]{sql_translator()}}
to define how R aggregate functions should be translated to SQL.
\itemize{
\item \code{sql_aggregate()} creates a SQL aggregate function translator for functions
with a single argument and an optional \code{na.rm} parameter (e.g., \code{SUM()},
\code{AVG()}).
\item \code{sql_aggregate_2()} creates a SQL aggregate function translator for
functions with exactly two arguments (e.g., \code{CORR()}, \code{COVAR_SAMP()}).
\item \code{sql_aggregate_n()} creates a SQL aggregate function translator for
functions with any number of arguments and an optional \code{na.rm} parameter
(e.g., \code{LEAST()}, \code{GREATEST()}).
\item \code{sql_not_supported()} creates a function that throws an informative error
when a function is not supported in SQL.
\item \code{sql_check_na_rm()} is a helper that you can use in aggregate functions
to direct the user towards setting \code{na.rm = TRUE}.
}
}
\seealso{
Other SQL translation helpers:
\code{\link{sql_translation_scalar}},
\code{\link{sql_translation_string}},
\code{\link{sql_translation_window}},
\code{\link[=sql_variant]{sql_variant()}}
}
\concept{SQL translation helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db-sql.R
\name{db-sql}
\alias{db-sql}
\alias{sql_expr_matches}
\alias{sql_translation}
\alias{sql_random}
\alias{sql_table_analyze}
\alias{sql_table_index}
\alias{sql_query_explain}
\alias{sql_query_fields}
\alias{sql_query_save}
\alias{sql_query_wrap}
\alias{sql_indent_subquery}
\alias{sql_query_rows}
\alias{supports_window_clause}
\alias{supports_star_without_alias}
\alias{sql_query_select}
\alias{sql_query_join}
\alias{sql_query_multi_join}
\alias{sql_query_semi_join}
\alias{sql_query_set_op}
\alias{sql_returning_cols}
\title{SQL generation generics}
\usage{
sql_expr_matches(con, x, y)

sql_translation(con)

sql_random(con)

sql_table_analyze(con, table, ...)

sql_table_index(con, table, columns, name = NULL, unique = FALSE, ...)

sql_query_explain(con, sql, ...)

sql_query_fields(con, sql, ...)

sql_query_save(con, sql, name, temporary = TRUE, ...)

sql_query_wrap(con, from, name = NULL, ..., lvl = 0)

sql_indent_subquery(from, con, lvl = 0)

sql_query_rows(con, sql, ...)

supports_window_clause(con)

supports_star_without_alias(con)

sql_query_select(
  con,
  select,
  from,
  where = NULL,
  group_by = NULL,
  having = NULL,
  window = NULL,
  order_by = NULL,
  limit = NULL,
  distinct = FALSE,
  ...,
  subquery = FALSE,
  lvl = 0
)

sql_query_join(
  con,
  x,
  y,
  vars,
  type = "inner",
  by = NULL,
  na_matches = FALSE,
  ...,
  lvl = 0
)

sql_query_multi_join(con, x, joins, table_vars, vars, ..., lvl = 0)

sql_query_semi_join(con, x, y, anti, by, vars, ..., lvl = 0)

sql_query_set_op(con, x, y, method, ..., all = FALSE, lvl = 0)

sql_returning_cols(con, cols, table, ...)
}
\description{
SQL translation:
\itemize{
\item \code{sql_expr_matches(con, x, y)} generates an alternative to \code{x = y} when a
pair of \code{NULL}s should match. The default translation uses a \verb{CASE WHEN}
as described in \url{https://modern-sql.com/feature/is-distinct-from}.
\item \code{sql_translation(con)} generates a SQL translation environment.
\item \code{sql_random(con)} generates SQL to get a random number which can be used
to select random rows in \code{slice_sample()}.
\item \code{supports_window_clause(con)} does the backend support named windows?
\item \code{supports_star_without_alias(con)} does the backend support using \code{*}
in a \code{SELECT} query without prefixing by a table alias?
}

Tables:
\itemize{
\item \code{sql_table_analyze(con, table)} generates SQL that "analyzes" the table,
ensuring that the database has up-to-date statistics for use in the query
planner. It called from \code{\link[=copy_to]{copy_to()}} when \code{analyze = TRUE}.
\item \code{sql_table_index()} generates SQL for adding an index to table.
}

Query manipulation:
\itemize{
\item \code{sql_query_explain(con, sql)} generates SQL that "explains" a query,
i.e. generates a query plan describing what indexes etc that the
database will use.
\item \code{sql_query_fields()} generates SQL for a 0-row result that is used to
capture field names in \code{\link[=tbl_sql]{tbl_sql()}}
\item \code{sql_query_save(con, sql)} generates SQL for saving a query into a
(temporary) table.
\item \code{sql_query_wrap(con, from)} generates SQL for wrapping a query into a
subquery.
}

Query indentation:
\itemize{
\item \code{sql_indent_subquery(from, con, lvl)} helps indenting a subquery.
}

Query generation:
\itemize{
\item \code{sql_query_select()} generates SQL for a \code{SELECT} query
\item \code{sql_query_join()} generates SQL for joins
\item \code{sql_query_semi_join()} generates SQL for semi- and anti-joins
\item \code{sql_query_set_op()} generates SQL for \code{UNION}, \code{INTERSECT}, and \code{EXCEPT}
queries.
}

Query generation for manipulation:
\itemize{
\item \code{sql_query_insert()} and \code{sql_query_append()} generate SQL for an \verb{INSERT FROM} query.
\item \code{sql_query_update_from()} generates SQL for an \verb{UPDATE FROM} query.
\item \code{sql_query_upsert()} generates SQL for an \code{UPSERT} query.
\item \code{sql_query_delete()} generates SQL for an \verb{DELETE FROM} query
\item \code{sql_returning_cols()} generates SQL for a \code{RETURNING} clause
}
}
\section{dbplyr 2.0.0}{


Many \verb{dplyr::db_*} generics have been replaced by \verb{dbplyr::sql_*} generics.
To update your backend, you'll need to extract the SQL generation out of your
existing code, and place it in a new method for a dbplyr \code{sql_} generic.
\itemize{
\item \code{dplyr::db_analyze()} is replaced by \code{dbplyr::sql_table_analyze()}
\item \code{dplyr::db_explain()} is replaced by \code{dbplyr::sql_query_explain()}
\item \code{dplyr::db_create_index()} is replaced by \code{dbplyr::sql_table_index()}
\item \code{dplyr::db_query_fields()} is replaced by \code{dbplyr::sql_query_fields()}
\item \code{dplyr::db_query_rows()} is no longer used; you can delete it
\item \code{dplyr::db_save_query()} is replaced by \code{dbplyr::sql_query_save()}
}

The query generating functions have also changed names. Their behaviour is
unchanged, so you just need to rename the generic and import from dbplyr
instead of dplyr.
\itemize{
\item \code{dplyr::sql_select()} is replaced by \code{dbplyr::sql_query_select()}
\item \code{dplyr::sql_join()} is replaced by \code{dbplyr::sql_query_join()}
\item \code{dplyr::sql_semi_join()} is replaced by \code{dbplyr::sql_query_semi_join()}
\item \code{dplyr::sql_set_op()} is replaced by \code{dbplyr::sql_query_set_op()}
\item \code{dplyr::sql_subquery()} is replaced by \code{dbplyr::sql_query_wrap()}
}

Learn more in \code{vignette("backend-2.0")}
}

\seealso{
Other generic: 
\code{\link{db_connection_describe}()},
\code{\link{db_copy_to}()},
\code{\link{sql_escape_logical}()}
}
\concept{generic}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rows.R
\name{get_returned_rows}
\alias{get_returned_rows}
\alias{has_returned_rows}
\title{Extract and check the \code{RETURNING} rows}
\usage{
get_returned_rows(x)

has_returned_rows(x)
}
\arguments{
\item{x}{A lazy tbl.}
}
\value{
For \code{get_returned_rows()}, a tibble.

For \code{has_returned_rows()}, a scalar logical.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{get_returned_rows()} extracts the \code{RETURNING} rows produced by
\code{\link[=rows_insert]{rows_insert()}}, \code{\link[=rows_append]{rows_append()}}, \code{\link[=rows_update]{rows_update()}}, \code{\link[=rows_upsert]{rows_upsert()}},
or \code{\link[=rows_delete]{rows_delete()}} if these are called with the \code{returning} argument.
An error is raised if this information is not available.

\code{has_returned_rows()} checks if \code{x} has stored RETURNING rows produced by
\code{\link[=rows_insert]{rows_insert()}}, \code{\link[=rows_append]{rows_append()}}, \code{\link[=rows_update]{rows_update()}}, \code{\link[=rows_upsert]{rows_upsert()}},
or \code{\link[=rows_delete]{rows_delete()}}.
}
\examples{
library(dplyr)

con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
DBI::dbExecute(con, "CREATE TABLE Info (
   id INTEGER PRIMARY KEY AUTOINCREMENT,
   number INTEGER
)")
info <- tbl(con, "Info")

rows1 <- copy_inline(con, data.frame(number = c(1, 5)))
rows_insert(info, rows1, conflict = "ignore", in_place = TRUE)
info

# If the table has an auto incrementing primary key, you can use
# the returning argument + `get_returned_rows()` its value
rows2 <- copy_inline(con, data.frame(number = c(13, 27)))
info <- rows_insert(
  info,
  rows2,
  conflict = "ignore",
  in_place = TRUE,
  returning = id
)
info
get_returned_rows(info)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-db2.R
\name{backend-db2}
\alias{dialect_db2}
\title{DB2 backend}
\usage{
dialect_db2()
}
\description{
This backend supports IBM DB2 databases, typically accessed via ODBC.
Use \code{dialect_db2()} with \code{lazy_frame()} to see simulated SQL without
connecting to a live database.

Key differences for this backend are:
\itemize{
\item Uses \verb{FETCH FIRST n ROWS ONLY} instead of \verb{LIMIT n}
\item Uses double quotes for identifier quoting
\item \code{paste()} uses \code{||}
\item DB2-specific data type names for casts (e.g. \code{VARCHAR(255)}, \code{DOUBLE})
\item Date component extraction via \code{YEAR()}, \code{MONTH()}, ..., \code{DAYOFYEAR()},
\code{DAYOFWEEK()}, \code{QUARTER()}, \code{WEEK()}
\item \code{str_flatten()} uses \code{LISTAGG}
\item Statistical summaries \code{sd()}, \code{var()}, \code{cor()}, \code{cov()}
\item \code{runif()} translates to \code{RAND()}
\item Regular expression functions (DB2 11.1+)
}

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_db2())
lf |> head()
lf |> transmute(x = paste0(d, " times"))
lf |> summarise(x = sd(b, na.rm = TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-distinct.R
\name{distinct.tbl_lazy}
\alias{distinct.tbl_lazy}
\title{Subset distinct/unique rows}
\usage{
\method{distinct}{tbl_lazy}(.data, ..., .keep_all = FALSE)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{...}{<\code{\link[dplyr:dplyr_data_masking]{data-masking}}> Variables, or functions of
variables. Use \code{\link[dplyr:desc]{desc()}} to sort a variable in descending order.}

\item{.keep_all}{If \code{TRUE}, keep all variables in \code{.data}.
If a combination of \code{...} is not distinct, this keeps the
first row of values.}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
This is a method for the dplyr \code{\link[=distinct]{distinct()}} generic. It adds the
\code{DISTINCT} clause to the SQL query.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(x = c(1, 1, 2, 2), y = c(1, 2, 1, 1))
db \%>\% distinct() \%>\% show_query()
db \%>\% distinct(x) \%>\% show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-sql.R
\name{build_sql}
\alias{build_sql}
\title{Build a SQL string.}
\usage{
build_sql(..., .env = parent.frame(), con = sql_current_con())
}
\arguments{
\item{...}{input to convert to SQL. Use \code{\link[=sql]{sql()}} to preserve
user input as is (dangerous), and \code{\link[=ident]{ident()}} to label user
input as sql identifiers (safe)}

\item{.env}{the environment in which to evaluate the arguments. Should not
be needed in typical use.}

\item{con}{database connection; used to select correct quoting characters.}
}
\description{
This is a convenience function that should prevent sql injection attacks
(which in the context of dplyr are most likely to be accidental not
deliberate) by automatically escaping all expressions in the input, while
treating bare strings as sql. This is unlikely to prevent any serious
attack, but should make it unlikely that you produce invalid sql.
}
\details{
This function should be used only when generating \code{SELECT} clauses,
other high level queries, or for other syntax that has no R equivalent.
For individual function translations, prefer \code{\link[=sql_expr]{sql_expr()}}.
}
\examples{
con <- simulate_dbi()
build_sql("SELECT * FROM TABLE", con = con)
x <- "TABLE"
build_sql("SELECT * FROM ", x, con = con)
build_sql("SELECT * FROM ", ident(x), con = con)
build_sql("SELECT * FROM ", sql(x), con = con)

# http://xkcd.com/327/
name <- "Robert'); DROP TABLE Students;--"
build_sql("INSERT INTO Students (Name) VALUES (", name, ")", con = con)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-teradata.R
\name{backend-teradata}
\alias{simulate_teradata}
\title{Backend: Teradata}
\usage{
simulate_teradata()
}
\description{
See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology. Key differences for this backend
are:
\itemize{
\item Uses \code{TOP} instead of \code{LIMIT}
\item Selection of user supplied translations
}

Use \code{simulate_teradata()} with \code{lazy_frame()} to see simulated SQL without
converting to live access database.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = simulate_teradata())
lf \%>\% head()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{in_schema}
\alias{in_schema}
\alias{in_catalog}
\title{Refer to a table in another schema/catalog}
\usage{
in_schema(schema, table)

in_catalog(catalog, schema, table)
}
\arguments{
\item{catalog, schema, table}{Names of catalog, schema, and table.
These will be automatically quoted; use \code{\link[=sql]{sql()}} to pass a raw name
that won't get quoted.}
}
\description{
\code{in_schema()} and \code{in_catalog()} can be used to refer to tables outside of
the current catalog/schema. However, we now recommend using \code{I()} as it's
typically less typing.
}
\examples{
# Previously:
in_schema("my_schema", "my_table")
in_catalog("my_catalog", "my_schema", "my_table")
in_schema(sql("my_schema"), sql("my_table"))

# Now
I("my_schema.my_table")
I("my_catalog.my_schema.my_table")
I("my_schema.my_table")

# Example using schemas with SQLite
con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")

# Add auxiliary schema
tmp <- tempfile()
DBI::dbExecute(con, paste0("ATTACH '", tmp, "' AS aux"))

library(dplyr, warn.conflicts = FALSE)
copy_to(con, iris, "df", temporary = FALSE)
copy_to(con, mtcars, I("aux.df"), temporary = FALSE)

con \%>\% tbl("df")
con \%>\% tbl(I("aux.df"))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-select.R
\name{simulate_vars}
\alias{simulate_vars}
\alias{simulate_vars_is_typed}
\title{Simulate variables to use in tidyselect}
\usage{
simulate_vars(x, drop_groups = FALSE)

simulate_vars_is_typed(x)
}
\arguments{
\item{x}{A lazy table}

\item{drop_groups}{Should groups be dropped?}
}
\value{
A 0 row tibble with the same columns names, and, if possible, types, as \code{x}.
}
\description{
Simulate variables to use in tidyselect
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-hive.R, R/backend-impala.R,
%   R/backend-mssql.R, R/simulate.R
\name{simulate_hive}
\alias{simulate_hive}
\alias{simulate_impala}
\alias{simulate_mssql}
\alias{simulate_dbi}
\title{Simulate database connections}
\usage{
simulate_hive()

simulate_impala()

simulate_mssql(version = "15.0")

simulate_dbi(class = character(), ...)
}
\description{
These functions generate S3 objects that have been designed to simulate
the action of a database connection, without actually having the database
available. Obviously, this simulation can only be incomplete, but most
importantly it allows us to simulate SQL generation for any database without
actually connecting to it.
}
\details{
Simulated SQL always quotes identifies with \code{`x`}, and strings with
\code{'x'}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-access.R
\name{backend-access}
\alias{simulate_access}
\title{Backend: MS Access}
\usage{
simulate_access()
}
\description{
See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology. Key differences for this backend
are:
\itemize{
\item \code{SELECT} uses \code{TOP}, not \code{LIMIT}
\item Non-standard types and mathematical functions
\item String concatenation uses \code{&}
\item No \code{ANALYZE} equivalent
\item \code{TRUE} and \code{FALSE} converted to 1 and 0
}

Use \code{simulate_access()} with \code{lazy_frame()} to see simulated SQL without
converting to live access database.
}
\examples{
library(dplyr, warn.conflicts = FALSE)
lf <- lazy_frame(x = 1, y = 2, z = "a", con = simulate_access())

lf \%>\% head()
lf \%>\% mutate(y = as.numeric(y), z = sqrt(x^2 + 10))
lf \%>\% mutate(a = paste0(z, " times"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test-frame.R
\name{testing}
\alias{testing}
\alias{test_register_src}
\alias{test_register_con}
\alias{src_test}
\alias{test_load}
\alias{test_frame}
\title{Infrastructure for testing dplyr}
\usage{
test_register_src(name, src)

test_register_con(name, ...)

src_test(name)

test_load(
  df,
  name = unique_table_name(),
  srcs = test_srcs$get(),
  ignore = character()
)

test_frame(..., srcs = test_srcs$get(), ignore = character())
}
\description{
Register testing sources, then use \code{test_load()} to load an existing
data frame into each source. To create a new table in each source,
use \code{test_frame()}.
}
\examples{
\dontrun{
test_register_src("sqlite", {
  DBI::dbConnect(RSQLite::SQLite(), ":memory:", create = TRUE)
})

test_frame(x = 1:3, y = 3:1)
test_load(mtcars)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/translate-sql-scalar.R
\name{sql_translation_scalar}
\alias{sql_translation_scalar}
\alias{sql_infix}
\alias{sql_prefix}
\alias{sql_cast}
\alias{sql_try_cast}
\alias{sql_log}
\alias{sql_cot}
\alias{sql_runif}
\title{SQL helpers for scalar functions}
\usage{
sql_infix(f, pad = TRUE)

sql_prefix(f, n = NULL)

sql_cast(type)

sql_try_cast(type)

sql_log()

sql_cot()

sql_runif(rand_expr, n = n(), min = 0, max = 1)
}
\arguments{
\item{f}{The name of the SQL function as a string.}

\item{pad}{If \code{TRUE}, the default, pad the infix operator with spaces.}

\item{n}{For \code{sql_prefix()}, an optional number of arguments to expect.
Will signal error if not correct.}

\item{type}{SQL type name as a string.}

\item{rand_expr}{A string giving an SQL expression that generates a
random number between 0 and 1, e.g. \code{"RANDOM()"}.}

\item{min, max}{Range of random values.}
}
\description{
These functions help you create custom scalar SQL translations when
implementing a new backend. They are typically used within \code{\link[=sql_translator]{sql_translator()}}
to define how R functions should be translated to SQL.
\itemize{
\item \code{sql_infix()} creates SQL infix operators like \code{+}, \code{-}, \code{*}, \code{/}.
\item \code{sql_prefix()} creates SQL prefix functions like \code{ABS()}, \code{SQRT()}.
\item \code{sql_cast()} creates SQL cast expressions like \verb{CAST(x AS type)}.
\item \code{sql_try_cast()} creates SQL try_cast expressions (for safe casting).
\item \code{sql_log()} creates a SQL logarithm function with optional base.
\item \code{sql_cot()} creates a SQL cotangent function (as \code{1 / TAN(x)}).
\item \code{sql_runif()} creates a SQL expression for generating uniform random
numbers.
}
}
\seealso{
Other SQL translation helpers:
\code{\link{sql_translation_agg}},
\code{\link{sql_translation_string}},
\code{\link{sql_translation_window}},
\code{\link[=sql_variant]{sql_variant()}}
}
\concept{SQL translation helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl-sql.R
\name{tbl_sql}
\alias{tbl_sql}
\title{Create an SQL tbl (abstract)}
\usage{
tbl_sql(subclass, src, from, ..., vars = NULL, check_from = deprecated())
}
\arguments{
\item{subclass}{name of subclass}

\item{src}{A \code{DBIConnection} object produced by \code{\link[DBI:dbConnect]{DBI::dbConnect()}}.}

\item{from}{Either a table identifier or a literal \code{\link[=sql]{sql()}} string.

Use a string to identify a table in the current schema/catalog or
\code{I()} for a table elsewhere, e.g. \code{I("schema.table")} or
\code{I("catalog.schema.table")}. For backward compatibility, you can also use
\code{\link[=in_schema]{in_schema()}}/\code{\link[=in_catalog]{in_catalog()}} or \code{\link[DBI:Id]{DBI::Id()}}.}

\item{...}{needed for agreement with generic. Not otherwise used.}

\item{vars}{Optionally, provide a character vector of column names. If
not supplied, will be retrieved from the database by running a simple
query. This argument is mainly useful for better performance when creating
many \code{tbl}s with known variables.}

\item{check_from}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}}
}
\description{
This function creates a lazy tbl object from a table in a database.
It's primarily intended for backend authors who need to create custom
subclasses; most users should use \code{\link[dplyr:tbl]{tbl()}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-expand.R
\name{expand.tbl_lazy}
\alias{expand.tbl_lazy}
\title{Expand SQL tables to include all possible combinations of values}
\usage{
\method{expand}{tbl_lazy}(data, ..., .name_repair = "check_unique")
}
\arguments{
\item{data}{A lazy data frame backed by a database query.}

\item{...}{Specification of columns to expand. See \link[tidyr:expand]{tidyr::expand} for
more details.}

\item{.name_repair}{Treatment of problematic column names:
\itemize{
\item \code{"minimal"}: No name repair or checks, beyond basic existence,
\item \code{"unique"}: Make sure names are unique and not empty,
\item \code{"check_unique"}: (default value), no name repair, but check they are
\code{unique},
\item \code{"universal"}: Make the names \code{unique} and syntactic
\item a function: apply custom name repair (e.g., \code{.name_repair = make.names}
for names in the style of base R).
\item A purrr-style anonymous function, see \code{\link[rlang:as_function]{rlang::as_function()}}
}

This argument is passed on as \code{repair} to \code{\link[vctrs:vec_as_names]{vctrs::vec_as_names()}}.
See there for more details on these terms and the strategies used
to enforce them.}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
This is a method for the \link[tidyr:expand]{tidyr::expand} generics. It doesn't sort the
result explicitly, so the order might be different to what \code{expand()}
returns for data frames.
}
\examples{
\dontshow{if (rlang::is_installed("tidyr", version = "1.0.0")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
fruits <- memdb_frame(
  type   = c("apple", "orange", "apple", "orange", "orange", "orange"),
  year   = c(2010, 2010, 2012, 2010, 2010, 2012),
  size = c("XS", "S",  "M", "S", "S", "M"),
  weights = rnorm(6)
)

# All possible combinations ---------------------------------------
fruits \%>\% tidyr::expand(type)
fruits \%>\% tidyr::expand(type, size)

# Only combinations that already appear in the data ---------------
fruits \%>\% tidyr::expand(nesting(type, size))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-.R
\name{backend-ansi}
\alias{dialect_ansi}
\alias{simulate_dbi}
\title{ANSI SQL backend}
\usage{
dialect_ansi()

simulate_dbi(class = character(), ...)
}
\arguments{
\item{class, ...}{No longer used.}
}
\description{
This is the base dialect for ANSI compliant SQL, forming the foundation
of all other dialects. Use \code{dialect_ansi()} with \code{lazy_frame()} to see
simulated SQL without connecting to a live database.

See \code{vignette("translation-function")} for a list of functions that are
translated.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_ansi())
lf |> transmute(x = mean(b, na.rm = TRUE))
lf |> transmute(x = log(b), y = log(b, base = 2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{named_commas}
\alias{named_commas}
\title{Provides comma-separated string out of the parameters}
\usage{
named_commas(...)
}
\arguments{
\item{...}{Arguments to be constructed into the string}
}
\description{
Provides comma-separated string out of the parameters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/translate-sql-helpers.R, R/backend-.R,
%   R/backend-odbc.R
\name{sql_variant}
\alias{sql_variant}
\alias{sql_translator}
\alias{base_scalar}
\alias{base_agg}
\alias{base_win}
\alias{base_no_win}
\alias{base_odbc_scalar}
\alias{base_odbc_agg}
\alias{base_odbc_win}
\title{Create an SQL translator}
\usage{
sql_variant(
  scalar = sql_translator(),
  aggregate = sql_translator(),
  window = sql_translator()
)

sql_translator(..., .funs = list(), .parent = new.env(parent = emptyenv()))

base_scalar

base_agg

base_win

base_no_win

base_odbc_scalar

base_odbc_agg

base_odbc_win
}
\arguments{
\item{scalar, aggregate, window}{The three families of functions that an SQL
variant can supply.}

\item{..., .funs}{Named functions, used to add custom converters from standard
R functions to SQL functions. Specify individually in \code{...}, or provide a
list of \code{.funs}.}

\item{.parent}{The SQL variant that this variant should inherit from.
Defaults to \code{base_agg} which provides a standard set of mappings for the
most common operators and functions.}
}
\description{
\code{sql_variant()} creates a SQL variant, a list of translators for scalar,
aggregate, and window functions. \code{sql_translator()} creates a translator,
an environment containing R to SQL translations. When creating a backend,
you'll use these functions to customize how R functions are converted to
SQL.

Learn more in \code{vignette("new-backend")}.
}
\section{Base translators}{


dbplyr provides the following base translators that implement standard SQL
semantics:
\itemize{
\item \code{base_scalar} - scalar functions and operators
\item \code{base_agg} - aggregate functions
\item \code{base_win} - window functions
\item \code{base_no_win} - versions of window functions that throw errors
}
}

\examples{
# An example of adding some mappings for the statistical functions that
# postgresql provides: http://bit.ly/K5EdTn
postgres_agg <- sql_translator(.parent = base_agg,
  cor = sql_aggregate_2("CORR"),
  cov = sql_aggregate_2("COVAR_SAMP"),
  sd =  sql_aggregate("STDDEV_SAMP", "sd"),
  var = sql_aggregate("VAR_SAMP", "var")
)

# Next we have to simulate a connection that uses this variant
con <- new_sql_dialect("test", quote = \(x) sql_quote(x, '"'))
sql_translation.sql_dialect_test <- function(x) {
  sql_variant(
    base_scalar,
    postgres_agg,
    base_no_win
  )
}

translate_sql(cor(x, y), con = con, window = FALSE)
translate_sql(sd(income / years, na.rm = TRUE), con = con, window = FALSE)
}
\seealso{
Other SQL translation helpers:
\code{\link{sql_translation_agg}},
\code{\link{sql_translation_scalar}},
\code{\link{sql_translation_string}},
\code{\link{sql_translation_window}}
}
\concept{SQL translation helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db.R
\name{sql_options}
\alias{sql_options}
\title{Options for generating SQL}
\usage{
sql_options(cte = FALSE, use_star = TRUE, qualify_all_columns = FALSE)
}
\arguments{
\item{cte}{If \code{FALSE}, the default, subqueries are used. If \code{TRUE} common
table expressions are used.}

\item{use_star}{If \code{TRUE}, the default, \code{*} is used to select all columns of
a table. If \code{FALSE} all columns are explicitly selected.}

\item{qualify_all_columns}{If \code{FALSE}, the default, columns are only
qualified with the table they come from if the same column name appears in
multiple tables.}
}
\value{
A <dbplyr_sql_options> object.
}
\description{
Options for generating SQL
}
\examples{
library(dplyr, warn.conflicts = FALSE)
lf1 <- lazy_frame(key = 1, a = 1, b = 2)
lf2 <- lazy_frame(key = 1, a = 1, c = 3)

result <- left_join(lf1, lf2, by = "key") \%>\%
  filter(c >= 3)

show_query(result)
sql_options <- sql_options(cte = TRUE, qualify_all_columns = TRUE)
show_query(result, sql_options = sql_options)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rows.R
\name{rows_insert.tbl_lazy}
\alias{rows_insert.tbl_lazy}
\alias{rows_append.tbl_lazy}
\alias{rows_update.tbl_lazy}
\alias{rows_patch.tbl_lazy}
\alias{rows_upsert.tbl_lazy}
\alias{rows_delete.tbl_lazy}
\title{Edit individual rows in the underlying database table}
\usage{
\method{rows_insert}{tbl_lazy}(
  x,
  y,
  by = NULL,
  ...,
  conflict = c("error", "ignore"),
  copy = FALSE,
  in_place = FALSE,
  returning = NULL,
  method = NULL
)

\method{rows_append}{tbl_lazy}(x, y, ..., copy = FALSE, in_place = FALSE, returning = NULL)

\method{rows_update}{tbl_lazy}(
  x,
  y,
  by = NULL,
  ...,
  unmatched = c("error", "ignore"),
  copy = FALSE,
  in_place = FALSE,
  returning = NULL
)

\method{rows_patch}{tbl_lazy}(
  x,
  y,
  by = NULL,
  ...,
  unmatched = c("error", "ignore"),
  copy = FALSE,
  in_place = FALSE,
  returning = NULL
)

\method{rows_upsert}{tbl_lazy}(
  x,
  y,
  by = NULL,
  ...,
  copy = FALSE,
  in_place = FALSE,
  returning = NULL,
  method = NULL
)

\method{rows_delete}{tbl_lazy}(
  x,
  y,
  by = NULL,
  ...,
  unmatched = c("error", "ignore"),
  copy = FALSE,
  in_place = FALSE,
  returning = NULL
)
}
\arguments{
\item{x}{A lazy table.
For \code{in_place = TRUE}, this must be a table instantiated with \code{\link[=tbl]{tbl()}} or
\code{\link[=compute]{compute()}}, not to a lazy query. The \code{\link[=remote_name]{remote_name()}} function is used to
determine the name of the table to be updated.}

\item{y}{A lazy table, data frame, or data frame extensions (e.g. a tibble).}

\item{by}{An unnamed character vector giving the key columns. The key columns
must exist in both \code{x} and \code{y}. Keys typically uniquely identify each row,
but this is only enforced for the key values of \code{y} when \code{rows_update()},
\code{rows_patch()}, or \code{rows_upsert()} are used.

By default, we use the first column in \code{y}, since the first column is
a reasonable place to put an identifier variable.}

\item{...}{Other parameters passed onto methods.}

\item{conflict}{For \code{rows_insert()}, how should keys in \code{y} that conflict
with keys in \code{x} be handled? A conflict arises if there is a key in \code{y}
that already exists in \code{x}.

One of:
\itemize{
\item \code{"error"}, the default, is not supported for database tables. To get the
same behaviour add a unique index on the \code{by} columns and use
\code{rows_append()}.
\item \code{"ignore"} will ignore rows in \code{y} with keys that conflict with keys in
\code{x}.
}}

\item{copy}{If \code{x} and \code{y} are not from the same data source,
and \code{copy} is \code{TRUE}, then \code{y} will be copied into the
same src as \code{x}.  This allows you to join tables across srcs, but
it is a potentially expensive operation so you must opt into it.}

\item{in_place}{Should \code{x} be modified in place? If \code{FALSE} will
generate a \code{SELECT} query that returns the modified table; if \code{TRUE}
will modify the underlying table using a DML operation (\code{INSERT}, \code{UPDATE},
\code{DELETE} or similar).}

\item{returning}{Columns to return. See \code{\link[=get_returned_rows]{get_returned_rows()}} for details.}

\item{method}{A string specifying the method to use. This is only relevant for
\code{in_place = TRUE}.}

\item{unmatched}{For \code{rows_update()}, \code{rows_patch()}, and \code{rows_delete()},
how should keys in \code{y} that are unmatched by the keys in \code{x} be handled?

One of:
\itemize{
\item \code{"error"}, the default, is not supported for database tables. Add a
foreign key constraint on the \code{by} columns of \code{y} to let the database
check this behaviour for you.
\item \code{"ignore"} will ignore rows in \code{y} with keys that are unmatched by the
keys in \code{x}.
}}
}
\value{
A new \code{tbl_lazy} of the modified data.
With \code{in_place = FALSE}, the result is a lazy query that prints visibly,
because the purpose of this operation is to preview the results.
With \code{in_place = TRUE}, \code{x} is returned invisibly,
because the purpose of this operation is the side effect of modifying rows
in the table behind \code{x}.
}
\description{
These are methods for the dplyr \code{\link[=rows_insert]{rows_insert()}}, \code{\link[=rows_append]{rows_append()}},
\code{\link[=rows_update]{rows_update()}}, \code{\link[=rows_patch]{rows_patch()}}, \code{\link[=rows_upsert]{rows_upsert()}}, and \code{\link[=rows_delete]{rows_delete()}}
generics.

When \code{in_place = TRUE} these verbs do not generate \code{SELECT} queries, but
instead directly modify the underlying data using \code{INSERT}, \code{UPDATE}, or
\code{DELETE} operators. This will require that you have write access to
the database: the connection needs permission to insert, modify or delete
rows, but not to alter the structure of the table.

The default, \code{in_place = FALSE}, generates equivalent lazy tables (using
\code{SELECT} queries) that allow previewing the result without actually
modifying the underlying table on the database.
}
\examples{
library(dplyr)

con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
DBI::dbExecute(con, "CREATE TABLE Ponies (
   id INTEGER PRIMARY KEY AUTOINCREMENT,
   name TEXT,
   cutie_mark TEXT
)")

ponies <- tbl(con, "Ponies")

applejack <- copy_inline(con, data.frame(
  name = "Apple Jack",
  cutie_mark = "three apples"
))

# The default behavior is to generate a SELECT query
rows_insert(ponies, applejack, conflict = "ignore")
# And the original table is left unchanged:
ponies

# You can also choose to modify the table with in_place = TRUE:
rows_insert(ponies, applejack, conflict = "ignore", in_place = TRUE)
# In this case `rows_insert()` returns nothing and the underlying
# data is modified
ponies
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-collect.R
\name{collect.tbl_sql}
\alias{collect.tbl_sql}
\title{Collect results into a local data frame}
\usage{
\method{collect}{tbl_sql}(
  x,
  ...,
  n = Inf,
  warn_incomplete = TRUE,
  sql_options = NULL,
  cte = deprecated()
)
}
\arguments{
\item{x}{A lazy data frame backed by a database query.}

\item{...}{Ignored.}

\item{n}{Number of rows to fetch. Defaults to \code{Inf}, meaning all rows.}

\item{warn_incomplete}{Warn if \code{n} is less than the number of result rows?}

\item{sql_options}{SQL rendering options generated by \code{\link[=sql_options]{sql_options()}}.}

\item{cte}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
Use the \code{sql_options} argument instead.}
}
\description{
\code{collect()} executes the query and retrieves the results into a local tibble.
This brings all the data from the database into R's memory, which is useful
once you've done as much as possible in the database, and now need to use
R functions.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(a = c(3, 4, 1, 2), b = c(5, 1, 2, NA))
db |> filter(a <= 2) |> collect()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-oracle.R
\name{backend-oracle}
\alias{dialect_oracle}
\alias{simulate_oracle}
\title{Oracle backend}
\usage{
dialect_oracle()

simulate_oracle()
}
\description{
This backend supports Oracle databases, typically accessed via
\code{OraConnection} created by \code{\link[DBI:dbConnect]{DBI::dbConnect()}}. Use \code{dialect_oracle()} with
\code{lazy_frame()} to see simulated SQL without connecting to a live database.

Key differences for this backend are:
\itemize{
\item Use \verb{FETCH FIRST} instead of \code{LIMIT}
\item Custom types
\item \code{paste()} uses \code{||}
\item Custom subquery generation (no \code{AS})
\item \code{setdiff()} uses \code{MINUS} instead of \code{EXCEPT}
}

Note that versions of Oracle prior to 23c have limited supported for
\code{TRUE} and \code{FALSE} and you may need to use \code{1} and \code{0} instead.
See \url{https://oracle-base.com/articles/23/boolean-data-type-23} for
more details.

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_oracle())
lf |> transmute(x = paste0(c, " times"))
lf |> setdiff(lf)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-count.R
\name{count.tbl_lazy}
\alias{count.tbl_lazy}
\alias{add_count.tbl_lazy}
\alias{tally.tbl_lazy}
\title{Count observations by group}
\usage{
\method{count}{tbl_lazy}(x, ..., wt = NULL, sort = FALSE, name = NULL)

\method{add_count}{tbl_lazy}(x, ..., wt = NULL, sort = FALSE, name = NULL, .drop = NULL)

\method{tally}{tbl_lazy}(x, wt = NULL, sort = FALSE, name = NULL)
}
\arguments{
\item{x}{A data frame, data frame extension (e.g. a tibble), or a
lazy data frame (e.g. from dbplyr or dtplyr).}

\item{...}{<\code{\link[dplyr:dplyr_data_masking]{data-masking}}> Variables, or functions of
variables. Use \code{\link[dplyr:desc]{desc()}} to sort a variable in descending order.}

\item{wt}{<\code{\link[dplyr:dplyr_data_masking]{data-masking}}> Frequency weights.
Can be \code{NULL} or a variable:
\itemize{
\item If \code{NULL} (the default), counts the number of rows in each group.
\item If a variable, computes \code{sum(wt)} for each group.
}}

\item{sort}{If \code{TRUE}, will show the largest groups at the top.}

\item{name}{The name of the new column in the output.

If omitted, it will default to \code{n}. If there's already a column called \code{n},
it will error, and require you to specify the name.}

\item{.drop}{Not supported for lazy tables.}
}
\description{
These are methods for the dplyr \code{\link[=count]{count()}} and \code{\link[=tally]{tally()}} generics. They
wrap up \code{\link[=group_by.tbl_lazy]{group_by.tbl_lazy()}}, \code{\link[=summarise.tbl_lazy]{summarise.tbl_lazy()}} and, optionally,
\code{\link[=arrange.tbl_lazy]{arrange.tbl_lazy()}}.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(g = c(1, 1, 1, 2, 2), x = c(4, 3, 6, 9, 2))
db \%>\% count(g) \%>\% show_query()
db \%>\% count(g, wt = x) \%>\% show_query()
db \%>\% count(g, wt = x, sort = TRUE) \%>\% show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-spark-sql.R
\name{backend-spark-sql}
\alias{dialect_spark_sql}
\alias{simulate_spark_sql}
\title{Databricks Spark SQL backend}
\usage{
dialect_spark_sql()

simulate_spark_sql()
}
\description{
This backend supports Databricks Spark SQL, typically accessed via the
Databricks ODBC or JDBC connector. Use \code{dialect_spark_sql()} with
\code{lazy_frame()} to see simulated SQL without connecting to a live database.

Key differences for this backend are better translation of statistical
aggregate functions (e.g. \code{var()}, \code{median()}) and use of temporary views
instead of temporary tables when copying data.

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, d = 2, c = "z", con = dialect_spark_sql())

lf |> summarise(x = median(d, na.rm = TRUE))
lf |> summarise(x = var(c, na.rm = TRUE), .by = d)

lf |> mutate(x = first(c))
lf |> mutate(x = first(c), .by = d)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sql-superseded.R
\name{ident_q}
\alias{ident_q}
\title{Declare a identifier as being pre-quoted.}
\usage{
ident_q(...)
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

No longer needed; please use \code{\link[=sql]{sql()}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sql-expr.R
\name{sql_expr}
\alias{sql_expr}
\alias{sql_call2}
\title{Generate SQL from R expressions}
\usage{
sql_expr(x, con = sql_current_con())

sql_call2(.fn, ..., con = sql_current_con())
}
\arguments{
\item{x}{A quasiquoted expression}

\item{con}{Connection to use for escaping. Will be set automatically when
called from a function translation.}

\item{.fn}{Function name (as string, call, or symbol)}

\item{...}{Arguments to function}
}
\description{
Low-level building block for generating SQL from R expressions.
Strings are escaped; names become bare SQL identifiers. User infix
functions have \verb{\%} stripped.
}
\details{
Using \code{sql_expr()} in package will require use of \code{\link[=globalVariables]{globalVariables()}}
to avoid \verb{R CMD check} NOTES. This is a small amount of additional pain,
which I think is worthwhile because it leads to more readable translation
code.
}
\examples{
con <- simulate_dbi() # not necessary when writing translations

sql_expr(f(x + 1), con = con)
sql_expr(f("x", "y"), con = con)
sql_expr(f(x, y), con = con)

x <- ident("x")
sql_expr(f(!!x, y), con = con)

sql_expr(cast("x" \%as\% DECIMAL), con = con)
sql_expr(round(x) \%::\% numeric, con = con)

sql_call2("+", quote(x), 1, con = con)
sql_call2("+", "x", 1, con = con)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db-escape.R
\name{db-quote}
\alias{sql_escape_logical}
\alias{sql_escape_date}
\alias{sql_escape_datetime}
\alias{sql_escape_raw}
\title{SQL escaping/quoting generics}
\usage{
sql_escape_logical(con, x)

sql_escape_date(con, x)

sql_escape_datetime(con, x)

sql_escape_raw(con, x)
}
\description{
These generics translate individual values into SQL. The core
generics are \code{\link[DBI:dbQuoteIdentifier]{DBI::dbQuoteIdentifier()}} and\link[DBI:dbQuoteString]{DBI::dbQuoteString}
for quoting identifiers and strings, but dbplyr needs additional
tools for inserting logical, date, date-time, and raw values into
queries.
}
\examples{
con <- simulate_dbi()
sql_escape_logical(con, c(TRUE, FALSE, NA))
sql_escape_date(con, Sys.Date())
sql_escape_date(con, Sys.time())
sql_escape_raw(con, charToRaw("hi"))
}
\seealso{
Other generic: 
\code{\link{db-sql}},
\code{\link{db_connection_describe}()},
\code{\link{db_copy_to}()}
}
\concept{generic}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/translate-sql-string.R
\name{sql_translation_string}
\alias{sql_translation_string}
\alias{sql_substr}
\alias{sql_str_sub}
\alias{sql_paste}
\alias{sql_paste_infix}
\title{SQL helpers for string functions}
\usage{
sql_substr(f = "SUBSTR")

sql_str_sub(subset_f = "SUBSTR", length_f = "LENGTH", optional_length = TRUE)

sql_paste(default_sep, f = "CONCAT_WS")

sql_paste_infix(default_sep, op, cast = sql_cast("text"))
}
\arguments{
\item{f}{The name of the SQL function as a string.}

\item{subset_f}{The name of the SQL substring function.}

\item{length_f}{The name of the SQL string length function.}

\item{optional_length}{Whether the length argument is optional in the SQL
substring function.}

\item{default_sep}{The default separator for paste operations.}

\item{op}{The SQL operator to use for infix paste operations.}

\item{cast}{A function to cast values to strings.}
}
\description{
These functions help you create custom string SQL translations when
implementing a new backend. They are typically used within \code{\link[=sql_translator]{sql_translator()}}
to define how R string functions should be translated to SQL.
\itemize{
\item \code{sql_substr()} creates a SQL substring function translator that converts
R's \code{substr(x, start, stop)} to SQL's \code{SUBSTR(x, start, length)}.
\item \code{sql_str_sub()} creates a SQL substring function translator that handles
stringr's \code{str_sub()} with support for negative indices.
\item \code{sql_paste()} creates a SQL paste function using \code{CONCAT_WS()} or similar.
\item \code{sql_paste_infix()} creates a SQL paste function using an infix operator
like \code{||}.
}
}
\seealso{
Other SQL translation helpers:
\code{\link{sql_translation_agg}},
\code{\link{sql_translation_scalar}},
\code{\link{sql_translation_window}},
\code{\link[=sql_variant]{sql_variant()}}
}
\concept{SQL translation helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db-sql.R
\name{sql_query_insert}
\alias{sql_query_insert}
\alias{sql_query_append}
\alias{sql_query_update_from}
\alias{sql_query_upsert}
\alias{sql_query_delete}
\title{Generate SQL for Insert, Update, Upsert, and Delete}
\usage{
sql_query_insert(
  con,
  x_name,
  y,
  by,
  ...,
  conflict = c("error", "ignore"),
  returning_cols = NULL,
  method = NULL
)

sql_query_append(con, x_name, y, ..., returning_cols = NULL)

sql_query_update_from(
  con,
  x_name,
  y,
  by,
  update_values,
  ...,
  returning_cols = NULL
)

sql_query_upsert(
  con,
  x_name,
  y,
  by,
  update_cols,
  ...,
  returning_cols = NULL,
  method = NULL
)

sql_query_delete(con, x_name, y, by, ..., returning_cols = NULL)
}
\arguments{
\item{con}{Database connection.}

\item{x_name}{Name of the table to update.}

\item{y}{A lazy tbl.}

\item{by}{An unnamed character vector giving the key columns. The key columns
must exist in both \code{x} and \code{y}. Keys typically uniquely identify each row,
but this is only enforced for the key values of \code{y} when \code{rows_update()},
\code{rows_patch()}, or \code{rows_upsert()} are used.

By default, we use the first column in \code{y}, since the first column is
a reasonable place to put an identifier variable.}

\item{...}{Other parameters passed onto methods.}

\item{conflict}{For \code{rows_insert()}, how should keys in \code{y} that conflict
with keys in \code{x} be handled? A conflict arises if there is a key in \code{y}
that already exists in \code{x}.

One of:
\itemize{
\item \code{"error"}, the default, will error if there are any keys in \code{y} that
conflict with keys in \code{x}.
\item \code{"ignore"} will ignore rows in \code{y} with keys that conflict with keys in
\code{x}.
}}

\item{returning_cols}{Optional. Names of columns to return.}

\item{method}{Optional. The method to use.}

\item{update_values}{A named SQL vector that specify how to update the columns.}

\item{update_cols}{Names of columns to update.}
}
\value{
A SQL query.
}
\description{
These functions generate the SQL used in \code{rows_*(in_place = TRUE)}.
}
\details{
Insert Methods
\subsection{\code{"where_not_exists"}}{

The default for most databases.

\if{html}{\out{<div class="sourceCode">}}\preformatted{INSERT INTO x_name
SELECT *
FROM y
WHERE NOT EXISTS <match on by columns>
}\if{html}{\out{</div>}}
}

\subsection{\code{"on_conflict"}}{

Supported by:
\itemize{
\item Postgres
\item SQLite
}

This method uses the \verb{ON CONFLICT} clause and therefore requires a unique
index on the columns specified in \code{by}.
}

Upsert Methods
\subsection{\code{"merge"}}{

The upsert method according to the SQL standard. It uses the \code{MERGE} statement

\if{html}{\out{<div class="sourceCode">}}\preformatted{MERGE INTO x_name
USING y
  ON <match on by columns>
WHEN MATCHED THEN
  UPDATE SET ...
WHEN NOT MATCHED THEN
  INSERT ...
}\if{html}{\out{</div>}}
}

\subsection{\code{"on_conflict"}}{

Supported by:
\itemize{
\item Postgres
\item SQLite
}

This method uses the \verb{ON CONFLICT} clause and therefore requires a unique
index on the columns specified in \code{by}.
}

\subsection{\code{"cte_update"}}{

Supported by:
\itemize{
\item Postgres
\item SQLite
\item Oracle
}

The classical way to upsert in Postgres and SQLite before support for
\verb{ON CONFLICT} was added. The update is done in a CTE clause and the unmatched
values are then inserted outside of the CTE.
}
}
\examples{
lf <- lazy_frame(
  carrier = c("9E", "AA"),
  name = c("Endeavor Air Inc.", "American Airlines Inc."),
  con = simulate_postgres()
)

sql_query_upsert(
  simulate_postgres(),
  ident("airlines"),
  lf,
  by = "carrier",
  update_cols = "name"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db.R
\name{db-misc}
\alias{db_connection_describe}
\alias{sql_join_suffix}
\alias{db_sql_render}
\alias{db_col_types}
\alias{dbplyr_edition}
\title{Miscellaneous database generics}
\usage{
db_connection_describe(con, ...)

sql_join_suffix(con, suffix, ...)

db_sql_render(con, sql, ..., cte = FALSE, sql_options = NULL)

db_col_types(con, table, call)

dbplyr_edition(con)
}
\arguments{
\item{con}{A database connection.}

\item{table}{A table identifier, or \code{NULL}. Use a string to refer to tables
in the current schema/catalog or \code{I()} to refer to tables in other
schemas/catalogs.}

\item{call}{The execution environment of a currently running function, used
to report errors.}
}
\description{
These are used when creating a new dbplyr backend and should generally not
be called directly.
\itemize{
\item \code{db_connection_describe()} provides a short string describing the
database connection, helping users tell which database a table comes
from. It should be a single line, and ideally less than 60 characters wide.
\item \code{dbplyr_edition()} declares which version of the dbplyr API you want.
\item \code{db_col_types()} introspects an existing table and returns a named
character vector mapping column names to their database-native SQL types.
The result is suitable for use as the \code{field.types} argument of
\code{\link[DBI:dbWriteTable]{DBI::dbWriteTable()}}, allowing dbplyr to preserve column types when
copying data with \verb{rows_*()}.

Returns \code{NULL} if the backend does not implement introspection.
}
}
\seealso{
Other generic:
\code{\link{db-sql}},
\code{\link[=db_copy_to]{db_copy_to()}},
\code{\link[=escape]{escape()}}
}
\concept{generic}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-select.R
\name{select.tbl_lazy}
\alias{select.tbl_lazy}
\alias{rename.tbl_lazy}
\alias{rename_with.tbl_lazy}
\alias{relocate.tbl_lazy}
\title{Subset, rename, and reorder columns using their names}
\usage{
\method{select}{tbl_lazy}(.data, ...)

\method{rename}{tbl_lazy}(.data, ...)

\method{rename_with}{tbl_lazy}(.data, .fn, .cols = everything(), ...)

\method{relocate}{tbl_lazy}(.data, ..., .before = NULL, .after = NULL)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{...}{<\code{\link[dplyr:dplyr_data_masking]{data-masking}}> Variables, or functions of
variables. Use \code{\link[dplyr:desc]{desc()}} to sort a variable in descending order.}

\item{.fn}{A function used to transform the selected \code{.cols}. Should
return a character vector the same length as the input.}

\item{.cols}{<\code{\link[dplyr:dplyr_tidy_select]{tidy-select}}> Columns to rename;
defaults to all columns.}

\item{.before, .after}{<\code{\link[dplyr:dplyr_tidy_select]{tidy-select}}> Destination of
columns selected by \code{...}. Supplying neither will move columns to the
left-hand side; specifying both is an error.}
}
\description{
These are methods for the dplyr \code{\link[=select]{select()}}, \code{\link[=rename]{rename()}}, and \code{\link[=relocate]{relocate()}}
generics. They generate the \code{SELECT} clause of the SQL query.

These functions do not support predicate functions, i.e. you can
not use \code{where(is.numeric)} to select all numeric variables.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(x = 1, y = 2, z = 3)
db \%>\% select(-y) \%>\% show_query()
db \%>\% relocate(z) \%>\% show_query()
db \%>\% rename(first = x, last = z) \%>\% show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lazy-ops.R
\name{lazy_ops}
\alias{lazy_ops}
\alias{lazy_base_query}
\alias{op_grps}
\alias{op_vars}
\alias{op_sort}
\alias{op_frame}
\title{Lazy operations}
\usage{
lazy_base_query(x, vars, class = character(), ...)

op_grps(op)

op_vars(op)

op_sort(op)

op_frame(op)
}
\description{
This set of S3 classes describe the action of dplyr verbs. These are
currently used for SQL sources to separate the description of operations
in R from their computation in SQL. This API is very new so is likely
to evolve in the future.
}
\details{
\code{op_vars()} and \code{op_grps()} compute the variables and groups from
a sequence of lazy operations. \code{op_sort()} and \code{op_frame()} tracks the
order and frame for use in window functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-redshift.R
\name{backend-redshift}
\alias{dialect_redshift}
\alias{simulate_redshift}
\title{Redshift backend}
\usage{
dialect_redshift()

simulate_redshift()
}
\description{
This backend supports Amazon Redshift databases, typically accessed via
a \code{RedshiftConnection} created by \code{\link[DBI:dbConnect]{DBI::dbConnect()}}. Use \code{dialect_redshift()}
with \code{lazy_frame()} to see simulated SQL without connecting to a live
database.

Base translations come from \link[=dialect_postgres]{PostgreSQL backend}. There
are generally few differences, apart from string manipulation.

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_redshift())
lf |> transmute(x = paste(c, " times"))
lf |> transmute(x = substr(c, 2, 3))
lf |> transmute(x = str_replace_all(c, "a", "z"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-copy-to.R
\name{copy_inline}
\alias{copy_inline}
\title{Use a local data frame in a dbplyr query}
\usage{
copy_inline(con, df, types = NULL)
}
\arguments{
\item{con}{A database connection.}

\item{df}{A local data frame. The data is written directly in the SQL query
so it should be small.}

\item{types}{A named character vector of SQL data types to use for the columns.
The data types are backend specific. For example for Postgres this could
be \code{c(id = "bigint", created_at = "timestamp", values = "integer[]")}.
If \code{NULL}, the default, the types are determined from \code{df}.}
}
\value{
A \code{tbl_lazy}.
}
\description{
This is an alternative to \code{\link[=copy_to]{copy_to()}} that does not need write access and
is faster for small data.
}
\details{
It writes the data directly in the SQL query via the \code{VALUES} clause.
}
\examples{
df <- data.frame(x = 1:3, y = c("a", "b", "c"))
con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")

copy_inline(con, df)

copy_inline(con, df) \%>\% dplyr::show_query()
}
\seealso{
\code{\link[=copy_to]{copy_to()}} to copy the data into a new database table.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-filter.R
\name{filter.tbl_lazy}
\alias{filter.tbl_lazy}
\alias{filter_out.tbl_lazy}
\title{Keep or drop rows that match a condition}
\usage{
\method{filter}{tbl_lazy}(.data, ..., .by = NULL, .preserve = FALSE)

\method{filter_out}{tbl_lazy}(.data, ..., .by = NULL, .preserve = FALSE)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{...}{<\code{\link[rlang:args_data_masking]{data-masking}}> Variables, or
functions of variables. Use \code{\link[dplyr:desc]{desc()}} to sort a variable in descending
order.}

\item{.by}{<\code{\link[dplyr:dplyr_tidy_select]{tidy-select}}> Optionally, a selection of columns to
group by for just this operation, functioning as an alternative to \code{\link[dplyr:group_by]{group_by()}}. For
details and examples, see \link[dplyr:dplyr_by]{?dplyr_by}.}

\item{.preserve}{Not supported by this method.}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
These are methods for the dplyr \code{\link[dplyr:filter]{dplyr::filter()}} and \code{\link[dplyr:filter_out]{dplyr::filter_out()}}
generics. They generate the \code{WHERE} clause of the SQL query.

\code{filter()} is translated directly to \code{WHERE}, which already matches dplyr's
behaviour of treating \code{NA} like \code{FALSE} (SQL's three-valued logic drops
\code{NULL} rows from \code{WHERE}).

\code{filter_out()} requires an additional step, where the combined condition
is wrapped in \code{is_distinct_from(., TRUE)}, which is then translated using
the backend (e.g. to \verb{IS DISTINCT FROM} on PostgreSQL, \verb{IS NOT} on SQLite).
This ensures that the SQL translation matches dplyr's semantics.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(x = c(2, NA, 5, NA, 10), y = 1:5)
db |> filter(x < 5) |> show_query()
db |> filter_out(x < 5) |> show_query()
db |> filter(is.na(x)) |> show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/with-dialect.R
\name{with_dialect}
\alias{with_dialect}
\title{Override the SQL dialect for a connection}
\usage{
with_dialect(con, dialect)
}
\arguments{
\item{con}{A database connection (class \code{DBIConnection}).}

\item{dialect}{A dialect object created by a \verb{dialect_*()} function
(e.g., \code{\link[=dialect_postgres]{dialect_postgres()}}, \code{\link[=dialect_sqlite]{dialect_sqlite()}}).}
}
\value{
A connection object that uses the specified dialect for SQL
generation.
}
\description{
\code{with_dialect()} overrides the default dialect assigned to a connection.
This is useful when dbplyr guesses the dialect incorrectly, which is most
likely to occur with ODBC/JDBC/ADBC backends.
}
\examples{
# Wrap an in-memory SQLite connection to use Postgres dialect
con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
pg_con <- with_dialect(con, dialect_postgres())

# SQL generation uses Postgres syntax
lf <- lazy_frame(x = 1, con = pg_con)
lf |> dplyr::mutate(y = sd(x))

DBI::dbDisconnect(con)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl-lazy.R
\name{tbl_lazy}
\alias{tbl_lazy}
\alias{lazy_frame}
\title{Create a local lazy tibble}
\usage{
tbl_lazy(df, con = NULL, src = NULL, name = "df")

lazy_frame(..., con = NULL, src = NULL, .name = "df")
}
\description{
These functions are useful for testing SQL generation without having to
have an active database connection. See \code{\link[=simulate_dbi]{simulate_dbi()}} for a list
available database simulations.
}
\examples{
library(dplyr)
df <- data.frame(x = 1, y = 2)

df_sqlite <- tbl_lazy(df, con = simulate_sqlite())
df_sqlite \%>\% summarise(x = sd(x, na.rm = TRUE)) \%>\% show_query()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-fill.R
\name{fill.tbl_lazy}
\alias{fill.tbl_lazy}
\title{Fill in missing values with previous or next value}
\usage{
\method{fill}{tbl_lazy}(.data, ..., .direction = c("down", "up", "updown", "downup"))
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{...}{Columns to fill.}

\item{.direction}{Direction in which to fill missing values. Currently
either "down" (the default) or "up". Note that "up" does not work when
\code{.data} is sorted by non-numeric columns. As a workaround revert the order
yourself beforehand; for example replace \code{arrange(x, desc(y))} by
\code{arrange(desc(x), y)}.}
}
\description{
Fill in missing values with previous or next value
}
\examples{
\dontshow{if (rlang::is_installed("tidyr", version = "1.0.0")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
squirrels <- tibble::tribble(
  ~group,    ~name,     ~role,     ~n_squirrels, ~ n_squirrels2,
  1,      "Sam",    "Observer",   NA,                 1,
  1,     "Mara", "Scorekeeper",    8,                NA,
  1,    "Jesse",    "Observer",   NA,                NA,
  1,      "Tom",    "Observer",   NA,                 4,
  2,     "Mike",    "Observer",   NA,                NA,
  2,  "Rachael",    "Observer",   NA,                 6,
  2,  "Sydekea", "Scorekeeper",   14,                NA,
  2, "Gabriela",    "Observer",   NA,                NA,
  3,  "Derrick",    "Observer",   NA,                NA,
  3,     "Kara", "Scorekeeper",    9,                 10,
  3,    "Emily",    "Observer",   NA,                NA,
  3, "Danielle",    "Observer",   NA,                NA
)
squirrels$id <- 1:12

tbl_memdb(squirrels) \%>\%
  window_order(id) \%>\%
  tidyr::fill(
    n_squirrels,
    n_squirrels2,
  )
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backward-compatibility.R
\name{escape_ansi}
\alias{escape_ansi}
\title{Escape ANSI characters}
\usage{
escape_ansi(x, parens = NA, collapse = "")
}
\arguments{
\item{x}{Object to escape.}

\item{parens, collapse}{Controls parens and collapsing. Passed on to
\code{\link[=escape]{escape()}}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

Use \code{\link[=escape]{escape()}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-mysql.R
\name{backend-mysql}
\alias{simulate_mysql}
\title{Backend: MySQL/MariaDB}
\usage{
simulate_mysql()
}
\description{
See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology. Key differences for this backend
are:
\itemize{
\item \code{paste()} uses \code{CONCAT_WS()}
\item String translations for \code{str_detect()}, \code{str_locate()}, and
\code{str_replace_all()}
\item Clear error message for unsupported full joins
}

Use \code{simulate_mysql()} with \code{lazy_frame()} to see simulated SQL without
converting to live access database.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = simulate_mysql())
lf \%>\% transmute(x = paste0(d, " times"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ident.R
\name{ident}
\alias{ident}
\alias{is.ident}
\title{Flag a character vector as SQL identifiers}
\usage{
ident(...)

is.ident(x)
}
\arguments{
\item{...}{A character vector, or name-value pairs.}

\item{x}{An object.}
}
\description{
\code{ident()} marks strings as database identifiers (e.g. table or column names)
quoting them using the identifier rules for your database. It is used
primarily in \code{\link[=translate_sql]{translate_sql()}} to label variables as identifiers; use
elsewhere should be regarded with suspicion.

\code{ident()} is for internal use only; if you need to supply an table name that
is qualified with schema or catalog use \code{I()}.
}
\examples{
con <- dialect_ansi()

# SQL92 quotes strings with '
escape("x", con = con)

# And identifiers with "
escape(ident("x"), con = con)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table-name.R
\name{is_table_path}
\alias{is_table_path}
\alias{table_path_name}
\alias{table_path_components}
\alias{check_table_path}
\alias{as_table_path}
\title{Table paths}
\usage{
is_table_path(x)

table_path_name(x, con)

table_path_components(x, con)

check_table_path(x, error_arg = caller_arg(x), error_call = caller_env())

as_table_path(x, con, error_arg = caller_arg(x), error_call = caller_env())
}
\description{
dbplyr standardises all the ways of referring to a table (i.e. a single
string, a string wrapped in \code{I()}, a \code{\link[DBI:Id]{DBI::Id()}} and the results of
\code{\link[=in_schema]{in_schema()}} and \code{\link[=in_catalog]{in_catalog()}}) into a table "path" of the form
\code{table}, \code{schema.table}, or \code{catalog.schema.path}. A table path is
always suitable for inlining into a query, so user input is quoted unless
it is wrapped in \code{I()}.

This is primarily for internal usage, but you may need to work with it if
you're implementing a backend, and you need to compute with the table path,
not just pass it on unchanged to some other dbplyr function.
\itemize{
\item \code{is_table_path()} returns \code{TRUE} if the object is a \code{table_path}.
\item \code{as_table_path()} coerces known table identifiers to a \code{table_path}.
\item \code{check_table_path()} throws an error if the object is not a \code{table_path}.
\item \code{table_path_name()} returns the last component of the table path (i.e.
the name of the table).
\item \code{table_path_components()} returns a list containing the components of each
table path.
}

A \code{table_path} object can technically be a vector of table paths, but
you will never see this in table paths constructed from user inputs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/translate-sql-window.R
\name{sql_translation_window}
\alias{sql_translation_window}
\alias{win_over}
\alias{win_rank}
\alias{win_aggregate}
\alias{win_aggregate_2}
\alias{win_recycled}
\alias{win_cumulative}
\alias{win_absent}
\alias{win_current_group}
\alias{win_current_order}
\alias{win_current_frame}
\title{SQL helpers for window functions}
\usage{
win_over(
  expr,
  partition = NULL,
  order = NULL,
  frame = NULL,
  con = sql_current_con()
)

win_rank(f, empty_order = FALSE)

win_aggregate(f)

win_aggregate_2(f)

win_cumulative(f)

win_absent(f)

win_current_group()

win_current_order()

win_current_frame()
}
\arguments{
\item{expr}{The window expression.}

\item{partition}{Variables to partition over.}

\item{order}{Variables to order by.}

\item{frame}{A numeric vector of length two defining the frame.}

\item{con}{A \link{sql_dialect} object or database connection. Connections are
supported for backward compatibility.}

\item{f}{The name of an SQL function as a string.}

\item{empty_order}{A logical value indicating whether to order by NULL if
\code{order} is not specified.}
}
\description{
These functions help you create custom window SQL translations when
implementing a new backend. They are typically used within \code{\link[=sql_translator]{sql_translator()}}
to define how R window functions should be translated to SQL.
\itemize{
\item \code{win_over()} makes it easy to generate the window function specification.
\item \code{win_absent()}, \code{win_rank()}, \code{win_aggregate()}, and \code{win_cumulative()}
provide helpers for constructing common types of window functions.
\item \code{win_current_group()} and \code{win_current_order()} allow you to access
the grouping and order context set up by \code{\link[dplyr:group_by]{group_by()}} and \code{\link[dplyr:arrange]{arrange()}}.
}
}
\examples{
con <- dialect_ansi()

win_over(sql("avg(x)"), con = con)
win_over(sql("avg(x)"), "y", con = con)
win_over(sql("avg(x)"), order = "y", con = con)
win_over(sql("avg(x)"), order = c("x", "y"), con = con)
win_over(sql("avg(x)"), frame = c(-Inf, 0), order = "y", con = con)
}
\seealso{
Other SQL translation helpers:
\code{\link{sql_translation_agg}},
\code{\link{sql_translation_scalar}},
\code{\link{sql_translation_string}},
\code{\link[=sql_variant]{sql_variant()}}
}
\concept{SQL translation helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sql-superseded.R
\name{as.sql}
\alias{as.sql}
\title{Convert to sql (deprecated)}
\usage{
as.sql(x, con)
}
\arguments{
\item{x}{Object to coerce}

\item{con}{Needed when \code{x} is directly supplied from the user so that
schema specifications can be quoted using the correct identifiers.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-pivot-longer.R
\name{pivot_longer.tbl_lazy}
\alias{pivot_longer.tbl_lazy}
\title{Pivot data from wide to long}
\usage{
\method{pivot_longer}{tbl_lazy}(
  data,
  cols,
  names_to = "name",
  names_prefix = NULL,
  names_sep = NULL,
  names_pattern = NULL,
  names_ptypes = NULL,
  names_transform = NULL,
  names_repair = "check_unique",
  values_to = "value",
  values_drop_na = FALSE,
  values_ptypes,
  values_transform = NULL,
  ...
)
}
\arguments{
\item{data}{A data frame to pivot.}

\item{cols}{Columns to pivot into longer format.}

\item{names_to}{A string specifying the name of the column to create
from the data stored in the column names of \code{data}.}

\item{names_prefix}{A regular expression used to remove matching text
from the start of each variable name.}

\item{names_sep, names_pattern}{If \code{names_to} contains multiple values,
these arguments control how the column name is broken up.}

\item{names_ptypes}{A list of column name-prototype pairs.}

\item{names_transform, values_transform}{A list of column name-function pairs.}

\item{names_repair}{What happens if the output has invalid column names?}

\item{values_to}{A string specifying the name of the column to create
from the data stored in cell values. If \code{names_to} is a character
containing the special \code{.value} sentinel, this value will be ignored,
and the name of the value column will be derived from part of the
existing column names.}

\item{values_drop_na}{If \code{TRUE}, will drop rows that contain only \code{NA}s
in the \code{value_to} column.}

\item{values_ptypes}{Not supported.}

\item{...}{Additional arguments passed on to methods.}
}
\description{
\code{pivot_longer()} "lengthens" data, increasing the number of rows and
decreasing the number of columns. The inverse transformation is
`tidyr::pivot_wider()]

Learn more in \code{vignette("pivot", "tidyr")}.

While most functionality is identical there are some differences to
\code{pivot_longer()} on local data frames:
\itemize{
\item the output is sorted differently/not explicitly,
\item the coercion of mixed column types is left to the database,
\item \code{values_ptypes} NOT supported.
}

Note that \code{build_longer_spec()} and \code{pivot_longer_spec()} do not work with
remote tables.
}
\details{
The SQL translation basically works as follows:
\enumerate{
\item split the specification by its key columns i.e. by variables crammed
into the column names.
\item for each part in the splitted specification \code{transmute()} \code{data} into the
following columns
}
\itemize{
\item id columns i.e. columns that are not pivotted
\item key columns
\item value columns i.e. columns that are pivotted
}
\enumerate{
\item combine all the parts with \code{union_all()}
}
}
\examples{
\dontshow{if (rlang::is_installed("tidyr", version = "1.0.0")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# See vignette("pivot") for examples and explanation

# Simplest case where column names are character data
memdb_frame(
  id = c("a", "b"),
  x = 1:2,
  y = 3:4
) \%>\%
  tidyr::pivot_longer(-id)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-adbc.R
\name{backend-adbc}
\alias{backend-adbc}
\title{ADBC backend}
\description{
This backend supports databases accessed via \code{AdbiConnection} created by
\code{adbi::adbi()} and \code{\link[DBI:dbConnect]{DBI::dbConnect()}}. dbplyr automatically detects the
underlying database type by querying the ADBC driver's vendor name and
uses the appropriate SQL dialect.

The following vendors are recognized:
\itemize{
\item PostgreSQL and PostgreSQL-compatible systems (CrateDB, CockroachDB,
Citus, Neon, ParadeDB, TimescaleDB, Yellowbrick, YugabyteDB, CedarDB):
\code{\link[=dialect_postgres]{dialect_postgres()}}
\item MySQL and MySQL-compatible systems (MariaDB, TiDB, Vitess): the
appropriate dialect (\code{\link[=dialect_mariadb]{dialect_mariadb()}} for MariaDB, \code{\link[=dialect_mysql]{dialect_mysql()}}
otherwise)
\item SQLite: \code{\link[=dialect_sqlite]{dialect_sqlite()}}
\item Microsoft SQL Server: \code{\link[=dialect_mssql]{dialect_mssql()}}
\item Snowflake: \code{\link[=dialect_snowflake]{dialect_snowflake()}}
\item Amazon Redshift: \code{\link[=dialect_redshift]{dialect_redshift()}}
}

If your database is not recognized, dbplyr will fall back to a generic ODBC
dialect. In this case, or if dbplyr guesses wrong, you can use
\code{\link[=with_dialect]{with_dialect()}} to choose a specific dialect.

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
# ADBC connections require the adbi package and an ADBC driver.
# Once connected, dbplyr automatically detects the database type:
#
# library(adbi)
# library(dplyr, warn.conflicts = FALSE)
#
# con <- DBI::dbConnect(adbi::adbi("adbcsqlite"), uri = ":memory:")
# tbl(con, "my_table") |> filter(x > 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-lahman.R
\name{lahman}
\alias{lahman}
\alias{lahman_sqlite}
\alias{lahman_postgres}
\alias{lahman_mysql}
\alias{copy_lahman}
\alias{has_lahman}
\alias{lahman_srcs}
\title{Cache and retrieve an \code{src_sqlite} of the Lahman baseball database.}
\usage{
lahman_sqlite(path = NULL)

lahman_postgres(dbname = "lahman", host = "localhost", ...)

lahman_mysql(dbname = "lahman", ...)

copy_lahman(con, ...)

has_lahman(type, ...)

lahman_srcs(..., quiet = NULL)
}
\arguments{
\item{...}{Other arguments passed to \code{src} on first
load. For MySQL and PostgreSQL, the defaults assume you have a local
server with \code{lahman} database already created.
For \code{lahman_srcs()}, character vector of names giving srcs to generate.}

\item{type}{src type.}

\item{quiet}{if \code{TRUE}, suppress messages about databases failing to
connect.}
}
\description{
This creates an interesting database using data from the Lahman baseball
data source, provided by Sean Lahman at
\url{https://www.seanlahman.com/baseball-archive/statistics/}, and
made easily available in R through the \pkg{Lahman} package by
Michael Friendly, Dennis Murphy and Martin Monkman. See the documentation
for that package for documentation of the individual tables.
}
\examples{
# Connect to a local sqlite database, if already created
\donttest{
library(dplyr)

if (has_lahman("sqlite")) {
  lahman_sqlite()
  batting <- tbl(lahman_sqlite(), "Batting")
  batting
}

# Connect to a local postgres database with lahman database, if available
if (has_lahman("postgres")) {
  lahman_postgres()
  batting <- tbl(lahman_postgres(), "Batting")
}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-postgres.R
\name{backend-postgres}
\alias{dialect_postgres}
\alias{simulate_postgres}
\title{PostgreSQL backend}
\usage{
dialect_postgres()

simulate_postgres()
}
\description{
This backend supports PostgreSQL databases, typically accessed via a
\code{PqConnection} created by \code{\link[DBI:dbConnect]{DBI::dbConnect()}}. Use \code{dialect_postgres()} with
\code{lazy_frame()} to see simulated SQL without connecting to a live database.

Key differences for this backend are:
\itemize{
\item Many stringr functions
\item lubridate date-time extraction functions
\item More standard statistical summaries
}

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = dialect_postgres())
lf |> summarise(x = sd(b, na.rm = TRUE))
lf |> summarise(y = cor(b, c), z = cov(b, c))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memdb.R
\name{memdb}
\alias{memdb}
\alias{memdb_frame}
\alias{local_memdb_frame}
\title{A temporary in-memory database}
\usage{
memdb()

memdb_frame(.name = unique_table_name(), ...)

local_memdb_frame(.name = unique_table_name(), ..., frame = caller_env())
}
\arguments{
\item{.name}{Name of table in database: defaults to a random name that's
unlikely to conflict with an existing table.}

\item{...}{<\code{\link[rlang:dyn-dots]{dynamic-dots}}>
A set of name-value pairs. These arguments are
processed with \code{\link[rlang:defusing-advanced]{rlang::quos()}} and support unquote via \code{\link[rlang:injection-operator]{rlang::!!}} and
unquote-splice via \code{\link[rlang:splice-operator]{rlang::!!!}}. Use \verb{:=} to create columns that start with a dot.

Arguments are evaluated sequentially.
You can refer to previously created elements directly or using the \link[rlang:dot-data]{rlang::.data}
pronoun.
To refer explicitly to objects in the calling environment, use \code{\link[rlang:injection-operator]{rlang::!!}} or
\link[rlang:dot-data]{rlang::.env}, e.g. \code{!!.data} or \code{.env$.data} for the special case of an object
named \code{.data}.}

\item{frame}{The created table is bound to this execution frame and will
be deleted when it ends. For expert use only.}
}
\description{
\code{memdb()} creates a temporary in-memory database that disappears when the
R session ends. It's a convenient way to learn about and experiment with
dbplyr without having to connect to a "real" database.

\code{memdb_frame()} works like \code{\link[tibble:tibble]{tibble::tibble()}}, but instead of creating a new
data frame in R, it creates a table in \code{memdb()}. \code{local_memdb_frame()}
is like \code{memdb_frame()} but the table will be automatically deleted when
the current scope ends. It's useful for tests. But beware: this function
will overwrite an existing table of the same name.
}
\examples{
library(dplyr)

# use memdb_frame() to create a new database table
df <- memdb_frame(x = runif(100), y = runif(100))
df |> arrange(x)
df |> arrange(x) |> show_query()

# Use memdb() + copy_to() to copy an existing data frame
iris_db <- copy_to(memdb(), iris)
iris_db
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbplyr.R
\docType{package}
\name{dbplyr-package}
\alias{dbplyr}
\alias{dbplyr-package}
\title{dbplyr: A 'dplyr' Back End for Databases}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A 'dplyr' back end for databases that allows you to work with remote database tables as if they are in-memory data frames. Basic features works with any database that has a 'DBI' back end; more advanced features require 'SQL' translation to be provided by the package author.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://dbplyr.tidyverse.org/}
  \item \url{https://github.com/tidyverse/dbplyr}
  \item Report bugs at \url{https://github.com/tidyverse/dbplyr/issues}
}

}
\author{
\strong{Maintainer}: Hadley Wickham \email{hadley@rstudio.com}

Authors:
\itemize{
  \item Maximilian Girlich
  \item Edgar Ruiz
}

Other contributors:
\itemize{
  \item RStudio [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sql-dialect.R
\name{sql_dialect}
\alias{sql_dialect}
\alias{new_sql_dialect}
\title{SQL dialects}
\usage{
sql_dialect(con)

new_sql_dialect(
  dialect,
  quote_identifier,
  has_window_clause = FALSE,
  has_table_alias_with_as = TRUE,
  has_star_table_prefix = FALSE
)
}
\arguments{
\item{con}{A database connection.}

\item{dialect}{A string giving the dialect name (e.g., "postgres", "mysql").}

\item{quote_identifier}{A function that quotes identifiers. Should accept
a character vector and return a \link{sql} vector.}

\item{has_window_clause}{Does the backend support named window
definitions (the \code{WINDOW} clause)?}

\item{has_table_alias_with_as}{Does the backend support using \code{AS}
when aliasing a table in a subquery?}

\item{has_star_table_prefix}{Does the backend require table prefixes when
selecting all columns in single-table queries (e.g., \verb{"table".*} vs \code{*})?}
}
\value{
\itemize{
\item \code{sql_dialect()} returns a dialect object (class \code{sql_dialect}) or
the connection itself for backward compatibility.
\item \code{new_sql_dialect()} returns a dialect object with class
\code{c("sql_dialect_{name}", "sql_dialect")}.
}
}
\description{
The dialect system allows multiple database connection classes to share
SQL generation code. A dialect object encapsulates the SQL syntax rules
for a particular database, independent of the connection mechanism.
\itemize{
\item \code{sql_dialect()} returns the dialect for a connection. For connections
that haven't implemented a dialect method, returns the connection itself
for backward compatibility.
\item \code{new_sql_dialect()} creates a new dialect object. This is primarily
intended for dbplyr backend authors.
}
}
\section{Dispatching on dialect}{
For backward compatibility, all \code{sql_} generics (and a handful of others)
call \code{sql_dialect()} on the \code{con} argument in order to dispatch further on
the dialect object, if possible:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{sql_generic <- function(con, arg1, arg2, ...) \{
  UseMethod("sql_generic", sql_dialect(con))
\}
}\if{html}{\out{</div>}}

Unfortunately, due to the way that \code{UseMethod()} works, this uses
\code{sql_dialect(con)} to control which method is selected, but still passes
the original \code{con} to the method. This means that if you are implementing
a method for a dialect and need to access dialect properties, you must call
\code{sql_dialect(con)} again inside the method.
}

\examples{
# Create a custom dialect
my_dialect <- new_sql_dialect(
  "custom",
  quote_identifier = function(x) sql_quote(x, "`"),
  has_window_clause = TRUE
)
class(my_dialect)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind-queries.R
\name{bind_queries}
\alias{bind_queries}
\title{Combine multiple lazy queries}
\usage{
bind_queries(...)
}
\arguments{
\item{...}{<\link[rlang:dyn-dots]{dynamic-dots}> Lazy tables to combine.}
}
\value{
A lazy query.
}
\description{
Combine multiple lazy queries into a single query using \verb{UNION ALL}.
This is a convenient wrapper around \code{purrr::reduce(tables, union_all)}.
Like \code{\link[dplyr:bind_rows]{dplyr::bind_rows()}} (and unlike \verb{UNION ALL}), \code{bind_queries()} will
automatically align columns based on their name, and fill in any missing
columns with missing values.
}
\examples{
lf1 <- lazy_frame(x = 1, y = "a")
lf2 <- lazy_frame(x = 2, y = "b")
bind_queries(lf1, lf2)

lf3 <- lazy_frame(y = "c", x = 3, z = 10)
bind_queries(lf2, lf3)

# If you already have a list, you can use splice operator
queries <- list(lf1, lf2)
bind_queries(!!!queries)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-pivot-wider.R
\name{pivot_wider.tbl_lazy}
\alias{pivot_wider.tbl_lazy}
\alias{dbplyr_pivot_wider_spec}
\title{Pivot data from long to wide}
\usage{
\method{pivot_wider}{tbl_lazy}(
  data,
  ...,
  id_cols = NULL,
  id_expand = FALSE,
  names_from = name,
  names_prefix = "",
  names_sep = "_",
  names_glue = NULL,
  names_sort = FALSE,
  names_vary = "fastest",
  names_expand = FALSE,
  names_repair = "check_unique",
  values_from = value,
  values_fill = NULL,
  values_fn = ~max(.x, na.rm = TRUE),
  unused_fn = NULL
)

dbplyr_pivot_wider_spec(
  data,
  spec,
  ...,
  names_repair = "check_unique",
  id_cols = NULL,
  id_expand = FALSE,
  values_fill = NULL,
  values_fn = ~max(.x, na.rm = TRUE),
  unused_fn = NULL,
  error_call = current_env()
)
}
\arguments{
\item{data}{A lazy data frame backed by a database query.}

\item{...}{Unused; included for compatibility with generic.}

\item{id_cols}{A set of columns that uniquely identifies each observation.}

\item{id_expand}{Unused; included for compatibility with the generic.}

\item{names_from, values_from}{A pair of
arguments describing which column (or columns) to get the name of the
output column (\code{names_from}), and which column (or columns) to get the
cell values from (\code{values_from}).

If \code{values_from} contains multiple values, the value will be added to the
front of the output column.}

\item{names_prefix}{String added to the start of every variable name.}

\item{names_sep}{If \code{names_from} or \code{values_from} contains multiple
variables, this will be used to join their values together into a single
string to use as a column name.}

\item{names_glue}{Instead of \code{names_sep} and \code{names_prefix}, you can supply
a glue specification that uses the \code{names_from} columns (and special
\code{.value}) to create custom column names.}

\item{names_sort}{Should the column names be sorted? If \code{FALSE}, the default,
column names are ordered by first appearance.}

\item{names_vary}{When \code{names_from} identifies a column (or columns) with
multiple unique values, and multiple \code{values_from} columns are provided,
in what order should the resulting column names be combined?
\itemize{
\item \code{"fastest"} varies \code{names_from} values fastest, resulting in a column
naming scheme of the form: \verb{value1_name1, value1_name2, value2_name1, value2_name2}. This is the default.
\item \code{"slowest"} varies \code{names_from} values slowest, resulting in a column
naming scheme of the form: \verb{value1_name1, value2_name1, value1_name2, value2_name2}.
}}

\item{names_expand}{Should the values in the \code{names_from} columns be expanded
by \code{\link[=expand]{expand()}} before pivoting? This results in more columns, the output
will contain column names corresponding to a complete expansion of all
possible values in \code{names_from}. Additionally, the column names will be
sorted, identical to what \code{names_sort} would produce.}

\item{names_repair}{What happens if the output has invalid column names?}

\item{values_fill}{Optionally, a (scalar) value that specifies what each
\code{value} should be filled in with when missing.}

\item{values_fn}{A function, the default is \code{max()}, applied to the \code{value}
in each cell in the output. In contrast to local data frames it must not be
\code{NULL}.}

\item{unused_fn}{Optionally, a function applied to summarize the values from
the unused columns (i.e. columns not identified by \code{id_cols},
\code{names_from}, or \code{values_from}).

The default drops all unused columns from the result.

This can be a named list if you want to apply different aggregations
to different unused columns.

\code{id_cols} must be supplied for \code{unused_fn} to be useful, since otherwise
all unspecified columns will be considered \code{id_cols}.

This is similar to grouping by the \code{id_cols} then summarizing the
unused columns using \code{unused_fn}.}

\item{spec}{A specification data frame. This is useful for more complex
pivots because it gives you greater control on how metadata stored in the
columns become column names in the result.

Must be a data frame containing character \code{.name} and \code{.value} columns.
Additional columns in \code{spec} should be named to match columns in the
long format of the dataset and contain values corresponding to columns
pivoted from the wide format.
The special \code{.seq} variable is used to disambiguate rows internally;
it is automatically removed after pivoting.}

\item{error_call}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}
}
\description{
\code{pivot_wider()} "widens" data, increasing the number of columns and
decreasing the number of rows. The inverse transformation is
\code{pivot_longer()}.
Learn more in \code{vignette("pivot", "tidyr")}.

Note that \code{pivot_wider()} is not and cannot be lazy because we need to look
at the data to figure out what the new column names will be.
If you have a long running query you have two options:
\itemize{
\item (temporarily) store the result of the query via \code{compute()}.
\item Create a spec before and use \code{dbplyr_pivot_wider_spec()} - dbplyr's version
of \code{tidyr::pivot_wider_spec()}. Note that this function is only a temporary
solution until \code{pivot_wider_spec()} becomes a generic. It will then be
removed soon afterwards.
}
}
\details{
The big difference to \code{pivot_wider()} for local data frames is that
\code{values_fn} must not be \code{NULL}. By default it is \code{max()} which yields
the same results as for local data frames if the combination of \code{id_cols}
and \code{value} column uniquely identify an observation.
Mind that you also do not get a warning if an observation is not uniquely
identified.

The translation to SQL code basically works as follows:
\enumerate{
\item Get unique keys in \code{names_from} column.
\item For each key value generate an expression of the form:

\if{html}{\out{<div class="sourceCode sql">}}\preformatted{value_fn(
  CASE WHEN (`names from column` == `key value`)
  THEN (`value column`)
  END
) AS `output column`
}\if{html}{\out{</div>}}
\item Group data by id columns.
\item Summarise the grouped data with the expressions from step 2.
}
}
\examples{
\dontshow{if (rlang::is_installed("tidyr", version = "1.0.0")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
memdb_frame(
  id = 1,
  key = c("x", "y"),
  value = 1:2
) \%>\%
  tidyr::pivot_wider(
    id_cols = id,
    names_from = key,
    values_from = value
  )
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyeval.R
\name{.sql}
\alias{.sql}
\title{Flag SQL function usage}
\usage{
.sql
}
\description{
Use \code{.sql$foo(x, y)} to make it clear that you're calling the SQL
\code{foo()} function, not the R \code{foo()} function. This also makes it easier to
reduce \verb{R CMD check} notes in packages; just import \code{.sql} from dbplyr with
e.g. \verb{@importFrom dbplyr .sql}.

Note that \code{.sql} itself does nothing and is just \code{NULL}; it is automatically
removed when dbplyr translates your R code to SQL.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- lazy_frame(x = 1, y = 2)
db |> mutate(z = .sql$CUMULATIVE_SUM(x, 1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-expand.R
\name{complete.tbl_lazy}
\alias{complete.tbl_lazy}
\title{Complete a SQL table with missing combinations of data}
\usage{
\method{complete}{tbl_lazy}(data, ..., fill = list())
}
\arguments{
\item{data}{A lazy data frame backed by a database query.}

\item{...}{Specification of columns to expand. See \link[tidyr:expand]{tidyr::expand} for
more details.}

\item{fill}{A named list that for each variable supplies a single value to
use instead of NA for missing combinations.}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
Turns implicit missing values into explicit missing values. This is a method
for the \code{\link[tidyr:complete]{tidyr::complete()}} generic.
}
\examples{
\dontshow{if (rlang::is_installed("tidyr", version = "1.0.0")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
df <- memdb_frame(
  group = c(1:2, 1),
  item_id = c(1:2, 2),
  item_name = c("a", "b", "b"),
  value1 = 1:3,
  value2 = 4:6
)

df \%>\% tidyr::complete(group, nesting(item_id, item_name))

# You can also choose to fill in missing values
df \%>\% tidyr::complete(group, nesting(item_id, item_name), fill = list(value1 = 0))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-summarise.R
\name{summarise.tbl_lazy}
\alias{summarise.tbl_lazy}
\title{Summarise each group to one row}
\usage{
\method{summarise}{tbl_lazy}(.data, ..., .by = NULL, .groups = NULL)
}
\arguments{
\item{.data}{A lazy data frame backed by a database query.}

\item{...}{<\code{\link[dplyr:dplyr_data_masking]{data-masking}}> Variables, or functions of
variables. Use \code{\link[dplyr:desc]{desc()}} to sort a variable in descending order.}

\item{.by}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

<\code{\link[=dplyr_tidy_select]{tidy-select}}> Optionally, a selection of columns to
temporarily group by using an inline alternative to \code{\link[=group_by]{group_by()}}.}

\item{.groups}{\Sexpr[results=rd]{lifecycle::badge("experimental")} Grouping structure of the result.
\itemize{
\item "drop_last": dropping the last level of grouping. This was the
only supported option before version 1.0.0.
\item "drop": All levels of grouping are dropped.
\item "keep": Same grouping structure as \code{.data}.
}

When \code{.groups} is not specified, it defaults to "drop_last".

In addition, a message informs you of that choice, unless the result is ungrouped,
the option "dplyr.summarise.inform" is set to \code{FALSE},
or when \code{summarise()} is called from a function in a package.}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
This is a method for the dplyr \code{\link[=summarise]{summarise()}} generic. It generates the
\code{SELECT} clause of the SQL query, and generally needs to be combined with
\code{group_by()}.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(g = c(1, 1, 1, 2, 2), x = c(4, 3, 6, 9, 2))
db \%>\%
  summarise(n()) \%>\%
  show_query()

db \%>\%
  group_by(g) \%>\%
  summarise(n()) \%>\%
  show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-sqlite.R
\name{backend-sqlite}
\alias{simulate_sqlite}
\title{Backend: SQLite}
\usage{
simulate_sqlite()
}
\description{
See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology. Key differences for this backend
are:
\itemize{
\item Uses non-standard \code{LOG()} function
\item Date-time extraction functions from lubridate
\item Custom median translation
\item Right and full joins are simulated using left joins
}

Use \code{simulate_sqlite()} with \code{lazy_frame()} to see simulated SQL without
converting to live access database.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

lf <- lazy_frame(a = TRUE, b = 1, c = 2, d = "z", con = simulate_sqlite())
lf \%>\% transmute(x = paste(c, " times"))
lf \%>\% transmute(x = log(b), y = log(b, base = 2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/escape.R
\name{escape}
\alias{escape}
\alias{db-quote}
\alias{sql_escape_ident}
\alias{sql_escape_logical}
\alias{sql_escape_date}
\alias{sql_escape_datetime}
\alias{sql_escape_string}
\alias{sql_escape_raw}
\alias{sql_vector}
\title{Escape/quote a value}
\usage{
escape(x, parens = NA, collapse = " ", con = NULL)

sql_escape_ident(con, x)

sql_escape_logical(con, x)

sql_escape_date(con, x)

sql_escape_datetime(con, x)

sql_escape_string(con, x)

sql_escape_raw(con, x)

sql_vector(x, parens = NA, collapse = " ", con = NULL)
}
\arguments{
\item{x}{An object to escape. Existing sql vectors will be left as is,
character vectors are escaped with single quotes, numeric vectors have
trailing \code{.0} added if they're whole numbers, identifiers are
escaped with double quotes.}

\item{parens, collapse}{Controls behaviour when multiple values are supplied.
\code{parens} should be a logical flag, or if \code{NA}, will wrap in
parens if length > 1.

Default behaviour: lists are always wrapped in parens and separated by
commas, identifiers are separated by commas and never wrapped,
atomic vectors are separated by spaces and wrapped in parens if needed.}

\item{con}{A \link{sql_dialect} object or database connection. Connections are
supported for backward compatibility.}
}
\value{
A \link{sql} vector.
}
\description{
\code{escape()} turns R values into SQL literals. It implements double dispatch
via two sets of generics: first \code{escape()} dispatches on the class of \code{x},
then that method calls \code{sql_escape_ident()}, \code{sql_escape_logical()}, etc,
which dispatch on \code{con}.

These generics translate individual values into SQL. The core
generics are \code{\link[DBI:dbQuoteIdentifier]{DBI::dbQuoteIdentifier()}} and \code{\link[DBI:dbQuoteString]{DBI::dbQuoteString()}}
for quoting identifiers and strings, but dbplyr needs additional
tools for inserting logical, date, date-time, and raw values into
queries.
}
\examples{
con <- dialect_ansi()

# Doubles vs. integers
escape(1:5, con = con)
escape(c(1, 5.4), con = con)

# String vs known sql vs. sql identifier
escape("X", con = con)
escape(sql("X"), con = con)
escape(ident("X"), con = con)

# Escaping is idempotent
escape("X", con = con)
escape(escape("X", con = con), con = con)

# Database specific generics
sql_escape_logical(con, c(TRUE, FALSE, NA))
sql_escape_date(con, Sys.Date())
sql_escape_date(con, Sys.time())
sql_escape_raw(con, charToRaw("hi"))
}
\seealso{
Other generic:
\code{\link{db-sql}},
\code{\link[=db_connection_describe]{db_connection_describe()}},
\code{\link[=db_copy_to]{db_copy_to()}}
}
\concept{generic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-head.R
\name{head.tbl_lazy}
\alias{head.tbl_lazy}
\title{Subset the first rows}
\usage{
\method{head}{tbl_lazy}(x, n = 6L, ...)
}
\arguments{
\item{x}{A lazy data frame backed by a database query.}

\item{n}{Number of rows to return}

\item{...}{Not used.}
}
\value{
Another \code{tbl_lazy}. Use \code{\link[=show_query]{show_query()}} to see the generated
query, and use \code{\link[=collect.tbl_sql]{collect()}} to execute the query
and return data to R.
}
\description{
This is a method for the \code{\link[=head]{head()}} generic. It is usually translated to the
\code{LIMIT} clause of the SQL query. Because \code{LIMIT} is not an official part of
the SQL specification, some database use other clauses like \code{TOP} or
\verb{FETCH ROWS}.

Note that databases don't really have a sense of row order, so what "first"
means is subject to interpretation. Most databases will respect ordering
performed with \code{arrange()}, but it's not guaranteed. \code{tail()} is not
supported at all because the situation is even murkier for the "last" rows.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(x = 1:100)
db \%>\% head() \%>\% show_query()

# Pretend we have data in a SQL server database
db2 <- lazy_frame(x = 1:100, con = simulate_mssql())
db2 \%>\% head() \%>\% show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/translate-sql-window.R, R/backend-teradata.R
\name{win_over}
\alias{win_over}
\alias{win_rank}
\alias{win_aggregate}
\alias{win_aggregate_2}
\alias{win_recycled}
\alias{win_cumulative}
\alias{win_absent}
\alias{win_current_group}
\alias{win_current_order}
\alias{win_current_frame}
\alias{win_rank_tdata}
\title{Generate SQL expression for window functions}
\usage{
win_over(
  expr,
  partition = NULL,
  order = NULL,
  frame = NULL,
  con = sql_current_con()
)

win_rank(f)

win_aggregate(f)

win_aggregate_2(f)

win_cumulative(f)

win_absent(f)

win_current_group()

win_current_order()

win_current_frame()

win_rank_tdata(f)
}
\arguments{
\item{expr}{The window expression}

\item{order}{Variables to order by}

\item{frame}{A numeric vector of length two defining the frame.}

\item{f}{The name of an sql function as a string}

\item{parition}{Variables to partition over}
}
\description{
\code{win_over()} makes it easy to generate the window function specification.
\code{win_absent()}, \code{win_rank()}, \code{win_aggregate()}, and \code{win_cumulative()}
provide helpers for constructing common types of window functions.
\code{win_current_group()} and \code{win_current_order()} allow you to access
the grouping and order context set up by \code{\link[=group_by]{group_by()}} and \code{\link[=arrange]{arrange()}}.
}
\examples{
con <- simulate_dbi()

win_over(sql("avg(x)"), con = con)
win_over(sql("avg(x)"), "y", con = con)
win_over(sql("avg(x)"), order = "y", con = con)
win_over(sql("avg(x)"), order = c("x", "y"), con = con)
win_over(sql("avg(x)"), frame = c(-Inf, 0), order = "y", con = con)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-collapse.R
\name{collapse.tbl_sql}
\alias{collapse.tbl_sql}
\title{Collapse a query into a subquery}
\usage{
\method{collapse}{tbl_sql}(x, ...)
}
\arguments{
\item{x}{A lazy data frame backed by a database query.}

\item{...}{Ignored.}
}
\description{
\code{collapse()} forces computation of a lazy query by wrapping it in a subquery.
This is not generally needed, but can be useful if you need to work around
database/dbplyr limitations.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

db <- memdb_frame(a = c(3, 4, 1, 2), b = c(5, 1, 2, NA))
db |> filter(a <= 2) |> show_query()
db |> filter(a <= 2) |> collapse() |> show_query()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verb-set-ops.R
\name{intersect.tbl_lazy}
\alias{intersect.tbl_lazy}
\alias{union.tbl_lazy}
\alias{union_all.tbl_lazy}
\alias{setdiff.tbl_lazy}
\title{SQL set operations}
\usage{
\method{intersect}{tbl_lazy}(x, y, copy = "none", ..., all = FALSE)

\method{union}{tbl_lazy}(x, y, copy = "none", ..., all = FALSE)

\method{union_all}{tbl_lazy}(x, y, copy = "none", ...)

\method{setdiff}{tbl_lazy}(x, y, copy = "none", ..., all = FALSE)
}
\arguments{
\item{x, y}{A pair of lazy data frames backed by database queries.}

\item{copy}{If \code{x} and \code{y} are not from the same data source, \code{copy}
controls how \code{y} is copied into the same source as \code{x}. There are three
options:
\itemize{
\item \code{"none"}, the default, will error if \code{y} needs to be copied. This ensures
that you don't accidentally copy large datasets from R to the database.
\item \code{"temp-table"}: copies \code{y} into a temporary table in the same
database as \code{x}. \verb{*_join()} will automatically run \code{ANALYZE} on the
created table in the hope that this will make your queries as efficient
as possible by giving more data to the query planner.
\item \code{"inline"}: \code{y} will be inlined into the query using \code{\link[=copy_inline]{copy_inline()}}.
This should be faster for small datasets and doesn't require write
access.
}

\code{TRUE} (\code{"temp-table"}) and \code{FALSE} (\code{"none"}) are also accepted for
backward compatibility.}

\item{...}{Must be empty.}

\item{all}{If \code{TRUE}, includes all matches in output, not just unique rows.}
}
\description{
These are methods for the dplyr generics \code{dplyr::intersect()},
\code{dplyr::union()}, and \code{dplyr::setdiff()}. They are translated to
\code{INTERSECT}, \code{UNION}, and \code{EXCEPT} respectively.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backend-jdbc.R
\name{backend-jdbc}
\alias{backend-jdbc}
\title{JDBC backend}
\description{
This backend supports databases accessed via \code{JDBCConnection} created by
\code{RJDBC::JDBC()} and \code{\link[DBI:dbConnect]{DBI::dbConnect()}}. dbplyr automatically detects the
underlying database type from the JDBC connection class and uses the
appropriate SQL dialect.

The following databases are recognized via their JDBC connection class:
\itemize{
\item PostgreSQL (\verb{org.postgresql.*}): \code{\link[=dialect_postgres]{dialect_postgres()}}
\item MySQL (\verb{com.mysql.*}): \code{\link[=dialect_mysql]{dialect_mysql()}}
\item MariaDB (\verb{org.mariadb.*}): \code{\link[=dialect_mariadb]{dialect_mariadb()}}
\item SQLite (\verb{org.sqlite.*}): \code{\link[=dialect_sqlite]{dialect_sqlite()}}
\item Oracle (\verb{oracle.*}): \code{\link[=dialect_oracle]{dialect_oracle()}}
\item SQL Server (\verb{com.microsoft.sqlserver.*}): \code{\link[=dialect_mssql]{dialect_mssql()}}
\item IBM DB2 (\verb{com.ibm.db2.*}): \code{\link[=dialect_db2]{dialect_db2()}}
\item SAP HANA (\verb{com.sap.db.*}): \code{\link[=dialect_hana]{dialect_hana()}}
\item Teradata (\verb{com.teradata.*}): \code{\link[=dialect_teradata]{dialect_teradata()}}
\item Apache Hive (\verb{org.apache.hive.*}): \code{\link[=dialect_hive]{dialect_hive()}}
\item Apache Spark (\verb{org.apache.spark.*} or \verb{com.simba.spark.*}): \code{\link[=dialect_spark_sql]{dialect_spark_sql()}}
\item Snowflake (\verb{net.snowflake.*}): \code{\link[=dialect_snowflake]{dialect_snowflake()}}
\item Impala (\verb{com.cloudera.impala.*}): \code{\link[=dialect_impala]{dialect_impala()}}
\item Amazon Redshift (\verb{com.amazon.redshift.*}): \code{\link[=dialect_redshift]{dialect_redshift()}}
}

If your database is not recognized, dbplyr will fall back to a generic ODBC
dialect. Please file an issue if you'd like support for additional databases.

See \code{vignette("translation-function")} and \code{vignette("translation-verb")} for
details of overall translation technology.
}
\examples{
# JDBC connections require the RJDBC package and a JDBC driver JAR file.
# Once connected, dbplyr automatically detects the database type:
#
# library(RJDBC)
# library(dplyr, warn.conflicts = FALSE)
#
# drv <- JDBC("org.postgresql.Driver", "postgresql.jar")
# con <- dbConnect(drv, "jdbc:postgresql://localhost/mydb", "user", "password")
# tbl(con, "my_table") |> filter(x > 1)
}
\seealso{
\code{\link[=with_dialect]{with_dialect()}} to use a different dialect if dbplyr guesses
incorrectly, or a more specific translation is available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remote.R
\name{last_sql}
\alias{last_sql}
\title{Retrieve the last SQL query generated}
\usage{
last_sql()
}
\value{
A SQL string, or \code{NULL} if no query has been generated yet.
}
\description{
This is a helper function that retrieves the most recent SQL query generated
by dbplyr, which can be useful for debugging.
}
\examples{
library(dplyr, warn.conflicts = FALSE)

df <- lazy_frame(x = 1:3)
df |> filter(x > 1)

last_sql()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backward-compatibility.R
\name{src_sql}
\alias{src_sql}
\title{Create a "sql src" object}
\usage{
src_sql(subclass, con, ...)
}
\arguments{
\item{subclass}{Name of subclass.}

\item{con}{The connection object.}

\item{...}{Other arguments passed on to individual methods.}
}
\description{
Deprecated: please use directly use a \code{DBIConnection} object instead.
}
\keyword{internal}
